#include <BasicInfo.h>
#include <iostream.h>
#include <string.h>

/***********************************************************************
     Test driver and simple wrapper for a basic information module
     in the Linear System Analyzer.  This routine reads in a
     Harwell/Boeing matrix file, and then calls getinfo().

     The information is returned in a vector of length 100 called "info".
     See the getinfo.f documentation for an explanation of the fields.
     Also note that if the "job" parameter passed to getinfo is 6,
     then getinfo will write output to standard out that shows the
     information as it is computed.  If job = 0, getinfo does no 
     output.

     This has been written and tested primarily for square matrices.
     It is not guaranteed to be correct for rectangular matrices.
     It will return nonsense values for a particular field in info[]
     if it cannot compute the requested number.  Negative numbers
     mainly, but the code assumes some familarity with sparse matrices.

       Randall Bramley
       Department of Computer Science
       Indiana University, Bloomington
       bramley@cs.indiana.edu
       Sun May 18 04:17:38 EST 1997
       Updated for CCAT: Mon Nov  8 10:16:21 EST 1999
************************************************************************/

int BasicInfo(BasicInfoParms& parms, SparseLinearSystem* sls){
  
  /* ----------------------------- */
  /* Matrix data structure things. */
  /* ----------------------------- */
  double* A2       = NULL;    // Copy of nonzero values in A  
  int*    colinds2 = NULL;    // Copy of column pointers in A 
  int*    rowptrs2 = NULL;    // Copy of row indices in A     
  int     nrow     = sls->n;  // Number of rows in A          
  int     ncol     = sls->n;  // Number of cols in A          
  int     job      = 0;
  double  Info[101];          // Array for information returned.

  /* ---------------------------------- */
  /* Set storage for transposed system. */
  /* ---------------------------------- */

  colinds2 = new int[sls->nnz];
  rowptrs2 = new int[nrow+1];
  A2       = new double[sls->nnz];

  if (A2 == NULL || colinds2 == NULL || rowptrs2 == NULL){
    delete[] colinds2;
    delete[] rowptrs2;
    delete[] A2;
    cout << "Can't allocate memory for copy of matrix in BasicInfo\n";
    return -1;
  }
  
  /* --------------------------------------------------------------- */
  /* If parameter = 0 then let the getinfo function display the info */
  /* to stdout.  Otherwise receive info and provide output myself    */
  /* Job here corresponds to Fortran I/O unit number in getinfo_     */
  /* --------------------------------------------------------------- */
  
  job = (parms.level == 1) * 6;
  cout << "In BI, job = " << job << endl;
  cout << "In BI, nrow = " << nrow << endl;
  cout << "In BI, rowptrs[0] = " << sls->rowptrs[0] << endl;
  cout << "In BI, rowptrs[nrow] = " << sls->rowptrs[nrow] << endl;
  cout << "In BI, colinds[0] = " << sls->colinds[0] << endl;
  cout.flush();

  getinfo_(&nrow, &ncol, &job, 
           sls->values, 
           sls->colinds, 
           sls->rowptrs, 
	   A2, colinds2, rowptrs2,
           Info);
  
  /* ----------- */
  /* Free memory */
  /* ----------- */

  delete[] A2;
  delete[] colinds2;
  delete[] rowptrs2;

  /* ------------------ */
  /* generate html page */
  /* ------------------ */

  cout << "On entry to CreateHtml, life seems guuud\n";
  cout.flush();
  CreateHtml(sls->values,
		sls->colinds,
		sls->rowptrs,
		sls->n,
		sls->nnz,
		sls->title,
		Info);

  cout << "On exit from CreateHtml, life seems guuud\n";
  cout.flush();

  CreateGif(sls->values,
		sls->colinds,
		sls->rowptrs,
		sls->n,
		sls->nnz,
		Info);
  cout << "On exit from CreateGIF, life seems guuud\n";
  cout.flush();
  return 0;
}

/* ------------------------------------- */
/* Build structure matrix and output GIF */
/* ------------------------------------- */

int CreateGif (double* values,
	 int* colinds,
	 int* rowptrs,
         int n, int nnz,
	 double* Info) {
  
  int nrow, ncol;
  const int xpix = 500;
  const int ypix = 500;
  float largest = -999, smallest = 0;
  float rowscale, colscale;
  float *data = new float[xpix*ypix];
  if (data==NULL) {
    cout << "Can't allocate memory for matrix structure plot in BI\n";
    return 1;
  }

  int matcol, matrow, latcol, latrow;
  int i,j,k;
  int x,y;
  FILE *gifout;
  gdImagePtr matImage[5];
  int white[5];
  int black[5];
  int yellow[5];
  int red[5];
  int lband, uband, colnnz;
  int rownnz[ypix];
  char *fname[] = { "basicinforesults/matstruc.gif", "basicinforesults/matuband.gif", "basicinforesults/matlband.gif",
                    "basicinforesults/matzeros.gif" };
  lband = (int) Info[22];
  uband = (int) Info[23];

  nrow = ncol = n;
  // the following scaling will make a rectangular matrix look like a
  // square matrix
  rowscale = (float) nrow/ (float) ypix ;
  colscale = (float) ncol/ (float) xpix ;

  for(i = 0; i < ypix*xpix; i++)
     data[i] = 0.0;
  
  for(i=0;i<nrow;i++){
    for(k=rowptrs[i];k<rowptrs[i+1];k++){
      if (values[k-1] == 0.0) continue;
      matcol = colinds[k-1];
      matrow = i+1;
      latrow = ccat_min(Round(matrow/rowscale),ypix);
      latrow = ccat_max(latrow,1);
      latcol = ccat_min(Round(matcol/colscale),xpix);
      latcol = ccat_max(latcol,1);
      for (x=0;x<=latrow - Round((matrow-1)/rowscale);x++) {
	if (latrow-x < 1 || latrow-x > ypix) continue;
	for (y=0;y<=latcol - Round((matcol-1)/colscale);y++) {
	  if (latcol-y < 1 || latcol-y > xpix) continue;
	  int jj = (latrow-x-1)+(latcol-y-1)*ypix;
	  if ( jj<0 || jj>xpix*ypix) {
	    cout << "Can't generate structure plot in GD\n";
	    return 3;
	  }
	  data[(latrow-x-1)+(latcol-y-1)*ypix] += (float) 1.0;
	}
      }
      largest = ccat_max( (data[(latrow-1)+(latcol-1)*ypix]), largest);
      smallest = ccat_min( (data[(latrow-1)+(latcol-1)*ypix]), smallest); 
    }
  }
  
  /* ------------ */
  /* generate gif */
  /* ------------ */
  /*   enum { structure, upper, lower, zeros, sym }; */


  for (i=0;i<4;i++) {
	 matImage[i] = gdImageCreate(xpix,ypix);
	 white[i] = gdImageColorAllocate(matImage[i], 255,255,255);
	 black[i] = gdImageColorAllocate(matImage[i], 0,0,0);
	 yellow[i] = gdImageColorAllocate(matImage[i],255,255,0);
	 red[i] = gdImageColorAllocate(matImage[i],255,0,0);
  }

  for (j=0;j<ypix;j++) rownnz[j] = 0;
  
  for (j=0;j<xpix;j++) {
    colnnz = 0;
    for (i=0;i<ypix;i++) {
      if (data[i+ypix*j] > 0 ) {
	colnnz++;
	rownnz[i]++;
	for (k=0;k<4;k++) 
	  gdImageSetPixel(matImage[k], j,i,black[k]);
      }
      else {
	gdImageSetPixel(matImage[0], j,i,white[0]);
	gdImageSetPixel(matImage[3], j,i,white[3]);
	if ( i>j && i<=ceil(j+lband/rowscale))
	  gdImageSetPixel(matImage[2],j,i,yellow[2]);
	else
	  gdImageSetPixel(matImage[2],j,i,white[2]);
	if ( i<j && i>=ceil(j-uband/rowscale))
	  gdImageSetPixel(matImage[1],j,i,yellow[1]);
	else
	  gdImageSetPixel(matImage[1],j,i,white[1]);
      }
    }
    if ( colnnz == 0 )
      gdImageLine(matImage[3], j,0,j,ypix-1, red[3]);
  }
  
  for (k=0;k<ypix;k++) {
    if (rownnz[k] == 0)
      gdImageLine(matImage[3], 0,k,xpix-1,k,red[3]);
  }
  for (k=0;k<(int) Info[52];k++)
    gdImageLine(matImage[3], Round(Info[60+k]/rowscale),0,
		Round(Info[60+k]/rowscale),ypix-1,red[3]);
  for (k=0;k<(int) Info[51];k++)
    gdImageLine(matImage[3], 0,Round(Info[70+k]/rowscale),
		xpix-1,Round(Info[70+k]/rowscale),red[3]);
  
  for (i=0;i<4;i++) {
    if ((gifout = fopen(fname[i],"w")) == NULL) {
      printf("can't open file %s\n",fname[i]);
      continue;
    }
    gdImageGif(matImage[i], gifout);
    fclose(gifout);	 
    gdImageDestroy(matImage[i]);
  }
  
  delete[] data;
  return 0;
  
}

int Round(double value) {
  double absvalue = fabs(value);
  int sign = -1 * (value<0) + (value>0);
  
  if ( absvalue-floor(absvalue) >= .5 )
	 return ( (int) value ) + sign;
  else
	 return ( (int) value );
}


int CreateHtml(double* values, int* colinds,
               int* rowptrs, int n, int nnz,
               char* title,
               double* Info) {
  FILE* dout;
  float csrbytes, bandbytes,skybytes,densebytes;
  csrbytes = 8*Info[2]+ 4*Info[2]+8*(Info[1]+1);
  bandbytes = 8*(Info[22]+2*Info[23]+1)*Info[1];
  skybytes = 8*Info[26];
  densebytes = 8*Info[1]*Info[1];

  if ((dout = fopen("basicinforesults/mat.html","w")) == NULL) {
    cout << "Can't open HTML file mat.html\n";
    return -1;
  }
  
  fprintf(dout,"<HTML>\n<HEAD>\n<TITLE>Basic Matrix Information</TITLE>\n");
  fprintf(dout,"<SCRIPT LANGUAGE=\"JavaScript\">\n<!--\n");
  fprintf(dout,"function openDSwin() {\n\tvar dswin = null;\n\tdswin = ");
  fprintf(dout,"window.open(\"\",\"DSWindow\",'toolbar=0,location=0,directories=0,status=0,menubar=0,scrollbars=0,resizable=0,width=530,height=450');\n");
  fprintf(dout,"\tself.location = \"#SKYLINE\";\n\tvar ds = dswin.document;\n\t");
  fprintf(dout,"ds.write('<H1 ALIGN=CENTER>Storage Cost Comparisons</H1>');\n\t");
  fprintf(dout,"ds.write('<H4>Representing the matrix in different data structures requires different amounts of storage base on the structure of the matrix.</H4>');");
  fprintf(dout,"ds.write('<TABLE BORDER=3 CELLPADDING=10><TR><TH COLSPAN=2></TH><TH COLSPAN=4>Ratio to</TH></TR>');\n\t");
  fprintf(dout,"ds.write('<TH></TH><TH>Bytes</TH><TH>CSR</TH><TH>Banded</TH><TH>Skyline</TH><TH>Dense</TH></TR>');\n\t");
  fprintf(dout,"ds.write('<TH>CSR</TH><TD>%s</TD>",FormatScientific(csrbytes,12,3));
  fprintf(dout,"<TD>%12.3f</TD><TD>%12.3f</TD>",1.0,csrbytes/bandbytes);
  fprintf(dout,"<TD>%12.3f</TD><TD>%12.3f</TD></TR>');\n\t",csrbytes/skybytes,csrbytes/densebytes);
  fprintf(dout,"ds.write('<TH>Banded</TH><TD>%s</TD>",FormatScientific(bandbytes,12,3));
  fprintf(dout,"<TD>%12.3f</TD><TD>%12.3f</TD>",bandbytes/csrbytes,1.0);
  fprintf(dout,"<TD>%12.3f</TD><TD>%12.3f</TD></TR>');\n\t",bandbytes/skybytes,bandbytes/densebytes);
  fprintf(dout,"ds.write('<TH>Skyline</TH><TD>%s</TD>",FormatScientific(skybytes,12,3));
  fprintf(dout,"<TD>%12.3f</TD><TD>%12.3f</TD>",skybytes/csrbytes,skybytes/bandbytes);
  fprintf(dout,"<TD>%12.3f</TD><TD>%12.3f</TD></TR>');\n\t",1.0,skybytes/densebytes);
  fprintf(dout,"ds.write('<TH>Dense</TH><TD>%s</TD>",FormatScientific(densebytes,12,3));
  fprintf(dout,"<TD>%12.3f</TD><TD>%12.3f</TD>",densebytes/csrbytes,densebytes/bandbytes);
  fprintf(dout,"<TD>%12.3f</TD><TD>%12.3f</TD></TR>');\n\t",densebytes/skybytes,1.0);
  fprintf(dout,"ds.write('</TABLE>');\n\tds.close();\n\treturn true;\n}\n// -->\n</SCRIPT>\n");

  fprintf(dout,"</HEAD>\n");
  fprintf(dout,"<BODY>\n<A NAME=\"TOP\"></A>\n<BR>\n");
  fprintf(dout,"<H1 ALIGN=CENTER>LINEAR SYSTEM ANALYZER</H1>");
  fprintf(dout,"<H2 ALIGN=CENTER>Basic Matrix Information</H2>\n<HR SIZE=4>\n");

  fprintf(dout,"<H3 ALIGN=CENTER>%s</H3>\n<BR>\n",title);

 
  fprintf(dout,"<A NAME=\"BASIC\"></A>\n");
  fprintf(dout,"<TABLE ALIGN=RIGHT BORDER=3 CELLPADDING=10>\n");
  fprintf(dout,"<TR>\n<TH COLSPAN=2>BASIC INFORMATION</TH>\n</TR>\n");
  fprintf(dout,"<TR>\n");
  fprintf(dout,"<TD><B>Dimension of Matrix</B></TD>");
  fprintf(dout,"<TD ALIGN=RIGHT><B>%s</B></TD>",FormatInteger((int)Info[1],14,0));
  fprintf(dout,"</TR>\n<TR>\n");
  fprintf(dout,"<TD><B>Number of explicitly stored elements</B></TD>");
  fprintf(dout,"<TD ALIGN=RIGHT><B>%s</B></TD>",FormatInteger((int)Info[2],14,0));
  fprintf(dout,"</TR>\n<TR>\n");
  fprintf(dout,"<TD><B>Max nonzeros in a column</B></TD>");
  fprintf(dout,"<TD ALIGN=RIGHT><B>%s</B></TD>",FormatInteger((int)Info[8],14,0));
  fprintf(dout,"</TR>\n<TR>\n");
  fprintf(dout,"<TD><B>Min nonzeros in a column</B></TD>");
  fprintf(dout,"<TD ALIGN=RIGHT><B>%s</B></TD>",FormatInteger((int)Info[9],14,0));
  fprintf(dout,"</TR>\n<TR>\n");
  fprintf(dout,"<TD><B>Max nonzeros in a row</B></TD>");
  fprintf(dout,"<TD ALIGN=RIGHT><B>%s</B></TD>",FormatInteger((int)Info[10],14,0));
  fprintf(dout,"</TR>\n<TR>\n");
  fprintf(dout,"<TD><B>Min nonzeros in a row</B></TD>");
  fprintf(dout,"<TD ALIGN=RIGHT><B>%s</B></TD>",FormatInteger((int)Info[11],14,0));
  fprintf(dout,"</TR>\n<TR>\n");
  fprintf(dout,"<TD><B>Average number of nonzero elements/Column</B></TD>");
  fprintf(dout,"<TD ALIGN=RIGHT><B>%s</B></TD>",FormatScientific(Info[3],14,4));
  fprintf(dout,"</TR>\n<TR>\n");
  fprintf(dout,"<TD><B>Standard deviation for above average</B></TD>");
  fprintf(dout,"<TD ALIGN=RIGHT><B>%s</B></TD>",FormatScientific(Info[4],14,4));
  fprintf(dout,"</TR>\n<TR>\n");
  fprintf(dout,"<TD><B>Maximum element in A in absolute value</B></TD>");
  fprintf(dout,"<TD ALIGN=RIGHT><B>%s</B></TD>",FormatScientific(Info[19],14,4));
  fprintf(dout,"</TR>\n</TABLE>\n<BR>\n");
  
  fprintf(dout,"<A HREF=\"matstruc.gif\">\n");
  fprintf(dout,"<IMG ALIGN=BOTTOM BORDER=4 HEIGHT=150 WIDTH=150 ALT=\"[Matrix Structure Plot]\" SRC=\"matstruc.gif\"></A>");
  fprintf(dout,"<H2 ALIGN=CENTER>Index</H2>\n");
  fprintf(dout,"<OL>\n<LI><A HREF=\"#BASIC\">Basic Structure</A>\n");
  fprintf(dout,"<LI><A HREF=\"#SYM\">Matrix Symmetry</A>\n");
  fprintf(dout,"<LI><A HREF=\"#DIAG\">Diagonal Dominance</A>\n");
  fprintf(dout,"<LI><A HREF=\"#SING\">Potential Singularity</A>\n");
  fprintf(dout,"<LI><A HREF=\"#SKYLINE\" onClick=\"openDSwin()\">Data Structure</A>\n<OL>\n");
  fprintf(dout,"<LI><A HREF=\"#SKYLINE\">Skyline</A>\n");
  fprintf(dout,"<LI><A HREF=\"#BANDED\">Banded</A>\n");
  fprintf(dout,"<LI><A HREF=\"#DENSE\">Dense</A>\n");
  fprintf(dout,"</OL>\n</OL>\n<BR><BR><BR><BR>\n");




  fprintf(dout,"<CENTER><A HREF=\"#TOP\">[Previous]</A><A HREF=\"#TOP\">[Index]</A><A HREF=\"#SYM\">[Next]</A></CENTER>\n");
  fprintf(dout,"<BR>\n<A NAME=\"SYM\"></A><HR SIZE=4><BR>\n");

  fprintf(dout,"<CENTER><TABLE ALIGN=CENTER BORDER=3 CELLPADDING=10>\n");
  fprintf(dout,"<TR>\n<TH COLSPAN=2>SYMMETRY INFORMATION</TH>\n</TR>\n");
  fprintf(dout,"<TR>\n");  
  fprintf(dout,"<TD><B>Nonzero elements in strict lower part</B></TD>");
  fprintf(dout,"<TD ALIGN=RIGHT><B>%s</B></TD>", FormatInteger((int)Info[5],14,0));
  fprintf(dout,"</TR>\n<TR>\n");
  fprintf(dout,"<TD><B>Nonzero elements in strict upper part</B></TD>");
  fprintf(dout,"<TD ALIGN=RIGHT><B>%s</B></TD>",FormatInteger((int)Info[6],14,0));
  fprintf(dout,"</TR>\n<TR>\n");
  fprintf(dout,"<TD><B>Number of entries matched in structural symmetry</B></TD>");
  fprintf(dout,"<TD ALIGN=RIGHT><B>%s</B></TD>",FormatInteger((int)Info[12],14,0));
  fprintf(dout,"</TR>\n<TR>\n");
  fprintf(dout,"<TD><B>Structural symmetry coefficient (symmetry=1)</B></TD>");
  fprintf(dout,"<TD ALIGN=RIGHT><B>%s</B></TD>",FormatScientific(Info[13],14,4));
  fprintf(dout,"</TR>\n<TR>\n");
  fprintf(dout,"<TD><B>Frobenius norm of A</B></TD>");
  fprintf(dout,"<TD ALIGN=RIGHT><B>%s</B></TD>",FormatScientific(Info[16],14,4));
  fprintf(dout,"</TR>\n<TR>\n");
  fprintf(dout,"<TD><B>Frobenius norm of symmetric part: &frac12;(A+A')</B></TD>");
  fprintf(dout,"<TD ALIGN=RIGHT><B>%s</B></TD>",FormatScientific(Info[17],14,4));
  fprintf(dout,"</TR>\n<TR>\n");
  fprintf(dout,"<TD><B>Frobenius norm of nonsymmetric part: &frac12;(A-A')</B></TD>");
  fprintf(dout,"<TD ALIGN=RIGHT><B>%s</B></TD>",FormatScientific(Info[18],14,4));
  fprintf(dout,"</TR>\n</TABLE></CENTER>\n<BR>\n");
  fprintf(dout,"<CENTER><A HREF=\"#BASIC\">[Previous]</A><A HREF=\"#TOP\">[Index]</A><A HREF=\"#DIAG\">[Next]</A></CENTER>\n");
  fprintf(dout,"<BR>\n<A NAME=\"DIAG\"></A><HR SIZE=4><BR>\n");
  
  fprintf(dout,"<CENTER><TABLE ALIGN=CENTER BORDER=3 CELLPADDING=10>\n");
  fprintf(dout,"<TR>\n<TH COLSPAN=2>DIAGONAL DOMINANCE INFORMATION</TH>\n</TR>\n");
  fprintf(dout,"<TR>\n"); 
  fprintf(dout,"<TD><B>Fraction of weakly diagonally dominant rows</B></TD>");
  fprintf(dout,"<TD ALIGN=RIGHT><B>%s</B></TD>",FormatScientific(Info[20],14,4));
  fprintf(dout,"</TR>\n<TR>\n");
  fprintf(dout,"<TD><B>Fraction of weakly diagonally dominant columns</B></TD>");
  fprintf(dout,"<TD ALIGN=RIGHT><B>%s</B></TD>",FormatScientific(Info[21],14,4));
  fprintf(dout,"</TR>\n</TABLE></CENTER>\n<BR>\n");
  fprintf(dout,"<CENTER><A HREF=\"#SYM\">[Previous]</A><A HREF=\"#TOP\">[Index]</A><A HREF=\"#SING\">[Next]</A></CENTER>\n");
  fprintf(dout,"<BR>\n<A NAME=\"SING\"></A><HR SIZE=4><BR>\n");
  
  fprintf(dout,"<CENTER><TABLE ALIGN=CENTER BORDER=3 CELLPADDING=10>\n");
  fprintf(dout,"<TR>\n<TH COLSPAN=2>POTENTIAL SINGULARITY INFORMATION</TH>\n</TR>\n");
  fprintf(dout,"<TR>\n");  
  fprintf(dout,"<TD><B>Number of zeros on main diagonal</B></TD>");
  fprintf(dout,"<TD ALIGN=RIGHT><B>%s</B></TD>",FormatInteger((int)Info[7],14,0));
  fprintf(dout,"</TR>\n<TR>\n");
  fprintf(dout,"<TD><B>Number of columns consisting of all zeros</B></TD>");
  fprintf(dout,"<TD ALIGN=RIGHT><B>%s</B></TD>",FormatInteger((int)Info[52],14,0));
  fprintf(dout,"</TR>\n<TR>\n");
  fprintf(dout,"<TD><B>Number of rows consisting of all zeros</B></TD>");
  fprintf(dout,"<TD ALIGN=RIGHT><B>%s</B></TD>",FormatInteger((int)Info[51],14,0));
  fprintf(dout,"</TR>\n</TABLE></CENTER>\n<BR>\n");
  fprintf(dout,"<CENTER><A HREF=\"#DIAG\">[Previous]</A><A HREF=\"#TOP\">[Index]</A><A HREF=\"#SKYLINE\">[Next]</A></CENTER>\n");
  fprintf(dout,"<BR>\n<A NAME=\"SKYLINE\"></A><HR SIZE=4><BR>\n");

  fprintf(dout,"<CENTER><TABLE ALIGN=CENTER BORDER=3 CELLPADDING=10>\n");
  fprintf(dout,"<TR>\n<TH COLSPAN=2>SKYLINE DATA STRUCTURE INFORMATION</TH>\n</TR>\n");
  fprintf(dout,"<TR>\n");  
  fprintf(dout,"<TD><B>Number of nonzeros in skyline storage</B></TD>");
  fprintf(dout,"<TD ALIGN=RIGHT><B>%s</B></TD>",FormatInteger((int)Info[26],14,0));
  fprintf(dout,"</TR>\n</TABLE></CENTER>\n<BR>\n");
  fprintf(dout,"<CENTER><A HREF=\"#SING\">[Previous]</A><A HREF=\"#TOP\">[Index]</A><A HREF=\"#BANDED\">[Next]</A></CENTER>\n");
  fprintf(dout,"<BR>\n<A NAME=\"BANDED\"></A><HR SIZE=4><BR>\n");

  fprintf(dout,"<CENTER><TABLE ALIGN=CENTER BORDER=3 CELLPADDING=10>\n");
  fprintf(dout,"<TR>\n<TH COLSPAN=2>BANDED STRUCTURE INFORMATION</TH>\n</TR>\n");
  fprintf(dout,"<TR>\n");  
  fprintf(dout,"<TD><B>Average distance of a(i,j)  from diagonal</B></TD>");
  fprintf(dout,"<TD ALIGN=RIGHT><B>%s</B></TD>",FormatScientific(Info[14],14,6));
  fprintf(dout,"</TR>\n<TR>\n");
  fprintf(dout,"<TD><B>Standard deviation for above average</B></TD>");
  fprintf(dout,"<TD ALIGN=RIGHT><B>%s</B></TD>",FormatScientific(Info[15],14,4));
  fprintf(dout,"</TR>\n<TR>\n");
  fprintf(dout,"<TD><B>Lower bandwidth  (max: i-j, a(i,j) .ne. 0)</B></TD>");
  fprintf(dout,"<TD ALIGN=RIGHT><B>%s</B></TD>",FormatInteger((int)Info[22],14,0));
  fprintf(dout,"</TR>\n<TR>\n");
  fprintf(dout,"<TD><B>Upper bandwidth  (max: j-i, a(i,j) .ne. 0)</B></TD>");
  fprintf(dout,"<TD ALIGN=RIGHT><B>%s</B></TD>",FormatInteger((int)Info[23],14,0));
  fprintf(dout,"</TR>\n<TR>\n");
  fprintf(dout,"<TD><B>Maximum Bandwidth</B></TD>");
  fprintf(dout,"<TD ALIGN=RIGHT><B>%s</B></TD>",FormatInteger((int)Info[24],14,0));
  fprintf(dout,"</TR>\n<TR>\n");
  fprintf(dout,"<TD><B>Average Bandwidth</B></TD>");
  fprintf(dout,"<TD ALIGN=RIGHT><B>%s</B></TD>",FormatScientific(Info[25],14,4));
  fprintf(dout,"</TR>\n<TR>\n");
  fprintf(dout,"<TD><B>90 percent of matrix is in the band of width</B></TD>");
  fprintf(dout,"<TD ALIGN=RIGHT><B>%s</B></TD>",FormatInteger((int)Info[27],14,0));
  fprintf(dout,"</TR>\n<TR>\n");
  fprintf(dout,"<TD><B>80 percent of matrix is in the band of width</B></TD>");
  fprintf(dout,"<TD ALIGN=RIGHT><B>%s</B></TD>",FormatInteger((int)Info[28],14,0));
  fprintf(dout,"</TR><TR>\n");
  fprintf(dout,"<TD><B>The total number of nonvoid diagonals is</B></TD>");
  fprintf(dout,"<TD ALIGN=RIGHT><B>%s</B></TD>",FormatInteger((int)Info[29],14,0));
  fprintf(dout,"</TR>\n</TABLE></CENTER>\n<BR>\n");
  fprintf(dout,"<CENTER><A HREF=\"#SKYLINE\">[Previous]</A><A HREF=\"#TOP\">[Index]</A><A HREF=\"#DENSE\">[Next]</A></CENTER>\n");
  fprintf(dout,"<BR>\n<A NAME=\"DENSE\"></A><HR SIZE=4><BR>\n");

  fprintf(dout,"<CENTER><TABLE ALIGN=CENTER BORDER=3 CELLPADDING=10>\n");
  fprintf(dout,"<TR>\n<TH COLSPAN=2>DENSE DATA STRUCTURE INFORMATION</TH>\n</TR>\n");
  fprintf(dout,"<TR>\n");
  fprintf(dout,"<TD><B>Number of double words to store matrix as dense is</B></TD>");
  fprintf(dout,"<TD ALIGN=RIGHT><B>%s</B></TD>",FormatScientific(Info[1]*Info[1],14,6));
  fprintf(dout,"</TR>\n");
  fprintf(dout,"</TABLE></CENTER>\n<BR>\n");
  fprintf(dout,"<CENTER><A HREF=\"#BANDED\">[Previous]</A><A HREF=\"#TOP\">[Index]</A><A HREF=\"#DENSE\">[Next]</A></CENTER>\n");
  fprintf(dout,"</BODY>\n</HTML>\n");

  fclose(dout);

  return 0;
}


/* ------------------------------------------------ */
/* Format integers for insertion into html web page */
/* ------------------------------------------------ */

char* FormatInteger( int value, int field, int base) {

  int len;
  int i, digit, commacount=0, roundup=0;
  char units[] = { ' ', 'x', 'x', 'k', 'x', 'x', 'M', 'x', 'x', 'G' };
  if ( field>99)
	 return NULL;
  len = field;
  CharBuffer[len] = '\0';
  if ( base > 0 ) {
    for (i=0;i<base;i++) {
      if (value %10 > 0)
	roundup = 1;
      value /= 10;
    }
    value += roundup;
    CharBuffer[--len] = units[base];
  }
  
  len--;
  i = len;
  if (value==0)
    CharBuffer[i--] = '0';
  while (i>=0) {
    if ( value > 0 ) {
      digit = value % 10;
      value /= 10;
      CharBuffer[i] = '0' + digit;
      if ( (len-i-commacount+1) % 3 == 0 && value > 0 ) { 
	CharBuffer[--i] = ',';
	commacount++;
      }
    }
    else
      CharBuffer[i] = ' ';
    i--;	 
  }
  return CharBuffer;
}

/* ------------------------------------------------------------ */
/* Format doubles (and floats) for insertion into html web page */
/* ------------------------------------------------------------ */

char* FormatScientific( double value, int field, int precision) {

  int len, i;
  char format[80];
  char text[80];
  char expstr[14];
  int exponent;
  int sindex=0;
  if (field>99)
	 return NULL;
  len = field;
  CharBuffer[len--] = '\0';
  sprintf(format,"%%.%de",precision);
  sprintf(text,format,value);
  while ( text[++sindex] != 'e')
	 ;
  text[sindex] = '\0';
  exponent = atoi(&text[sindex+1]);
  if ( exponent != 0 ) {
	 sprintf(expstr,"%d",exponent);
	 sindex += strlen(expstr) + 5;
	 sprintf(format," &times; 10<SUP>%s</SUP>",expstr);
  }
  else
	 format[0] = '\0';
  for (i=0;i<field-sindex;i++)
	 CharBuffer[i] = ' ';
  strcpy(&CharBuffer[i],text);
  strcat(CharBuffer,format);
  return CharBuffer;
}
