/*
// --------------------------------------------------------------------------
// The domain is the unit square, with M+1 x N+1 mesh lines.
// The mesh lines corresponding to
//  x = 0
//  x = M+1
//  y = 0
//  y = N+1
// are Dirichlet boundary conditions, so
// the total number of unknowns is (M-1)*(N-1).
//
// The mesh is assumed to be M x N, and the lines corresponding to
// x = 0, x = M, y = 0, and y = N are boundary conditions from
// the function BC().  So for M = 5 and N = 4, the mesh looks like
//
//             B   B   B   B   B   B
//             B   8   9   10  11  B
//             B   4   5   6   7   B
//             B   0   1   2   3   B
//             B   B   B   B   B   B
//
// where B = boundary condition node and the numbered nodes are the
// indexing of the unknowns.  Furthermore, the boundary condition vector
// boundary[] returns the values in the boundary nodes, with the indexing
//
//             6    7    8    9    10   11
//             14   X    X    X    X    17
//             13   X    X    X    X    16
//             12   X    X    X    X    15
//             0    1    2    3    4    5
// --------------------------------------------------------------------------
*/

#include <Geometry.h>
#include <assert.h>
#include <stdlib.h>
#include "../../../../framework/util/macros.h"
#include <hpcxx_rts.h>

Geometry::Geometry(){

    this->M = this->N = this->size = 0;
    this->x = 0;
    this->y = 0;
    this->left = 0;   
    this->right = 0;  
    this->top = 0;    
    this->bottom = 0; 
};

Geometry::Geometry(const Geometry &g) {
    copy(g);
}

void
Geometry::operator=(const Geometry &g) {
    free();
    copy(g);
}

Geometry::~Geometry(){

    if (size > 0) {
        assert(size == (M + 1)*(N + 1));
    }
    free();
}

void
Geometry::copy(const Geometry &g) {

	int i;

    M = g.M;
    N = g.N;
    size = g.size;

    x = new double[M + 1];
    y = new double[N + 1];

    left = new double[N - 1];
    right = new double[N - 1];
    top = new double[M + 1];
    bottom = new double[M + 1];

    for (i = 0; i < M + 1; i++) {
        x[i] = g.x[i];
    }
    for (i = 0; i < M + 1; i++) {
        y[i] = g.y[i];
    }

    for (i = 0; i < N - 1; i++) {
        left[i] = g.left[i];
    }
    for (i = 0; i < N - 1; i++) {
        right[i] = g.right[i];
    }

    for (i = 0; i < M + 1; i++) {
        top[i] = g.top[i];
    }
    for (i = 0; i < M + 1; i++) {
        bottom[i] = g.bottom[i];
    }
}

void
Geometry::free() {

    DELETE(x);
    DELETE(y);

    DELETE(left);
    DELETE(right);
    DELETE(top);
    DELETE(bottom);
}

void
hpcxx_pack(HPCxx_Buffer& buffer, Geometry *g, int count) {

    for(int i = 0; i < count; i++) {

        assert(g->size == (g->M + 1)*(g->N + 1));

        hpcxx_pack(buffer, &g->M, 1);
        hpcxx_pack(buffer, &g->N, 1);
        hpcxx_pack(buffer, &g->size, 1);

        hpcxx_pack(buffer, g->x, g->M + 1);
        hpcxx_pack(buffer, g->y, g->N + 1);

        hpcxx_pack(buffer, g->left, g->N - 1);
        hpcxx_pack(buffer, g->right, g->N - 1);

        hpcxx_pack(buffer, g->top, g->M + 1);
        hpcxx_pack(buffer, g->bottom, g->M + 1);

        g++;
    }
}

void
hpcxx_unpack(HPCxx_Buffer& buffer, Geometry *g, int count) {

    for(int i = 0; i < count; i++) {

        hpcxx_unpack(buffer, &g->M, 1);
        hpcxx_unpack(buffer, &g->N, 1);
        hpcxx_unpack(buffer, &g->size, 1);
        assert(g->size == (g->M + 1)*(g->N + 1));

        g->x = new double[g->M + 1];
        g->y = new double[g->N + 1];

        g->left = new double[g->N - 1];
        g->right = new double[g->N - 1];
        g->top = new double[g->M + 1];
        g->bottom = new double[g->M + 1];

        hpcxx_unpack(buffer, g->x, g->M + 1);
        hpcxx_unpack(buffer, g->y, g->N + 1);

        hpcxx_unpack(buffer, g->left, g->N - 1);
        hpcxx_unpack(buffer, g->right, g->N - 1);

        hpcxx_unpack(buffer, g->top, g->M + 1);
        hpcxx_unpack(buffer, g->bottom, g->M + 1);

        g++;
    }
}
