/*
//  --------------------------------------------------------------------------------
//  Class Mungings holds the data from a sequence of reorderings and scalings done
//  by LSA modules.  This is needed so that after such a sequence and a solve of
//  the resulting mangled system, we can retrieve the solution vector to the
//  original system.  A sequence of reorderings, of course, can simply be
//  represented as the composition of those reorderings, requiring just a single
//  integer vector to store.  Similarly a sequence of scalings can be combined into
//  a single double array.  The complexity in Mungings comes because it is possible
//  to feed the linear system through an arbitrary, possibly alternating, sequence
//  of reorderings and scalings.  Each time a new munging is added, we may have to
//  allocate another vector.
//  
//  Mungings does not need to, but does keep track of the total numbers of
//  scalings (nscalings) and of reorderings (nreorderings).  
//  
//  Two-dimensional arrays called reorderings and scalings are used to store
//  each of the munging flavors.  Each reordering or scaling vector is allocated
//  dynamically as needed, but the pointers to pointers start out with 8 each.
//  It is unlikely anyone will require that many of either, but this saves us from
//  costly reallocation on every added munging vector.  The integers
//  allocated_reorderings and allocated_scalings indicate how many pointers have
//  been allocated - if they have to be increased, they are done so in increments
//  of 8 also.  
//  
//  To keep track of the sequencing of the operations (which must be undone in
//  precisely reverse order for retrieving the true solution vector) an integer
//  lastoper is used.  It indexes into the munging vector arrays as:
//      lastoper < 0:  means scaling was the last munging done, and
//                      scaling[-lastoper] was the last one done.
//      lastoper > 0:  means reordering was the last munging done, and
//                      reordering[-lastoper] was the last one done.
//      lastoper = 0:  means no mungings have been done.
//  Because we compose munging operations whenever possible, this suffices to
//  indicate what sequence is needed; scalings and reorderings have to alternate in
//  the unmunging operation.
//  
//  Randall Bramley
//  Department of Computer Science
//  Indiana University, Bloomington
//  bramley@cs.indiana.edu                             
//  Mon Oct 25 15:11:55 EST 1999
//  --------------------------------------------------------------------------------
*/

#include "Mungings.h"

#include <iostream.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>

Mungings::Mungings() {

      n            = 0;     // Length of scale/reorder vecs
      nreorderings = 0;     // Total number of reorderings
      nscalings    = 0;     // Total number of scalings
      nreordering_vectors = 0;  // Number of reordering vectors
      nscaling_vectors    = 0;  // Number of scaling vectors
      allocated_reorderings = 8;    // Number of allocated pointers
      allocated_scalings    = 8;    // Number of allocated pointers
      lastoper = 0;
      reorderings = new int*[8];
      scalings    = new double*[8];
}

Mungings::Mungings(const Mungings &m){
    assign(m);
}

void
Mungings::operator=(const Mungings &m) {
    free();
    assign(m);
}

Mungings::~Mungings(){
    free();
}

void
Mungings::free() {
    
    int k;

    /* XXX Memory leak.
    for (k = 0; k < nreordering_vectors; k++)
        delete[] reorderings[k];

    for (k = 0; k < nscaling_vectors; k++)
        delete[] scalings[k];

    delete [] reorderings;
    delete [] scalings;
    */
}

void
Mungings::assign(const Mungings &m) {

    int k;

    n                     = m.n;
    nreorderings          = m.nreorderings;
    nscalings             = m.nscalings;
    nreordering_vectors   = m.nreordering_vectors;
    nscaling_vectors      = m.nscaling_vectors;
    allocated_reorderings = m.allocated_reorderings;
    allocated_scalings    = m.allocated_scalings;
    lastoper              = m.lastoper;

    reorderings = new int*[allocated_reorderings];
    if (reorderings == 0) {
        cout << "Outta memory inside Mungings" << endl;
        exit(-1);
    }
    for (k = 0; k < nreorderings; k++) {
        reorderings[k] = new int[n];
        assert(reorderings[k] != NULL);
        for (int i = 0; i < n; i++) {
            reorderings[k][i] = m.reorderings[k][i];
        }
    }

    scalings = new double*[allocated_scalings];
    if (scalings == 0) {
        cout << "Outta memory inside Mungings" << endl;
        exit(-1);
    }
    for (k = 0; k < nscalings; k++) {
        scalings[k] = new double[n];
        assert(scalings[k] != NULL);
        for (int i = 0; i < n; i++) {
            scalings[k][i] = m.scalings[k][i];
        }
    }
}


// -----------------------
// Add in a new reordering
// -----------------------
int Mungings::AddReordering(int *newreordering){

    // First check to see if last operation done was a 
    // reordering.   If it was, we just need to compose
    // the new reordering with the old one in reverse order.
    nreorderings++;
    if (lastoper > 0) {
        ivperm_(&n, reorderings[nreordering_vectors-1], newreordering);
        return 0;
    }

    // Otherwise, need to add the new reordering vector to the list
    // Worse case: need more pointers to reordering vectors than
    // have currently allocated; need to do a realloc.
    if (nreordering_vectors > allocated_reorderings) {

        // Allocate new pointer to pointers, up to 8 additional vecs
        int **temp = new int*[allocated_reorderings + 8];

        if (temp == 0) {
            cout << "Outta memory in Mungings::AddReordering" << endl;
            exit(-1);
        }

        for (int k = 0; k < allocated_reorderings; k++)
            temp[k] = reorderings[k];

        delete[] reorderings;
        reorderings = temp;
        allocated_reorderings += 8;
    }

    // Now copy over pointer from new reordering vector
    nreordering_vectors++;
    reorderings[nreordering_vectors-1] = newreordering;
    return 0;
}

// -----------------------
// Add in a new scaling
// -----------------------
int Mungings::AddScaling(double *newscaling){
  // First check to see if last operation done was a 
  // scaling.   If it was, we just need to compose
  // the new scaling with the previous one.
   nscalings++;
  if (lastoper < 0) {
     double alpha = 1.0;
     int one = 1;
     daxpy_(&n, &alpha, newscaling, &one, scalings[nscaling_vectors-1], &one); 
     return 0;
  }
  // Otherwise, have to add the new scaling vector to the list.
   // Worse case: need more pointers to scaling vectors than
   // have currently allocated; need to do a realloc.
   if (nscaling_vectors >= allocated_scalings) {
     // Allocate new pointer to pointers, up to 8 additional vecs
     double **temp = new double*[allocated_scalings + 8];
     if (temp == 0) {
       cout << "Outta memory in Mungings::AddScaling" << endl;
       exit(-1);
     }
     for (int k = 0; k < allocated_scalings; k++)
       temp[k] = scalings[k];
     delete[] scalings;
     scalings = temp;
     allocated_scalings += 8;
   }
   // Now copy over pointer from new scaling vector
   nscaling_vectors++;
   scalings[nscaling_vectors-1] = newscaling;
   return 0;
}

// How many mungings done? Sum of reorderings and scalings
int Mungings::getNumMungings(){return(nreorderings + nscalings);}

// How many reorderings done?
int Mungings::getNumReorderings(){return(nreorderings);}

// How many scalings done?
int Mungings::getNumScalings(){return(nscalings);}

// Storage currently used.
int Mungings::getBytes(){
 return (8*n*nreordering_vectors
       + 4*n*nscaling_vectors
       + 4*(7+nscaling_vectors+nreordering_vectors));}

// -----------------------------------------------
// Retrieve the solution by unapplying all of the
// scalings/reorderings in reverse order.  Do this
// in-place with the vector x.
// -----------------------------------------------
int Mungings::UnMunge(double *x) {
  int k, i, ks, kr;
  if (lastoper == 0) return 0;
  kr = nreordering_vectors;
  ks = nscaling_vectors;

  // Case 1: last operation was scaling
  if (lastoper < 0) {
    for (k = ks+kr; k > 0; k--) {
      if (ks > 0) {
    unscalx_(&n, x, scalings[ks-1]);
    ks--; k--;
      }
      if (kr > 0) {
    dvperm_(&n, x, reorderings[kr-1]);
    kr--; k--;
      }
    }
  }

  // Case 2: last operation was reordering
  if (lastoper > 0) {
    for (k = ks+kr; k > 0; k--) {
      if (kr > 0) {
    dvperm_(&n, x, reorderings[kr-1]);
    kr--; k--;
      }
      if (ks > 0) {
    unscalx_(&n, x, scalings[ks-1]);
    ks--; k--;
      }
    }
  }
  return 0;

}
void hpcxx_pack(HPCxx_Buffer& buffer, Mungings* p, int count){ 

    for (int i=0; i<count; i++) { 

        hpcxx_pack(buffer, &(p->n), 1); // no of rows

        hpcxx_pack(buffer, &(p->nreorderings), 1);
        hpcxx_pack(buffer, &(p->nreordering_vectors), 1); // no of ordeings 
        for (int j = 0; j < p->nreordering_vectors; j++) {
            hpcxx_pack(buffer, p->reorderings[j], p->n);
        }

        hpcxx_pack(buffer, &(p->nscalings), 1);
        hpcxx_pack(buffer, &(p->nscaling_vectors), 1); //no of cols
        for (int j = 0; j < p->nscaling_vectors; j++) {
            hpcxx_pack(buffer, p->scalings[j], p->n);
        }

        hpcxx_pack(buffer, &(p->lastoper), 1);

        p++;
    }
}

void hpcxx_unpack(HPCxx_Buffer& buffer, Mungings* p, int count){

    for (int i=0; i<count; i++) {

        p->free();

        hpcxx_unpack(buffer, &(p->n), 1);

        hpcxx_unpack(buffer, &(p->nreorderings), 1);
        hpcxx_unpack(buffer, &(p->nreordering_vectors), 1);
        p->reorderings = new int*[p->nreordering_vectors];
        for (int j = 0; j < p->nreordering_vectors; j++) {
            p->reorderings[j] = new int[p->n];
            hpcxx_unpack(buffer, p->reorderings[j], p->n);
        }

        hpcxx_unpack(buffer, &(p->nscalings), 1);
        hpcxx_unpack(buffer, &(p->nscaling_vectors), 1);
        p->scalings = new double*[p->nscaling_vectors];
        for (int j = 0; j < p->nscaling_vectors; j++) {
            p->scalings[j] = new double[p->n];
            hpcxx_unpack(buffer, p->scalings[j], p->n);
        }

        hpcxx_unpack(buffer, &(p->lastoper), 1);

        p->allocated_reorderings = p->nreordering_vectors;
        p->allocated_scalings = p->nscaling_vectors;

        p++;
    }
}
