#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <unistd.h>
#include <errno.h>
#include <iostream.h>
#include <assert.h>
#include <mmio.h>
#include <SparseLinearSystem.h>

#ifdef REMOVE_
#define readmt_ readmt
#define ssrcsr_ ssrcsr
#endif

#ifdef sgi
extern "C" int symlink(const char *, const char *);
#endif


// ------------------
// Fortran functions:
// ------------------
extern "C" {
  void readmt_(int *nmax, int *nzmax, int *job, int *iounit,
               double *a, int *colinds, int *rowptrs, double *b,
               int *nrhs, char guesol[4], int *nrow, int *ncol,
               int *nnz, char title[80], char key[16], char type[8],
               int *ierr);

  void ssrcsr_(int*, double*, int*, int*, int*, double*, int*, int*, int*,
               int*);
           };

// ----------------------
// Local helper functions
// ----------------------
  void COOtoCSR(double *cooval, int *coocol, int *coorow);
  int  SSRtoCSR();
  int  MissingDiagonalElements(int *diag); 
  void InsertDiagonalZeros(int ndiag, int *diag);
  void GenerateRhs();

// Enumeration of matrix file types possible here
enum MatrixType { MMReal, MMComplex, MMPattern, MMInteger, HBReal };
char*           MatrixFileName;
char*           RhsFileName;  // Ain't gonna be used.  Hate those.

enum MatrixType KindOfMatrix;
int             NumRows;
int             NumNonZeros;
double*         MatrixValues;
double*         RightHandSide;
int*            ColumnIndices;
int*            RowPointers;
char*           MatrixTitle;
char*           MatrixKey;
char*           MatrixType;
int             ErrorCode;
// EventResults*   Results;

SparseLinearSystem *
ReadSparseMatrix(char *MatrixFileName, int *errflag) {

  /* ------------------- */
  /* Interface variables */
  /* ------------------- */

  int i=0, k=0;          /* Loop control variables */  
  int bsize;             /* Number of doubles for rhs b */
  double* b = NULL;

  /* ----------------------------- */
  /* Matrix data structure things. */
  /* ----------------------------- */

  int nrow = 0;             /* Number of rows in A            */
  int ncol = 0;             /* Number of cols in A            */
  int ierr = 0;             /* Error flag                     */
  int extra;                /* # of extra rhs vectors in HB file */
  
  /* ----------------------------------- */
  /* Quantities needed for H/B matrix IO */
  /* ----------------------------------- */
  
  int nmax = 0;             /* Max number of rows in matrix */
  int nzmax = 0;            /* Max number of nonzeros in matrix */
  int iounit = 6;           /* Unit number for Fortran I/O  */
  int nrhs = 0;             /* Number of right hand sides */
  char guesol[4];           /* Whether sol'ns or init vectors given */
  int job = 0;              /* Job specifier for readmt, prtmt */
  int ndiag = 0;
  int *tempintptr;
  double *tempdblptr;
  int src;
  int *diag;
  int rrow,rcol;

  int *coorow, *coocol;		// Matrix Market COO row/col indices
  double *cooval;		// Matrix Market COO values

  FILE *matfile;
  MM_typecode matcode;
  char destFile[80];

  int symmetric = 0;


  /* ---------------------------------------------------------------- */
  /* module only reads from a file name matrix, so symbolic link is   */
  /* used to access files of other names                              */
  /* ---------------------------------------------------------------- */

  ErrorCode = 0;
  *errflag = ErrorCode;

  strncpy(destFile,"matrix",80);
  if ( MatrixFileName == NULL || MatrixFileName[0] == '\0' ) {
    ErrorCode = 21; /* Matrix file not specified */
    *errflag = ErrorCode;
    return (new SparseLinearSystem());
  }

  if ( access(MatrixFileName, R_OK) ) {
    ErrorCode = 18; /* MatrixFile is not readable */
    *errflag = ErrorCode;
    return (new SparseLinearSystem());
  }

  if ((matfile=fopen(MatrixFileName,"r")) == NULL) {
    ErrorCode = 1;   /* can't open file */
    *errflag = ErrorCode;
    return (new SparseLinearSystem());
  }

  if ( strcmp(MatrixFileName,destFile) != 0) {
    remove(destFile);
    if (symlink(MatrixFileName, destFile)) {
      ErrorCode = 2;   /* can't make link with matrix file */
      *errflag = ErrorCode;
      return (new SparseLinearSystem());
    }
  }
  

  /*====================================================================*/
  /*=====   Read in matrix header from a file                       ====*/
  /*====================================================================*/
  
  /* --------------------------------------------------------------------- */
  /* First check if the file is in Matrix Market format, if not assume     */
  /* it is a Harwell-Boeing file. In either case read header to get size.  */
  /* --------------------------------------------------------------------- */


  MatrixType  = new char[8];
  MatrixKey   = new char[16];
  MatrixTitle = new char[80];

  if (MatrixTitle==NULL || MatrixKey==NULL || MatrixType==NULL) {
    ErrorCode = 17;
    *errflag = ErrorCode;
    return (new SparseLinearSystem());
  }


  if (mm_read_banner(matfile,&matcode) == 0) {
    /* ------------------- */
    /* MatrixMarket format */
    /* ------------------- */
    if ( !( mm_is_matrix(matcode) && mm_is_sparse(matcode))) {
      ErrorCode = 3;  /* this matrixmarket file type is not supported */
      *errflag = ErrorCode;
      fclose(matfile);
      return (new SparseLinearSystem());
    }
    if ( mm_read_mtx_crd_size(matfile,&nrow,&ncol,&NumNonZeros) != 0) {
      fclose(matfile);
      ErrorCode = 4;  /* can't read matrix header */
      *errflag = ErrorCode;
      return (new SparseLinearSystem());
    }

    strncpy(MatrixTitle,MatrixFileName,80);
    MatrixTitle[72] = '\0';

    strncpy(MatrixKey,MatrixFileName,8);
    MatrixKey[8] = '\0';

    strncpy(MatrixType,"  ",3);
    MatrixType[3] = '\0';

    strncpy(guesol," ",2);
    guesol[2] = '\0';
    nrhs = 0;
    extra = 1;
    if (mm_is_real(matcode))    KindOfMatrix = MMReal;
    if (mm_is_complex(matcode)) KindOfMatrix = MMComplex;
    if (mm_is_pattern(matcode)) KindOfMatrix = MMPattern;
    if (mm_is_integer(matcode)) KindOfMatrix = MMInteger;
    if (mm_is_symmetric(matcode)) symmetric = 1;
  }
  else {
    /* ------------------------------------------------------- */
    /* Try reading the file as a Harwell-Boeing matrix         */
    /* First, call readmt with job = 0, to just read the header*/
    /* information and get the sizes of arrays needed.         */
    /* ------------------------------------------------------- */
    job = 0;
    readmt_(&nmax, &nzmax, &job, &iounit, MatrixValues, ColumnIndices,
            RowPointers, b, &nrhs, guesol, &nrow, &ncol, &NumNonZeros,
            MatrixTitle, MatrixKey, MatrixType, &ierr);
    fclose(matfile);
   
    KindOfMatrix = HBReal;
    extra = 1;

    if (guesol[0]=='n' || guesol[0]=='N') extra++;
    if (guesol[1]=='x' || guesol[0]=='X') extra++;
    char tempTitle[73];
    if (tempTitle==NULL) {
      ErrorCode = 17;
      *errflag = ErrorCode;
      return (new SparseLinearSystem());
    }
    symmetric = 0;
  }

  /* ---------------------------------------------------------------- */
  /* Allocate  necessary memory to read matrix for both HB & MM files */
  /* ---------------------------------------------------------------- */

  NumRows = nrow;

  ColumnIndices   = new int[NumNonZeros];
  RowPointers     = new int[nrow+1];
  if (KindOfMatrix == MMComplex) {
    RightHandSide = new double[2*nrow];
    MatrixValues  = new double[2*NumNonZeros];
  } else {
    RightHandSide = new double[nrow];
    MatrixValues  = new double[NumNonZeros];
  }
  if (MatrixValues == NULL || RightHandSide == NULL
      || ColumnIndices == NULL || RowPointers == NULL) {
    delete[] MatrixValues;
    delete[] RightHandSide;
    delete[] RowPointers;
    delete[] ColumnIndices;
    ErrorCode = 5; /* Not enough memory available to read matrix */
    *errflag = ErrorCode;
    return (new SparseLinearSystem());
  }

  if ( KindOfMatrix == HBReal ) {
    /* -------------------------------------------------------------- */
    /* Now call readmt for matrix data.  Try to read rhs if supplied. */
    /* -------------------------------------------------------------- */
    if ( nrhs > 1 || extra > 1) {
      b = new double[nrhs*nrow*extra];
      if ( b ==NULL ) {
        ErrorCode = 22;
        *errflag = ErrorCode;
        delete []MatrixValues;
        delete []RightHandSide;
        delete []RowPointers;
        delete []ColumnIndices;
        return (new SparseLinearSystem());
      }
    }
    else
      b = RightHandSide;

    job = 3; /* Read matrix and rhs if supplied */
    nzmax = NumNonZeros;
    nmax  = nrow+1;
    bsize = nrhs*nrow*extra; /* Beware; on entry nrhs is length of b! */
    
    readmt_(&nmax, &nzmax, &job, &iounit, MatrixValues, ColumnIndices,
            RowPointers, b, &bsize, guesol, &nrow, &ncol, &NumNonZeros,
            MatrixTitle, MatrixKey, MatrixType, &ierr);
    // -----------------------------------------------------
    // Null-terminate the strings; Fortran does not do this.
    // -----------------------------------------------------
    MatrixTitle[72] = '\0';
    MatrixKey[8] = '\0';

    if ( ierr != 0) {
      if ( ierr != 5) {
        delete []   MatrixValues;
        delete []   RightHandSide;
        delete []   RowPointers;
        delete []   ColumnIndices;
        ErrorCode = 6; /* can't read HB file data */
        *errflag = ErrorCode;
        return (new SparseLinearSystem());
      }
      else {
      }
    }

    if ( nrhs > 1 || extra > 1) {
      for (i=0; i<nrow; i++)
        RightHandSide[i] = b[i];
      delete [] b ;
    }
  }
  else {  // Reading a Matrix Market file
    coorow   = new  int[NumNonZeros];
    coocol   = new  int[NumNonZeros];
    if (KindOfMatrix == MMComplex)
      cooval = new  double[2*NumNonZeros];
    else
      cooval = new  double[NumNonZeros];
    
    if (cooval==NULL) {
      ErrorCode = 7; /* Out of memory for COO structure */
      *errflag = ErrorCode;
      fclose(matfile);
      delete [] MatrixValues;
      delete [] RightHandSide;
      delete [] RowPointers;
      delete [] ColumnIndices;
      delete [] coorow;
      delete [] coocol;
      return (new SparseLinearSystem()) ;
    }
    for (i=0; i < NumNonZeros; i++) {
      if (feof(matfile)) {
        ErrorCode = 8;      /* unexpected end of file */ 
        *errflag = ErrorCode;
        fclose(matfile);
        delete [] MatrixValues;
        delete [] RightHandSide;
        delete [] RowPointers;
        delete [] ColumnIndices;
        delete [] coorow;
        delete [] coocol;
        delete [] cooval;
        return (new SparseLinearSystem());
      }
      switch(KindOfMatrix) {
      case MMReal:
        fscanf(matfile,"%d %d %lg\n",&coorow[i], &coocol[i], &cooval[i]);
        break;
      case MMComplex:
        fscanf(matfile,"%d %d %lg %lg\n",&coorow[i],
           &coocol[i], &cooval[i], &cooval[i+NumNonZeros]);
        break;
      case MMPattern:
        fscanf(matfile,"%d %d\n",&coorow[i], &coocol[i]);
        cooval[i] = 1.0;
        break;
      case MMInteger:
        fscanf(matfile,"%d %d %d\n",&coorow[i], &coocol[i], &k);
        cooval[i] = (double) k;
        break;
      }
    }
    fclose(matfile);

    /* ----------------------- */
    /* convert from COO to CSR */
    /* ----------------------- */
    COOtoCSR(cooval,coocol,coorow);
    delete [] coorow;
    delete [] coocol;
    delete [] cooval;
  }

  /* ------------------------------------------------------------- */
  /* Check that rowptrs and colinds are correct, or at least won't */
  /* cause a segmentation fault in later modules                   */
  /* ------------------------------------------------------------- */
  if ( RowPointers[0] != 1 || RowPointers[nrow] != NumNonZeros+1 ) {
    ErrorCode = 9; /* row pointers are incorrect */
    *errflag = ErrorCode;
    delete [] MatrixValues;
    delete [] RightHandSide;
    delete [] RowPointers;
    delete [] ColumnIndices;
    return (new SparseLinearSystem());
  }
  
  diag = new  int[nrow];
  if ( diag == NULL) {
    ErrorCode = 10; /* out of memory for diag array */
    *errflag = ErrorCode;
    delete []   MatrixValues;
    delete []   RightHandSide;
    delete []   RowPointers;
    delete []   ColumnIndices;
    return (new SparseLinearSystem());
  }
  
  ndiag = MissingDiagonalElements(diag);
  if ( ndiag < 0 ) {
    delete []   MatrixValues;
    delete []   RightHandSide;
    delete []   RowPointers;
    delete []   ColumnIndices;
    delete [] diag;
    ErrorCode = 10 -ndiag;
    *errflag = ErrorCode;
    return (new SparseLinearSystem());
  }

  /* -------------------------------------------------------------------- */
  /* if ndiag > 0 then the zero elements on the diagonal must be added to */
  /* the data structure                                                   */
  /* -------------------------------------------------------------------- */
  if ( ndiag > 0 ) {
    src = NumNonZeros;
    NumNonZeros += ndiag;

    if (KindOfMatrix == MMComplex) {
      ErrorCode = 23;
      *errflag = ErrorCode;
      return (new SparseLinearSystem()) ;
    }
    else {
      tempintptr    = ColumnIndices;
      ColumnIndices = new int[NumNonZeros];
      if (ColumnIndices != NULL) {
	for (k = 0; k < src; k++) ColumnIndices[k] = tempintptr[k];
      }
      delete [] tempintptr ;
      tempdblptr = MatrixValues;
      MatrixValues = new double[NumNonZeros];
      
      if ( MatrixValues == NULL || ColumnIndices==NULL) {
        ErrorCode = 13; /* out of memory for zero diagonal entries */
        *errflag = ErrorCode;
        delete [] MatrixValues;
        delete [] RightHandSide;
        delete [] RowPointers;
        delete [] tempdblptr;
        delete [] ColumnIndices;
        return (new SparseLinearSystem()) ;
      }
      if (MatrixValues != NULL) {
	for (k = 0; k < src; k++) MatrixValues[k] = tempdblptr[k];
      }
      delete [] tempdblptr;
    }

    if ( MatrixValues == NULL) {
      ErrorCode = 13; /* out of memory for zero diagonal entries */
      *errflag = ErrorCode;
      delete [] MatrixValues;
      delete [] RightHandSide;
      delete [] RowPointers;
      delete [] ColumnIndices;
      return (new SparseLinearSystem()) ;
    }
    if (KindOfMatrix == MMComplex) {
      for (i = 2*src-1; i >= src; i--) 
        MatrixValues[i+ndiag] = MatrixValues[i];
    }

    InsertDiagonalZeros(ndiag,diag);
  }

  if ( symmetric ) { 
      ErrorCode = 29;
      *errflag = ErrorCode;
      return (new SparseLinearSystem());
  }
  
  /* ------------------------------ */
  /* Matrix looks good, so continue */
  /* ------------------------------ */
  
  /*    ---------------------------------------------------------------- */
  /*    If right hand side was not supplied, then it needs to be created */
  /*    artifically.  That is done in setb.                              */
  /*    ---------------------------------------------------------------- */

  if (nrhs <= 0 && (RhsFileName==NULL || RhsFileName[0] == '\0') ) {
    GenerateRhs();
  }
  else if ( RhsFileName != NULL && RhsFileName[0] != '\0' ) {
    if ((matfile=fopen(RhsFileName,"r")) == NULL ||
        mm_read_banner(matfile,&matcode) != 0 ) {
      ErrorCode = 14; /* can't access file, for rhs -- generate one */
      *errflag = ErrorCode;
      fclose(matfile);
      return (new SparseLinearSystem());
    }
    if ( !( mm_is_real(matcode) ) ) {
      ErrorCode = 15;
      *errflag = ErrorCode;
      fclose(matfile); /* this matrix file type is not supported */
      return (new SparseLinearSystem());
    }
    if (mm_read_mtx_array_size(matfile,&rrow,&rcol) != 0) {
      ErrorCode = 15; /* this matrix file type is not supported */
      *errflag = ErrorCode;
      fclose(matfile);
      return (new SparseLinearSystem());
    }
    if ( rcol !=1 || rrow != nrow) {
      ErrorCode = 16; /* rhs is wrong size */
      *errflag = ErrorCode;
      fclose(matfile);
      return (new SparseLinearSystem());
    }
    for (i=0;i<rrow;i++) {
      if (feof(matfile)) {
        fclose(matfile);
        ErrorCode = 8; /* unexpected end of file */
        *errflag = ErrorCode;
        return (new SparseLinearSystem());
      }
      fscanf(matfile,"%lg\n",(RightHandSide)+i);
    }
    fclose(matfile);
  }
  else {
    /* fprintf(stderr,"right hand side #%d supplied with H/B file\n",1); */
  }

  /* -------------------------------------------- */
  /* Now load up the SparseLinearSystem with data */
  /* -------------------------------------------- */
  SparseLinearSystem *sls = new SparseLinearSystem(MatrixTitle,
						  MatrixKey,
						  nrow, 
						  NumNonZeros, 
						  MatrixValues, 
						  RowPointers,
						  ColumnIndices, 
						  b,
						  1,
						  0,
						  *(new Mungings));

  remove(destFile);
  return sls;
}

void COOtoCSR(double *cooval, int *coocol, int *coorow) {
    int k=0;
    int i=0;
    int j=0;
    for ( i=0; i < NumRows; i++) {
        RowPointers[i] = k+1;
        for ( j=0; j< NumNonZeros; j++) {
            if ( coorow[j] == i+1) {
              MatrixValues[k] = cooval[j];
              if (KindOfMatrix == MMComplex)
                MatrixValues[k+NumNonZeros] = cooval[j+NumNonZeros];
              ColumnIndices[k] = coocol[j];
              k++;
            }
        }
    }
    RowPointers[i] = k+1;
}

int MissingDiagonalElements(int *diag) {
  int ndiag = 0, i, j;
  for (i=0;i < NumRows;i++) {
    diag[i] = 0;
    if (RowPointers[i+1] < RowPointers[i]) {
      return -1; /* row pointer are incorrect */
    }
    for (j = RowPointers[i]; j< RowPointers[i+1]; j++) {
      if (ColumnIndices[j-1] < 0 || ColumnIndices[j-1] > NumRows) {
        return -2; /* column pointers are incorrect */
      }
      if (ColumnIndices[j-1] == i+1)
        diag[i] = j;
    }
    ndiag += (diag[i] == 0);
  }
  return ndiag;
}

void InsertDiagonalZeros(int ndiag, int *diag) {

  int destlast, dest = NumNonZeros, src = NumNonZeros - ndiag;
  int i, j;
  for (i = NumRows-1;i>=0;i--) {
    destlast = dest;
    for (j = RowPointers[i+1]-1; j >= RowPointers[i]; j--) {
      dest--;
      src--;
      if ( diag[i]==0 && ColumnIndices[src]<i+1 ) {
        MatrixValues[dest] = (double) 0.0;
        if (KindOfMatrix == MMComplex)
          MatrixValues[dest+NumNonZeros] = (double) 0.0;
        ColumnIndices[dest] = i+1;
        diag[i] = dest;
        dest--;
      }
      MatrixValues[dest] = MatrixValues[src];
      if (KindOfMatrix == MMComplex)
        MatrixValues[dest+NumNonZeros] = MatrixValues[src+NumNonZeros];

      ColumnIndices[dest] = ColumnIndices[src];
    }
    RowPointers[i+1] = destlast+1;
  }
  RowPointers[NumRows] = NumNonZeros+1;
  RowPointers[0] = 1;
}


void GenerateRhs() {
  int i;
  for (i=0;i < NumRows; i++)
    RightHandSide[i] = (double)1.0 - 2.0/(double)(i+1);
}

