/*
//  --------------------------------------------------------------------------------
//  The SparseLinearSystem class is the sparse system data structure and
//  usual constructors, destructor, and assignment operator.  We use a
//  compressed sparse row (CSR) sparse matrix format, but have flexibility to
//  use others in the future.  So the data field storageType is currently not
//  used, but will allow for the use of different matrix data structures such as
//  CSC, COO, MSR, etc.
//  
//  Much of the heritage of the SparseLinearSystem class comes from Harwell/Boeing
//  formatted matrices, and we keep the "title" and "key" fields of those to give a
//  long and short name for the system, respectively.
//  
//  
//    titlelen: length of matrix title including the null terminator
//              (so, titlelen = ::strlen(title) + 1).
//    title: pointer to the title string ( null terminated ).
//  
//    keylen: length of matrix key including the null terminator.
//    key:  pointer to the key string ( null terminated ).
//  
//    n: order of matrix. Later we may need to change this to allow nonsquare
//       matrices, but for now the numbers of rows and columns are always the
//       same.
//  
//    nnz: number of non-zeros in the sparse matrix data structure.  More
//         precisely, the number of stored entries - since some zeros may be
//         explicitly stored in some applications.  We use standard terminology,
//         however, and call the stored entries "nonzeros".
//  
//    colinds: pointer to the column indices of the matrix nonzeros.  This
//             array must be at least of length nnz, and uses 1-based indexing.
//  
//    rowptrs: pointer to the row pointers. This array
//             must be of length at least n+1, and uses 1-based indexing.
//  
//    values:  pointer to the matrix values, of length at least nnz.
//  
//    rhs: pointer to the values of the right-hand side vector. This array
//         must be of at least length n.
//
//    x: pointer to the solution vector, set up by solve routines.
//  
//    nparts:  for matrices partitioned for shared memory machines, this
//             indicates how many "block rows" are present.  
//    bstarts: array of pointers to the first row in each block row.  So,
//             for example, rows bstarts[k] to bstarts[k+1]-1 comprise the
//             k-th block row of the matrix A, where 0 < k < nparts+1.
//             This array must be of length at least nparts+1.
//             
//    Mungings: a class that contains information needed to retrieve the solution
//              to a sparse system that has undergone various reorderings and
//              scalings.
//  
//  
//  Randall Bramley
//  Department of Computer Science
//  Indiana University, Bloomington
//  bramley@cs.indiana.edu                             
//  Sun Oct 17 10:43:04 EST 1999
//  --------------------------------------------------------------------------------
*/

#include "SparseLinearSystem.h"
#include "../../../../framework/util/HPCxx_String.h"

#ifdef sgi
extern "C" char *strdup(const char *);
#endif

SparseLinearSystem::SparseLinearSystem() {

  storageType   = CSR;

  titlelen      = 0;
  title         = NULL;

  keylen        = 0;
  key           = NULL;

  n             = 0;
  nnz           = 0;

  colinds       = NULL;
  rowptrs       = NULL;
  values        = NULL;
  rhs           = NULL;
  x             = NULL;

  nparts        = 0;
  bstarts       = NULL;
}

SparseLinearSystem::SparseLinearSystem(const SparseLinearSystem& sls)
 : mungings(sls.mungings) {

    copy(sls);
}

// Randy, sorry to change the input variable
// names with an underscore in the end.
// This gets rid of the warnings.
SparseLinearSystem::SparseLinearSystem(
                                    char *title_,
                                    char *key_,
                                    int n_, 
                                    int nnz_, 
                                    double* values_, 
                                    int* rowptrs_,
                                    int* colinds_, 
                                    double *rhs_,
                                    int nparts_,
                                    int* bstarts_,
                                    Mungings mungings_){

  this->storageType = CSR;
  this->titlelen    = titlelen;
  this->title       = strdup(title_);
  this->keylen      = keylen;
  this->key         = strdup(key_);
  this->n           = n_;
  this->nnz         = nnz_;
  this->colinds     = colinds_;
  this->rowptrs     = rowptrs_;
  this->values      = values_;
  this->rhs         = rhs_;
  this->x           = NULL;
  this->nparts      = nparts_;
  this->bstarts     = bstarts_;
  this->mungings    = mungings_;
}

// Randy, sorry to change the input variable
// names with an underscore in the end.
// This gets rid of the warnings.
SparseLinearSystem::SparseLinearSystem(char *title_,
                       char *key_,
                           int n_, 
                                       int nnz_, 
                                       double* values_, 
                                       int* rowptrs_,
                                       int* colinds_, 
                                       double *rhs_){
  this->storageType = CSR;

  this->titlelen    = titlelen;
  this->title       = title_;

  this->keylen      = keylen;
  this->key         = key_;

  this->n           = n_;
  this->nnz         = nnz_;

  this->colinds     = colinds_;
  this->rowptrs     = rowptrs_;
  this->values      = values_;
  this->rhs         = rhs_;
  this->x           = NULL;

  this->nparts      = 0;
  this->bstarts     = NULL;
}

SparseLinearSystem::SparseLinearSystem(int n_, 
                                       int nnz_, 
                                       double* values_, 
                                       int* rowptrs_,
                                       int* colinds_){
  this->storageType = CSR;

  this->titlelen    = 0;
  this->title       = NULL;

  this->keylen      = 0;
  this->key         = NULL;

  this->n           = n_;
  this->nnz         = nnz_;

  this->colinds     = colinds_;
  this->rowptrs     = rowptrs_;
  this->values      = values_;
  this->rhs         = NULL;
  this->x           = NULL;

  this->nparts      = 0;
  this->bstarts     = NULL;
}

void
SparseLinearSystem::operator=(const SparseLinearSystem &sls) {

    free();
    copy(sls);
    mungings = sls.mungings;
}

SparseLinearSystem::~SparseLinearSystem(){
    free();
}

void
SparseLinearSystem::copy(const SparseLinearSystem& sls) {

    int i;

    storageType = sls.storageType;
    titlelen    = sls.titlelen;
    title       = strdup(sls.title);
    keylen      = sls.keylen;
    key         = strdup(sls.key);
    n           = sls.n;
    nnz         = sls.nnz;

    colinds = new int[nnz];         // nnz
    for (i = 0; i < nnz; i++) {
        colinds[i] = sls.colinds[i];
    }

    rowptrs = new int[n + 1];       // n + 1
    for (i = 0; i < n + 1; i++) {
        rowptrs[i] = sls.rowptrs[i];
    }

    values = new double[nnz];       // nnz
    for (i = 0; i < nnz; i++) {
        values[i] = sls.values[i];
    }

    rhs = new double[n];            // n
    for (i = 0; i < n; i++) {
        rhs[i] = sls.rhs[i];
    }

    if (sls.x != NULL) {            // n or null.
        x = new double[n];
        for (i = 0; i < n; i++) {
            x[i] = sls.x[i];
        }
    } else {
        x = NULL;
    }

    nparts      = sls.nparts;

    bstarts = new int[nparts];      // nparts
    for (i = 0; i < nparts; i++) {
        bstarts[i] = sls.bstarts[i];
    }
}

void SparseLinearSystem::free() {

    ::free(title);
    ::free(key);

    delete[] colinds;
    delete[] rowptrs;
    delete[] values;
    delete[] rhs;
    delete[] x;
    delete[] bstarts;
}

void hpcxx_pack(HPCxx_Buffer& buffer, SparseLinearSystem* p, int count){

  for(int i=0; i<count; i++) {

    cout << "SparseLinearSystem:: hpcxx_pack() called " << endl;

    int t = p->storageType;
    hpcxx_pack(buffer, &t, 1);

    hpcxx_pack(buffer, &(p->titlelen), 1);
    HPCxx_String title(p->title);
    hpcxx_pack(buffer, &title, 1);
    cout << "SparseLinearSystem:: hpcxx_pack title: " << title << endl;
    cout.flush();

    hpcxx_pack(buffer, &(p->keylen), 1);
    HPCxx_String key(p->key);
    hpcxx_pack(buffer, &key, 1);
    cout << "SparseLinearSystem:: hpcxx_pack title: " << key << endl;
    cout.flush();

    hpcxx_pack(buffer, &(p->n), 1);
    hpcxx_pack(buffer, &(p->nnz), 1);
    cout << "SparseLinearSystem:: hpcxx_pack n and nnz: " << p->n << " " << p->nnz  << endl;
    cout.flush();

    // colinds should be at least be nnz
    hpcxx_pack(buffer, p->colinds, p->nnz);
    cout << "SparseLinearSystem:: hpcxx_pack colinds done " << endl;
    // rowptrs should be at least n+1
    hpcxx_pack(buffer, p->rowptrs, (p->n + 1));
    cout << "SparseLinearSystem:: hpcxx_pack rowptrs done " << endl;
    // value should be at least be nnz
    hpcxx_pack(buffer, p->values, p->nnz);
    cout << "SparseLinearSystem:: hpcxx_pack values  done " << endl;
    // rsh should be at least be n
    hpcxx_pack(buffer, p->rhs, p->n);
    cout << "SparseLinearSystem:: hpcxx_pack rsh  done " << endl;
    // FIXME FIXME
    // I am *guessing* that x is of size n
    // Need to confirm this from Randy.
    // If you think you know more Linear System
    // stuff than me, then fix it! or else
    // just ask Randy.
    int xNull;
    if (p->x == NULL) {
      xNull = 1;
    } else {
      xNull = 2;
    }
    hpcxx_pack(buffer, &xNull, 1);

    if (xNull == 2) {
      hpcxx_pack(buffer, p->x, p->n);
    }
    cout << "SparseLinearSystem:: hpcxx_pack x done " << endl;
    cout.flush();
    hpcxx_pack(buffer, &(p->nparts), 1);
    // bstarts should be at least be nparts +1 long
    hpcxx_pack(buffer, p->bstarts, (p->nparts+ 1));
    hpcxx_pack(buffer, &(p->mungings), 1);
    cout << "SparseLinearSystem:: hpcxx_pack() done " << endl;
    cout.flush();
  }
}

void
hpcxx_unpack(HPCxx_Buffer& buffer, SparseLinearSystem* p, int count){
const char *xfn = "hpcxx_unpack";

    for(int i=0; i<count; i++) {

        cout << xfn << " called." << endl << flush;

        int t;
        hpcxx_unpack(buffer, &t, 1);
        p->storageType = (ds_type) t;

        hpcxx_unpack(buffer, &(p->titlelen), 1);
        HPCxx_String title;
        hpcxx_unpack(buffer, &title, 1);
        p->title = strdup((char *) title);
        cout << xfn << " title: " << p->title << endl << flush;

        hpcxx_unpack(buffer, &(p->keylen), 1);
        HPCxx_String key;
        hpcxx_unpack(buffer, &key, 1);
        p->key = strdup(key);
        cout << xfn << " key: " << p->key << endl << flush;

        hpcxx_unpack(buffer, &(p->n), 1);
        hpcxx_unpack(buffer, &(p->nnz), 1);
        cout << xfn << " n: " << p->n << endl << flush;
        cout << xfn << " nnz: " << p->nnz << endl << flush;

        // colinds should be at least be nnz
        p->colinds = new int[p->nnz];
        hpcxx_unpack(buffer, p->colinds, p->nnz);
        cout << xfn << " colinds: " << p->colinds << endl << flush;

        // rowptrs should be at least n+1
        p->rowptrs = new int[p->n+1];
        hpcxx_unpack(buffer, p->rowptrs, (p->n + 1));
        cout << xfn << " rowptrs: " << p->rowptrs << endl << flush;

        // value should be at least be nnz
        p->values = new double[p->nnz];
        hpcxx_unpack(buffer, p->values, p->nnz);
        cout << xfn << " values: " << p->values << endl << flush;

        // rsh should be at least be n
        p->rhs =  new double[p->n];
        hpcxx_unpack(buffer, p->rhs, p->n);
        cout << xfn << " rhs: " << p->rhs << endl << flush;

        // FIXME FIXME
        // I am *guessing* that x is of size n
        // Need to confirm this from Randy.
        // If you think you know more Linear System
        // stuff than me, then fix it! or else
        // just ask Randy and send me an email
        // apologizing for daring to doubt my guess.
        int isXNULL = 0;
        hpcxx_unpack(buffer, &isXNULL, 1);
        if (isXNULL == 2) {
          p->x = new double[p->n];
          hpcxx_unpack(buffer, p->x, p->n);
        } else {
          p->x = NULL;
        }

        // bstarts should be at least be nparts long
        hpcxx_unpack(buffer, &(p->nparts), 1);
        p->bstarts = new int[p->nparts + 1];
        hpcxx_unpack(buffer, p->bstarts, p->nparts + 1);
        cout << xfn << " bstarts: " << p->bstarts << endl << flush;

        hpcxx_unpack(buffer, &(p->mungings), 1);
        cout << "SparseLinearSystem:: hpcxx_unpack() done" << endl;
        cout.flush();
    }
}
