/*
//  --------------------------------------------------------------------------------
//  The SparseLinearSystem class is the sparse system data structure and
//  usual constructors, destructor, and assignment operator.  We use a
//  compressed sparse row (CSR) sparse matrix format, but have flexibility to
//  use others in the future.  So the data field storageType is currently not
//  used, but will allow for the use of different matrix data structures such as
//  CSC, COO, MSR, etc.
//  
//  Much of the heritage of the SparseLinearSystem class comes from Harwell/Boeing
//  formatted matrices, and we keep the "title" and "key" fields of those to give a
//  long and short name for the system, respectively.
//  
//  
//    titlelen: length of matrix title including the null terminator
//              (so, titlelen = ::strlen(title) + 1).
//    title: pointer to the title string ( null terminated ).
//  
//    keylen: length of matrix key including the null terminator.
//    key:  pointer to the key string ( null terminated ).
//  
//    n: order of matrix. Later we may need to change this to allow nonsquare
//       matrices, but for now the numbers of rows and columns are always the
//  	 same.
//  
//    nnz: number of non-zeros in the sparse matrix data structure.  More
//  	   precisely, the number of stored entries - since some zeros may be
//  	   explicitly stored in some applications.  We use standard terminology,
//  	   however, and call the stored entries "nonzeros".
//  
//    colinds: pointer to the column indices of the matrix nonzeros.  This
//             array must be at least of length nnz, and uses 1-based indexing.
//  
//    rowptrs: pointer to the row pointers. This array
//             must be of length at least n+1, and uses 1-based indexing.
//  
//    values:  pointer to the matrix values, of length at least nnz.
//  
//    rhs: pointer to the values of the right-hand side vector. This array
//         must be of at least length n.
//  
//    nparts:  for matrices partitioned for shared memory machines, this
//             indicates how many "block rows" are present.  
//    bstarts: array of pointers to the first row in each block row.  So,
//             for example, rows bstarts[k] to bstarts[k+1]-1 comprise the
//             k-th block row of the matrix A, where 0 < k < nparts+1.
//             This array must be of length at least nparts+1.
//             
//    Mungings: a class that contains information needed to retrieve the solution
//              to a sparse system that has undergone various reorderings and
//              scalings.
//  
//  
//  Randall Bramley
//  Department of Computer Science
//  Indiana University, Bloomington
//  bramley@cs.indiana.edu                             
//  Sun Oct 17 10:43:04 EST 1999
//  --------------------------------------------------------------------------------
*/

#ifndef SPARSELINEARSYSTEM_H
#define SPARSELINEARSYSTEM_H

#include "Mungings.h"
#include "MatrixDataStructures.h"

class SparseLinearSystem {

public:
	  // LIFECYCLE
	  SparseLinearSystem();
	  SparseLinearSystem( const SparseLinearSystem& sls);
	  void operator=(const SparseLinearSystem &sls);

	  SparseLinearSystem(char *title,
			     char *key,
			     int n, 
			     int nnz, 
			     double* values, 
			     int* rowptrs,
			     int* colinds, 
			     double *rhs,
			     int nparts,
			     int* bstarts,
			     Mungings mungings);
	  SparseLinearSystem(char *title,
			     char *key,
			     int n, 
			     int nnz, 
			     double* values, 
			     int* rowptrs,
			     int* colinds, 
			     double *rhs);
	  SparseLinearSystem(int n, 
                             int nnz, 
                             double* values, 
                             int* rowptrs,
                             int* colinds, 
                             double *rhs);
	  SparseLinearSystem(int n, 
                             int nnz, 
                             double* values, 
                             int* rowptrs,
                             int* colinds);

	  ~SparseLinearSystem();
	

	  // DATA FIELDS
	  ds_type  storageType;
	
	  int      titlelen;
	  char     *title;
	 
	  int      keylen;
	  char     *key;
	
	  int      n;
	  int      nnz;
	
	  int      *colinds;
	  int      *rowptrs;
	  double   *values;
	  double   *rhs;
	  double   *x;
	
	  int      nparts;
	  int      *bstarts;
	  
	  Mungings mungings;
	  
	  friend void hpcxx_pack(HPCxx_Buffer& b, SparseLinearSystem* sls, int count);
	  friend void hpcxx_unpack(HPCxx_Buffer& b, SparseLinearSystem* p, int count);
			    
	  private:

		  void free();
		  void copy(const SparseLinearSystem &);
};

#endif
 

