#include "DiscretizerComponent.h"
#include "Discretizer.h"
#include "TimeStep_impl.h"
#include "DiscretizerParms_impl.h"

#include "../../../../framework/ccacore/Port.h"
#include "../../../../framework/ccacore/CCATServicesProxy.h"
#include "../../../../framework/ccacore/RegisterType.h"
#include "../../../../framework/ccacore/FRAMEWORK_REGISTRATION_ID.h"

#include "../../../../framework/idl/sparseLinearSystem/UsesSparseLinearSystem.h"
#include "../../../../framework/idl/geometry/UsesGeometry.h"

DiscretizerComponent::DiscretizerComponent() {

    int ec = -1;

    m_core = NULL;
    m_timeStepPort = new TimeStep_impl(this);
    m_discretizerParmsPort = new DiscretizerParms_impl(this);
    m_computing = false;
    m_stopped = false;

    /*
     * Init mutex
     */

    pthread_mutexattr_t attr;

    pthread_mutexattr_init(&attr);

    #if defined(DEBUG) && !defined(__linux__)
        ec = pthread_mutexattr_settype(&attr, PTHREAD_MUTEX_ERRORCHECK);
        assert(ec == 0);
    #endif

    pthread_mutex_init(&m_lock, &attr);

    pthread_mutexattr_destroy(&attr);

    /*
     * Init condition variable.
     */

    pthread_cond_init(&m_cv, NULL);

    /*
     * Init pthread attr
     */

    ec = pthread_attr_init(&m_threadAttr);
    assert(ec == 0);
    ec = pthread_attr_setdetachstate(&m_threadAttr, PTHREAD_CREATE_DETACHED);
    assert(ec == 0);
    //ec = pthread_attr_setscope(&m_threadAttr, PTHREAD_SCOPE_SYSTEM);
    //assert(ec == 0);
}

DiscretizerComponent::~DiscretizerComponent() {

    int ec = -1;

    if (m_core != NULL) {
        delete m_core;
    }

    delete m_timeStepPort;
    delete m_discretizerParmsPort;

    ec = pthread_attr_destroy(&m_threadAttr);
    assert(ec == 0);

    ec = pthread_cond_destroy(&m_cv);
    assert(ec == 0);

    ec = pthread_mutex_destroy(&m_lock);
    assert(ec == 0);
}

void
DiscretizerComponent::setServices(Services *cc) {

    if (cc == NULL) {

        assert(m_core != NULL);
        delete m_core;
        m_core = NULL;
        return;

    } else {

        m_core = cc;

	   // Coming in from a time generator component
        m_core->addProvidesPort(m_timeStepPort,
         new DefaultPortInfo("inputTimeStep", "TimeStep_idl"));

	// Coming in from a the GUI
        m_core->addProvidesPort(m_discretizerParmsPort,
         new DefaultPortInfo("inputDiscretizerParms", "DiscretizerParms_idl"));

	   // Going out to an assembler component
        m_core->registerUsesPort(new DefaultPortInfo("outputGeometry",
         "Geometry_idl"));

	   // Going out to a linear system solver/analyzer component
        m_core->registerUsesPort(new DefaultPortInfo("outputSLS",
         "SparseLinearSystem_idl"));
    }
}

struct Info {
    Info(DiscretizerComponent *cop, TimeStep *ts_)
     : component(cop), ts(ts_) {}
    DiscretizerComponent *const component;
    TimeStep *const ts;
};

void *
DiscretizerComponent::compute_wrapper(void *vp) {

    Info *ip = static_cast<Info *>(vp);
    Info info = *ip;
    delete ip;

    info.component->compute(info.ts);

    return NULL;
}

void
DiscretizerComponent::startComputation(TimeStep *ts) {

    pthread_t tid;

    Info *info = new Info(this, ts);
    assert(info != NULL);

    pthread_create(&tid, &m_threadAttr, compute_wrapper, info);
}

void
DiscretizerComponent::compute(TimeStep *ts) {
const char *xfn = "DiscretizerComponent::compute";

    cout << xfn << " called." << endl;
  
    assert(ts != NULL);

    SparseLinearSystem *sls = new SparseLinearSystem();
    Geometry *geom          = new Geometry();

    int ec = Discretizer(ts->M, ts->N, ts->time, sls, geom);
    assert(ec == 0);

    // send the SLS
    UsesSparseLinearSystem *upSLS = dynamic_cast<UsesSparseLinearSystem *>(m_core
     ->getPort("outputSLS"));
    cout << xfn << "sendSparseLinearSystem() attempting ..." << endl << flush;
    upSLS->sendSparseLinearSystem(sls);
    m_core->releasePort("outputSLS");
    cout << xfn << "sendSparseLinearSystem() done " << endl << flush;

    // send the Geometry
    UsesGeometry *upGeom = dynamic_cast<UsesGeometry *>(m_core
     ->getPort("outputGeometry"));
    cout << xfn << "sendGeometry() attempting ..." << endl << flush;
    upGeom->sendGeometry(geom);
    m_core->releasePort("outputGeometry");
    cout << xfn << "sendGeometry() done" << endl << flush;
    
    pthread_mutex_lock(&m_lock);
    m_computing = false;
    pthread_cond_broadcast(&m_cv);
    pthread_mutex_unlock(&m_lock);

    //    delete ts;
    //    delete sls;
    //    delete geom;
}
