#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <iostream.h>

/*--------------------------------------------------------------------------------*/
/*---------------------  setup.c    ----------------------------------------------*/
/*--------------------------------------------------------------------------------*/
/*                                                                                */
/* Function to create the discretized linear system corresponding the linear PDE  */
/*                                                                                */
/*     a*u_xx + b*u_yy + c*u_x * d*u_y + e*u = f,                                 */
/*                                                                                */
/* where all of the coefficient functions of a-f can be any functions that depend */
/* on (x,y).  The functions are appended to this file, but can be supplied and    */
/* compiled separately.  The PDE is presumed to be on the unit square, with a     */
/* uniform M x N discretization mesh.  Boundary conditions are supplied in a      */
/* function BC(x,y, time).                                                        */
/*                                                                                */
/* The arguments are the discretization size, and four pointers.  This routine    */
/* malloc's the memory needed for the arrays those pointers will point to.        */
/* This preliminary version actually creates the sparse matrix A in               */
/* coordinatewise (COO) format, but will be changed to CSR as soon as it is       */
/* validated and tested.  The vector b is the right hand side vector.             */
/*                                                                                */
/* The function getsten returns the stencil operator values corresponding to the  */
/* four arms of standard second order centered differences:                       */
/*                                                                                */
/*                            stencil[4]                                          */
/*                                |                                               */
/*                                |                                               */
/*                                |                                               */
/*                                |                                               */
/*      stencil[1]------------stencil[0]--------------stencil[3]                  */
/*                                |                                               */
/*                                |                                               */
/*                                |                                               */
/*                                |                                               */
/*                            stencil[2]                                          */
/*                                                                                */
/* Additionally, the component stencil[5] contains the right hand side component  */
/* corresponding to the forcing function f(x,y).                                  */
/*                                                                                */
/* The domain is the unit square, with M+1 x N+1 mesh lines.                      */
/* The mesh lines corresponding to                                                */
/* 	x = 0                                                                     */
/* 	x = M+1                                                                   */
/* 	y = 0                                                                     */
/* 	y = N+1                                                                   */
/* are assumed to correspond to Dirichlet boundary conditions,                    */
/* so the total number of unknowns is (M-1)*(N-1).                                */
/*                                                                                */
/* Because C is not great for returning multiple values of different types,       */
/* I have done the usual Fortran trick: have it return a pointer to a vector      */
/* of pointers to doubles, then cast the pointers to integers if necessary.       */
/* This way setup.c creates the storage as well as the values for the matrix      */
/* A and its data structure.  This also means finagling for nnz.                  */
/*                                                                                */
/* The interpretation of the return vector lhs[] is:                              */
/*      lhs[0] = pointer to Avals  (length nnz)                                   */ 
/*      lhs[1] = pointer to row    (length n+1)                                   */
/*      lhs[2] = pointer to col    (length nnz)                                   */
/*      lhs[3] = pointer to b      (length n)                                     */
/*      lhs[4] = pointer to pointer to nnz                                        */
/*      lhs[5] = pointer to pointer to boundary condition vector (length 2(M+N))  */
/*                                                                                */
/* The boundary condition vector, called boundary[], is set up to have            */
/* the components in order:                                                       */
/*      the x = 0 boundary, with corners                                          */
/*      the x = 1 boundary, with corners                                          */
/*      the y = 0 boundary, without corners                                       */
/*      the y = 1 boundary, without corners                                       */
/*                                                                                */
/* The mesh is assumed to be M x N, and the lines corresponding to                */
/* ix = 0, ix = M, iy = 0, and iy = N are boundary conditions, specified          */
/* in the function BC().  So for M = 5 and N = 4, the mesh looks like             */
/*                                                                                */
/*             B   B   B   B   B   B                                              */
/*             B   8   9   10  11  B                                              */
/*             B   4   5   6   7   B                                              */
/*             B   0   1   2   3   B                                              */
/*             B   B   B   B   B   B                                              */
/*                                                                                */
/* where B = boundary condition node and the numbered nodes are the               */
/* indexing of the unknowns.  Furthermore, the boundary condition vector          */
/* boundary[] returns the values in the boundary nodes, with the indexing         */
/*                                                                                */
/*             6    7    8    9    10   11                                        */
/*             14   X    X    X    X    17                                        */
/*             13   X    X    X    X    16                                        */
/*             12   X    X    X    X    15                                        */
/*             0    1    2    3    4    5                                         */
/*                                                                                */
/*                                                                                */
/*--------------------------------------------------------------------------------*/
/*                                                                                */
/*      Randall Bramley                                                           */
/*      Department of Computer Science                                            */
/*      Indiana University                                                        */
/*      Phone: (812) 855-7790                                                     */
/*      email: bramley@cs.indiana.edu                                             */
/*      Sun Sep 19  8:32:01 EST 1999                                              */
/*        Modified to return boundary condition vector                            */
/*        Sun Sep 19 11:36:04 EST 1999                                            */
/*                                                                                */
/*--------------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------------*/


/*-----------------------*/
/* Function prototypes:  */
/*-----------------------*/

      /* ------------------------------------------ */
      /* Fixed, independent of problem PDE to solve */
      /* ------------------------------------------ */
      void getsten(int i,int j,int M,int N, double time, double stencil[6]); 

extern "C" {
      void qs2i1d_(int *row, int *col, double *Aval, int *nnz_ptr, int *incdecr);
      void rowinc_(int *nzrow, int *col, double *Aval);
}

      /* -------------------------------------------------- */
      /* Dependent on problem PDE to solve                  */
      /* BC is: double BC(double x, double y, double time); */
      /* -------------------------------------------------- */
#include "BC.h"

/*---------------------------------------------------------------------*/
/* Function setup: create the sparse matrix for five point differences */
/*---------------------------------------------------------------------*/
double **setup(int M,         /* Number of mesh lines in x-direction  */
               int N,         /* Number of mesh lines in y-direction  */
               double time)   /* Time in the time stepping routine    */

{
 int n;                 /* Number of unknowns in system   */
 int nnz;               /* Number of nonzeros in matrix A */
 int *nnz_ptr;          /* Pointer to nnz for return values */
 int offset;            /* Offset for indexing  */
 int incdecr = 1;       /* Flag for sorting */
 int bottom = 0;        /* Start index in boundary vector for bottom BC */
 int top    = M+1;      /* Start index in boundary vector for top BC    */
 int left   = 2*M+2;    /* Start index in boundary vector for left BC   */
 int right  = 2*M+N+1;  /* Start index in boundary vector for right BC  */

 register int node, ind, ind1, ind2, ind3, ind4, i, j;
 int *rowptr;
 int ibegin, iend, nzrow, thisrow, k;
 double stencil[6];

 double *Aval;
 double *b;
 double *boundary;
 int *row;
 int *col;

 double **lhs;

 lhs = (double **) malloc(6*sizeof(double *));

/* ------------------------------------ */
/* nnz = number of nonzeros in matrix A */
/* ------------------------------------ */
 nnz_ptr = (int *) malloc(sizeof(int));
 nnz = 5*(M-1)*(N-1) - 2*(M+N-2);
 *nnz_ptr = nnz;
 
 Aval = (double *) malloc(nnz*sizeof(double));
 row  = (int *) malloc(nnz*sizeof(int));
 col  = (int *) malloc(nnz*sizeof(int));
 b    = (double *) malloc((M-1)*(N-1)*sizeof(double));
 boundary = (double *) malloc(2*(N+M)*sizeof(double));

 lhs[0] = Aval;
 lhs[1] = (double *) row;
 lhs[2] = (double *) col;
 lhs[3] = b;
 lhs[4] = (double *) nnz_ptr;
 lhs[5] = boundary;

/* --------------------------  */
/* Indices for loop unrolling  */
/* --------------------------  */

 ind  = 0;
 ind1 = 1;
 ind2 = 2;
 ind3 = 3;
 ind4 = 4;

/*-------------------------------------------*/
/* Main block of inner nodes without BC arms */
/*-------------------------------------------*/
 for( j = N-4; j >= 0; j--)
 {
  offset = 2+(j+1)*(M-1);
  for( i = M-4; i >=0 ;i--)
  {
   node = i+offset-1;

   getsten(i+2,j+2,M,N,time,stencil);


   Aval[ind] = stencil[1];
   Aval[ind1] = stencil[2];
   Aval[ind2] = stencil[0];
   Aval[ind3] = stencil[3];
   Aval[ind4] = stencil[4];

   row[ind] = node;
   row[ind1] = node;
   row[ind2] = node;
   row[ind3] = node;
   row[ind4] = node;

   col[ind] = node - 1;
   col[ind1] = node - (M-1);
   col[ind2] = node;
   col[ind3] = node + 1;
   col[ind4] = node + M-1;

   b[node] = stencil[5];

   ind = ind + 5;
   ind1 = ind1 + 5;
   ind2 = ind2 + 5;
   ind3 = ind3 + 5;
   ind4 = ind4 +5;
  
 }
}

/*---------------------*/
/* Left boundary nodes */
/*---------------------*/
offset = 2-M;
boundary[left++] = BC(0.0,1.0/(double)N, time); 
for( j = 2; j < N-1; j++) {

   node = j*(M-1) + offset - 1;
   getsten(1,j,M,N,time,stencil);

   b[node] = stencil[5];
  
   Aval[ind] = stencil[2];
   Aval[ind1] = stencil[0];
   Aval[ind2] = stencil[3];
   Aval[ind3] = stencil[4];
	
   row[ind] = node;
   row[ind1] = node;
   row[ind2] = node;
   row[ind3] = node;

   col[ind] = node - (M-1);
   col[ind1] = node;
   col[ind2] = node + 1;
   col[ind3] = node + M-1;

   ind = ind +4;
   ind1 = ind1 +4;
   ind2 = ind2 +4;
   ind3 = ind3 +4;
  
   boundary[left] = BC(0.0,(double) j/N, time); 
   b[node] = b[node] - stencil[1]*boundary[left];
   left++;
}
boundary[left] = BC(0.0,(double)(N-1)/(double)N, time); 

/*----------------------*/
/* Right boundary nodes */
/*----------------------*/
boundary[right++] = BC(1.0,1.0/(double)N, time); 
for( j = 2; j < N-1; j++) {

   node = j*(M-1) - 1;
   getsten(M-1,j,M,N,time,stencil);

   b[node] = stencil[5];

   Aval[ind] = stencil[1];
   Aval[ind1] = stencil[2];
   Aval[ind2] = stencil[0];
   Aval[ind3] = stencil[4];

   row[ind] = node;
   row[ind1] = node;
   row[ind2] = node;
   row[ind3] = node;

   col[ind] = node - 1;
   col[ind1] = node - (M-1);
   col[ind2] = node;
   col[ind3] = node + M-1;

   ind = ind +4;
   ind1 = ind1 +4;
   ind2 = ind2 +4;
   ind3 = ind3 +4;

   boundary[right] = BC(1.0,(double) j/N, time);
   b[node] = b[node] - stencil[3]*boundary[right];
   right++;
}
boundary[right] = BC(1.0,(double)(N-1)/(double)N, time); 

/*-----------------------*/
/* Bottom boundary nodes */
/*-----------------------*/
boundary[bottom++] = BC(0.0,0.0, time);
boundary[bottom++] = BC(1.0/(double)M,0.0, time);
for( i = 2; i < M-1; i++) {

   node = i - 1;
   getsten(i,1,M,N,time,stencil);

   b[node] = stencil[5];

   Aval[ind] = stencil[1];
   Aval[ind1] = stencil[0];
   Aval[ind2] = stencil[3];
   Aval[ind3] = stencil[4];

   row[ind] = node;
   row[ind1] = node;
   row[ind2] = node;
   row[ind3] = node;

   col[ind] = node - 1;
   col[ind1] = node;
   col[ind2] = node + 1;
   col[ind3] = node + M-1;
  
   boundary[bottom] = BC((double)i/M,0.0, time);
   b[node] = b[node] - stencil[2]*boundary[bottom];
   bottom++;

   ind = ind +4;
   ind1 = ind1 +4;
   ind2 = ind2 +4;
   ind3 = ind3 +4;
}
boundary[bottom++] = BC((double)(M-1)/(double)M,0.0, time);
boundary[bottom++] = BC(1.0,0.0, time);


/*--------------------*/
/* Top boundary nodes */
/*--------------------*/

boundary[top++] = BC(0.0,1.0, time);
boundary[top++] = BC(1.0/(double)M,1.0, time);
offset = (N-2)* (M-1);
for( i = 2; i < M-1; i++) {

   node = i + offset - 1 ;
   getsten(1,N-1,M,N,time,stencil);

   b[node] = stencil[5];

   Aval[ind] = stencil[1];
   Aval[ind1] = stencil[2];
   Aval[ind2] = stencil[0];
   Aval[ind3] = stencil[3];

   row[ind] = node;
   row[ind1] = node;
   row[ind2] = node;
   row[ind3] = node;

   col[ind] = node - 1;
   col[ind1] = node - (M-1);
   col[ind2] = node;
   col[ind3] = node + 1;

   ind = ind +4;
   ind1 = ind1 +4;
   ind2 = ind2 +4;
   ind3 = ind3 +4;
   boundary[top] = BC((double)i/M, 1.0, time);
   b[node] = b[node] - stencil[4]*boundary[top];
   top++;
 
}
boundary[top++] = BC((double)(M-1)/(double)M,1.0, time);
boundary[top++] = BC(1.0,1.0, time);

/* ------------------- */
/* Corner nodes: (0,0) */
/* ------------------- */

 node = 0;
 getsten(1,1,M,N,time,stencil);

 b[node] = stencil[5];

 Aval[ind] = stencil[0];
 row[ind] = node;
 col[ind] = node;
 ind++;

 b[node] = b[node] - stencil[1]*BC(0.0,1.0/(double)N, time);
 Aval[ind] = stencil[3];
 row[ind] = node;
 col[ind] = node + 1;
 ind++;

 Aval[ind] = stencil[4];
 row[ind] = node;
 col[ind] = node + M-1;
 ind++;

 b[node] = b[node] - stencil[2]*BC(1.0/(double)M,0.0, time);
 
/* --------------------- */
/* Corner nodes: (M-1,0) */
/* --------------------- */

   node = M-2;
   getsten(M-1,1,M,N,time,stencil);
  
   b[node] = stencil[5];

   Aval[ind] = stencil[1];
   row[ind] = node;
   col[ind] = node - 1;
   ind++;

   Aval[ind] = stencil[0];
   row[ind] = node;
   col[ind] = node;
   ind++;

   b[node] = b[node] - stencil[3]*BC(1.0,1.0/(double)N, time);

   Aval[ind] = stencil[4];
   row[ind] = node;
   col[ind] = node + M-1;
   ind++; 

   b[node] = b[node] - stencil[2]*BC(1.0/(double)M,0.0, time);


/* --------------------- */
/* Corner nodes: (0,N-1) */
/* --------------------- */

   node = 1 + (N-2)*(M-1) - 1;

   getsten(1,N-1,M,N,time,stencil);

   b[node] = stencil[5];
 
   Aval[ind] = stencil[2];
   row[ind] = node;
   col[ind] = node - (M-1);
   ind++;

   Aval[ind] = stencil[0];
   row[ind] = node;
   col[ind] = node;
   ind++;

   b[node] = b[node] - stencil[4]*BC(1.0/(double)M,1.0, time);

   Aval[ind] = stencil[3];
   row[ind] = node;
   col[ind] = node + 1;
   ind++;

   b[node] = b[node] - stencil[1]*BC(0,1.0/(double)N, time);

/* ----------------------- */
/* Corner nodes: (M-1,N-1) */
/* ----------------------- */
 
   node = (M-1)*(N-1) - 1;
   getsten(M-1,N-1,M,N,time,stencil);
  
   b[node] = stencil[5];

   Aval[ind] = stencil[1];
   row[ind] = node;
   col[ind] = node - 1;
   ind++;
 
   Aval[ind] = stencil[2];
   row[ind] = node;
   col[ind] = node - (M-1);
   ind++;

   b[node] = b[node] - stencil[4]*BC( 1.0/(double)M,1, time );

   Aval[ind] = stencil[0];
   row[ind] = node;
   col[ind] = node;
   ind++;
   
   b[node] = b[node] - stencil[3]*BC(1,(double)1/N, time);

/*  -----------------------------------------------------*/
/*  Now convert to a CSR format.  First sort to have row */
/*  indices in increasing order.                         */
/*  -----------------------------------------------------*/

   qs2i1d_(row, col, Aval, nnz_ptr, &incdecr);
   n = (M-1)*(N-1);

/*  Begin debugging output for COO format -------------------- */
/*
     printf("---------- Inside setup.c:  ------\n");
     printf("k    row[k]     col[k]\n");
        for (k = 0; k < nnz; k++)
          {
            printf("%d   %d   %d\n", k, row[k], col[k]);
          }
     printf("---------- Outside setup.c:  ------\n");
*/
/*  End debugging output for COO format ----------------------- */

/*  ------------------- */
/*  Set up row pointers */
/*  ------------------- */
   rowptr = (int *) malloc((n+2)*sizeof(int));

   rowptr[0] = 0;
   rowptr[1] = -1;
   thisrow = 1;

   for (k = 0; k < nnz; k++)
     {
       rowptr[thisrow]++;
       if (row[k] == thisrow)
          {thisrow++ ;
          rowptr[thisrow] = rowptr[thisrow-1];}
     };
   rowptr[n] = nnz;

/*  ------------------------------------------------- */
/*  Now sort entries in each row to be col-increasing */
/*  ------------------------------------------------- */
   for (thisrow = 0; thisrow < n; thisrow++)
     {
       ibegin = rowptr[thisrow];
       iend   = rowptr[thisrow+1];
       nzrow  = iend - ibegin;
       rowinc_(&nzrow, &col[ibegin], &Aval[ibegin]);
     }

   free(row);
   lhs[1] = (double *) rowptr;
   return(lhs);
}
