#include <math.h>
#include <stdio.h>
#include <iostream.h>
#include <stdlib.h>
#include <string.h>
#include <orderings.h>

/*******************************************************************
//   This C function takes a sparse linear system, applies the
//   reordering method specified in the parameter object, creating
//   a new sparse system with the given ordering.  It also returns
//   the row and column permutation vectors - which for symmetric
//   orderings, are inverses of each other.
//
//   The data structure for the sparse matrices keeps the
//   Fortran 1-based indexing throughout, because it calls all
//   Fortran routines.
//   The code has QMD, RCM, ND, and SYMQMD (symmetric
//   min degree) reorderings.
//  
//       Randall Bramley
//       Department of Computer Science
//       Indiana University, Bloomington
//       bramley@cs.indiana.edu
//       Sun Nov  7 08:31:21 EST 1999
********************************************************************/

#include <SparseLinearSystem.h>
#include <ReorderParms.h>
#include <Reorder.h>

// ------------------
// Fortran functions:
// ------------------
#include "forts.h"

// ----------
// C routines
// ----------

extern "C"
void reorder(double *A, 
             int *colinds, 
             int *rowptrs,         
             double *A2, 
             int *colinds2, 
             int *rowptrs2,         
             int *n, int *nonzeros, 
             double *b,             
             double *b2,             
             int *order,            
             int *errflg,           
             int *RowPerms, int *ColPerms);

SparseLinearSystem* Reorder(ReorderParms* p,
                       	SparseLinearSystem* OriginalSystem,
                       	int* ErrorCode){

      /* ----------------------------- */
      /* Matrix data structure things. */
      /* ----------------------------- */
      double *A = NULL;         /* Vector of nonzero values in A  */
      int    *colinds = NULL;   /* Vector of column pointers in A */
      int    *rowptrs = NULL;   /* Vector of row indices in A     */

      double *A2 = NULL;        /* Copy of nonzero values in A  */
      int    *colinds2 = NULL;  /* Copy of column pointers in A */
      int    *rowptrs2 = NULL;  /* Copy of row indices in A     */
      SparseLinearSystem* ReorderedSystem;

      int nrow = 0;             /* Number of rows in A            */
      int  nnz = 0;             /* Number of nonzeros in A        */
      size_t len ;              /* Lengths for string copies      */

      double *b = NULL;         /* Right hand side vector         */
      double *b2 = NULL;        /* Permuted hand side vector      */

      /* ----------------------- */
      /* Assorted junk and stuff */
      /* ----------------------- */
      int order = 0;
      int k = 0;
      int errflg = 0;

      /* --------------------------------------------- */
      /* Generic vectors for reordering, permutations: */
      /* --------------------------------------------- */
      int *RowPerms  ;             /* Working vector for row order */
      int *ColPerms  ;             /* Working vector for col order */

      if (p->reorder_method == NONE)   order = 0;
       else if (p->reorder_method == RCM)    order = 2;
       else if (p->reorder_method == QMD)    order = 1;
       else if (p->reorder_method == ND)     order = 3;
       else if (p->reorder_method == SYMQMD) order = 4;
	 else {
              cout << "Unknown reordering method."<< endl;
              *ErrorCode = -1;
              return NULL;
	 }

/*    ---------------------------------------------------- */
/*    Unpack the sparse linear system for use in reorder.c */
/*    ---------------------------------------------------- */
	A       = OriginalSystem->values;
	colinds = OriginalSystem->colinds;
	rowptrs = OriginalSystem->rowptrs;
	nrow    = OriginalSystem->n;
	nnz     = OriginalSystem->nnz; 
	b       = OriginalSystem->rhs; 

/*    -------------------------------- */
/*    Set storage for permuted system. */
/*    -------------------------------- */
      colinds2 = (int *) malloc(nnz*sizeof(int));
      rowptrs2 = (int *) malloc((nrow+1)*sizeof(int));
      b2 = (double *) malloc(nrow*sizeof(double));
      A2 = (double *) malloc(nnz*sizeof(double));
      if (A2 == 0){
         printf("*** Not enough memory available to create matrix copy\n");
         *ErrorCode = -2;
         return NULL;
      }

/*    ---------------------------------------------------- */
/*    Set up auxillary vectors; more may be needed later   */
/*    ---------------------------------------------------- */
      RowPerms  = (int *) malloc(nrow*sizeof(int));
      ColPerms  = (int *) malloc(nrow*sizeof(int));
      if (ColPerms == 0){
         printf("*** Not enough memory available for perm vectors\n");
         *ErrorCode = -3;
         return NULL;
      }

/*    ------------------------------------------------------ */
/*    Call the routine to compute and perform the reordering */
/*    ------------------------------------------------------ */

      reorder(A , colinds , rowptrs ,              
              A2, colinds2, rowptrs2,   
              &nrow, &nnz, 
              b, b2,
              &order,   
              &errflg,   
              RowPerms, ColPerms);

/*    ------------------------------------------------------- */
/*    Create the new Sparse system from its constituent parts */
/*    ------------------------------------------------------- */

      ReorderedSystem = new SparseLinearSystem(nrow, nnz, A2,
					rowptrs2, colinds2);


      len = strlen (OriginalSystem->key);
      ReorderedSystem->key = new char[len+1];
      (void) strcpy(ReorderedSystem->key, OriginalSystem->key);
      ReorderedSystem->keylen = OriginalSystem->keylen;

      len = strlen (OriginalSystem->title);
      ReorderedSystem->title = new char[len+1];
      (void) strcpy(ReorderedSystem->title, OriginalSystem->title);
      ReorderedSystem->titlelen = OriginalSystem->titlelen;

      ReorderedSystem->storageType = OriginalSystem->storageType;
      ReorderedSystem->rhs = b2;
      ReorderedSystem->nparts = OriginalSystem->nparts;
      ReorderedSystem->bstarts = new int[OriginalSystem->nparts];

      for (k = 0; k < OriginalSystem->nparts; k++) 
        (ReorderedSystem->bstarts)[k] = (OriginalSystem->bstarts)[k];

/*    THIS REQUIRES A DEEP COPY! */
/*    DEBUG LATER .............. */

      ReorderedSystem->mungings = OriginalSystem->mungings;

      (ReorderedSystem->mungings).AddReordering(ColPerms);

      *ErrorCode = errflg;
      return ReorderedSystem;
}
