#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include "orderings.h"
#include "forts.h"

/***********************************************************************
//   Apply a matrix reordering to the CSR matrix A. The data structure 
//   for the sparse matrices uses Fortran 1-based indexing throughout,
//   because it calls all Fortran routines.  The array returned has been
//   symmetrically reordered, as well as the right hand side vector b.
//
//   The calling routine will need to reorder the solution vector x 
//   after whatever solution method is applied to the system, so the row
//   and column permutations are returned as well.  They should be 
//   preallocated.
//
//   Currently the code has QMD, RCM, and ND reorderings.
//   A symmetrized version of QMD, called SYMQMD, is available which
//   works on the nonzero structure of A + A' instead of just A.
//
//   Fortran Compatibility:
//   ----------------------
//   Because this routine needs to be called from Fortran as
//   well as C, it requires that the user supply the storage
//   for the reordered matrix.  On return, the CSR data structure
//   (A, colinds, rowptrs) may be deallocated.  If no reordering
//   is specified, (A, colinds, rowptrs) is simply copied over
//   to (A2, colinds2, rowptrs2).  Also, all arguments are passed
//   in by address - including scalars.
//
//   Memory Usage: 
//   ------------- 
//   This function will allocate all arrays needed by the reordering
//   routines, and then free them up on return.  Storage needed depends
//   on the reordering routine called.  However, the calling function is
//   responsible for allocating (A, colinds, rowptrs),
//   (A2, colinds2, rowptrs2), and the row/col perm vectors md_r, md_c.
//
//   External Functions:
//   -------------------
//         gennd.f
//         genrcm.f
//         genqmd.f
//
//         rowinc.f
//         invertp.f
//
//         dperm.f
//         dvperm.f
//         bcopy.c  (System block copy)
//
//   On return, errflg = 0 if all's well.
//              errflg = 1 if memory allocation failed
//              errflg = 2 if an unknown ordering was specified.
//                         In this case, no reordering is done.
//              errflg > 2 if something else bad happened.
//  
//       Randall Bramley
//       Department of Computer Science
//       Indiana University, Bloomington
//       bramley@cs.indiana.edu
//       Sun Nov  7 08:02:18 EST 1999
//
************************************************************************/

int getapat(double *A , int *colinds , int *rowptrs, 
            double *A2, int *colinds2, int *rowptrs2, 
            int n, int nnz); /* Computes structure of A+A' */

void reorder(double *A, 
             int *colinds, 
             int *rowptrs,              /* CSR structure for A */
             double *A2, 
             int *colinds2, 
             int *rowptrs2,             /* CSR structure for A2 */
             int *n, int *nonzeros,     /* Sizes passed in */
             double *b,                 /* RHS vector */
             double *b2,                /* Permuted RHS vector */
             int *order,                /* Job specifier */
             int *errflg,               /* Problem portent */
             int *md_r, int *md_c) {    /* Row/col perms */

      int nrow = *n;             /* Number of rows in A            */
      int  nnz = *nonzeros;      /* Number of nonzeros in A        */

      enum Ordering reorder_method; /* Job specifier */
      int job = 0;
      int ibegin = 0;
      int ilen = 0;
      int i = 0;	       /* Generic loop index */
      size_t length = 0;       /* Size of things to copy in bcopy */
      int  nnzc = 0;
      int  *colinds3 = NULL;   /* Needed only for SYMQMD */

      /* --------------------------------------------- */
      /* Generic vectors for reordering, permutations: */
      /* --------------------------------------------- */
      int nofsub = 0;          /* Needed for qmd */

      /* ------------------ */
      /* Min degree vectors */
      /* ------------------ */
      int *degree;             /* Degree of a node */
      int *marker;             /* Mark if node has been visited */
      int *rchset;             /* Reachable set denotation */
      int *nbrhd ;             /* Neighbor marker vector */
      int *qsize ;             /* Quotient graph marker  */
      int *qlink ;             /* Used for links in quotient graph */
      int *rw    ;             /* Used to pass rowptrs */
      int *cw    ;             /* Used to pass colinds */

      /* ------------------------- */
      /* Nested dissection vectors */
      /* ------------------------- */
      int *mask;               /* Used for masking spikes */
      int *xls;                /* Used for level structure */
      int *ls;                 /* Used for level structure */

/*=================================================================*/
/*==== Start of executable statements.  ===========================*/
/*=================================================================*/

      
/*    --------------------------------- */
/*    Set the ordering method specified */
/*    --------------------------------- */
      *errflg = 0;
      if (*order == 0) reorder_method = NONE;
      if (*order == 1) reorder_method = QMD;
      if (*order == 2) reorder_method = RCM;
      if (*order == 3) reorder_method = ND;
      if (*order == 4) reorder_method = SYMQMD;
      if ((*order < 0) || (*order > 4))
      {
           *errflg = 2;
           printf("\nUnknown reordering method!  Specify\n");
           printf("\n   NONE for no reordering\n");
           printf("   RCM for reverse Cuthill-Mckee\n");
           printf("   QMD for quotient min degree\n");
           printf("    ND for nested dissection\n");
           printf("SYMQMD for nested dissection\n");
           printf("Using no reordering method.\n\n");
           reorder_method = NONE;
      }

/*    ------------------------------------- */
/*    If no reordering, copy to A2 and quit */
/*    ------------------------------------- */
      if ( reorder_method == NONE ) {
         length = (size_t) nnz*sizeof(double);
         bcopy(A, A2, length);
         length = (size_t) nnz*sizeof(int);
         bcopy(colinds, colinds2, length);
         length = (size_t) (nrow+1)*sizeof(int);
         bcopy(rowptrs, rowptrs2, length);
         return;
      }


/*    ------------------------------------------------------------ */
/*    Reordering methods switch.                                   */
/*         reorder_method = NOP ---> no reordering.                */
/*         reorder_method = QMD ---> perform min degree            */
/*         reorder_method = RCM ---> perform reverse Cuthill-Mckee */
/*         reorder_method =  ND ---> perform nested dissection     */
/*    ------------------------------------------------------------ */

      if ( (reorder_method == QMD) || (reorder_method == SYMQMD) ) {

/*        --------------------------  */
/*        Call min degree reordering  */
/*        --------------------------  */

         /* ------------------------------------------------------ */
         /* If symmetrized version needed, get structure of A + A' */
         /* into (A2,colinds2,rowptrs2).  Otherwise, just copy the */
         /* colinds over to colinds2.                              */
         /* ------------------------------------------------------ */

         if (reorder_method == SYMQMD) {
           printf("Calling getapat\n");
           colinds3 = (int *) malloc(2*nnz*sizeof(int));
           if (colinds3 == NULL){
              printf("*** Not enough memory available for reorder function\n");
              printf("*** Failure occured on colinds3                     \n");
              *errflg = 1;
              return;
           }

           nnzc = getapat(A, colinds , rowptrs , 
                         A2, colinds3, rowptrs2, 
                         nrow, nnz);
           printf("Getapat returns nnzc = %d\n", nnzc);
           if (nnzc < 0) { 
              printf("Error in getapat; probably memory\n");
              exit(-1);  /* NOTE: should probably just do copy to A2 */
           }
           if (nnzc == 0) { 
              printf("Error in apmbt\n");
              exit(-1);  /* NOTE: should probably just do copy to A2 */
           }
           rw = rowptrs2;
         }

         else {
            length = (size_t) nnz*sizeof(int);
            bcopy(colinds, colinds2, length);
            rw = rowptrs;
         }

         degree = (int *) malloc(nrow*sizeof(int));
         marker = (int *) malloc(nrow*sizeof(int));
         rchset = (int *) malloc(nrow*sizeof(int));
         nbrhd  = (int *) malloc(nrow*sizeof(int));
         qsize  = (int *) malloc(nrow*sizeof(int));
         qlink  = (int *) malloc(nrow*sizeof(int));
         if (qlink == NULL){
            printf("*** Not enough memory available for reorder function\n");
            *errflg = 1;
            return;
         }

         printf("Calling genqmd ...");
         cw = colinds2;
         if (reorder_method == SYMQMD) cw = colinds3;

         genqmd_(&nrow , rw    ,      cw , md_r,
                 md_c  , degree,  marker ,
                 rchset, nbrhd ,  qsize  ,
                 qlink , &nofsub);

         free(qlink);
         free(qsize);
         free(nbrhd);
         free(rchset);
         free(marker);
         free(degree);
         if (reorder_method == SYMQMD) free(colinds3);
         printf("... completed\n");
      }

      else if (reorder_method == RCM) {
/*       -------------------  */
/*       Call RCM reordering  */
/*       -------------------  */
         printf("Calling genrcm ...");
         length = (size_t) nnz*sizeof(int);
         bcopy(colinds, colinds2, length);
         degree = (int *) malloc(nrow*sizeof(int));
         marker = (int *) malloc(nrow*sizeof(int));
         if (marker == NULL){
            printf("*** Not enough memory available for reorder function\n");
            printf("*** Failure occured on marker vector for RCM        \n");
            *errflg = 1;
            return;
         }
         colinds3 = (int *) malloc(2*nnz*sizeof(int));
         if (colinds3 == NULL){
            printf("*** Not enough memory available for reorder function\n");
            printf("*** Failure occured on colinds3 in RCM              \n");
            *errflg = 1;
            return;
         }
         nnzc = getapat(A, colinds , rowptrs , 
                        A2, colinds3, rowptrs2, 
                        nrow, nnz);
         printf("Getapat returns nnzc = %d\n", nnzc);
         if (nnzc < 0) { 
            printf("Error in getapat; probably memory\n");
            exit(-1);  /* NOTE: should probably just do copy to A2 */
         }
         if (nnzc == 0) { 
            printf("Error in apmbt\n");
            exit(-1);  /* NOTE: should probably just do copy to A2 */
         }
         rw = rowptrs2;
         genrcm_(&nrow ,     rw , colinds3, md_r,
                 degree, marker);                  

         invertp_(&nrow, md_r, md_c);

/*
         printf("\nmdr = [ ...\n");
         for (i = 0; i < nrow; i++) printf("%d\n",md_r[i]);
         printf("];\n");
         printf("\nmdc = [ ...\n");
         for (i = 0; i < nrow; i++) printf("%d\n",md_c[i]);
         printf("];\n");
*/
         
         free(marker);
         free(degree);
         printf("... completed\n");
      }


      else if (reorder_method == ND) {
/*       ------------------  */
/*       Call ND reordering  */
/*       ------------------  */
         printf("Calling gennd ...");
         length = (size_t) nnz*sizeof(int);
         bcopy(colinds, colinds2, length);
         mask = (int *) malloc(nrow*sizeof(int));
         xls  = (int *) malloc(nrow*sizeof(int));
         ls   = (int *) malloc(nrow*sizeof(int));
         gennd_(&nrow , rowptrs, colinds2, mask,
                md_r, xls, ls);                  
         invertp_(&nrow, md_r, md_c);
         free(xls);
         free(ls);
         free(mask);
         printf("... completed\n");
      }

/*    ----------------------------------------------------------------------- */
/*    Permute (A,colinds,rowptrs) into matrix (A2,colinds2, rowptrs2)         */
/*    ----------------------------------------------------------------------- */

      job = 3;
      printf("Permuting A and b ... ");
      dperm_(&nrow , A  , colinds , rowptrs,
                     A2 , colinds2, rowptrs2,
                    md_c, md_c, &job);
      for (i = 0; i < nrow; i++) b2[i] = b[i];
      dvperm_(&nrow, b2, md_c);
      printf("... completed\n");

/*    ----------------------------------- */
/*    ... get rid of extra copy of matrix */
/*    ----------------------------------- */
/*      free(A); A = A2;                  */
/*      free(colinds); colinds = colinds2;*/
/*      free(rowptrs); rowptrs = rowptrs2;*/

/*  -----------------------------------------------------------  */
/*  Sort each row of the matrix to column-increasing order.      */
/*  -----------------------------------------------------------  */
      printf("Sorting rows of A ... ");
      for (i = 0; i < nrow; i++) {
          ibegin = rowptrs2[i];       /* Start index of row i */
          ilen   = rowptrs2[i+1] - ibegin;       /* Length of row i */
          rowinc_( &ilen , &colinds2[ibegin-1], &A2[ibegin-1] );
      }
      printf("... completed\n");

      return;
}
