#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <iostream.h>

/*******************************************************************
//   This C function takes a sparse linear system and scales it
//   according to the method specified in the parameter object.
//   It then creates a new sparse system with the given scaling.
//   The scaling vectors are implicitly returned in mungings.
//
//       Randall Bramley
//       Department of Computer Science
//       Indiana University, Bloomington
//       bramley@cs.indiana.edu
//       Sun Nov  7 12:19:50 EST 1999
********************************************************************/

#include <SparseLinearSystem.h>
#include <ScaleParms.h>
#include <Scale.h>

// ------------------
// Fortran functions:
// ------------------
#include "forts.h"

// ----------
// C routines
// ----------

SparseLinearSystem* Scale(ScaleParms* p,
                       	SparseLinearSystem* OriginalSystem,
                       	int* ErrorCode){

      /* ----------------------------- */
      /* Matrix data structure things. */
      /* ----------------------------- */
      double *A = NULL;         /* Vector of nonzero values in A  */
      int    *colinds = NULL;   /* Vector of column pointers in A */
      int    *rowptrs = NULL;   /* Vector of row indices in A     */

      double *A2 = NULL;        /* Copy of nonzero values in A  */
      int    *colinds2 = NULL;  /* Copy of column pointers in A */
      int    *rowptrs2 = NULL;  /* Copy of row indices in A     */
      SparseLinearSystem* ScaledSystem;

      int nrow = 0;             /* Number of rows in A            */
      int  nnz = 0;             /* Number of nonzeros in A        */
      size_t len ;              /* Lengths for string copies      */

      double *b = NULL;         /* Right hand side vector         */
      double *b2 = NULL;        /* Permuted hand side vector      */

      /* ----------------------- */
      /* Assorted junk and stuff */
      /* ----------------------- */
      int sm = p->scale_method;
      int k = 0;
      int errflg = 0;

      /* --------------------------------------------- */
      /* Generic vectors for scaleing, permutations: */
      /* --------------------------------------------- */
      double* Dl  ;             /* Left scaling vector  */
      double* Dr  ;             /* Right scaling vector */

      if (p->scale_method < 0 || p->scale_method > 4)
	 {
              cout << "Unknown scaling method."<< endl;
              *ErrorCode = -1;
              return NULL;
	 }

/*    -------------------------------------------------- */
/*    Unpack the sparse linear system for use in scala.f */
/*    -------------------------------------------------- */
	A       = OriginalSystem->values;
	colinds = OriginalSystem->colinds;
	rowptrs = OriginalSystem->rowptrs;
	nrow    = OriginalSystem->n;
	nnz     = OriginalSystem->nnz; 
	b       = OriginalSystem->rhs; 

/*    ------------------------------ */
/*    Set storage for scaled system. */
/*    ------------------------------ */
      colinds2 = (int *) malloc(nnz*sizeof(int));
      rowptrs2 = (int *) malloc((nrow+1)*sizeof(int));
      b2 = (double *) malloc(nrow*sizeof(double));
      A2 = (double *) malloc(nnz*sizeof(double));
      if (A2 == 0){
         printf("*** Not enough memory available to create matrix copy\n");
         *ErrorCode = -2;
         return NULL;
      }

/*    -------------------------------------- */
/*    Copy A over to scaled system location. */
/*    -------------------------------------- */
      len = (size_t) nnz*sizeof(double);
      bcopy(A, A2, len);
      len = (size_t) nnz*sizeof(int);
      bcopy(colinds, colinds2, len);
      len = (size_t) (nrow+1)*sizeof(int);
      bcopy(rowptrs, rowptrs2, len);
      len = (size_t) nrow*sizeof(double);
      bcopy(b, b2, len);

/*    ---------------------------------------------------- */
/*    Set up auxillary vectors; more may be needed later   */
/*    ---------------------------------------------------- */
      if ( sm == 1 || sm == 2 || sm == 4 ) {
         Dl  = new double[nrow];
         if (Dl == NULL){
            printf("*** Not enough memory available for scale vectors\n");
            *ErrorCode = -3;
            return NULL;
         }
         for (k = 0; k < nrow; k++) Dl[k] = 1.0;
	}

      if ( sm == 3 || sm == 4 ) {
        Dr = new double[nrow];
        if (Dr == NULL){
           printf("*** Not enough memory available for scale vectors\n");
           *ErrorCode = -3;
           return NULL;
        }
        for (k = 0; k < nrow; k++) Dr[k] = 1.0;
      }  


/*    --------------------------------------------------- */
/*    Call the routine to compute and perform the scaling */
/*    --------------------------------------------------- */
      int one = 1;
      int two = 2;
      int three = 3;
  
      if ( (sm < 0) || (sm > 4) ) sm = 0;
    
      if ( sm == 1 ) 
        scala_(&one, A2, rowptrs2, colinds2, &nrow, b2, Dl);
      if ( (sm == 2) || (sm == 4) ) 
        scala_(&two,A2 ,rowptrs2, colinds2, &nrow, b2, Dl);
      if ( (sm == 3) || (sm == 4) )
        scala_(&three, A2, rowptrs2, colinds2, &nrow, b2, Dr);


/*    ------------------------------------------------------- */
/*    Create the new Sparse system from its constituent parts */
/*    ------------------------------------------------------- */

      ScaledSystem = new SparseLinearSystem(nrow, nnz, A2,
					rowptrs2, colinds2);

      len = strlen (OriginalSystem->key);
      ScaledSystem->key = new char[len+1];
      (void) strcpy(ScaledSystem->key, OriginalSystem->key);
      ScaledSystem->keylen = OriginalSystem->keylen;

      len = strlen (OriginalSystem->title);
      ScaledSystem->title = new char[len+1];
      (void) strcpy(ScaledSystem->title, OriginalSystem->title);
      ScaledSystem->titlelen = OriginalSystem->titlelen;

      ScaledSystem->storageType = OriginalSystem->storageType;
      ScaledSystem->rhs = b2;
      ScaledSystem->nparts = OriginalSystem->nparts;
      ScaledSystem->bstarts = new int[OriginalSystem->nparts];

      for (k = 0; k < OriginalSystem->nparts; k++) 
        (ScaledSystem->bstarts)[k] = (OriginalSystem->bstarts)[k];

/*    THIS REQUIRES A DEEP COPY! */
/*    DEBUG LATER .............. */

      ScaledSystem->mungings = OriginalSystem->mungings;

      (ScaledSystem->mungings).AddScaling(Dr);

      *ErrorCode = errflg;
      return ScaledSystem;
}
