/*
//  --------------------------------------------------------------------------------
//  Class Mungings holds the data from a sequence of reorderings and scalings done
//  by LSA modules.  This is needed so that after such a sequence and a solve of
//  the resulting mangled system, we can retrieve the solution vector to the
//  original system.  A sequence of reorderings, of course, can simply be
//  represented as the composition of those reorderings, requiring just a single
//  integer vector to store.  Similarly a sequence of scalings can be combined into
//  a single double array.  The complexity in Mungings comes because it is possible
//  to feed the linear system through an arbitrary, possibly alternating, sequence
//  of reorderings and scalings.  Each time a new munging is added, we may have to
//  allocate another vector.
//  
//  Mungings does not need to, but does keep track of the total numbers of
//  scalings (nscalings) and of reorderings (nreorderings).  
//  
//  Two-dimensional arrays called reorderings and scalings are used to store
//  each of the munging flavors.  Each reordering or scaling vector is allocated
//  dynamically as needed, but the pointers to pointers start out with 8 each.
//  It is unlikely anyone will require that many of either, but this saves us from
//  costly reallocation on every added munging vector.  The integers
//  allocated_reorderings and allocated_scalings indicate how many pointers have
//  been allocated - if they have to be increased, they are done so in increments
//  of 8 also.  
//  
//  To keep track of the sequencing of the operations (which must be undone in
//  precisely reverse order for retrieving the true solution vector) an integer
//  lastoper is used.  It indexes into the munging vector arrays as:
//  	lastoper < 0:  means scaling was the last munging done, and
//  					scaling[-lastoper] was the last one done.
//  	lastoper > 0:  means reordering was the last munging done, and
//  					reordering[-lastoper] was the last one done.
//  	lastoper = 0:  means no mungings have been done.
//  Because we compose munging operations whenever possible, this suffices to
//  indicate what sequence is needed; scalings and reorderings have to alternate in
//  the unmunging operation.
//  
//  Randall Bramley
//  Department of Computer Science
//  Indiana University, Bloomington
//  bramley@cs.indiana.edu                             
//  Sun Oct 17 10:22:01 EST 1999
//  --------------------------------------------------------------------------------
*/

#ifndef MUNGINGS_H
#define MUNGINGS_H

#ifdef REMOVE_
#define dvperm_ dvperm
#define ivperm_ ivperm
#define unscalx_ unscalx
#define daxpy_ daxpy
#endif

// ------------------
// Fortran functions:
// ------------------
extern "C" {
  void dvperm_(int* n, double* x, int* perm);
  void ivperm_(int* n, int* ix, int* perm);
  void unscalx_(int* n, double* x, double* scaling);
  void daxpy_(int* n, double*alpha, double* x, int* incx,
                                    double* y, int* incy);
};


class Mungings{
public:
      // ---------
      // LIFECYCLE
      // ---------
      Mungings();
      Mungings( const Mungings& m);
      ~Mungings();
 
      // ----------
      // OPERATIONS
      // ----------

	  int AddReordering(int *newreordering);// Add in a new reordering
	  int AddScaling(double *newscaling);	// Add in a new scaling
	  int UnMunge(double *x);		// Retrieve the solution
	  					//  by unapplying all of the
	  					//  scalings/reorderings in
	  					//  reverse order.  Does this
	  					//  in-place with vector x.
	  int getNumMungings();			// How many mungings done?
	  int getNumReorderings();		// How many reorderings done?
	  int getNumScalings();			// How many scalings done?
	  int getBytes();			// Storage currently used.

	  // -----------
	  // DATA FIELDS
	  // -----------
	  int      n;				// Length of reorder/scale vecs
	  int      nreorderings;		// Total number of reorderings
	  int      nscalings;			// Total number of scalings

	  int      nreordering_vectors;		// Number of reordering vectors; will
						//   be less than nreorderings if any
						//   of the reorderings were
						//   consecutive
	  int      nscaling_vectors;		// Number of scaling vectors; will
						//   be less than nscalings if any
						//   of the scalings were consecutive

	  int      allocated_reorderings;	// Number of allocated pointers
	  int      allocated_scalings;		// Number of allocated pointers
	  int      lastoper;
	  
	  int**    reorderings;
	  double** scalings;
};
#endif
