#include <fstream.h>
#include <Splib.h>
#include <SparseLinearSystem.h>
#include <Mungings.h>

// ------------------------------------------------------------
// Check the arguments for Splib.  Very basic error checking is
// done here, but that will OK for the most part since SPLIB
// fixes most incorrect control parameters.  More deadly is
// making sure that arrays have been allocated.
// ------------------------------------------------------------
int CheckArguments(SplibParms& parms, SparseLinearSystem* sls) {

  if ( parms.Scaling < 0 || parms.Scaling > 4) {
    return 2;
  }
  if ( parms.relative != 0 && parms.relative != 1) {
    return 3;
  }
  if ( sls->n < 1 ) {
    return 4;
  }

// Following causes problems; n*n exceeds 32 bits for n > 32k
//  if ( sls->nnz < sls->n || sls->nnz > (sls->n)*(sls->n)) {
//    return 5;
//  }

  if ( sls->values == NULL || sls->colinds == NULL
       || sls->rowptrs == NULL) {
    return 6;
  }
  if ( sls->rhs == NULL) {
    return 7;
  }

  if (sls->nnz != sls->rowptrs[sls->n] - 1) {
	return 8;
  }
  
  return 0;
}

int Splib(SplibParms& parms, SparseLinearSystem* sls) {

  int ErrorCode = CheckArguments(parms,sls);

  switch (ErrorCode) {
  case 0:
    break;
  case 1:
    cout << "Pre-Splib Error: Wrong number/kind of control parameters." << endl;
    break;
  case 2:
    cout << "Pre-Splib Error: Scaling parameter out of range." << endl;
    break;
  case 3:
    cout << "Pre-Splib Error: Relative parameter out of range." << endl;
    break;    
  case 4:
    cout << "Pre-Splib Error: Number of rows in matrix is zero or less." << endl;
    break;
  case 5:
    cout << "Pre-Splib Error: Number of non-zeros in matrix is out of range." << endl;
    break;
  case 6:
    cout << "Pre-Splib Error: Matrix is NULL." << endl;
    break;
  case 7:
    cout << "Pre-Splib Error: Right-hand side vector is NULL" << endl;
    break;   
  case 8:
    cout << "Pre-Splib Error: nnz is wrong" << " nnz: " << sls->nnz << " sls expr: " << sls->rowptrs[sls->n] - 1 << endl;
    break;   
  default:
    cout << "Unknown Pre-Splib error code occurred." << endl;
    break;
  }    

  double* Dl = NULL;       /* Left diagonal scaling vector */
  double* Dr = NULL;       /* Right diagonal scaling vector */
  double rparms[2] = {0,0};  /* Real parameters for preconditioners */
  double* work = NULL;     /* Work vector of space for SPLIB */
  int iparms[2] = {0,20};  /* Integer parameters */
  int errflg = 0;          /* Error flag for SPLIB */
  int i = 0, nrow;
  int iredo;
  int AvailSpace = parms.AllocatedSpace;
 
  enum { NoPrecond = 0, ILU, MILU, ILUT, SSOR, TRID, ILU0, ECIMGS };
  enum { BiCG = 1, CGNE, CGNR, CGSQ, CGSTAB, GMRES, QMR, TCGSTAB, TGMRES,
         Jacobi, GaussSeidel, SOR, Orthomin };

  nrow = sls->n;

  // --------------------------------------------------------------
  // Can't change OutputLevel after the first call to this function 
  // --------------------------------------------------------------
  
  if ( (parms.SolveMethod == GMRES) || 
       (parms.SolveMethod == TGMRES) || 
       (parms.SolveMethod == Orthomin) )
    iparms[1] = parms.KrylovVectors;
    
  if ( parms.SolveMethod == SOR )
    rparms[0] = parms.SORRelaxParm;

  if ( (parms.SolveMethod==Jacobi) || 
       (parms.SolveMethod==GaussSeidel) || 
       (parms.SolveMethod==SOR) ) {
    // cannot precondition Jacobi, GS, or SOR
    parms.PrecondMeth = 0;
    iredo = 0;
  }
  else {
    // whether to recompute preconditioner, or use earlier one
    iredo = parms.RecomputePrecond;
  }

  if ( (parms.PrecondMeth == ILU) || (parms.PrecondMeth == MILU) ) 
    // Have possibly nonzero levels of fill for these preconditioners
    iparms[0] = parms.FillLevels;

  /* -------------------------------------------- */
  /* MILU and SSOR require a relaxation parameter */
  /* -------------------------------------------- */

  if ( (parms.PrecondMeth == MILU) || (parms.PrecondMeth == SSOR) ) 
    rparms[0] = parms.MILUParam;

  /* --------------------------------------------------------------- */
  /* ILUT requires an max additional entries per row parameter and a */
  /* drop tolerance                                                  */
  /* --------------------------------------------------------------- */

  if (parms.PrecondMeth == ILUT) {
    iparms[0] = parms.rows;
    rparms[0] = parms.DropTol;
  }

  /* ----------------------------------- */
  /* TRID requires a blocksize parameter */
  /* ----------------------------------- */

  if (parms.PrecondMeth == TRID) 
    iparms[0] = parms.BlockSize;

  if (parms.PrecondMeth == ECIMGS) 
    rparms[0] = parms.SparsityTarget;
  
  /* ---------------------------------- */
  /* set up data structure for solution */
  /* ---------------------------------- */

  if (sls->x == NULL) {
	printf("Allocating new solution of %d rows.\n", nrow);
	sls->x = new double[nrow];
  }
  if (sls->x == NULL) {
    ErrorCode =1; /* can't allocate memory for solution */
    return -1;
  }

  // ----------------------------------------------
  // Initialize to zero values, to avoid accidental
  // over/under flows from having garabage values
  // ----------------------------------------------

  for (i = 0; i < nrow; i++) {
    sls->x[i] = 0.0; 
  }

  // --------------------------------------------------------------
  // Allocate work space; first try specified amount, then back off
  // by halving each time until succeed or have less than 20n words
  // --------------------------------------------------------------
  
  work = (double *) malloc((AvailSpace/2)*sizeof(double));
  while ((work == NULL) && (AvailSpace > 20*nrow)) {
    AvailSpace = AvailSpace/2;
    work = (double *) malloc((AvailSpace/2)*sizeof(double));
  }
  
  if (work == NULL) {
    ErrorCode = 2; /* Can't allocate work space */
    return -1;
  }
  
  if (parms.Scaling> 0) {
    Dl = (double *) malloc(nrow*sizeof(double));
    Dr = (double *) malloc(nrow*sizeof(double));
  
    if (Dr == NULL || Dl == NULL) {
      ErrorCode = 3; /* Can't allocate scaling vectors */
      free(work);
      if (Dl!=NULL) free(Dl);
      if (Dr != NULL) free(Dr);
      return -1;
    }
  }

  openunits_(parms.iunit); 
  parms.UnitsOpen = 1;
  // On entry to splib, gives max allowed its.
  // On return, has actual number exectuted.
  int ActualIterations = parms.MaximumIterations;

//   cout << "Splib.C values " << sls->values << endl;
//   cout << "Splib.C colinds " << sls->colinds << endl;
//   cout << "Splib.C rowptrs " << sls->rowptrs << endl;
//   cout << "Splib.C n " << sls->n << endl;
//   cout << "Splib.C x " << sls->x << endl;
//   cout << "Splib.C rhs " << sls->rhs << endl;
//   cout << "Splib.C work " << work << endl;
//   cout << "Splib.C availSpace " << AvailSpace << endl;
  cout << "Splib.C PreCondMeth " << parms.PrecondMeth << endl;
  cout << "Splib.C SolveMethod " << parms.SolveMethod << endl;
  cout << "Splib.C Scaling " << parms.Scaling << endl;
  cout.flush();

  splib_(sls->values,
         sls->colinds,
         sls->rowptrs, 
         &(sls->n),           /* CSR sparse matrix data structure */
	 sls->x,              /* Solution vector on return (doubles) */
	 sls->rhs,            /* Right hand side vector (doubles) */
	 work, &AvailSpace,   /* Large vector of double-aligned work
			      storage for preconditioners, etc. */
	 &parms.Tolerance,   /* Tolerance for stopping test; usually
			      want this around 1.0e-6  */
	 &parms.relative,    /* Integer with value 1 to indicate using
			      relative residual stopping test */
	 &ActualIterations,/* Integer giving maximum number of 
			      iterations.  Don't make it too big */
	 &parms.PrecondMeth, /* Preconditioning method (integer)  */
	 &parms.SolveMethod, /* Iterative solver choice (integer) */
	 &parms.Scaling,     /* Integer specifying scaling strategy */ 
	 Dl, Dr,           /* Double vectors of length n giving
			       the diagonal scaling factors.      */
	 iparms,           /* Vector of length 2, giving parameters
			       for preconditioner, Krylov subspace size */
	 rparms,           /* Double vector of length 1, giving
			       relaxation parameter for SOR, etc. */
	 &iredo,           /* Integer specifying whether or not to
			       recompute the preconditioner */
	 &errflg,          /* Return integer telling what went wrong */
	 parms.iunit,        /* Vector of length 20, for I/O unit numbers
			        used by SPLIB */
	 &parms.OutputLevel  /* Integer specifying the instrumentation
			     level; amount of informative output and
			     Xgraph files of output */
	 );
  flushunits_(parms.iunit); 
  
  free(work);
  free(Dl);
  free(Dr);
  ErrorCode = errflg;

  return ErrorCode;

}
