#include "SplibComponent.h"
#include "Splib.h"
#include "SparseLinearSystem_impl.h"
#include "SplibParms_impl.h"
#include "../Common/SolutionVector.h"

#include "../../../../framework/ccacore/Port.h"
#include "../../../../framework/ccacore/CCATServicesProxy.h"
#include "../../../../framework/ccacore/RegisterType.h"
#include "../../../../framework/ccacore/FRAMEWORK_REGISTRATION_ID.h"

#include "../../../../framework/idl/solutionVector/UsesSolutionVector.h"

SplibComponent::SplibComponent() {

    int ec = -1;

    m_core = NULL;
    m_sparseLinearSystemPort = new SparseLinearSystem_impl(this);
    m_splibParmsPort = new SplibParms_impl(this);
    m_computing = false;
    m_stopped = false;

    m_parms_cache = new SplibParms;

    /*
     * Init mutex
     */

    pthread_mutexattr_t attr;

    pthread_mutexattr_init(&attr);

    #if defined(DEBUG) && !defined(__linux__)
        ec = pthread_mutexattr_settype(&attr, PTHREAD_MUTEX_ERRORCHECK);
        assert(ec == 0);
    #endif

    pthread_mutex_init(&m_lock, &attr);

    pthread_mutexattr_destroy(&attr);

    /*
     * Init condition variable.
     */

    pthread_cond_init(&m_cv, NULL);

    /*
     * Init pthread attr
     */

    ec = pthread_attr_init(&m_threadAttr);
    assert(ec == 0);
    ec = pthread_attr_setdetachstate(&m_threadAttr, PTHREAD_CREATE_DETACHED);
    assert(ec == 0);
    //ec = pthread_attr_setscope(&m_threadAttr, PTHREAD_SCOPE_SYSTEM);
    //assert(ec == 0);
}

SplibComponent::~SplibComponent() {

    int ec = -1;

    if (m_core != NULL) {
        delete m_core;
    }

    delete m_sparseLinearSystemPort;
    delete m_splibParmsPort;

    ec = pthread_attr_destroy(&m_threadAttr);
    assert(ec == 0);

    ec = pthread_cond_destroy(&m_cv);
    assert(ec == 0);

    ec = pthread_mutex_destroy(&m_lock);
    assert(ec == 0);
}

void
SplibComponent::setServices(Services *cc) {

    if (cc == NULL) {

        assert(m_core != NULL);
        delete m_core;
        m_core = NULL;
        return;

    } else {

        m_core = cc;

        m_core->addProvidesPort(m_sparseLinearSystemPort,
         new DefaultPortInfo("inputSLS", "SparseLinearSystem_idl"));
        m_core->addProvidesPort(m_splibParmsPort,
         new DefaultPortInfo("inputSplibParms", "SplibParms_idl"));
        m_core->registerUsesPort(new DefaultPortInfo("outputSV",
         "SolutionVector_idl"));
    }
}

struct Info {
    Info(SplibComponent *cop, SparseLinearSystem *sls_, SplibParms *ps)
     : component(cop), sls(sls_), parms(ps) {}
    SplibComponent *const component;
    SparseLinearSystem *const sls;
    SplibParms *const parms;
};

void *
SplibComponent::compute_wrapper(void *vp) {

    Info *ip = static_cast<Info *>(vp);
    Info info = *ip;
    delete ip;

    info.component->compute(info.sls, info.parms);

    return NULL;
}

void
SplibComponent::startComputation(SparseLinearSystem *sls, SplibParms *sp) {

    pthread_t tid;

    Info *info = new Info(this, sls, sp);
    assert(info != NULL);

    pthread_create(&tid, &m_threadAttr, compute_wrapper, info);
}

void
SplibComponent::compute(SparseLinearSystem *sls, SplibParms *parms) {
const char *xfn = "SplibComponent::compute";

    cout << xfn << " called." << endl;
  
    assert(sls != NULL);
    assert(parms != NULL);

    int ec = Splib(*parms, sls);
    assert(ec == 0);

    //(sls->mungings).UnMunge(sls->x);

    SolutionVector sv(sls->n, sls->x);

    /*
    cout << xfn << " solution Vector is " << endl;
    for (int i =0; i<sls->n; i++) {
        cout << (sls->x)[i]  << endl;
    }
    cout.flush();
    */

    UsesSolutionVector *up = dynamic_cast<UsesSolutionVector *>(m_core
     ->getPort("outputSV"));
    up->sendSolutionVector(&sv);
    m_core->releasePort("outputSV");
    cout << xfn << " solution output." << endl;

    pthread_mutex_lock(&m_lock);
    m_computing = false;
    pthread_cond_broadcast(&m_cv);
    pthread_mutex_unlock(&m_lock);

    delete sls;
    /*
	 delete parms;
    */
}
