c======================================================================== 
      subroutine ilu0(n,a,colind,rwptr,relax,pivots,diag,ierr)

c**********************************************************************
c                   SPLIB Copyright (C) 1995-1999                    **
c                     Indiana University                             **
c**********************************************************************

c======================================================================== 
c                    *** ILU0 ***					*
c======================================================================== 
c                                                                     	*
c      Incomplete LU factorization for 0 levels of fill-in.	      	*
c      This is the version that does not modify the offdiagonal		*
c      nonzeros of A, so only the pivot elements need be computed 	*
c      and stored.  This corresponds to "-1" levels of fill-in, 	*
c      rather than the "0" levels normally referred to.  Note that	*
c      a great deal of confusion exists in the literature about what	*
c      an ILU(0) preconditioner is .  Here, the preconditioner  	*
c      does not allow modification of an off-diagonal entry, even    	*
c      it is a nonzero in the original matrix A.                     	*
c									*
c      For matrices from quasi-uniform regular meshes, the resulting	*
c      preconditioner should be the same as ILU(0).			*
c                                                                     	*
c      The implicit splitting is                                      	*
c          LU = (D + LA) (D)^-1 (D + UA)                            	*
c             = (I + LA(D)^-1) (D + UA)                            	*
c             = (D + LA) (I + (D)^-1 UA)                            	*
c       where LA, UA are the strictly lower and 			*
c       strictly upper triangular parts of A.                         	*
c                                                                     	*
c  NOTE: This code will sort the entries in each row of A to be		*
c        in column increasing order.	                         	*
c  NOTE: A tolerance parameter SMALL is used for handling small       	*
c        pivots; this replaces a diagonal entry in some cases.        	*
c        SMALL is set to machine tolerance.				*
c                                                                     	*
c-----------------------------------------------------------------------*
c									*
c     Randall Bramley						      	*
c     Department of Computer Science				      	*
c     Indiana University					      	*
c     email: bramley@cs.indiana.edu				      	*
c     Mon May 30 07:36:14 EST 1994					*
c									*
c======================================================================== 
c
c parameters
c-----------
c
c on entry:
c========== 
c n       = integer. the dimension of the matrix a.
c
c a,colind,rwptr = matrix stored in compressed sparse row format,
c           each row ordered in increasing column order.
c
c relax   = relaxation parameter.  Gives the multiplier to use before
c           adding discarded off-diagonal modification to the diagonal.
c           Straightforward ILU(-1) uses relax = 1.0d0.
c
c on return:
c=========== 
c
c pivots  = vector of length n of pivot elements
c                                                                        
c  ierr is an error flag:
c        ierr  = -i --> near zero pivot in step i
c        ierr  = 0  --> all's OK
c        ierr  = 1  --> not enough storage; check mneed.
c        ierr  = 2  --> illegal parameter
c
c Work arrays:
c=============
c
c  diag    = Indices of diagonal entries of A in a, colind, rwptr data structure.
c
c Storage requirements:
c======================
c
c   diag:	n 		integers.
c   pivots:     n 		reals.
c
c  Warnings:
c===========
c
c 1) The CSR data structure will be row ordered in increasing column
c    order.  Currently this is done by calls to rowinc, which can mess
c    up any reorderings the user is keeping around.
c
c 2) The diagonal entries of A are assumed to be nonzero., and in particular
c    to have storage provided for them in the CSR data structure.
c
c============================================================================ 
      implicit none

      integer   n, ierr, rwptr(n+1), colind(*), diag(n)
      integer   i, j, k, ie, is
      logical found
      real*8    pivots(n), a(*), relax, subdiag
      real*8 SMALL, dlamch

c     -------------------------------------------------
c     Set divide by zero tolerance to machine tolerance
c     -------------------------------------------------
      SMALL = sqrt(dlamch('E'))
      ierr  =  0

c     -------------------------------------------------------
c     Sort the matrix to have rows in increasing column order
c     -------------------------------------------------------
      do k = 1, n
         is = rwptr(k)
         ie = rwptr(k+1)
         call rowinc( ie - is, colind(is), a(is) )
      end do

c     ---------------------------------------------
c     Get the diagonal entry indices in array diag,
c     and get the diagonal entries in array pivots
c     ---------------------------------------------
      do 80 i = 1, n
         do 40 j = rwptr(i), rwptr(i+1)-1
             if (colind(j) .eq. i) then
                 pivots(i) = a(j)
                 diag(i) = j
                 go to 80
             end if 
 40      continue

c         ------------------------------------------------------------
c         If we make it to here, then the diagonal entry of A is not
c         part of the CSR data structure.  This is an error condition.
c         ------------------------------------------------------------
         ierr = -i
         return
 80   continue

c     -------------------------
c     Perform the factorization
c     -------------------------
      do 140 i = 1, n

c        ------------------------------------------------------
c        Invert the i-th pivot element, to avoid extra divides.
c        ------------------------------------------------------
         if (abs(pivots(i)) .le. SMALL) then
            ierr = -i
            pivots(i) = sign(SMALL,pivots(i))
         end if
         pivots(i) = 1.0d0/pivots(i)
c        ----------------------------------------------------------
c        For each nonzero A(i,j) in upper triangular part of row i,
c        update the corresponding diagonal entry A(j,j) by
c            A(j,j) = A(j,j) - A(i,j)*A(j,i)/A(i,i)
c        ----------------------------------------------------------
         do 120 j = diag(i)+1, rwptr(i+1)-1

c        -----------------------------------------------
c        Find the subdiagonal entry A(j,i), it it exists
c        -----------------------------------------------
            found = .false.
            do 100 k = rwptr(colind(j)), diag(colind(j))-1
               if (colind(k) .eq. i) then
                  found = .true.
                  subdiag = a(k)
               end if 
 100        continue
         if (found) pivots(colind(j)) = pivots(colind(j))
c     1              - subdiag * a(j) * pivots(i)
     1              - relax * subdiag * a(j) * pivots(i)
 120     continue
 140  continue

      return

c====================  End of ILU0 ==================================== 

      end
