c========================================================================
      subroutine tridiag(n,a,colind,rwptr,quad,ipvt,iskip,ierr)
c========================================================================

c**********************************************************************
c                   SPLIB Copyright (C) 1995-1999                    **
c                     Indiana University                             **
c**********************************************************************

c                 *** Tridiagonal preconditioner ***			*
c----------------------------------------------------------------------== 
c									*
c  This implements block diagonal preconditioning, where		*
c  each diagonal block is from the corresponding tridiagonal part	*
c  of the matrix A.  The size of the blocks is given implicitly in	*
c  the integer iskip, stored in jlu(n).  Factors are set up for		*
c  contiguous blocks of iskip rows and so by using iskip = n the	*
c  factor can be for the entire matrix's tridiagonal part.		*
c  Using iskip = 1 is an inefficient way of performing diagonal		*
c  preconditioning, while 1 < iskip < n is for a block tridiagonal	*
c  preconditioner.  Gaussian elimination with pivoting is used for	*
c  factoring each block, so it is necessary to store the pivot		*
c  sequence in jlu(1:n-1).  quad has four vectors giving the nonzeros	*
c  of the tridiagonal preconditioner.  The method performs Gaussian	*
c  elimination with partial pivoting on the tridiagonal part of the	*
c  matrix; this is used as a preconditioner via routines tridsl.f and	*
c  trifac.f.								*
c									*
c  This is generally not a very good preconditioner!			*
c									*
c----------------------------------------------------------------------==
c
c     Randall Bramley
c     Department of Computer Science
c     Indiana University
c     email: bramley@cs.indiana.edu
c     Tue Jan 18 11:14:46 EST 1994
c
c----------------------------------------------------------------------==
c
c parameters
c-----------
c
c on entry:
c========== 
c n       = integer. the dimension of the matrix a.
c
c a,colind,rwptr = matrix stored in compressed sparse row format.
c
c iskip   = integer specifying the block size to use for a block
c           diagonal preconditioner, with each block a tridiagonal matrix.
c
c on return:
c=========== 
c
c quad    = matrix storing the four diagonal needed for the LU factors
c           of the tridiagonal part of the matrix A.  quad must be n x 4
c           double precision.
c
c ierr is an error flag:
c        ierr  = -i --> near zero pivot in step i
c        ierr  = 0  --> all's OK
c        ierr  = 1  --> not enough storage; check mneed.
c        ierr  = 2  --> illegal parameter
c
c work arrays:
c=============
c
c ipvt     = integer work array of length >= n.
c            The integer iskip is placed in ipvt(n), for passing to the
c            preconditioner application programs.
c
c----------------------------------------------------------------------- 

      implicit none
      integer   n, ipvt(n), ierr, rwptr(n+1), colind(*),k,kk
      real*8    a(*), quad(n,*)
      integer   length, iskip, job, ndz, ioff, blksz

c=======================================================================
c     Start of Executable Statements
c=======================================================================

      ierr  =  0
      if (iskip .le. 0) ierr = 2
      if (n .le. 0) ierr = 2
      if (ierr .ne. 0) return
      blksz = min(n,iskip)

c     -----------------------------------------------------------------
c     Extract the central tridiagonal part of A:
c     Mapping to quad: quad(*,1) --> first subdiagonal
c                      quad(*,2) --> main diagonal
c                      quad(*,3) --> first superdiagonal
c                      quad(*,4) --> second superdiagonal (for fill-in)
c     -----------------------------------------------------------------
      job = 0
      ioff = -1
      call getdia(n,n,job,a,colind,rwptr,ndz,quad(1,1),ipvt,ioff)
      ioff = 0
      call getdia(n,n,job,a,colind,rwptr,ndz,quad(1,2),ipvt,ioff)
      ioff = 1
      call getdia(n,n,job,a,colind,rwptr,ndz,quad(1,3),ipvt,ioff)

c     ----------------------------------------------------------------
c     Perform the LU factorization.  Note that because getdia extracts
c     the diagonal assuming they start at row number 1,
c     the first subdiagonal     begins in quad(2,1)
c     the main diagonal         begins in quad(1,2)
c     the first superdiagonal   begins in quad(1,3)
c     ----------------------------------------------------------------

      do 100 k = 1, n, blksz

c        ------------------------------------------
c        Length of the tridiagonal matrix to factor
c        ------------------------------------------
         length = min(blksz,n-k+1)

c        --------------------------------------------------
c        Offset into the quad array for this block to start
c        --------------------------------------------------
         ioff = k

c        --------------------
c        Do the factorization
c        --------------------
         call trifac(length,quad(1+ioff,1),
     &                      quad(ioff,2),
     &                      quad(ioff,3),
     &                      quad(ioff,4),
     &                      ipvt(ioff),ierr)

c        ----------------
c        Check error code
c        ----------------
         if (ierr.ne. 0) then
            write(6,*) 'Error condition in trifac on block beginning'
            write(6,*) '       at position ',k
            write(6,*) '       Occurs in tridiag.f'
            write(6,*) '       Error code = ',ierr
            write(6,*) '       Setting block to identity'
            do kk = ioff,ioff+length-1
               quad(kk,1) = 0.0d0
               quad(kk,2) = 1.0d0
               quad(kk,3) = 0.0d0
               quad(kk,4) = 0.0d0
               ipvt(kk) = kk-ioff+1
            end do
            ierr = 0
         end if
 100     continue
         ipvt(n) = blksz


 1000  format(128(2(4(d16.8,3x),/),/))
 1010  format(128(8(i5,3x),/))
      return
      end
c
c
c========================================================================
c
c
      subroutine trifac( n, dl, d, du, du2, ipiv, info )
c========================================================================
c       **** Factor a tridiagonal matrix, with pivoting ***		*
c========================================================================
c
c     Factor a linear system Ax = b, where A is tridiagonal and has
c     been factored by trifac.  This routine is modified from a Lapack
c     routine (dgttrf.f).  The factorization uses partial pivoting, so
c     an extra vector of storage is needed.  Furthermore, the factors
c     overwrite the input matrix.  This routine is from LAPACK, with
c     some minor changes.
c
c     Randall Bramley
c     Department of Computer Science
c     Indiana University
c     email: bramley@cs.indiana.edu
c     Tue Jan 18 10:01:23 EST 1994
c
c  -- original lapack routine (version 1.1) --
c     univ. of tennessee, univ. of california berkeley, nag ltd.,
c     courant institute, argonne national lab, and rice university
c     march 31, 1993
c
c========================================================================
c
c
c  Scalar arguments
      implicit none
      integer            info, n
c  Array arguments 
      integer            ipiv( * )
      double precision   d( * ), dl( * ), du( * ), du2( * )
c
c  purpose
c  =======
c
c  trifac computes an lu factorization of a real tridiagonal matrix a
c  using elimination with partial pivoting and row interchanges.
c
c  the factorization has the form
c     a = l * u
c  where l is a product of permutation and unit lower bidiagonal
c  matrices and u is upper triangular with nonzeros in only the main
c  diagonal and first two superdiagonals.
c
c  arguments
c  =========
c
c  n       (input) integer
c          the order of the matrix a.
c
c  dl      (input/output) double precision array, dimension (n-1)
c          on entry, dl must contain the (n-1) subdiagonal elements of
c          a.
c          on exit, dl is overwritten by the (n-1) multipliers that
c          define the matrix l from the lu factorization of a.
c
c  d       (input/output) double precision array, dimension (n)
c          on entry, d must contain the diagonal elements of a.
c          on exit, d is overwritten by the n diagonal elements of the
c          upper triangular matrix u from the lu factorization of a.
c
c  du      (input/output) double precision array, dimension (n-1)
c          on entry, du must contain the (n-1) superdiagonal elements
c          of a.
c          on exit, du is overwritten by the (n-1) elements of the first
c          superdiagonal of u.
c
c  du2     (output) double precision array, dimension (n-2)
c          on exit, du2 is overwritten by the (n-2) elements of the
c          second superdiagonal of u.
c
c  ipiv    (output) integer array, dimension (n)
c          the pivot indices; for 1 <= i <= n, row i of the matrix was
c          interchanged with row ipiv(i).  ipiv(i) will always be either
c          i or i+1; ipiv(i) = i indicates a row interchange was not
c          required.
c
c  info    (output) integer
c          = 0:  successful exit
c          < 0:  if info = -i, the i-th argument had an illegal value
c          > 0:  if info = i, u(i,i) is exactly zero. the factorization
c                has been completed, but the factor u is exactly
c                singular, and division by zero will occur if it is used
c                to solve a system of equations.
c
c========================================================================
c========================================================================
c
c  Local scalars:
      integer   i
      double precision   fact, temp
c  Intrinsic functions:
      intrinsic abs
c  Parameters:
      double precision   zero
      parameter ( zero = 0.0d+0 )
c
c========================================================================
c
      info = 0
      if( n.lt.0 ) then
         info = -1
         write(6,*) 'Error in trifac; info = ',info
         return
      end if
c
c     quick return if possible
c
      if( n.eq.0 ) return
c
c     initialize ipiv(i) = i
c
      do 10 i = 1, n
         ipiv( i ) = i
   10 continue
c
      do 20 i = 1, n - 1
         if( dl( i ).eq.zero ) then
c
c           subdiagonal is zero, no elimination is required.
c
            if( d( i ).eq.zero .and. info.eq.0 ) info = i
C            if( i.lt.n-1 ) du2( i ) = zero
         else if( abs( d( i ) ).ge.abs( dl( i ) ) ) then
c
c           no row interchange required, eliminate dl(i)
c
            fact = dl( i ) / d( i )
            dl( i ) = fact
            d( i+1 ) = d( i+1 ) - fact*du( i )
C            if( i.lt.n-1 ) du2( i ) = zero
         else
c
c           interchange rows i and i+1, eliminate dl(i)
c
            fact = d( i ) / dl( i )
            d( i ) = dl( i )
            dl( i ) = fact
            temp = du( i )
            du( i ) = d( i+1 )
            d( i+1 ) = temp - fact*d( i+1 )
            if( i.lt.n-1 ) then
               du2( i ) = du( i+1 )
               du( i+1 ) = -fact*du( i+1 )
            end if
            ipiv( i ) = ipiv( i ) + 1
         end if
   20 continue
c
      if( d( n ).eq.zero .and. info.eq.0 ) then
         info = n
         return
      end if
c
      return
c
c  ---------------> End of trifac <-----------------
c
      end
