c========================================================================
      subroutine tridsl( trans, n, dl, d, du, du2, ipiv, b, info )


c**********************************************************************
c                   SPLIB Copyright (C) 1995-1999                    **
c                     Indiana University                             **
c**********************************************************************

c========================================================================
c      ***** Solve a factored tridiagonal system *****			*
c========================================================================
c
c     Solve a linear system Ax = b, where A is tridiagonal and has
c     been factored by trifac.  This routine is modified from a Lapack
c     routine (dgttrs.f).  Modifications include changing the code to
c     handle only one right hand side, and eliminating xerbla calls.
c     Because of the change in the calling sequence, the error codes
c     have also been changed.
c
c------------------------------------------------------------------------
c     Randall Bramley
c     Department of Computer Science
c     Indiana University
c     email: bramley@cs.indiana.edu
c     Tue Jan 18 10:01:23 EST 1994
c------------------------------------------------------------------------
c
c  -- original Lapack routine (Version 1.1) --
c     Univ. of Tennessee, Univ. of California Berkeley, NAG Ltd.,
c     Courant Institute, Argonne National Lab, and Rice University
c     March 31, 1993
c
c
c  Scalar arguments 
      implicit none
      integer            info, n
c  Array arguments 
      integer            ipiv( * )
      double precision   b( * ), d( * ), dl( * ), du( * ), du2( * )
c  Character arguments 
      character trans(*)
c
c========================================================================
c
c  purpose
c  =======
c
c  tridsl solves one of the systems of equations
c     a*x = b  or  a'*x = b,
c  with a tridiagonal matrix a using the lu factorization computed
c  by trifac.f.
c
c  arguments
c  =========
c
c  trans   (input) character
c          specifies the form of the system of equations:
c          = 'n':  a * x = b  (no transpose)
c          = 't':  a'* x = b  (transpose)
c
c  n       (input) integer
c          the order of the matrix a.
c
c  dl      (input) double precision array, dimension (n-1)
c          the (n-1) multipliers that define the matrix l from the
c          lu factorization of a.
c
c  d       (input) double precision array, dimension (n)
c          the n diagonal elements of the upper triangular matrix u from
c          the lu factorization of a.
c
c  du      (input) double precision array, dimension (n-1)
c          the (n-1) elements of the first superdiagonal of u.
c
c  du2     (input) double precision array, dimension (n-2)
c          the (n-2) elements of the second superdiagonal of u.
c
c  ipiv    (input) integer array, dimension (n)
c          the pivot indices; for 1 <= i <= n, row i of the matrix was
c          interchanged with row ipiv(i).  ipiv(i) will always be either
c          i or i+1; ipiv(i) = i indicates a row interchange was not
c          required.
c
c  b       (input/output) double precision array, dimension (n)
c          on entry, the right hand side matrix b.
c          on exit, b is overwritten by the solution matrix x.
c
c  info    (output) integer
c          = 0:  successful exit
c          < 0:  if info = -i, some argument had an illegal value
c
c========================================================================
c========================================================================
c
c  Local scalars: 
      logical            notran
      integer            i
      double precision   temp
c  External functions: 
      logical            lsame
      external           lsame
c  Intrinsic functions:
      intrinsic          max
c
c========================================================================
c
      info = 0
      notran = trans(1) .eq. 'n'  .or. trans(1) .eq. 'N'
      if( n.lt.0 ) then
         info = -2
      end if
      if( info.ne.0 ) then
         write(6,*)'Error in tridsl; info = ',info
         return
      end if
      if( n.eq.0) return
c
      if( notran ) then
c
c        solve a*x = b using the lu factorization of a,
c        overwriting right hand side vector with its solution.
c
c           solve l*x = b.
c
            do 10 i = 1, n - 1
               if( ipiv( i ).eq.i ) then
                  b( i+1) = b( i+1 ) - dl( i )*b( i )
               else
                  temp = b( i )
                  b( i ) = b( i+1 )
                  b( i+1 ) = temp - dl( i )*b( i )
               end if
   10       continue
c
c           solve u*x = b.
c
            b( n ) = b( n ) / d( n )
            if( n.gt.1 )
     &         b( n-1 ) = ( b( n-1 )-du( n-1 )*b( n ) ) /
     &                       d( n-1 )
            do 20 i = n - 2, 1, -1
               b( i ) = ( b( i )-du( i )*b( i+1 )-du2( i )*
     &                     b( i+2 ) ) / d( i )
   20       continue
      else
c
c        solve a' * x = b.
c
c
c           solve u'*x = b.
c
            b( 1 ) = b( 1 ) / d( 1 )
            if( n.gt.1 )
     &         b( 2 ) = ( b( 2 )-du( 1 )*b( 1 ) ) / d( 2 )
            do 40 i = 3, n
               b( i ) = ( b( i )-du( i-1 )*b( i-1 )-du2( i-2 )*
     &                     b( i-2 ) ) / d( i )
   40       continue
c
c           solve l'*x = b.
c
            do 50 i = n - 1, 1, -1
               if( ipiv( i ).eq.i ) then
                  b( i ) = b( i ) - dl( i )*b( i+1 )
               else
                  temp = b( i+1 )
                  b( i+1 ) = b( i ) - dl( i )*temp
                  b( i ) = temp
               end if
   50       continue
      end if
c
c  End of tridsl
c
      return
      end
