c=======================================================================
      subroutine utltsl (n, y, x, lu, jlu, uptr, pcmeth, a, colind,
     &                   rwptr)

c**********************************************************************
c                   SPLIB Copyright (C) 1995-1999                    **
c                     Indiana University                             **
c**********************************************************************

c=======================================================================*
c                *** UTLTSL: Solve M'x = y ***				*
c									*
c  Apply transpose preconditioner, for methods such as CGNE which	*
c  require transpose matrix operations.					*
c									*
c=======================================================================*
c
c  x       : real*8 vector of length n (output).
c  y       : real*8 vector of length n (input).
c
c  n       : Order of matrix A, and system Mx = y where M is the
c            preconditioning matrix.
c  A       : real*8 array for nonzero entries of matrix A; of
c            length nnz = rwptr(n+1)-1.
c  colind  : Integer array of column indices for entries in A; of
c            length nnz = rwptr(n+1)-1.
c  rwptr   : Integer array of pointers to beginning position of rows
c            in arrays A and colind.  Must be of length n+1.
c  uptr, 
c  lu,jlu  : Data structure for the preconditioner M, with different meanings
c            for different preconditioners.
c
c            ILU(s), MILU(s), or ILUt --->
c                 (lu,jlu) gives the combined L/U factors for the preconditioner
c                 in a modified sparse row (MSR) format.  L is unit lower 
c                 triangular and U is upper triangular.  The diagonal of U is
c                 stored in lu(1:n) inverted, to replace divisions in the algorithm
c                 with multiplications.  Each i-th row of the combined L/U matrix in
c                 the (lu,jlu) data structure contains the i-th row of L (excluding
c                 the diagonal entry = 1), followed by the i-th row of u.
c                 uptr is a integer vector of length n containing pointers to the
c                 start position of each row of U in (lu, jlu).
c
c            SSOR(omega)   --->
c                 lu is the scalar acceleration parameter omega.
c                 jlu is not used, and uptr is a vector of pointers to the
c                 diagonal entries in the CSR data structure for A.
c
c            TRID(s)   --->
c                 This implements block diagonal preconditioning, where
c                 each diagonal block is from the corresponding tridiagonal part
c                 of the matrix A.  The size of the blocks is given implicitly in
c                 the integer iskip, stored in jlu(n).  Factors are set up for 
c                 contiguous blocks of iskip rows and so by using iskip = n the 
c                 factor can be for the entire matrix's tridiagonal part.
c                 Using iskip = 1 is an inefficient way of performing diagonal
c                 preconditioning, while 1 < iskip < n is for a block tridiagonal
c                 preconditioner.  Gaussian elimination with pivoting is used for 
c                 factoring each block, so it is necessary to store the pivot
c                 sequence in jlu(1:n-1).  lu has the four vectors giving the nonzeros
c                 of the tridiagonal preconditioner.  Vector uptr is not used.
c
c            ILU0   --->
c                 This implements the version of ILU(0) that does not change any
c                 off-diagonal entries of A, and so only needs to store the n-vector
c                 of diagonal pivots.  lu(1:n) stores those pivots, and jlu(1:n) contains
c                 pointers to the diagonal entry in each row of A.
c
c  pcmeth  : preconditioning method to use.
c		pcmeth = 0  ---> no preconditioner, i.e. m = identity matrix.
c		pcmeth = 1  ---> ILU(c).
c		pcmeth = 2  ---> MILU(c).
c		pcmeth = 3  ---> ILUT(c).
c		pcmeth = 4  ---> SSOR(omega).
c		pcmeth = 5  ---> TRID.
c		pcmeth = 6  ---> ILU0.
c		pcmeth = 7  ---> ECIMGS.
c
c  Routines
c  ========
c
c  lutsolt : Solves u^t l^t x = y.
c            used with ILU, MILU, ILUT, ECIMGS
c  lutssor : Solves u^t l^t x = y.
c            used with SSOR
c  lutsol0 : Solves u^t l^t x = y.
c            used with ILU0
c  tridsl  : Applies transposed tridiagonal block preconditioner.
c  dcopy   : Blas routine for copying vectors
c
c-------------------------------------------------------------------------
c
c     Randall Bramley
c     Department of Computer Science
c     Indiana University, Bloomington
c     bramley@cs.indiana.edu
c     Wed Jul 12 10:03:18 EST 1995
c
c-------------------------------------------------------------------------

c     Declarations of arguments
c     -------------------------
      implicit none
      integer n
      real*8 x(n),y(n),lu(*)
      real*8 a(*)
      integer jlu(*),uptr(n)
      integer colind(*), rwptr(n+1)
      integer pcmeth, ierr

c     Local variables
c     ---------------
      integer k, length

c     Timing variables
c     ----------------
      include '../include/heads.inc'
      real*8 mytime, tstart
      external mytime


c========================================================================
c      Beginning of Executable Statements
c========================================================================

c      tstart = mytime()

c     ----------------
c     Copy y over to x
c     ----------------
      call dcopy(n,y,1,x,1)

c     ------------------------------------------
c     Unknown pcmeth (should not be encountered)
c     ------------------------------------------
      if (pcmeth.lt.0 .or. pcmeth .gt. 7) then
         write(6,*) 'Unknown method in utltsl'
         stop
      endif

c     ------------------
c     No preconditioning
c     ------------------
      if (pcmeth.eq.0) then
         return
      endif

c     --------------------------
c     ILU, MILU, ILUT, or ECIMGS
c     --------------------------
      if ( ((pcmeth.ge.1) .and. (pcmeth.le.3)) .or. pcmeth .eq. 7)
     +   call lutsolt (n, x, lu, jlu, uptr)

c     --------------------
c     SSOR preconditioning
c     --------------------
      if (pcmeth.eq.4)
     +   call lutssor(n, x, lu, jlu, a, colind, rwptr)

c     --------------------
c     ILU0 preconditioning
c     --------------------
      if (pcmeth.eq.6)
     +   call lutsol0(n, x, lu, jlu, a, colind, rwptr)

c     --------------------
c     TRID preconditioning
c     --------------------
      if (pcmeth.eq.5) then
         do 100 k = 1, n, jlu(n)
            length = min(jlu(n),n-k+1)
            call tridsl('T',length,lu(1+k),
     &                             lu(k+n),
     &                             lu(k+2*n),
     &                             lu(k+3*n),
     &                             jlu(k),x(k),ierr)
            if (ierr .ne. 0) then
               write(6,*)'Error in preconditioner; tridsl ierr = ' ,ierr
               write(6,*)'        length = ', length
               stop
            end if
 100     continue
      end if

c      times(5) = times(5) + mytime() - tstart

      return

c===================== End of LUSL =====================================

      end

c====================================================================== 
      subroutine lutsolt (n, x, lu, jlu, uptr) 
c====================================================================== 
c
c   Perform u^t l^t x = x for LU matrix as produced by ILU, MILU, ILUT,
c   or ECIMGS.
c
c====================================================================== 

        implicit none
	integer n, jlu(*),uptr(n),i,k 
        real*8 x(n), lu(*)

c       -------------
c       Solve U'x = x
c       -------------
	do i = 1, n
           x(i) = lu(i)*x(i)
	   do k=uptr(i),jlu(i+1)-1 
              x(jlu(k)) = x(jlu(k)) - lu(k)*x(i)
           end do
        end do

c       -------------
c       Solve L'x = x
c       -------------
        do i = n,1,-1 
           do k = jlu(i), uptr(i)-1 
              x(jlu(k)) = x(jlu(k)) - lu(k)* x(i) 
           end do
        end do

  	return

c================ End of lutsolt =========================================

  	end


c======================================================================= 
      subroutine lutsol0(n, x, pivots, diag, a,colind,rwptr) 
c======================================================================= 
c  Solve the system (LU)'x = x, where L and U were created using
c  ILU(-1) preconditioner, the space saver version of ILU(0).
c  The implicit splitting is
c          LU = (I + LA(D)^-1) (D + UA)
c  where D is the diagonal matrix of pivots, and LA, UA are the strictly
c  lower and upper triangular parts of A.
c======================================================================= 

      implicit none
      integer n, colind(*), rwptr(n+1), diag(n)
      real*8 x(n), pivots(n)
      real*8 a(*), t
      integer i, j

      do 40 i = 1, n
         t = pivots(i)*x(i)
         do 20 j = diag(i)+1,rwptr(i+1)-1
            x(colind(j)) = x(colind(j)) - t*a(j)
  20     continue
  40  continue

      do 80 i = n, 1, -1
         x(i) = pivots(i)*x(i)
         do 60 j = rwptr(i), diag(i)-1
            x(colind(j)) = x(colind(j)) - a(j)*x(i)
  60     continue
  80  continue
      return

c=================== End of lutsol0 ======================================= 

      end


c======================================================================= 
      subroutine lutssor(n, x, omega, diag, a, colind, rwptr)
c======================================================================= 
c
c  Solve the transposed SSOR preconditioning system
c   M'x = x, where M' =  (I + U' * omega * D^-1) * ( D/omega + L')
c
c======================================================================= 
      implicit none
      integer n, colind(*), rwptr(n+1), diag(n)
      real*8 x(n), omega
      real*8 a(*), dk
      integer k, j

      do 40 k = 1, n
         dk = x(k)*omega/a(diag(k))
         do 20 j = diag(k)+1, rwptr(k+1)-1
           x(colind(j)) = x(colind(j)) - dk*a(j)
  20     continue
  40  continue
c
      do 80 k = n, 1, -1
         dk = x(k)*omega/a(diag(k))
         x(k) = dk
         do 60 j = rwptr(k), diag(k)-1
           x(colind(j)) = x(colind(j)) - dk*a(j)
  60     continue
  80  continue
      
      return
      end
c=================== End of lutssor ======================================= 
