c========================================================================
      subroutine  cgne(A,colind,rwptr,n,
     1                 x,r,b,
     2                 tolrnc,reltv,maxits,ierr,
     3                 work,wksz,
     4                 lu,jlu,uptr,pcmeth,iunit,instlvl)

c**********************************************************************
c                   SPLIB Copyright (C) 1995-1999                    **
c                     Indiana University                             **
c**********************************************************************

c-----------------------------------------------------------------------*
c           		*** CGNE ***					*
c									*
c  Method originally proposed in Hestenes and Stiefel's classical	*
c  paper on the CG method.						*
c									*
c  Solve a nonsymmetric sparse system Ax = b by the conjugate gradient	*
c  iteration applied to the normal equations,				*
c									*
c                        AA'y = b     x = A'y				*
c									*
c  The sparse matrix is stored in the compressed sparse row (CSR)	*
c  format (a, colind, rwptr).						*
c									*
c  This uses two sided preconditioning, implicitly working on the	*
c  linear system							*
c									*
c     inv(LU)*A*A'*inv(LU)' v = inv(LU)*b,				*
c                           x = A'*inv(LU)'				*
c									*
c  The algorithm does not directly use v, and instead via a change of	*
c  variables it directly updates the vector x.  This two sided version	*
c  uses the changes of variables					*
c  	d = inv(LU)'*dhat, where dhat is the usual search direction,	*
c  	q = A'*d,							*
c  	x = A'*y = A'* inv(LU)'*v,					*
c  	y = inv(LU)'*v.							*
c  The residual vector r actually being kept here is for the residual	*
c      r = inv(LU)*(b-A*x).						*
c									*
c-----------------------------------------------------------------------*
c
c  Arguments:
c  ----------
c
c     (Data structure for A)
c  n       : Order of matrix A.
c  A       : Real*8 array for nonzero entries of matrix A; of
c            length nnz = rwptr(n+1)-1.
c  colind  : Integer array of column indices for entries in A; of
c            length nnz = rwptr(n+1)-1.
c  rwptr   : Integer array of pointers to beginning position of rows
c            in arrays A and colind.  Must be of length n+1.
c
c     (Required vectors)
c  x       : Vector of unknowns to be solved for.  Contains initial
c            estimate of x on entry, and contains final approximate
c            on return.
c  r       : Residual vector r = b - A*x.
c  b       : Right hand side vector of system to be solved: Ax = b
c  work    : Double precision array of additional storage.  CGNE
c            requires 3 vectors of length n, and if instrumentation
c            routine instr.f is called, an additional 2 vectors of
c            length n are required.  If not enough storage is given,
c            nothing is done and an error code is returned specifying
c            the additional storage needed (in double precision words).
c  wksz    : Length of array work, in double precision words.  
c
c     (Preconditioner data structure)
c  lu,jlu  : Preconditioner in Modified Sparse Row format containing
c            the L and U factors together. The diagonal of U (stored in
c            lu(1:n) ) is inverted.  The i-th row of the lu,jlu matrix 
c            contains the i-th row of L (excluding the diagonal entry) 
c            followed by the i-th row of U.  The diagonal of L has
c            all ones, and so is not stored explicitly.
c  uptr    : Integer array of length n containing the pointers to        
c            the beginning of each row of U in the matrix lu,jlu. 
c  pcmeth  : Preconditioning method to use.  See splib.f documentation
c            for the interpretation of this integer.
c
c     (Iteration control)
c  tolrnc  : Tolerances on norm(s) used in stopping test.  This is
c            passed in as a vector in case future versions
c            set pdr and SMALL, rather than using
c            the dlamch() function.  Since Krylov methods generally
c            use the dotproduct r'*r, a local variable tolsq is used
c            in order to avoid square roots.  Note: because the
c            preconditioned residual norm || inv(M)*r || is the natural
c            one tracked by the iterative method, but users generally
c            want the true residual norm || r || as the stopping test,
c            this routine will exit the iterations when the precond
c            residual is small enough and then test the true residual.
c            If it is not sufficiently reduced, then the local tolerance
c            tolsq is reduced and iterations resume.  This is done
c            a maximum of eight times before failure is declared and
c            the routine returns.
c  reltv   : Logical variable indicating whether or not to use a
c            relative residual as the stopping test.  If this is
c            false, the stopping test is || r_k || < tolrnc(1).  If it
c            is true, the test is || r_k || < || r_0 || * tolrnc(1).
c            Note that it is possible to modify SPLIB to use other
c            stopping tests (see routine instr.f).
c  maxits  : Maximum number of iterations allowed.
c
c      (Other stuff)
c  ierr    : Error flag; see below.
c  iunit   : Integer vector of unit numbers for the instrumentation
c            routine instr.f to use.  Not referenced by cgne, just
c            passed along to instr.f
c  instlvl : Integer giving instrumentation level.  If instlvl = 0,
c            then instr.f is not called and array work() need not
c            provide the 2n double precision words used by instr.f
c
c  Some internal variables:
c  ------------------------
c
c  first   : Logical parameter that specifies to instr.f if this is
c            the first iteration, and so if new headers need to be
c            written to the Xgraph output files.
c  SMALL   : Parameter that determines when it is not safe to divide
c            by a scalar quantity.  Error condition is set if a divisor
c            is smaller than SMALL.  Set by dlamch() function.
c  pdr     : Parameter to determine if preconditioner is no good.
c            Experience shows that if the preconditioned residual norm
c            is more than 1/eps, where eps is the machine epsilon, then
c            even if the preconditioned residual converges, the original
c            system residual does not.  If you are not comfortable with
c            this try setting pdr = 1.0d200.  
c  urnorm  : Unpreconditioned initial residual.  Used for tests to see
c            if initial residual is good enough, and if preconditioner
c            is no good (see pdr above).
c
c  w,d,q   : Integer pointers in the array work(*) that give the
c            starting position of the corresponding vector used by CGNE.
c  wins1,
c  wins2   : Integer pointers in the array work(*) that give starting
c            positions of two vectors used by instrumentation routines.
c
c======================================================================
c
c Return values:                                                           
c --------------
c
c  x      : Contains an approximate solution (upon successful return).  
c
c  ierr   : Error flag with the following meanings:          
c           ierr = -m --> Not enough workspace provided; m more double
c                         precision words needed in array work().
c           ierr = 0  --> Successful return; tolerance achieved.
c           ierr = 1  --> The initial guess satisfies the stopping test.
c                         This is often a clue that rhs not set properly.
c           ierr = 2  --> Tolerance reset 8 times in attempt to
c                         get true residual small enough.
c           ierr = 3  --> convergence not achieved in maxits iterations.  
c           ierr = 4  --> Preconditioned residual is 1/eps times
c                         larger than unpreconditioned residual.
c           ierr = 5  --> error condition for bs(qtq) <= SMALL.
c           ierr = 6  --> error condition for abs(rho) <= SMALL.
c
c  maxits : Contains the number of actual iterations performed.
c  tolrnc : May be modified if reltv = .true.
c                                                                       
c======================================================================
c  Routines
c  --------
c
c  lusl    : Solves M z = r.
c  utltsl  : Solves M'z = r.
c  bmux    : Computes matrix-vector product with A.
c  btmux   : Computes matrix-vector product with A'.
c  instr   : Instrumentation routine for timing, etc.
c  dlamch  : Function from LAPACK project, for determining machine
c            constants.  This provides machine epsilon and the safe
c            invertible floating point minimum. 
c  BLAS routines: ddot, daxpy
c
c======================================================================
c
c     Randall Bramley
c     Department of Computer Science
c     Indiana University
c     email: bramley@cs.indiana.edu
c     Thu Jul 28 13:40:51 EST 1994
c     Documented Wed Jul  5 12:56:48 EST 1995
c
c----------------------------------------------------------------------
       implicit none

c     -------------------------
c     Declaration of arguments:
c     -------------------------
      integer n
      real*8 A(*)
      real*8 x(*), b(*), r(*)
      real*8 work(n,*)
      real*8 tolrnc(*)
      real*8 lu(*)
c
      integer colind(*), rwptr(*)
      integer jlu(*), uptr(*)
      integer maxits, ierr, iunit(*), instlvl
      integer pcmeth, wksz
      logical reltv

c     ------------------
c     External functions
c     ------------------
      real*8 ddot, dlamch

c     -------------
c     Local storage 
c     ------------- 
      real*8 urnorm, pdr, SMALL, t
      integer its, i, resets
      real*8 alpha, beta, temp, rho, tolsq
      logical first

c     -------------------------------
c     Pointers for work array vectors
c     -------------------------------
      integer w, d, q, wins1, wins2

c=======================================================================
c     Start of executable statements
c=======================================================================

c     ------------------------------
c     Set up square root of safe min
c     ------------------------------
      SMALL = sqrt(dlamch('S'))

c     ---------------------------------
c     Set up inverse of machine epsilon
c     ---------------------------------
      pdr = 1.0d0/dlamch('E')

c----------------------------------------------------------------------
c  Key to work storage:
c     w -> work(:,1)  (product vector w = A*p, etc.)
c     d -> work(:,2)  (search direction d)
c     q -> work(:,3)  (preconditioned residual vector)
c----------------------------------------------------------------------

c     --------------------------------------
c     Set up pointers for work array vectors
c     --------------------------------------
      if (instlvl .gt. 0) then
         wins1 = 1
         wins2 = wins1 + 1
      else
         wins1 = 0
         wins2 = 0
      end if
      w = wins2 + 1
      d = w + 1
      q = d + 1

c     -------------------------
c     Initialize for iterations
c     -------------------------
      its = 0
      resets = 0
      tolsq = tolrnc(1)*tolrnc(1)
      first = .true.

c     ----------------------------------------
c     Return if not enough workspace provided:
c     ----------------------------------------
      if (wksz .lt. (wins2+3)*n) then
         ierr = wksz - (wins2+3)*n
         maxits = 0
         return
      end if 

c     --------------------------------------
c     Check initial true residual w = b - Ax 
c     --------------------------------------
      call bmux(n, n, x, work(1,w), A, colind, rwptr)
      do 10 i = 1, n
         work(i,w) = b(i) - work(i,w)
  10  continue
      urnorm = ddot(n, work(1,w), 1, work(1,w), 1)

c     ------------------------------------------------
c     Adjust tolerance if relative residual to be used
c     ------------------------------------------------
      if (reltv) then
          tolsq = tolsq*urnorm
          tolrnc(1) = tolrnc(1)*sqrt(urnorm)
      end if

c     -----------------------------------------------------
c     Check for initial guess satisfying stopping condition
c     -----------------------------------------------------
      if (urnorm .le. tolsq) go to 900

c     ------------------------------
c     r = inv(LU)*w = inv(LU)*(b-Ax)
c     ------------------------------
      call lusl(n, work(1,w), r, lu, jlu, uptr, pcmeth, A,colind,rwptr)

c     ---------
c     rho = r'r
c     ---------
      rho = ddot(n, r, 1, r, 1)

c     --------------
c     d = inv(LU)'*r
c     --------------
      call utltsl(n, r, work(1,d), lu, jlu, uptr, pcmeth,
     1            A,colind,rwptr)

c  ---------------------------------------------------------
c  Check if preconditioned residual is too large relative to
c  unpreconditioned residual:
c  ---------------------------------------------------------
      if (ddot(n,work(1,d),1,work(1,d),1) .gt. pdr*pdr*urnorm) then
         ierr = 4
         maxits = 0
         return
      end if 

c=======================================================================
c     Begin main loop of iterations.  Loop is equivalent to
c      do while (rnorm .gt. tolsq) .and. (its .le. maxits)
c=======================================================================

  30   continue

      call instr(A,colind,rwptr,n, x,b, work(1,wins1), lu,jlu,uptr,
     &               pcmeth,first,its, iunit,instlvl)
      its = its + 1

c     -------
c     q = A'd
c     -------
      call btmux(n,n, work(1,d), work(1,q), A, colind, rwptr)

c     ---------------
c     w = inv(LU)*A*q
c     ---------------
      call bmux(n, n, work(1,q), work(1,w), A, colind, rwptr)
      call lusl(n, work(1,w), work(1,w), lu,jlu,uptr, pcmeth, A,
     &          colind,rwptr)

c     ---------------
c     alpha = rho/q'q
c     ---------------
      t = ddot(n, work(1,q), 1, work(1,q), 1)
      if (abs(t) .le. SMALL) go to 930
      alpha = rho / t

c     ---------------
c     x = x + alpha*q
c     ---------------
      call daxpy(n, alpha, work(1,q), 1, x, 1)

c     ---------------
c     r = r - alpha*w
c     ---------------
      call daxpy(n, -alpha, work(1,w), 1, r, 1)

c     ---------------------------------------
c     temp = r'r, beta = temp/rho, rho = temp
c     ---------------------------------------
      temp = ddot(n, r, 1, r, 1)
      if (abs(rho) .le. SMALL) go to 940
      beta = temp / rho
      rho = temp

c     -----------------------
c     d = inv(LU)'*r + beta*d
c     -----------------------
      call utltsl(n, r, work(1,w), lu, jlu, uptr, pcmeth, A,
     &            colind,rwptr)
      do 50 i = 1,n
                work(i,d) = work(i,w) + beta*work(i,d)
  50  continue
c
      if (rho .le. tolsq) go to 910
  60  continue
      if (its .gt. maxits) go to 920
      go to 30

c=======================================================================
c                  Stopping conditions
c=======================================================================

c     ====================
c     (Almost) exact guess
c     ====================
 900  continue
      ierr = 1
      maxits = 0
      return

c     =========================================
c     Possible convergence; check true residual
c     =========================================
 910  continue
      ierr = 0
      resets = resets + 1
      if (resets .gt. 8) then
         ierr = 2
         go to 915
      end if
      call bmux(n, n,x,work(1,w),A,colind,rwptr)
      do i = 1, n
         work(i,w) = b(i) - work(i,w)
      end do
      temp = ddot(n,work(1,w),1,work(1,w),1)
      if (temp .gt. tolrnc(1)*tolrnc(1)) then
          tolsq = tolsq*tolrnc(1)*tolrnc(1)/temp
c          write(6,*)'>>>>> tolerance reset to ', sqrt(tolsq),'  <<<<<<'
          go to 60
      end if
 915  continue
      maxits = its
      call instr(A,colind,rwptr,n, x,b, work(1,wins1), lu,jlu,uptr,
     &               pcmeth,first,its, iunit,instlvl)
      return

c     =========================================
c     No convergence in max allowed iterations.
c     =========================================
 920  continue
      ierr = 3
      maxits = its
      call instr(A,colind,rwptr,n, x,b, work(1,wins1), lu,jlu,uptr,
     &               pcmeth,first,its, iunit,instlvl)
      return

c     ================================
c     Failure from abs(q'q) too small.
c     ================================
 930  continue
      ierr = 5
      maxits = its
      call instr(A,colind,rwptr,n, x,b, work(1,wins1), lu,jlu,uptr,
     &               pcmeth,first,its, iunit,instlvl)
      return

c     ================================
c     Failure from abs(rho) too small.
c     ================================
 940  continue
      ierr = 6
      maxits = its
      call instr(A,colind,rwptr,n, x,b, work(1,wins1), lu,jlu,uptr,
     &               pcmeth,first,its, iunit,instlvl)
      return

c================= End of CGNE ======================================== 

      end
