c========================================================================
      subroutine  cgnr(A,colind,rwptr,n,
     1                 x,r,b,
     2                 tolrnc,reltv,maxits,ierr,
     3                 work,wksz,
     4                 lu,jlu,uptr,pcmeth,iunit,instlvl)

c**********************************************************************
c                   SPLIB Copyright (C) 1995-1999                    **
c                     Indiana University                             **
c**********************************************************************

c-----------------------------------------------------------------------*
c           		*** CGNR ***					*
c									*
c  Solve a symmetric sparse system Ax = b by the conjugate gradient	*
c  iteration applied to the normal equations				*
c									*
c                             A'Ax = A'b				*
c									*
c  The sparse matrix is stored in the compressed sparse row (CSR)	*
c  format (A, colind, rwptr).  Preconditioning is two-sided, on		*
c  the system								*
c									*
c           inv(U'L')*A'*A*inv(LU)*y = inv(U'L')*A'b,			*
c                                  y = LU*x.				*
c									*
c  The usual CG algorithm is applied to the system, but with the	*
c  change of variables d = inv(LU)*dhat, where dhat is the usual	*
c  search direction and w = inv(LU)'*what, where "what" is the usual	*
c  product vector with the coefficient matrix.  And Who is on first.	*
c									*
c-----------------------------------------------------------------------*
c
c  Arguments:
c  ----------
c
c     (Data structure for A)
c  n       : Order of matrix A.
c  A       : Real*8 array for nonzero entries of matrix A; of
c            length nnz = rwptr(n+1)-1.
c  colind  : Integer array of column indices for entries in A; of
c            length nnz = rwptr(n+1)-1.
c  rwptr   : Integer array of pointers to beginning position of rows
c            in arrays A and colind.  Must be of length n+1.
c
c     (Required vectors)
c  x       : Vector of unknowns to be solved for.  Contains initial
c            estimate of x on entry, and contains final approximate
c            on return.
c  r       : Residual vector r = b - A*x.
c  b       : Right hand side vector of system to be solved: Ax = b
c  work    : Double precision array of additional storage.  CGNR
c            requires 3 vectors of length n, and if instrumentation
c            routine instr.f is called, an additional 2 vectors of
c            length n are required.  If not enough storage is given,
c            nothing is done and an error code is returned specifying
c            the additional storage needed (in double precision words).
c  wksz    : Length of array work, in double precision words.  
c
c     (Preconditioner data structure)
c  lu,jlu  : Preconditioner in Modified Sparse Row format containing
c            the L and U factors together. The diagonal of U (stored in
c            lu(1:n) ) is inverted.  The i-th row of the lu,jlu matrix 
c            contains the i-th row of L (excluding the diagonal entry) 
c            followed by the i-th row of U.  The diagonal of L has
c            all ones, and so is not stored explicitly.
c  uptr    : Integer array of length n containing the pointers to        
c            the beginning of each row of U in the matrix lu,jlu. 
c  pcmeth  : Preconditioning method to use.  See splib.f documentation
c            for the interpretation of this integer.
c
c     (Iteration control)
c  tolrnc  : Tolerances on norm(s) used in stopping test.  This is
c            passed in as a vector in case future versions
c            set pdr and SMALL, rather than using
c            the dlamch() function.  Since Krylov methods generally
c            use the dotproduct r'*r, a local variable tolsq is used
c            in order to avoid square roots.  Note: because the
c            preconditioned residual norm || inv(M)*r || is the natural
c            one tracked by the iterative method, but users generally
c            want the true residual norm || r || as the stopping test,
c            this routine will exit the iterations when the precond
c            residual is small enough and then test the true residual.
c            If it is not sufficiently reduced, then the local tolerance
c            tolsq is reduced and iterations resume.  This is done
c            a maximum of eight times before failure is declared and
c            the routine returns.
c  reltv   : Logical variable indicating whether or not to use a
c            relative residual as the stopping test.  If this is
c            false, the stopping test is || r_k || < tolrnc(1).  If it
c            is true, the test is || r_k || < || r_0 || * tolrnc(1).
c            Note that it is possible to modify SPLIB to use other
c            stopping tests (see routine instr.f).
c  maxits  : Maximum number of iterations allowed.
c
c      (Other stuff)
c  ierr    : Error flag; see below.
c  iunit   : Integer vector of unit numbers for the instrumentation
c            routine instr.f to use.  Not referenced by cgne, just
c            passed along to instr.f
c  instlvl : Integer giving instrumentation level.  If instlvl = 0,
c            then instr.f is not called and array work() need not
c            provide the 2n double precision words used by instr.f
c
c  Some internal variables:
c  ------------------------
c
c  first   : Logical parameter that specifies to instr.f if this is
c            the first iteration, and so if new headers need to be
c            written to the Xgraph output files.
c  SMALL   : Parameter that determines when it is not safe to divide
c            by a scalar quantity.  Error condition is set if a divisor
c            is smaller than SMALL.  Set by dlamch() function.
c  pdr     : Parameter to determine if preconditioner is no good.
c            Experience shows that if the preconditioned residual norm
c            is more than 1/eps, where eps is the machine epsilon, then
c            even if the preconditioned residual converges, the original
c            system residual does not.  If you are not comfortable with
c            this try setting pdr = 1.0d200.  
c  urnorm  : Unpreconditioned initial residual.  Used for tests to see
c            if initial residual is good enough, and if preconditioner
c            is no good (see pdr above).
c
c  w,d,q   : Integer pointers in the array work(*) that give the
c            starting position of the corresponding vector used by CGNR.
c  wins1,
c  wins2   : Integer pointers in the array work(*) that give starting
c            positions of two vectors used by instrumentation routines.
c
c======================================================================
c
c Return values:                                                           
c --------------
c
c  x      : Contains an approximate solution (upon successful return).  
c
c  ierr   : Error flag with the following meanings:          
c           ierr = -m --> Not enough workspace provided; m more double
c                         precision words needed in array work().
c           ierr = 0  --> Successful return; tolerance achieved.
c           ierr = 1  --> The initial guess satisfies the stopping test.
c                         This is often a clue that rhs not set properly.
c           ierr = 2  --> Tolerance reset 8 times in attempt to
c                         get true residual small enough.
c           ierr = 3  --> convergence not achieved in maxits iterations.  
c           ierr = 4  --> Preconditioned residual is 1/eps times
c                         larger than unpreconditioned residual.
c           ierr = 5  --> error condition for bs(w'd) <= SMALL.
c           ierr = 6  --> error condition for abs(rho) <= SMALL.
c
c  maxits : Contains the number of actual iterations performed.
c  tolrnc : May be modified if reltv = .true.
c                                                                       
c======================================================================
c  Routines
c  --------
c
c  lusl    : Solves M z = r.
c  utltsl  : Solves M'z = r.
c  bmux    : Computes matrix-vector product with A.
c  btmux   : Computes matrix-vector product with A'.
c  instr   : Instrumentation routine for timing, etc.
c  dlamch  : Function from LAPACK project, for determining machine
c            constants.  This provides machine epsilon and the safe
c            invertible floating point minimum. 
c  BLAS routines: ddot, daxpy
c
c======================================================================
c
c  Original version developed at:
c     Center for Supercomputing Research and Development
c     University of Illinois - Urbana
c     1 February 1991
c
c  Revised:
c     Sat Jan  1 
c     (to output timing data)
c     Fri Jan  7 07:54:26 EST 1994
c     (to use only 3n workspace instead of 4n).
c     Thu Jan 20 09:04:16 EST 1994
c     (to implement TRID preconditioning).
c     Mon Jun  6 09:43:49 EST 1994
c     (to use new calling sequence and instrumentation routines)
c     
c     Randall Bramley
c     Department of Computer Science
c     Indiana University
c     email: bramley@cs.indiana.edu
c     Documented Thu Jul  6 05:20:12 EST 1995
c
c----------------------------------------------------------------------
      implicit none

c     -------------------------
c     Declaration of arguments:
c     -------------------------
      integer n
      real*8 A(*)
      real*8 lu(*)
      real*8 x(*), b(*), r(*)
      real*8 work(n,*)
      real*8 tolrnc(*)

      integer colind(*), rwptr(*)
      integer jlu(*), uptr(*)
      integer maxits, ierr
      integer wksz
      integer pcmeth, iunit(*), instlvl
      logical reltv

c     ------------------
c     External functions
c     ------------------
      real*8 ddot, dlamch

c     ------------- 
c     Local storage 
c     ------------- 
      logical first
      integer its, i, resets
      real*8 rho, alpha, temp, beta, tolsq, t
      real*8 pdr, urnorm, SMALL

c     -------------------------------
c     Pointers to work array vectors:
c     -------------------------------
      integer w,d,z,wins1,wins2

c=======================================================================
c     Start of executable statements
c=======================================================================

c     ------------------------------
c     Set up square root of safe min
c     ------------------------------
      SMALL = sqrt(dlamch('S'))

c     -------------------------------------------------
c     Set tolerance for size of preconditioned residual
c     -------------------------------------------------
      pdr = 1.0d0/dlamch('E')

c     -----------------------------------
c     Set pointers for work array vectors
c     -----------------------------------
      if (instlvl .gt. 0) then
         wins1 = 1
         wins2 = wins1 + 1
      else
         wins1 = 0
         wins2 = 0
      end if
      w = wins2 + 1
      d = w + 1
      z = d + 1

c----------------------------------------------------------------------
c
c  Key to additional work storage:
c        w -> work(:,1)  (product vector w = A'Ad)
c        d -> work(:,2)  (search direction vector )
c        z -> work(:,3)  (work vector)
c----------------------------------------------------------------------

c     ------------------------
c     Initialize for iteration
c     ------------------------
      its = 0
      resets = 0
      tolsq = tolrnc(1)*tolrnc(1)
      first = .true.
      if (wksz .lt. (wins2+3)*n) then
         ierr = wksz - (wins2+3)*n
         maxits = 0
         return
      end if

c     ---------- 
c     w = b - Ax 
c     ---------- 
      call bmux(n, n, x, work(1,w), A, colind, rwptr)
      do 10 i = 1, n
         work(i,w) = b(i) - work(i,w)
  10  continue
      urnorm = ddot(n, work(1,w), 1, work(1,w), 1)

c     -------------------------------------------------
c     Adjust tolerance if relative residual to be used:
c     -------------------------------------------------
      if (reltv) then
          tolsq = tolsq*urnorm
          tolrnc(1) = tolrnc(1)*sqrt(urnorm)
      end if

c     -----------------------------------------------------
c     Early exit if first residual satisfies stopping test.
c     -----------------------------------------------------
      if (urnorm .le. tolsq) go to 900

c     -----------------
c     r = inv(LU)'*A'*w
c     -----------------
      call btmux(n, n, work(1,w), r, A, colind, rwptr)
      call utltsl(n, r, r, lu, jlu, uptr, pcmeth, A,colind,rwptr)

c     -------------
c     d = inv(LU)*r
c     -------------
      call lusl(n, r, work(1,d), lu, jlu, uptr, pcmeth, A,colind,rwptr)
      rho = ddot(n, r, 1, r, 1)

c     -------------------------------------------------------------
c     Check to see if preconditioned residual is too large relative
c     to unpreconditioned residual
c     -------------------------------------------------------------
      if (rho .gt. pdr*pdr*urnorm) then
         ierr = 4
         maxits = 0
         return
      end if 

c=======================================================================
c     Begin main loop of iterations.  Loop is equivalent to
c      do while (rnorm .gt. tolsq) .and. (its .le. maxits)
c=======================================================================

  30   continue

      call instr(A,colind,rwptr,n, x,b, work(1,wins1), lu,jlu,uptr,
     &               pcmeth,first,its, iunit,instlvl)
      its = its + 1

c     ----------
c     w = A'*A*d
c     ----------
      call bmux(n,n, work(1,d), work(1,z), A, colind, rwptr)
      call btmux(n,n, work(1,z), work(1,w), A, colind, rwptr)
      t = ddot(n, work(1,w), 1, work(1,d), 1)
      if (abs(t) .le. SMALL) go to 930
      alpha = rho / t

c     ---------------
c     x = x + alpha*d
c     ---------------
      call daxpy(n, alpha, work(1,d), 1, x, 1)

c     ------------------------
c     r = r - alpha*inv(LU)'*w
c     ------------------------
      call utltsl(n,work(1,w),work(1,z),lu,jlu,uptr,pcmeth, A,colind,
     &            rwptr)
      call daxpy(n, -alpha, work(1,z), 1, r, 1)

c     ---------------------------------------
c     temp = r'r; beta = temp/rho; rho = temp
c     ---------------------------------------
      temp = ddot(n, r, 1, r, 1)
      if (abs(rho) .le. SMALL) go to 940
      beta = temp / rho
      rho = temp

c     ----------------------
c     d = inv(LU)*r + beta*d
c     ----------------------
      call lusl(n, r, work(1,z), lu, jlu, uptr, pcmeth, A,colind,rwptr)
      do 50 i = 1, n
         work(i,d) = work(i,z) + beta*work(i,d)
  50  continue

      if (rho .le. tolsq) go to 910
  60  continue
      if (its .ge. maxits) go to 920
      go to 30

c=======================================================================
c                  Stopping conditions
c=======================================================================

c     ====================
c     (Almost) exact guess
c     ====================
 900  continue
      ierr = 1
      maxits = 0
      return

c     =========================================
c     Possible convergence; check true residual
c     =========================================
 910  continue
      ierr = 0
      resets = resets + 1
      if (resets .gt. 8) then
         ierr = 2
         go to 915
      end if
      call bmux(n, n,x,work(1,w),A,colind,rwptr)
      do i = 1, n
         work(i,w) = b(i) - work(i,w)
      end do
      temp = ddot(n,work(1,w),1,work(1,w),1)
      if (temp .gt. tolrnc(1)*tolrnc(1)) then
          tolsq = tolsq*tolrnc(1)*tolrnc(1)/temp
c          write(6,*)'>>>>> tolerance reset to ', sqrt(tolsq),'  <<<<<<'
          go to 60
      end if
 915  continue
      maxits = its
      call instr(A,colind,rwptr,n, x,b, work(1,wins1), lu,jlu,uptr,
     &               pcmeth,first,its, iunit,instlvl)
      return

c     ================================================
c     No convergence within allowed maximum iterations
c     ================================================
 920  continue
      ierr = 3
      maxits = its
      call instr(A,colind,rwptr,n, x,b, work(1,wins1), lu,jlu,uptr,
     &               pcmeth,first,its, iunit,instlvl)
      return

c     ================================
c     Failure from abs(w'd) too small.
c     ================================
 930  continue
      ierr = 5
      maxits = its
      call instr(A,colind,rwptr,n, x,b, work(1,wins1), lu,jlu,uptr,
     &               pcmeth,first,its, iunit,instlvl)
      return

c     ================================
c     Failure from abs(rho) too small.
c     ================================
 940  continue
      ierr = 6
      maxits = its
      call instr(A,colind,rwptr,n, x,b, work(1,wins1), lu,jlu,uptr,
     &               pcmeth,first,its, iunit,instlvl)
      return

c================= End of CGNR ======================================== 

      end

