c=======================================================================*
      subroutine  cgstab(A,colind,rwptr,n,
     1                   x,r,b,
     2                   tolrnc,reltv,maxits,ierr,
     3                   work,wksz,
     4                   lu,jlu,uptr,pcmeth, iunit,instlvl)

c**********************************************************************
c                   SPLIB Copyright (C) 1995-1999                    **
c                     Indiana University                             **
c**********************************************************************

c-----------------------------------------------------------------------*
c                           *** CGSTAB ***				*
c  The stabilized biconjugate gradient algorithm of van der Vorst.	*
c									*
c  See bicgs.f for the corresponding Templates version.			*
c									*
c-----------------------------------------------------------------------*
c
c  Arguments:
c  ----------
c
c     (Data structure for A)
c  n       : Order of matrix A.
c  A       : Real*8 array for nonzero entries of matrix A; of
c            length nnz = rwptr(n+1)-1.
c  colind  : Integer array of column indices for entries in A; of
c            length nnz = rwptr(n+1)-1.
c  rwptr   : Integer array of pointers to beginning position of rows
c            in arrays A and colind.  Must be of length n+1.
c
c     (Required vectors)
c  x       : Vector of unknowns to be solved for.  Contains initial
c            estimate of x on entry, and contains final approximate
c            on return.
c  r       : Residual vector r = b - A*x.
c  b       : Right hand side vector of system to be solved: Ax = b
c  work    : Double precision array of additional storage.  CGSTAB
c            requires 6 vectors of length n, and if instrumentation
c            routine instr.f is called, an additional 2 vectors of
c            length n are required.  If not enough storage is given,
c            nothing is done and an error code is returned specifying
c            the additional storage needed (in double precision words).
c  wksz    : Length of array work, in double precision words.  
c
c     (Preconditioner data structure)
c  lu,jlu  : Preconditioner in Modified Sparse Row format containing
c            the L and U factors together. The diagonal of U (stored in
c            lu(1:n) ) is inverted.  The i-th row of the lu,jlu matrix 
c            contains the i-th row of L (excluding the diagonal entry) 
c            followed by the i-th row of U.  The diagonal of L has
c            all ones, and so is not stored explicitly.
c  uptr    : Integer array of length n containing the pointers to        
c            the beginning of each row of U in the matrix lu,jlu. 
c  pcmeth  : Preconditioning method to use.  See splib.f documentation
c            for the interpretation of this integer.
c
c     (Iteration control)
c  tolrnc  : Tolerances on norm(s) used in stopping test.  This is
c            passed in as a vector in case future versions
c            set pdr and SMALL, rather than using
c            the dlamch() function.  Since Krylov methods generally
c            use the dotproduct r'*r, a local variable tolsq is used
c            in order to avoid square roots.  Note: because the
c            preconditioned residual norm || inv(M)*r || is the natural
c            one tracked by the iterative method, but users generally
c            want the true residual norm || r || as the stopping test,
c            this routine will exit the iterations when the precond
c            residual is small enough and then test the true residual.
c            If it is not sufficiently reduced, then the local tolerance
c            tolsq is reduced and iterations resume.  This is done
c            a maximum of eight times before failure is declared and
c            the routine returns.
c  reltv   : Logical variable indicating whether or not to use a
c            relative residual as the stopping test.  If this is
c            false, the stopping test is || r_k || < tolrnc(1).  If it
c            is true, the test is || r_k || < || r_0 || * tolrnc(1).
c            Note that it is possible to modify SPLIB to use other
c            stopping tests (see routine instr.f).
c  maxits  : Maximum number of iterations allowed.
c
c      (Other stuff)
c  ierr    : Error flag; see below.
c  iunit   : Integer vector of unit numbers for the instrumentation
c            routine instr.f to use.  Not referenced by cgstab, just
c            passed along to instr.f
c  instlvl : Integer giving instrumentation level.  If instlvl = 0,
c            then instr.f is not called and array work() need not
c            provide the 2n double precision words used by instr.f
c
c  Some internal variables:
c  ------------------------
c
c  early   : Logical parameter that indicates if a tentative exit
c            was made because of early termination.  This is needed
c            because of the way SPLIB uses unpreconditioned residuals
c            as the stopping test.
c  first   : Logical parameter that specifies to instr.f if this is
c            the first iteration, and so if new headers need to be
c            written to the Xgraph output files.
c  SMALL   : Parameter that determines when it is not safe to divide
c            by a scalar quantity.  Error condition is set if a divisor
c            is smaller than SMALL.  Set by dlamch() function.
c  pdr     : Parameter to determine if preconditioner is no good.
c            Experience shows that if the preconditioned residual norm
c            is more than 1/eps, where eps is the machine epsilon, then
c            even if the preconditioned residual converges, the original
c            system residual does not.  If you are not comfortable with
c            this try setting pdr = 1.0d200.  
c  urnorm  : Unpreconditioned initial residual.  Used for tests to see
c            if initial residual is good enough, and if preconditioner
c            is no good (see pdr above).
c
c  temp,q,
c  v,p,s,t : Integer pointers in the array work(*) that give the
c            starting position of the corresponding vectors.
c  wins1,
c  wins2   : Integer pointers in the array work(*) that give starting
c            positions of two vectors used by instrumentation routines.
c
c======================================================================
c
c Return values:                                                           
c --------------
c
c  x      : Contains an approximate solution (upon successful return).  
c
c  ierr   : Error flag with the following meanings:          
c           ierr = -m --> Not enough workspace provided; m more double
c                         precision words needed in array work().
c           ierr = 0  --> Successful return; tolerance achieved.
c           ierr = 1  --> The initial guess satisfies the stopping test.
c                         This is often a clue that rhs not set properly.
c           ierr = 2  --> Tolerance reset 8 times in attempt to
c                         get true residual small enough.
c           ierr = 3  --> convergence not achieved in maxits iterations.  
c           ierr = 4  --> Preconditioned residual is 1/eps times
c                         larger than unpreconditioned residual.
c           ierr = 5  --> alpha = 0 during iteration.
c           ierr = 6  --> beta = 0 during iteration.
c           ierr = 7  --> tau = 0 during iteration.
c
c  maxits : Contains the number of actual iterations performed.
c  tolrnc : May be modified if reltv = .true.
c                                                                       
c======================================================================
c  Routines
c  --------
c
c  lusl    : Solves M z = r.
c  bmux    : Computes matrix-vector product with A.
c  instr   : Instrumentation routine for timing, etc.
c  dlamch  : Function from LAPACK project, for determining machine
c            constants.  This provides machine epsilon and the safe
c            invertible floating point minimum. 
c  BLAS routines: ddot, daxpy, dcopy
c
c======================================================================
c
c     Randall Bramley
c     Department of Computer Science
c     Indiana University
c     email: bramley@cs.indiana.edu
c     Mon Jun  6 11:13:35 EST 1994
c     Documented Sun Jul  9 16:51:23 EST 1995
c
c======================================================================

c     -------------------------
c     Declaration of arguments:
c     -------------------------
      implicit none

      integer n
      real*8 A(*)
      real*8 x(*), b(*), r(*)
      real*8 work(n,*)
      real*8 tolrnc(*)
      real*8 lu(*)

      integer colind(*), rwptr(*)
      integer wksz
      integer maxits, ierr
      integer jlu(*), uptr(*)
      integer pcmeth, iunit(*), instlvl
      logical reltv

c     ------------------
c     External functions
c     ------------------
      real*8 ddot, dlamch
      external dlamch, ddot
      real*8 Oetli
      external Oetli

c     -------------
c     Local storage 
c     -------------
      logical first
      integer i, its, resets
      real*8 alfa, beta, betah, omega, omegah, sigma, tau, res, tolsq
      real*8 SMALL, urnorm, pdr
      real*8 Oetli0, Oetli1

c     ------------------------------
c     Pointers to work array vectors
c     ------------------------------
      integer temp,q,v,p,s,t, wins1, wins2

c=======================================================================
c     Start of executable statements
c=======================================================================

c     ------------------------------
c     Set up square root of safe min
c     ------------------------------
      SMALL = sqrt(dlamch('S'))

c     ---------------------------------
c     Set up inverse of machine epsilon
c     ---------------------------------
      pdr = 1.0d0/dlamch('E')

c----------------------------------------------------------------------
c
c  Key to additional work storage requirements (if no instrumentation):
c
c       temp -> work(:,1)
c       q    -> work(:,2)
c       v    -> work(:,3)
c       p    -> work(:,4)
c       s    -> work(:,5)
c       t    -> work(:,6)
c
c----------------------------------------------------------------------

c     ----------------------------------------
c     Set up meaningful names for work vectors
c     ----------------------------------------
      if (instlvl .gt. 0) then
         wins1 = 1
         wins2 = wins1 + 1
      else
         wins1 = 0
         wins2 = 0
      end if
       temp = wins2 + 1
       q    = temp + 1
       v    = q + 1
       p    = v + 1
       s    = p + 1
       t    = s + 1

c     -------------------------------------
c     Stop if not enough workspace provided
c     -------------------------------------
      if (wksz .lt. (wins2+6)*n) then
         ierr = wksz - (wins2+6)*n
         maxits = 0
         return
      end if

c     ---------------------
c     Initialize iterations
c     ---------------------
      resets = 0
      first = .true.
      tolsq = tolrnc(1)*tolrnc(1)
      its = 0

c     ------------------
c     t = A*x; v = b - t
c     ------------------
      call bmux(n, n,x,work(1,t),A,colind,rwptr)
      do i = 1,n
         work(i,v) = b(i) - work(i,t)
      end do
      urnorm = ddot(n, work(1,v), 1, work(1,v), 1)

c     -------------------------------------------------
c     Adjust tolerance if relative residual to be used:
c     -------------------------------------------------
      if (reltv) then
          tolsq = tolsq*urnorm
          tolrnc(1) = tolrnc(1)*sqrt(urnorm)
      end if

      if (urnorm .lt. tolsq) go to 900

c     ---------------------------------------
c     r = preconditioned residual = inv(LU)*v
c     ---------------------------------------
      call lusl(n, work(1,v), r, lu, jlu, uptr, pcmeth, A,colind,rwptr)

c     -------------------------------------------------------------
c     Check to see if preconditioned residual is too large relative
c     to unpreconditioned residual:
c     -------------------------------------------------------------
      if (ddot(n,r,1,r,1) .gt. pdr*pdr*urnorm) then
         ierr = 4
         maxits = 0
         return
      end if 

      do i = 1,n
         work(i,p) = r(i)
         work(i,v) = 0.0d0
         work(i,q) = 0.0d0
      end do

      omegah = 1.0d0
      beta = 1.0d0
      alfa = 1.0d0
      Oetli0 = Oetli(A,rwptr,colind,n, b, x, first)

c=======================================================================
c     Begin main loop of iterations.  Loop is equivalent to
c      do while (rnorm .gt. tolsq) .and. (its .le. maxits)
c=======================================================================

 2       continue

         call instr(A,colind,rwptr,n, x,b, work(1,wins1), lu,jlu,uptr,
     &               pcmeth,first,its, iunit,instlvl)
         its = its + 1

         betah = ddot(n, work(1,p), 1, r, 1)
         if (abs(beta) .le. SMALL) go to 930

         omega = (betah/beta)*(omegah/alfa)
         beta = betah
         do i = 1,n
            work(i,q) = r(i)+omega*(work(i,q)-alfa*work(i,v))
         end do

         call bmux(n, n,work(1,q),work(1,t),A,colind,rwptr)
         call lusl(n,work(1,t),work(1,v),lu,jlu,uptr,pcmeth, 
     &             A,colind,rwptr)

         tau = ddot(n, work(1,p), 1, work(1,v), 1)
         if (abs(tau) .le. SMALL) go to 950

         omegah = betah/tau
         do i = 1,n
            work(i,s) = r(i)-omegah*work(i,v)
         end do

         call bmux(n,n,work(1,s),work(1,temp),A,colind,rwptr)
         call lusl(n,work(1,temp),work(1,t),lu,jlu,uptr,pcmeth, 
     &             A,colind,rwptr)

         sigma = ddot(n, work(1,t), 1, work(1,s), 1)
         tau = ddot(n, work(1,t), 1, work(1,t), 1)

c        ---------------------------
c        Check for early termination
c        ---------------------------
         if (abs(tau) .le. SMALL) then
            call daxpy(n, omegah, work(1,q), 1, x, 1)
            go to 950
         end if 

         alfa = sigma/tau
         if (abs(alfa) .le. SMALL) go to 940

         do i = 1,n
            x(i) = x(i)+omegah*work(i,q)+alfa*work(i,s)
         end do

         do i = 1,n
            r(i) = work(i,s)-alfa*work(i,t)
         end do

         res = ddot(n, r, 1, r, 1)
         if (its .ge. maxits) go to 920
         if (res .lt. tolsq) go to 910

        go to 2

c=======================================================================
c           Stopping conditions
c=======================================================================


c     ===================== 
c     (Almost) exact guess 
c     ===================== 
 900  continue
      ierr = 1
      maxits = 0
      return

c     ========================================= 
c     Possible convergence; check true residual 
c     ========================================= 
 910  continue
      ierr = 0
      resets = resets + 1
      if (resets .gt. 8) then
         ierr = 2
         go to 915
      end if
      call bmux(n,n,x,work(1,s),A,colind,rwptr)
      do i = 1, n
         work(i,s) = b(i) - work(i,s)
      end do
      res = ddot(n,work(1,s),1,work(1,s),1)
      if (res .gt. tolrnc(1)*tolrnc(1)) then
          Oetli1 = Oetli(A,rwptr,colind,n, b, x, first)
          if (Oetli1/Oetli0 .lt. max(10.0d0/pdr, tolsq) ) then
             ierr = 100
             go to 915
          end if
          tolsq = tolsq*tolrnc(1)*tolrnc(1)/res
c         write(6,*)'>>>>> tolerance reset to ', sqrt(tolsq),'  <<<<<<'
          go to 2
      end if
 915  continue
      maxits = its
      call instr(A,colind,rwptr,n, x,b, work(1,wins1), lu,jlu,uptr,
     &               pcmeth,first,its, iunit,instlvl)
      return

c     ======================================== 
c     No convergence in max allowed iterations 
c     ======================================== 
 920  continue
      ierr = 3
      maxits = its
      call instr(A,colind,rwptr,n, x,b, work(1,wins1), lu,jlu,uptr,
     &               pcmeth,first,its, iunit,instlvl)
      return

c     ======== 
c     beta = 0
c     ======== 
 930  continue
      ierr = 6
      maxits = its
      call instr(A,colind,rwptr,n, x,b, work(1,wins1), lu,jlu,uptr,
     &               pcmeth,first,its, iunit,instlvl)
      return

c     ========= 
c     alpha = 0 
c     ========= 
 940  continue
      ierr = 5
      maxits = its
      call instr(A,colind,rwptr,n, x,b, work(1,wins1), lu,jlu,uptr,
     &               pcmeth,first,its, iunit,instlvl)
      return

c     ======= 
c     tau = 0 
c     ======= 
 950  continue
      ierr = 7
      maxits = its
      call instr(A,colind,rwptr,n, x,b, work(1,wins1), lu,jlu,uptr,
     &               pcmeth,first,its, iunit,instlvl)
      return

c================= End of CGSTAB =======================================

      end
