c======================================================================*
      subroutine gmres(a,colind,rwptr,n,
     1                 x,r,b,
     2                 tolrnc,reltv,maxits,ierr,
     3                 work,wksz,
     4                 lu,jlu,uptr,pcmeth,iunit,instlvl)

c**********************************************************************
c                   SPLIB Copyright (C) 1995                         **
c                     Indiana University                             **
c**********************************************************************

c-----------------------------------------------------------------------*
c                       *** GMRES ***					*
c									*
c  Standard (nonflexible) version of GMRES(m).  The grandfather of	*
c  this code is one supplied by Yousef Saad while at CSRD, and many	*
c  of the names of variables have been retained.			*
c									*
c  Unlike other routines in SPLIB, this uses right preconditioning.	*
c  That allows keeping track of the unpreconditioned residual, without	*
c  having to worry about the preconditioned residual norm giving an	*
c  (too) early termination.  Nevertheless, it is possible for the	*
c  preconditioner to be unstable, and the usual check is supplied for	*
c  that (see the ierr = 4 condition).					*	
c									*
c  Also note that maxits is used as a vector here, to pass along the	*
c  maximal Krylov subspace size.					*
c									*
c-----------------------------------------------------------------------*
c
c  Arguments:
c  ----------
c
c     (Data structure for A)
c  n       : Order of matrix A.
c  A       : Real*8 array for nonzero entries of matrix A; of
c            length nnz = rwptr(n+1)-1.
c  colind  : Integer array of column indices for entries in A; of
c            length nnz = rwptr(n+1)-1.
c  rwptr   : Integer array of pointers to beginning position of rows
c            in arrays A and colind.  Must be of length n+1.
c
c     (Required vectors)
c  x       : Vector of unknowns to be solved for.  Contains initial
c            estimate of x on entry, and contains final approximate
c            on return.
c  r       : Residual vector r = b - A*x.
c  b       : Right hand side vector of system to be solved: Ax = b
c  work    : Double precision array of additional storage.  GMRES
c            requires a matrix H for the reduced approximation in
c            the Krylov subspace, as well as a few vectors. If the
c            routine instr.f is called, an additional 2 vectors of
c            length n are required.  If not enough storage is given,
c            nothing is done and an error code is returned specifying
c            the additional storage needed (in double precision words).
c            The array must have a length (see wksz) such that
c                wksz  >=   (1+im)*n   +  im   + im*(im+1)  + 3*im+1,
c                for      Krylov space + work  + Hessenberg + c,s,rs
c            Add 2*n if instrumentation is used.
c  wksz    : Length of array work, in double precision words.  
c
c     (Preconditioner data structure)
c  lu,jlu  : Preconditioner in Modified Sparse Row format containing
c            the L and U factors together. The diagonal of U (stored in
c            lu(1:n) ) is inverted.  The i-th row of the lu,jlu matrix 
c            contains the i-th row of L (excluding the diagonal entry) 
c            followed by the i-th row of U.  The diagonal of L has
c            all ones, and so is not stored explicitly.
c  uptr    : Integer array of length n containing the pointers to        
c            the beginning of each row of U in the matrix lu,jlu. 
c  pcmeth  : Preconditioning method to use.  See splib.f documentation
c            for the interpretation of this integer.
c
c     (Iteration control)
c  tolrnc  : Tolerances on norm(s) used in stopping test.  This is
c            passed in as a vector in case future versions
c            set pdr and SMALL, rather than using
c            the dlamch() function.  Since Krylov methods generally
c            use the dotproduct r'*r, a local variable tolsq is used
c            in order to avoid square roots.  Note: because the
c            preconditioned residual norm || inv(M)*r || is the natural
c            one tracked by the iterative method, but users generally
c            want the true residual norm || r || as the stopping test,
c            this routine will exit the iterations when the precond
c            residual is small enough and then test the true residual.
c            If it is not sufficiently reduced, then the local tolerance
c            tolsq is reduced and iterations resume.  This is done
c            a maximum of eight times before failure is declared and
c            the routine returns.
c  reltv   : Logical variable indicating whether or not to use a
c            relative residual as the stopping test.  If this is
c            false, the stopping test is || r_k || < tolrnc(1).  If it
c            is true, the test is || r_k || < || r_0 || * tolrnc(1).
c            Note that it is possible to modify SPLIB to use other
c            stopping tests (see routine instr.f).
c  maxits  : Integer array of length two.
c            maxits(1) ---> Maximum number of iterations allowed.
c            maxits(2) ---> Dimension im of Krylov subspace to use.
c
c      (Other stuff)
c  ierr    : Error flag; see below.
c  iunit   : Integer vector of unit numbers for the instrumentation
c            routine instr.f to use.  Not referenced by gmres, just
c            passed along to instr.f
c  instlvl : Integer giving instrumentation level.  If instlvl = 0,
c            then instr.f is not called and array work() need not
c            provide the 2n double precision words used by instr.f
c
c  Some internal variables:
c  ------------------------
c
c  first   : Logical parameter that specifies to instr.f if this is
c            the first iteration, and so if new headers need to be
c            written to the Xgraph output files.
c  SMALL   : Parameter that determines when it is not safe to divide
c            by a scalar quantity.  Error condition is set if a divisor
c            is smaller than SMALL.  Set by dlamch() function.
c  pdr     : Parameter to determine if preconditioner is no good.
c            Experience shows that if the preconditioned residual norm
c            is more than 1/eps, where eps is the machine epsilon, then
c            even if the preconditioned residual converges, the original
c            system residual does not.  If you are not comfortable with
c            this try setting pdr = 1.0d200.  
c  rs,
c  hh,c,s  : Integer pointers in the array work(*) that give the
c            starting position of the matrices and vectors used by
c            GMRES.  Note that this is much more complicated here 
c            than in the other routines of SPLIB, because of the need
c            to store a full im x im array in work().  The 
c            correspondance between GMRES algorithm arrays and the
c            work vector indexing are given by
c               hh(i,j)  --> work( (j-1)*im + i, hh)
c                  c(i)  --> work( c + i, hh)
c                  s(i)  --> work( s + i, hh)
c                 rs(i)  --> work(rs + i, hh)
c            To make things a little easier to understand, each update
c            in the work() array is accompanied by a commented out 
c            assignment in the corresponding vectors rs(), c(), etc.
c  wins1,
c  wins2   : Integer pointers in the array work(*) that give starting
c            positions of two vectors used by instrumentation routines.
c
c======================================================================
c
c Return values:                                                           
c --------------
c
c  x      : Contains an approximate solution (upon successful return).  
c
c  ierr   : Error flag with the following meanings:          
c           ierr = -m --> Not enough workspace provided; m more double
c                         precision words needed in array work().
c           ierr = 0  --> Successful return; tolerance achieved.
c           ierr = 1  --> The initial guess satisfies the stopping test.
c                         This is often a clue that rhs not set properly.
c           ierr = 2  --> Tolerance reset 8 times in attempt to
c                         get true residual small enough.
c           ierr = 3  --> convergence not achieved in maxits iterations.  
c           ierr = 4  --> Preconditioned residual is 1/eps times
c                         larger than unpreconditioned residual.
c
c  maxits : Contains the number of actual iterations performed
c           in maxits(1); the Krylov subspace size maxits(2) is
c           unchanged.
c  tolrnc : May be modified if reltv = .true.
c                                                                       
c======================================================================
c  Routines
c  --------
c
c  lusl    : Solves M z = r.
c  bmux    : Computes matrix-vector product with A.
c  instr   : Instrumentation routine for timing, etc.
c  dlamch  : Function from LAPACK project, for determining machine
c            constants.  This provides machine epsilon and the safe
c            invertible floating point minimum. 
c  BLAS routines: ddot, daxpy, dscal, dgemv
c
c======================================================================
c
c     Randall Bramley
c     Department of Computer Science
c     Indiana University
c     email: bramley@cs.indiana.edu
c     Thu Jul 21 09:14:53 EST 1994
c
c     Original version coded by Y. Saad, from SPARSKIT I 
c
c----------------------------------------------------------------------* 

c     -------------------------
c     Declaration of arguments:
c     -------------------------
      implicit none
      integer n, maxits(2), ierr
      real*8 work(n,*), b(n), x(n), a(*), lu(*), r(*)
      integer colind(*), rwptr(n+1), jlu(*), uptr(n)
      real*8 tolrnc(*)
      integer wksz
      logical reltv
      real*8 timers(10), t0
      real*8 mytime
      external mytime

c     ------------- 
c     Local storage 
c     ------------- 
      logical first
      real*8 ro, t, eps1, gam
      integer im, nnz
      integer j, i, k, ii, k1, jj, i1
      integer pcmeth, its, resets, iunit(*), instlvl
      real*8 epsmac, SMALL, pdr

c     -------------------------------
c     Pointers for work array vectors
c     -------------------------------
      integer hh, c, s, rs, wins1, wins2

c     ------------------
c     External functions
c     ------------------
      real*8 ddot, dnrm2, dlamch
      external dlamch, ddot, dnrm2

c=======================================================================
c     Start of executable statements
c=======================================================================

c     ---------------------------------------------------------
c     Set up square root of safe min, the smallest number sfmin
c     such that 1.0d0/sfmin does not overflow:
c     ---------------------------------------------------------
      SMALL = sqrt(dlamch('S'))

c     ------------------------
c     Set up machine tolerance
c     ------------------------
      epsmac = dlamch('E')

c     ---------------------------------
c     Set up inverse of machine epsilon
c     ---------------------------------
      pdr = 1.0d0/epsmac

c     ----------------------------
c     Unpack Krylov subspace size.
c     ----------------------------
      im = maxits(2)

c------------------------------------------------------------------------
c     Alias workspace pointers.  First im columns of work() are for
c     the Krylov basis vectors, next im*(im+1) are for Hessenberg
c     matrix hh, then vectors c, s, and rs of lengths im, im , im+1.
c     The pointer references are then
c                hh(i,j)  --> work( (j-1)*im + i, hh)
c                   c(i)  --> work( c + i, hh)
c                   s(i)  --> work( s + i, hh)
c                  rs(i)  --> work(rs + i, hh)
c------------------------------------------------------------------------

      if (instlvl. gt. 0) then
          wins1 = 1
          wins2 = wins1 + 1
      else
          wins1 = 0
          wins2 = 0
      end if
      hh = im + 2 + wins2
      c = im*(im+1)
      s = c + im
      rs = s + im

c     -------------------------------------
c     Stop if not enough workspace provided
c     -------------------------------------
      if (wksz .lt. wins2*n + (im*(n+im+5)+n + 1) ) then
           ierr = wksz - (wins2*n + (im*(n+im+5)+n + 1))
           maxits(1) = 0
	   return
      end if

c     Initialize for iterations
c     -------------------------
      resets = 0
      its = 0
      first = .true.

      do k = 1, 10
	timers(k) = 0.0d0
      end do

c=============================================================
c             Outer Loop Starts Here 
c=============================================================

 20     continue

c       ----------------------------------------------
c       Compute initial residual vector for this cycle
c       ----------------------------------------------
	t0 = mytime()
	call bmux (n, n, x, work(1,wins2+1), a, colind, rwptr)
	timers(1) = mytime() - t0 + timers(1)
	do j = 1,n
	   work(j,wins2+1) = b(j) - work(j,wins2+1)
	end do 
	ro = dnrm2(n,work(1,wins2+1),1)

	call instr(a,colind,rwptr,n, x,b, work(1,wins1), lu,jlu,uptr,
     &             pcmeth,first,its, iunit,instlvl)

c       ---------------------------
c       Check for early termination
c       ---------------------------
	if (abs(ro) .le. SMALL) go to 999 
	t = 1.0d0/ ro
	call dscal(n,t,work(1,wins2+1),1)

c       -----------------------------------------
c       Adjust tolerance if first outer iteration
c       -----------------------------------------
	if (its .eq. 0) then
	      eps1 = tolrnc(1)
	      if (reltv) then
		  tolrnc(1) = tolrnc(1)*ro
		  eps1 = eps1*ro
	      end if
	end if

c       --------------------------------------------------------
c       Initialize first term of RHS of Hessenberg system.  Here
c       rs is the right hand side vector for the least squares 
c       subproblem that GMRES must solve.  It is iteratively
c       updated with the Givens rotations generated below.
c       --------------------------------------------------------
c       rs(1) = ro
	work(rs+1,hh) = ro

c       -----------------------------------------
c       Size of Krylov subspace built so far is i
c       -----------------------------------------
	i = 0

c       ======================================
c       Inner loop to build up Krylov subspace
c       ======================================
 60         continue
	    i = i + 1
	    its = its + 1
	    i1 = i + 1

c           --------------------------------------------
c           Matrix-vector product for next Krylov vector
c           begins here, and ends with call to bmux.    
c           --------------------------------------------
	    call lusl (n, work(1,wins2+i), r, lu, jlu, uptr, pcmeth, 
     &                   a,colind,rwptr) 

c           ------------------------------------------------------
c           Check to see if initial preconditioned residual is too
c           large relative to initial unpreconditioned residual.
c           ------------------------------------------------------
	    if (its .eq. 1 ) then
		 if (ddot(n,r,1,r,1) .gt. pdr*pdr) then
		    ierr = 4
		    maxits(1) = 0
		    return
                 end if 
            end if 

c           ---------------------------
c           Matrix-vector product ends.
c           ---------------------------
        t0 = mytime()
            call bmux (n,n, r, work(1,wins2+i1), a, colind, rwptr) 
        timers(1) = mytime() - t0 + timers(1)

c           ========================================================
c           At this point, the upper Hessenberg matrix H = V'AV has
c           been factored and stored as H = Q*hh, with Q the product
c           of Givens rotations stored in c ( = cosines) c and s
c           ( = sines), and hh the (m+1) x m matrix
c
c                          x  x  x  x  x
c                          0  x  x  x  x
c             hh =         0  0  x  x  x
c                          0  0  0  x  x
c                          0  0  0  0  x
c                          0  0  0  0  0
c
c           Next, use Gram-Schmidt process to orthonormalize v(:,i+1)
c           vs. previous v's.   This adjoins another column and row 
c           to hh, as determined by the new Krylov vector V(:,i+1),
c           to get
c
c                          x  x  x  x  x n
c                          0  x  x  x  x n
c             hh =         0  0  x  x  x n
c                          0  0  0  x  x n
c                          0  0  0  0  x n
c                          0  0  0  0  0 s
c                          0  0  0  0  0 t
c           ==========================================================

        t0 = mytime()
            do j = 1, i
                 t = ddot(n, work(1,wins2+j),1,work(1,wins2+i1),1)
c                hh(j,i) = t
                 work( (i-1)*im+j,hh) = t
                 call daxpy(n, -t, work(1,wins2+j), 1,
     1                      work(1,wins2+i1), 1)
            end do

            t = dnrm2(n,work(1,wins2+i1),1)
c           hh(i1,i) = t
            work( (i-1)*im+i1,hh) = t
	    if ( abs(t) .le. SMALL) go to 100
	    t = 1.0d0/t
	    call dscal(n,t,work(1,wins2+i1),1)
        timers(2) = mytime() - t0 + timers(2)

c           -----------------------------------------------------
c           Done with Modified Gram-Schimdt and Arnoldi step.
c           Update factorization of hh; apply previous rotations.
c           The new hh still has the form given above, but with 
c           different entries n and s.
c           -----------------------------------------------------
 100        continue

c           ----------------------------------------------------
c           Apply previous transformations to i-th column of hh
c           ----------------------------------------------------
        t0 = mytime()
            do k = 2, i
                 k1 = k-1
c                t = hh(k1,i)
c                hh(k1,i) = c(k1)*t + s(k1)*hh(k,i)
c                hh(k,i) = -s(k1)*t + c(k1)*hh(k,i)
                 t = work( (i-1)*im+k1, hh )
                 work( (i-1)*im+k1, hh) = work(c+k1,hh)*t +
     &                    work(s+k1,hh)*work( (i-1)*im+k, hh)
                 work( (i-1)*im+k, hh) = -work(s+k1,hh)*t +
     &                    work(c+k1,hh)*work( (i-1)*im+k,hh)
            end do
c
c           gam = sqrt(hh(i,i)**2 + hh(i1,i)**2)
            gam = sqrt(work( (i-1)*im+i,hh)**2 + 
     &                   work( (i-1)*im+i1,hh)**2)
            if (abs(gam) .le. SMALL) gam = epsmac

c           -----------------------------------------------------
c           Determine next plane rotation to zilch out 't' in hh 
c           shown above, and apply it to rs; note that rs(j+1) is
c           still zero so the update uses only rs(i) to create
c           the new values for rs(i) and rs(i+1).  Similarly,
c           hh(i,k) is still zero for k < i (see representation
c           of H above).
c           -----------------------------------------------------
c           c(i) = hh(i,i)/gam
            work(c+i,hh) = work( (i-1)*im+i,hh)/gam
c           s(i) = hh(i1,i)/gam
            work(s+i,hh) = work ((i-1)*im+i1,hh)/gam
c           rs(i1) = -s(i)*rs(i)
            work(rs+i1,hh) = -work(s+i,hh)*work(rs+i,hh)
c           rs(i) =  c(i)*rs(i)
            work(rs+i,hh) =  work(c+i,hh)*work(rs+i,hh)
c           hh(i,i) = c(i)*hh(i,i) + s(i)*hh(i1,i)
            work( (i-1)*im+i,hh) = work(c+i,hh)*work( (i-1)*im+i,hh)
     &                       + work(s+i,hh)*work( (i-1)*im+i1,hh)

c           ---------------------------------------------------
c           Determine new residual norm, which is just the size
c           of rs(i+1)
c           ---------------------------------------------------
            ro = abs(work(rs+i1,hh))
        timers(3) = mytime() - t0 + timers(3)

c           ---------------------------------------------------------
c           Three convergence tests are used: first, if the Krylov 
c           subspace is of the specified maximum allowed size,
c           secondly if in middle of building up the Krylov subspace
c           the residual is small, and thirdly if the total number
c           of mat-vec multiplies (iterations) has been reached.
c           ---------------------------------------------------------
            if (i .lt. im
     &            .and. (ro .gt. eps1)
     &            .and. its .le. maxits(1) )  go to 60

c          ==========================================
c          End inner loop to build up Krylov subspace
c          ==========================================

c       -----------------------------------------------------------
c       Now compute solution x. First solve upper triangular system
c       hh(1:i,1:i) * rs(1:i) = rs(1:i)
c       -----------------------------------------------------------
c       rs(i) = rs(i)/hh(i,i)

        t0 = mytime()
        work(rs+i,hh) = work(rs+i,hh)/work( (i-1)*im+i,hh)
        do ii = 2,i
              k = i-ii+1
              k1 = k+1
c             t = rs(k)
              t = work(rs+k,hh)
              do j = k1,i
c                t = t-hh(k,j)*rs(j)
                 t = t-work( (j-1)*im+k,hh)*work(rs+j,hh)
              end do
c             rs(k) = t/hh(k,k)
              work(rs+k,hh) = t/work( (k-1)*im+k,hh)
        end do
        timers(4) = mytime() - t0 + timers(4)

c       ----------------------------------------------------------
c       Form linear combination of work(*,i)'s to get solution.
c       Note that the original code for this is commented out, and
c       replaced by a call to the BLAS2 routine for mat*vec.  That
c       original code for mat-vec product was of form
c     
c                t = rs(1)
c                do k = 1, n
c                   r(k) = work(k,1)*t
c                end do
c                do j = 2, i
c                   t = rs(j)
c                   do k = 1, n
c                      r(k) = r(k)+t*work(k,j) 
c                   end do
c                end do
c       ----------------------------------------------------------
        t0 = mytime()
        call dgemv ('N', n, i, 1.0d0, work(1,wins2+1), n,
     &                 work(rs+1,hh), 1, 0.0d0, r, 1)

        timers(5) = mytime() - t0 + timers(5)
        call lusl (n, r, r, lu, jlu, uptr, pcmeth, a,colind,rwptr)

        do k=1, n
              x(k) = x(k) + r(k)
        end do

c       Restart outer loop when necessary
c       ---------------------------------
        if (ro .le. eps1) go to 990
 260    continue
        if (its .gt. maxits(1)) go to 991

c       Else compute residual vector and continue
c       -----------------------------------------
        do j = 1,i
             jj = i1-j+1
c            rs(jj-1) = -s(jj-1)*rs(jj)
             work(rs+jj-1,hh) = -work(s+jj-1,hh)*work(rs+jj,hh)
c            rs(jj) = c(jj-1)*rs(jj)
             work(rs+jj,hh) = work(c+jj-1,hh)*work(rs+jj,hh)
        end do

        do j = 1,i1
c            t = rs(j)
             t = work(rs+j,hh)
             if (j .eq. 1)  t = t-1.0d0
             call daxpy (n, t, work(1,wins2+j), 1, work(1,wins2+1), 1) 
        end do

c=============================================================
c        Outer Loop Ends Here
c=============================================================

         go to 20


c=======================================================================
c           Stopping conditions
c=======================================================================


c     (Almost) exact guess
c     =====================
 999   continue
       ierr = 1
       maxits(1) = its
       return

c     Possible convergence; check true residual
c     =========================================
 990   continue
       ierr = 0
       resets = resets + 1
       if (resets .gt. 8) then
         ierr = 2
         go to 995
       end if
       call bmux(n,n,x,r,a,colind,rwptr)
       do j = 1, n
          r(j) = r(j) - b(j)
       end do
       t = dnrm2(n,r,1)
       if (t .gt. tolrnc(1)) then
          eps1 = eps1*tolrnc(1)/t
c          write(6,*)'>>>>> tolerance reset to ', eps1,'  <<<<<<'
          go to 260
       end if
 995   continue
       maxits(1) = its
       call instr(a,colind,rwptr,n, x,b, work(1,wins1), lu,jlu,uptr,
     &               pcmeth,first,its, iunit,instlvl)
       nnz = rwptr(n+1)-1
       Write(iunit(1),222) n,nnz, im,(timers(i),i = 1, 5)
       return
 222   format('AAA', 1x,i5,3x,i7,3x,i3,3x, 6(1pe10.3,3x))

c      No convergence within allowed iterations 
c      ======================================== 
 991   continue
       ierr = 3
       maxits(1) = its
       call instr(a,colind,rwptr,n, x,b, work(1,wins1), lu,jlu,uptr,
     &               pcmeth,first,its, iunit,instlvl)
       Write(iunit(1),222) n,nnz, im,(timers(i),i = 1, 5)
       return

c================= End of GMRES ====================================== 
           end
