c======================================================================== 
      	subroutine jacobi(A,colind,rwptr,n,
     1                    x,r,b,
     2                    tol,reltv,maxits,ierr,
     3                    work,wksz,
     4                    lu,jlu,diag,pcmeth,iunit,instlvl)

c**********************************************************************
c                   SPLIB Copyright (C) 1995-1999                    **
c                     Indiana University                             **
c**********************************************************************

c======================================================================== 
c                      *** Jacobi iteration ***				*
c                      --------------------------- 			*
c									*
c  Classical Jacobi iteration.  This is not meant to be a serious	*
c  implementation, and is provided because it is useful in teaching	*
c  courses in iterative methods.  There are severe limitations:		*
c     1) The matrix in CSR format must have its rows ordered so that	*
c        in each row first all of the lower triangular,	then the	*
c        diagonal, and finally the upper triangular entries occur.	*
c     2) No preconditioning is allowed with this implementation. 	*
c        not a limitation of the alogrithm, but of the current SPLIB.	*
c     3) Tolerances (tolrnc(*)) is replaced by scalar tol.        	*
c     4) The data structure for A may be changed, to have conditions	*
c        listed in (1) above satisfied.  Done by calling ssor.f		*
c									*
c  This routine also differs from the usual calling sequence of SPLIB.	*
c  Since no preconditioning is allowed, the arrays alu and jlu are	*
c  dummies.  The array lu, however, is used to set up pointers to the	*
c  diagonal entries in each row, so must be of length n.  This makes 	*
c  required upper/lower triangular solves more efficienct.		*
c									*
c  Finally note that for efficiency reasons, this code makes two	*
c  iterations per loop index and so might take one extra iteration	*
c  compared to a standard formulation.					*
c									*
c=======================================================================*
c
c  Arguments:
c  ----------
c
c     (Data structure for A)
c  n       : Order of matrix A.
c  A       : Real*8 array for nonzero entries of matrix A; of
c            length nnz = rwptr(n+1)-1.
c  colind  : Integer array of column indices for entries in A; of
c            length nnz = rwptr(n+1)-1.
c  rwptr   : Integer array of pointers to beginning position of rows
c            in arrays A and colind.  Must be of length n+1.
c
c     (Required vectors)
c  x       : Vector of unknowns to be solved for.  Contains initial
c            estimate of x on entry, and contains final approximate
c            on return.
c  r       : Residual vector r = b - A*x.
c  b       : Right hand side vector of system to be solved: Ax = b
c  work    : Double precision array; used here only if instrumentation
c            is turned on.
c  wksz    : Length of array work, in double precision words.  
c            Must be >= 2n if instrumentation is used.
c
c     (Preconditioner data structure)
c  lu,jlu  : Dummy arguments here.
c  diag    : Integer array of length n used to contain pointers to
c            diagonal entries in A's data structure.
c  pcmeth  : Preconditioning method to use.  WARNING: this version 
c            does not allow preconditioning.
c
c     (Iteration control)
c  tol     : Tolerance on norm used in stopping test.
c  reltv   : Logical variable indicating whether or not to use a
c            relative residual as the stopping test.  If this is
c            false, the stopping test is || r_k || < tol.  If it
c            is true, the test is || r_k || < || r_0 || * tol.
c            Note that it is possible to modify SPLIB to use other
c            stopping tests (see routine instr.f).
c  maxits  : Maximum number of iterations allowed.
c
c      (Other stuff)
c  ierr    : Error flag; see below.
c  iunit   : Integer vector of unit numbers for the instrumentation
c            routine instr.f to use.  Not referenced by cgne, just
c            passed along to instr.f
c  instlvl : Integer giving instrumentation level.  If instlvl = 0,
c            then instr.f is not called and array work() need not
c            provide the 2n double precision words used by instr.f
c
c  Some internal variables:
c  ------------------------
c
c  first   : Logical parameter that specifies to instr.f if this is
c            the first iteration, and so if new headers need to be
c            written to the Xgraph output files.
c======================================================================
c
c Return values:                                                           
c --------------
c
c  x      : Contains an approximate solution (upon successful return).  
c
c  ierr   : Error flag with the following meanings:          
c           ierr = -m --> Not enough workspace provided; m more double
c                         precision words needed in array work().
c           ierr = 0  --> Successful return; tolerance achieved.
c           ierr = 1  --> The initial guess satisfies the stopping test.
c                         This is often a clue that rhs not set properly.
c           ierr = 2  --> Tolerance reset 8 times in attempt to
c                         get true residual small enough.
c           ierr = 3  --> Convergence not achieved in maxits iterations.  
c           ierr = 4  --> Diagonal entry of A not safely invertible.
c
c  maxits : Contains the number of actual iterations performed.
c  tol    : May be modified if reltv = .true.
c
c  diag   : Integer vector of pointers to diagonal entries in 
c           the data structure for A.
c
c  A      : Although the *matrix* A is not changed by this routine,
c           it may reorder the matrix to have each row in column 
c           increasing order.  So A, colind may both be changed.  This
c           is primarily a problem if you have computed a preconditioner
c           when using another method, and want to follow a call to
c           gauss.f with a call to another solver that uses the
c           previously computed preconditioner.  BEWARE!
c                                                                       
c----------------------------------------------------------------------
c  Routines
c  --------
c
c  bmux    : Computes matrix-vector product with A.
c  instr   : Instrumentation routine for timing, etc.
c  BLAS routines: dnrm2
c
c=======================================================================*
c									*
c     Randall Bramley							*
c     Department of Computer Science					*
c     Indiana University						*
c     email: bramley@cs.indiana.edu					*
c     Wed Sep 21 07:53:21 EST 1994					*
c     Documented: Mon Jul 10 10:56:01 EST 1995				*
c									*
c=======================================================================*

	implicit none

c       -------------------------
c       Declaration of arguments:
c       -------------------------
        real*8 lu
        integer wksz, jlu, pcmeth
        logical reltv
        integer iunit(*), instlvl
	integer maxits, n

	integer rwptr(n+1), colind(*), diag(n), ierr
	real*8 work(n,*), a(*), tol
	real*8 x(n), b(n), r(n)

c       ---------------
c       Local variables
c       ---------------
	real*8 t, omega, toloc
        real*8 dummy, rho
	integer i,k, iters, resets
        logical first

c       ------------------
c       External functions
c       ------------------
	real*8 dnrm2
	external dnrm2


c=======================================================================
c     Start of executable statements
c=======================================================================

c       --------------------------- 
c       Check for adequate worksize 
c       --------------------------- 
        if (instlvl. gt. 0 .and. wksz .lt. 2*n) then
               ierr = wksz - 2*n 
               return
        end if
c
	ierr = 0
c       ---------------------------------------------------------
c       First, massage the data structure of A to get pointers to
c       diagonal entries.  Use the ssor.f routine for this:
c       ---------------------------------------------------------
      	omega = 1.0d0
      	call ssor (n,a,colind,rwptr,omega,dummy,jlu,diag,work,wksz,ierr)

c       -----------------------------------------------
c       Return if some diagonal entry of A is too small
c       -----------------------------------------------
        if (ierr .lt. 0) then
           ierr = 4
           return
        end if

c       -----------------------------------------------------------
c       Now matrix data structure is set up.  Get initial residual
c       and quit if it is small enough.  This version will perform
c       two iterations on every sweep through the loop starting at
c       statement 80, since if it ends after one the vector work 
c       will have to be copied back over anyway.
c       -----------------------------------------------------------
        call bmux(n, n, x, r, A, colind, rwptr)
        do i = 1, n
           r(i) = b(i) - r(i)
        end do
        rho = dnrm2(n,r,1)
        if (reltv) toloc = toloc*rho

c       --------------------------
c       Initialize for iterations.
c       --------------------------
        first = .true.
        toloc = tol
        iters = 0
        resets = 0

c       =================
c       Outer loop begins
c       =================

 80     continue

        call instr(a,colind,rwptr,n, x,b, work, lu,jlu,diag,pcmeth,
     &               first,iters,iunit,instlvl)
        do k = 1, n
           r(k) = 0.0d0
           do i = rwptr(k), diag(k)-1
              r(k) = r(k) + a(i)*x(colind(i))
           end do
           do i = diag(k)+1, rwptr(k+1)-1
              r(k) = r(k) + a(i)*x(colind(i))
           end do
           r(k) = (b(k) - r(k))/a(diag(k))
        end do 

        rho = 0.0d0

        do k = 1, n
           x(k) = 0.0d0
           do i = rwptr(k), diag(k)-1
              x(k) = x(k) + a(i)*r(colind(i))
           end do
           do i = diag(k)+1, rwptr(k+1)-1
              x(k) = x(k) + a(i)*r(colind(i))
           end do
           t = (b(k) - x(k))/a(diag(k))
           rho = rho + abs(a(diag(k)) * (t - r(k)))
           x(k) = t
        end do 

        iters = iters + 2
        if (rho .gt. toloc .and. iters .lt. maxits) go to 80

        if (iters .gt. maxits) then
           ierr = 3
           return
        end if 

        if (rho .le. toloc) then
           call bmux(n, n, x, r, A, colind, rwptr)
           do i = 1, n
              r(i) = b(i) - r(i)
           end do
           t = dnrm2(n,r,1)
           if (t .le. tol) then
              maxits = iters
              return
           end if 
           resets = resets + 1
           if (resets .gt. 8) then
              ierr = 2
              return
           end if
           toloc =  0.9 * toloc * tol/t
           go to 80
        end if 

	return

c===================== End of Jacobi ==================================== 

	end			
