c=======================================================================
      subroutine  orthomin(a,colind,rwptr,n,
     1                     x,r,b,
     2                     tolrnc,reltv,maxits,ierr,
     3                     work,wksz,
     4                     alu,jlu,ju,pcmeth,iunit,instlvl)

c**********************************************************************
c                   SPLIB Copyright (C) 1995-1999                    **
c                     Indiana University                             **
c**********************************************************************

c-----------------------------------------------------------------------*
c                           *** ORTHOMIN ***				*
c									*
c  The ORTHOMIN(m) algorithm of Vinsome.  This version keeps 2m vectors	*
c  where m is the maximal Krylov subspace size.  The additional m 	*
c  vectors are to store the product vectors to avoid recomputing matrix	*
c  vector products.  							*
c									*
c-----------------------------------------------------------------------*
c
c  Arguments:
c  ----------
c
c     (Data structure for A)
c  n       : Order of matrix A.
c  A       : Real*8 array for nonzero entries of matrix A; of
c            length nnz = rwptr(n+1)-1.
c  colind  : Integer array of column indices for entries in A; of
c            length nnz = rwptr(n+1)-1.
c  rwptr   : Integer array of pointers to beginning position of rows
c            in arrays A and colind.  Must be of length n+1.
c
c     (Required vectors)
c  x       : Vector of unknowns to be solved for.  Contains initial
c            estimate of x on entry, and contains final approximate
c            on return.
c  r       : Residual vector r = b - A*x.
c  b       : Right hand side vector of system to be solved: Ax = b
c  work    : Double precision array of additional storage.  ORTHOMIN 
c            requires 2m vectors of length n for the Krylov subspace
c            basis and matrix-vector product vectors, as well as
c            6 vectors of length n.  If instrumentation
c            routine instr.f is called, an additional 2 vectors of
c            length n are required.  If not enough storage is given,
c            nothing is done and an error code is returned specifying
c            the additional storage needed (in double precision words).
c  wksz    : Length of array work, in double precision words.  
c
c     (Preconditioner data structure)
c  lu,jlu  : Preconditioner in Modified Sparse Row format containing
c            the L and U factors together. The diagonal of U (stored in
c            lu(1:n) ) is inverted.  The i-th row of the lu,jlu matrix 
c            contains the i-th row of L (excluding the diagonal entry) 
c            followed by the i-th row of U.  The diagonal of L has
c            all ones, and so is not stored explicitly.
c  uptr    : Integer array of length n containing the pointers to        
c            the beginning of each row of U in the matrix lu,jlu. 
c  pcmeth  : Preconditioning method to use.  See splib.f documentation
c            for the interpretation of this integer.
c
c     (Iteration control)
c  tolrnc  : Tolerances on norm(s) used in stopping test.  This is
c            passed in as a vector in case future versions
c            set pdr and SMALL, rather than using
c            the dlamch() function.  Since Krylov methods generally
c            use the dotproduct r'*r, a local variable tolsq is used
c            in order to avoid square roots.  Note: because the
c            preconditioned residual norm || inv(M)*r || is the natural
c            one tracked by the iterative method, but users generally
c            want the true residual norm || r || as the stopping test,
c            this routine will exit the iterations when the precond
c            residual is small enough and then test the true residual.
c            If it is not sufficiently reduced, then the local tolerance
c            tolsq is reduced and iterations resume.  This is done
c            a maximum of eight times before failure is declared and
c            the routine returns.
c  reltv   : Logical variable indicating whether or not to use a
c            relative residual as the stopping test.  If this is
c            false, the stopping test is || r_k || < tolrnc(1).  If it
c            is true, the test is || r_k || < || r_0 || * tolrnc(1).
c            Note that it is possible to modify SPLIB to use other
c            stopping tests (see routine instr.f).
c  maxits  : Maximum number of iterations allowed.
c
c      (Other stuff)
c  ierr    : Error flag; see below.
c  iunit   : Integer vector of unit numbers for the instrumentation
c            routine instr.f to use.  Not referenced by orthomin, just
c            passed along to instr.f
c  instlvl : Integer giving instrumentation level.  If instlvl = 0,
c            then instr.f is not called and array work() need not
c            provide the 2n double precision words used by instr.f
c
c  Some internal variables:
c  ------------------------
c
c  next    : Next available position in Krylov basis matrix
c  kdim    : Dimension of Krylov subspace so far
c  first   : Logical parameter that specifies to instr.f if this is
c            the first iteration, and so if new headers need to be
c            written to the Xgraph output files.
c  SMALL   : Parameter that determines when it is not safe to divide
c            by a scalar quantity.  Error condition is set if a divisor
c            is smaller than SMALL.  Set by dlamch() function.
c  pdr     : Parameter to determine if preconditioner is no good.
c            Experience shows that if the preconditioned residual norm
c            is more than 1/eps, where eps is the machine epsilon, then
c            even if the preconditioned residual converges, the original
c            system residual does not.  If you are not comfortable with
c            this try setting pdr = 1.0d200.  
c  urnorm  : Unpreconditioned initial residual.  Used for tests to see
c            if initial residual is good enough, and if preconditioner
c            is no good (see pdr above).
c
c  D, Beta,
c  W, WtW  : Integer pointers in the array work(*) that give the
c            starting position of the corresponding vectors.
c            Beware that beta is used slightly differently, as the
c            row number.  Beta begins in work(Beta,WtW), as is done
c            with the vectors in the gmres.f routine.
c  wins1,
c  wins2   : Integer pointers in the array work(*) that give starting
c            positions of two vectors used by instrumentation routines.
c
c======================================================================
c
c Return values:                                                           
c --------------
c
c  x      : Contains an approximate solution (upon successful return).  
c
c  ierr   : Error flag with the following meanings:          
c           ierr = -m --> Not enough workspace provided; m more double
c                         precision words needed in array work().
c           ierr = 0  --> Successful return; tolerance achieved.
c           ierr = 1  --> The initial guess satisfies the stopping test.
c                         This is often a clue that rhs not set properly.
c           ierr = 2  --> Tolerance reset 8 times in attempt to
c                         get true residual small enough.
c           ierr = 3  --> convergence not achieved in maxits iterations.  
c           ierr = 4  --> Preconditioned residual is 1/eps times
c                         larger than unpreconditioned residual.
c           ierr = 5  --> wtw = 0 during iteration.
c
c  maxits : Contains the number of actual iterations performed.
c  tolrnc : May be modified if reltv = .true.
c                                                                       
c======================================================================
c  Routines
c  --------
c
c  lusl    : Applies preconditioner; solves M z = r.
c  bmux    : Computes matrix-vector product with A.
c  instr   : Instrumentation routine for timing, etc.
c  dlamch  : Function from LAPACK project, for determining machine
c            constants.  This provides machine epsilon and the safe
c            invertible floating point minimum. 
c  BLAS routines: ddot, daxpy, dcopy
c
c======================================================================
c
c     Randall Bramley
c     Department of Computer Science
c     Indiana University
c     email: bramley@cs.indiana.edu
c     Sat Dec  3 14:28:57 CST 1994
c
c======================================================================

       implicit none

c     -------------------------
c     Declaration of arguments:
c     -------------------------
      integer n, wksz
      integer colind(*), rwptr(*)
      integer maxits(2), ierr
      integer jlu(*), ju(*)
      integer pcmeth
      integer iunit(*), instlvl
      logical reltv

      real*8 a(*)
      real*8 x(*), b(*), r(*)
      real*8 work(n,*)
      real*8 tolrnc(*)
      real*8 alu(*)

c     ------------------
c     External functions
c     ------------------
      real*8 ddot, dlamch
      external dlamch, ddot

c     ------------- 
c     Local storage 
c     ------------- 
      integer krylov
      logical first
      integer k, its, resets, kdim, next, z, v
      real*8 alpha ,tolsq, t
      real*8 SMALL, urnorm, pdr

c     ------------------------------
c     Pointers to work array vectors
c     ------------------------------
      integer wins1, wins2, W, D, WtW, Beta

c=======================================================================
c     Start of executable statements
c=======================================================================

c     ------------------------------
c     Set up square root of safe min
c     ------------------------------
      SMALL = sqrt(dlamch('S'))

c     ---------------------------------
c     Set up inverse of machine epsilon
c     ---------------------------------
      pdr = 1.0d0/dlamch('E')

c     ----------------------------------------
c     Set up meaningful names for work vectors
c     ----------------------------------------
       if (instlvl .gt. 0) then
          wins1 = 1
          wins2 = wins1 + 1
       else 
          wins1 = 0
          wins2 = 0
       end if
       z    = wins2 + 1
       v    = z + 1
       W    = v + 1
       D    = W + krylov 
       WtW  = D + krylov
c      ------------------------------------------------------------
c      Warning: unlike other indices, Beta is a row number in array
c      work.  So the first entry in Beta is work(Beta,WtW)
c      ------------------------------------------------------------
       Beta = krylov + 1

c----------------------------------------------------------------------
c
c  Key to additional work storage requirements (with instrumentation):
c
c       w1          -> work(:,1)
c       w2          -> work(:,2)
c       z           -> work(:,3)
c       v           -> work(:,4)
c       W           -> work(:,5:krylov+4)
c       D           -> work(:,krylov+5:2krylov+4)
c       [WtW,Beta]  -> work(:,2krylov+5)
c
c----------------------------------------------------------------------

c     -------------------------------------
c     Stop if not enough workspace provided
c     -------------------------------------
       krylov = maxits(2)
       k = n*(2*krylov+2) + 2*krylov
       if (k .gt. wksz) then
         ierr = wksz - k
         maxits(1) = 0
         return
       end if

c     ---------------------
c     Initialize iterations
c     ---------------------
      resets = 0
      first = .true.
      tolsq = tolrnc(1)*tolrnc(1)
      its = 0
      next = 0
      kdim = 1

c     -----------
c     r = b - A*x
c     -----------
      call bmux(n,n,x,r,A,colind,rwptr)
      do k = 1,n
         r(k) = b(k) - r(k)
      end do
      urnorm = ddot(n, r, 1, r, 1)

c     -------------------------------------------------
c     Adjust tolerance if relative residual to be used:
c     -------------------------------------------------
      if (reltv) then
          tolsq = tolsq*urnorm
          tolrnc(1) = tolrnc(1)*sqrt(urnorm)
      end if

c     ------------------------------------------------
c     Exit if initial residual is already small enough
c     ------------------------------------------------
      if (urnorm .lt. tolsq) go to 900

c     ---------------------------------------
c     z = preconditioned residual = inv(LU)*r
c     ---------------------------------------
      call lusl(n, r, work(1,D), alu, jlu, ju, pcmeth, a,colind,rwptr)

c     -------------------------------------------------------------
c     Check to see if preconditioned residual is too large relative
c     to unpreconditioned residual:
c     -------------------------------------------------------------
      if (ddot(n,work(1,D),1,work(1,D),1) .gt. pdr*pdr*urnorm) then
         ierr = 4
         maxits(1) = 0
         return
      end if 

c=======================================================================
c     Begin main loop of iterations.  Loop is equivalent to
c      do while (rnorm .gt. tolsq) .and. (its .le. maxits)
c=======================================================================

 2       continue
c
         call instr(A,colind,rwptr,n, x,b, work(1,wins1), alu,jlu,ju,
     1              pcmeth, first,its,iunit,instlvl)
         its = its + 1
c
         call bmux(n,n,work(1,D+next),work(1,W+next),A,colind,rwptr)
         t = ddot(n,r,1,work(1,W+next),1)
         work(next+1, WTW) = ddot(n,work(1,W+next),1,work(1,W+next),1)
         if (work(next+1, WTW) .lt. SMALL) go to 940
         alpha = t/work(next+1,WTW)
         call daxpy(n, alpha, work(1,D+next), 1, x, 1)
         call daxpy(n, -alpha, work(1,W+next), 1, r, 1)
         urnorm = ddot(n,r,1,r,1)
         if (urnorm .le. tolsq) go to 910
 3       continue
         call lusl(n,r,work(1,z),alu,jlu,ju,pcmeth, A,colind,rwptr)
         call bmux(n,n,work(1,z),work(1,v),A,colind,rwptr)
c
         do k = 1, kdim
            work(Beta+k-1,WtW) = - ddot(n,work(1,v),1,work(1,W+k-1),1)
     1                             /work(k,WtW)
         end do 

         next = next + 1
         if (next .eq. krylov) next =  0

c        -----------------------------------------------
c        Multiply out z = z + work(1:n,D:D+kdim-1) c
c                           * work(Beta:Beta+kdim-1,WtW)
c        -----------------------------------------------
c         call dgemv('N',n,kdim,1.0d0,work(1,D),n,work(Beta,WTW),1,
c     1               1.0d0, work(1,z),1)
c
         do k = 1, kdim
            call daxpy(n,work(Beta+k-1,WtW), work(1,D+k-1), 1,
     +                 work(1,z),1)
         end do

         call dcopy(n, work(1,z), 1, work(1,D+next),1)

         kdim = min(krylov,kdim+1)
         if (its. gt. maxits(1)) go to 920

c       -------------
c       End main loop
c       -------------
        go to 2

c=======================================================================
c           Stopping conditions
c=======================================================================

c     =====================
c     (Almost) exact guess
c     =====================
 900  continue
      ierr = 1
      maxits(1) = 0
      return

c     =========================================
c     Possible convergence; check true residual
c     =========================================
 910  continue
      ierr = 0
      resets = resets + 1
      if (resets .gt. 8) then
         ierr = 2
         go to 915
      end if
      call bmux(n,n,x,work(1,v),A,colind,rwptr)
      do k = 1, n
         work(k,v) = b(k) - work(k,v)
      end do
      urnorm = ddot(n,work(1,v),1,work(1,v),1)
      if (urnorm .gt. tolrnc(1)*tolrnc(1)) then
          tolsq = tolsq*tolrnc(1)*tolrnc(1)/urnorm
c          write(6,*)'>>>>> tolerance reset to ', sqrt(tolsq),'  <<<<<<'
          go to 3
      end if
 915  continue
      maxits(1) = its
      call instr(A,colind,rwptr,n, x,b, work(1,wins1), alu,jlu,ju,
     1           pcmeth, first,its,iunit,instlvl)
      return

c     ========================================
c     No convergence in max allowed iterations
c     ========================================
 920  continue
      ierr = 3
      maxits(1) = its
      call instr(A,colind,rwptr,n, x,b, work(1,wins1), alu,jlu,ju,
     1           pcmeth, first,its,iunit,instlvl)
      return

c     ======= 
c     wtw = 0 
c     ======= 
 940  continue
      ierr = 5
      maxits(1) = its
      call instr(A,colind,rwptr,n, x,b, work(1,wins1), alu,jlu,ju,
     1           pcmeth, first,its,iunit,instlvl)
      return

c================= End of ORTHOMIN =======================================

      end
