c========================================================================
      subroutine splib(A, colind, rwptr, n, 
     1                 x, b, work, space,
     2                 tol, ireltv, maxits, pcmeth, solmeth,
     3                 scale, Dl, Dr,
     4                 iparms, rparms, iredo, errflg, iunit, instlvl)
c========================================================================
c           *** SPLIB: sparse matrix iterative library ***
c========================================================================

c**********************************************************************
c                   SPLIB Copyright (C) 1995-1999                    **
c                     Indiana University                             **
c**********************************************************************

c
c  SPLIB is a package of iterative solvers for sparse nonsymmetric linear 
c  systems, and has 13 iterative methods and 7 preconditioning methods.
c  The purpose is to provide a collection that allows practical testing
c  on systems of nontrivial size without massive amounts of coding 
c  effort.  In particular, SPLIB relieves the user of the need to code
c  up preconditioners, which are often the most difficult and error-prone
c  part of iterative solvers.  The price to be paid is that the matrix
c  must be in a specific data structure: compressed sparse row.  This stores
c  the nonzero entries of A in a double precision singly dimensioned vector,
c  a corresponding integer vector of the column indices, and an integer
c  vector of length n+1 with pointers to the beginning of each row in the
c  matrix ( entry n+1 points to one position beyond the end of the arrays
c  for nonzero values and column indices).
c
c  The target users for SPLIB are scientists and engineers in application
c  areas who wish to experiment with iterative solvers in their codes to
c  find which of the many methods may be practical for them.  SPLIB is
c  also useful for teaching iterative methods, allowing students to 
c  implement methods without having to develop the massive amounts of code
c  needed for instrumentation, etc.
c
c  The code is written in Fortran 77, with a few exceptions.
c  Include files are used for some parameters and the memory manager.
c  One function, getr.c, is necessarily written in C.  That
c  function accesses the getrusage structure and returns it to the 
c  Fortran 77 routine prtres.  A timing function (mytime)
c  is used.  It is set to use etime(), but has commented out options
c  for using mclock(), second(), etc. on machines without etime().
c  BLAS routines used include: ddot, daxpy, dcopy, dgemv,
c  dasum, drot, dscal, drotg, and dtrsv.  Routines lsame and xerbla
c  from the LAPACK project are also provided in the directory "blas",
c  so that SPLIB can be compiled and run on machines without a native
c  BLAS library.  LAPACK also supplies the function dlamch, for finding
c  machine constants.  Finally, "implicit none" is used in all the routines.
c
c  This package differs from others in seven ways:
c    1) It uses a sophisticated memory managment module, which allows
c       the package to be easily adapted and used by applications
c       programs.  All they need do is provide a work array, from which
c       the necessary arrays are divvied out using pointers.  This helps
c       keep the calling sequence short.
c    2) It provides implementations of levels of fill LU preconditioning,
c       well documented and with various options.
c    3) The matrices dealt with are always kept in standard data structures,
c       so that they can be easily extracted and examined/processed using
c       other tools, such as emily for visualization or Sparskit for basic
c       manipulations.  
c    4) Detailed instrumentation of the code is provided for, at the risk of
c       losing some portability.  In particular, this code allows tracking
c       several different measures of error norms, including a backward
c       error analysis based measure.  Furthermore, the error norms are written
c       out to files which can then be plotted via xgraph, giving graphical
c       comparison of different methods and preconditioners.  Finally, the
c       norms are plotted against both iteration number and time, allowing
c       a more fair comparison than the usual "number of iterations" results.
c       The plotting against time is especially crucial, since methods can
c       vary drastically in the amount of work per iteration.
c    5) Further instrumentation provides access to rusage structures, which
c       allows information about memory utilization, page faults, context 
c       switches, and other important performance measures.
c    6) Implementation of the partial orthogonalization preconditioners of
c       Bramley, Gallivan, and Wang.  These generally produce better factors
c       than the ILU type methods, and are more robust.  However, they require
c       more intermediate storage to compute. 
c    7) Stringent tests for having too large of a preconditioned
c       residual, for dividing by small quantities, etc.
c
c  History of SPLIB:
c  -----------------
c
c  This code has evolved through many hands.  Originally it started
c  out as RP-PACK, for examining row projection methods.  It then
c  was converted to a front end for CG-like iterative methods.
c  Ulrike Meier and G. Skinner did most of the preliminary work, for
c  the Center for Supercomputer Research and Development at the University 
c  of Illinois.  They created a uniform calling sequence and data naming 
c  convention that tied many disparate codes together, as well as writing and
c  testing heroic amounts of codes.  Xiaoge Wang has done a great deal
c  of work, especially on the incomplete preconditioners based on levels
c  of fill-in, and all of the ECIMGS material.
c
c  Youcef Saad wrote the original GMRES(k) solver, and the package
c  heavily uses modified SPARSKIT routines provided by him and Edmond Chow.
c  The memory manager is a modification of one developed by Dominique
c  Pelletier of Ecole Polytechnique.  The machine constant function,
c  dlamch(), is from the LAPACK project headed by J. Dongarra.
c
c  Usage of SPLIB:
c  ---------------
c
c  Compilation is via the makefile.  There are two features to note:
c  1)  You should first edit the include file intdbl.inc, to give the
c      number of integer words per double precision word.  For most
c      machines this is intdbl = 2, but for Cray machines it is
c      generally intdbl = 1.  For 64 bit machines with long 
c      integers, this is also intdbl = 1.  This is used to force
c      double word alignment in the memory manager.
c  2)  The functions in dlamch() will compute machine epsilon and
c      safe invertible minimum with methods that cause under- and
c      over-flow.  If you have IEEE traps set, this will cause the
c      program to abend.  Turn off IEEE traps if they are by default
c      on for your Fortran compiler.
c  
c  Compilation sets up the archive library splib.a, which you can 
c  link with your code.
c
c  Calling SPLIB requires you to:
c  1)  Set up the matrix in CSR format.
c      The main sticking point in using SPLIB is that the matrix must be
c      provided in compressed sparse row (CSR) format.  This is necessary
c      because the preconditioners need to be able to access the 
c      entries of the matrix.  Unpreconditioned iterative methods based on 
c      Krylov subspaces generally only need to access the matrix via 
c      matrix-vector products, but they also generally do not work.  The 
c      tradeoff here is that most applications scientists find it relatively
c      easy to write a data structure conversion routine to put their 
c      coefficient matrix into a CSR format, but writing robust factorization
c      based preconditioners are much more difficult.
c  2)  Provide n-vectors x and b for the solution and right hand side: A x = b.
c  3)  Provide an integer array work(1:space), aligned on a double word
c      boundary.  All of the additional storage needed by SPLIB will be 
c      carved out from this array.  To assure double word alignment, you can 
c      declare it as a double precision array of length m, then pass it in 
c      with space = intdbl*m.
c  4)  Set the control parameters.  Defaults are in parenthesis, and are 
c      triggered by giving an invalid value for the parameter.
c  	  solmeth -- solution method (CG-Stabilized, solmeth = 5)
c  	  pcmeth  -- preconditioning method (No preconditioning, pcmeth = 0)
c  	  iredo   -- whether or not to recompute the preconditioner.
c                    (0 for recompute, 1 to not).
c  	  tol     -- tolerance on residual  (1.0e-6)
c  	  ireltv  -- whether or not to use relative residual for stopping
c                    (0 for not, 1 to use relative residual).
c  	  maxits  -- maximal allowed iterations (100)
c  	  scale   -- scaling of the system (0)
c  	  instlvl -- instrumentation level (0)
c      In addition, some methods require setting iparms and rparms.  For 
c      example, GMRES(m) requires a value m for the Krylov subspace size (but
c      this is set to 10 by default).  ILU(s) factorization requires a level
c      of fill value, which by default is 0.  Because of the wide variety of
c      methods available, it is necessary to check the individual routines
c      to find all the possibilities.
c  5)  If instrumentation is to be used, a vector of 16 logical unit numbers
c      must be provided.  SPLIB will open files with those unit numbers, to
c      write out summary data and Xgraph files tracing the iteration history.
c     
c------------------------------------------------------------------------------
c
c     Randall Bramley
c     Department of Computer Science
c     Indiana University, Bloomington
c     bramley@cs.indiana.edu
c     Wed Jul 12 10:03:18 EST 1995
c     Updated Mon Apr 12 12:59:46 EST 1999 for HTML output
c
c------------------------------------------------------------------------------
c
c  Parameters:
c  -----------
c
c  intdbl     : number of integer words per double precision word for
c               target machine.   This is used for alignment within
c               the memory manager routines.
c
c  Arguments:
c  ----------
c  n       : Order of the matrix A.
c  A       : Real*8 array for nonzero entries of matrix A; of
c            length nnz = rwptr(n+1)-1.
c  colind  : Integer array of column indices for entries in A; of
c            length nnz = rwptr(n+1)-1.
c  rwptr   : Integer array of pointers to beginning position of rows
c            in arrays A and colind.  Must be of length n+1.
c
c  x       : real*8 vector of unknowns to be solved for.
c  b       : real*8 right hand side vector of system to be solved: Ax = b.
c            See lrhs and lsol for options on this vector.
c
c  tol     : tolerance on norm used in stopping test.
c  maxits  : Maximum number of iterations allowed.  On return, contains actual
c            number performed.
c
c  solmeth : Solution method to use.
c             solmeth = 1  ---> BiCG
c             solmeth = 2  ---> CG on normal equations AA^Ty = b; x = A^Ty.
c             solmeth = 3  ---> CG on normal equations A^TA x = A^Tb
c             solmeth = 4  ---> CGS
c             solmeth = 5  ---> CGSTAB
c             solmeth = 6  ---> GMRES(k) 
c             solmeth = 7  ---> transpose-free QMR
c             solmeth = 8  ---> Bi-CG stabilized (Templates version of CG-Stab).
c             solmeth = 9  ---> Templates version of GMRES 
c             solmeth = 10 ---> Jacobi
c             solmeth = 11 ---> Gauss-Seidel
c             solmeth = 12 ---> SOR
c             solmeth = 13 ---> Orthomin(m)
c
c  pcmeth  : Preconditioning method to use.
c             pcmeth = 0  ---> no preconditioner.
c             pcmeth = 1  ---> ILU(s)
c             pcmeth = 2  ---> MILU(s,rpcprm)
c             pcmeth = 3  ---> ILUT(s,rpcprm)
c             pcmeth = 4  ---> SSOR(rpcprm)
c             pcmeth = 5  ---> TRID(s), where s is the block size.
c             pcmeth = 6  ---> ILU(-1), the space saver version of ILU(0)
c             pcmeth = 7  ---> ECIMGS.
c
c  rparms  : real preconditioner parameter, with interpretation depending
c            on preconditioning method to be used.
c             - tolerance for numerical dropping with ILUT;
c             - omega parameter for SSOR(omega);
c             - relaxation constant for MILU or ILU0.
c             - sparsity pattern indicator for ECIMGS
c                   rparms = 0.0        Use nonzero pattern of A for L/U
c                   rparms = 1.0        Use a diagonal matrix as precond.
c                   0.0 < rparms < 1.0  Use a pattern sparser than A.
c                   rparms > 1.0        Use int(rparms) levels of fill-in.
c
c  iparms  : integer parameters, with interpretation depending on methods used.
c             - iparms(1) is levels of fill for ILU(s), MILU(s), ILUT
c             - iparms(1) is block size for TRID.
c             - iparms(2) is maximum Kyrlov subspace size k
c                         for GMRES(k), Orthomin(k).
c
c  instlvl : Level of instrumentation.  The level is cummulative, so specifying
c                  output of Oetli/Prager norms implies residual norms, etc.
c             instlvl < 0  ---> no output from splib
c             instlvl = 0  ---> only summary data
c             instlvl = 1  ---> residual norm data
c             instlvl = 2  ---> preconditioned residual norm
c             instlvl = 3  ---> relative residual norm
c             instlvl = 4  ---> relative preconditioned residual norm
c             instlvl = 5  ---> Oetli/Prager norms
c             instlvl = 6  ---> error norms (if true solution vector is provided
c                               via a common block)
c
c  iunit   : vector of I/O unit numbers to use for output.
c            The I/O unit numbers are provided in the vector iunit(17).
c            The units are optional, depending on the error norm(s) you wish to
c            plot/use.  However, if you specify instlvl >= 0,
c            the following units will be opened for writing:
c            
c            iunit(1):    File for output of summary data
c            iunit(2):    Not used here (control parms in my driver)
c            iunit(3):    Not used here (input H/B matrix file in my driver)
c            iunit(4):    || Ax - b ||  vs. iter
c            iunit(5):    || inv(M) (Ax-b) || vs. iter
c            iunit(6):    || Ax - b || / || r_0 || vs. iter
c            iunit(7):    || inv(M) (Ax-b) || / || inv(M) r_0 || vs. iter
c            iunit(8):    || Ax - b ||_inf / ( ||A||_inf ||x||_1 + ||b||_inf)
c                             vs. iter
c            iunit(9):    || x - xstar || vs. iter
c            iunit(10):   Not used.
c            iunit(11):   || Ax - b || vs. time
c            iunit(12):   || inv(M) (Ax-b) || vs. time
c            iunit(13):   || Ax - b || / || r_0 || vs. time
c            iunit(14):   || inv(M) (Ax-b) || / || inv(M) r_0 || vs. time
c            iunit(15):   || Ax - b ||_inf / ( ||A||_inf ||x||_1 + ||b||_inf)
c                             vs. time
c            iunit(16):   || x - xstar || vs. time
c            iunit(17):   HTML description of SPLIB.
c
c  work    : Array of work storage, for preconditioners, solvers, and.
c            instrumentation routines.  The matrices and vectors used
c            are all indexed from this integer array, using pointers.
c
c  space   : maximum number of integer words allowed for all problem
c            dependent arrays.  This includes the preconditioners
c            and the additional work vectors needed in the solvers.
c            On return, space holds the actual amount of storage
c            required by SPLIB. 
c
c  iredo   : Recompute preconditioner if redo = 0, reuse if redo = 1.
c            This is useful if you wish to reuse a preconditioner from
c            an earlier call to SPLIB.  Converted internally to logical
c            variable redo
c
c  Variables (also see the descriptions in include file instr.inc):
c  --------------------------------------------------------------
c
c  r       : (int pointer) real*8 residual vector.
c
c  uptr, 
c  lu,jlu  : Data structure for the preconditioner M, with different meanings
c            for different preconditioners.  All three of uptr, lu, jlu are
c            simply pointers into the work() array in splib.f, but are
c            used as the names of the real*8 arrays when passed into the
c            subroutines called by splib.f.
c
c            ILU(s), MILU(s), or ILUt --->
c                 (lu,jlu) gives the combined L/U factors for the
c                 preconditioner in a modified sparse row (MSR) format.
c                 L is unit lower triangular and U is upper triangular.
c                 The diagonal of U is stored in lu(1:n) inverted, to 
c                 replace divisions in the algorithm with multiplications.
c                 Each i-th row of the combined L/U matrix in the (lu,jlu)
c                 data structure contains the i-th row of L (excluding
c                 the diagonal entry = 1), followed by the i-th row of u.
c                 uptr is a integer vector of length n containing pointers
c                 to the start position of each row of U in (lu, jlu).
c
c            SSOR(omega)   --->
c                 lu is the scalar acceleration parameter omega.
c                 jlu is not used, and uptr is a vector of pointers to the
c                 diagonal entries in the CSR data structure for A.
c
c            TRID(s)   --->
c                 This implements block diagonal preconditioning, where
c                 each diagonal block is from the corresponding tridiagonal
c                 part of the matrix A.  The size of the blocks is given
c                 implicitly in the integer iskip, stored in jlu(n).
c                 Factors are set up for contiguous blocks of iskip rows
c                 and so by using iskip = n the factor can be for the entire
c                 matrix tridiagonal part.  Using iskip = 1 is an 
c                 inefficient way of performing diagonal preconditioning,
c                 while 1 < iskip < n is for a block tridiagonal 
c                 preconditioner.  Gaussian elimination with pivoting is 
c                 used for factoring each block, so it is necessary to store
c                 the pivot sequence in jlu(1:n-1).  lu has the four vectors
c                 giving the nonzeros of the tridiagonal preconditioner.
c                 Vector uptr is not used for TRID.
c
c            ILU0   --->
c                 This implements the version of ILU(0) that does not change
c                 any off-diagonal entries of A, and so only needs to store
c                 the n-vector of diagonal pivots.  lu(1:n) stores those 
c                 pivots, and jlu(1:n) contains pointers to the diagonal
c                 entry in each row of A.
c
c
c  next    : points to next available location in array work()
c  remain  : number of integer words remaining in work() to be used.
c
c  ireltv  : variable to indicate whether stopping test is relative or not.
c             converted internally to logical variable reltv
c  iters   : Actual number of iterations used, returned by solver.
c            Second component contains Krylov subspace size for GMRES.
c  krylov  : Size of Krylov subspace to use in GMRES; k above.
c  ipcprm  : integer preconditioning parameter, usually levels of fill.
c            For tridiag preconditioning, it means the block size.
c  nzlu    : Integer giving number of nonzeros in preconditioner.
c  mneed   : Integer parameter from preconditioner call indicating
c            the amount of additional integer word storage needed.
c            Warning: this is only an approximation in general.
c  nnz     : Number of nonzero entries in A.
c
c  errflg  : Integer variable indicating error conditions.  For the
c            solvers, the interpretation is: 
c             errflg = -m --> Not enough workspace provided; m more double
c                             precision words needed in array work().
c             errflg = 0  --> Successful return; tolerance achieved.
c             errflg = 1  --> The initial guess satisfies the stopping test.
c                             This is often a clue that rhs not set properly.
c             errflg = 2  --> Tolerance reset 8 times in attempt to
c                             get true residual small enough.
c             errflg = 3  --> Convergence not achieved in maxits iterations.  
c             errflg = 4  --> Preconditioned residual is too large relative
c                             to the unpreconditioned residual.  Indicates
c                             an unstable preconditioner.
c             4 < errflg < 9 --> Breakdown condition, usually from an attempt to
c                                divide by a near-zero number in algorithm.
c             errflg = 10+k  --> Error condition k occured in preconditioner.
c             Note that any of conditions 0, 1, 2 may mean success.
c
c  res     : Final residual norm (2-norm)
c  err     : Final error norm (2-norm)
c  bnorm   : Two-norm of the right hand side vector b.  Used for relative
c            residual stopping tests: ||r_k||/||r_0|| < tol.
c  sptime  : Integer pointer to array of times in memory manager.
c            Used for interacting with CADYF program.
c
c  scale   : Job control parameter for scaling of system.
c             scale = 0  ---> perform no scaling
c             scale = 1  ---> set max(max(A)) = 1.0
c             scale = 2  ---> set max of each row to be 1.0
c             scale = 3  ---> scale each column so diagonal entry is 1.0
c             scale = 4  ---> perform first scale=2, then scale=3
c            If scale = 0 or 1, Dl and Dr can be dummy variables.
c            If scale = 2 or 4, Dl must provided (see below)
c            If scale = 3 or 4, Dr must provided (see below)
c            Note: SPLIB will unscale the solution vector so that
c                  it is the solution to the original system when
c                  scale = 3 or 4.
c
c  Dl, Dr  : Double arrays of length n, used to hold scaling of matrix.
c            system actually solved is Dl*A*Dr y = Dl b.  On return, the
c            matrix A has been scaled and x = Dr*y.
c
c  Instrumentation Common Block Variables:
c  ---------------------------------------
c
c  xstar   : Vector with true solution, for error norm checking.
c  title   : Character array to hold title of Harwell/Boeing matrix.
c  key     : Character array for short title of Harwell/Boeing matrix.
c  times   : vector containing timing data, for plotting convergence
c            history versus time.
c         times(1) --> time spent in computing the preconditioner.
c         times(2) --> time spent in iterative solver, including pc solves
c         times(3) --> time spent in A*v multiplications
c                        (includes instrumentation calls to A*v)
c         times(4) --> time spent in A^T*v multiplications 
c                        (includes instrumentation calls to A^T*v)
c         times(5) --> time spent in applying preconditioner
c                        (includes instrumentation calls to lusl)
c         times(6) --> time spent in remainder of solver  (not yet done)
c         times(7) --> time to be excluded because spent in instrumentation.
c         times(8) --> time to be excluded because of other initializations,
c                    or because of earlier runs with the same matrix.
c  
c
c  I/O Files
c  ---------
c
c  results : File containing summary of results from run.
c  About_SPLIB.html : File containing description of SPLIB
c
c  resi    : File with residual norm vs. iteration number.
c  rest    : File with residual norm vs. elapsed time.
c
c  reli    : File with relative residual norm vs. iteration number.
c  relt    : File with relative residual norm vs. elapsed time.
c
c  Mresi   : File with preconditioned residual norm vs. iteration number.
c  Mrest   : File with preconditioned residual norm vs. elapsed time.
c
c  Mreli   : File with preconditioned relative residual norm 
c            vs. iteration number.
c  Mrelt   : File with preconditioned relative residual norm 
c            vs. elapsed time.
c
c  oetli   : File with Oettli and Prager norm vs. iteration number.
c  oetlt   : File with Oettli and Prager norm vs. elapsed time.
c
c  errori  : File with error norm vs. iteration number (if lsol = true).
c  errort  : File with error norm vs. time (if lsol = true).
c
c
c  External Routines
c  -----------------
c
c  bmux    : Sparse matrix multiply (Sparskit)
c  ddot    : Blas routine for dotproduct
c  dcopy   : Blas routine for copy of a vector
c  dasum   : Blas routine for max norm of vector
c  dnrm2   : Blas routine for two-norm of vector
c  dgemv   : Blas routine for matrix-vector product
c
c  Workspace
c  ---------
c
c  The amount of workspace depends on the methods used, the levels
c  of fill-in (and way of defining fill-in levels), and the number
c  of Krylov subspace vectors used.  The following guide excludes 
c  the storage for x, r, and b which appear in the calling
c  sequence for all the routines.  Note that GMRES(k), GMREST(k), and 
c  ORTHOMIN(k) depend on the size k of the Krylov subspace.  Also note
c  that the methods with fill-in will require more storage for lu, jlu,
c  uptr, but the actual amount is not predictable in advance.  The size
c  shown is for the preconditioners with no fill-in allowed, and does
c  not include the storage that depends on nnz.  Even for zero levels
c  of fill, ECIMGS requires intermediate storage that is unpredictable.
c
c     Routine		Reals		Integers
c     -------		-----		--------
c
c     (m)ilus 		n          	2n
c     ilut    		2n + 1		3n
c     ssor    		1		n (but uptr is recycled for this)
c     trid    		4n		n
c     ilu0    		n		n
c     ecimgs  		?		? (problem dependent)
c
c     Routine		Reals		
c     -------		-----		
c     bicg		6n              }
c     cgne		3n              }
c     cgnr		3n              }  Add 2n for all of these if
c     cgs 		6n              }  the instrumentation routine
c     cgstab 		6n              }  is to be used at all; the
c     gmres(k)		k*(n+k+5)+n + 1 }  extra is for work arrays
c     qmrtf   		10n             }  used in residual computation.
c     bicgs	 	6n              }
c     orthomin(k) 	(2k+6)*n 	}
c     jacobi             0		}
c     gauss              0		}
c     SOR                0		}
c

c========================================================================
c  Memory management module
c========================================================================
      implicit none

      include 'include/memory.inc'
      include 'include/intdbl.inc'

c========================================================================
c   Declarations of variables
c========================================================================
      include 'include/symbols.inc'
c
c     Double precision arrays and pointers:
c     -------------------------------------
      integer n, space
      real*8 a(*), x(n), b(n), rparms(*)
      real*8 Dl(n), Dr(n)
      integer lu, r, work(space), errflg, wu, wl

c     Double precision scalars:
c     -------------------------
      real*8 tol, rpcprm, loctol
      real*8 res,bnorm

c     Timing variables:
c     -----------------
      real*8 mytime

c     Function programs from BLAS:
c     ----------------------------
      real*8 ddot,dasum,dnrm2
      external ddot,dasum,dnrm2

c     Function programs from memory manager:
c     -------------------------------------
      integer addtbl, loctbl, sizeof, storof, getnext, memleft
      logical istbl
      external addtbl, loctbl, sizeof, storof, istbl
      external getnext, memleft

c     Other function programs
c     -----------------------
      real*8 Oetli
      external Oetli

c     Integer arrays:
c     ---------------
      integer rwptr(n+1), colind(*)

c     Integer array pointers:
c     -----------------------
      integer uptr, jlu, diag, itvecs
      integer next, remain
      integer jr, jwl, jwu, ofclu
      integer sptime

c     Integers:
c     ---------
      integer ierr, maxits, nnz, iparms(*)
      integer iters(2), mneed, wksz, loctot, num_entries
      integer k
      integer pcmeth, solmeth, krylov, nzlu
      integer alloc
      integer scale
      integer storage

c     Character variables:
c     --------------------
      character*8 from

c     Logical variables:
c     ------------------
      logical reltv, redo, results, nogood, force, ofl
      integer ireltv, iredo

c     Instrumentation variables:
c     --------------------------
      integer iunit(17)
      integer instlvl
      include 'include/heads.inc'

c========================================================================
c      Beginning of Executable Statements
c========================================================================

c     ------------------------------------------------------------
c     Convert to logical variables the redo and reltv input vars.
c     ------------------------------------------------------------
      redo = .true.
      if (iredo .eq. 1) redo = .false.
      reltv = .true.
      if (ireltv.eq. 0) reltv = .false.

c     ------------------------------------------------------------
c     Determine if any printing is to be done by checking iunit(1)
c     ------------------------------------------------------------
      results = (instlvl .ge. 0)

c     -------------------------------------------------------------
c     Initialize memory management.  Have it write to results file.
c     -------------------------------------------------------------
      call initbl(iunit(1),space)
      storage = getnext()

c     ----------------------------------------------
c     Set up array to return times to CADYF program.
c     ----------------------------------------------
      from = 'splib'
      if (.not. istbl('sptime', from )) then
         sptime = addtbl('sptime', 8, 'real', from)
      else
         sptime = loctbl('sptime', from)
      end if

c     ------------------------------------------
c     Set pointer to scaling vectors for system.
c     ------------------------------------------
c      if (.not. istbl('Dl', from )) then
c        Dl = addtbl('Dl', n, 'real', from)
c      else
c        Dl = loctbl('Dl', from)
c      end if
c      if (.not. istbl('Dr', from )) then
c        Dr = addtbl('Dr', n, 'real', from)
c      else
c        Dr = loctbl('Dr', from)
c      end if
      
c     --------------------------------------------------
c     Set pointer to next available location in work():
c     --------------------------------------------------
      next = getnext()
      remain = memleft()

c     ------------------------------------------------------
c     Call getrusage data to initialize resource usage file:
c     ------------------------------------------------------
      if (results) call prtres(0,iunit(1))
      ierr = 0

c     -----------------------------------------------------------
c     Set up strings giving solution method, preconditioner names
c     -----------------------------------------------------------
      call strings(solmeth, solver, pcmeth, precond)

c     ----------------------------------------------------------
c     Create an HTML page describing SPLIB; the output results
c     file (if any) will have links to this page as explanation.
c     ----------------------------------------------------------
      if (results) then
         call descr(iunit(17))
      end if

c     ----------------------------------------------------------------
c     Set number of nonzeros, and unpack parameter for preconditioners
c     ----------------------------------------------------------------
      nnz = rwptr(n+1) - 1
      rpcprm = rparms(1)
      ipcprm = iparms(1)
      iters(1) = maxits
      if (solmeth .eq. sGMRES .or.
     1    solmeth .eq. sGMRESt .or.
     2    solmeth .eq. sORTHM) then
           krylov = iparms(2)
      end if

c     ---------------------------------------
c     Compute norm of right hand side vector.
c     ---------------------------------------
      bnorm = dnrm2(n,b,1)

c     -----------------------------
c     Echo some info to output file
c     -----------------------------
      if (results) then
          write(iunit(1),2040)
          write(iunit(1),8000)
          write(iunit(1),8005)
          write(iunit(1),6010)
          write(iunit(1),6020) n, nnz
          write(iunit(1),2020) bnorm

          write(iunit(1),6050) solver,precond
          if (solmeth .eq. sGMRES .or.
     1        solmeth .eq. sGMRESt .or.
     2        solmeth .eq. sORTHM) then
            write(iunit(1),6030) krylov
          end if
          write(iunit(1),6060) ipcprm
          write(iunit(1),6090) rpcprm
          write(iunit(1),6040) tol
          write(iunit(1),6070) maxits
          if (scale .eq. NOSC)  write(iunit(1),7000)
          if (scale .eq. MXSC)  write(iunit(1),7010)
          if (scale .eq. RWSC)  write(iunit(1),7020)
          if (scale .eq. COLSC) write(iunit(1),7030)
          if (scale .eq. RCSC)  write(iunit(1),7040)
          write(iunit(1),7045)
      end if 

c================================================================================
c     Set default values if not specified by calling routine,
c     and perform some simple validity checks.
c================================================================================

c     ------------------------------------------
c     Valid solution method?  Use CGSTAB if not.
c     ------------------------------------------
      if (solmeth .le. 0 .or. solmeth .gt. 14) then
         if (results) write(iunit(1),3000)
         solmeth = 5
      end if 

c     ---------------------------------------
c     Valid precond method?  Use NoPC if not.
c     ---------------------------------------
      if (pcmeth .lt. pNOPC .or. pcmeth .gt. pEIMGS ) then
         if (results) write(iunit(1),3010)
         pcmeth = pNOPC
      end if

c     -------------------------------------------------
c     Valid maximal iterations allowed?  Use 100 if not.
c     -------------------------------------------------
      if (maxits .le. 0 ) then
         if (results) write(iunit(1),3020)
         iters(1) = 100
      end if

c     ------------------------------------------------
c     Set local tolerance and then check for valid
c     tolerance on residual.  Use 1.0d-6 if not valid.
c     ------------------------------------------------
      loctol = tol
      if (tol .lt. 0.0d0) then
         if (results) write(iunit(1),3030) tol
         loctol    = 1.0d-6
      end if

c     ----------------------------------------------------------------
c     If Jacobi, Gauss-Seidel, or SOR then cannot use preconditioning.
c     ----------------------------------------------------------------
      nogood = ((solmeth .eq. 10) .or. (solmeth .eq. 11) .or.
     1          (solmeth .eq. 12)) .and. (pcmeth .ne. pNOPC)
      if (nogood) then
         if (results) write(iunit(1),3040) 
         pcmeth = pNOPC
      end if

c     ---------------------------------------------------------------
c     If GMRES, GMRESt, or Orthomin then Krylov subspace size must be
c     at least equal to one.  Reset to 10 otherwise.
c     ---------------------------------------------------------------
      nogood = (solmeth .eq. 6 .or. solmeth .eq. 9 .or. solmeth .eq. 13)
     1         .and. (krylov .le. 0)
      if (nogood) then
         if (results) write(iunit(1),3050) 
         krylov = 10
      end if

c     ----------------------------------------------------------
c     Set up space for the residual vector r if not already done
c     ----------------------------------------------------------
      if (.not. istbl('r', from)) then
         r = addtbl('r', n, 'real', from)
      else
         r = loctbl('r', from)
      end if
      remain = memleft()
      next = getnext()
      if (remain .le. 0) then
         errflg = -1.0d6
         if (results) write(iunit(1),5040) next
         return
      end if

c     ----------------------------------------
c     Scale the matrix A and right hand side b
c     ----------------------------------------
      if (scale .lt. NOSC .or. scale .gt. RCSC) scale = NOSC
      if (scale .eq. MXSC) call scala(MXSC, A, rwptr, colind, n, b, Dl)
      if (scale .eq. RWSC .or. scale .eq. RCSC)
     &    call scala(RWSC, A, rwptr, colind, n, b, Dl)
      if (scale .eq. COLSC .or. scale .eq. RCSC)
     &    call scala(COLSC, A, rwptr, colind, n, b, Dr)


c     ------------------------
c     Initialize timing vector
c     ------------------------
      do k = 1, 8
         times(k) = 0.0d0
      end do
      times(1) = mytime()

c     --------------------------------------------------------
c     Time to exclude from instrumentation routines, because
c     it was spent in initialization or other solution methods
c     for the same matrix.
c     --------------------------------------------------------
      times(8) = mytime()

c=======================================================================
c             *** Preconditioner Computation ***
c=======================================================================

c     -----------------------------------------------------------------
c     If we are to reuse the preconditioner from a previous run, skip
c     computation of it.  Otherwise need to go through and set it up.
c     Later extend this to whether or not to compute structure as well.
c     -----------------------------------------------------------------
      if (redo .or. .not. istbl('lu',from)) then

c         ---------------------------------------------------
c         Free up space used by the previous preconditioner
c         ---------------------------------------------------
          if (istbl('lu',from)) 
     1        call rmvtbl(work(1),'lu',from)
          if (istbl('jlu',from)) 
     1        call rmvtbl(work(1),'jlu',from)
          if (istbl('uptr', from)) 
     1        call rmvtbl(work(1),'uptr' ,from)

c         -----------------------------------------------------------------
c         Call the routines to set up the preconditioner.
c         Note that ilus must align everything on double word boundaries,
c         so it is assumed that mneed is multiple of intdbl if ierr .ne. 0.
c         First set pointers in case no preconditioner is used.
c         -----------------------------------------------------------------
          next = getnext()
          remain = memleft()
          uptr = next
          jlu = next
          lu = next
          nzlu = 0

c         --------------------------------------
c         Preconditioner set up for ILU or MILU:
c         --------------------------------------
          if (pcmeth .eq. pILUS) then
	     call ilus (n,a,colind,rwptr,
     1                  ipcprm,0.0d0,
     2                  lu,jlu,uptr,work(next),remain,
     3                  ierr,mneed)
          end if 

          if (pcmeth .eq. pMILU) then
	     call ilus (n,a,colind,rwptr,
     1                  ipcprm,rpcprm,
     2                  lu,jlu,uptr,work(next),remain,
     3                  ierr,mneed)
          end if 

c         -----------------------------------------------------------
c         Update memory management pointers if pcmeth is ILU or MILU:
c         -----------------------------------------------------------
          if ((pcmeth .eq. pILUS .or. pcmeth .eq. pMILU)
     1                                .and. ierr .eq. 0) then
              nzlu = work(jlu+next-1+n) - 1
              mneed = (n + (1 + intdbl)*nzlu) - remain
              if (mneed .gt. 0) then
                 ierr = 1
                 go to 200
              end if
              from = '(M)ILU'
              uptr = addtbl('uptr', n, 'integer', from)
              jlu  = addtbl('jlu', nzlu, 'integer', from)
              lu   = addtbl('lu', nzlu, 'real', from)
          end if

c         -------------------------------
c         Preconditioner set up for ILUT:
c         -------------------------------
          if (pcmeth .eq. pILUT) then
             next = getnext()
             remain = memleft()
             loctot = next + remain - intdbl
             mneed = 2*intdbl *(n+1) + 4*n - remain
             if (mneed .gt. 0) then
                ierr = 1
                go to 200
             end if

c            -----------------------------------------------------
c            Make sure allocated length n is a multiple of intdbl
c            -----------------------------------------------------
             alloc = n + mod(n,intdbl)

c            -------------------------------------------------------
c            The integer work arrays need not be double word aligned,
c            since they will be ignored outside of ilut.f
c            -------------------------------------------------------
             jwu = loctot - alloc
             jwl = jwu - alloc
             jr =  jwl - alloc

c            ------------------------------------------------------------
c            The real*8 arrays need to be aligned, since they might cause
c            data alignment errors within ilut.f otherwise.
c            ------------------------------------------------------------
             alloc = intdbl*(n+1)
             wl = jr - alloc - intdbl
             wl = wl + mod(wl-1,intdbl)
             wu = wl - alloc - intdbl
             wu = wu + mod(wu-1,intdbl)

c            ------------------------------------------------------
c            Unless ilut fails, uptr will be allocated permanently,
c            so do it here.
c            ------------------------------------------------------
             from = 'ILUT'
             uptr    = addtbl('uptr',n,'integer',from)

c            ------------------------------------------------------------------
c            Partition the rest of available memory between jlu and lu, keeping
c            in mind that for the same number of entries lu requires twice as
c            many integer words.
c            ------------------------------------------------------------------
             next = getnext()
             remain = memleft()
             num_entries = (wu - next)/(intdbl + 1)
             jlu = next
             lu  = next + num_entries
             lu = lu + mod(lu-1,intdbl)

             call ilut(n, a, colind, rwptr,
     &                 ipcprm, rpcprm,
     &                 work(lu), work(jlu), work(uptr), num_entries,
     &                 work(wu), work(wl), work(jr), work(jwl),
     &                 work(jwu), ierr)

             if (ierr .gt. 0) then
                if (results) write(iunit(1),5000) ierr
                go to 100
             end if 
             if (ierr .eq. -1 .or. ierr .eq. -5) then
                if (results) write(iunit(1),5030) ierr 
                ierr = -ierr
                go to 100
             end if
             if (ierr .eq. -4) then
                if (results) write(iunit(1),5033) ipcprm
                ierr = -ierr
                go to 100
             end if
             if (ierr. ne. 0) then
                mneed = (1 + intdbl) * ( 2*n*(ipcprm+1) + nnz)
                nzlu = work(jlu+n)-1
                go to 100
             end if

c            ------------
c            Allocate jlu
c            ------------
             from = 'ILUT'
             mneed = num_entries*(1+intdbl) - remain
             if (mneed .gt. 0) then 
                ierr = 1
                go to 200
             end if
             k = addtbl('jlu', num_entries, 'integer', from)

c            ------------------------------------------------------
c            Allocate the lu array, and then shift the computed one
c            up in the work array to fill in any empty spaces.
c            ------------------------------------------------------
             ofclu = addtbl('lu', num_entries, 'real', from)
             call dcopy(num_entries, work(lu), 1, work(ofclu), 1)
             lu = ofclu

c            -----------------
c            End of ILUT setup
c            -----------------
 100         continue
          end if

c         -------------------------------
c         Preconditioner set up for SSOR:
c         -------------------------------
          if (pcmeth .eq. pSSOR) then
             nzlu = n
             mneed = (intdbl + n) - remain
             if (mneed .gt. 0) then
                 ierr = 1
                 go to 200
             end if
             from = 'SSOR'
             jlu  = addtbl('jlu', n, 'integer', from)
             lu   = addtbl('lu', 1, 'real', from)
             uptr = jlu
             call ssor(n,a,colind,rwptr,rpcprm,work(lu),
     &                 work(jlu),work(jlu),work(next),mneed,ierr)
          end if

c         ---------------------------------
c         Preconditioner set up for TRDIAG:
c         ---------------------------------
          if (pcmeth .eq. pTRID) then
             nzlu = 4*n
             mneed = (n + (1 + intdbl)*nzlu) - remain
             if (mneed .gt. 0) then
                ierr = 1
                go to 200
             end if
             from = 'TRDIAG'
             jlu  = addtbl('jlu', n, 'integer', from)
             lu   = addtbl('lu', nzlu, 'real', from)
             uptr = jlu
             call tridiag(n,a,colind,rwptr,
     &                    work(lu),work(jlu),ipcprm,ierr)
          end if

c         ----------------------------------
c         Preconditioner set up for ILU(-1):
c         ----------------------------------
          if (pcmeth .eq. pILU0) then
             nzlu = n
             mneed = (n + (1 + intdbl)*nzlu) - remain
             if (mneed .gt. 0) then
                ierr = 1
                go to 200
             end if
             from = 'ILU0'
             jlu  = addtbl('jlu', n, 'integer', from)
             lu   = addtbl('lu', nzlu, 'real', from)
             uptr = jlu
             call ilu0(n, a, colind, rwptr, rpcprm, 
     &                 work(lu), work(jlu), ierr)
          end if

c         -------------------------------
c         Preconditioner set up for EIMGS
c         --------------------------------
          if (pcmeth .eq. pEIMGS) then
             call ecimgs(n,a,colind,rwptr, rpcprm, jlu,
     1                   lu, uptr, work(next),
     2                   remain, ierr, mneed)
             nzlu  = work(jlu+n+next-1) - 1
             mneed = (n + (1 + intdbl)*nzlu) - remain
             if (mneed .gt. 0) then
                ierr = 1
                go to 200
             end if
             from = 'ECMGS'
	     uptr  = addtbl('uptr', n, 'integer', from)
	     jlu   = addtbl('jlu', nzlu, 'integer', from)
	     lu    = addtbl('lu', nzlu, 'real', from)
          end if

          next = getnext()
          remain = memleft()

c         -------------------------------------------------
c         Error conditions from computing preconditioners:
c         -------------------------------------------------

 200      continue

          if (ierr .ne. 0) then

             errflg = 10 + ierr
c            -----------------------------------------------------
c            Error codes in TRID differ from other preconditioners
c            -----------------------------------------------------
             if (pcmeth .eq. pTRID .and. results) then
                write(iunit(1),4040)
                return
             end if 

c            -----------------------
c            Small pivot encountered
c            -----------------------
             if (ierr .lt. 0 .and. results) write(iunit(1),4090) ierr

c            -----------------
c            Not enough memory
c            -----------------
             if (ierr .eq. 1 .and. results) then
                write(iunit(1),4080) mneed
             end if 

c            -----------------
c            Illegal parameter
c            -----------------
             if (ierr .eq. 2 .and. results) write(iunit(1),4070)

c            ----------------------------------------------------------
c            List memory usage and clean up mess in case of error in PC
c            ----------------------------------------------------------
             from = 'PCFail'
             if (results) call listbl(from)
             from = 'splib'
             if (istbl('lu',from)) 
     1             call rmvtbl(work(1),'lu',from)
             if (istbl('jlu',from)) 
     1             call rmvtbl(work(1),'jlu',from)
             if (istbl('uptr',from)) 
     1             call rmvtbl(work(1),'uptr',from)
             return
          else 

c            ------------------------------------------------------
c            No error in preconditioner; output memory requirements
c            ------------------------------------------------------
          end if

c      ----------------------------------------------------
c      If not recomputing preconditioner, just loctbl them:
c      ----------------------------------------------------
       else
          lu   = loctbl('lu'  , from)
          jlu  = loctbl('jlu' , from)
          uptr = loctbl('uptr', from)

c      --------------------------------------------------
c      End if block for (re)computing the preconditioner
c      --------------------------------------------------
       end if

c     ==================================================================
c             *** End Preconditioner Computation ***
c     ==================================================================

c     ----------------------------------
c     Time for computing preconditioner:
c     ----------------------------------
      times(1) = mytime() - times(1)

c     ==================================================================
c             *** Call Solver Routine ***
c     ==================================================================

      times(2) = mytime()
      wksz = memleft()/intdbl
c
      if (solmeth .eq. sBICG) then
        k = 6*n
        if (memleft() - intdbl*k .lt. 0) then
           errflg = memleft() - intdbl*k
           return
        end if 
        from = 'BiCG'
	itvecs   = addtbl('itvecs', k, 'real', from)
        call bicg(
     1            a,colind,rwptr,n,
     2            x,work(r),b,
     3            loctol,reltv,iters,ierr,
     4            work(next),wksz,
     5            work(lu),work(jlu),work(uptr),
     6            pcmeth, iunit, instlvl)
      end if

      if (solmeth .eq. sCGNE)  then
        k = 3*n
        if (memleft() - intdbl*k .lt. 0) then
           errflg = memleft() - intdbl*k
           return
        end if 
        from = 'CGNE'
	itvecs   = addtbl('itvecs', k, 'real', from)
        call cgne(
     1            a,colind,rwptr,n,
     2            x,work(r),b,
     3            loctol,reltv,iters,ierr,
     4            work(next),wksz,
     5            work(lu),work(jlu),work(uptr),
     6            pcmeth, iunit, instlvl)
      end if

      if (solmeth .eq. sCGNR) then
        k = 3*n
        if (memleft() - intdbl*k .lt. 0) then
           errflg = memleft() - intdbl*k
           return
        end if 
        from = 'CGNR'
	itvecs   = addtbl('itvecs', k, 'real', from)
        call cgnr(
     1            a,colind,rwptr,n,
     2            x,work(r),b,
     3            loctol,reltv,iters,ierr,
     4            work(next),wksz,
     5            work(lu),work(jlu),work(uptr),
     6            pcmeth, iunit, instlvl)
      end if

      if (solmeth .eq. sCGS) then
        k = 6*n
        if (memleft() - intdbl*k .lt. 0) then
           errflg = memleft() - intdbl*k
           return
        end if 
        from = 'CGS'
	itvecs   = addtbl('itvecs', k, 'real', from)
        call cgs (
     1            a,colind,rwptr,n,
     2            x,work(r),b,
     3            loctol,reltv,iters,ierr,
     4            work(next),wksz,
     5            work(lu),work(jlu),work(uptr),
     6            pcmeth, iunit, instlvl)
      end if

      if (solmeth .eq. sCGSTAB) then
        k = 6*n
        if (memleft() - intdbl*k .lt. 0) then
           errflg = memleft() - intdbl*k
           return
        end if 
        from = 'CGStab'
	itvecs   = addtbl('itvecs', k, 'real', from)
        call cgstab(
     1            a,colind,rwptr,n,
     2            x,work(r),b,
     3            loctol,reltv,iters,ierr,
     4            work(next),wksz,
     5            work(lu),work(jlu),work(uptr),
     6            pcmeth, iunit, instlvl)
      end if

      if (solmeth .eq. sGMRES) then
         iters(2) = krylov
         k = krylov*(n+krylov+5)+n
        if (memleft() - intdbl*k .lt. 0) then
           errflg = memleft() - intdbl*k
           return
        end if 
         from = 'GMRES'
	 itvecs   = addtbl('itvecs', k, 'real', from)
         call gmres(
     1            a,colind,rwptr,n,
     2            x,work(r),b,
     3            loctol,reltv,iters,ierr,
     4            work(next),wksz,
     5            work(lu),work(jlu),work(uptr),
     6            pcmeth, iunit, instlvl)
      end if

      if (solmeth .eq. sQMR) then
        k = 6*n
        if (memleft() - intdbl*k .lt. 0) then
           errflg = memleft() - intdbl*k
           return
        end if 
        from = 'QMR'
	itvecs   = addtbl('itvecs', k, 'real', from)
        call qmrtf(
     1            a,colind,rwptr,n,
     2            x,work(r),b,
     3            loctol,reltv,iters,ierr,
     4            work(next),wksz,
     5            work(lu),work(jlu),work(uptr),
     6            pcmeth, iunit, instlvl)
      end if

      if (solmeth .eq. sBICGS) then
        k = 6*n
        if (memleft() - intdbl*k .lt. 0) then
           errflg = memleft() - intdbl*k
           return
        end if 
        from = 'BiCGS'
	itvecs   = addtbl('itvecs', k, 'real', from)
        call bicgs(
     1            a,colind,rwptr,n,
     2            x,work(r),b,
     3            loctol,reltv,iters,ierr,
     4            work(next),wksz,
     5            work(lu),work(jlu),work(uptr),
     6            pcmeth, iunit, instlvl)
      end if

      if (solmeth .eq. sGMRESt) then
         iters(2) = krylov
         k = krylov*(n+krylov+5)+n
         from = 'GMRESt'
	 itvecs   = addtbl('itvecs', k, 'real', from)
         call gmrest(
     1            a,colind,rwptr,n,
     2            x,work(r),b,
     3            loctol,reltv,iters,ierr,
     4            work(next),wksz,
     5            work(lu),work(jlu),work(uptr),
     6            pcmeth, iunit, instlvl)
      end if

      if (solmeth .eq. sJACOBI) then
        k = n
        if (memleft() - intdbl*k .lt. 0) then
           errflg = memleft() - intdbl*k
           return
        end if 
        from = 'jacobi'
	itvecs   = addtbl('itvecs', k, 'real', from)
	if (.not. istbl('diag', from)) then
            diag    = addtbl('diag', n, 'integer', from)
            next = getnext()
            wksz = memleft()/intdbl
	end if 
        call jacobi(
     1            a,colind,rwptr,n,
     2            x,work(r),b,
     3            loctol,reltv,iters,ierr,
     4            work(next),wksz,
     5            work(lu),work(jlu),work(diag),
     6            pcmeth, iunit, instlvl)
      end if

      if (solmeth .eq. sGAUSS) then
        k = n
        if (memleft() - intdbl*k .lt. 0) then
           errflg = memleft() - intdbl*k
           return
        end if 
        from = 'gauss'
	itvecs   = addtbl('itvecs', k, 'real', from)
	if (.not. istbl('diag', from)) then
            diag = addtbl('diag', n, 'integer', from)
            next = getnext()
            wksz = memleft()/intdbl
	end if 
        call gauss(
     1            a,colind,rwptr,n,
     2            x,work(r),b,
     3            loctol,reltv,iters,ierr,
     4            work(next),wksz,
     5            work(lu),work(jlu),work(diag),
     6            pcmeth, iunit, instlvl)
      end if

      if (solmeth .eq. sSUCR) then
        k = n
        if (memleft() - intdbl*k .lt. 0) then
           errflg = memleft() - intdbl*k
           return
        end if 
        from = 'sor'
	itvecs   = addtbl('itvecs', k, 'real', from)
	if (.not. istbl('diag', from)) then
            diag = addtbl('diag', n, 'integer', from)
            next = getnext()
            wksz = memleft()/intdbl
	end if 
        call sor(
     1            a,colind,rwptr,n,
     2            x,work(r),b,
     3            loctol,reltv,iters,ierr,
     4            work(next),wksz,
     5            rparms,work(jlu),work(diag),
     6            pcmeth, iunit, instlvl)
      end if

      if (solmeth .eq. sORTHM) then
        from = 'orthomin'
        iters(2) = krylov
        k        = 2*krylov*n
        if (memleft() - intdbl*k .lt. 0) then
           errflg = memleft() - intdbl*k
           return
        end if 
	itvecs   = addtbl('itvecs', k, 'real', from)
        call orthomin(
     1            a,colind,rwptr,n,
     2            x,work(r),b,
     3            loctol,reltv,iters,ierr,
     4            work(next),wksz,
     5            work(lu),work(jlu),work(uptr),
     6            pcmeth, iunit, instlvl)
      end if

      if (solmeth .eq. sCSCGS) then
        k = 17*n
        if (memleft() - intdbl*k .lt. 0) then
           errflg = memleft() - intdbl*k
           return
        end if 
        from = 'CSCGS'
        force = .false.
        itvecs   = addtbl('itvecs', k, 'real', from)
        call cscgstab(
     1            a,colind,rwptr,n,
     2            x,work(r),b,
     3            loctol,reltv,iters,ierr,force,
     4            work(next),wksz,
     5            work(lu),work(jlu),work(uptr),
     6            pcmeth, iunit, instlvl)
      end if

c     ==================================================================
c             *** End Solver Routines ***
c     ==================================================================


c     --------------------------
c     Time for iterating solver:
c     --------------------------
      times(2) = mytime() - times(2) - times(7)


c     ---------------------------------------------------
c     Write out message indicating success/fail of method
c     ---------------------------------------------------
      if (results) then
         write(iunit(1),2100)
         if (ierr .lt. 0) write(iunit(1),2120) -intdbl*ierr
         if (ierr .eq. 0) write(iunit(1),2130)
         if (ierr .eq. 1) write(iunit(1),2140)
         if (ierr .eq. 2) write(iunit(1),2150)
         if (ierr .eq. 3) write(iunit(1),2160) 
         if (ierr .eq. 4) write(iunit(1),2180) 
         if (ierr .gt. 4 .and. ierr .lt. 100) write(iunit(1),2170) ierr
         if (ierr .eq. 100) then
           write(iunit(1),2190) Oetli(A,rwptr,colind,n, b, x, .false.)
         end if
         write(iunit(1),2030) nzlu, dfloat(nzlu)/dfloat(nnz)
      end if
      errflg = ierr

c     ----------------------------
c     Compute final residual norms
c     ----------------------------
      call bmux(n,n,x,work(r),a,colind,rwptr)
      call daxpy(n,-1.0d0,b,1,work(r),1)
      res = dnrm2(n,work(r),1)
      if (results) then
        write(iunit(1),2000) res
        write(iunit(1),2090) iters(1)
        write(iunit(1),7045)
      end if

c     --------------------------------------------
c     Output timing summary and total memory used:
c     --------------------------------------------
      space = getnext() - storage
      if (results) then
         write(iunit(1),2060) 
         write(iunit(1),2080) times(3),times(4),times(5)
         write(iunit(1),2070) times(1),times(2),times(1)+times(2)
         write(iunit(1),7045) 
      end if

c     ----------------------------
c     Output memory usage summary:
c     ----------------------------
      if (results) then
         write(iunit(1),2095)
         write(iunit(1),2085) space
         from = 'summem'
         call listbl(from)
      end if 

      if (istbl('diag', from))
     1             call rmvtbl(work(1),'diag',from)
      if (istbl('itvecs',from)) 
     1             call rmvtbl(work(1),'itvecs',from)

c     -----------------------------------------------
c     Call getrusage data for timings file,
c     and call the end of output HTML file indicator.
c     -----------------------------------------------
      if (results) then
          write(iunit(1),7050) 
          call prtres(1,iunit(1))
          write(iunit(1),7045) 
          write(iunit(1),2050)
      end if
      maxits = iters(1)

c     -------------------------
c     Copy times over for CADYF
c     -------------------------
      call dcopy(8,times,1,work(sptime),1)

c     ------------------------------
c     Unscale x vector to reflect Dr
c     ------------------------------
      if(scale .eq. COLSC .or. scale .eq. RCSC) 
     $   call unscalx(n, x, Dr)

c     ------------------------------------------
c     Unscale the matrix A and right hand side b
c     ------------------------------------------
      if (scale .eq. RWSC .or. scale .eq. RCSC)
     &    call unscala(RWSC, A, rwptr, colind, n, b, Dl)
      if (scale .eq. COLSC .or. scale .eq. RCSC)
     &    call unscala(COLSC, A, rwptr, colind, n, b, Dr)


      return

c===========================================================================
c                 *** Format Statements ***
c===========================================================================

 1000  format(128(2(4(d16.8,3x),/),/))
 1010  format(128(8(i5,3x),/))
 1090  format(i4,3x,e11.5)

 6010  format('<H3> Linear System Information </H3>',/,'<UL>')
 6020  format('<LI> Order of matrix = ',i6,/,
     1        '<LI> Number of nonzeros ',i8)
 2020  format('<LI> Right hand side norm: ',1pe12.5,/,'</UL>',/)


 6050  format('<H3> Information about iterative method </H3>',/,'<UL>',
     1    /,  '<LI> Solver is ',a8,' with ',a8, ' preconditioning')
 6030  format('<LI> Krylov subspace size ',i4)
 6060  format('<LI> Levels of fill allowed in precond: ',i3)
 6090  format('<LI> Acceleration parameter: ',f8.2)
 6040  format('<LI> Tolerance on residual = ',1pe10.4)
 6070  format('<LI> Max allowed iterations: ',i6)
 7000  format('<LI> No scaling of system is performed')
 7010  format('<LI> Max value scaling of system is performed')
 7020  format('<LI> Row scaling of system is performed')
 7030  format('<LI> Column scaling of system is performed')
 7040  format('<LI> Row and column scaling of system is performed')
 7045  format('</UL>',/)

 2100  format('<H3>Results of Execution</H3>',/,'<UL>')
 2030  format('<LI> Precond used: ',i8,' words for ',f8.2,' times nnz')
 2130  format('<LI> <FONT COLOR="#008800"><B>Successful return;',
     1        ' tolerance achieved </B></FONT>')
 2120  format('<LI> <FONT COLOR="#FF0000"><B>Additional workspace',
     1        ' needed in solver',/,
     1        '    need ', i6, '  additional integer words</B></FONT>')
 2140  format('<LI> <FONT COLOR="#008800"><B>Initial guess satisfies',
     1        ' stopping test</B></FONT>')
 2150  format('<LI> <FONT COLOR="#008800"><B>Tolerance reset 8 times',
     1        ' in effort to achieve',/,'    true residual small',
     2        ' enough </B></FONT>')
 2160  format('<LI> <FONT COLOR="#008800"><B>Maximum allowed',
     1        ' iterations reached</B></FONT>')
 2170  format('<LI> <FONT COLOR="#FF0000"><B>Breakdown in algorithm;',
     1        ' probable divide',
     2     /, '    by zero; error code = ', i4,'</B></FONT>')
 2180  format('<LI> <FONT COLOR="#FF0000"><B>Preconditioned residual',
     1        ' is too large',
     2     /, '    Try a different preconditioner!</B></FONT>')
 2190  format('<LI> <FONT COLOR="#FF0000"><B>Oetli-Prager norm less',
     1        ' than max(10*macheps,tol)',
     1     /, '    Final O-P norm is', 1pe12.5,'</B></FONT>')
 2000  format('<LI> Final residual 2-norm: ',1pe12.5)
 2010  format('<LI> Final error 2-norm: ',1pe12.5)
 2090  format('<LI> Iterations performed: ', i5)

 2040  format(/,'<HTML>',/, '<HEAD>',/,
     1   '<TITLE>Sparse Iterative Library (SPLIB) Results</TITLE>',/,
     2   '</HEAD>',/,'<BODY TEXT="#000000" BGCOLOR="#FDF5E6">',/,
     3   '<HR size=8 noshade><H1>Sparse Iterative Library (SPLIB)',/,
     4   ' Results</H1>')

 2045  format(/,'<HTML>',/, '<HEAD>',/,
     1   '<TITLE>About Sparse Iterative Library (SPLIB)</TITLE>',/,
     2   '</HEAD>',/,'<BODY TEXT="#000000" BGCOLOR="#FDF5E6">',/,
     3   '<HR size=8 noshade><H1>About Sparse Iterative Library',/,
     4   ' (SPLIB) </H1>')

 2050  format(/,'</BODY>',/,'</HTML>')

 2060  format('<H3>Timing information breakdown</H3>',/,'<UL>')
 2080  format('<LI> Time for A*x (incl. instrumentation): ',1pe12.5,/,
     +   '<LI> Time for At*x (incl. instrumentation): ', 1pe12.5,/,
     +   '<LI> Time for LU-solve (incl. instrumentation): ', 1pe12.5)
 2070  format('<LI> Time to compute preconditioner: ',1pe12.5,/,
     +        '<LI> Time to iterate solver: ', 1pe12.5,/,
     +        '<LI> Total time: ', 1pe12.5)

 2095  format('<H3>Memory usage information</H3>')
 2085  format('Total SPLIB storage used in integer words: ', i8,'<BR>')

 7050  format(/,'<H3>System resource usage information</H3>',/,'<UL>')

 3000  format('*** Invalid solution method parameter: solmeth = ',i5,
     1      /,'    Solution method being reset to CG-Stabilized')
 3010  format('*** Invalid precond method parameter: pcmeth = ',i5,
     1      /,'    Precond method being reset to No Precond')
 3020  format('*** Invalid maximum iterations allowed: ',i5,
     1      /,'    maxits being reset to 100')
 3030  format('*** Invalid tolerance: tol = ',1pe12.5,/,
     1        '    tol being reset to 1.0e-6')
 3040  format('*** Not allowed to use preconditioning with Jacobi, ',
     1      /,'    Gauss-Seidel, or SOR.  Pcmeth reset to no precond')
 3050  format('*** Krylov subspace size must be positive with GMRES, ',
     1      /,'    GMRESt, or Orthomin.  krylov reset to 10')

 4040  format('*** Error in preconditioner TRID ')
 4050  format('*** n exceeds nmax; n,nmax = ',i8,3x,i8)
 4060  format('*** nnz exceeds nzmax; nnz,nzmax = ',i8,3x,i8)
 4070  format('*** Illegal parameter in computing preconditioner')
 4080  format('*** Additional storage needed for preconditioner;',
     +        ' need ',i8,' more words')
 4090  format(' *** Small pivot in computing preconditioner on step ',
     +        i5)
 5000  format('*** Error in ILUT; zero pivot on row = ',i4)
 5010  format('*** Error in readmt ; mneed = ',i7)
 5030  format('*** Error in matrix data for ILUT: ierr = ',i4)
 5033  format('*** Bad value of lfil for ILUT: lfil = ',i6)
 5040  format('*** Error in splib; need more storage.',
     &        '     next = ',i7)

 8000  format('SPLIB is a package of iterative solvers for ',/,
     2        'sparse nonsymmetric linear ',/,
     3        'systems, and has 13 iterative methods and ',/,
     4        '7 preconditioning methods.',/,
     5        'General details about the library can be ',/,
     6        'found in the attached Web page:',/,
     7        '<UL>')
 8005  format('<LI> <A HREF="About_SPLIB.html"> About SPLIB </A>',/,
     1  '<LI> <A HREF="About_SPLIB.html#Usage"> Usage of SPLIB </A>',/,
     1  '<LI> <A HREF="About_SPLIB.html#Methods">Methods and their ',/,
     1  'specification</A>',/,
     1  '<LI> <A HREF="About_SPLIB.html#Input"> Input variables to',/,
     1  ' SPLIB </A>',/,
     1  '<LI> <A HREF="About_SPLIB.html#IO">I/O files </A>',/,
     1  '<LI> <A HREF="About_SPLIB.html#Memory"> Memory usage in',/,
     1  ' SPLIB  </A>',/,
     1  '<LI> <A HREF="About_SPLIB.html#Error"> Error flags',/,
     1  ' returned</A>',/,
     1        '</UL><P> Results about this run follow immediately.',/,
     1        '<HR size=8 noshade>')

c================ End of SPLIB ==========================================

	end
