c======================================================================== 
c                    **** utilsm.f *****
c========================================================================        
c
c  utilsm.f contains routines and functions that are generally useful for 
c  direct sparse matrix manipulation.
c  Currently contains the routines:
c
c      bmux    : CSR A - multiply: y = A*x
c      btmux   : CSR A'- multiply: y = A'*x
c      btmuxp  : CSR A'- multiply: y = y + A'*x
c      btmuxm  : CSR A'- multiply: y = y - A'*x
c      spbmux  : CSR A - multiply using sparse dotproduct function calls.
c      scala   : Scaling of matrix A and right hand side b
c      unscalx : Unscale solution x at end of computations.
c
c     Randall Bramley
c     Department of Computer Science
c     Indiana University
c     Phone: (812) 855-7790
c     email: bramley@cs.indiana.edu
c     Mon Dec 16 11:31:50 EST 1996
c
c
c========================================================================        

c**********************************************************************
c                   SPLIB Copyright (C) 1995-1999                    **
c                     Indiana University                             **
c**********************************************************************

      subroutine bmux (m, n, x, y, a, ja, ia) 
c-----------------------------------------------------------------------
c          y = A*x
c----------------------------------------------------------------------- 
c Multiplies a matrix times a vector using the dot product form
c A is stored in compressed sparse row storage.  This version is 
c instrumented, by adding its total time to the times() array on
c each invocation.  
c
c on entry:
c----------
c m     = row dimension of A
c n     = column dimension of A
c x     = real array of length equal to the column dimension of
c         the A matrix.
c a, ja,
c    ia = input matrix in compressed sparse row format.
c
c on return:
c-----------
c y     = real array of length m, containing the product y=Ax
c
c-----------------------------------------------------------------------
     
      implicit none

c     Arguments
c     =========
      integer m, n, ja(*), ia(*)
      real*8  x(n), y(m), a(*) 

c     Local variables
c     ===============
      real*8 t
      integer i, k

c     Instrumentation variables for timing.
c     =====================================
      include '../include/heads.inc'
      real*8 mytime, tstart
      external mytime

c=======================================================================
c      Start Executable Statements
c=======================================================================

      tstart = mytime()

      do 100 i = 1, m

         y(i) = 0.0d0
c        ------------------------------------------------
c        Compute the inner product of row i with vector x
c        ------------------------------------------------
         do k = ia(i), ia(i+1)-1 
            y(i) = y(i) + a(k)*x(ja(k))
         end do

 100  continue
c
      times(3) = times(3) + mytime() - tstart
      return

c=============== End of bmux ==========================================

      end
c
c=======================================================================
      subroutine btmux (m, n, x, y, a, ja, ia)
c-----------------------------------------------------------------------
c         y =  A' * x
c
c Multiplies the transpose of a matrix times a vector when the original
c matrix is stored in compressed sparse row storage. Can also be
c used as the product of a matrix by a vector when the original
c matrix is stored in the compressed sparse column format.
c-----------------------------------------------------------------------
c
c on entry:
c ----------
c m     = row dimension of A
c n     = column dimension of A
c x     = real array of length equal to the row dimension of
c         the A matrix.
c a, ja,
c    ia = input matrix in compressed sparse row format.
c
c on return:
c -----------
c y     = real array of length equal to column dimmension of the A matrix, 
c	  containing the product y=transp(A)*x
c
c-----------------------------------------------------------------------

      implicit none

c     Arguments
c     =========
      integer m, n, ia(*), ja(*)
      real*8 x(m), y(n), a(*)

c     Local variables
c     ===============
      integer i, k

c     Instrumentation variables for timing.
c     =====================================
      include '../include/heads.inc'
      real*8 mytime, tstart
      external mytime

c=======================================================================
c      Start of Executable Statements
c=======================================================================

      tstart = mytime()

c     ----------------------
c     Zero out output vector
c     ----------------------
      do i = 1,n
         y(i) = 0.0
      end do

c     -------------------------------------------------
c     Loop over the rows, using saxpy form of multiply.
c     -------------------------------------------------
      do i = 1, m
         do k = ia(i), ia(i+1)-1
            y(ja(k)) = y(ja(k)) + x(i)*a(k)
         end do
      end do

      times(4) = times(4) + mytime() - tstart
      return

c=================== End of btmux ======================================

      end

      subroutine bmuxp(m, n, x, y, a, ja, ia) 
c-----------------------------------------------------------------------
c          y = y + A*x
c----------------------------------------------------------------------- 
c Multiplies a matrix times a vector using the dot product form
c A is stored in compressed sparse row storage.  This version is 
c instrumented, by adding its total time to the times() array on
c each invocation.  
c
c on entry:
c----------
c m     = row dimension of A
c n     = column dimension of A
c x     = real array of length equal to the column dimension of
c         the A matrix.
c a, ja,
c    ia = input matrix in compressed sparse row format.
c
c on return:
c-----------
c y     = real array of length m, containing the product y=Ax
c
c-----------------------------------------------------------------------
     
      implicit none

c     Arguments
c     =========
      integer m, n, ja(*), ia(*)
      real*8  x(n), y(m), a(*) 

c     Local variables
c     ===============
      real*8 t
      integer i, k

c     Instrumentation variables for timing.
c     =====================================
      include '../include/heads.inc'
      real*8 mytime, tstart
      external mytime

c=======================================================================
c      Start Executable Statements
c=======================================================================

      tstart = mytime()

      do 100 i = 1, m

c        ------------------------------------------------
c        Compute the inner product of row i with vector x
c        ------------------------------------------------
         do k = ia(i), ia(i+1)-1 
            y(i) = y(i) + a(k)*x(ja(k))
         end do

 100  continue
c
      times(3) = times(3) + mytime() - tstart
      return

c=============== End of bmuxp =========================================

      end
c
c=======================================================================
      subroutine btmuxp(m, n, x, y, a, ja, ia)
c-----------------------------------------------------------------------
c         y =  y + A' * x
c
c Multiplies the transpose of a matrix times a vector when the original
c matrix is stored in compressed sparse row storage. Can also be
c used as the product of a matrix by a vector when the original
c matrix is stored in the compressed sparse column format.  This version
c updates with an add to y.
c-----------------------------------------------------------------------
c
c on entry:
c ----------
c m     = row dimension of A
c n     = column dimension of A
c x     = real array of length equal to the row dimension of
c         the A matrix.
c a, ja,
c    ia = input matrix in compressed sparse row format.
c
c on return:
c -----------
c y     = real array of length equal to column dimmension of the A matrix, 
c	  containing the product y=transp(A)*x
c
c-----------------------------------------------------------------------

      implicit none

c     Arguments
c     =========
      integer m, n, ia(*), ja(*)
      real*8 x(m), y(n), a(*)

c     Local variables
c     ===============
      integer i, k

c     Instrumentation variables for timing.
c     =====================================
      include '../include/heads.inc'
      real*8 mytime, tstart
      external mytime

c=======================================================================
c      Start of Executable Statements
c=======================================================================

      tstart = mytime()

c     -------------------------------------------------
c     Loop over the rows, using saxpy form of multiply.
c     -------------------------------------------------
      do i = 1, m
         do k = ia(i), ia(i+1)-1
            y(ja(k)) = y(ja(k)) + x(i)*a(k)
         end do
      end do

      times(4) = times(4) + mytime() - tstart
      return

c=================== End of btmuxp ======================================

      end
 
c=======================================================================
      subroutine btmuxm(m, n, x, y, a, ja, ia)
c-----------------------------------------------------------------------
c         y =  y - A' * x
c
c Multiplies the transpose of a matrix times a vector when the original
c matrix is stored in compressed sparse row storage. Can also be
c used as the product of a matrix by a vector when the original
c matrix is stored in the compressed sparse column format.  This version
c updates with a subtraction from y.
c-----------------------------------------------------------------------
c
c on entry:
c ----------
c m     = row dimension of A
c n     = column dimension of A
c x     = real array of length equal to the row dimension of
c         the A matrix.
c a, ja,
c    ia = input matrix in compressed sparse row format.
c
c on return:
c -----------
c y     = real array of length equal to column dimmension of the A matrix, 
c	  containing the product y=transp(A)*x
c
c-----------------------------------------------------------------------

      implicit none

c     Arguments
c     =========
      integer m, n, ia(*), ja(*)
      real*8 x(m), y(n), a(*)

c     Local variables
c     ===============
      integer i, k

c     Instrumentation variables for timing.
c     =====================================
      include '../include/heads.inc'
      real*8 mytime, tstart
      external mytime

c=======================================================================
c      Start of Executable Statements
c=======================================================================

      tstart = mytime()

c     -------------------------------------------------
c     Loop over the rows, using saxpy form of multiply.
c     -------------------------------------------------
      do i = 1, m
         do k = ia(i), ia(i+1)-1
            y(ja(k)) = y(ja(k)) - x(i)*a(k)
         end do
      end do

      times(4) = times(4) + mytime() - tstart
      return

c=================== End of btmuxm ======================================

      end

c=======================================================================

      subroutine scala(job,A, rwptr, colind, n, b, D)
      implicit none

c----------------------------------------------------------------
c     Scale a CSR matrix to have certain properties, 
c     depending on the job parameter.
c       job = 1  ---> set max(max(A)) = 1.0
c       job = 2  ---> set max of each row to be 1.0
c       job = 3  ---> scale each column so diagonal entry is 1.0
c
c     If job > 1, a vector is returned in D that gives
c     the scaling vector used, so that
c          Anew = D*A   (job = 2)  or
c          Anew = A*D   (job = 3)
c
c     In case job < 3, the right hand side vector is also scaled.
c
c     Warning: this code assumes that A is not the zero matrix!
c       Also, if the max-norm of a row is zero when job = 2, 
c       or if the diagonal element is zero when job = 3, the
c       code sets the corresponding D element to be zero and
c       does not scale that row or column.
c----------------------------------------------------------------

      integer rwptr(*), n, colind(*), k, i, nnz
      integer s, job

      real*8 A(*), t
      real*8 b(n), D(n)

      integer idamax
      external idamax

c---------- Scaling one: max(max(A)) = 1.0 -------------
      if (job .eq. 1) then
         nnz = rwptr(n+1)-1
         k = idamax(nnz, A, 1)
         t = 1.0d0/A(k)
         call dscal(nnz,t,A,1)
         call dscal(n,t,b,1)
         return
      end if 

c---------- Scaling two: max(A') = 1.0 -------------
      if (job .eq. 2) then
         do i = 1, n
             s = rwptr(i)
             nnz = rwptr(i+1) - s
             t = abs(A(s))
             do k = s+1, rwptr(i+1)-1
               if (abs(A(k)) .gt. t) t = abs(A(k))
             end do 
             if ( t .ne. 0.0d0) then
                D(i) = 1.0d0/t
                call dscal(nnz,D(i),A(s),1)
                b(i) = D(i)*b(i)
             else
                D(i) = 1.0d0
             end if
         end do 
         return
      end if 
c---------- Scaling three: A*D has diag(A) = I --------
      if (job .eq. 3) then
         do i = 1, n
            D(i) = 1.0d0
            do k = rwptr(i), rwptr(i+1)-1
               if (colind(k) .eq. i .and. A(k).ne. 0.0d0) then
                  D(i) = 1.0d0/A(k)
                  go to 100
               end if
            end do
 100     continue
         end do 
         nnz = rwptr(n+1)-1
         do i = 1, nnz
            A(i) = A(i)*D(colind(i))
         end do 
         return
      end if 

      return
      end 
c
c==============================================================
c
      subroutine unscalx(n, x, Dr)
      implicit none
      integer n, k
      real*8 x(n), Dr(n)
      do k = 1, n
         x(k) = x(k)*Dr(k)
      end do
      return
      end

c-----------------------------------------------------------------------
c
      subroutine unscala(job,A, rwptr, colind, n, b, D)
      implicit none

c----------------------------------------------------------------
c     Unscale a CSR matrix to reverse effects of scala.
c----------------------------------------------------------------

      integer rwptr(*), n, colind(*), k, i, nnz
      integer s, job

      real*8 A(*), t
      real*8 b(n), D(n)

      integer idamax
      external idamax

c---------- Scaling one: max(max(A)) = 1.0 -------------
      if (job .eq. 1) then
c         nnz = rwptr(n+1)-1
c         k = idamax(nnz, A, 1)
c         t = 1.0d0/A(k)
c         call dscal(nnz,t,A,1)
c         return
      end if 

c---------- Scaling two: max(A') = 1.0 -------------
      if (job .eq. 2) then
         do i = 1, n
             s = rwptr(i)
             nnz = rwptr(i+1) - s
             if (D(i) .ne. 0.0d0) then
                t = 1.0d0/D(i)
                call dscal(nnz,t,A(s),1)
                b(i) = t*b(i)
             end if 
         end do 
         return
      end if 
c---------- Scaling three: A*D has diag(A) = I --------
      if (job .eq. 3) then
         nnz = rwptr(n+1)-1
         do i = 1, nnz
            A(i) = A(i)/D(colind(i))
         end do 
         return
      end if 

      return
      end 
c
c-----------------------------------------------------------------------
