#include <fstream.h>
#include <SuperLU.h>
#include <SuperLUParms.h>
#include <SparseLinearSystem.h>
#include <Mungings.h>

// --------------------------------
// Check the arguments for SuperLU.
// --------------------------------
int CheckArguments(SuperLUParms& parms, SparseLinearSystem* sls) {

  if ( parms.FacJob < -4 || parms.FacJob > 3) {
    cout << "Bum FacJob in SuperLU; FacJob = "
         << parms.FacJob << endl;
    cout << "Resetting FacJob = 3" << endl;
    parms.FacJob = 3;
  }

  if ( parms.Markovitz < 0.0 || parms.Markovitz > 1.0) {
    cout << "Bum Markovitz in SuperLU; Markovitz = "
         << parms.Markovitz << endl;
    cout << "Resetting Markovitz = 0.01" << endl;
    parms.Markovitz = 0.01;
    
  }
  if ( sls->n < 1 ) {
    cout << "Bad matrix in SuperLU; n = " << sls->n << endl;
    return 4;
  }

// Following causes problems; n*n exceeds 32 bits for n > 32k
//  if ( sls->nnz < sls->n || sls->nnz > (sls->n)*(sls->n)) {
//    return 5;
//  }

  if ( sls->values == NULL || sls->colinds == NULL
       || sls->rowptrs == NULL) {
    cout << "Bad matrix in SuperLU; ptrs screwed up" << endl;
    return 6;
  }
  if ( sls->rhs == NULL) {
    cout << "No rhs provided to SuperLU" << endl;
    return 7;
  }
  
  return 0;
}



int SuperLU(SuperLUParms& parms, SparseLinearSystem* sls) {

  int ErrorCode = CheckArguments(parms,sls);
  if (ErrorCode != 0) return ErrorCode;
  int i = 0;

  int nrow = sls->n;

  /* ---------------------------------- */
  /* set up data structure for solution */
  /* ---------------------------------- */

  if (sls->x == NULL) sls->x = new double[nrow];
  if (sls->x == NULL) {
    ErrorCode =1; /* can't allocate memory for solution */
    return -1;
  }

  // ----------------------------------------------
  // Initialize to zero values, to avoid accidental
  // over/under flows from having garabage values
  // ----------------------------------------------
  for (i = 0; i < nrow; i++) 
    sls->x[i] = 0; 
  
  int remain = 0;   /* Specifies C memory allocation */
  int keep = 0;     /* Destroy LU at end of solve */	 
  int one = 1;
  int info;
  int ierr;
  int *etree;
  int *piv_r;
  int *piv_c;
  int *md_r;
  int *md_c;
  double work;

  // -------------------------------
  // Error and memory analysis stuff
  // -------------------------------
  double rpg; 
  double RecipCondition; 
  double ForwardError; 
  double BackwardError;
  float Memory; 
  float Mem_forLU;
  int Mem_expansions;
  int LowerNnz; 
  int UpperNnz; 

  int n = sls->n;

  etree = new int[n];
  md_r  = new int[n];
  md_c  = new int[n];
  piv_r = new int[n];
  piv_c = new int[n];
  
  if ( piv_c == NULL || 
       piv_r == NULL || 
       md_c == NULL  || 
       md_r == NULL  ||
       etree == NULL) {

    delete[] etree;
    delete[] md_r;
    delete[] md_c;
    delete[] piv_r;
    delete[] piv_c;
    ErrorCode = -11;
    return ErrorCode;
  }
  
  /* ------------ */
  /* call SuperLU */
  /* ------------ */
  
  info = dsolve_(sls->values,
                 sls->colinds,
                 sls->rowptrs,
                 &n,
		 &n, 
                 sls->rhs, 
                 sls->x, 
                 &one,
		 &(parms.panelSize), 
		 &(parms.relaxation), 
                 &(parms.Markovitz), 
                 &(parms.threshold),
		 md_r, 
                 md_c, 
                 etree, 
                 piv_r, 
                 piv_c, 
                 &work, 
                 &remain,
		 &rpg, 
                 &RecipCondition, 
                 &ForwardError, 
                 &BackwardError,
		 &Memory, 
                 &Mem_forLU,
                 &Mem_expansions,
                 &LowerNnz, 
                 &UpperNnz, 
                 &(parms.FacJob), 
                 &keep, 
                 &ierr);
  
  delete[] md_r;
  delete[] md_c;
  delete[] piv_r;
  delete[] piv_c;
  if (ierr != 0) return info;

  ofstream info_file;
   info_file.open("superluresults/info.html", ios::out);
   if (!info_file) {
         cout << "Can't open file for error analysis info" << endl;
         ErrorCode = 5;
         return ErrorCode;
   }
   else {
   
   info_file << "<HTML>" << endl;
   info_file << "<HEAD>" << endl;
   info_file << "<TITLE>SuperLU Results</TITLE>" << endl;
   info_file << "</HEAD>" << endl;
   info_file << "<BODY TEXT=\"#000000\" BGCOLOR=\"#FDF5E6\">" << endl;
   info_file << "<HR size=8 noshade><H1>SuperLU Results </H1>" << endl;
   info_file << "SuperLU is a sparse direct solver from Xiaoye Li of" << endl;
   info_file << "UC-Berkeley and NSERC, which uses a supernodal approach and" << endl;
   info_file << "the BLAS level 2 libraries as much as possible.  " << endl;
   info_file << "<H3> SuperLU factorization</H3>" << endl;
   info_file << "The user specifies what job to perform in SuperLU with the" << endl;
   info_file << "<B>FacJob</B> parameter, which takes the values:" << endl;
   info_file << "<UL>" << endl;
   info_file << "  <LI> -3:  Factor AX = B using previously determined" << endl;
   info_file << "      pivot sequence from earlier factorization." << endl;
   info_file << "      This is for when the values in A change, but" << endl;
   info_file << "      the nonzero structure is the same." << endl;
   info_file << "  <LI> -2:  Factor AX = B only." << endl;
   info_file << "  <LI> -1:  Solve AX = B using previously computed L and U." << endl;
   info_file << "  <LI> 0:   Factor and solve AX = B." << endl;
   info_file << "  <LI> 1:   All in FacJob = 0, and also compute " << endl;
   info_file << "            the reciprocal pivot growth" << endl;
   info_file << "  <LI> 2:   All in FacJob = 1, and also " << endl;
   info_file << "  compute reciprocal condition number." << endl;
   info_file << "  <LI> 3:   All in FacJob = 2, and also perform " << endl;
   info_file << "  one step of iterative" << endl;
   info_file << "  refinement with forward/backward error estimates." << endl;
   info_file << "</UL>" << endl;
   info_file << "<H3> SuperLU parameters </H3>" << endl;
   info_file << "A user can set three parameters that have " << endl;
   info_file << "major impact on the computational" << endl;
   info_file << "accuracy, resource requirements, and " << endl;
   info_file << "computational efficiency:" << endl;
   info_file << "<OL>" << endl;
   info_file << "<LI> <B>panelsize</B>.  A panel consists of " << endl;
   info_file << "consecutive columns in the" << endl;
   info_file << "matrix during Gaussian elimination, and is treated as a single" << endl;
   info_file << "block.  Much of the work done in supernodal methods consists of" << endl;
   info_file << "generating consecutive block columns that have essentially the" << endl;
   info_file << "same nonzero structure.  Panels made of those \"supernodes\" can" << endl;
   info_file << "be treated with efficient BLAS-2 calls." << endl;
   info_file << "<LI> <B>relaxation</B>.  If the number of" << endl;
   info_file << "nodes (columns) in a subtree of the elimination tree is less" << endl;
   info_file << "than relaxation, the entire subtree is treated as one supernode," << endl;
   info_file << "regardless of its nonzero structure." << endl;
   info_file << "<LI> <B>threshold</B>.  This is the standard Markovitz pivoting" << endl;
   info_file << "parameter, traditionally set at 0.01 based on extensive" << endl;
   info_file << "computational experience from Iain Duff and his colleagues." << endl;
   info_file << "This prevents pivoting from being done until the maximum element" << endl;
   info_file << "in an unreduced column is over 0.01 times the size of the diagonal" << endl;
   info_file << "element, and maintains good preliminary reorderings of the matrix." << endl;
   info_file << "</OL>" << endl;
   info_file << "Other parameters that SuperLU provides, " << endl;
   info_file << "but which have not been exposed" << endl;
   info_file << "to the LSA user, include the maximum size for a supernode," << endl;
   info_file << "the minimum row dimension for 2-D blocking to be used," << endl;
   info_file << "and the minimum column dimension for 2-D blocking to be used." << endl;
   info_file << "" << endl;
   info_file << "<HR size=8 noshade>" << endl;
   info_file << "" << endl;
   info_file << "<H3>Results from this run:</H3>" << endl;
   
   info_file << "<H4>Error analysis</H4>" << endl;
   
         info_file << "<UL>" << endl;
         info_file << setprecision(8);
   
         if (parms.FacJob > 0)
           info_file << "<LI> Reciprocal pivot growth: " << rpg << endl;
   
         if (parms.FacJob > 1)
           info_file << "<LI> Estimate of the condition number: " 
                     << 1.0/RecipCondition << endl;
   
         if (parms.FacJob > 2) {
            info_file << "<LI> Estimated forward error bound: " 
                      << ForwardError << endl;
            info_file << "<LI> Componentwise relative error: " 
                      << BackwardError << endl;
         }
         info_file << "</UL>" << endl;
   
   info_file << "<H4>Memory Usage Analysis </H4>" << endl;
         info_file << "<UL>" << endl;
         info_file << "<LI> Memory used in SuperLU: " 
                   << Memory << " bytes" << endl;
         info_file << "<LI> Memory used for LU factors: " 
                   << Mem_forLU << " bytes" << endl;
         info_file << "<LI> Number of memory expansions: " 
                   << Mem_expansions << endl;
         info_file << "<LI> Nonzeros in lower triangular factor: " 
                   << LowerNnz << endl;
         info_file << "<LI> Nonzeros in upper triangular factor: " 
                   << UpperNnz << endl;
   
         info_file << "</UL></BODY> </HTML> " << endl;
         info_file.close();
       }
  return 0;
  
}
  
