#include "SuperLUComponent.h"
#include "SuperLU.h"
#include "SparseLinearSystem_impl.h"
#include "SuperLUParms_impl.h"
#include "../Common/SolutionVector.h"

#include "../../../../framework/ccacore/Port.h"
#include "../../../../framework/ccacore/CCATServicesProxy.h"
#include "../../../../framework/ccacore/RegisterType.h"
#include "../../../../framework/ccacore/FRAMEWORK_REGISTRATION_ID.h"

#include "../../../../framework/idl/solutionVector/UsesSolutionVector.h"

SuperLUComponent::SuperLUComponent() {

    int ec = -1;

    m_core = NULL;
    m_sparseLinearSystemPort = new SparseLinearSystem_impl(this);
    m_superLUParmsPort = new SuperLUParms_impl(this);
    m_computing = false;
    m_stopped = false;

    // set the paramters to be the default one
    // we cannot do the same for sls.
    m_parms_cache = new SuperLUParms;

    /*
     * Init mutex
     */

    pthread_mutexattr_t attr;

    pthread_mutexattr_init(&attr);

    #if defined(DEBUG) && !defined(__linux__)
        ec = pthread_mutexattr_settype(&attr, PTHREAD_MUTEX_ERRORCHECK);
        assert(ec == 0);
    #endif

    pthread_mutex_init(&m_lock, &attr);

    pthread_mutexattr_destroy(&attr);

    /*
     * Init condition variable.
     */

    pthread_cond_init(&m_cv, NULL);

    /*
     * Init pthread attr
     */

    ec = pthread_attr_init(&m_threadAttr);
    assert(ec == 0);
    ec = pthread_attr_setdetachstate(&m_threadAttr, PTHREAD_CREATE_DETACHED);
    assert(ec == 0);
    // stack size being set to 10M
    ec = pthread_attr_setstacksize(&m_threadAttr, 10000000);
    assert(ec == 0);
    //ec = pthread_attr_setscope(&m_threadAttr, PTHREAD_SCOPE_SYSTEM);
    //assert(ec == 0);
}

SuperLUComponent::~SuperLUComponent() {

    int ec = -1;

    if (m_core != NULL) {
        delete m_core;
    }

    delete m_sparseLinearSystemPort;
    delete m_superLUParmsPort;

    ec = pthread_attr_destroy(&m_threadAttr);
    assert(ec == 0);

    ec = pthread_cond_destroy(&m_cv);
    assert(ec == 0);

    ec = pthread_mutex_destroy(&m_lock);
    assert(ec == 0);
}

void
SuperLUComponent::setServices(Services *cc) {

    if (cc == NULL) {

        assert(m_core != NULL);
        delete m_core;
        m_core = NULL;
        return;

    } else {

        m_core = cc;

        m_core->addProvidesPort(m_sparseLinearSystemPort,
         new DefaultPortInfo("inputSLS", "SparseLinearSystem_idl"));
        m_core->addProvidesPort(m_superLUParmsPort,
         new DefaultPortInfo("inputSuperLUParms", "SuperLUParms_idl"));
        m_core->registerUsesPort(new DefaultPortInfo("outputSV",
         "SolutionVector_idl"));
    }
}

// when executing using pthreads, we can pass only
// one parameter. So pack the paramters needed for
// compute in a struct and pass that. Standard 
// thing to do in C and pthreads. The parameter 
// needs to be of type void *. Info will be cast
// to that when sent to the compute_wrapper method.
struct Info {
    Info(SuperLUComponent *cop, SparseLinearSystem *sls_, SuperLUParms *ps)
     : component(cop), sls(sls_), parms(ps) {}
    SuperLUComponent *const component;
    SparseLinearSystem *const sls;
    SuperLUParms *const parms;
};

// do the object oriented thing.
// invoke a member method to do the
// computation. This function is 
// being called by pthreads and so it 
// takes a void *. Info was created by 
// passing "this" as the first member
// of the struct. Info object is created
// in the startComputation() method.
void *
SuperLUComponent::compute_wrapper(void *vp) {

    Info *ip = static_cast<Info *>(vp);
    Info info = *ip;
    delete ip;

    info.component->compute(info.sls, info.parms);

    return NULL;
}

void
SuperLUComponent::startComputation(SparseLinearSystem *sls, SuperLUParms *sp) {

    pthread_t tid;

    Info *info = new Info(this, sls, sp);
    assert(info != NULL);

    pthread_create(&tid, &m_threadAttr, compute_wrapper, info);
}

void
SuperLUComponent::compute(SparseLinearSystem *sls, SuperLUParms *parms) {
const char *xfn = "SuperLUComponent::compute";

    cout << xfn << " called." << endl;
  
    assert(sls != NULL);
    assert(parms != NULL);

    int ec = SuperLU(*parms, sls);
    assert(ec == 0);

    
    //(sls->mungings).UnMunge(sls->x);

    SolutionVector sv(sls->n, sls->x);

    
    cout << xfn << " solution Vector is " << endl;
   // for (int i =0; i<sls->n; i++) {
   //    cout << (sls->x)[i]  << endl;
   //}
	
    cout.flush();
    

    // Send solution vector out
    cout << xfn << " sending solution vector" << endl;
    UsesSolutionVector *up = dynamic_cast<UsesSolutionVector *>(m_core->getPort("outputSV"));
    up->sendSolutionVector(&sv);
    m_core->releasePort("outputSV");
    cout << xfn << " Sent solution vector!" << endl;
    
    // entering critical section, so grab lock
    pthread_mutex_lock(&m_lock);
    // its okay for another computation to start
    // if it is waiting for this computation to finish
    m_computing = false;
    cout << xfn << " setting m_computing: " << m_computing << endl;
    pthread_cond_broadcast(&m_cv);
    // leavingcritical section, so release lock
    pthread_mutex_unlock(&m_lock);
    
    /*
      delete sls;
      delete parms;
    */
}
