#include "dsp_defs.h"
#include "util.h"

int dsolve_(
            double *a,           /* A in CSR format (1-based indices)  */
            int    *colind,      /* Column indices for A */
            int    *rwptr,       /* Row pointers for A */
            int    *nrows,       /* Number of rows ... */
            int    *ncols,       /* ... and columns of A */
            double *rhs,         /* Pointer to rhs vector */
            double *xact,        /* Pointer to solution vector */
            int    *Nrhs,        /* Number of right hand sides ; usually 1 */

            int    *Panel_Size,  /* control parameters; usually fixed */
            int    *Relax,
            double *Threshold,
            double *Drop_Tol,

            int    *perm_c,      /* Column permutation vector of length n */
            int    *perm_r,      /* Row perm vector of length n */
            int    *etree,       /* Elimination tree of length n */
            int    *localperm_c, /* Column permutation vector of length n */
            int    *localperm_r, /* Row perm vector of length n */

            void   *work,        /* Work array of storage; not working yet */
            int    *Work_size_in_bytes,   /* Guess what ? */

            double *rpg,         /* rpg, rcond, ferr, berr: error estimates */
            double *rcond,
            double *ferr,
            double *berr,

            float *Mem_Total,   /* number of bytes needed by SuperLU */
            float *For_LU,      /* number of bytes for LU factors*/
            int   *expansions,  /* number of mem expansions in SuperLU*/
            int   *Lnnz,        /* Number of nonzeros in L */
            int   *Unnz,        /* Number of nonzeros in U */

            int  *FacJob,       /* Job control parameter */
            int  *keep,         /* Keep L,U storage? */
            int  *ierr)
/*
*  Driver subroutine for calling dgssvx.c.  Changes made include
*  using a job control parameter FacJob, which determines what
*  dgssvx.c is to do.  FacJob has the following interpretation:
*
*    FacJob = -3  ---> Factor AX = B using previously determined
*                        pivot sequence from earlier factorization.
*                        This is for when the values in A change, but
*                        the nonzero structure is the same.
*    FacJob = -2  ---> Factor AX = B only.
*    FacJob = -1  ---> Solve AX = B using previously computed L,U.
*    FacJob = 0   ---> Factor and solve AX = B.
*    FacJob = 1:  ---> " and also compute reciprocal_pivot_growth
*    FacJob = 2:  ---> " and also compute reciprocal cond no.
*    FacJob = 3:  ---> " and also perform one step of iterative refinement
*                        with forward/backward error estimates.
*
*/

{
    char           equed[2], fact[2], trans[2], refact[2];
    Matrix         A;
    Matrix         B, X;
    static Matrix  *L;
    static Matrix  *U;
    NCformat       *Astore, *Ustore;
    SCformat       *Lstore;

    double         read_time;
    double         *R = NULL;
    double         *C = NULL;
    factor_param_t iparam;
    int            info = 0;
    int            lwork = *Work_size_in_bytes;
    int            nrhs = *Nrhs;
    int            ldb  = *nrows;
    int            ldx  = *ncols;
    int            m = *nrows;
    int            n = *ncols;
    int            nnz = rwptr[*nrows] - rwptr[0];
    int            needsolve = (*FacJob >= -1);
    int            *tempwork;

    int            i, j, rowequ, colequ, firstfact;
    extern LUStat_t LUStat;
    double         *utime;
    flops_t        *ops;
    mem_usage_t    mem_usage;

    int job;

    /*  Set up job to be done; needsolve is already set */
    equed[0] = 'N';      /* No equilibration */
    trans[0] = 'T';      /* Input is assumed CSR, so need transpose solve */
    if (*FacJob >=  0) {fact[0] = 'N'; refact[0] = 'N';};
    if (*FacJob == -1) {fact[0] = 'F'; refact[0] = 'N';};
    if (*FacJob == -2) {fact[0] = 'N'; refact[0] = 'N';};
    if (*FacJob == -3) {fact[0] = 'N'; refact[0] = 'Y';};
    /*  Just in case these need be strings somewhere later */
    equed[1] = '\0';
    trans[1] = '\0';
    fact[1] = '\0';
    refact[1] = '\0';

    /*------------------------------------------------------------------*/
    /*  If A is already factored, or if we are factoring a matrix
        with same sparsity pattern and want to reuse the pivot sequence,
        then firstfact is set to be True */
    /*------------------------------------------------------------------*/
    firstfact = lsame_(fact, "F") || lsame_(refact, "Y");

    /*----------------------------------------------------*/
    /* Set up data structure for factorization parameters */
    /*----------------------------------------------------*/
    iparam.panel_size        = *Panel_Size;
    iparam.relax             = *Relax;
    iparam.diag_pivot_thresh = *Threshold;
    iparam.drop_tol          = -1;

    /* Shift indices back by one to have 0-based C indices  */
    for (i = 0; i < n+1; i++) rwptr[i]--;
    for (i = 0; i < nnz; i++) colind[i]--;

    /* Set up the Matrix data structure needed by SuperLU */
    dCreate_CompCol_Matrix(&A, m, n, nnz, a, colind, rwptr, NC, _D, GE);
    Astore = A.Store;

    /* Set up the right hand side and solution vectors in SuperLU Matrix structure */
    dCreate_Dense_Matrix(&B, m, nrhs, rhs, m, DN, _D, GE);
    dCreate_Dense_Matrix(&X, m, nrhs, xact, m, DN, _D, GE);
    if (*FacJob == -3  || *FacJob == -1)  { 
        for (i = 0; i < n; i++) localperm_r [i] = perm_r[i];
        for (i = 0; i < n; i++) localperm_c [i] = perm_c[i];
      }
    else {
        for (i = 0; i < n; i++) localperm_r[i] = i;
        for (i = 0; i < n; i++) localperm_c[i] = i;
      };

    if (*FacJob == -2 || *FacJob > -1) {
       L = (Matrix *) malloc(sizeof(Matrix));
       U = (Matrix *) malloc(sizeof(Matrix));
    }

    /* Call dgssvx */

    dgssvx(fact, trans, refact, &A, &iparam, localperm_r, localperm_c, etree,
	   equed, R, C, L, U, work, lwork, &B, &X, rpg, rcond,
	   ferr, berr, &mem_usage, *FacJob, &info, needsolve);

    if ( info == 0 )
      {
        /* Unpack the performance statistics stuff */
        /*-----------------------------------------*/
        *Mem_Total = mem_usage.total_needed;
        *For_LU    = mem_usage.for_lu;
        *expansions = mem_usage.expansions;
        *ierr = 0;
        Lstore = (SCformat *) L->Store;
        Ustore = (NCformat *) U->Store;
        *Lnnz = Lstore->nnz;
        *Unnz = Ustore->nnz;
      }
      else if ( info > 0 && lwork != -1 )
      {
	printf("dgssvx(): info %d\n", info);
	if (*FacJob >= 1) printf("Recip. pivot growth = %e\n", *rpg);
	if ( info == A.ncol+1  && *FacJob >= 2) 
                 printf("RCOND (est) = %e\n", *rcond);	    
      }

    /* Shift indices forward by one to restore 1-based Fortran indices  */

    for (i = 0; i < n+1; i++) rwptr[i]++;
    for (i = 0; i < nnz; i++) colind[i]++;

    if (*keep == 0)
    {
       Destroy_SuperNode_Matrix(L);
       Destroy_CompCol_Matrix(U);
       /*       Destroy_Dense_Matrix(&B);
       Destroy_CompCol_Matrix(&A); */
    }
    return info ;
}
