#include <map>
#include <iostream>
#include <pthread.h>
#include <assert.h>
#include "EventQueueManager.h"
#include "EventQueue.h"


#ifdef __linux__

/*
 * Linux does not yet have read/write locks.
 */

static int
pthread_rwlock_init(pthread_rwlock_t *l, void *p) {
	return pthread_mutex_init(l, (pthread_mutexattr_t *) p);
}

static int
pthread_rwlock_destroy(pthread_rwlock_t *l) {
	return pthread_mutex_destroy(l);
}

static int
pthread_rwlock_rdlock(pthread_rwlock_t *rwlock) {
	return pthread_mutex_lock(rwlock);
}

static int
pthread_rwlock_unlock(pthread_rwlock_t *rwlock) {
	return pthread_mutex_unlock(rwlock);
}

static int
pthread_rwlock_wrlock(pthread_rwlock_t *rwlock) {
	return pthread_mutex_lock(rwlock);
}

#endif


/*----------------------------------------------------------------------------\
|                                                                             |
|                                                                             |
|  EventQueueManager                                                          |
|                                                                             |
|                                                                             |
\----------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------
 *
 *  Types-Constants
 */

/*------------------------------------------------------------------------------
 *
 *  Constructors-Initters-Assignment-Conversions-Destructors
 */

EventQueueManager::EventQueueManager() {

    pthread_rwlock_init(&m_lock, NULL);
}

EventQueueManager::~EventQueueManager() {

	int ec = -1;

	ec = pthread_rwlock_destroy(&m_lock);
	assert(ec == 0);
}

/*------------------------------------------------------------------------------
 *
 *  Access Operators-Methods
 */

/*------------------------------------------------------------------------------
 *
 *  Operators
 */

/*------------------------------------------------------------------------------
 *
 *  Methods
 */

Ptr<EventQueue>
EventQueueManager::lookup(const std::string &type) {

	assert(m_core != NULL);

	TypeMap::iterator it;
	Ptr<EventQueue> eq;

	pthread_rwlock_rdlock(&m_lock);

	it = m_typeMap.find(type);
	if (it != m_typeMap.end()) {
		eq = (*it).second;
	}

	pthread_rwlock_unlock(&m_lock);

	return eq;
}

std::string
EventQueueManager::addListener(const std::string &id, const std::string &port,
 const std::string &type, bool *new_type) {

	assert(m_core != NULL);

	TypeMap::iterator it;
	Ptr<EventQueue> eq;
	std::string my_port = type + "_uses";

	*new_type = false;

	pthread_rwlock_wrlock(&m_lock);
std::cerr << "adding listener" << std::endl;

	it = m_typeMap.find(type);

	if (it == m_typeMap.end()) {

		eq = new EventQueue(type, m_core);
		std::pair<TypeMap::iterator, bool> res
		 = m_typeMap.insert(std::pair<const std::string,
		 Ptr<EventQueue> >(type, eq));
		assert(res.second);
		eq->typeMapPos = res.first;
		eq->start();

		std::cerr << "added eq for " << type << std::endl;

		*new_type = true;

	} else {

		eq = (*it).second;
	}

	Ptr<Listener> l = new Listener(id, port, type);
	l->queue = eq;

	l->queueSetPos = eq->addListener(l);

	std::pair<ListenerMap::iterator, bool> res
	 = m_listenerMap.insert(std::pair<const std::string,
	 Ptr<Listener> >(id + "_" + port, l));
	assert(res.second);

	pthread_rwlock_unlock(&m_lock);

	return my_port;
}

bool
EventQueueManager::removeListener(const std::string &id,
 const std::string &port, std::string *port_name) {

	assert(m_core != NULL);

	bool rc = false;

	pthread_rwlock_wrlock(&m_lock);
std::cerr << "removing listener" << std::endl;

	// Look up the listener record using "<id>_<port_name>" string.
	ListenerMap::iterator it = m_listenerMap.find(id + "_" + port);

	// Commenting this out because when we get an unregister call, we
	// don't know if it is for a listener or a publisher.
	//assert(it != m_listenerMap.end());

	// If we found it...
	if (it != m_listenerMap.end()) {

		Ptr<Listener> l = (*it).second;

		*port_name = l->queue->type() + "_uses";

		// Remove the listener record from the listener map.
		m_listenerMap.erase(it);

		// Remove the listener record from the event queue set.
		l->queue->removeListener(l->queueSetPos);

		/* XXX This is commented out for now.
		// If no more listeners for this queue, remove it from the type map.
		// It should get cleaned up automatically when it's ref count goes
		// to 0, which will happen when the destructor for l gets called.
		if (l->queue->nListeners() == 0) {
			m_typeMap.erase(l->queue->typeMapPos);
			l->queue->stop();
			*remove_port = true;
		}
		*/

		rc = true;
	}

	pthread_rwlock_unlock(&m_lock);

	return rc;
}

/*------------------------------------------------------------------------------
 *
 *  Static Methods and Helper Functions
 */

/*------------------------------------------------------------------------------
 *
 *  Static Data
 */



/*----------------------------------------------------------------------------\
|                                                                             |
|                                                                             |
|  EventQueueManager::Cmp                                                     |
|                                                                             |
|                                                                             |
\----------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------
 *
 *  Methods
 */

/*
bool
EventQueueManager::Cmp::operator()(const Key &k1, const Key &k2) const {

	if (k1.id.stringify() < k2.id.stringify()) {
		return true;
	} else if (k1.id.stringify() == k2.id.stringify()) {
		return k1.port_name < k2.port_name;
	} else {
		return false;
	}
}
*/
