/**
 * This is the graphical view of a component.<br>
 * It catches creation events and handles component display.
 */

package compositiontool.view;

import compositiontool.model.ComponentModel;
import compositiontool.model.PortModel;
import compositiontool.model.ParamFields;
import util.Info;
import idsbrowser.InformationDirectoryIconRenderer;
import ids.InformationInfo;
import ids.InformationDirectoryNode;
import registry.InstanceInfo;
import idsbrowser.DraggableInformationDirectoryTreeNode;
import registrybrowser.DraggableRegistryTreeNode;
import java.awt.dnd.DragGestureListener;
import java.awt.dnd.DragGestureEvent;
import java.awt.dnd.DragGestureRecognizer;
import java.awt.dnd.DragSourceListener;
import java.awt.dnd.DragSourceEvent;
import java.awt.dnd.DragSourceDragEvent;
import java.awt.dnd.DragSourceDropEvent;
import java.awt.dnd.DnDConstants;
import java.awt.dnd.DragSource;
import java.awt.datatransfer.Transferable;
import java.awt.datatransfer.DataFlavor;
import java.awt.datatransfer.UnsupportedFlavorException;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.BorderLayout;
import java.awt.GridLayout;
import java.awt.Dimension;
import java.awt.Point;
import java.awt.Color;
import java.awt.Cursor;
import java.awt.Toolkit;
import javax.swing.Icon;
import java.awt.Image;
import java.awt.Window;
import java.awt.Frame;
import java.awt.Rectangle;
import java.awt.Graphics;
import java.beans.PropertyVetoException;
import javax.swing.JInternalFrame;
import javax.swing.BoxLayout;
import javax.swing.JDialog;
import javax.swing.JFrame;
import javax.swing.JButton;
import javax.swing.JComponent;
import javax.swing.Timer;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTextField;
import javax.swing.JComboBox;
import javax.swing.JLabel;
import javax.swing.ImageIcon;
import javax.swing.DefaultComboBoxModel;
import javax.swing.BorderFactory;
import javax.swing.border.Border;
import javax.swing.event.InternalFrameAdapter;
import javax.swing.event.InternalFrameEvent;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import java.lang.reflect.Method;
import java.util.Vector;
import java.util.Enumeration;

public class ComponentView extends JInternalFrame 
    implements MouseListener, DragGestureListener, ActionListener {
    private ComponentModel componentModel;
    private ComponentIcon iconifiedFrame;
    private int numInputPorts;
    private int numOutputPorts;
    private Vector inputPortList; // of PortView objects
    private Vector outputPortList; // of PortView objects
    private JPanel infoPane;
    private JPanel controlPane;
    private JPanel buttonPane;
    private JComboBox machines;
    private JComboBox creationMechanisms;
    private ComponentButton createButton;
    private JLabel icon;
    private WorkingAreaController controller;
    private DragSource dragSource  = DragSource.getDefaultDragSource();
    private static int FLASH_INTERVAL = 250; // interval for flashes while waiting for events
    private Timer flashTimer;
    private Border componentBorder;
    public final static DragSourceListener dragSourceListener = 
	new ComponentDragSourceListener();

    /**
     * Create a component view 
     */
    public ComponentView(String name,
			 ComponentModel componentModel, 
			 WorkingAreaController controller) {
	// create an internal frame
	super(name,true,true,false,true);
	this.componentModel = componentModel;
	this.controller = controller;
	// set up component border
	this.setBorder(BorderFactory.createMatteBorder(1,5,1,1,Color.cyan));
	// set up timer
	flashTimer = new Timer(FLASH_INTERVAL,this);
	JComponent contentPane = (JComponent) getContentPane();
	// create a panel for control of the component
	controlPane = new JPanel();
	controlPane.setLayout(new GridLayout(1,2));
	contentPane.add(controlPane,BorderLayout.CENTER);
	// create a panel to hold input ports
	JPanel inputPane = new JPanel();
	contentPane.add(inputPane,BorderLayout.WEST);
	// create a panel to hold output ports
	JPanel outputPane = new JPanel();
	contentPane.add(outputPane,BorderLayout.EAST);
	// create a panel to hold machine and creation information
	infoPane = new JPanel();
	contentPane.add(infoPane,BorderLayout.SOUTH);
	icon = null;
	// if this is a dummy component, add a create button to the control
	// panel
	String imageFile = MainCompositionToolWindow.CONFIG_DIR+"icons/instancedead.gif";
	if (!componentModel.isReal()) {
	    createButton = new ComponentButton(this,"createComponent","Create");
	    createButton.setToolTipText("Create instance");
	    controlPane.add(createButton,BorderLayout.WEST);
	    // add the icon for this component
	    if (InformationDirectoryIconRenderer.iconChooser != null)
		icon = new JLabel(InformationDirectoryIconRenderer.iconChooser.getMatchingIcon((InformationDirectoryNode)componentModel.getInfo(),false));
	    else 
		icon = new JLabel(new ImageIcon(imageFile));
	} else {
	    imageFile = MainCompositionToolWindow.CONFIG_DIR+"icons/instancealive.gif";
	    icon = new JLabel(new ImageIcon(imageFile));
	}
	controlPane.add(icon,BorderLayout.EAST);
	// get all the input ports and create port views for them
	PortModel [] inputPorts = componentModel.getInputPorts();
	inputPortList = new Vector();
	for (int i=0; i<inputPorts.length; i++) {
	    PortView newView = new PortView(controller,this);
	    newView.setPortModel(inputPorts[i]);
	    inputPorts[i].setPortView(newView);
	    // add the port view to the input port panel
	    System.out.println("Added one port view to inputs");
	    inputPane.add(newView);
	    inputPortList.add(newView);
	    numInputPorts++;
	}
	// get all the output ports and create port views for them
	outputPortList = new Vector();
	PortModel [] outputPorts = componentModel.getOutputPorts();
	for (int i=0; i<outputPorts.length; i++) {
	    PortView newView = new PortView(controller,this);
	    newView.setPortModel(outputPorts[i]);
	    outputPorts[i].setPortView(newView);
	    // add the port view to the output port panel
	    System.out.println("Added one port view to outputs");
	    outputPane.add(newView);
	    outputPortList.add(newView);
	    numOutputPorts++;
	}
	if (!componentModel.isReal()) {
	    // if this is a dummy component, parse the information object
	    // to extract all the installation information and display that
	    // in the appropriate panel using lists
	    addInstallationInfo();
	} else {
	    // otherwise this component refers to a running instance, just 
	    // display the machine the instance is running on and the
	    // creation mechanism that was used
	    JLabel machineName = new JLabel(componentModel.getMachineName());
	    JLabel creationMech = new JLabel(componentModel.getCreationMechanism());
	}
	// add an internal frame listener so that a window close event
	// can invoke a kill on the component
	final ComponentModel model = componentModel;
	this.addInternalFrameListener(new InternalFrameAdapter() {
	    public void internalFrameClosed(InternalFrameEvent event) {
		// kill the component
		model.killComponent();
	    }
	});
	// pack the frame and make it visible
	this.pack();
	this.setVisible(true);
	// add a mouse listener to the glass pane associated with this frame
	this.getGlassPane().addMouseListener(this);
	// add the drag source listener
	dragSource.createDefaultDragGestureRecognizer(icon,
						      DnDConstants.ACTION_COPY,
						      this);
    }

    public void paint(Graphics g) {
	super.paint(g);
    }

    /**
     * Access the input port views
     */
    public Enumeration getInputPortViews() {
	return inputPortList.elements();
    }

    /**
     * Access the output port views
     */
    public Enumeration getOutputPortViews() {
	return outputPortList.elements();
    }

    /**
     * Create lists of machine names and creation mechanisms form the installation
     * information and add them to the info panel
     */
    private void addInstallationInfo() {
	// get the installation information
	Info [] installationInfo = componentModel.getInstallationInfo();
	String [] machineNames = new String[installationInfo.length];
	Vector possibleCreationMechanisms = new Vector();
	int numMachines;
	// create the machine list from the machine names
	System.out.println("There are "+installationInfo.length+" installations");
	for (int i=0; i<installationInfo.length; i++) {
	    Info currentInstallation = installationInfo[i];
	    // get the machine name
	    machineNames[i] = currentInstallation.getValue("host-name")[0];
	    System.out.println("Added machine "+machineNames[i]);
	}
	machines = new JComboBox(machineNames);
	machines.setEditable(false);
	// add this class as a selection listener for the machine list
	machines.addActionListener(this);
	creationMechanisms = new JComboBox();
	// add this class as a selection listener for the creation mechanism list
	creationMechanisms.addActionListener(this);
	creationMechanisms.setEditable(false);
	machines.setSelectedIndex(0);
	// add the list of machines and possible creation mechanisms to the panel
	infoPane.add(machines,BorderLayout.WEST);
	infoPane.add(creationMechanisms,BorderLayout.EAST);
	infoPane.setToolTipText("Installation to use");
    }

    /**
     * Called whenever the value on the machine or creation 
     * mechanism list gets changed
     */
    public void actionPerformed(ActionEvent event) {
	if (event.getSource() == machines) {
	    // get the new value
	    String newMachineName = machines.getSelectedItem().toString();
	    // set this on the component model
	    componentModel.setMachineName(newMachineName);
	    // update the creation method list to display possibilities for
	    // this machine
	    addCreationMechanismInfo(newMachineName);
	    // update the creation mechanism on the model
	    componentModel.setCreationMechanism(creationMechanisms.getSelectedItem().toString());
	} else if (event.getSource() == creationMechanisms) {
	    // came here because the creation mechanism list got updated
	    // update the creation mechanism on the model
	    componentModel.setCreationMechanism(creationMechanisms.getSelectedItem().toString());
	} else {
	    // flash timer went off
	    // toggle the component border between the current border color
	    // and a white matte border
	    Border border = this.getBorder();
	    if (border == componentBorder)
		this.setBorder(BorderFactory.createMatteBorder(1,5,1,1,Color.red));
	    else
		this.setBorder(componentBorder);
	}
    }

    /**
     * Called when there is an action on the create button
     */
    public void createComponent() {
      // disable the create button
      createButton.setEnabled(false);
	controller.setStatusText("Attempting creation...");
	// create the darned thing, in a separate thread
	new Thread() {
	    public void run() {
		componentModel.create();
	    }
	}.start();
    }

    public void creationImpending() {
	// inform the workspace that a creation is impending
	controller.creationImpending(this);
	componentBorder = this.getBorder();
	// make the component flash
	flashTimer.start();
    }

    public void creationAborted() {
	// unsuccessful creation attempt
	controller.creationAborted(this);
	// stop flashing
	flashTimer.stop();
	// set the old background
	this.setBorder(componentBorder);
    }

    public void creationCompleted() {
	System.out.println("Stopping flashes...");
	// stop flashing
	flashTimer.stop();
	// set the old background
	this.setBorder(componentBorder);
	// remove the create button
	controlPane.remove(createButton);
	// add a label for the machine name the instance is 
	// running on and its creation mechanism
	JLabel machineName = new JLabel(machines.getSelectedItem().toString());
	JLabel creationMechName = new JLabel(creationMechanisms.getSelectedItem().toString());
	// remove the installation information
	infoPane.remove(machines);
	infoPane.remove(creationMechanisms);
	// add the lebels
	infoPane.add(machineName,BorderLayout.WEST);
	infoPane.add(creationMechName,BorderLayout.EAST);
	infoPane.setToolTipText("Installation information");
	// change the icon 
	icon.setIcon(new ImageIcon(MainCompositionToolWindow.CONFIG_DIR+
				   "icons/instancealive.gif"));
	// inform the workspace that a creation completed
	controller.creationCompleted(this);
    }	

    // method to add an execute button on the component view
    public void addExecuteButton() {
	if (buttonPane == null) {
	    buttonPane = new JPanel();
	    controlPane.add(buttonPane,BorderLayout.WEST);
	}
	JButton executeButton = new ComponentButton(this,"executeComponent","Execute");
	buttonPane.add(executeButton,BorderLayout.NORTH);
    }

    // method to add a set params button on the component view
    public void addSetParamsButton() {
	if (buttonPane == null) {
	    buttonPane = new JPanel();
	    controlPane.add(buttonPane);
	}
	JButton setParamsButton = new ComponentButton(this,"setParams","Set parameters");
	buttonPane.add(setParamsButton,BorderLayout.SOUTH);
    }

    // method to process an event on the execute button
    public void executeComponent() {
	componentModel.execute();
    }

    // method to process an event on the set params button
    public void setParams() {
	System.out.println("Starting up param GUI");
	// make the component stop execution while params are being set
	componentModel.preventFurtherExecution();
	// bring up the GUI for setting this component's parameters
	// get the list of params with default values
	final ParamFields [] params = componentModel.getParamFields();
	final JTextField [] paramValueFields = new JTextField[params.length];
	// create a dialog for the user to enter the values
	final JFrame paramWindow = new JFrame();
	JPanel panel = new JPanel();
	panel.setLayout(new BoxLayout(panel,BoxLayout.Y_AXIS));
	// create a label and text field for each parameter
	for (int i=0; i<params.length; i++) {
	    // create a panel to hold this row of inputs
	    JPanel row = new JPanel();
	    System.out.println("Adding to param dialog: field "+params[i].name+" with class name "+params[i].className);
	    row.add(new JLabel(params[i].name+" "),BorderLayout.WEST);
	    paramValueFields[i] = new JTextField(params[i].defaultValue);
	    if (paramValueFields[i].getColumns() < 10)
		paramValueFields[i].setColumns(10);
	    else
		paramValueFields[i].setColumns(paramValueFields[i].getColumns()+5);
	    row.add(paramValueFields[i],BorderLayout.CENTER);
	    panel.add(row);
	}
	// add a done button at the bottom
	JButton done = new JButton("Set");
	done.addActionListener(new ActionListener() {
	    public void actionPerformed(ActionEvent event) {
		// update the real values of the params
		for (int i=0; i<params.length; i++) {
		    params[i].realValue = paramValueFields[i].getText();
		}
		// send the changed params to the model
		componentModel.setParams(params);
		// this is an SC 99 hack
		// let the component start execution now   (unless of course it is TimeGen
		// in which case we make the user do an explicit execute
		// ******************************
		// this is an SC 99 hack
		// ******************************
		if (!componentModel.getName().endsWith("TimeGen")) {
		  System.out.println("Executing component "+componentModel.getName());
		  componentModel.execute();
		}
		// ******************************
		// this was an SC 99 hack
		// ******************************
		// kill the dialog
		paramWindow.setVisible(false);
	    }
	});
	panel.add(done);
	panel.setPreferredSize(new Dimension(180,300));
	JScrollPane scrollPane = new JScrollPane(panel);
	paramWindow.setTitle("Set parameters");
	paramWindow.setContentPane(scrollPane);
	paramWindow.pack();
	paramWindow.setVisible(true);
    }

    private void addCreationMechanismInfo(String machineName) {
	Vector possibleCreationMechanisms = new Vector();
	// get the installation info
	Info [] installations = componentModel.getInstallationInfo();
	// iterate through it until we reach the installation on this machine
	for (int i=0; i<installations.length; i++) {
	    Info currentInstallation = installations[i];
	    // get the machine name
	    String currentMachine = currentInstallation.getValue("host-name")[0];
	    // if this is the same as the specified machine name, iterate through
	    // the creation info for the installations on this machine
	    if (currentMachine.equals(machineName)) {
		// create the creation mechanism list
		System.out.println("Looking up creation info for host "+machineName);
		Info [] creationInfo = currentInstallation.getInfo("creation-info");
		for (int j=0; j<creationInfo.length; j++) {
		    System.out.println("Added "+creationInfo[j].getValue("creation-method")[0]);
		    possibleCreationMechanisms.add(creationInfo[j].getValue("creation-method")[0]);
		}
		break;
	    }
	}
	// set the model for the creation mechanism list to this data
	DefaultComboBoxModel choiceBoxModel = new DefaultComboBoxModel(possibleCreationMechanisms);
	creationMechanisms.setModel(choiceBoxModel);
	// make sure the first creation mechanism is selected by default
	creationMechanisms.setSelectedIndex(0);
    }

    /**
     * get the model associated with this view
     */
    public ComponentModel getComponentModel() {
	return componentModel;
    }

    /**
     * This component's iconified view will be set to the specified icon.
     */
    public void setComponentIcon(ComponentIcon icon) {
	iconifiedFrame = icon;
    }

    /**
     * Gets the bounding rectangle for the component's icon. If the component is
     * not iconified, this returns the bounding rectangle for the full-sized 
     * component.
     * @return the bounding rectangle of the component icon or component.
     */
    public Rectangle getIconLocation() {
	if (iconifiedFrame != null)
	    return iconifiedFrame.getBounds(null);
	return getBounds(null);
    }

    /**
     * Computes the preferred size of the component based on the number of 
     * input and output ports
     * @return the component's preferred size
     */
    public Dimension getPreferredSize() {
	int maxPortsInColumn = (numInputPorts > numOutputPorts)? numInputPorts : numOutputPorts;
	int height = maxPortsInColumn*38+50;
	if (maxPortsInColumn <= 1)
	    height = 105;
	int width = 360;
	if (numInputPorts == 0)
	    width -= 10;
	if (numOutputPorts == 0)
	    width -= 10;
	return new Dimension(width,height);
    }

    // mouse listener implementation

    /**
     * Called when the mouse has entered the component view.
     * Selects the component.
     *
     * @param event the mouse enter event
     *
     */
    public void mouseEntered(MouseEvent event) {
	try {
	    setSelected(true);
	} catch (PropertyVetoException exception) {
	    System.err.println("Unable to select component due to exception "+exception);
	}
    }

    /** Not implemented */
    public void mouseClicked(MouseEvent event) {}

    /** Not implemented */
    public void mousePressed(MouseEvent event) {}

    /** Not implemented */
    public void mouseReleased(MouseEvent event) {}

    /** Not implemented */
    public void mouseExited(MouseEvent event) {}

    // drag gesture listener implementation

    /**
     * Called when a drag gesture has been recognized. This method will create
     * the transferable object and initiate the drag.
     * @param dragGestureEvent the event that resulted in the drag gesture 
     * being recognized.
     */
    public void dragGestureRecognized(DragGestureEvent dragGestureEvent) {
	// get the info object from the component
	Info info = componentModel.getInfo();
	// create a draggable object containing the information
	// create a draggable ids node or a draggable registry node
	// if the component is alive, we will create a draggable registry node
	// else we will create a draggable ids node
	Object draggableComponentInformation = null;
	if (componentModel.isReal()) {
	    draggableComponentInformation = 
		new DraggableRegistryTreeNode((InstanceInfo)info);
	} else {
	    draggableComponentInformation = 
		new DraggableInformationDirectoryTreeNode((InformationInfo)info);
	}
	// this draggable info is the transferable object
	Transferable informationCarrier = (Transferable) draggableComponentInformation;
	System.out.println("Starting drag from component...");
	// try setting the cursor to the one used by the component
	Cursor dragCursor = null;
	try {
	    Icon infoIcon = icon.getIcon();
	    Image cursorImage = ((ImageIcon)infoIcon).getImage();
	    dragCursor = Toolkit.getDefaultToolkit().createCustomCursor(cursorImage,new Point(infoIcon.getIconWidth()/2,infoIcon.getIconHeight()/2),"componentinfocursor");
	} catch (Exception exception) {
	    // unable to set the cursor as desired, just use the default cursor
	    dragCursor = DragSource.DefaultCopyDrop;
	}
	dragSource.startDrag(dragGestureEvent,
			     DragSource.DefaultCopyDrop,
			     informationCarrier,
			     dragSourceListener);
    }

    /**
     * <tt>ComponentDragSourceListener</tt> implements the <tt>DragSourceListener</tt>
     * interface. None of the methods actually do anything, but some implementation (even
     * an empty one) is required in order to allow a component to be a drag source.
     * @see java.awt.dnd.DragSourceListener
     */
    public static class ComponentDragSourceListener implements DragSourceListener {
	/** Not implemented */
	public void dragDropEnd(DragSourceDropEvent dragSourceDropEvent) { }
       	/** Not implemented */
	public void dragEnter(DragSourceDragEvent dragSourceDragEvent) { }
	/** Not implemented */
	public void dragExit(DragSourceEvent dragSourceEvent) { }
	/** Not implemented */
	public void dragOver(DragSourceDragEvent dragSourceDragEvent) { }
	/** Not implemented */
	public void dropActionChanged(DragSourceDragEvent dragSourceDragEvent) { }
    }

    /** 
     * <tt>ComponentButton</tt> encapsulates the functionality required of a button 
     * on the component. Modifications made to the base class include  
     * allowing specification of the handler method for button events
     * in the constructor.
     */
    public class ComponentButton extends JButton implements ActionListener {
	private Method eventHandler;
	private ComponentView view;

	/**
	 * Create a component button with the specified name and event handler.
	 * @param eventHandlerName the name of the method in the component view
	 * that will handle events on this button.
	 * @param buttonName the button's name.
	 */
	public ComponentButton(ComponentView view, String eventHandlerName, 
				  String buttonName) {
	    super(buttonName);
	    try {
		eventHandler = view.getClass().getDeclaredMethod(eventHandlerName,new Class[0]);
	    } catch (Exception exception) {
		System.err.println("Unable to look up method named "+eventHandlerName+" in component controller\n");
		exception.printStackTrace();
	    }
	    this.addActionListener(this);
	    this.view = view;
	}

	/**
	 * Catch events occurring on the button and invoke the button's event handler.
	 * @param event the button event.
	 */
	public void actionPerformed(ActionEvent event) {
	    // invoke the handler
	    try {
		eventHandler.invoke(view,new Object[0]);
	    } catch (Exception exception) {
		System.err.println("Unable to invoke method "+eventHandler.getName()+" due to exception "+exception);
		exception.printStackTrace();
	    }
	}
    }

}
