/**
 * This is the graphical view of a port
 * It has an associated port model and is 
 * contained inside a component view
 * Users can interact with the port for making connections
 * and disconnections to and from this port
 */

package compositiontool.view;

import compositiontool.model.PortModel;
import javax.swing.JRadioButton;
import javax.swing.JInternalFrame;
import javax.swing.ImageIcon;
import java.awt.event.MouseMotionListener;
import java.awt.event.MouseListener;
import java.awt.event.MouseEvent;
import java.awt.Point;
import java.awt.Component;
import java.awt.IllegalComponentStateException;

public class PortView extends JRadioButton 
    implements MouseListener, MouseMotionListener {
    // the port model this view is associated with
    private PortModel portModel;
    // the working area all components reside in
    private WorkingAreaController workingArea;
    // the component view that contains this port
    private ComponentView componentView;
    // the origin of the port w.r.t. the workspace
    private Point origin;

    /**
     * Create a port view 
     */
    public PortView(WorkingAreaController workingArea, ComponentView componentView) {
	this.addMouseListener(this);
	this.addMouseMotionListener(this);
	this.workingArea = workingArea;
	this.componentView = componentView;
	this.setIcon(new ImageIcon(MainCompositionToolWindow.CONFIG_DIR+"icons/Plug.gif"));
	this.setSelectedIcon(new ImageIcon(MainCompositionToolWindow.CONFIG_DIR+"icons/NewPlug.gif"));
    }

    /**
     * Set the port model this port view is associated with
     */
    public void setPortModel(PortModel portModel) {
	this.portModel = portModel;
	this.setToolTipText(portModel.getName());
    }

    /**
     * Get the port model this port is associated with
     */
    public PortModel getPortModel() {
	return portModel;
    }

    /**
     * Indicate to the working area that a connection
     * to the specified port is impending
     */
    public void connectionImpending(PortModel otherEnd) {
	// tell the working area that a connection
	// establishment is underway
	if (otherEnd.isInput())
	    workingArea.connectionImpending(otherEnd.getPortView(),this);
	else
	    workingArea.connectionImpending(this,otherEnd.getPortView());
    }

    /**
     * Indicate to the working area that a connection
     * to the specified port has been completed
     */
    public void connectionCompleted(PortModel otherEnd) {
	// tell the working area that a connection
	// establishment to this port has been completed
	if (otherEnd.isInput())
	    workingArea.connectionCompleted(otherEnd.getPortView(),this);
	else
	    workingArea.connectionCompleted(this,otherEnd.getPortView());
    }

    /**
     * Indicate to the working area that a connection
     * to the specified port has been aborted
     */
    public void connectionAborted(PortModel otherEnd) {
	// tell the working area that a connection
	// establishment to this port has been aborted
	if (otherEnd.isInput())
	    workingArea.connectionAborted(otherEnd.getPortView(),this);
	else
	    workingArea.connectionAborted(this,otherEnd.getPortView());
    }

    /**
     * Indicate to the working area that a disconnection
     * to the specified port is impending
     */
    public void disconnectionImpending(PortModel otherEnd) {
	// tell the working area that a disconnection
	// effort is underway
	if (otherEnd.isInput())
	    workingArea.disconnectionImpending(otherEnd.getPortView(),this);
	else
	    workingArea.disconnectionImpending(this,otherEnd.getPortView());
    }

    /**
     * Indicate to the working area that a disconnection
     * to the specified port has been completed
     */
    public void disconnectionCompleted(PortModel otherEnd) {
	// tell the working area that a disconnection
	// effort to this port has been completed
	if (otherEnd.isInput())
	    workingArea.disconnectionCompleted(otherEnd.getPortView(),this);
	else
	    workingArea.disconnectionCompleted(this,otherEnd.getPortView());
    }

    /**
     * Indicate to the working area that a disconnection
     * to the specified port has been aborted
     */
    public void disconnectionAborted(PortModel otherEnd) {
	// tell the working area that a disconnection
	// effort to this port has been aborted
	if (otherEnd.isInput())
	    workingArea.disconnectionAborted(otherEnd.getPortView(),this);
	else
	    workingArea.disconnectionAborted(this,otherEnd.getPortView());
    }

    // implement the mouse listener interface
    public void mouseClicked(MouseEvent event) {
	// nothing done for a mouse click
    }

    public void mousePressed(MouseEvent event) {
	if (event.getModifiers() == MouseEvent.BUTTON1_MASK) {
	    // alert the workspace controller that a connection 
	    // is being made from here
	    workingArea.beginConnection(portModel);
	    // select this port 
	    this.setSelected(true);
	    // ask the workspace to select compatible ports
	    workingArea.selectCompatiblePorts(this);
	    // set the new origin
	    refreshOrigin();
	}
    }	

    public void mouseReleased(MouseEvent event) {
	// alert the workpace controller that a connection
	// ended here
	if (event.getModifiers() == MouseEvent.BUTTON1_MASK) {
	    Point connectionEnd = event.getPoint();
	    connectionEnd.translate(origin.x,origin.y);
	    workingArea.endConnection(connectionEnd);
	}
    }

    public void mouseEntered(MouseEvent event) {
	// nothing to be done here
    }

    public void mouseExited(MouseEvent event) {
	// nothing to be done here
    }

    // implement the mouse motion listener interface
    public void mouseDragged(MouseEvent event) {
	// update the current connection line on 
	// the workspace
	if (event.getModifiers() == MouseEvent.BUTTON1_MASK) {
	    Point endOfMouse = event.getPoint();
	    endOfMouse.translate(origin.x,origin.y);
	    workingArea.drawCurrentConnection(endOfMouse);
	}
    }

    public void mouseMoved(MouseEvent event) {
	// nothing to be done here
    }

    /**
     * Get the origin of this port view in the workspace.
     * This will translate the coordinates so that they make sense to the CAT workspace.
     *
     * @return the coordinates of this port view's top left corner in the workspace.
     * @see cat.view.CATWorkingArea
     *
     */
    public Point getOrigin() {
	refreshOrigin();
	return origin;
    }

    public void refreshOrigin() {
	try {
	    Point newOrigin = getLocationOnScreen();
	    newOrigin.translate(getWidth()/2,getHeight()/2);
	    Point desktopLocation = workingArea.getLocationOnScreen();
	    newOrigin.translate(-desktopLocation.x,-desktopLocation.y);
	    origin = newOrigin;
	} catch (IllegalComponentStateException e) {
	    // oops, can't get locaiton on screen
	    Point tmpOrigin = workingArea.getComponentLocation(this);
	    tmpOrigin.translate(0,40);
	    origin = tmpOrigin;
	}
    }

    public void refreshOriginOld() {
	// port's parent is a JPanel which is inside a 
	// JInternalFrame which is in the workspace JPanel
	if (!this.isShowing() || !this.isValid()) {
	    // get the location of the component in the workspace's coordinate system
	    Point tmpOrigin = workingArea.getComponentLocation(this);
	    // ** HACK
	    // translate this by the mysterious observed difference in the Y coordinate
	    tmpOrigin.translate(0,40);
	    // ** END HACK
	    origin = tmpOrigin;
	}
	Component portParent = getParent();
	Component portGParent = portParent.getParent();
	Component portGGParent = portGParent.getParent();
	Component portGGGParent = portGGParent.getParent();
	Component portGGGGParent = portGGGParent.getParent();
	Point graphicsOrigin = null;
	if (portGGGGParent != null) {
	    graphicsOrigin = new Point(portGGGGParent.getLocation());
	} else {
	    graphicsOrigin = new Point(0,0);
	}
	//	graphicsOrigin.translate(portGGGGParent.getX(),portGGGGParent.getY());
	graphicsOrigin.translate(portGGGParent.getX(),portGGGParent.getY());
	graphicsOrigin.translate(portGGParent.getX(),portGGParent.getY());
	graphicsOrigin.translate(portGParent.getX(),portGParent.getY());
	graphicsOrigin.translate(portParent.getX(),portParent.getY());
	graphicsOrigin.translate(getX()+getWidth()/2,getY()+getHeight()/2);
	origin = graphicsOrigin;
    }

}
					     
