/**
 * This class is the controller for the entire working area
 * and handles all high level events dealing with connection
 * establishment and disconnection.<br>
 * This class also stores the state of the workspace and
 * directs drawing of connection lines.
 * @author Nirmal (nmukhi)
 */

package compositiontool.view;

import compositiontool.model.PortModel;
import compositiontool.model.ComponentModel;
import creation.CreationService;
import ccacore.ComponentID;
import ids.InformationDirectoryService;
import idl.eventPort.CCATEvent;
import registry.RegistryService;
import connection.ConnectionService;
import java.util.Vector;
import java.util.Enumeration;
import java.awt.Point;
import java.awt.Rectangle;
import javax.swing.ImageIcon;
import java.awt.Image;

public class WorkingAreaController  {
    // things required for maintaining correct visual state
    private Vector portViews; // list of PortView objects
    private Vector componentViews; // list of ComponentView objects
    private WorkingAreaView workingAreaView;
    private StatusBarView statusBar;
    // things required for drawing connection lines while they 
    // are being made
    private boolean connectionAlive;
    private PortView startOfCurrentConnection;
    private static int FLASH_INTERVAL = 100; // interval for flashes while waiting for events
    // things required for framework interaction
    private CreationService creationService;
    private ConnectionService connectionService;
    private RegistryService registryService;
    private InformationDirectoryService idsService;
    private EventListener eventListener;
    
    /**
     * Create a working area given all the associated services 
     * from the framework
     */
    public WorkingAreaController(CreationService creationService,
				 ConnectionService connectionService,
				 RegistryService registryService,
				 InformationDirectoryService idsService,
				 WorkingAreaView workingAreaView,
				 StatusBarView statusBar,
				 EventListener eventListener) {
	// set the working area view
	this.workingAreaView = workingAreaView;
	workingAreaView.setController(this);
	// set the status bar view
	this.statusBar = statusBar;
	// set all the services
	this.registryService = registryService;
	this.idsService = idsService;
	this.connectionService = connectionService;
	this.creationService = creationService;
	// set the event listener
	this.eventListener = eventListener;
	eventListener.setGUIController(this);
	// initialize the lists
	portViews = new Vector();
	componentViews = new Vector();
    }

    // ------------------------------------------------------------
    // things relating to accessing stuff stored here and reqd by others
    // ------------------------------------------------------------

    /**
     * get the creation service
     */
    public CreationService getCreationService() {
	return creationService;
    }

    /**
     * get the connection service
     */
    public ConnectionService getConnectionService() {
	return connectionService;
    }

    /**
     * get the IDS service
     */
    public InformationDirectoryService getIDSService() {
	return idsService;
    }

    /**
     * get the registry service
     */
    public RegistryService getRegistryService() {
	return registryService;
    }

    // ------------------------------------------------------------
    // things relating to dragging in new stuff into the workspace
    // ------------------------------------------------------------
    
    /**
     * add a new component to the workspace
     */
    public void addComponent(ComponentView component) {
	// add the component view to the list of component views
	componentViews.add(component);
	// get the component model
	ComponentModel model = component.getComponentModel();
	// get the list of input ports and add 
	// their views to the workspace
	Enumeration inputPorts = component.getInputPortViews();
	while(inputPorts.hasMoreElements()) {
	    portViews.add(inputPorts.nextElement());
	}
	// get the list of output ports and add 
	// their views to the workspace
	Enumeration outputPorts = component.getOutputPortViews();
	while(outputPorts.hasMoreElements()) {
	    portViews.add(outputPorts.nextElement());
	}
	// add the component view to the workspace view
	workingAreaView.add(component);
    }

    // ------------------------------------------------------------
    // things relating to connections
    // ------------------------------------------------------------

    /**
     * Tell the workspace that a connection has begun from the specified port.
     * @param portModel the port from where the connection has begun.
     * @see WorkingAreaController#endConnection
     */
    public void beginConnection(PortModel portModel) {
	// set the start of the connection
	connectionAlive = true;
	startOfCurrentConnection = portModel.getPortView();
	// inidicate on the status bar that a connection 
	// is being initiated
	statusBar.setText("Attempting connection...");
    }

    /**
     * Tell the workspace that the currently active connection has ended at 
     * the specified point in the workspace.
     * If this point corresponds to a valid port, the connection is attempted.
     * @see WorkingAreaController#beginConnection
     */
    public void endConnection(Point endOfConnection) {
	// check if there is an active connection
	if (!connectionAlive)
	    return;
	connectionAlive = false;
	statusBar.setText("Initiating connection...");
	// update the working area view
	workingAreaView.drawConnection(null,null);
	workingAreaView.repaint();
	// figure out which port contains this point
	Enumeration portList = portViews.elements();
	while (portList.hasMoreElements()) {
	    PortView currentPort = (PortView) portList.nextElement();
	    // deselect this port
	    currentPort.setSelected(false);
	    currentPort.refreshOrigin();
	    Point origin = currentPort.getOrigin();
	    Point connectionPoint = new Point(endOfConnection.x,endOfConnection.y);
	    connectionPoint.translate(-origin.x,-origin.y);
	    if (currentPort.contains(connectionPoint)) {
		// this is the end of the connection
		PortModel portAtOtherEnd = currentPort.getPortModel();
		// attempt to connect
		startOfCurrentConnection.getPortModel().connectPort(portAtOtherEnd);
	    }
	}
	// no active connection now
	workingAreaView.updateConnectionLines();
    }

    /**
     * Selects the ports currently in the workspace that are compatible 
     * with the specified one.
     *
     * @param port the port against which compatiblity is being checked.
     * @see cat.model.CATPortModel#isCompatibleWith
     *
     */
    public void selectCompatiblePorts(PortView port) {
	// get the port model of this port
	PortModel modelToCompareWith = port.getPortModel();
	// go through all port views and get their corresponding port
	// models
	Enumeration portList = portViews.elements();
	while (portList.hasMoreElements()) {
	    PortView currentPortView = (PortView) portList.nextElement();
	    PortModel currentPortModel = currentPortView.getPortModel();
	    // check if this port is compatible with the given port
	    if (currentPortModel.isCompatibleWith(modelToCompareWith)) {
		currentPortView.setSelected(true);
	    }
	}
    }

    /**
     * this method attempts to remove the connection described
     * by the specified connection line
     */
    public void disconnect(Line connectionLine) { 
	// get the port views on either end
	PortView output = connectionLine.getOutput();
	PortView input = connectionLine.getInput();
	// disconnect
	output.getPortModel().disconnectPort(input.getPortModel(),true);
    }

    // ------------------------------------------------------------
    // things relating to drawing connection lines accurately
    // ------------------------------------------------------------

    /**
     * Get all the connection lines to be drawn
     */
    public Line [] getConnectionLines() {
	Vector lines = new Vector();
	// go through all port views
	Enumeration portList = portViews.elements();
	while(portList.hasMoreElements()) {
	    // get the corresponding port model
	    PortView view = (PortView)portList.nextElement();
	    Point origin = view.getOrigin();
	    PortModel model = view.getPortModel();
	    // get all the connections from here (dummy and real)
	    Enumeration dummyConnections = model.getDummyConnections();
	    while(dummyConnections.hasMoreElements()) {
		// get the port view on the other end for each connection
		PortModel connectedPort = (PortModel)dummyConnections.nextElement();
		PortView connectedView = connectedPort.getPortView();
		// construct a line from the output to the input
		Line newLine = null;
		if (model.isInput()) {
		    newLine = new Line(connectedView.getOrigin(),origin,view,connectedView,true);
		} else {
		    newLine = new Line(origin,connectedView.getOrigin(),view,connectedView,true);
		}
		lines.add(newLine);
	    }
	    Enumeration realConnections = model.getRealConnections();
	    while(realConnections.hasMoreElements()) {
		// get the port view on the other end for each connection
		PortModel connectedPort = (PortModel)realConnections.nextElement();
		PortView connectedView = connectedPort.getPortView();
		// construct a line from the output to the input
		Line newLine = null;
		if (model.isInput()) {
		    newLine = new Line(connectedView.getOrigin(),origin,view,connectedView,false);
		} else {
		    newLine = new Line(origin,connectedView.getOrigin(),view,connectedView,false);
		}
		lines.add(newLine);
	    }
	}
	// return the list of connection lines
	Line [] lineArray = new Line[lines.size()];
	for (int i=0; i<lineArray.length; i++) {
	    lineArray[i] = (Line) lines.elementAt(i);
	}
	return lineArray;
    }


    /**
     * Draw a connection line from the current port being connected
     * to this point
     */
    public void drawCurrentConnection(Point currentEndOfConnection) {
	// tell the view to draw this line
	workingAreaView.drawConnection(startOfCurrentConnection.getOrigin(),
				       currentEndOfConnection);
    }

    /**
     * Called when a connection from the specified input port to
     * the specified output port is impending
     */
    public void connectionImpending(PortView input,PortView output) {
      statusBar.setText("Attempting connection...");
    }

    /**
     * Called when the connection from the specified input port to 
     * the specified output port is aborted
     */
    public void connectionAborted(PortView input, PortView output) {
	// display status
	statusBar.setText("Connection unsuccessful");
    }
    
    /**
     * Called when the connection from the specified input port to 
     * the specified output port is completed
     */
    public void connectionCompleted(PortView input, PortView output) {
	// display status
	statusBar.setText("Connection successful");
    }

    /**
     * Called when the creation of a component is impending
     */
    public void creationImpending(ComponentView component) {
	statusBar.setText("Attempting creation...");
    }

    /**
     * Called when the creation is aborted
     */
    public void creationAborted(ComponentView component) {
	statusBar.setText("Creation aborted");
    }
    
    /**
     * Called when the creation is complete
     */
    public void creationCompleted(ComponentView component) {
	workingAreaView.updateConnectionLines();
	statusBar.setText("Creation completed");
    }

    /**
     * Called when a disconnection from the specified input port to
     * the specified output port is impending
     */
    public void disconnectionImpending(PortView input,PortView output) {
	statusBar.setText("Attempting disconnection...");
    }

    /**
     * Called when the disconnection from the specified input port to 
     * the specified output port is aborted
     */
    public void disconnectionAborted(PortView input, PortView output) {
	statusBar.setText("Disconnection unsuccessful");
    }
    
    /**
     * Called when the disconnection from the specified input port to 
     * the specified output port is completed
     */
    public void disconnectionCompleted(PortView input, PortView output) {
	statusBar.setText("Disconnection successful");
	workingAreaView.updateConnectionLines();
    }

    /**
     * Get the bounding rectangle of the component which owns this port
     */
    public Point getComponentLocation(PortView portView) {
	// the component view is the port's parent's parent
	ComponentView component = portView.getPortModel().getOwner().getComponentView();
	Rectangle bounds = component.getIconLocation();
	return bounds.getLocation();
    }

    public Point getLocationOnScreen() {
	return workingAreaView.getLocationOnScreen();
    }

    public void setStatusText(String text) {
	statusBar.setText(text);
    }

    public void changeBackground(String backgroundName) {
	// the background is the name of a file
	// create an image from this file
	String imageFileName = MainCompositionToolWindow.CONFIG_DIR+"backgrounds/"+backgroundName;
	Image image = (new ImageIcon(imageFileName)).getImage();
	// make this the background for the working area
	workingAreaView.setBackgroundImage(image);
	workingAreaView.repaint();
    }

    public void quit() {
	// kill all components (and hence remove all connections)
	Enumeration components = componentViews.elements();
	while(components.hasMoreElements()) {
	    ComponentView currentComponent = (ComponentView) components.nextElement();
	    System.out.println("Killing "+currentComponent.getComponentModel().getName());
	    currentComponent.getComponentModel().killComponent();
	}
    }

  public synchronized void componentKilled(ComponentModel componentModel) {
    Enumeration components = componentViews.elements();
    while(components.hasMoreElements()) {
      // make the window for the killed component invisible
      ComponentView view = (ComponentView)components.nextElement();
      ComponentModel model = view.getComponentModel();
	  if (componentModel == model)
		view.setVisible(false);
    }
  }

    // ------------------------------------------------------------
    // things to do with receiving events form the event channel
    // ------------------------------------------------------------

    public void connectionEventReceived(CCATEvent event) {
	// figure out the name of the ports involved in this 
	// connection
	String connectionDetails = event.getDetails();
	// this contains names of the two ports involved
	// in the connection
	// the names are separated by colons
	int colonPosition = connectionDetails.indexOf(":");
	String firstPortName = connectionDetails.substring(0,colonPosition);
	String secondPortName = connectionDetails.substring(colonPosition+1);
	System.out.println("Received connection event for connection between "+
			   firstPortName+" and "+secondPortName);
	// send both these port names to the port model
	// involved in the current connection
	Enumeration ports = portViews.elements();
	while(ports.hasMoreElements()) {
	    PortModel currentPortModel = ((PortView)ports.nextElement()).getPortModel();
	    if (currentPortModel.getName().equals(firstPortName))
	      currentPortModel.connectionEventOccurred(secondPortName);
	    else
	      currentPortModel.connectionEventOccurred(firstPortName);
	    workingAreaView.updateConnectionLines();
	}
    }

    public void disconnectionEventReceived(CCATEvent event) {
	// figure out the name of the ports involved in this 
	// connection
	String disconnectionDetails = event.getDetails();
	// this contains names of the two ports involved
	// in the connection
	// the names are separated by colons
	int colonPosition = disconnectionDetails.indexOf(":");
	String firstPortName = disconnectionDetails.substring(0,colonPosition);
	String secondPortName = disconnectionDetails.substring(colonPosition+1);
	System.out.println("Received disconnection event for connection between "+
			   firstPortName+" and "+secondPortName);
	// send both these port names to all port models
	// since we have no idea who is involved in the current 
	// disconnection attempt
	Enumeration allPorts = portViews.elements();
	while(allPorts.hasMoreElements()) {
	    PortModel currentPortModel = ((PortView)
					  allPorts.nextElement()).getPortModel();
	    if (currentPortModel.getName().equals(firstPortName))
	      currentPortModel.disconnectionEventOccurred(secondPortName);
	    if (currentPortModel.getName().equals(secondPortName))
	      workingAreaView.updateConnectionLines();
	}
    }

    public void creationEventReceived(CCATEvent event) {
	System.out.println("Creation event received in GUI controller");
	// check if this descrption  matches an event id for one 
	// of the components
	// get the event description
	String description = event.getSource();
	Enumeration components = componentViews.elements();
	while(components.hasMoreElements()) {
	    // get the corresponding component model
	    ComponentModel componentModel = ((ComponentView)
					     components.nextElement()).getComponentModel();
	    // inform the component model that a creation was completed
	    componentModel.creationEventOccurred(description);
	    workingAreaView.updateConnectionLines();
	}
    }

    public void idsAdditionEventReceived(CCATEvent event) {
	// TODO
	System.out.println("IDS addition event received in GUI controller");
    }

    public void idsModificationEventReceived(CCATEvent event) {
	// TODO
	System.out.println("IDS modification event received in GUI controller");
    }

    public void registryBindEventReceived(CCATEvent event) {
	// TODO
	System.out.println("Registry bind event received in GUI controller");
    }

    public void registryUnbindEventReceived(CCATEvent event) {
	// TODO
	System.out.println("Registry unbind event received in GUI controller");
    }
}
