/**
 * This is the graphical working area that holds all the
 * components.
 * Users interact with this by moving component windows
 * around, iconifying & deiconifying them. 
 * This also handles drawing of connection lines from 
 * one port view to another
 */

package compositiontool.view;

import compositiontool.model.PortModel;
import compositiontool.model.ComponentModel;
import idsbrowser.DraggableInformationDirectoryTreeNode;
import registrybrowser.DraggableRegistryTreeNode;
import ids.InformationXMLUtils;
import ids.InformationInfo;
import registry.InstanceXMLUtils;
import registry.InstanceInfo;
import javax.swing.JDesktopPane;
import javax.swing.JInternalFrame;
import javax.swing.JComponent;
import javax.swing.JInternalFrame.JDesktopIcon;
import javax.swing.DefaultDesktopManager;
import java.awt.dnd.DropTargetListener;
import java.awt.dnd.DropTarget;
import java.awt.dnd.DropTargetEvent;
import java.awt.dnd.DropTargetDragEvent;
import java.awt.dnd.DropTargetDropEvent;
import java.awt.dnd.DnDConstants;
import java.awt.datatransfer.Transferable;
import java.awt.datatransfer.DataFlavor;
import java.awt.datatransfer.UnsupportedFlavorException;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Cursor;
import java.awt.Image;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.io.IOException;
import java.io.FileInputStream;
import java.io.File;
import java.util.LinkedList;


public class WorkingAreaView extends JDesktopPane implements DropTargetListener, MouseListener, MouseMotionListener {
    private DropTarget dropTarget;
    private WorkingAreaController controller;
    private Point startOfCurrentConnection;
    private Point endOfCurrentConnection;
    private Line [] connectionLines;
    private ComponentCommunicator communicator;
    private Image background;

    // colors to be used
    private static Color DUMMY_CONNECTION_COLOR = Color.black;
    private static Color REAL_CONNECTION_COLOR = Color.blue;
    private static Color CURRENT_CONNECTION_COLOR = Color.red;

    /**
     * Create an empty working area view
     */
    public WorkingAreaView(ComponentCommunicator componentCommunicator) {
	// make this a drop target
	dropTarget = new DropTarget(this,this);
	// make things look nice
	setBorder(javax.swing.border.LineBorder.createGrayLineBorder());
	setPreferredSize(new Dimension(1000,700));
	this.setDoubleBuffered(true);
	// no connections initially
	connectionLines = new Line[0];
	// no background initially
	this.background = null;
	// set the component communicator
	this.communicator = componentCommunicator;
	// make this a mouse listener in order to detect disconnection attempts
	this.addMouseListener(this);
	this.addMouseMotionListener(this);
	// customize the way things get iconified and maximized
	this.setDesktopManager(new DefaultDesktopManager() {
	    public void iconifyFrame(JInternalFrame frame) {
		ComponentView component = (ComponentView)frame;
		// get the bounds of the frame
		Rectangle bounds = component.getBounds();
		// get the parent desktop pane 
		JDesktopPane desktopPane = component.getDesktopPane();
		// remove the frame from the parent desktop pane
		desktopPane.remove(component);
		// create a component icon
		// use only the leaf name of the component
		String fullName = component.getComponentModel().getName();
		String leafName = fullName;
		if (fullName.lastIndexOf("/") != -1)
		    leafName = fullName.substring(fullName.lastIndexOf("/")+1);
		ComponentIcon icon = 
		    new ComponentIcon(leafName,component);
		component.setComponentIcon(icon);
		// add the icon to the parent desktop pane at the desired 
		// position
		desktopPane.add(icon,2);
		icon.setLocation(new Point((int)(bounds.getX()),(int)(bounds.getY())));
		// update lines and repaint the desktop
		updateConnectionLines();
	    }

	    public void maximizeFrame(JInternalFrame frame) {
		JDesktopPane desktopPane = frame.getDesktopPane();
		if (!frame.isIcon())
		    return;
		// get the icon's frame
		ComponentView component = ((ComponentIcon)frame).getFullView();
		component.setComponentIcon(null);
		// get the icon's location
		Point location = frame.getLocation();
		// remove the icon from the desktop pane
		desktopPane.remove(frame);
		// add the frame to the desktop pane
		// at the desired position
		desktopPane.add(component,2);
		component.setLocation(location);
		// update lines and repaint the desktop
		updateConnectionLines();
	    }

	    public void endDraggingFrame(JComponent f) {
		super.endDraggingFrame(f);
		// repaint so that lines get drawn properly
		JDesktopPane desktopPane = ((JInternalFrame)f).getDesktopPane();
		// update lines and repaint the desktop
		updateConnectionLines();
	    }

	    // kill a component if the frame is closed
	    public void closeFrame(JComponent f) {
	      // this is a component view
	      ComponentView component = (ComponentView) f;
	      component.getComponentModel().killComponent();
	    }
	});
    }

    public void setController(WorkingAreaController controller) {
	this.controller = controller;
    }

    public void setBackgroundImage(Image background) {
	this.background = background;
    }

    public void paintChildren(Graphics g) {
	super.paintChildren(g);
	paintLines(g);
    }

    public void paintComponent(Graphics graphics) {
	super.paintComponent(graphics);
	if (background != null) {
	    // figure out the size of the image
	    // if it is smaller than the current workspace size, tile it
	    int imgWidth = background.getWidth(this);
	    int imgHeight = background.getHeight(this);
	    if (imgWidth != -1 && imgHeight != -1) {
		// check if the width and height are acceptable
		int spaceWidth = this.getWidth();
		int spaceHeight = this.getHeight();
		if (imgWidth < spaceWidth*0.5 || imgHeight < spaceHeight*0.5) {
		    // unacceptable image size, tile it
		    int numAlongHorizontal = (int)Math.ceil((1.0*spaceWidth)/imgWidth);
		    int numAlongVertical = (int)Math.ceil((1.0*spaceHeight)/imgHeight);
		    for (int i=0; i<numAlongHorizontal; i++) {
			for (int j=0; j<numAlongVertical; j++) {
			    graphics.drawImage(background,i*imgWidth,j*imgHeight,imgWidth,imgHeight,this);
			}
		    }
		} else {
		    graphics.drawImage(background,0,0,this.getWidth(),this.getHeight(),this);
		}
	    } else {
		graphics.drawImage(background,0,0,this.getWidth(),this.getHeight(),this);
	    }
	}
    }

    /**
     * Draws a connection line from the specified starting point to 
     * the specified ending point. The coordinates must make sense 
     * according to the workspace's graphics context.
     * @param start the coordinates of the start of the desired connection line.
     * @param end the coordinates of the end of the desired connection line.
     */
    public void drawConnection(Point start, Point end) {
	startOfCurrentConnection=start;
	endOfCurrentConnection=end;
	repaint();
    }

    public void updateConnectionLines() {
	connectionLines = controller.getConnectionLines();
	repaint();
    }

    public void paintLines(Graphics graphics) {
	// draw the active connection, if any
	if (startOfCurrentConnection != null && endOfCurrentConnection != null) {
	    drawConnectionLine(graphics,startOfCurrentConnection.x,
			       startOfCurrentConnection.y,endOfCurrentConnection.x,
			       endOfCurrentConnection.y,true,CURRENT_CONNECTION_COLOR);
	}
	// draw existing connections
	// get all the controllers and check their existing connections
	for (int i=0; i<connectionLines.length; i++) {
	    Line currentLine = connectionLines[i];
	    // if this line is a dummy line, the color it in DUMMY_CONNECTION_COLOR
	    // otherwise use REAL_CONNECTION_COLOR
	    // if it is not visible, don't color at all
	    if (currentLine.isVisible()) {
		if (currentLine.isDummy())
		    drawConnectionLine(graphics,currentLine.getStart().x,
				       currentLine.getStart().y,currentLine.getEnd().x,
				       currentLine.getEnd().y,false,DUMMY_CONNECTION_COLOR);
		else
		    drawConnectionLine(graphics,currentLine.getStart().x,
				       currentLine.getStart().y,currentLine.getEnd().x,
				       currentLine.getEnd().y,false,REAL_CONNECTION_COLOR);
	    }
	}
    }

    /**
     * Method to draw a connection line
     */
    private void drawConnectionLine(Graphics graphics, int x1, int y1, int x2, 
				    int y2, boolean currentConnection, 
				    Color arrowColor) {
	int r = 16;
	int h = 8;
	int [] X = new int[3];
	int [] Y = new int[3];
	double slope, term, xr, yr;
	// draw line connecting "source" point (x1,y1) to "sink" point (x2,y2)
	graphics.setColor(arrowColor);
	graphics.drawLine(x1,y1,x2,y2);
	// assume head points in the direction of (x2,y2)
	X[0] = x2;
	Y[0] = y2;
	// figure out slope of line from (x1,y1) to (x2,y2)
	if (x1==x2) {
	    xr = X[0];
	    // make sure (xr,yr) comes _before_ (x2,y2)
	    if (y2<y1) {
		yr = Y[0] + r;
	    } else {
		yr = Y[0] - r;
	    }
	    X[1] = (int)(xr - h);
	    Y[1] = (int)yr;
	    X[2] = (int)(xr + h);
	    Y[2] = (int)yr;
	} else {
	    slope = 1.0d*(y2-y1)/(x2-x1);
	    term = (r/Math.sqrt(1.0d+slope*slope));
	    xr = X[0]+term;
	    yr = Y[0]+slope*term;
	    // make sure (xr,yr) comes _before_ (x2,y2)
	    if ((xr*xr - 2*(x1*xr+y1*yr) + yr*yr) > (x2*x2 - 2*(x1*x2+y1*y2) + y2*y2)) {
		xr = X[0]-term;
		yr = Y[0]-slope*term;
	    }
	    // figure out slope of NORMAL line perpendicular to line (x1,y1)-(x2,y2)
	    slope = 1.0d/slope;
	    if (Math.abs(slope)>100.0d) {
		X[1] = (int)xr;
		Y[1] = (int)(yr + h);
		X[2] = (int)xr;
		Y[2] = (int)(yr - h);
	    } else {
		term = (h/Math.sqrt(1.0d+slope*slope));
		X[1] = (int)(xr+term);
		Y[1] = (int)(yr-slope*term);
		X[2] = (int)(xr-term);
		Y[2] = (int)(yr+slope*term);
	    }
	}
	graphics.setColor(arrowColor);
	graphics.drawPolygon(X,Y,3);
    }

    // methods to support the mouse listener interface
    public void mouseClicked(MouseEvent event) {
	// check if the middle mouse button was pressed
	if (event.getModifiers() == MouseEvent.BUTTON1_MASK) {
	    // possible disconnection attempt
	    // get the point where this occurred
	    Point disconnectionAttemptLocation = event.getPoint();
	    // check if this lies on any of the connection lines
	    for (int i=0; i<connectionLines.length; i++) {
		if (connectionLines[i].containsPoint(disconnectionAttemptLocation)) {
		    // remove this connection
		    controller.disconnect(connectionLines[i]);
		    break;
		}
	    }
	}
    }

    public void mouseEntered(MouseEvent event) { }
    public void mouseExited(MouseEvent event) { }
    public void mousePressed(MouseEvent event) { }
    public void mouseReleased(MouseEvent event) { }

    // methods to support mouse motion listener interface
    public void mouseMoved(MouseEvent event) {
	// get the point
	Point newPoint = event.getPoint();
	// check if this lies on a line
	for (int i=0; i<connectionLines.length; i++) {
	    if (connectionLines[i].containsPoint(newPoint)) {
		// change the cursor
		addNotify();
		this.setCursor(Cursor.getPredefinedCursor(Cursor.CROSSHAIR_CURSOR));
		return;
	    }
	}
	addNotify();
	this.setCursor(Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR));
    }

    public void mouseDragged(MouseEvent event) { }

    // methods to support drop target interface

    /** 
     * Nothing done here, required for interface implementation.
     */
    public void dragEnter (DropTargetDragEvent dropTargetDragEvent) { }

    /** 
     * Nothing done here, required for interface implementation.
     */
    public void dragExit (DropTargetEvent dropTargetEvent) { }
    
    /** 
     * Nothing done here, required for interface implementation.
     */
    public void dragOver (DropTargetDragEvent dropTargetDragEvent) { }

    /** 
     * Nothing done here, required for interface implementation.
     */
    public void dropActionChanged (DropTargetDragEvent dropTargetDragEvent) { }

    /**
     * Handles a drop action (as in drag-and-drop) in the workspace. If the transferable
     * object is in a recognizable flavor (such as a file or a string), this assumes it is
     * an XML representation of a component/instance and tries to create the 
     * corresponding component/instance.
     * If this is successfully done, the newly created component/instance is then added to 
     * the workspace.
     * @param dropTargetDropEvent the drop event.
     */
    public synchronized void drop (DropTargetDropEvent dropTargetDropEvent) {
	try {
	    // get the transferable object
	    Transferable tr = dropTargetDropEvent.getTransferable();
	    // accept it only if the transferable object has a flavor we recognise
	    if (tr.isDataFlavorSupported(DraggableInformationDirectoryTreeNode.STRING_FLAVOR) ||
		tr.isDataFlavorSupported(DraggableInformationDirectoryTreeNode.FILE_FLAVOR)) {
		String xmlDescription = null;
		if (tr.isDataFlavorSupported(DraggableInformationDirectoryTreeNode.STRING_FLAVOR)) {
		    // assume this is a description of a component
		    xmlDescription = (String) tr.getTransferData
			(DraggableInformationDirectoryTreeNode.STRING_FLAVOR);
		} else if (tr.isDataFlavorSupported(DraggableInformationDirectoryTreeNode.FILE_FLAVOR)) {
		    // assume that the file contains the xml information 
		    LinkedList files = (LinkedList) tr.getTransferData
			(DraggableInformationDirectoryTreeNode.FILE_FLAVOR);
		    FileInputStream inStream = new FileInputStream((File)files.getFirst());
		    byte [] contents = new byte[10000];
		    try {
		        int size = inStream.read(contents);
			xmlDescription = new String(contents,0,size);
		    } catch (IOException exception) {
			System.err.println("Got exception while reading XML information from file");
			exception.printStackTrace();
		    }
		}
		// convert the xml description into an information info object
		InformationInfo componentInfo = 
		    InformationXMLUtils.createComponentFromXML(xmlDescription);
		// if this is null, we could not parse the data, so ignore it
		if (componentInfo != null) {
		    // aha, we have a component
		    dropTargetDropEvent.acceptDrop(DnDConstants.ACTION_COPY_OR_MOVE);
		    // create the component model
		    System.out.println("Component info is "+componentInfo);
		    ComponentModel model = new ComponentModel(controller.getCreationService(),
							      controller.getConnectionService(),
							      communicator,
							      controller,
							      componentInfo);
		    // create the component view
		    // use only the leaf name of the component
		    String fullName = model.getName();
		    String leafName = fullName;
		    if (fullName.lastIndexOf("/") != -1)
			leafName = fullName.substring(fullName.lastIndexOf("/")+1);
		    ComponentView view = new ComponentView(leafName,model,controller);
		    int componentX = dropTargetDropEvent.getLocation().x;
		    int componentY = dropTargetDropEvent.getLocation().y;
		    view.reshape(componentX,componentY,view.getWidth(),view.getHeight());
		    model.setComponentView(view);
		    // add the newly created component to the workspace
		    controller.addComponent(view);
		    return;
		}
	    } 
	    if (tr.isDataFlavorSupported(DraggableRegistryTreeNode.STRING_FLAVOR) ||
		tr.isDataFlavorSupported(DraggableRegistryTreeNode.FILE_FLAVOR)) {
		String xmlDescription = null;
		if (tr.isDataFlavorSupported(DraggableRegistryTreeNode.STRING_FLAVOR)) {
		    // assume this is a description of an instance
		    xmlDescription = (String) tr.getTransferData
			(DraggableRegistryTreeNode.STRING_FLAVOR);
		} else if (tr.isDataFlavorSupported(DraggableRegistryTreeNode.FILE_FLAVOR)) {
		    // assume that the file contains the xml information 
		    LinkedList files = (LinkedList) tr.getTransferData
			(DraggableRegistryTreeNode.FILE_FLAVOR);
		    FileInputStream inStream = new FileInputStream((File)files.getFirst());	
		    try {
		        byte [] contents = new byte[10000];
			int size = inStream.read(contents);
			xmlDescription = new String(contents,0,size);
		    } catch (IOException exception) {
			System.err.println("Got exception while reading XML information from file");
			exception.printStackTrace();
		    }
		}
		// convert the xml descrption into an info object
		InstanceInfo instanceInfo = 
		    InstanceXMLUtils.createInstanceInfoFromXML(xmlDescription);
		// if the conversion failed, reject the drop
		if (instanceInfo == null) {
		    System.err.println("Dropped information was not parsable XML component or instance information");
		    System.out.println("Dropped info was "+instanceInfo);
		    dropTargetDropEvent.rejectDrop();
		    getToolkit().beep();
		    return;
		}
		dropTargetDropEvent.acceptDrop(DnDConstants.ACTION_COPY_OR_MOVE);
		// create a compoennt model and view for this instance
		ComponentModel model = new ComponentModel(controller.getCreationService(),
							  controller.getConnectionService(),
							  communicator,
							  controller,
							  instanceInfo);
		String fullName = model.getName();
		String leafName = fullName;
		if (fullName.lastIndexOf("/") != -1)
		    leafName = fullName.substring(fullName.lastIndexOf("/"));
		ComponentView view = new ComponentView(leafName,model,controller);
		int componentX = dropTargetDropEvent.getLocation().x;
		int componentY = dropTargetDropEvent.getLocation().y;
		view.reshape(componentX,componentY,view.getWidth(),view.getHeight());
		model.setComponentView(view);
		model.updateInstanceView();
		// add the new component to the workspace
		controller.addComponent(view);
	    }
	} catch (UnsupportedFlavorException exception) {
	    System.out.println("Got exception "+exception+" while handling drop");
	    exception.printStackTrace();
	    dropTargetDropEvent.rejectDrop();
	    getToolkit().beep();
	} catch (IOException exception) {
	    System.out.println("Got exception "+exception+" while handling drop");
	    exception.printStackTrace();
	    dropTargetDropEvent.rejectDrop();
	    getToolkit().beep();
	}
    }
}
