/**
 * This class is used to specify parameters for selection of an
 * icon for an info object in the information tree.
 * @author Nirmal (nmukhi)
 * @see InformationDirectoryIconChooser
 */

package idsbrowser;

import javax.swing.Icon;
import java.util.Vector;
import java.util.StringTokenizer;
import util.GeneralInfoUtils;
import util.Info;

public class InfoIconSelector {
    private String keyPathExpression;
    private String valueExpression;
    private Icon iconToUse;
    
    /**
     * This is the wild card string to be used to match anything
     */
    public static String WILD_CARD = "*";

    /**
     * This is the path separator string used in the key path
     */
    public static String KEY_PATH_SEPARATOR = "/";

    /**
     * Create an info icon selector with the specified key path, value
     * and icon. 
     * @param keyPath the path to be used in the Info object for 
     * accessing a specific key. If this is *, it will be
     * interpreted to mean "all keys"
     * @param value the value required of the above specified key for 
     * a match. If this is "*", it will be taken to mean "any value"
     * @param icon the icon to be used for an info object that has the 
     * specified key and corresponding value
     */
    public InfoIconSelector(String keyPath, String value, Icon icon) {
	// set the key, value and icon tuple
	this.keyPathExpression = keyPath;
	this.valueExpression = value;
	this.iconToUse = icon;
    }

    /**
     * Checks if the given info object had the key associated with this object
     * and that its value matched the required value of this key. This will
     * return the corresponding icon.
     * @param info the Info object being examined for match
     * @return the icon for matching objects or null if the info object
     * did not match the specification
     * @see util.GeneralInfoUtils#deepGetValue(Info)
     */
    public Icon getMatchingIcon(Info info) {
	// parse the key path and get the value objects at the end
	String [] valuesToCheck = GeneralInfoUtils.deepGetValue(info,keyPathExpression);
	// if we have a wild card value expression, it matches anything
	if (valueExpression.equals(InfoIconSelector.WILD_CARD)) {
	    // match!
	    return iconToUse;
	}
	// no wild card
	// check all the values the key path mapped to
	if (valuesToCheck == null) {
	    // no match
	    return null;
	} else {
	    // non wild value expression, check if it matches all 
	    // the values with this key path
	    boolean matchObtained = true;
	    for (int i=0; i<valuesToCheck.length; i++) {
		if (!valuesToCheck[i].equals(valueExpression)) {
		    matchObtained = false;
		    break;
		}
	    }
	    if (matchObtained) {
		// match!
		return iconToUse;
	    } else {
		return null;
	    }
	}
    }

    /**
     * Get the key path expression for this icon selector
     * @return the key path expression this icon selector uses 
     * while trying to match an icon
     */
    public String getKeyPathExpression() {
	return keyPathExpression;
    }

    /**
     * Get the value expression for this icon selector
     * @return the value expression this icon selector uses 
     * while trying to match an icon
     */
    public String getValueExpression() {
	return valueExpression;
    }

    /**
     * Get the icon used by this icon selector
     * @return the icon this icon selector uses 
     */
    public Icon getIcon() {
	return iconToUse;
    }
}
