/**
 * This class is used for choosing an icon for an info object.
 * It makes its selection given icons to be used for expanded and
 * unexpanded contexts. It uses the 
 * <a href="InfoIconSelector">InfoIconSelector</a> to select an
 * icon for leaf info objects 
 * @see InfoIconSelector
 */

package idsbrowser;

import javax.swing.Icon;
import javax.swing.JPanel;
import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.JScrollPane;
import java.awt.Window;
import java.awt.Image;
import javax.swing.JTextField;
import javax.swing.JButton;
import javax.swing.ImageIcon;
import javax.swing.JDialog;
import javax.swing.Icon;
import javax.swing.BoxLayout;
import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.Color;
import javax.swing.border.EmptyBorder;
import java.awt.event.ActionEvent;
import javax.swing.event.ListSelectionListener;
import javax.swing.event.ListSelectionEvent;
import java.awt.event.ActionListener;
import java.util.Vector;
import java.util.Enumeration;
import java.io.File;
import ids.InformationDirectoryNode;
import compositiontool.view.MainCompositionToolWindow;
import util.Info;

public class InformationDirectoryIconChooserImplementation 
    implements InformationDirectoryIconChooser {
    private Icon contextIcon;
    private Icon expandedContextIcon;
    private Vector infoIconSelectors;
    // things used while browsing icon filters
    private JTextField keyLabel;
    private JTextField valueLabel;
    private JLabel iconLabel;
    private JPanel iconPanel;
    private int currentIconFilter;

    /**
     * Create an icon chooser with the specified icons to be used for a context
     * and the given list of info icon selectors
     * @param contextIcon the icon to be used for contexts
     * @param expandedContextIcon the icon to be used for an expanded context
     * @param infoIconSelectors the list of info icon selectors used for
     * selecting an icon for a leaf
     */
    public InformationDirectoryIconChooserImplementation(Icon contextIcon,
							 Icon expandedContextIcon,
							 Vector infoIconSelectors) {
	// set the context icons and info icon selectors
	this.contextIcon = contextIcon;
	this.expandedContextIcon = expandedContextIcon;
	this.infoIconSelectors = infoIconSelectors;
    }

    /**
     * Get the matching icon for the specified info object
     * @param info the IDS node object for which an icon is desired
     * @expanded whether this info object has been expanded or not
     * in the tree
     * @return the icon to be used for this info object. Note that for leaf
     * info objects, the list of info icon selectors is examined for a match
     * and the first selector which can match this info object is used.
     */
    public Icon getMatchingIcon(InformationDirectoryNode infoNode, 
				     boolean expanded) {
	// check if this is a context
	if (infoNode.isContext()==1) {
	    // we have a context, return expanded or normal icon as appropriate
	    if (expanded)
		return expandedContextIcon;
	    else
		return contextIcon;
	} 
	if (infoNode.isInformation()==1) {
	    Info info = (Info)infoNode;
	    // we have a leaf, go through the list of info icon selectors
	    Enumeration iconSelectorList = infoIconSelectors.elements();
	    while(iconSelectorList.hasMoreElements()) {
		InfoIconSelector iconSelector = (InfoIconSelector)iconSelectorList.nextElement();
		// check if this info icon selector can match this info object
		Icon matchingIcon = iconSelector.getMatchingIcon(info);
		if (matchingIcon != null) {
		    // got a match, return the icon
		    return matchingIcon;
		}
	    }
	    // no match, return null
	    return null;
	}
	// wasn't a context or info, could not find a matching icon
	return null;
    }

    public void browseIconFilters() {
	// create a window to get the required input for the icon customization
	final JDialog iconChooserWindow = new JDialog();
	iconChooserWindow.setTitle("Customize Icons");
	iconChooserWindow.setSize(new Dimension(450,180));
	// this window will contain the current icon filter display and options 
	// to see the next and previous icon filters (if any)
	// there are also buttons to remove the current icon filter and 
	// insert an icon filter before the one being displayed
	// an icon filter display consists of a label showing the search path
	// and value being searched as well as the icon used for documents matching 
	// that search

	// the top panel containing everything
	JPanel panel = new JPanel();
	// panel containing search key and value for this icon filter
	JPanel paramsPanel = new JPanel();
	// panel containing the icon for this icon filter
	iconPanel = new JPanel();
	// panel which containing previous/next buttons
	JPanel browsePanel = new JPanel();
	// panel containing insert/remove buttons
	JPanel modifyPanel = new JPanel();
	
	// do the layout
	panel.setLayout(new BoxLayout(panel,BoxLayout.X_AXIS));
	panel.setPreferredSize(new Dimension(450,150));
	panel.add(paramsPanel);
	panel.add(iconPanel);
	panel.add(browsePanel);
	// do the params panel
	paramsPanel.setLayout(new BoxLayout(paramsPanel,BoxLayout.Y_AXIS));
	paramsPanel.setPreferredSize(new Dimension(200,50));
	// display the icon filters first to last
	currentIconFilter = 0;
	keyLabel = new JTextField(16);
	keyLabel.setEditable(false);
	keyLabel.setFont(keyLabel.getFont().deriveFont((float)16.0));
	valueLabel = new JTextField(16);
	valueLabel.setEditable(false);
	valueLabel.setFont(keyLabel.getFont().deriveFont((float)16.0));
	paramsPanel.add(keyLabel);
	paramsPanel.add(valueLabel);
	iconLabel = new JLabel();
	iconPanel.setPreferredSize(new Dimension(70,60));
	iconPanel.add(iconLabel);
	// get the current icon selector and display its information
	try {
	    InfoIconSelector currentSelector = (InfoIconSelector) 
		infoIconSelectors.elementAt(currentIconFilter);
	    displayIconFilter(currentSelector);
	} catch (ArrayIndexOutOfBoundsException exception) {
	    System.out.println("Invalid index in icon filter vector "+currentIconFilter);
	    exception.printStackTrace();
	}
	// add the left/right buttons to the browse panel
	JButton previous = new JButton(new ImageIcon(MainCompositionToolWindow.CONFIG_DIR+"newicons/Left.gif"));
	previous.addActionListener(new ActionListener() {
	    public void actionPerformed(ActionEvent event) {
		if (currentIconFilter==0)
		    return;
		currentIconFilter--;
		try {
		    InfoIconSelector currentSelector = (InfoIconSelector) 
			infoIconSelectors.elementAt(currentIconFilter);
		    displayIconFilter(currentSelector);
		} catch (ArrayIndexOutOfBoundsException exception) {
		    System.out.println("Invalid index in icon filter vector "+currentIconFilter);
		    exception.printStackTrace();
		}
	    }
	});
	browsePanel.setLayout(new BoxLayout(browsePanel,BoxLayout.Y_AXIS));
	browsePanel.setPreferredSize(new Dimension(60,80));
	browsePanel.add(previous);
	JButton next = new JButton(new ImageIcon(MainCompositionToolWindow.CONFIG_DIR+"newicons/Right.gif"));
	next.addActionListener(new ActionListener() {
	    public void actionPerformed(ActionEvent event) {
		if (currentIconFilter==(infoIconSelectors.size()-1))
		    return;
		currentIconFilter++;
		try {
		    InfoIconSelector currentSelector = (InfoIconSelector) 
			infoIconSelectors.elementAt(currentIconFilter);
		    displayIconFilter(currentSelector);
		} catch (ArrayIndexOutOfBoundsException exception) {
		    System.out.println("Invalid index in icon filter vector "+currentIconFilter);
		    exception.printStackTrace();
		}
	    }
	});
	browsePanel.add(next);
	// do the modify panel
	modifyPanel.setLayout(new BoxLayout(modifyPanel,BoxLayout.X_AXIS));
	JButton insert = new JButton("Insert");
	insert.addActionListener(new ActionListener() {
	    public void actionPerformed(ActionEvent event) {
		// bring up a dialog box to insert an icon filter
		final JDialog iconInsertWindow = new JDialog();
		iconInsertWindow.setTitle("Add icon filter");
		JPanel insertPanel = new JPanel();
		// fix up a panel to hold the input for search path and value
		JPanel inputPanel = new JPanel();
		inputPanel.setLayout(new BoxLayout(inputPanel,BoxLayout.Y_AXIS));
		// specify the search key path
		final JTextField keyPathInput = new JTextField(28);
		keyPathInput.setPreferredSize(new Dimension(145,30));
		inputPanel.add(new JLabel("Key path: "));
		inputPanel.add(keyPathInput);
		// specify the search value
		final JTextField valueInput = new JTextField();
		valueInput.setPreferredSize(new Dimension(145,30));
		inputPanel.add(new JLabel("Value: "));
		inputPanel.add(valueInput);
		// fix up a panel to hold the list of icons and their images
		File iconDirectory = new File(MainCompositionToolWindow.CONFIG_DIR+"icons");
		File [] icons = iconDirectory.listFiles();
		String [] iconNames = new String [icons.length];
		for (int i=0; i<iconNames.length; i++) {
		    iconNames[i] = new String(icons[i].getName());
		}
		// create a list box with possible icons
		final JList listBox = new JList(iconNames);
		final JLabel currentIcon = new JLabel();
		listBox.addListSelectionListener(new ListSelectionListener() {
		    public void valueChanged(ListSelectionEvent event) {
			JList iconList = (JList) event.getSource();
			String iconName = (String) iconList.getSelectedValue();
			// display this icon in the label
			// first resize the icon so it has the same size as the default icon
			Icon defaultIcon = InformationDirectoryIconChooserFactory.DEFAULT_INFO_ICON;
			ImageIcon icon = new ImageIcon(MainCompositionToolWindow.CONFIG_DIR+"icons/"+iconName);
			Image newIconImage = icon.getImage();
			Image resizedImage = newIconImage.getScaledInstance(defaultIcon.getIconWidth(),
									    defaultIcon.getIconHeight(),
									    Image.SCALE_DEFAULT);
			currentIcon.setIcon(new ImageIcon(resizedImage));
		    }
		});
		// put the list in a scrollable pane
		JScrollPane iconPane = new JScrollPane(listBox);
		JPanel iconDisplayPanel = new JPanel();
		iconDisplayPanel.add(currentIcon);
		JPanel mainIconPane = new JPanel();
		mainIconPane.add(iconPane,BorderLayout.WEST);
		mainIconPane.add(iconDisplayPanel,BorderLayout.CENTER);
		// add a done button to signal the end of input
		JButton done = new JButton("Done");
		done.addActionListener(new ActionListener() {
		    public void actionPerformed(ActionEvent event) {
			// create an InfoIconSelector with the current parameters
			String iconFile = (String) listBox.getSelectedValue();
			iconFile = MainCompositionToolWindow.CONFIG_DIR+"icons/"+iconFile;
			InfoIconSelector newIconFilter = new InfoIconSelector(keyPathInput.getText(),
									      valueInput.getText(),
									      currentIcon.getIcon());
			// add this icon selector before the current icon being displayed
			try {
			    infoIconSelectors.insertElementAt(newIconFilter,currentIconFilter);
			} catch (ArrayIndexOutOfBoundsException exception) {
			    System.err.println("Invalid index in icon filter vector "+currentIconFilter);
			    exception.printStackTrace();
			}
			// close the dialog
			iconInsertWindow.dispose();
		    }
		});
		insertPanel.add(inputPanel,BorderLayout.WEST);
		insertPanel.add(mainIconPane,BorderLayout.EAST);
		insertPanel.add(done,BorderLayout.SOUTH);
		insertPanel.setPreferredSize(new Dimension(250,250));
		iconInsertWindow.setSize(260,260);
		iconInsertWindow.getContentPane().add(insertPanel);
		//Window owner = (Window) iconChooserWindow.getParent();
		//Dimension d1 = iconChooserWindow.getSize();
		//Dimension d2 = owner.getSize();
		//int x = Math.max((d2.width-d1.width)/2, 0);
		//int y = Math.max((d2.height-d1.height)/2, 0);
		//iconChooserWindow.setBounds(x, y, d1.width, d1.height);
		iconInsertWindow.setVisible(true);
	    }
	});
	JButton delete = new JButton("Delete");
	delete.addActionListener(new ActionListener() {
	    public void actionPerformed(ActionEvent event) {
		// remove the currently displayed icon filter and 
		// display the next one
		// don't allow the last icon filter (wild card icon
		// filter) to be removed ever
		if (currentIconFilter == (infoIconSelectors.size()-1)) {
		    // last one, don't allow it to be removed
		    return;
		} else {
		    infoIconSelectors.remove(currentIconFilter);
		    try {
			InfoIconSelector currentSelector = (InfoIconSelector) 
			    infoIconSelectors.elementAt(currentIconFilter);
			displayIconFilter(currentSelector);
		    } catch (ArrayIndexOutOfBoundsException exception) {
			System.out.println("Invalid index in icon filter vector "+currentIconFilter);
			exception.printStackTrace();
		    }
		}
	    }
	});
	modifyPanel.add(insert);
	modifyPanel.add(delete);
	// put everything in the icon chooser window
	iconChooserWindow.getContentPane().add(modifyPanel,BorderLayout.SOUTH);
	iconChooserWindow.getContentPane().add(panel,BorderLayout.NORTH);
	iconChooserWindow.setVisible(true);
    }
	
    private void displayIconFilter(InfoIconSelector currentSelector) {
	// get the key path, value and icon for the current icon selector
	String keyPath = currentSelector.getKeyPathExpression();
	String valueExpression = currentSelector.getValueExpression();
	Icon icon = currentSelector.getIcon();
	// display these
	keyLabel.setText(keyPath);
	valueLabel.setText(valueExpression);
	iconLabel.setIcon(icon);
    }
}
