/**
 * Thi class acts as a cell renderer for icons displayed in the
 * <a href="InformationDirectoryTreeView">InformationDirectoryTreeView</a>.
 * It chooses the right icon with the help of an
 * <a href="InformationDirectoryIconChooser">InformationDirectoryIconChooser</a>
 * This is a flexible convenient way of having customized icons for
 * nodes in the tree.
 * @author Nirmal (nmukhi)
 */

package idsbrowser;

import java.awt.Color;
import java.awt.Component;
import java.awt.Graphics;
import javax.swing.Icon;
import javax.swing.JLabel;
import javax.swing.JTree;
import javax.swing.UIManager;
import javax.swing.tree.TreeCellRenderer;
import ids.InformationDirectoryNode;

public class InformationDirectoryIconRenderer extends JLabel 
    implements TreeCellRenderer {
    public static InformationDirectoryIconChooser iconChooser;
    private boolean selected;
    // graphical information required for rendering
    private Color textSelectionColor;
    private Color textNonSelectionColor;
    private Color bkSelectionColor;
    private Color bkNonSelectionColor;
    private Color borderSelectionColor;

    /**
     * Creates an icon renderer with the default icon chooser
     */
    public InformationDirectoryIconRenderer() {
	// create a default icon chooser from the icon chooser 
	// factory
	// call the one param contructor with this icon chooser
	this(InformationDirectoryIconChooserFactory.createDefaultIconChooser());
    }

    /**
     * Creates an icon renderer with the specified icon chooser
     * @param iconChooser the icon chooser to be used for this cell
     * renderer
     */
    public InformationDirectoryIconRenderer(InformationDirectoryIconChooser iconChooser) {
	super();
	// set the default colors for the icon
	textSelectionColor = UIManager.getColor("Tree.selectionForeground");
	textNonSelectionColor = UIManager.getColor("Tree.textForeground");
	bkSelectionColor = UIManager.getColor("Tree.selectionBackground").brighter();
	bkNonSelectionColor = UIManager.getColor("Tree.textBackground").brighter();
	borderSelectionColor = UIManager.getColor("Tree.selectionBorderColor");
	setOpaque(false);
	// set the icon chooser if needed
	if (this.iconChooser == null)
	    this.iconChooser = iconChooser;
    }

    public void browseIconFilters() {
	iconChooser.browseIconFilters();
    }

    /**
     * Figure out the icon to use for the specified node
     * @param tree the tree containing the icon to be rendered
     * @param value the value contained in this node of the tree
     * @param sel whether this node is currently selected or not
     * @param expanded whether this node has been expanded or not
     * @param leaf whether this node is a leaf node or not
     * @param row which row of the tree this node belongs in
     * @param hasFocus whether or not the node currently has focus
     * @return Component the component to be painted in order to 
     * render the icon for this node
     */
    public Component getTreeCellRendererComponent(JTree tree, Object value, 
						  boolean sel, boolean expanded, 
						  boolean leaf, int row, 
						  boolean hasFocus) {
	// create a new renderer component
	// this is a temporary way of getting around painting problems 
	// created when only one renderer is used and will be fixed
	// if there is time (Nirmal)
	InformationDirectoryIconRenderer renderer = 
	    new InformationDirectoryIconRenderer(iconChooser);
	// get the information object stored inside the node
	InformationDirectoryNode infoNode = (InformationDirectoryNode)
	    ((InformationDirectoryTreeNode)value).getUserObject();
	// use the icon chooser to get an icon for this file
	// given information on whether it is expanded 
	renderer.setIcon(iconChooser.getMatchingIcon(infoNode,expanded));
	// set the text for the node to be rendered
	renderer.setText(value.toString());
	renderer.selected = sel;
	// return the endering component
	return renderer;
    }

    /**
     * Paint this component. This will paint an icon on the tree.
     * @param graphics the graphics context to be used for 
     * painting.
     */
    public void paintComponent(Graphics graphics) {
	// draw the background with icon if any
	Color bColor = getBackground();
	Icon icon = getIcon();
	graphics.setColor(bColor);
	int offset = 0;
	if(icon != null && getText() != null) 
	    offset = (icon.getIconWidth() + getIconTextGap());
	graphics.fillRect(offset, 0, getWidth() - 1 - offset,
			  getHeight() - 1);
	// draw a border if selected
	if (selected) {
	    graphics.setColor(borderSelectionColor);
	    graphics.drawRect(offset, 0, getWidth()-1-offset, getHeight()-1);
	}
	// call superclass paint to do children and other stuff
	super.paintComponent(graphics);
    }
}

