/**
 * This class is used to display a set of icons in rows.
 * A list of icons can be supplied for display, and the display
 * can be cleared. It is used to diplay search results when
 * searching the information directory using the browser.
 * @author Nirmal (nmukhi)
 * @see MainBrowserWindow
 */

package idsbrowser;

import java.awt.dnd.DragGestureListener;
import java.awt.dnd.DragGestureEvent;
import java.awt.dnd.DragSourceListener;
import java.awt.dnd.DragSourceEvent;
import java.awt.dnd.DragSourceDragEvent;
import java.awt.dnd.DragSourceDropEvent;
import java.awt.dnd.DnDConstants;
import java.awt.dnd.DragSource;
import java.awt.dnd.DropTargetListener;
import java.awt.dnd.DropTarget;
import java.awt.dnd.DropTargetEvent;
import java.awt.dnd.DropTargetDragEvent;
import java.awt.dnd.DropTargetDropEvent;
import java.awt.datatransfer.Transferable;
import java.awt.datatransfer.DataFlavor;
import java.awt.datatransfer.UnsupportedFlavorException;
import java.awt.event.MouseListener;
import java.awt.event.MouseEvent;
import java.awt.Point;
import java.awt.Cursor;
import java.awt.Component;
import java.awt.GridLayout;
import javax.swing.tree.TreeCellRenderer;
import javax.swing.JPanel;
import ids.InformationDirectoryNode;
import ids.InformationInfo;

public class InformationDirectorySearchWindow extends JPanel 
    implements DragGestureListener, MouseListener {
    public final static DragSourceListener dragSourceListener = new SearchWindowDragSourceListener();
    private DragSource dragSource  = DragSource.getDefaultDragSource();
    private DraggableInformationDirectoryTreeNode [] searchResults;
    private InformationDirectoryTreeView tree;

    /** 
     * Create a search window with no icons
     */
    public InformationDirectorySearchWindow(InformationDirectoryTreeView tree) {
	super();
	this.tree = tree;
	searchResults = null;
	this.setLayout(new GridLayout());
	// make the background the same as the tree's backgrouns
	this.setBackground(tree.getBackground());
	// add myself as a mouse listener
	this.addMouseListener(this);
	// make this a drag source
	dragSource.createDefaultDragGestureRecognizer(this, 
						      DnDConstants.ACTION_COPY, 
						      this);
    }

    /**
     * Display the provided set of nodes in the search window
     */
    public void displaySearchResults(DraggableInformationDirectoryTreeNode [] results) {
	// clear what's being displayed now
	clear();
	// iterate through the search results
	for (int i=0; i<results.length; i++) {
	    // use the cell renderer to get a component which will 
	    // render the current item
	    TreeCellRenderer renderer = tree.getCellRenderer();
	    Component renderingComponent = 
		renderer.getTreeCellRendererComponent(tree, results[i], false,
						      false, true, 0, false);
	    // set the background color to match the current one
	    renderingComponent.setBackground(getBackground());
	    // set the name of the rendering component
	    String nodeName = 
		((InformationDirectoryNode)results[i].getUserObject()).getName();
	    renderingComponent.setName(nodeName);
	    // add the component to the search window
	    this.add(renderingComponent);
	}
	this.searchResults = results;
	this.revalidate();
    }

    /**
     * Clear the displayed nodes from the search window
     */
    public void clear() {
	System.out.println("Clearing search window...");
	// remove all the components in the search window
	this.removeAll();
	// paint the window again to make children disappear
	this.repaint();
    }	

    /**
     * This method is required by the <var>dragGestureListener</var> interface
     * and is triggered by the initiation of a drag 
     * @param dragGestureEvent the event that was interpreted as a drag 
     * gesture
     */
    public void dragGestureRecognized(DragGestureEvent dragGestureEvent) {
	// get the origin of this event
	Point dragOrigin = dragGestureEvent.getDragOrigin();
	// get the node which is placed here
	Component componentDragged = this.getComponentAt(dragOrigin);
	String componentName = componentDragged.getName();
	DraggableInformationDirectoryTreeNode draggableNode = null;
	// iterate through the search results to match this component
	// with a draggable node
	for (int i=0; i<searchResults.length; i++) {
	    String resultName = 
		((InformationDirectoryNode)searchResults[i].getUserObject()).getName();
	    if (resultName.equals(componentName)) {
		draggableNode = searchResults[i];
		break;
	    }
	}
	if (draggableNode == null) {
	    // weird, should never happen
	    return;
	}
	// start the drag
	System.out.println("Starting to drag "+draggableNode.getUserObject());
	dragSource.startDrag(dragGestureEvent,
			     DragSource.DefaultCopyDrop,
			     draggableNode,
			     dragSourceListener);
    }

    // methods for implementing the mouse listener interface
    public void mouseClicked(MouseEvent event) { 
	// check if this is a double click
	if (event.getClickCount() == 2) {
	    // get the node over which the click occurred
	    Point clickOrigin = event.getPoint();
	    // get the component at that point
	    Component clickedComponent = this.getComponentAt(clickOrigin);
	    String componentName = clickedComponent.getName();
	    DraggableInformationDirectoryTreeNode node = null;
	    // iterate through the search results to match this component
	    // with a node
	    for (int i=0; i<searchResults.length; i++) {
		String resultName = ((InformationDirectoryNode)
				     searchResults[i].getUserObject()).getName();
		if (resultName.equals(componentName)) {
		    node = searchResults[i];
		    break;
		}
	    }
	    if (node == null) {
		// weird, should never happen
		return;
	    }
	    // get the information directory node which is in here
	    InformationDirectoryNode infoNode = 
		(InformationDirectoryNode) node.getUserObject();
	    // check if this corresponds to information
	    if (infoNode.isInformation() == 1) {
		// make this an information object
		InformationInfo info = (InformationInfo) infoNode;
		// open a window to display this in a pretty format
		Cursor currentCursor = getCursor();
		setCursor(new Cursor(Cursor.WAIT_CURSOR));
		tree.displayInfo(info);
		setCursor(currentCursor);
	    }
	}
    }

    public void mouseEntered(MouseEvent event) { }

    public void mouseExited(MouseEvent event) { }

    public void mousePressed(MouseEvent event) { }

    public void mouseReleased(MouseEvent event) { }

    /**
     * This class is a default drag source listener
     */
    public static class SearchWindowDragSourceListener implements DragSourceListener {
	public void dragDropEnd(DragSourceDropEvent dragSourceDropEvent) { }
	public void dragEnter(DragSourceDragEvent dragSourceDragEvent) { }
	public void dragExit(DragSourceEvent dragSourceEvent) { }
	public void dragOver(DragSourceDragEvent dragSourceDragEvent) { }
	public void dropActionChanged(DragSourceDragEvent dragSourceDragEvent) { }
    }
}
    
