/**
 * <tt>InformationDirectoryTreeNode</tt> represents a node in the
 * <A href="InformationDirectoryTreeView">InformationDirectoryTreeView</a>.
 * It contains an <tt>InformationDirectoryNode</tt> object in addition to 
 * other information required for the graphical tree structure.
 * @author Nirmal (nmukhi)
 * @see InformationDirectoryTreeView
 */

package idsbrowser;

import javax.swing.tree.DefaultMutableTreeNode;
import ids.InformationDirectoryNode;
import ids.InformationInfo;
import ids.InformationDirectoryContext;
import registry.InvalidContextException;

public class InformationDirectoryTreeNode extends DefaultMutableTreeNode {
    private boolean hasBeenRead;

    /**
     * Create a node given the IDS node it is associated with
     * @param idsNode the information directory node this graphical
     * node represents
     */
    public InformationDirectoryTreeNode(InformationDirectoryNode idsNode) {
	super(idsNode);
	hasBeenRead = false;
    }

    /**
     * Look up this tree's children and add them to the tree model
     * @return true if this node was read, false if it had already
     * been read before
     * @see #readChildren(boolean)
     */
    public boolean readTree() {
	return readTree(false);
    }

    /**
     * Refresh the subtree rooted at this node.
     * This will remove all children of this node,
     * and read the subtree again from scratch
     */
    public void refresh() {
	this.removeAllChildren();
	hasBeenRead = false;
	readTree();
    }
    
    /**
     * Look up this tree's children and add them to the tree model.
     * @param readChildren recursively call this method on the 
     * children if true, do not recurse if false
     * @return true if this node was read, false if it had already
     * been read before
     */
    public boolean readTree(boolean readChildren) {
	// if this node has already been read, return false
	if(hasBeenRead)
	    return false;
	// get a list of all the children of the information directory
	// node this corresponds to
	InformationDirectoryNode idsNode = (InformationDirectoryNode)getUserObject();
	if (idsNode.isContext()!=1) {
	    // not a context, no children
	    return false;
	}
	// this is a context
	InformationDirectoryContext idsContext = (InformationDirectoryContext)idsNode;
	try {
	    // get its children
	    InformationDirectoryNode [] children = idsContext.getChildren();
	    // go through all the children and create tree nodes for each
	    for (int i=0; i<children.length; i++) {
		InformationDirectoryNode currentNode = children[i];
		InformationDirectoryTreeNode subNode = null;
		if (currentNode.isContext()==1) {
		    // this is a context
		    // first create a node for it
		    subNode = new InformationDirectoryTreeNode(currentNode);
		    // if this is a deep read, read from this context as well
		    if (readChildren)
			subNode.readTree(readChildren);
		} 
		if (currentNode.isInformation()==1) {
		    // this represents component information
		    // create a draggable node for it
		    subNode = 
			new DraggableInformationDirectoryTreeNode((InformationInfo)currentNode);
		}
		// add the newly created subnode if any as a child of this node
		if (subNode!=null)
		    add(subNode);
	    }
	    hasBeenRead=true;
	} catch (InvalidContextException exception) {
	    // context invalid, could not get children
	    System.err.println("Got exception "+exception);
	    // return false 
	    return false;
	}
	// return true since there was a change
	return true;
    }

    /**
     * Returns a string representation of the node
     * This is rendered next to the icon for this node
     * @return the name of the node
     */
    public String toString() {
	InformationDirectoryNode idsNode = (InformationDirectoryNode)getUserObject();
	// get the name of the ids node, but only use the last portion of the
	// name
	String fullName = idsNode.getName();
	// if this is a context, get the last but one slash
	int lastSlashPos = -1;
	if (idsNode.isContext()==1) {
	    // skip the last character (which is a slash)
	    lastSlashPos = fullName.substring(0,fullName.length()-1).lastIndexOf("/");
	} else {
	    lastSlashPos = fullName.lastIndexOf("/");
	}
	String truncatedName = fullName;
	if (lastSlashPos!=-1)
	    truncatedName = fullName.substring(lastSlashPos+1);
	return truncatedName;
    }

    /**
     * Checks whether this node corresponds to a leaf or not
     * @return false if this node corresponds to a context, 
     * true otherwise
     */
    public boolean isLeaf() {
	InformationDirectoryNode node = (InformationDirectoryNode)getUserObject();
	return (!(node.isContext()==1));
    }
}
