/**
 * This class is the main window for the information directory browser.
 * It contains a tree of directory information that can be navigated 
 * easily. Items can be dragged into and out of the window. The
 * window also has its own trash can where items can be dropped for 
 * deletion to take place.
 * @author Nirmal (nmukhi)
 */

package idsbrowser;

import javax.swing.JFrame;
import javax.swing.JComponent;
import javax.swing.JSplitPane;
import javax.swing.JScrollPane;
import javax.swing.JMenuBar;
import javax.swing.JMenu;
import javax.swing.JButton;
import javax.swing.JMenuItem;
import javax.swing.JDialog;
import javax.swing.ImageIcon;
import javax.swing.Icon;
import javax.swing.JLabel;
import javax.swing.JTextField;
import java.awt.TextField;
import javax.swing.JTextArea;
import javax.swing.AbstractButton;
import javax.swing.JRadioButton;
import javax.swing.JPanel;
import javax.swing.JToolBar;
import javax.swing.ButtonGroup;
import javax.swing.UnsupportedLookAndFeelException;
import javax.swing.UIManager;
import javax.swing.BoxLayout;
import javax.swing.Box;
import javax.swing.KeyStroke;
import java.awt.event.KeyEvent;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.Dimension;
import java.awt.Window;
import java.awt.Cursor;
import java.awt.Color;
import java.awt.GridLayout;
import java.awt.BorderLayout;
import ids.InformationDirectoryContext;
import ids.InformationDirectoryService;
import ids.ServerNotFoundException;
import registry.NotFoundException;

public class MainBrowserWindow extends JFrame implements ActionListener {
    private InformationDirectoryContext context;
    private InformationDirectoryService idsService;
    private InformationDirectoryTreeView treeWindow;
    private InformationDirectorySearchWindow searchWindow;
    private TextField contextField;
    private JButton searchButton;
    private JButton newBrowser;
    private JButton customIcons;
    private JButton quitButton;

    // things required to get input during dialog interaction
    // this is untidy, will get rid of it later
    private JDialog dialog;
    private JTextField keyPathInput;
    private JTextField valueInput;
    private ButtonGroup group;

    /**
     * Create a main browser window
     * @param idsService the information directory service
     * @param context the initial context to be used
     */
    public MainBrowserWindow(InformationDirectoryService idsService,
			     InformationDirectoryContext context,
			     String fullContextNameString) {
	// set the title of the frame
	super("Information Directory Service Browser");
	// use the Java look and feel
	try {
	    UIManager.setLookAndFeel(UIManager.getCrossPlatformLookAndFeelClassName());
	} catch (Exception exception) { 
	    System.out.println("Unable to set look and feel, got exception "+exception);
	    exception.printStackTrace();
	}
	// set the ids service
	this.idsService = idsService;
	// set the initial context
	this.context = context;
	// create the top panel
	JPanel topPanel = new JPanel();
	topPanel.setLayout(new BorderLayout());
	JPanel buttonPanel = new JPanel();
	buttonPanel.setLayout(new GridLayout(2,1));
	buttonPanel.setPreferredSize(new Dimension(500,70));
	topPanel.add(buttonPanel,BorderLayout.NORTH);
	// create the text field to hold the full context name
	contextField = new TextField(fullContextNameString);
	contextField.addActionListener(this);
	contextField.setBackground(Color.white);
	// add to the top panel
	buttonPanel.add(contextField);
	// create the toolbar to hold the search and quit buttons
	JToolBar toolBar = new JToolBar();
	// add the toolbar to the top panel
	buttonPanel.add(toolBar);
	// add search and quit buttons
	Icon searchIcon = new ImageIcon(compositiontool.view.MainCompositionToolWindow.CONFIG_DIR+"icons/searchicon.gif");
	Icon quitIcon = new ImageIcon(compositiontool.view.MainCompositionToolWindow.CONFIG_DIR+"icons/Delete.gif");
	searchButton = new JButton(searchIcon);
	searchButton.setToolTipText("search context");
	searchButton.addActionListener(this);
	newBrowser = new JButton("New");
	newBrowser.setToolTipText("new browser");
	newBrowser.addActionListener(this);
	customIcons = new JButton("Icons");
	customIcons.setToolTipText("customise icons");
	customIcons.addActionListener(this);
	quitButton = new JButton(quitIcon);
	quitButton.setToolTipText("quit");
	quitButton.addActionListener(this);
	toolBar.add(searchButton);
	toolBar.add(newBrowser);
	toolBar.add(customIcons);
	toolBar.add(quitButton);
	// scroll pane for tree window
	// create tree (with trashcan) and add to left panel
	treeWindow = new InformationDirectoryTreeView(context);
	JScrollPane treePanel = new JScrollPane(treeWindow,
						JScrollPane.VERTICAL_SCROLLBAR_AS_NEEDED,
						JScrollPane.HORIZONTAL_SCROLLBAR_AS_NEEDED);
	// set the preferred size of the tree
	treePanel.setPreferredSize(new Dimension(250,400));
	// scroll pane for search window
	// create search window
	searchWindow = new InformationDirectorySearchWindow(treeWindow);
	JScrollPane searchPanel = new JScrollPane(searchWindow,
						  JScrollPane.VERTICAL_SCROLLBAR_AS_NEEDED,
						  JScrollPane.HORIZONTAL_SCROLLBAR_AS_NEEDED);
	// set the preferred size of the search window
	searchPanel.setPreferredSize(new Dimension(250,400));
	// split pane for main window
	JSplitPane mainPanel = new JSplitPane(JSplitPane.HORIZONTAL_SPLIT,
					      treePanel, searchPanel);
	mainPanel.setPreferredSize(new Dimension(520,410));
	topPanel.add(mainPanel,BorderLayout.CENTER);
	// add split pane to this frame
	this.setContentPane(topPanel);
	// pack and display
	this.pack();
	this.setVisible(true);
    }

    /**
     * This method is required by the action listener interface.
     * It will be triggered whenever there is an action on either
     * of the menu items
     */
    public void actionPerformed(ActionEvent event) {
	// get the source of the action
	Object source = event.getSource();
	if (source == searchButton) {
	    Cursor currentCursor = getCursor();
	    this.addNotify();
	    setCursor(new Cursor(Cursor.WAIT_CURSOR));
	    // create a window to get the required input for the search
	    JDialog searchParamWindow = new JDialog(this,"Search Parameters",true);
	    dialog = searchParamWindow;
	    JPanel panel = new JPanel();
	    searchParamWindow.getContentPane().add(panel);
	    panel.setLayout(new BoxLayout(panel,BoxLayout.Y_AXIS));
	    // specify the search key path
	    keyPathInput = new JTextField(28);
	    keyPathInput.setPreferredSize(new Dimension(145,30));
	    panel.add(new JLabel("Key path: "));
	    panel.add(keyPathInput);
	    // specify the search value
	    valueInput = new JTextField();
	    valueInput.setPreferredSize(new Dimension(145,30));
	    panel.add(new JLabel("Value: "));
	    panel.add(valueInput);
	    // specify if this is a deep search or not
	    JRadioButton deepButton = new JRadioButton("Deep search");
	    deepButton.setActionCommand("deep");
	    JRadioButton shallowButton = new JRadioButton("Shallow search");
	    shallowButton.setActionCommand("shallow");
	    shallowButton.setSelected(true);
	    group = new ButtonGroup();
	    group.add(deepButton);
	    group.add(shallowButton);
	    panel.add(deepButton);
	    panel.add(shallowButton);
	    // add a done button to signal the end of input
	    JButton done = new JButton("Done");
	    done.addActionListener(this);
	    panel.add(done);
	    panel.setPreferredSize(new Dimension(150,250));
	    searchParamWindow.setSize(160,260);
	    Window owner = (Window) searchParamWindow.getParent();
	    //searchParamWindow.setResizable(false);
	    Dimension d1 = searchParamWindow.getSize();
	    Dimension d2 = owner.getSize();
	    int x = Math.max((d2.width-d1.width)/2, 0);
	    int y = Math.max((d2.height-d1.height)/2, 0);
	    searchParamWindow.setBounds(x, y, d1.width, d1.height);
	    searchParamWindow.setVisible(true);
	    this.addNotify();
	    setCursor(currentCursor);
	} else if (source == quitButton) {
	    quit();
	} else if (source == contextField) {
	    Cursor currentCursor = getCursor();
	    this.addNotify();
	    setCursor(new Cursor(Cursor.WAIT_CURSOR));
	    // assume the context dialog was completed
	    System.out.println("Setting new context...");
	    String newContextName = contextField.getText();
	    this.changeContext(newContextName);
	    this.addNotify();
	    setCursor(currentCursor);
	} else if (source == customIcons) {
	    // bring up dialog for icon display
	    Cursor currentCursor = getCursor();
	    this.addNotify();
	    setCursor(new Cursor(Cursor.WAIT_CURSOR));
	    treeWindow.browseIconFilters();
	    this.addNotify();
	    setCursor(currentCursor);
	} else if (source == newBrowser) {
	    new MainBrowserWindow(idsService,context,contextField.getText());
	} else {
	    // search dialog must have been completed
	    Cursor currentCursor = getCursor();
	    this.addNotify();
	    setCursor(new Cursor(Cursor.WAIT_CURSOR));
	    // assume the search dialog was completed
	    System.out.println("Processing search dialog content...");
	    String searchParam = keyPathInput.getText();
	    String searchValue = valueInput.getText();
	    boolean deepSearch = true;
	    if (!group.getSelection().getActionCommand().equals("deep")) {
		deepSearch = false;
	    }
	    search(searchParam,searchValue,deepSearch);
	    keyPathInput = null;
	    valueInput = null;
	    group = null;
	    this.addNotify();
	    setCursor(currentCursor);
	    // kill the dialog
	    dialog.dispose();
	}
    }

    /**
     * Carries out a search with the specified search parameters
     * and displays icons for matching information in the search 
     * window
     * @param searchParameter the paramter being searched
     * @param searchParameterValue the value the parameter should have
     * @param deepSearch search in all subcontexts recursively if true,
     * otherwise search only the children of this context without
     * exploring subcontexts
     */
    public void search(String searchParameter, String searchParameterValue,
		       boolean deepSearch) {
	// call the tree method to perform this search
	DraggableInformationDirectoryTreeNode [] results = 
	    treeWindow.search(searchParameter, searchParameterValue, deepSearch);
	// give the list of matching information to the 
	// search window for display
	if (results != null) {
	    // get the number of results
	    int numResults = results.length;
	    // display results 4 per row
	    GridLayout layout = (GridLayout)searchWindow.getLayout();
	    layout.setColumns(4);
	    layout.setRows(numResults/4+2);
	    System.out.println("Displaying search results in a "+layout.getRows()+"x"+layout.getColumns()+" window");
	    searchWindow.displaySearchResults(results);
	}
    }

    /**
     * Changes the directory context and clears the search window
     * @param newContextName the new directory context to be used
     */
    public void changeContext(String newContextName) {
	try {
	    // get the new context using the ids service
	    InformationDirectoryContext newContext = idsService.getContext(newContextName);
	    this.context = newContext;
	    // call the tree method to change the context
	    treeWindow.changeContext(newContext);
	} catch (NotFoundException exception) {
	    System.out.println("Could not change to context "+newContextName+
			       ", got exception "+exception);
	    exception.printStackTrace();
	} catch (ServerNotFoundException exception) {
	    System.out.println("Could not find server in context "+newContextName+
			       " got exception "+exception);
	    exception.printStackTrace();
	}
	// call the search window method to clear the window
	// of icons if there are any
	searchWindow.clear();
    }

    /** 
     * Closes down all windows and exits the program
     * normally.
     */
    public void quit() {
	this.setVisible(false);
    }
    
    /**
     * main program for testing stuff out
     */
    public static void main(String [] args) {
	try {
	    // get the ids service
	    ids.InformationDirectoryServiceImplementation idsService = 
		new ids.InformationDirectoryServiceImplementation();
	    // create an initial context
	    InformationDirectoryContext context = 
		idsService.getContext(args[0]);
	    // create a browser for this
	    new MainBrowserWindow(idsService,context,args[0]);
	} catch (Exception exception) {
	    exception.printStackTrace();
	}
    }
}
