/**
 * <tt>DraggableRegistryTreeNode</tt> represents a node in the
 * <a href="RegistryTreeView">RegistryTreeView</a>
 * that can be dragged. When the node is dragged, a text description
 * of the user object will be transferred along with the dragged icon.
 * The drop target can then interpret this text for data transfer to
 * take place via drag and drop.
 * @see RegistryTreeView
 */

package registrybrowser;

import java.awt.datatransfer.Transferable;
import java.awt.datatransfer.DataFlavor;
import java.awt.datatransfer.UnsupportedFlavorException;
import java.io.File;
import java.io.FileWriter;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.LinkedList;
import registry.InstanceInfo;
import registry.InstanceXMLUtils;

public class DraggableRegistryTreeNode 
    extends RegistryTreeNode implements Transferable {

    /**
     * <tt>STRING_FLAVOR</tt> is a <tt>String</tt> object with 
     * contents of the <tt>InstanceInfo</tt> object stored in this 
     * node
     */
    public final static DataFlavor STRING_FLAVOR = DataFlavor.stringFlavor;
    public final static DataFlavor FILE_FLAVOR = DataFlavor.javaFileListFlavor;
    public static final int MAX_TMP_FILES = 256;

    /**
     * All the available flavors for the data.
     *
     */
    public static DataFlavor flavors[] = {
	STRING_FLAVOR,
	FILE_FLAVOR
    };

    /**
     * Return the list of flavors supported.
     * @return the list of flavors
     */
    public DataFlavor[] getTransferDataFlavors() {
	return flavors;
    }

    /**
     * Create a node given the IDS information object it is associated with
     * @param idsInfo the information directory information this graphical
     * node represents
     */
    public DraggableRegistryTreeNode(InstanceInfo info) {
	super(info);
    }

    /**
     * Get the transferable object corresponding to the specified data flavor.
     * @param flavor the data flavor.
     * @return the data object corresponding to the specified data flavor.
     */
    public Object getTransferData(DataFlavor flavor) 
	throws UnsupportedFlavorException, IOException {
	Object returnObject = null;
	if (flavor.equals(STRING_FLAVOR)) {
	    // get the information info object
	    InstanceInfo info = (InstanceInfo)getUserObject();
	    // write the info into an XML string
	    String xmlDescription = InstanceXMLUtils.createXMLFromInfo(info,info.getName());
	    returnObject = xmlDescription;
	} else if (flavor.equals(FILE_FLAVOR)) {
	    // write the XML stuff into a file
	    InstanceInfo info = (InstanceInfo)getUserObject();
	    // write the info into an XML string
	    String xmlDescription = InstanceXMLUtils.createXMLFromInfo(info,info.getName());
	    // create the tmp directory where the file will reside
	    File tmpDir = new File(compositiontool.view.MainCompositionToolWindow.CONFIG_DIR+"tmp/");
	    if (!tmpDir.exists()) {
		tmpDir.mkdir();
	    }
	    // write the string into a file in the tmp directory
	    // first create a file with a unique name
	    File xmlFile = null;
	    for (int i=0; i<MAX_TMP_FILES; i++) {
		xmlFile = new File(tmpDir,"xml_information"+i+".txt");
		if (!xmlFile.exists())
		    break;
	    }
	    // write the stuff into the file
	    FileWriter fileWriter = new FileWriter(xmlFile);
	    try {
		fileWriter.write(xmlDescription);
		fileWriter.close();
	    } catch (IOException exception) {
		System.err.println("Could not write XML information into tmp file, got exception");
		exception.printStackTrace();
	    }
	    // make an array of file objects
	    LinkedList data = new LinkedList();
	    data.add(xmlFile);
	    return data;
	} else {
	    // some other flavor was requested
	    throw new UnsupportedFlavorException(flavor);
	}
	return returnObject;
    }

    /**
     * Checks if the specified data flavor is supported by this object.
     * @param flavor the specified data flavor.
     * @return true if the specified falvor is supported, false otherwise.
     */
    public boolean isDataFlavorSupported(DataFlavor flavor) {
	if (flavor.equals(STRING_FLAVOR))
	    return true;
	return false;
    }
}
