/**
 * This class is used for choosing an icon for an info object.
 * It makes its selection given icons to be used for expanded and
 * unexpanded contexts. It uses the <tt>InfoIconSelector</tt> 
 * to select an icon for leaf info objects 
 * @see idsbrowser.InfoIconSelector
 */

package registrybrowser;

import javax.swing.Icon;
import java.util.Vector;
import java.util.Enumeration;
import registry.RegistryNode;
import util.Info;
import idsbrowser.InfoIconSelector;

public class RegistryIconChooserImplementation 
    implements RegistryIconChooser {
    private Icon contextIcon;
    private Icon expandedContextIcon;
    private Vector infoIconSelectors;

    /**
     * Create an icon chooser with the specified icons to be used for a context
     * and the given list of info icon selectors
     * @param contextIcon the icon to be used for contexts
     * @param expandedContextIcon the icon to be used for an expanded context
     * @param infoIconSelectors the list of info icon selectors used for
     * selecting an icon for a leaf
     */
    public RegistryIconChooserImplementation(Icon contextIcon,
					     Icon expandedContextIcon,
					     Vector infoIconSelectors) {
	// set the context icons and info icon selectors
	this.contextIcon = contextIcon;
	this.expandedContextIcon = expandedContextIcon;
	this.infoIconSelectors = infoIconSelectors;
    }

    /**
     * Get the matching icon for the specified info object
     * @param infoNode the registry node object for which an icon is desired
     * @expanded whether this node has been expanded or not in the tree
     * @return the icon to be used for this info object. Note that for leaf
     * info objects, the list of info icon selectors is examined for a match
     * and the first selector which can match this info object is used.
     */
    public Icon getMatchingIcon(RegistryNode infoNode, boolean expanded) {
	// check if this is a context
	if (infoNode.isContext()==1) {
	    // we have a context, return expanded or normal icon as appropriate
	    if (expanded)
		return expandedContextIcon;
	    else
		return contextIcon;
	} 
	if (infoNode.isInstance()==1) {
	    Info info = (Info)infoNode;
	    // we have a leaf, go through the list of info icon selectors
	    Enumeration iconSelectorList = infoIconSelectors.elements();
	    while(iconSelectorList.hasMoreElements()) {
		InfoIconSelector iconSelector = (InfoIconSelector)iconSelectorList.nextElement();
		// check if this info icon selector can match this info object
		Icon matchingIcon = iconSelector.getMatchingIcon(info);
		if (matchingIcon != null) {
		    // got a match, return the icon
		    return matchingIcon;
		}
	    }
	    // no match, return null
	    return null;
	}
	// wasn't a context or info, could not find a matching icon
	return null;
    }
}
