/**
 * This class represents a tree view of the ifnormation in the
 * registry.
 * It uses methods from the registry context to
 * navigate through the instances and displays it appropriately
 * in this view.
 * This view can optionally include a trash can icon which is removed
 * at the end of the session
 * @author Nirmal (nmukhi)
 * @see MainBrowserWindow
 */

package registrybrowser;

import java.awt.dnd.DragGestureListener;
import java.awt.dnd.DragGestureEvent;
import java.awt.dnd.DragSourceListener;
import java.awt.dnd.DragSourceEvent;
import java.awt.dnd.DragSourceDragEvent;
import java.awt.dnd.DragSourceDropEvent;
import java.awt.dnd.DnDConstants;
import java.awt.dnd.DragSource;
import java.awt.dnd.DropTargetListener;
import java.awt.dnd.DropTarget;
import java.awt.dnd.DropTargetEvent;
import java.awt.dnd.DropTargetDragEvent;
import java.awt.dnd.DropTargetDropEvent;
import java.awt.datatransfer.Transferable;
import java.awt.datatransfer.DataFlavor;
import java.awt.datatransfer.UnsupportedFlavorException;
import javax.swing.JTextArea;
import javax.swing.JFrame;
import javax.swing.JTree;
import javax.swing.JOptionPane;
import javax.swing.JScrollPane;
import javax.swing.tree.TreePath;
import javax.swing.tree.DefaultTreeModel;
import java.awt.event.MouseListener;
import java.awt.event.MouseEvent;
import java.awt.Point;
import java.awt.Dimension;
import java.awt.Color;
import java.awt.Cursor;
import java.awt.Font;
import java.util.Vector;
import java.util.Enumeration;
import java.io.IOException;
import registry.RegistryContext;
import registry.RegistryNode;
import registry.InstanceInfo;
import registry.InstanceXMLUtils;
import registry.InvalidContextException;
import registry.NotFoundException;
import registry.BadContextNameException;
import util.GeneralInfoUtils;
import util.Info;
import ccacore.ComponentID;

public class RegistryTreeView extends JTree 
    implements DragGestureListener, DropTargetListener, MouseListener {
    private RegistryContext context;
    public final static DragSourceListener dragSourceListener = new TreeDragSourceListener();
    private DropTarget dropTarget;
    private DragSource dragSource  = DragSource.getDefaultDragSource();
    private DefaultTreeModel model;

    /**
     * Create a registry tree view with the
     * specified context
     * @param context the context to be used
     */
    public RegistryTreeView(RegistryContext context) {
	super();
	// set the context
	this.context = context;
	// build the root node
	RegistryNode root = (RegistryNode)context;
	RegistryTreeNode rootNode = new RegistryTreeNode(root);
	// read the first set of children
	rootNode.readTree(false);
	// create the tree itself
	model = new DefaultTreeModel(rootNode);
	this.setModel(model);
	// set the cell renderer for the tree
	// use a default icon chooser
	this.setCellRenderer(new RegistryIconRenderer());
	// make this double buffered so icon rendering is smooth
	this.setDoubleBuffered(true);
	// create a tree expansion listener
	this.addTreeExpansionListener(new RegistryTreeExpansionListener(model));
	// add myself as a mouse listener
	this.addMouseListener(this);
	// make this a drag source and drop target
	dropTarget = new DropTarget(this,this);
	dragSource.createDefaultDragGestureRecognizer(this, 
						      DnDConstants.ACTION_COPY_OR_MOVE, 
						      this);
    }

    /**
     * Search the context for the nodes with specified parameter 
     * values. 
     * @param searchParameter the paramter being searched
     * @param searchParameterValue the value the parameter should have
     * @param deepSearch search in all subcontexts recursively if true,
     * otherwise search only the children of this context without
     * exploring subcontexts
     * @param searchWindow the search window where results should be 
     * displayed
     */
    public DraggableRegistryTreeNode [] search(String searchParameter, 
					       String searchParameterValue,
					       boolean deepSearch) {
	// get the list of information objects in the scope of the search
	InstanceInfo [] candidates = null;
	Vector searchResults = new Vector(); 
	// the search result vector will contain 
	// DraggableRegistryTreeNode objects
	if (deepSearch) {
	    // deep search, get all components below this context
	    try {
		candidates = context.getAllInstances();
	    } catch (InvalidContextException exception) {
		System.err.println("Got exception "+exception);
		exception.printStackTrace();
		return null;
	    }
	} else {
	    // get children in this context
	    RegistryNode [] children = null;
	    try {
		children = context.getChildren();
	    } catch (InvalidContextException exception) {
		System.err.println("Got exception "+exception);
		exception.printStackTrace();
		return null;
	    }
	    Vector candidateVector = new Vector();
	    // iterate through children and pick out the components 
	    for (int i=0; i<children.length; i++) {
		if (children[i].isInstance() == 1) {
		    candidateVector.add(children[i]);
		}
	    }
	    Enumeration componentList = candidateVector.elements();
	    candidates = new InstanceInfo[candidateVector.size()];
	    int position = 0;
	    while(componentList.hasMoreElements()) {
		candidates[position++] = (InstanceInfo) componentList.nextElement();
	    }
	}
	// iterate through the match candidates
	System.out.println("There are "+candidates.length+" possible matching components");
	for (int i=0; i<candidates.length; i++) {
	    int j = 0;
	    String [] values = null;
	    if (!searchParameterValue.equals(GeneralInfoUtils.WILD_CARD)) {
		// check for a match
		values = GeneralInfoUtils.deepGetValue(candidates[i],searchParameter);
		for (j=0; j<values.length; j++) {
		    String trimmedValue = values[j].trim();
		    String trimmedParam = searchParameterValue.trim();
		    System.out.println("Checking against "+trimmedValue);
		    if (!trimmedValue.equals(trimmedParam))
			break;
		}
	    }
	    if (searchParameterValue.equals(GeneralInfoUtils.WILD_CARD) ||
		j == values.length) {
		System.out.println("Matched!");
		// got a match, add to the list of matching information
		DraggableRegistryTreeNode matchingNode = 
		    new DraggableRegistryTreeNode((InstanceInfo)candidates[i]);
		searchResults.add(matchingNode);
	    }
	}
	// copy stuff from the search result vector into an array
	DraggableRegistryTreeNode [] finalNodeSet = 
	    new DraggableRegistryTreeNode[searchResults.size()];
	for (int i=0; i<searchResults.size(); i++) {
	    finalNodeSet[i] = (DraggableRegistryTreeNode)searchResults.elementAt(i);
	}
	// return the search results
	return finalNodeSet;
    }

    /**
     * Sets the context to the specfied context
     * @param context the new context the tree view
     * must display
     */
    public void changeContext(RegistryContext context) {
	// set the context
	this.context = context;
	// build the root node
	RegistryNode root = (RegistryNode)context;
	RegistryTreeNode rootNode = new RegistryTreeNode(root);
	rootNode.readTree();
	// reset the root node of the current tree
	DefaultTreeModel model = new DefaultTreeModel(rootNode);
	this.setModel(model);
	this.model = model;
    }

    /**
     * Opens a separate window that displays the provided information 
     * in a pretty format.<br>
     * It is assumed that the information object describes an instance
     * @param info the information object
     */
    public void displayInfo(InstanceInfo info) {
	// at this point we could get the XML string describing the instance
	// and use a full-fledged XML browser to view it
	// for now we just display a pretty string form of the info

	// convert the info into a pretty string describing the info
	String description = GeneralInfoUtils.prettyToString(info);
	// create a frame 
	JFrame frame = new JFrame(info.getName());
	// create a text area to display the information
	JTextArea displayArea = new JTextArea();
	// write the data to the text area
	displayArea.setFont(new Font("SansSerif",Font.BOLD,20));
	displayArea.append(info.getName());
	displayArea.append("\n");
	displayArea.setFont(new Font("SansSerif",Font.PLAIN,14));
	displayArea.append(description);
	displayArea.setEditable(false);
	// add the display area to the panel and display the frame
	JScrollPane panel = new JScrollPane(displayArea);
	panel.setPreferredSize(new Dimension(400,400));
	// add the scroll pane to the frame
	frame.setContentPane(panel);
	frame.pack();
	frame.setVisible(true);
    }

    /**
     * This method is required by the <var>dragGestureListener</var> interface
     * and is triggered by the initiation of a drag 
     * @param dragGestureEvent the event that was interpreted as a drag 
     * gesture
     */
    public void dragGestureRecognized(DragGestureEvent dragGestureEvent) {
	// get the origin of this event
	Point dragOrigin = dragGestureEvent.getDragOrigin();
	// check which node where the drag began at
	TreePath path = this.getPathForLocation(dragOrigin.x,dragOrigin.y);
	System.out.println("Drag event at "+path);
	RegistryTreeNode node = 
	    (RegistryTreeNode) path.getLastPathComponent();
	RegistryNode infoNode = (RegistryNode) node.getUserObject();
	// drags allowed only for information objects, check if this is one
	if (infoNode.isInstance() != 1) {
	    // not information object, return
	    return;
	}
	DraggableRegistryTreeNode draggableNode = 
	    (DraggableRegistryTreeNode) node;
	// if the action was move, remove this node from the tree and the context
	int action = dragGestureEvent.getDragAction();
	if (action == DnDConstants.ACTION_COPY) {
	    System.out.println("Detected move");
	    System.out.println("Removing node from parent...");
	    try {
		ComponentID id = ((InstanceInfo)infoNode).getComponentID();
		context.unbind(id);
		// check if this node has siblings, if not check parents and so 
		// on
		// refresh the first node without siblings or the root, 
		// whichever comes first
		RegistryTreeNode root = (RegistryTreeNode)model.getRoot();
		RegistryTreeNode nodeToRefresh = (RegistryTreeNode)node.getParent();
		while(nodeToRefresh.getChildCount()==1 && nodeToRefresh!=root) {
		    nodeToRefresh = (RegistryTreeNode)nodeToRefresh.getParent();
		}
		// refresh the root of the tree
		System.out.println("Refreshing node "+nodeToRefresh);
		nodeToRefresh.refresh();
		// redo the way the tree is rendered since its structure changed
		this.setSelectionPath(this.getPathForRow(0));
		model.nodeStructureChanged(nodeToRefresh);
	    } catch (InvalidContextException exception) {
		System.err.println("Got invalid context exception, could not remove info");
		exception.printStackTrace();
	    } catch (NotFoundException exception) {
		System.err.println("info does not exist in context "+context);
		exception.printStackTrace();
	    }
	}
	// start the drag
	System.out.println("Starting drag...");
	dragSource.startDrag(dragGestureEvent,
			     DragSource.DefaultCopyDrop,
			     draggableNode,
			     dragSourceListener);
    }

    // methods to support drop target interface
    public void dragEnter (DropTargetDragEvent dropTargetDragEvent) { }

    public void dragExit (DropTargetEvent dropTargetEvent) { }
    
    public void dragOver (DropTargetDragEvent dropTargetDragEvent) { }

    public void dropActionChanged (DropTargetDragEvent dropTargetDragEvent) { }

    public synchronized void drop (DropTargetDropEvent dropTargetDropEvent) {
	try {
	    // get the transferable object
	    Transferable tr = dropTargetDropEvent.getTransferable();
	    // accept it only if the transferable object has a flavor we recognise
	    if (tr.isDataFlavorSupported(DraggableRegistryTreeNode.STRING_FLAVOR)) {
		dropTargetDropEvent.acceptDrop(DnDConstants.ACTION_COPY_OR_MOVE);
		// get the string being transferred
		// assume this is a description of a component
		String xmlInstanceDescription = 
		    (String) tr.getTransferData(DraggableRegistryTreeNode.STRING_FLAVOR);
		// get the target directory, determined by the location of the drop
		Point dropOrigin = dropTargetDropEvent.getLocation();
		// check which subcontext the drop happened over
		TreePath path = this.getPathForLocation(dropOrigin.x,dropOrigin.y);
		RegistryTreeNode node = 
		    (RegistryTreeNode)path.getLastPathComponent();
		RegistryNode contextNode = 
		    (RegistryNode) node.getUserObject();
		// if this is not a context, reject the drop since we can't drop over
		// information
		if (contextNode.isContext() == 0) {
		    System.out.println("Drops only allowed over a context icon");
		    dropTargetDropEvent.rejectDrop();
		    getToolkit().beep();
		    return;
		}
		// convert this into a context object
		RegistryContext targetContext = 
		    (RegistryContext) contextNode;
		// convert the xml descrption into an info object
		InstanceInfo instanceInfo = 
		    InstanceXMLUtils.createInstanceInfoFromXML(xmlInstanceDescription);
		// if the conversion failed, reject the drop
		if (instanceInfo == null) {
		    System.err.println("Dropped information was not parsable XML instance information");
		    System.out.println("Dropped info was "+instanceInfo);
		    dropTargetDropEvent.rejectDrop();
		    getToolkit().beep();
		    return;
		}
		// add the new information to the target context
		// bring up a dialog to get the name under which the info 
		// should be bound
		Info [] information = instanceInfo.getInfo("instance-info");
		System.out.println("Showing input dialog...");
		String instanceName = JOptionPane.showInputDialog(this, 
								   "Enter binding name: ");
		if (instanceName == null) {
		    return;
		}
		System.out.println("Creating a instance with name "+
				   instanceName+" in context "+targetContext);
		try {
		    targetContext.bind(instanceInfo,instanceName);
		} catch (InvalidContextException exception) {
		    System.err.println("Could not add instance in this context");
		    dropTargetDropEvent.rejectDrop();
		    getToolkit().beep();
		    return;
		} 
		// refresh the node where the new instance was dropped
		node.refresh();
		// redo the way the node structure is displayed
		model.nodeStructureChanged(node);
	    }
	} catch (IOException exception) {
	    System.err.println("IO Exception "+exception);
	    dropTargetDropEvent.rejectDrop();
	    getToolkit().beep();
	} catch (UnsupportedFlavorException exception) {
	    System.err.println("Unsupported flavor exception "+exception);
	    dropTargetDropEvent.rejectDrop();
	    getToolkit().beep();
	}
    }

    // methods for implementing the mouse listener interface
    public void mouseClicked(MouseEvent event) { 
	// check if this is a double click
	if (event.getClickCount() == 2) {
	    // get the node over which the click occurred
	    Point clickOrigin = event.getPoint();
	    TreePath path = this.getPathForLocation(clickOrigin.x,clickOrigin.y);
	    RegistryTreeNode node = 
		(RegistryTreeNode) path.getLastPathComponent();
	    // get the information directory node which is in here
	    RegistryNode infoNode = 
		(RegistryNode) node.getUserObject();
	    // check if this corresponds to information
	    if (infoNode.isInstance() == 1) {
		// make this an information object
		InstanceInfo info = (InstanceInfo) infoNode;
		// open a window to display this in a pretty format
		Cursor currentCursor = getCursor();
		setCursor(new Cursor(Cursor.WAIT_CURSOR));
		displayInfo(info);
		setCursor(currentCursor);
	    }
	}
    }

    public void mouseEntered(MouseEvent event) { }

    public void mouseExited(MouseEvent event) { }

    public void mousePressed(MouseEvent event) { }

    public void mouseReleased(MouseEvent event) { }

    /**
     * This class is a default drag source listener
     */
    public static class TreeDragSourceListener implements DragSourceListener {
	public void dragDropEnd(DragSourceDropEvent dragSourceDropEvent) { }
	public void dragEnter(DragSourceDragEvent dragSourceDragEvent) { }
	public void dragExit(DragSourceEvent dragSourceEvent) { }
	public void dragOver(DragSourceDragEvent dragSourceDragEvent) { }
	public void dropActionChanged(DragSourceDragEvent dragSourceDragEvent) { }
    }
}
