
% Matlab file that runs a CCAT Demo.
%
% Please setenv JAVA_HOME to jdk1.1 before starting
% Matlab. Also, the classpath.txt file for Matlab 
% should contain the paths for the Java classes
% that we want to use for CCAT. 
%
% The components invovled in this LSA application
% Demo are:
%
% TimeGen: This generates a time step for the 
% discretizer component.
%
% Discretizer: Component that does a discretization of 
% a PDE, given a single double corresponding to a 
% simulated time value. This component produces a 
% a sparse linear system object which gets sent to
% some solver system.
%
% Splib: This component uses a library of
%        preconditioned iterative solvers to
%        solve a linear system.
%
% Assemble: Assembles and visualizes the solution.
%
% These components are composed in the following
% manner:
% 
%         TimeGen -> Discretizer
%
%         Discretizer -> Splib
%
%         Splib -> Assemble
%
%         Assemble -> TimeGen (feedback)
%
% -- Last modified by Madhuri on 1 March, 2000.



% Turn on the java option so that 
% we can use Java classes in Matlab.
java on


% Create an environment in which we can use CCAT.
compositionTool


% Create component models for the components 
% TimeGen, Discretizer, BasicInfo, Splib, and Assemble.
timegen = createComponent ('TimeGen')
discretizer = createComponent ('Discretizer')
splib = createComponent ('Splib')
assemble = createComponent ('Assemble')


% specify the machines where we want to create 
% instances of the components.
setMachineName (timegen, 'baldy.extreme.indiana.edu')
setMachineName (discretizer, 'baldy.extreme.indiana.edu')
setMachineName (assemble, 'baldy.extreme.indiana.edu')
setMachineName (splib, 'baldy.extreme.indiana.edu')


% specify the creation mechanisms that are
% to be used while creating the components.
setCreationMechanism (timegen, 'gram')
setCreationMechanism (discretizer, 'gram')
setCreationMechanism (assemble, 'gram')
setCreationMechanism (splib, 'gram')


% Create live instances of the components.
createInstance (timegen)
createInstance (discretizer)
createInstance (splib)
createInstance (assemble)


% Pause for a few seconds. This is done so
% that the instance creation is completed before
% we attempt to make and connections.
pause(10)


% Connects input and output ports in the following 
% manner:
%         TimeGen -> Discretizer
%
%         Discretizer -> splib
%
%         Splib -> Assemble
%
%         Assemble -> TimeGen (feedback)
connectPorts (timegen, 'outputTimeStep', discretizer, 'inputTimeStep')
connectPorts (discretizer, 'outputGeometry', assemble, 'inputGeometry')
connectPorts (discretizer, 'outputSLS', splib, 'inputSLS')
connectPorts (splib, 'outputSV', assemble, 'inputSV')
connectPorts (assemble, 'outputRunItAgain', timegen, 'inputRunItAgain')


% Set the parameters of the TimeGen component to
% their default values.
setParams (timegen)


% Execute the TimeGen component. This will automotically
% execute all the other components.
execute (timegen)
