/**
 * This is the composition tool's view of a port
 * @author Nirmal (nmukhi)
 * @see ccacore.Port
 */

package python;

import util.Info;
import exceptions.ConnectionException;
import exceptions.DisconnectionException;
import ccacore.PortInfo;
import ccacore.ComponentID;
import connection.ConnectionService;
import java.util.Vector;
import java.util.Enumeration;

public class PortModel {
    private ConnectionService connectionService;
    private String name;
    private String type;
    private boolean input;
    private ComponentModel ownerComponent;
    private ComponentID ownerComponentID;
    private String description;
    // Note: connections go _from_ the output _to_ the input
    private Vector dummyConnectionsFromHere; // list of PortModel objects
    private Vector dummyConnectionsToHere; // list of PortModel objects
    private Vector realConnectionsFromHere; // list of PortModel objects
    private Vector realConnectionsToHere; // list of PortModel objects
    // things required for current connection/disconnection
    private PortModel currentConnectionOtherEnd;
    private PortModel currentDisconnectionOtherEnd;
    public static double DISCONNECT_TIMEOUT = 5;
    /**
     * Create a port model with no associated information
     * and empty lists of connections
     */
    public PortModel(ConnectionService connectionService) {
	dummyConnectionsFromHere = new Vector();
	dummyConnectionsToHere = new Vector();
	realConnectionsFromHere = new Vector();
	realConnectionsToHere = new Vector();
	this.connectionService = connectionService;
    }

    /**
     * Create a port from its description in an info object
     */
    public PortModel(ComponentModel owner, Info description, 
		     ConnectionService connectionService) {
	this(connectionService);
	// get the required information and initialise everything
	this.ownerComponent = owner;
	this.ownerComponentID = owner.getComponentID();
	// get stuff from the component info
	// this is assumed to contain below "port"
	// from the DTD
	this.name = description.getValue("port-name")[0];
	this.type = description.getValue("my-port-type")[0];
	this.input = false;
	String direction = description.getValue("port-dir")[0];
	if (direction.equals("input"))
		this.input = true;
    }

    /**
     * Create a port from its corresponding PortInfo object
     * and an indication of whether it is input or not.<br>
     */
    public PortModel(ComponentModel owner, PortInfo portInfo, 
		     ConnectionService connectionService, boolean input) {
	this(connectionService);
	// get the required information from the port info
	// object 
	this.type = portInfo.getType();
	this.name = portInfo.getName();
	// remember whether this is input or not
	this.input = input;
	// set the owner and id
	this.ownerComponent = owner;
	this.ownerComponentID = owner.getComponentID();
    }

    /**
     * Make a connection from this port to another port.
     */
    public synchronized boolean connectPort(PortModel otherEnd) {
	// check if this is a compatible port
	if (!this.isCompatibleWith(otherEnd)) {
	    return false;
	}
	// ok we can attempt a connection
	currentConnectionOtherEnd = otherEnd;
	// check if this is a dummy port or not
	if (this.isReal() && currentConnectionOtherEnd.isReal()) {
	    System.out.println("Making real connection...");
	    // alert port view of impending connection
	    // use connection service to actually connect the ports
	    // if this is a real connection
	    // do that in a separate thread
	    final String otherEndName = currentConnectionOtherEnd.getName();
	    final ComponentID otherEndParent = currentConnectionOtherEnd.getOwnerID();
	    // attempt the connection
	    try {
		// figure out which the uses port is
		if (input) {
		    // this is the provides port
		    connectionService.connect(otherEndParent,otherEndName,
					      ownerComponentID,name);
		} else {
		    // this is the uses port
		    connectionService.connect(ownerComponentID,name,
					      otherEndParent,otherEndName);
		}
	    } catch (ConnectionException exception) {
		System.err.println("Got exception "+exception+" while attempting connection");
		exception.printStackTrace();
	    }
	} else {
	    System.out.println("Making dummy connection...");
	    // alert port view of impending connection
	    // add to list of dummy connections from/to here
	    // depending on whether this is an input or output port
	    if (input) {
		// this is the provides port
		dummyConnectionsToHere.addElement (currentConnectionOtherEnd);
		currentConnectionOtherEnd.dummyConnectionsFromHere.addElement (this);
	    } else {
		// this is the uses port
		dummyConnectionsFromHere.addElement (currentConnectionOtherEnd);
		currentConnectionOtherEnd.dummyConnectionsToHere.addElement (this);
	    }
	    // no current connection now
	    currentConnectionOtherEnd = null;
	    System.out.println("Connection completed");
	}
	// connection process initiated successfullly, return true
	return true;
    }
    
    /**
     * Update existing dummy connections
     */
    public void updateDummyConnections() {
	// if this is an input port, just update the dummy connections
	// to here (since there can't be any from here)
	if (input)
	    updateDummyConnectionsToHere();
	else
	    updateDummyConnectionsFromHere();
    }

    /**
     * Update existing dummy connections from here and try and make 
     * real connections in their place
     */
    public void updateDummyConnectionsFromHere() {
      // go through the list of connections from here
      Enumeration connectionsToCheck = dummyConnectionsFromHere.elements();
      while(connectionsToCheck.hasMoreElements()) {
	PortModel otherEnd = (PortModel) connectionsToCheck.nextElement();
	System.out.println("Attempting to make dummy connection from "+getName()+" to "+otherEnd.getName()+" real");
	// check the port model at the other end to see if it
	// is now real
	// if real, attempt a real connection
	if (otherEnd.isReal()) {
	  boolean successful = connectPort(otherEnd);
	  // remove from list of dummy connections
	  dummyConnectionsFromHere.remove(otherEnd);
	  otherEnd.dummyConnectionsToHere.remove(this);
	  if (!successful) {
	    System.out.println("Warning >>> dummy connection from "+getName()+" to "+otherEnd.getName()+" was incompatible!!!");
	  }
	}
      }
    }

    public void updateDummyConnectionsToHere() {
      // go through the list of connections from here
      Enumeration connectionsToCheck = dummyConnectionsToHere.elements();
      while(connectionsToCheck.hasMoreElements()) {
	PortModel otherEnd = (PortModel) connectionsToCheck.nextElement();
	System.out.println("Attempting to make dummy connection from "+getName()+" to "+otherEnd.getName()+" real");
	// check the port model at the other end to see if it
	// is now real
	    // if real, attempt a real connection
	if (otherEnd.isReal()) {
	  boolean successful = connectPort(otherEnd);
	  dummyConnectionsToHere.remove(otherEnd);
	  otherEnd.dummyConnectionsFromHere.remove(this);
	  // if the connection was not successful, print a warning!
	  if (!successful)
	    System.out.println("Warning >>> dummy connection from "+getName()+" to "+otherEnd.getName()+" was incompatible!!??!");
	}
      }
    }

    // removes all connections from and to this port, whether dummy or real
    public void removeAllConnections() {
	if (input) {
	    // can only have connections _to_ here
	    Enumeration dummyConnectionsList = dummyConnectionsToHere.elements();
	    while(dummyConnectionsList.hasMoreElements()) {
		PortModel otherEnd = (PortModel) dummyConnectionsList.nextElement();
		disconnectPort(otherEnd,false);
	    }
	    Enumeration realConnectionsList = realConnectionsToHere.elements();
	    while(realConnectionsList.hasMoreElements()) {
		PortModel otherEnd = (PortModel) realConnectionsList.nextElement();
		disconnectPort(otherEnd,false);
	    }
	} else {
	    // can only have connections _from_ here
	    Enumeration dummyConnectionsList = dummyConnectionsFromHere.elements();
	    while(dummyConnectionsList.hasMoreElements()) {
		PortModel otherEnd = (PortModel) dummyConnectionsList.nextElement();
		disconnectPort(otherEnd,false);
	    }
	    Enumeration realConnectionsList = realConnectionsFromHere.elements();
	    while(realConnectionsList.hasMoreElements()) {
		PortModel otherEnd = (PortModel) realConnectionsList.nextElement();
		disconnectPort(otherEnd,false);
	    }
	}
    }

    /**
     * Disconnect the specified port from this one
     */
    public synchronized boolean disconnectPort(PortModel otherEnd,boolean useSeparateThread) {
	// check if this is connected to the other end or not
	// skip that for now
	
	currentDisconnectionOtherEnd = otherEnd;
	if (this.isReal() && otherEnd.isReal()) {
	    // if this was a real connection, use the connection
	    // service to disconnect
	    // do that in a separate thread if reqd
	    final String otherEndName = currentDisconnectionOtherEnd.getName();
	    final ComponentID otherEndParent = currentDisconnectionOtherEnd.getOwnerID();
	    if (useSeparateThread) {
	      // attempt the disconnection
	      new Thread() {
		public void run() {
		  try {
		    // figure out which the uses port is
		    if (input) {
		      // this is the provides port
		      connectionService.disconnect(otherEndParent,otherEndName,
						   ownerComponentID,name,DISCONNECT_TIMEOUT);
		    } else {
		      // this is the uses port
		      connectionService.disconnect(ownerComponentID,name,
						   otherEndParent,otherEndName,DISCONNECT_TIMEOUT);
		    }
		  } catch (DisconnectionException exception) {
		    System.err.println("Got exception "+exception+" while attempting connection");
		    exception.printStackTrace();
		  }
		}
	      }.start();
	    } else {
	        try {
		    // figure out which the uses port is
		    if (input) {
		      // this is the provides port
		      connectionService.disconnect(otherEndParent,otherEndName,
						   ownerComponentID,name,DISCONNECT_TIMEOUT);
		    } else {
		      // this is the uses port
		      connectionService.disconnect(ownerComponentID,name,
						   otherEndParent,otherEndName,DISCONNECT_TIMEOUT);
		    }
		  } catch (DisconnectionException exception) {
		    System.err.println("Got exception "+exception+" while attempting connection");
		    exception.printStackTrace();
		  }
	    }
	} else {
	    // not a real connection, just remove from dummy lists
	    // on either end
	    if (input) {
		this.dummyConnectionsToHere.remove(otherEnd);
		otherEnd.dummyConnectionsFromHere.remove(this);
	    } else {
		this.dummyConnectionsFromHere.remove(otherEnd);
		otherEnd.dummyConnectionsToHere.remove(this);
	    }
	    // no current disconnection now
	    currentDisconnectionOtherEnd = null;
	}
	// return true since disconnection process was initiated successfully
	return true;
    }


    /**
     * Get the owner's component id
     */
    public ComponentID getOwnerID() {
	return  ownerComponentID;
    }

    /**
     * Set the owner's component id
     */
    public void setOwnerID(ComponentID id) {
	ownerComponentID = id;
    }


    /**
     * Get the owner component
     */
    public ComponentModel getOwner() {
	return  ownerComponent;
    }

    /**
     * Get this port's name
     */
    public String getName() {
	return name;
    }

    /**
     * Check if this port is an input port or not
     */
    public boolean isInput() {
	// return true if input, false otherwise
	return input;
    }

    /**
     * Check if this port is "real" or not
     * It is real if it's parent ID is non null
     */
    public boolean isReal() {
	// check if parent is real
	return ownerComponent.isReal();
    }

    /**
     * Get all the dummy connections from here
     */
    public Enumeration getDummyConnections() {
	return dummyConnectionsFromHere.elements();
    }

    /**
     * Get all the real connections from here
     */
    public Enumeration getRealConnections() {
	return realConnectionsFromHere.elements();
    }

    /**
     * Check whether the specified port can be connected to this one
     */
    public boolean isCompatibleWith(PortModel otherPort) {
	// they are compatible if they have different directions
	// and have the same type
	return ((this.input && !otherPort.isInput() ||
	       !this.input && otherPort.isInput()) &&
		this.hasSameTypeAs(otherPort));
    }

    /**
     * Get a description of this port if any
     */
    public String getDescription() {
	// return the port's description
	return description;
    }



    /**
     * A string representation of the information here
     */
    public String toString() {
	// put the port's name, type
	// direction and description in
	// a buffer and return that
	StringBuffer sb = new StringBuffer();
	sb.append("Port name: ");
	sb.append(name);
	sb.append("\nPort type: ");
	sb.append(type);
	if (input)
	    sb.append("\nInput port");
	else
	    sb.append("\nOutput port");
	sb.append("\nDescription: ");
	sb.append(description);
	sb.append("\n");
	return sb.toString();
    }

    /**
     * Check whether two port info objects have the same type
     */
    private boolean hasSameTypeAs(PortModel otherPort) {
	// check if the two types are the same
	return type.equals(otherPort.type);
    }

    /**
     * Called when a connection event occurs 
     */
  public synchronized void connectionEventOccurred(String otherPortName) {
    // make sure that this name matches the name of the 
    // current other end
    if (currentConnectionOtherEnd != null &&
	currentConnectionOtherEnd.getName().equals(otherPortName)) {
      // ok we got the event corresponding to the current connection
      // make changes to the conneciton lists on either end
      // add to list of real connections from/to here
      // depending on whether this is an input or output port
      // and ditto for the other end
      if (input) {
	// this is the provides port
	realConnectionsToHere.addElement (currentConnectionOtherEnd);
	currentConnectionOtherEnd.realConnectionsFromHere.addElement (this);
	System.out.println("Added a real connection to port "+getName());
      } else {
	// this is the uses port
	realConnectionsFromHere.addElement (currentConnectionOtherEnd);
	currentConnectionOtherEnd.realConnectionsToHere.addElement (this);
	System.out.println("Added a real connection to port "+getName());
      }
      // no current connection now
      currentConnectionOtherEnd = null;
    }
  }

    /**
     * Called when a disconnection event occurs
     */
    public synchronized void disconnectionEventOccurred(String otherPortName) {
	// make sure that this name matches the name of the 
	// current other end
	if (currentDisconnectionOtherEnd != null &&
	    currentDisconnectionOtherEnd.getName().equals(otherPortName)) {
	    // remove the connection
	    if (input) {
		this.realConnectionsToHere.remove(currentDisconnectionOtherEnd);
		currentDisconnectionOtherEnd.realConnectionsFromHere.remove(this);
	    } else {
		this.realConnectionsFromHere.remove(currentDisconnectionOtherEnd);
		currentDisconnectionOtherEnd.realConnectionsToHere.remove(this);
	    }
	    // disconnection completed
	    // no current disconnection now
	    currentDisconnectionOtherEnd = null;
	}
    }

  
    /** return the name of the port 
      */
    public String getPortName () {

        return this.name;
    }
}
