/**
 * This class is the controller for the entire working area
 * and handles all high level events dealing with connection
 * establishment and disconnection.<br>
 * This class also stores the state of the workspace and
 * directs drawing of connection lines.
 * @author Nirmal (nmukhi)
 * modified by Madhuri.
 */

package python;

import python.PortModel;
import python.ComponentModel;
import creation.CreationService;
import ccacore.ComponentID;
import ids.InformationDirectoryService;
import idl.eventPort.CCATEvent;
import registry.RegistryService;
import connection.ConnectionService;
import java.util.Vector;
import java.util.Enumeration;
import java.awt.Point;
import java.awt.Rectangle;
import javax.swing.ImageIcon;
import java.awt.Image;
import java.io.FileInputStream;
import java.io.IOException;
import ids.InformationInfo;
import registry.InstanceInfo;
import ids.InformationXMLUtils;
import registry.InstanceXMLUtils;
import java.io.File;
import java.io.FileNotFoundException;
import ids.InformationDirectoryContext;
import registry.RegistryContext;
import java.io.BufferedInputStream;
import python.SetParms;




public class WorkingAreaController  {
    // things required for maintaining correct visual state
    private Vector portModels; // list of PortModel objects
    private Vector componentModels; // list of ComponentModel objects
    private WorkingAreaView workingAreaView;
    private StatusBarView statusBar;
    private ComponentCommunicator communicator;
    // things required for drawing connection lines while they 
    // are being made
    private boolean connectionAlive;
    private PortModel startOfCurrentConnection;
    private static int FLASH_INTERVAL = 100; // interval for flashes while waiting for events
    // things required for framework interaction
    private CreationService creationService;
    private ConnectionService connectionService;
    private RegistryService registryService;
    private InformationDirectoryService idsService;
    private EventListener eventListener;
    private String idsURI;
    private String registryURI;
    
    /**
     * Create a working area given all the associated services 
     * from the framework
     */
    public WorkingAreaController(CreationService creationService,
				 ConnectionService connectionService,
				 RegistryService registryService,
				 InformationDirectoryService idsService,                                         EventListener eventListener,
                                 ComponentCommunicator componentCommunicator) {
	// set all the services
	this.registryService = registryService;
	this.idsService = idsService;
	this.connectionService = connectionService;
	this.creationService = creationService;
        // set the event listener
        this.eventListener = eventListener;
        eventListener.setGUIController(this);
        communicator = componentCommunicator;

	// initialize the lists
	portModels = new Vector();
	componentModels = new Vector();

        // set the default IDS location
        idsURI = new String ("http://lookout.extreme.indiana.edu:4242/ccat/XML");

        // set the default Registry location
        registryURI = new String ("http://lookout.extreme.indiana.edu:4242/registry_test");

    }

    // ------------------------------------------------------------
    // things relating to accessing stuff stored here and reqd by others
    // ------------------------------------------------------------

    /**
     * get the creation service
     */
    public CreationService getCreationService() {
	return creationService;
    }

    /**
     * get the connection service
     */
    public ConnectionService getConnectionService() {
	return connectionService;
    }

    /**
     * get the IDS service
     */
    public InformationDirectoryService getIDSService() {
	return idsService;
    }

    /**
     * get the registry service
     */
    public RegistryService getRegistryService() {
	return registryService;
    }

    // ------------------------------------------------------------
    // things relating to dragging in new stuff into the workspace
    // ------------------------------------------------------------
    
    /**
     * add a new component to the workspace
     */
    public void addComponent(ComponentModel component) {
	// add the component model to the list of component models
        // componentModels.add(component);
        componentModels.addElement(component);

        // get the list of input ports and add
        // their models to the workspace
        PortModel [] inputPorts = component.getInputPorts();
        for (int i=0; i<inputPorts.length; i++) {
            // portModels.add(inputPorts[i]);
            portModels.addElement(inputPorts[i]);
        }
        // get the list of output ports and add
        // their models to the workspace
        PortModel [] outputPorts = component.getOutputPorts();
        for(int i=0; i<outputPorts.length; i++) {
            // portModels.add(outputPorts[i]);
            portModels.addElement(outputPorts[i]);
        }


    }

    // ------------------------------------------------------------
    // things relating to connections
    // ------------------------------------------------------------




    public void quit() {
	// kill all components (and hence remove all connections)
	Enumeration components = componentModels.elements();
	while(components.hasMoreElements()) {
	    ComponentModel currentComponent = (ComponentModel) components.nextElement();
	    System.out.println("Killing "+currentComponent.getName());
	    currentComponent.killComponent();
	}
    }

  public synchronized void componentKilled(ComponentID componentID) {
    Enumeration components = componentModels.elements();
    while(components.hasMoreElements()) {
      // make the window for the killed component invisible
      ComponentModel model = (ComponentModel)components.nextElement();
      // if (componentID.equals(model.getComponentID()))
      //	view.setVisible(false);
    }
  }

    // ------------------------------------------------------------
    // things to do with receiving events form the event channel
    // ------------------------------------------------------------

    public void connectionEventReceived(CCATEvent event) {
	// figure out the name of the ports involved in this 
	// connection
	String connectionDetails = event.getDetails();
	// this contains names of the two ports involved
	// in the connection
	// the names are separated by colons
	int colonPosition = connectionDetails.indexOf(":");
	String firstPortName = connectionDetails.substring(0,colonPosition);
	String secondPortName = connectionDetails.substring(colonPosition+1);
	System.out.println("Received connection event for connection between "+
			   firstPortName+" and "+secondPortName);
	// send both these port names to the port model
	// involved in the current connection
	Enumeration ports = portModels.elements();
	while(ports.hasMoreElements()) {
	    PortModel currentPortModel = (PortModel)ports.nextElement();
	    if (currentPortModel.getName().equals(firstPortName))
	      currentPortModel.connectionEventOccurred(secondPortName);
	    else
	      currentPortModel.connectionEventOccurred(firstPortName);
	    // workingAreaView.updateConnectionLines();
	}
    }

    public void disconnectionEventReceived(CCATEvent event) {
	// figure out the name of the ports involved in this 
	// connection
	String disconnectionDetails = event.getDetails();
	// this contains names of the two ports involved
	// in the connection
	// the names are separated by colons
	int colonPosition = disconnectionDetails.indexOf(":");
	String firstPortName = disconnectionDetails.substring(0,colonPosition);
	String secondPortName = disconnectionDetails.substring(colonPosition+1);
	System.out.println("Received disconnection event for connection between "+
			   firstPortName+" and "+secondPortName);
	// send both these port names to all port models
	// since we have no idea who is involved in the current 
	// disconnection attempt
	Enumeration allPorts = portModels.elements();
	while(allPorts.hasMoreElements()) {
	    PortModel currentPortModel = (PortModel) allPorts.nextElement();
	    if (currentPortModel.getName().equals(firstPortName))
	      currentPortModel.disconnectionEventOccurred(secondPortName);
	    if (currentPortModel.getName().equals(secondPortName));
	     // workingAreaView.updateConnectionLines();
	}
    }

    public void creationEventReceived(CCATEvent event) {
	System.out.println("Creation event received in GUI controller");
	// check if this descrption  matches an event id for one 
	// of the components
	// get the event description
	String description = event.getSource();
	Enumeration components = componentModels.elements();
	while(components.hasMoreElements()) {
	    // get the corresponding component model
	    ComponentModel componentModel = (ComponentModel) components.nextElement();
	    // inform the component model that a creation was completed
	    componentModel.creationEventOccurred(description);
	    // workingAreaView.updateConnectionLines();
	}
    }

    public void idsAdditionEventReceived(CCATEvent event) {
	// TODO
	System.out.println("IDS addition event received in GUI controller");
    }

    public void idsModificationEventReceived(CCATEvent event) {
	// TODO
	System.out.println("IDS modification event received in GUI controller");
    }

    public void registryBindEventReceived(CCATEvent event) {
	// TODO
	System.out.println("Registry bind event received in GUI controller");
    }

    public void registryUnbindEventReceived(CCATEvent event) {
	// TODO
	System.out.println("Registry unbind event received in GUI controller");
    }


    // create a component
    public synchronized ComponentModel create(String xmlFileName) {

        ComponentModel model = null;
        String xmlDescription = null;
        FileInputStream inStream = null;
        byte [] contents = null;

        System.out.println ("Got xml file " + xmlFileName);

        try {

                        inStream = new FileInputStream(xmlFileName);
                        contents = new byte[10000];
                    } catch (FileNotFoundException exception) {
                        System.out.println ("Could not find the specified xml file" + xmlFileName+"in the current direcotry");
                         //exception.printStactTrace();
                   }

        try {
                        int size = inStream.read(contents);
                        xmlDescription = new String(contents,0,size);
                    } catch (Exception exception) {
                        System.err.println("Got exception while reading XML information from file");
                        exception.printStackTrace();
                    }

        System.out.println ("The xml file is " + xmlDescription );

        // convert the xml description into an information info object
        InformationInfo componentInfo =
                InformationXMLUtils.createComponentFromXML(xmlDescription);
        // if this is null, we could not parse the data, so ignore it
       if (componentInfo != null) {

                    // create the component model
                    System.out.println("Component info is "+componentInfo);
                    model = new ComponentModel(getCreationService(),
                                                              getConnectionService(),
                                                              communicator,
                                                              this,
                                                              componentInfo);

                                  
                    // add the new component to the workspace
                    addComponent (model);


       }
       else {

                    System.out.println ("Could not create info object from xml description");
       }

       return model;

   }


   public synchronized void setIDSLocation (String idsLocation) {

      idsURI = idsLocation;

   }


   public synchronized void setRegistryLocation (String registryLocation) {

      registryURI = registryLocation;

   }


   public void lookupIDS () {

      System.out.println ("Got context " + idsURI);
           

      InformationDirectoryContext idsContext = getIDSContext();

      IDSRequestHandler.handleIDSLookup (idsContext);


      }


    public synchronized ComponentModel createInstanceFromIDS (String componentName) {


      ComponentModel componentModel = null;

      InformationDirectoryContext idsContext = getIDSContext();

      componentModel = IDSRequestHandler.handleIDSCreate (componentName, idsContext, getCreationService(), getConnectionService(), this, communicator);


      return componentModel;

    }


   private InformationDirectoryContext getIDSContext() {


      InformationDirectoryContext idsContext = null;

      try {
             idsContext = idsService.getContext(idsURI);

      } catch (Exception exception) {

             System.out.println ("couldnt create information dir context for the default directory");
             exception.printStackTrace();
      }

      return idsContext;
   }



   public void lookupRegistry () {



      System.out.println ("Got registry context " + registryURI);
           

      RegistryContext registryContext = getRegistryContext();

      RegistryRequestHandler.handleRegistryLookup (registryContext);


   }


    public synchronized ComponentModel copyInstanceFromRegistry (String instanceName) {


      ComponentModel componentModel = null;

      RegistryContext registryContext = getRegistryContext();

      componentModel = RegistryRequestHandler.handleRegistryCopy (instanceName, registryContext, getCreationService(), getConnectionService(), this, communicator);

      return componentModel;

    }


    public void bindInstance (ComponentModel componentModel,
                              String bindingName) {


      RegistryContext registryContext = getRegistryContext();


      RegistryRequestHandler.bindInstance (componentModel, bindingName, registryContext);

    }



    public void displayBoundInstances() {
     

      RegistryContext registryContext = getRegistryContext();

      RegistryRequestHandler.displayBoundInstances (registryContext);

    }


    public void unbindInstance (String instanceName) {

     

      RegistryContext registryContext = getRegistryContext();

      RegistryRequestHandler.unBind (instanceName, registryContext);

    }


    private RegistryContext getRegistryContext() {

     

      RegistryContext registryContext = null;

      try {
             registryContext = registryService.getContext (registryURI);

      } catch (Exception exception) {

             System.out.println ("couldnt create registry dir context for the default directory");
             exception.printStackTrace();
      }
      
      return registryContext;

    }

   
    public void setParams (ComponentModel component, 
                           String paramName, 
                           String paramValue) {

      SetParms.setParameters (component, paramName, paramValue, communicator);

    }
                                

    public void displayDefaultParams (ComponentModel component) {

    SetParms.displayDefaultParams (component);

    }
}
