# Python file that runs a CCAT Demo.
# Please setnv CLASSPATH before starting jpython.
#
# The components invovled in this LSA application
# Demo are:
#
# TimeGen: This generates a time step for the
# discretizer component.
#
# Discretizer: Component that does a discretization of
# a PDE, given a single double corresponding to a
# simulated time value. This component produces a
# a sparse linear system object which gets sent to
# some solver system.
#
# Splib: This component uses a library of
#        preconditioned iterative solvers to
#        solve a linear system.
#
# Assemble: Assembles and visualizes the solution.
#
# These components are composed in the following
# manner:
# 
#         TimeGen -> Discretizer
#
#         Discretizer -> Splib
#                  \
#                   \
#                    \
#                     v
#         Splib -> Assemble
#
#         Assemble -> TimeGen (feedback)
#
# Last modified by Madhuri on 1 March, 2000.


# import the ccat API
import ccat



# Create component models for the components
# TimeGen, Discretizer, BasicInfo, Splib, and Assemble.
timegen = ccat.createComponent ('TimeGen')
discretizer = ccat.createComponent ('Discretizer')
splib = ccat.createComponent ('Splib')
assemble = ccat.createComponent ('Assemble')




# specify the machines where we want to create
# instances of the components.
ccat.setMachineName (timegen, 'baldy.extreme.indiana.edu')
ccat.setMachineName (discretizer, 'baldy.extreme.indiana.edu')
ccat.setMachineName (assemble, 'baldy.extreme.indiana.edu')
ccat.setMachineName (splib, 'baldy.extreme.indiana.edu')



# specify the creation mechanisms that are
# to be used while creating the components.
ccat.setCreationMechanism (timegen, 'gram')
ccat.setCreationMechanism (discretizer, 'gram')
ccat.setCreationMechanism (assemble, 'gram')
ccat.setCreationMechanism (splib, 'gram')



# Create live instances of the components.
ccat.createInstance (timegen)
ccat.createInstance (discretizer)
ccat.createInstance (assemble)
ccat.createInstance (splib)



# Connects input and output ports in the following
# manner:
#         TimeGen -> Discretizer
#
#         Discretizer -> Splib
#                  \
#                   \
#                    \
#                     v
#         Splib -> Assemble
#
#         Assemble -> TimeGen (feedback)
ccat.connectPorts (timegen, 'outputTimeStep', discretizer, 'inputTimeStep')
ccat.connectPorts (discretizer, 'outputGeometry', assemble, 'inputGeometry')
ccat.connectPorts (discretizer, 'outputSLS', splib, 'inputSLS')
ccat.connectPorts (splib, 'outputSV', assemble, 'inputSV')
ccat.connectPorts (assemble, 'outputRunItAgain', timegen, 'inputRunItAgain')



# Set the parameters of the TimeGen component to
# their default values.
ccat.setParams (timegen)


# Execute the TimeGen component. This will automotically
# execute all the other components.
ccat.execute (timegen)

