/**
 * This class is a handle to a collection on a DAV server.
 * It will be used through the <var>InformationDirectoryContext</var> 
 * interface which it implements to navigate the hierarchy
 * and add and remove objects from it.
 * @author Nirmal (nmukhi)
 * @see util.webdav.client.DAVConnection
 * @see util.webdav.client.DAVUtilityCalls
 */
package ids;

import util.webdav.client.DAVConnection;
import util.webdav.client.DAVConnectionFactory;
import util.webdav.client.DAVUtilityCalls;
import util.webdav.client.DAVServerResponse;
import util.webdav.client.DirectoryNode;
import util.webdav.client.DAVResourceNotFoundException;
import util.Info;
import registry.NotFoundException;
import registry.InvalidContextException;
import registry.BadContextNameException;
import java.util.Vector;

public class DAVContext extends Context implements InformationDirectoryContext {
    private DAVConnection connection;
    private String host;
    private int port;

    /**
     * the default host to which a connection is made
     */
    public static String DEFAULT_HOST = "lookout.extreme.indiana.edu";

    /**
     * the default port at which the connection is made
     */
    public static int DEFAULT_PORT = 4242;

    /**
     * the default initial resource on the server which the context
     * points to
     */
    public static String DEFAULT_RESOURCE = "/";

    /** 
     * Creates a context to a default server and port with a default
     * initial resource
     */
    public DAVContext() {
	// create an information directory context node with the default name
	super(DEFAULT_RESOURCE);
	// create the DAV connection to talk to the server
	connection = DAVConnectionFactory.createConnection(DEFAULT_HOST,DEFAULT_PORT);
	// set the host and port
	this.host = DEFAULT_HOST;
	this.port = DEFAULT_PORT;
    }

    /**
     * Creates a context to the specified server at a default port
     * and sets a default initial resource
     * @param host the name of the host where the DAV server is 
     * to be contacted
     */
    public DAVContext(String host) {
	// create an information directory context node with the default name
	super(DEFAULT_RESOURCE);
	// create the DAV connection to talk to the server
	connection = DAVConnectionFactory.createConnection(host,DEFAULT_PORT);
	// set the host and port
	this.host = host;
	this.port = DEFAULT_PORT;
    }

    /**
     * Creates a context to the specified server at the specified port
     * and sets a default initial resource
     * @param host the name of the host where the DAV server is running
     * @param port the port where the server can be contacted
     */
    public DAVContext(String host, int port) {
	// create an information directory context node with the default name
	super(DEFAULT_RESOURCE);
	// create the DAV connection to talk to the server
	connection = DAVConnectionFactory.createConnection(host,port);
	// set the host and port
	this.host = host;
	this.port = port;
    }

    /**
     * Creates a context to the specified server at a default port
     * and sets the specified initial resource
     * @param host the name of the host where the DAV server is running
     * @param initialResourceName the collection on the DAV server this 
     * context points to
     */
    public DAVContext(String host, String initialResourceName) {
	// create an information directory context node with the specified name
	super(initialResourceName);
	// create the DAV connection to talk to the server
	connection = DAVConnectionFactory.createConnection(host,DEFAULT_PORT);
	// set the host and port
	this.host = host;
	this.port = DEFAULT_PORT;
    }

    /**
     * Creates a context to the specified server at the specified port
     * and points to the specified initial resource
     * @param host the name of the host where the DAV server is running
     * @param port the port at which the DAV server can be contacted
     * @param initialResourceName the name of the collection on the DAV server
     * this context should point to
     */
    public DAVContext(String host, int port, String initialResourceName) {
	// create an information directory context node with the specified name
	super(initialResourceName);
	// create the DAV connection to talk to the server
	connection = DAVConnectionFactory.createConnection(host,port);
	this.host = host;
	this.port = port;
    }

    // implementation of the InformationDirectoryContext interface

    /**
     * Looks for the information bound under the specified name. 
     * @param name the name under which the information has
     * been bound
     * @return an <var>InformationInfo</var> object describing
     * the information bound under the specified name
     * @exception NotFoundException if there is no binding under the
     * provided <var>name</var> in this context
     * @exception InvalidContextException if the context being used
     * becomes invalid for some reason such as its deletion from the 
     * information directory hierarchy
     */
    public InformationInfo lookup(String instanceName) 
	throws NotFoundException, InvalidContextException {
	// look for the specified instance
	// create the full path name of the instance
	String fullInstanceName = getName()+instanceName;
	System.out.println("Looking for "+fullInstanceName);
	try {
	    DAVServerResponse response = connection.get(fullInstanceName);
	    // make sure the response was successful
	    if (!response.isSuccess()) {
		// network problem or some DAV error
		// handle this below
		throw new DAVResourceNotFoundException();
	    }
	    // the response's data is the file content
	    // which should be parsable XML describing the information
	    String resourceXMLInformation = new String(response.getData());
	    // parse the information and create an InformationInfo object
	    InformationInfo info = 
		InformationXMLUtils.createComponentFromXML(resourceXMLInformation);
	    // if this returns null, there was a parsing error of some
	    // kind, so throw the not found exception 
	    if (info == null)
		throw new NotFoundException("Could not parse information corresponding to "+
					    fullInstanceName);
	    // parsed information into info object successfully
	    return info;
	} catch (DAVResourceNotFoundException exception) {
	    checkContextValidity();
	    // context was valid but getting children threw an
	    // exception, which means the instance name was incorrect
	    throw new NotFoundException();
	}
    }

    /**
     * Gets a context for the specified subcontext name.
     * @param subcontext the name of a valid context accessible 
     * from this context.
     * @return a handle to the named context
     * @exception BadContextNameException if the specified 
     * context does not correspond to a context
     * @exception InvalidContextException if the context being used
     * becomes invalid for some reason such as its deletion from
     * the information directory
     */
    public InformationDirectoryContext getSubContext(String subcontext)
	throws BadContextNameException, InvalidContextException, NotFoundException {
	// construct the full path name of the subcontext
	String fullSubContextName = getName()+subcontext;
	// make sure that this is a valid resource
	try {
	    DAVServerResponse response = connection.get(fullSubContextName);
	    if (!response.isSuccess()) {
		// network problem or other DAV error
		throw new DAVResourceNotFoundException();
	    } else {
		// this is a valid resource
		// make sure it points to a context and not an instance
		if (DAVUtilityCalls.describesCollection(new String(response.getData()))) {
		    // ok, create the DAV connection to this subcontext
		    DAVContext newContext = new DAVContext(host,port,fullSubContextName);
		    System.out.println("Returning subcontext "+newContext);
		    return newContext;
		} else {
		    // not a collection
		    throw new BadContextNameException();
		}
	    }
	} catch (DAVResourceNotFoundException exception) {
	    // could not find the specified subcontext
	    // make sure the context is valid
	    checkContextValidity();
	    // context was valid, so the only possibility is that
	    // the subcontext name was wrong
	    throw new NotFoundException();
	}
    }
		
		    
    /**
     * Gets a list of nodes that are immediate children of this context
     * in the information directory hierarchy.
     * @return a list of nodes that are children of the current context
     * @exception InvalidContextException if the context being used
     * becomes invalid for some reason
     */
    public InformationDirectoryNode [] getChildren() throws InvalidContextException {
	try {
	    // get all the children of this context
	    String [] childNames = DAVUtilityCalls.getChildren(connection,getName());
	    // if this was not successful, something weird happened
	    // (network error or context has been deleted or some such thing)
	    if (childNames == null)
		throw new InvalidContextException();
	    // temporarily store the nodes to be returned in a vector
	    Vector nodes = new Vector();
	    // iterate through the child names
	    for (int i=0; i<childNames.length; i++) {
		String currentChildName = getName()+childNames[i];
		// get this resource from the server
		try {
		    DAVServerResponse response = connection.get(currentChildName);
		    // check whether this is a context or not
		    String fileContent = new String(response.getData());
		    if(DAVUtilityCalls.describesCollection(fileContent)) {
			// create an information directory context for this node
			DAVContext newContext = new DAVContext(host,port,currentChildName);
			// add this context to the nodes to be returned
			nodes.addElement (newContext);
		    } else {
			// create an information object for this node
			InformationInfo currentInfo = 
			    InformationXMLUtils.createComponentFromXML(fileContent);
			// if this returns null, there was a parsing error of some
			// kind, so skip the instance
			if (currentInfo != null)
			    nodes.addElement (currentInfo);
		    }
		} catch (DAVResourceNotFoundException exception) {
		    // the resource got removed just before I asked for it or
		    // there was a network error or some other DAV error
		    // make sure the ocntext is ok
		    checkContextValidity();
		    // skip this child
		}
	    }
	    // now copy the things from the vector to the array which will 
	    // be returned
	    InformationDirectoryNode [] directoryNodes = new 
		InformationDirectoryNode[nodes.size()];
	    for (int i=0; i<nodes.size(); i++) {
		directoryNodes[i] = (InformationDirectoryNode)nodes.elementAt(i);
	    }
	    return directoryNodes;
	} catch (DAVResourceNotFoundException exception) {
	    // could not get children for some reason
	    // the context no longer exists or there was some network
	    // or DAV error
	    throw new InvalidContextException();
	}
    }

    /**
     * Gets a list of all components that are below this context in the
     * information directory hierarchy 
     * @exception InvalidContextException thrown if this context becomes
     * invalid for some reason such as its deletion from the information directory
     */
    public InformationInfo [] getAllComponents() throws InvalidContextException {
	try {
	    // get all the leaves of this context
	    String [] childNames = DAVUtilityCalls.getAllChildren(connection,getName());
	    // if this was not successful, something weird happened
	    // (network error or context has been deleted or some such thing)
	    if (childNames == null)
		throw new InvalidContextException();
	    // temporarily store the instances to be returned in a vector
	    Vector resourceInfo = new Vector();
	    // iterate through the child names
	    for (int i=0; i<childNames.length; i++) {
		String currentChildName = childNames[i];
		// get this resource from the server
		try {
		    DAVServerResponse response = connection.get(currentChildName);
		    // if the request was unsuccessful, skip this child
		    if (response.isSuccess()) {
			// create an information object for this node
			String xmlInfo = new String(response.getData());
			InformationInfo currentComponent = 
			    InformationXMLUtils.createComponentFromXML(xmlInfo);
			// if this returns null, there was a parsing error of some
			// kind, so skip the instance
			if (currentComponent != null)
			    resourceInfo.addElement (currentComponent);
		    }
		} catch (DAVResourceNotFoundException exception) {
		    // the resource got removed just before I asked for it or
		    // there was a network error or some other DAV error
		    // make sure the context is ok
		    checkContextValidity();
		    // skip this child
		}
	    }
	    // now copy the things from the vector to the array which will 
	    // be returned
	    InformationInfo [] components = new InformationInfo[resourceInfo.size()];
	    for (int i=0; i<resourceInfo.size(); i++) {
		components[i] = (InformationInfo) resourceInfo.elementAt(i);
	    }
	    return components;
	} catch (DAVResourceNotFoundException exception) {
	    // could not get all children for some reason
	    // the context no longer exists or there was some network
	    // or DAV error
	    throw new InvalidContextException();
	}
    }


    /**
     * Binds the specified <var>info</var> into this context under the 
     * specified name.
     * @param name a name that can be used to lookup this
     * information. If there is already some information associated
     * with the specified name, it will be overwritten with the 
     * new information.
     * @param info the information to be bound.
     * @see InformationInfo
     */
    public void bind(Info info, String name) throws InvalidContextException {
	// create the full name 
	String fullResourceName = getName()+name;
	// create the XML file for this information object
	String resourceData = InformationXMLUtils.createXMLFromInfo(info,fullResourceName);
	// the resource gets put into the collection described by this name
	// upto the last slash
	int lastSlashPosition = fullResourceName.lastIndexOf('/');
	// note that there must be a slash because getName() when called on
	// a context must return a name which ends with a /
	String resourceCollection = fullResourceName.substring(0,lastSlashPosition+1);
	String resourceFileName = fullResourceName.substring(lastSlashPosition+1);
	System.out.println("Binding a new component into "+resourceCollection);
	DAVContext resourceContext = null;
	// check if there is already a collection with the required name
	// if not, create one
	try {
	    DAVServerResponse response = connection.get(resourceCollection);
	    System.out.println("Collection "+resourceCollection+" exists...");
	    // collection exists
	} catch (DAVResourceNotFoundException exception) {
	    // create a collection with the required name
	    // this may necessitate creating a whole set of collections
	    // e.g. if the collection name is something like /foo/bar/foobar/
	    // and this does not exist, we have to check if foo exists otherwise
	    // create it, likewise for /foo/bar/ and /foo/bar/foobar/
	    System.out.println("Collection does not exist, creating it...");
	    try {
		int endOfCollectionName = name.lastIndexOf('/');
		if (endOfCollectionName != -1) {
		    String collectionToCreate = name.substring(0,endOfCollectionName+1);
		    System.out.println("Creating collection "+
				       collectionToCreate+" starting from path "+getName());
		    DAVUtilityCalls.createCollectionPath(connection,collectionToCreate,getName());
		    System.out.println("Creation complete");
		}
	    } catch (DAVResourceNotFoundException contextException) {
		// something went wrong with the context or there was
		// a network or DAV exception
		throw new InvalidContextException();
	    }
	}
	// get the file name to be used for the instance
	System.out.println("Component file name will be "+resourceFileName+
			   ", putting this in the collection");
	// check if the file already exists, in which case we are overwriting 
	// information about a resource
	try {
	    DAVServerResponse response = connection.get(resourceCollection+resourceFileName);
	    // the resource is existing at the server end
	    System.out.println("Modifying existing data on server");
	    // generate component modified event here
	    // TODO
	} catch (DAVResourceNotFoundException exception) {
	    // the resource isn't there right now
	    System.out.println("Putting new data into server");
	    // generate component added event here
	    // TODO
	}
	// create a resource with this name and put the instance
	// data in it
	try {
	    connection.put(resourceCollection+resourceFileName,resourceData);
	    System.out.println("Put complete");
	} catch (DAVResourceNotFoundException exception) {
	    // can't see how this can happen unless we have a weird network
	    // or DAV error
	    throw new InvalidContextException();
	}
    }

    /**
     * Removes the binding of the specified name from the information directory.
     * @param name the name under which the information was bound
     * @exception NotFoundException thrown if nothing was bound with the
     * specified name in this context
     */
    public void unbind(String name) throws NotFoundException, InvalidContextException {
	// if any prefix of the provided name matches a suffix of the context name,
	// ignore that portion
	int lastSlashPos = name.lastIndexOf("/");
	String nameRelativeToContext = name.substring(lastSlashPos+1);
	String contextName = this.getName();
	while(lastSlashPos != -1 && lastSlashPos != 0) {
	    String possibleMatchingPrefix = name.substring(0,lastSlashPos+1);
	    System.out.println(possibleMatchingPrefix+" may match...");
	    if (contextName.endsWith(possibleMatchingPrefix)) {
		// the name relative to the context is whatever follows 
		// the prefix
		nameRelativeToContext = name.substring(lastSlashPos+1);
		System.out.println("Prefix matched, name relative to this "+
				   " context is "+nameRelativeToContext);
		break;
	    }
	    lastSlashPos = name.substring(0,lastSlashPos).lastIndexOf("/");
	}
	// construct the full resource name
	String fullResourceName = getName()+nameRelativeToContext;
	System.out.println("Attempting to unbind component "+fullResourceName);
	try {
	    DAVUtilityCalls.strongDelete(connection,fullResourceName);
	    // generate a component unbound event here
	    // TODO
	    return;
	} catch (DAVResourceNotFoundException exception) {
	    // make sure the context is valid
	    checkContextValidity();
	    // context is valid, so name was not bound
	    throw new NotFoundException();
	}
    }

    /**
     * @return a string representation of this context
     */
    public String toString() {
	StringBuffer sb = new StringBuffer();
	sb.append("Name: ");
	sb.append(getName());
	sb.append("\nHost: ");
	sb.append(host);
	sb.append("\nPort: ");
	sb.append(port);
	sb.append("\n");
	return sb.toString();
    }

    /**
     * Check if the context is valid.
     * This is done by requesting this context from the server
     * and making sure there was no error. 
     * @exception InvalidContextException thrown if the context
     * is not valid
     */
    private void checkContextValidity() throws InvalidContextException {
	// request the context from the server
	try {
	    DAVServerResponse response = connection.get(getName());
	    // got response, check if the request was successful
	    if (!response.isSuccess())
		throw new InvalidContextException();
	} catch (DAVResourceNotFoundException exception) {
	    throw new InvalidContextException();
	}
    }
}
