/**
 * This interface is a handle to the information hierarchy on an
 * information directory.
 * It provides methods to traverse the information hierarchy, bind
 * information into the directory, lookup previously bound information, etc.
 * Logically a directory context is a pointer to a specific directory
 * on a server that maintains the information.
 * Details on the API are available from 
 * <a href="http://www.extreme.indiana.edu/ccat/directory_service.html">here</a>.
 * @author Nirmal (nmukhi)
 * @see InformationNode
 */

package ids;

import registry.NotFoundException;
import registry.BadContextNameException;
import registry.InvalidContextException;
import util.Info;

public interface InformationDirectoryContext extends InformationDirectoryNode {

    /**
     * Looks for information bound in this directory that has the 
     * specified name. 
     * @param name the name of the information being looked up
     * @return an <var>InformationInfo</var> object describing
     * the information with the specified name.
     * @exception NotFoundException if there is no information with the
     * provided <var>name</var>.
     * @exception InvalidContextException if the context being used
     * becomes invalid for some reason such as its deletion from the 
     * information repository
     */
    public InformationInfo lookup(String name)
	throws NotFoundException, InvalidContextException;


    /**
     * Gets a context for the specified subcontext name.
     * @param subcontext the name of a valid context accessible 
     * from this context.
     * @return a handle to the named context
     * @exception BadContextNameException if the specified 
     * context does not correspond to a context
     * @exception InvalidContextException if the context being used
     * becomes invalid for some reason such as its deletion from
     * the registry
     * @exception NotFoundException thrown if the specified context
     * is not found in the directory
     */
    public InformationDirectoryContext getSubContext(String subcontext)
	throws BadContextNameException, InvalidContextException,
	       NotFoundException;

    /**
     * Gets a list of nodes that are immediate children of this context
     * in the information hierarchy.
     * @return a list of nodes that are children of the current context
     * @exception InvalidContextException if the context being used
     * becomes invalid for some reason
     */    
    public InformationDirectoryNode [] getChildren()
	throws InvalidContextException;

    /**
     * Gets a list of all information that is below this context in the
     * information hierarchy 
     * @exception InvalidContextException thrown if this context becomes
     * invalid for some reason such as its deletion from the information 
     * directory
     */
    public InformationInfo [] getAllComponents()
	throws InvalidContextException;
    
    /**
     * Binds the specified <var>info</var> into this context under the 
     * specified name.
     * @param name a name that can be used to lookup this
     * information. If there is already some information associated
     * with the specified name, it will be overwritten with the 
     * new information.
     * @param info the information to be bound.
     * @see InformationInfo
     */
    public void bind(Info info, String name) throws InvalidContextException;
    
    /**
     * Removes the binding of the specified name from the information directory.
     * @param name the name under which the information was bound
     * @exception NotFoundException thrown if nothing was bound with the
     * specified name in this context
     */
    public void unbind(String name) throws NotFoundException, InvalidContextException;
}
              
