/**
 * This class implements the <var>InformationDirectoryService</var> interface.
 * Given a named context, it chooses an appropriate context 
 * using  context service providers and instantiates the it.<br>
 * @author Nirmal (nmukhi)
 * @see InformationDirectoryContext
 */

package ids;

import util.Info;
import registry.NotFoundException;
import util.webdav.client.DAVConnection;
import util.webdav.client.DAVConnectionFactory;
import util.webdav.client.DAVServerResponse;

public class InformationDirectoryServiceImplementation 
    implements InformationDirectoryService {
    // constants
    /**
     * protocol name for local registry
     */
    public static final String LOCAL_PROTOCOL_NAME = "file";

    /**
     * protocol name for DAV registry
     */
    public static final String DAV_PROTOCOL_NAME = "http";

    /**
     * Gets a context for an information directory
     * @param context the address of the context, for example
     * http://lookout.extreme.indiana.edu:4242/ccat/resources<br>
     * @return a context object for the named context or null if 
     * the protocol was not recognized
     * @exception NotFoundException thrown if the specified context
     * did not correspond to an actual information directory
     * @exception ServerNotFoundException thrown if the server could
     * not be located
     */    
    public InformationDirectoryContext getContext(String context) 
	throws NotFoundException, ServerNotFoundException {
	// parse the context name

	// get the name of the protocol to be used
	// this is everything upto the first "://"
	int numCharsBeforeSeparator = context.indexOf("://");
	// if the protocol was not specified, throw a NotFoundException
	if (numCharsBeforeSeparator == -1)
	    throw new NotFoundException("Context name "+context+" had invalid format");
	String protocol = new String(context.substring(0,numCharsBeforeSeparator));

	// the context that will finally be returned
	InformationDirectoryContext directoryContext = null;
	// the initial resource name on the registry server
	String resourceName = "";

	// if protocol is local, the rest of the name just refers to a local
	// resource
	if (protocol.equals(InformationDirectoryServiceImplementation.LOCAL_PROTOCOL_NAME)) {
	    resourceName = new String(context.substring(numCharsBeforeSeparator+2));
	    // now we can create the local context
	    directoryContext = new LocalContext(resourceName);
	} else if (protocol.equals(InformationDirectoryServiceImplementation.DAV_PROTOCOL_NAME)) {
	    // the portion of the name upto the first / (if any) describes
	    // the server and port (if any)
	    int endOfServerDescription = context.indexOf('/',numCharsBeforeSeparator+3);
	    String serverDescription = "", host = "";
	    int port = -1;
	    if (endOfServerDescription == -1) {
		// no slash, add one
		context = context+"/";
		endOfServerDescription = context.length()-1;
	    }
	    // everything up to the end of the server description is 
	    // server, port stuff
	    // everything after that is the resource name
	    resourceName = context.substring(endOfServerDescription);
	    serverDescription = 
		new String(context.substring(numCharsBeforeSeparator+3,
					     endOfServerDescription));
	    // check if the server description includes a port number
	    // it does if it contains a :
	    int beginningOfPortNumber = serverDescription.lastIndexOf(':');
	    if (beginningOfPortNumber == -1) {
		// no port number
		// the server description just contains the host name
		host = serverDescription;
	    } else {
		// everything up to the port number is the host name
		host = serverDescription.substring(0,beginningOfPortNumber);
		String portNumber = serverDescription.substring(beginningOfPortNumber+1);
		port = new Integer(portNumber).intValue();
	    }
	    // check if the server is alive
	    // need to go one layer below to use core connection calls
	    // and request a resource which must exist
	    DAVConnection connection = util.webdav.client.
		DAVConnectionFactory.createConnection(host,port);
	    try {
		connection.get("/");
	    } catch (util.webdav.client.DAVResourceNotFoundException exception) {
		// could not get /
		// conclude that server wasn't alive
		throw new ServerNotFoundException();
	    }
	    // now we know the resource name, host and port so we can create
	    // a DAV context object
	    // if the port number is still -1, it was not specified so let 
	    // the context use the default
	    if (port == -1) {
		directoryContext = new DAVContext(host,resourceName);
	    } else {
		directoryContext = new DAVContext(host,port,resourceName);
	    }
	}
	return directoryContext;
    }

    /**
     * Main to test the information directory service code
     */
    public static void main(String [] args) {
	InformationDirectoryService ids = new InformationDirectoryServiceImplementation();
	// assume args[0] is a context string
	if (args.length != 1) {
	    System.out.println("Usage\n\tjava InformationDirectoryServiceImplementation <initial-context>\nwhere <initial-context> is a pointer to a DAV or local context such as http://lookout.extreme.indiana.edu/testdir or file://u/jsmith/testdir");
	    System.exit(1);
	}
	// create the context
	try {
	    InformationDirectoryContext context = ids.getContext(args[0]);
	    // lookup a component
	    System.out.println("\n\n------------------------------\nLooking up a component named fictional");
	    InformationInfo componentInfo = context.lookup("fictional");
	    System.out.println("Got component description\n"+componentInfo);
	    // bind this info object
	    System.out.println("\n\n------------------------------\nBinding the same component under the name test/foo/fictional-copy");
	    context.bind(componentInfo,"test/foo/fictional-copy");
	    // get all children of the context
	    System.out.println("\n\n------------------------------\nGetting all children of this context");
	    InformationInfo [] allChildren = context.getAllComponents();
	    for (int i=0; i<allChildren.length; i++) {
		System.out.println("Child #"+(i+1)+" has info "+allChildren[i]+"\n");
	    }
	    // get children of test
	    System.out.println("\n\n------------------------------\nGetting children of context test");
	    InformationDirectoryContext subcontext = context.getSubContext("test");
	    InformationDirectoryNode [] nodes = subcontext.getChildren();
	    System.out.println("test has "+nodes.length+" children");
	    for (int i=0; i<nodes.length; i++) {
		System.out.print("Child #"+(i+1)+" is ");
		if (nodes[i].isContext()==1) {
		    System.out.println("A context "+nodes[i]);
		} else {
		    System.out.println("A component with info "+nodes[i]);
		}
	    }
	    System.out.println("\n\n------------------------------\nUnbinding test/foo/fictional-copy");
	    context.unbind("test/foo/fictional-copy");
	} catch (Exception exception) {
	    System.out.println("Got exception "+exception+", quitting...");
	    System.exit(1);
	}
	System.out.println("Testing done");
    }
}
             
