/**
 * This class is an implementation of the 
 * <var>InformationInfo</var> interface. It allows access to
 * information stored in a hierarchical format indexed
 * by keys.
 * @author Nirmal (nmukhi)
 */

package ids;

import util.InfoImplementation;
import util.EnvObj;
import util.Info;
import java.util.Date;

public class InformationInfoImplementation extends InfoImplementation 
    implements InformationInfo {
    private String name;

    /** 
     * Creates information for an instance given the name
     * of the information node.
     * @param name the name for this node in the information 
     * directory
     */
    public InformationInfoImplementation(String name) {
	// default InfoImplementation constructor
	super();
	// set the name
	this.name = name;
    }

    /**
     * Creates information given the name and an 
     * <var>InfoImplementation</var> object containing the 
     * associated hierarchical information.
     */
    public InformationInfoImplementation(String name, InfoImplementation info) {
	// store the given info
	super(info);
	// set the name
	this.name = name;
    }

    // implementation of InformationDirectoryNode interface

    /**
     * Gets the name of this node
     * @return the name of the node.
     */
    public String getName() {
	return name;
    }

    /**
     * @return 1 since this node refers to information
     */
    public int isInformation() {
	return 1;
    }
    
    /**
     * @return 0 since this does not refer to a context
     */
    public int isContext() {
	return 0;
    }

    /**
     * This method uses the information describing this component
     * to get a creation environment suitable for instantiation of
     * that component on the specified machine using the specified
     * creation mechanism.<br>
     * Note that this will not be a complete creation environment 
     * and some things may have to be added before it can be used 
     * for creation.
     * @return a partial creation environment suitable for instantiation
     * of the component described in this information
     */
    public EnvObj getCreationEnvironment(String host, String creationMech) {
	// things required for the creation environment:
	// user-name, creation-time, creation-mech, host
	// things that are optional: 
	// LD_LIBRARY_PATH, PATH, WD, TMPDIR, SHELL, registry-host,
	// registry-port, ids-host, ids-port

	System.out.println("Getting creation env for machine "+host+" creation method "+creationMech);
	// create the env object to be returned
	EnvObj env = new EnvObj();
	// get the user name and add to the env
	String userName = System.getProperty("user.name");
	env.put("user-name",userName);
	// get the current time and add it as the creation time
	String creationTime = new Date().toString().replace(' ','_');
	env.put("creation-time",creationTime);
	// choose the very first installation as the default and
	// get the host name, default creation mechanism and creation
	// environment from there
	Info [] installationInfo = this.getInfo("component-info")[0].getInfo("installation");
	System.out.println("Got "+installationInfo.length+" installations");
	if (installationInfo.length == 0) {
	    // there was no info, so we can't add anything
	    // more to the environment
	    return null;
	}
	for (int i=0; i<installationInfo.length; i++) {
	    Info currentInstallation = installationInfo[i];
	    // check if the host matches the one desired
	    // get the host name
	    String machine = currentInstallation.getValue("host-name")[0];
	    if (machine.equals(host)) {
		// aha, this is the machine we want
		// go through installations on this machine
		Info [] creationInfo = currentInstallation.getInfo("creation-info");
		for (int j=0; j<creationInfo.length; j++) {
		    // get the creation mechanism and see if it matches
		    String creationMechanism = 
			creationInfo[j].getValue("creation-method")[0];
		    if(creationMechanism.equals(creationMech)) {
			// aha, this is the environment we want
			String executablePath = creationInfo[j].getValue("path")[0];
			env.put("host",host);
			env.put("creation-mech",creationMechanism);
			env.put("exec-path",executablePath);
			// get the creation environment
			Info creationEnv = creationInfo[j].getInfo("creation-env")[0];
			Info [] nameValuePairs = creationEnv.getInfo("name-value-pair");
			// the creation service expects this to include
			// at least the following:
			// creation-mech-path, exec-name, globus-run-script-path,
			// port-for-component-callback, instance-name
			// go through all these and add them to the env
			for(int k=0; k<nameValuePairs.length; k++) {
			    env.put(nameValuePairs[k].getValue("name")[0],
				    nameValuePairs[k].getValue("value")[0]);
			}
			// return the created environment
			return env;
		    }
		}
		// did not find the desired creation mechanism
		return null;
	    }
	} 
	// did not find the desired host 
	return null;
    }

    /**
     * Convert this information to a formatted string
     */
    public String toString() {
	StringBuffer sb = new StringBuffer();
	sb.append("Component name: ");
	sb.append(name);
	sb.append(super.toString());
	return sb.toString();
    }
}
