/**
 * This class contains a set of utility methods for accessing resource
 * information stored in XML format and for creating XML-ified versions
 * of resource information in the information directory
 * @author Nirmal (nmukhi)
 * @see InformationInfo
 */
package ids;

import java.io.ByteArrayOutputStream;
import java.util.Enumeration;
import com.ms.xml.om.Element;
import com.ms.xml.parser.ParseException;
import com.ms.xml.om.Document;
import com.ms.xml.util.Name;
import com.ms.xml.util.XMLOutputStream;
import util.GeneralXMLUtils;
import util.Info;
import util.InfoImplementation;

public class InformationXMLUtils extends GeneralXMLUtils {
    /**
     * This method takes the XML data describing a component and
     * creates the associated InformationInfo object from it. This
     * object will describe the data in the XML document bu hide the 
     * use of XML from the user.
     * @param xmlData the string containing the XML description of 
     * a component
     * @return the correpsonding information object describing the
     * component
     */
    public static InformationInfo createComponentFromXML(String xmlData) { 
	// create the Info object
	InfoImplementation info = GeneralXMLUtils.getInfoFromXML(xmlData);
	// if this returned null, the XML information could not be parsed, 
	// return null
	if (info == null)
	    return null;
	// get the name of the component
	Info [] componentInfo = info.getInfo("component-info");
	if (componentInfo.length== 0) {
	    // wasn't a valid component XML doc
	    return null;
	}
	// make sure the ports have valid types and that the parameter
	// block has paramters with valid parent ports and parameter names
	// TODO
	String name = componentInfo[0].getValue("component-name")[0];
	// create the information info object for this component
	return new InformationInfoImplementation(name,info);
    }
	
    /**
     * This method takes an Info object describing a component
     * and converts it into the corresponding XML document. 
     * @param componentInfo the component info object to 
     * be converted into XML
     * @return a string with the corresponding XML description of the 
     * component
     */
    public static String createXMLFromInfo(Info info, String componentName) { 
	// create an XML document
	Document document = new Document();
	document.setOutputStyle(XMLOutputStream.PRETTY);
	document.setVersion("1.0");
	document.setEncoding("us-ascii");
	// create the DTD node
	Element dtdNode = document.createElement(Element.DTD);
	dtdNode.setAttribute("NAME",Name.create("component-info"));
	dtdNode.setAttribute("URL",Name.create("http://www.cs.indiana.edu/~nmukhi/cat/test/component-info.dtd"));
	document.addChild(dtdNode,null);
	// create the root node
	Element rootNode = document.createElement(Element.ELEMENT,"component-info");
	document.addChild(rootNode,null);
	// begin adding children to the document, with the current parent element
	// as the root
	GeneralXMLUtils.createXMLLevelFromInfo(document,info.getInfo("component-info")[0],rootNode);
	// adjust the component name in the XML tree to match the provided component
	// name
	Enumeration rootChildren = rootNode.getElements();
	while(rootChildren.hasMoreElements()) {
	    Element current = (Element) rootChildren.nextElement();
	    if (current.getTagName().getName().equals("component-name")) {
		// get the only child of this element, which will be the
		// PCDATA corresponding to the value of the component name
		Element name = current.getChild(0);
		// make sure the name is the same as the name under 
		// which it was bound
		name.setText(componentName);
		break;
	    }
	}
	// write the XML tree into a file
	try {
	    ByteArrayOutputStream outStream = new ByteArrayOutputStream();
	    document.save(document.createOutputStream(outStream));
	    return outStream.toString();
	} catch (Exception exception) {
	    System.err.println("Could not create XML file, got exception "+exception);
	}
	return null;
    }
}
