/**
 * This class is a handle to a directory on the local filesystem.
 * It is used through the <var>InformationDirectoryContext</var> interface
 * to navigate the file hierarchy and add and remove resource 
 * information from it.
 * @author Nirmal (nmukhi)
 */

package ids;

import java.io.File;
import java.io.FileReader;
import java.io.StringWriter;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.FileNotFoundException;
import java.util.Vector;
import java.util.Enumeration;
import com.ms.xml.util.StringInputStream;
import registry.NotFoundException;
import registry.InvalidContextException;
import registry.BadContextNameException;
import util.Info;

public class LocalContext extends Context implements InformationDirectoryContext {
    private File contextDirectory;

    /**
     * The default directory where the context will point to
     */
    public static String DEFAULT_DIRECTORY = "resources";

    /**
     * Create a local context pointing to the default directory
     */
    public LocalContext() {
	this(DEFAULT_DIRECTORY);
    }

    /**
     * Create a local context pointing to the specified directory
     */
    public LocalContext(String directory) {
	super(directory);
	contextDirectory = new File(directory);
    }

    /**
     * Looks for information bound in this directory that has the 
     * specified name. 
     * @param name the name of the information being looked up
     * @return an <var>InformationInfo</var> object describing
     * the information with the specified name.
     * @exception NotFoundException if there is no information with the
     * provided <var>name</var>.
     * @exception InvalidContextException if the context being used
     * becomes invalid for some reason such as its deletion from the 
     * information repository
     */
    public InformationInfo lookup(String name)
	throws NotFoundException, InvalidContextException {
	// treat the name as a file in this context
	// and look for it
	File resource = new File(contextDirectory,name);
	if (!resource.exists() || !resource.isFile()) {
	    // make sure that the context is valid
	    checkContextValidity();
	    // context was valid, so the resource was not 
	    // there at all
	    throw new NotFoundException();
	}
	String fileData = readFile(resource);
	if (fileData == null) {
	    // something went wrong while reading this file
	    // throw an invalid context exception
	    throw new InvalidContextException();
	} else {
	    // assume the data is XML describing a component, and try
	    // parsing it to create an information object
	    InformationInfo resourceInfo = 
		InformationXMLUtils.createComponentFromXML(fileData);
	    // if this is null, there was a parsing error, so
	    // throw a not found exception
	    if (resourceInfo == null) {
		throw new NotFoundException("Could not parse information correspondsing to "+resource.getAbsolutePath());
	    }
	    // return the information 
	    return resourceInfo;
	}
    }

    /**
     * Gets a context for the specified subcontext name.
     * @param subcontext the name of a valid context accessible 
     * from this context.
     * @return a handle to the named context
     * @exception BadContextNameException if the specified 
     * context does not correspond to a context
     * @exception InvalidContextException if the context being used
     * becomes invalid for some reason such as its deletion from
     * the information directory
     */
    public InformationDirectoryContext getSubContext(String subcontext)
	throws BadContextNameException, InvalidContextException, NotFoundException {
	// get the file object for the subcontext
	File subContext = new File(contextDirectory,subcontext);
	// if this does not exist, throw not found exception
	// after checking that the context is valid
	if (!subContext.exists()) {
	    checkContextValidity();
	    throw new NotFoundException();
	}
	// if this does not correspond to a directory, throw 
	// BadContextName exception
	if (!subContext.isDirectory())
	    throw new BadContextNameException();
	// it exists, and is a directory!
	// create a local context for it and return
	return new LocalContext(subContext.getAbsolutePath());
    }

    /**
     * Gets a list of nodes that are immediate children of this context
     * in the information directory.
     * @return a list of nodes that are children of the current context
     * @exception InvalidContextException if the context being used
     * becomes invalid for some reason
     */
    public InformationDirectoryNode [] getChildren() throws InvalidContextException {
	try {
	    // get all children of this context
	    File [] children = contextDirectory.listFiles();
	    // if the call returned null, there was a problem 
	    // with the context
	    if (children == null)
		throw new InvalidContextException();
	    // create a vector of directory nodes to store the nodes in 
	    // the context
	    Vector nodes = new Vector();
	    // iterate through these
	    for (int i=0; i<children.length; i++) {
		File currentChild = children[i];
		if (currentChild.isDirectory()) {
		    // this is a directory, create a local 
		    // context for it
		    nodes.add(new LocalContext(currentChild.getAbsolutePath()));
		} else {
		    // this is a file, assume it describes a component and try
		    // parsing it
		    String resourceXMLData = readFile(currentChild);
		    // if the data was null, there was an error while reading the
		    // file, so just skip this child
		    if (resourceXMLData != null) {
			InformationInfo info = 
			    InformationXMLUtils.createComponentFromXML(resourceXMLData);
			// if the information was null, there was an error while
			// parsing the XML, so ignore this child otherwise add it
			// to our node list
			if (info != null)
			    nodes.add(info);
		    }
		}
	    }
	    // now copy the things from the vector to the array which will 
	    // be returned
	    InformationDirectoryNode [] directoryNodes = 
		new InformationDirectoryNode[nodes.size()];
	    for (int i=0; i<nodes.size(); i++) {
		directoryNodes[i] = (InformationDirectoryNode)nodes.elementAt(i);
	    }
	    return directoryNodes;
	} catch (SecurityException exception) {
	    // could not list files in the context
	    throw new InvalidContextException();
	}
    }
    
    /**
     * Gets a list of all information that is below this context in the
     * information hierarchy 
     * @exception InvalidContextException thrown if this context becomes
     * invalid for some reason such as its deletion from the information
     * directory
     */
    public InformationInfo [] getAllComponents()
	throws InvalidContextException {
	// do an iterative breadth first search of the directory
	// tree 
	Vector directoriesRemaining = new Vector();
	Vector resourceInfo = new Vector();
	// add this context to the list of directories to be searched
	directoriesRemaining.add(contextDirectory);
	// go through all the directories that remain to be searched
	while(directoriesRemaining.size()!=0) {
	    // get the first remaining directory
	    File currentDir = (File)directoriesRemaining.remove(0);
	    try {
		File [] currentChildren = currentDir.listFiles();
		// if the call returned null, there was a problem with the 
		// currentDir
		if (currentChildren == null)
		    throw new InvalidContextException("Could not get files under "+currentDir);
		// go through all the children
		for (int i=0; i<currentChildren.length; i++) {
		    File currentChild = currentChildren[i];
		    if (currentChild.isDirectory()) {
			// a subdirectory, add it to the directories remaining to 
			// be searched
			directoriesRemaining.add(currentChild);
		    } else {
			// a file, assume it describes a component and try
			// parsing it
			String componentXMLData = readFile(currentChild);
			// if this was null, there was an error while reading the 
			// file, skip this child
			if (componentXMLData!=null) {
			    InformationInfo info = 
				InformationXMLUtils.createComponentFromXML(componentXMLData);
			    // if this was null, ther was a parsing error, skip the file
			    // otherwise add it to our resource list
			    if (info != null)
				resourceInfo.add(info);
			}
		    }
		}
	    } catch (SecurityException exception) {
		// could not list files for some reason
		throw new InvalidContextException();
	    }
	}
	// now copy the things from the vector to the array which will 
	// be returned
	InformationInfo [] resources = new InformationInfo[resourceInfo.size()];
	for (int i=0; i<resourceInfo.size(); i++) {
	    resources[i] = (InformationInfo) resourceInfo.elementAt(i);
	}
	return resources;
    }

    /**
     * Binds the specified <var>info</var> into this context under the 
     * specified name.
     * @param name a name that can be used to lookup this
     * information. If there is already some information associated
     * with the specified name, it will be overwritten with the 
     * new information.
     * @param info the information to be bound.
     * @see InformationInfo
     */
    public void bind(Info info, String name) throws InvalidContextException {
	// create the full name 
	String fullResourceName = getName()+name;
	// create the XML string describing this resource
	String resourceData = InformationXMLUtils.createXMLFromInfo(info,fullResourceName);
	// make sure the context is valid before trying to create the
	// resource directory
	checkContextValidity();
	// the resource gets put into the collection described by this name
	// upto the last slash
	int lastSlashPosition = fullResourceName.lastIndexOf('/');
	// note that there must be a slash because getName() when called on
	// a context must return a name which ends with a /
	String resourceDirectory = fullResourceName.substring(0,lastSlashPosition+1);
	String resourceFileName = fullResourceName.substring(lastSlashPosition+1);
	System.out.println("Binding a new component into "+resourceDirectory);
	File resourceDir = new File(resourceDirectory);
	try {
	    // create the directory in which the information will live
	    // if it doesn't already exist
	    boolean dirCreated = resourceDir.mkdirs();
	    if (dirCreated)
		System.out.println("Created "+resourceDir);
	    else
		System.out.println(resourceDir+" exists");
	    // put the info file in the  directory with the specified data
	    // first create the file
	    File infoFile = new File(resourceDir,resourceFileName);
	    boolean fileCreated = infoFile.createNewFile();
	    if (fileCreated) {
		System.out.println("Created "+infoFile+" in "+resourceDir);
		// generate component added event here
		// TODO
	    } else {
		System.out.println(infoFile+" already exists in "+
				   resourceDir+", overwriting the existing data...");
		// generate a component modified event here
		// TODO
	    }
	    // now fill it with the data
	    FileOutputStream outStream = new FileOutputStream(infoFile);
	    StringInputStream inStream = new StringInputStream(resourceData);
	    int currentChar = inStream.read();
	    while (currentChar!=-1) {
		outStream.write(currentChar);
		currentChar = inStream.read();
	    }
	    // close the streams
	    inStream.close();
	    outStream.close();
	} catch (SecurityException exception) {
	    // could not create the resource directory or the resource file
	    throw new InvalidContextException();
	} catch (IOException exception) {
	    // could not write the resource file
	    throw new InvalidContextException();
	}
    }

    /**
     * Removes the binding of the specified name from the information directory.
     * @param name the name under which the information was bound
     * @exception NotFoundException thrown if nothing was bound with the
     * specified name in this context
     */
    public void unbind(String name) 
	throws NotFoundException, InvalidContextException {
	// if any prefix of the provided name matches a suffix of the context name,
	// ignore that portion
	int lastSlashPos = name.lastIndexOf("/");
	String nameRelativeToContext = name.substring(lastSlashPos+1);
	String contextName = this.getName();
	while(lastSlashPos != -1 && lastSlashPos != 0) {
	    String possibleMatchingPrefix = name.substring(0,lastSlashPos+1);
	    System.out.println(possibleMatchingPrefix+" may match...");
	    if (contextName.endsWith(possibleMatchingPrefix)) {
		// the name relative to the context is whatever follows 
		// the prefix
		nameRelativeToContext = name.substring(lastSlashPos+1);
		System.out.println("Prefix matched, name relative to this "+
				   " context is "+nameRelativeToContext);
		break;
	    }
	    lastSlashPos = name.substring(0,lastSlashPos).lastIndexOf("/");
	}
	// construct the full resource name
	String fullResourceName = getName()+nameRelativeToContext;
	System.out.println("Attempting to unbind component "+fullResourceName);
	try {
	    // remove this resource
	    // use strong deletion so empty parent directories will
	    // also be removed
	    String dirName = getName();
	    String fileName = fullResourceName;
	    System.out.println("Deleting  "+fileName);
	    // delete file
	    File fileToDelete = new File(fileName);
	    // make sure the file exists
	    if (!fileToDelete.exists() || !fileToDelete.isFile()) {
		// make sure the ocntext is valid
		checkContextValidity();
		throw new NotFoundException(nameRelativeToContext+" does not correspond to a valid resource in this context");
	    }
	    File parentDir = fileToDelete.getParentFile();
	    boolean deleted = fileToDelete.delete();
	    while(deleted && parentDir!=null) {
		// keep deleting parent directories while it is 
		// possible to do so
		fileToDelete = parentDir;
		parentDir = fileToDelete.getParentFile();
		deleted = fileToDelete.delete();
	    }
	    // generate a component removed event here
	    // TODO
	} catch (SecurityException exception) {
	    // deletion was not allowed somewhere
	    throw new InvalidContextException();
	}
    }

    /**
     * Check if this context is valid. This is done by making sure that
     * a directory with the context name exists
     */
    private void checkContextValidity() throws InvalidContextException {
	// check that the context directory exists, is a directory
	// and can be read form and written to
	if (!contextDirectory.exists() || !contextDirectory.canRead() ||
	    !contextDirectory.canWrite() || !contextDirectory.isDirectory())
	    throw new InvalidContextException(contextDirectory+" is not a valid context");
    }
    
    /**
     * Read information from the specified file and return it as 
     * a string
     * @param file the file to be read
     * @return the data in the file
     */
    private String readFile(File file) {
	// open a stream for reading
	FileReader reader = null;
	try {
	    reader = new FileReader(file);
	} catch (FileNotFoundException exception) {
	    // could not find file
	    return null;
	}
	// open a stream for writing
	StringWriter writer = new StringWriter();
	// read one character at a time till the whole
	// file has been read
	try {
	    // read a character
	    int currentCharacter = reader.read();
	    // write the character as long as we have not reached the end
	    while(currentCharacter != -1) {
		writer.write(currentCharacter);
		currentCharacter = reader.read();
	    }
	    // we've reached the end, return what we have so far
	    reader.close();
	    writer.close();
	    return writer.toString();
	} catch (IOException exception) {
	    // problem while reading, just return null
	    return null;
	}
    }

    /**
     * Creates a string representation of the context and returns it
     * @return a string representation of the context
     */
    public String toString() {
	return contextDirectory.toString();
    }
}
