/**
 * This is the internal implementation of a context node.
 * It stores the name of the context and implements 
 * the RegistryNode interface. It is the base class
 * for any registry context.
 * @author Nirmal (nmukhi)
 *
 */

package registry;

import ccacore.ComponentID;
import util.EnvObj;
import ids.InformationInfo;

public abstract class Context implements RegistryNode {
    private String name;

    /**
     * Create a context node with the specified name
     *
     */
    public Context(String name) {
	if (!name.endsWith("/"))
	    name+="/";
	this.name = name;
    }


    // implementation of the RegistryNode interface

    /**
     * Gets the name of this node
     * @returns the name of the node.
     */
    public String getName() {
	return name;
    }

    /**
     * Checks if this node corresponds to an instance
     * @returns 0 since this node is a context
     */
    public int isInstance() {
	return 0;
    }

    /**
     * Checks if this node corresponds to a context
     * @returns 1 since this node is a context
     */
    public int isContext() {
	return 1;
    }

    /**
     * Binds the specified <var>id</var> into this context under the 
     * specified name. The environment object contains additional 
     * information about the instance
     * @param id the handle to the component itself
     * @param instanceName a name that can be used to lookup this
     * instance. This need not be unique.
     * @param env the environment object used to create this 
     * instance
     * @param componentInfo the information object for the component
     * of which this is an instance
     * @see InstanceInfo
     * @see InstanceXMLUtils.createXMLFromInfo(ComponentID,String,EnvObj,Info)
     */
    public void bind(ComponentID id, String instanceName, EnvObj env,
		     InformationInfo componentInfo) 
	throws InvalidContextException {
	// create the XML file for this instance
	// make sure the instance name in the XML description is the FQN (i.e.
	// includes the context name in front)
	String instanceData = InstanceXMLUtils.createXMLFromInfo(id,
								 getName()+instanceName,
								 env,componentInfo);
	// bind this information under the given name
	bind(instanceData,instanceName,id);
    }

    /**
     * Binds the specified <var>instanceInfo</var> into this context
     * under the specified name. 
     * @param instanceInfo the <tt>InstanceInfo</tt> object describing 
     * the instance
     * @param instanceName a name that can be used to lookup this
     * instance. This need not be unique
     * @see InstanceInfo
     * @see InstanceXMLUtils#createXMLFromInfo(InstanceInfo,String)
     */
    public void bind(InstanceInfo instanceInfo, String instanceName) 
	throws InvalidContextException {
	// create the XML string describing this instance
	String fullInstanceName = getName()+instanceName;
	String instanceData = InstanceXMLUtils.createXMLFromInfo(instanceInfo, 
								 instanceName);
	// bind this information under the given name
	bind(instanceData,instanceName,instanceInfo.getComponentID());
    }

    /**
     * Binds the given stringified instance information into this
     * context under the given name
     */
    protected abstract void bind(String xmlInstanceDescription, String instanceName,
				 ComponentID id) throws InvalidContextException;
}
              
    
