/**
 * This class is a handle to a collection on a DAV server.
 * It will be used through the <var>RegistryContext</var> 
 * interface which it implements to navigate the hierarchy
 * and add and remove objects from it.
 * @author Nirmal (nmukhi)
 * @see util.webdav.client.DAVConnection
 * @see util.webdav.client.DAVUtilityCalls
 */
package registry;

import util.webdav.client.DAVConnection;
import util.webdav.client.DAVConnectionFactory;
import util.webdav.client.DAVUtilityCalls;
import util.webdav.client.DAVServerResponse;
import util.webdav.client.DirectoryNode;
import util.webdav.client.DAVResourceNotFoundException;
import ccacore.ComponentID;
import ccacore.ComponentIDregistry;
import nexusrmi.RemoteException;
import java.util.Vector;

public class DAVContext extends Context implements RegistryContext {
    private DAVConnection connection;
    private String host;
    private int port;

    /**
     * the default host to which a connection is made
     */
    public static String DEFAULT_HOST = "lookout.extreme.indiana.edu";

    /**
     * the default port at which the connection is made
     */
    public static int DEFAULT_PORT = 4242;

    /**
     * the default initial resource on the server which the context
     * points to
     */
    public static String DEFAULT_RESOURCE = "/";

    /** 
     * Creates a context to a default server and port with a default
     * initial resource
     */
    public DAVContext() {
	// create a registry context node with the default name
	super(DEFAULT_RESOURCE);
	// create the DAV connection to talk to the server
	connection = DAVConnectionFactory.createConnection(DEFAULT_HOST,DEFAULT_PORT);
	// set the host and port
	this.host = DEFAULT_HOST;
	this.port = DEFAULT_PORT;
    }

    /**
     * Creates a context to the specified server at a default port
     * and sets a default initial resource
     * @param host the name of the host where the DAV server is 
     * to be contacted
     */
    public DAVContext(String host) {
	// create a registry context node with the default name
	super(DEFAULT_RESOURCE);
	// create the DAV connection to talk to the server
	connection = DAVConnectionFactory.createConnection(host,DEFAULT_PORT);
	// set the host and port
	this.host = host;
	this.port = DEFAULT_PORT;
    }

    /**
     * Creates a context to the specified server at the specified port
     * and sets a default initial resource
     * @param host the name of the host where the DAV server is running
     * @param port the port where the server can be contacted
     */
    public DAVContext(String host, int port) {
	// create a registry context node with the default name
	super(DEFAULT_RESOURCE);
	// create the DAV connection to talk to the server
	connection = DAVConnectionFactory.createConnection(host,port);
	// set the host and port
	this.host = host;
	this.port = port;
    }

    /**
     * Creates a context to the specified server at a default port
     * and sets the specified initial resource
     * @param host the name of the host where the DAV server is running
     * @param initialResourceName the collection on the DAV server this 
     * context points to
     */
    public DAVContext(String host, String initialResourceName) {
	// create a registry context node with the specified name
	super(initialResourceName);
	// create the DAV connection to talk to the server
	connection = DAVConnectionFactory.createConnection(host,DEFAULT_PORT);
	// set the host and port
	this.host = host;
	this.port = DEFAULT_PORT;
    }

    /**
     * Creates a context to the specified server at the specified port
     * and points to the specified initial resource
     * @param host the name of the host where the DAV server is running
     * @param port the port at which the DAV server can be contacted
     * @param initialResourceName the name of the collection on the DAV server
     * this context should point to
     */
    public DAVContext(String host, int port, String initialResourceName) {
	// create a registry context node with the specified name
	super(initialResourceName);
	// set the host and port
	this.host = host;
	this.port = port;
	// create the DAV connection to talk to the server
	connection = DAVConnectionFactory.createConnection(host,port);
    }

    // implementation of the RegistryContext interface

    /**
     * Looks for instances that have the specified name. There may
     * be more than one since this is a user-provided instance name
     * supplied at the time of the <a href="#bind">bind</a> call and 
     * cannot be guaranteed unique.
     * @param instanceName the name of the instance
     * @return a list of <var>InstanceInfo</var> objects describing
     * the instances with the specified name.
     * @exception NotFoundException if there is no instance with the
     * provided <var>instanceName</var>.
     * @exception InvalidContextException if the context being used
     * becomes invalid for some reason such as its deletion from the 
     * registry
     */
    public InstanceInfo [] lookup(String instanceName) 
	throws NotFoundException, InvalidContextException {
	// look for the specified instance
	// create the full path name of the instance
	String fullInstanceName = getName()+instanceName;
	// make sure this has a trailing slash
	if (!fullInstanceName.endsWith("/"))
	    fullInstanceName += "/";
	System.out.println("Looking for "+fullInstanceName);
	try {
	    // get the children of the collection using the above name
	    // this should return all instances that have been registered
	    // under this name
	    String [] children = DAVUtilityCalls.getChildren(connection,
							     fullInstanceName);
	    if (children == null) {
		// there was a network error of some kind (most likely
		// a bad server name, port or something)
		// or there was an error in DAV (e.g. the instanceName
		// did not correspond to a collection but was a file 
		// instead)
		throw new InvalidContextException();
	    } else {
		// temporarily store instance information in a vector
		Vector instanceInfo = new Vector();
		// iterate through the children and get each file
		for (int i=0; i<children.length; i++) {
		    String currentChildName = fullInstanceName+children[i];
		    System.out.println("Got instance file name "+currentChildName+
				       ",attempting to get instance file...");
		    try {
			DAVServerResponse response = connection.get(currentChildName);
			// make sure the response was successful
			if (!response.isSuccess()) {
			    // network problem or some DAV error
			    // throw a resource not found exception
			    throw new DAVResourceNotFoundException();
			}
			// the response's data is the file content
			// which should be parsable XML describing the instance
			String instanceXMLInformation = new String(response.getData());
			// parse the information and create an InstanceInfo object
			InstanceInfo currentInstance = 
			    InstanceXMLUtils.createInstanceInfoFromXML(instanceXMLInformation);
			// if this returns null, there was a parsing error of some
			// kind, so skip the instance
			if (currentInstance != null)
			    instanceInfo.addElement (currentInstance);
		    } catch (DAVResourceNotFoundException exception) {
			// make sure that this context is still valid
			checkContextValidity();
			// weird, maybe the instance info file got deleted 
			// between the time its name came here and the time
			// I asked for it...just skip this child
		    }
		}
		// copy instance information from the vector
		// and return that
		InstanceInfo [] instances = new InstanceInfo[instanceInfo.size()];
		for (int i=0; i<instanceInfo.size(); i++) {
		    instances[i] = (InstanceInfo)instanceInfo.elementAt(i);
		}
		// return the information on the requested instances
		return instances;
	    }
	} catch (DAVResourceNotFoundException exception) {
	    checkContextValidity();
	    // context was valid but getting children threw an
	    // exception, which means the instance name was incorrect
	    throw new NotFoundException();
	}
    }

    /**
     * Gets a context for the specified subcontext name.
     * @param subcontext the name of a valid context accessible 
     * from this context.
     * @return a handle to the named context
     * @exception BadContextNameException if the specified 
     * context does not correspond to a context
     * @exception InvalidContextException if the context being used
     * becomes invalid for some reason such as its deletion from
     * the registry
     */
    public RegistryContext getSubContext(String subcontext)
	throws BadContextNameException, InvalidContextException, NotFoundException {
	// construct the full path name of the subcontext
	String fullSubContextName = getName()+subcontext;
	// make sure that this is a valid resource
	try {
	    DAVServerResponse response = connection.get(fullSubContextName);
	    if (!response.isSuccess()) {
		// network problem or other DAV error
		throw new DAVResourceNotFoundException();
	    } else {
		// this is a valid resource
		// make sure it points to a context and not an instance
		if (DAVUtilityCalls.describesCollection(new String(response.getData()))) {
		    // ok, create the DAV connection to this subcontext
		    DAVContext newContext = new DAVContext(host,port,fullSubContextName);
		    return newContext;
		} else {
		    // not a collection
		    throw new BadContextNameException();
		}
	    }
	} catch (DAVResourceNotFoundException exception) {
	    // could not find the specified subcontext
	    // make sure the context is valid
	    checkContextValidity();
	    // context was valid, so the only possibility is that
	    // the subcontext name was wrong
	    throw new NotFoundException();
	}
    }
		
		    
    /**
     * Gets a list of nodes that are immediate children of this context
     * in the registry hierarchy.
     * @return a list of nodes that are children of the current context
     * @exception InvalidContextException if the context being used
     * becomes invalid for some reason
     */
    public RegistryNode [] getChildren() throws InvalidContextException {
	try {
	    // get all the children of this context
	    String [] childNames = DAVUtilityCalls.getChildren(connection,getName());
	    // if this was not successful, something weird happened
	    // (network error or context has been deleted or some such thing)
	    if (childNames == null)
		throw new InvalidContextException();
	    // temporarily store the nodes to be returned in a vector
	    Vector nodes = new Vector();
	    // iterate through the child names
	    for (int i=0; i<childNames.length; i++) {
		String currentChildName = getName()+childNames[i];
		// get this resource from the server
		try {
		    DAVServerResponse response = connection.get(currentChildName);
		    // if this response was not successful skip this child
		    if (response.isSuccess()) {
			// check whether this is a context or not
			String fileContent = new String(response.getData());
			if(DAVUtilityCalls.describesCollection(fileContent)) {
			    // create a registry context for this node
			    DAVContext newContext = new DAVContext(host,port,currentChildName);
			    // add this context to the nodes to be returned
			    nodes.addElement (newContext);
			} else {
			    // create an instance info for this node
			    InstanceInfo currentInstance = 
				InstanceXMLUtils.createInstanceInfoFromXML(fileContent);
			    // if this returns null, there was a parsing error of some
			    // kind, so skip the instance
			    if (currentInstance != null)
				nodes.addElement (currentInstance);
			}
		    }
		} catch (DAVResourceNotFoundException exception) {
		    // the resource got removed just before I asked for it or
		    // there was a network error or some other DAV error
		    // make sure the ocntext is ok
		    checkContextValidity();
		    // skip this child
		}
	    }
	    // now copy the things from the vector to the array which will 
	    // be returned
	    RegistryNode [] registryNodes = new RegistryNode[nodes.size()];
	    for (int i=0; i<nodes.size(); i++) {
		registryNodes[i] = (RegistryNode)nodes.elementAt(i);
	    }
	    return registryNodes;
	} catch (DAVResourceNotFoundException exception) {
	    // could not get children for some reason
	    // the context no longer exists or there was some network
	    // or DAV error
	    throw new InvalidContextException();
	}
    }

    /**
     * Gets a list of all instances that are below this context in the
     * registry hierarchy 
     * @exception InvalidContextException thrown if this context becomes
     * invalid for some reason such as its deletion from the registry
     */
    public InstanceInfo [] getAllInstances() throws InvalidContextException {
	try {
	    // get all the leaves of this context
	    String [] childNames = DAVUtilityCalls.getAllChildren(connection,getName());
	    // if this was not successful, something weird happened
	    // (network error or context has been deleted or some such thing)
	    if (childNames == null)
		throw new InvalidContextException();
	    // temporarily store the instances to be returned in a vector
	    Vector instanceInfo = new Vector();
	    // iterate through the child names
	    for (int i=0; i<childNames.length; i++) {
		String currentChildName = childNames[i];
		// get this resource from the server
		try {
		    System.out.println("Getting info from "+currentChildName);
		    DAVServerResponse response = connection.get(currentChildName);
		    // if the request was unsuccessful, skip this child
		    if (response.isSuccess()) {
			// create an instance info for this node
			String xmlInfo = new String(response.getData());
			InstanceInfo currentInstance = 
			    InstanceXMLUtils.createInstanceInfoFromXML(xmlInfo);
			// if this returns null, there was a parsing error of some
			// kind, so skip the instance
			if (currentInstance != null)
			    instanceInfo.addElement (currentInstance);
		    }
		} catch (DAVResourceNotFoundException exception) {
		    // the resource got removed just before I asked for it or
		    // there was a network error or some other DAV error
		    // make sure the context is ok
		    checkContextValidity();
		    // skip this child
		}
	    }
	    // now copy the things from the vector to the array which will 
	    // be returned
	    InstanceInfo [] instances = new InstanceInfo[instanceInfo.size()];
	    for (int i=0; i<instanceInfo.size(); i++) {
		instances[i] = (InstanceInfo) instanceInfo.elementAt(i);
	    }
	    return instances;
	} catch (DAVResourceNotFoundException exception) {
	    // could not get all children for some reason
	    // the context no longer exists or there was some network
	    // or DAV error
	    throw new InvalidContextException();
	}
    }


	
    /**
     * Binds the given stringified instance information into this
     * context under the given name
     */
    protected void bind(String xmlInstanceDescription, String instanceName,
			ComponentID id)
	throws InvalidContextException {
	// create the full name and make sure it has a trailing slash
	String instanceCollection = getName()+instanceName;
	System.out.println("Binding a new instance into "+instanceCollection);
	DAVContext instanceContext = null;
	if (!instanceName.endsWith("/"))
	    instanceCollection += "/";
	// check if there is already a collection with the specified instance
	// name
	// check if the required collection for the instance exists
	// if not, create one
	try {
	    DAVServerResponse response = connection.get(instanceCollection);
	    System.out.println("Collection "+instanceCollection+" exists...");
	    // collection exists
	} catch (DAVResourceNotFoundException exception) {
	    // create a collection with the required name
	    // this may necessitate creating a whole set of collections
	    // e.g. if the collection name is something like /foo/bar/foobar/
	    // and this does not exist, we have to check if foo exists otherwise
	    // create it, likewise for /foo/bar/ and /foo/bar/foobar/
	    System.out.println("Collection does not exist, creating it...");
	    try {
		DAVUtilityCalls.createCollectionPath(connection,instanceName,getName());
		System.out.println("Creation complete");
	    } catch (DAVResourceNotFoundException contextException) {
		// something went wrong with the context or there was
		// a network or DAV exception
		throw new InvalidContextException();
	    }
	}
	// get the file name to be used for the instance
	String instanceFileName = 
	    InstanceXMLUtils.getInstanceFileName(xmlInstanceDescription);
	System.out.println("Instance file name will be "+instanceFileName+
			   ", putting this in the collection");
	// create a resource with this name and put the instance
	// data in it
	try {
	    connection.put(instanceCollection+instanceFileName,xmlInstanceDescription);
	    // put the instance name into the ocmponent id
	    ((ComponentIDregistry)id).setName(instanceCollection);
	    // generate an instance bound event here
	    // TODO
	    System.out.println("Put complete");
	} catch (DAVResourceNotFoundException exception) {
	    // can't see how this can happen unless we have a weird network
	    // or DAV error
	    throw new InvalidContextException();
	} catch (RemoteException exception) {
	    System.out.println("Got exception while setting name of component id");
	    exception.printStackTrace();
	}
    }

    /**
     * Removes the binding of the specified id from the registry.
     * @param id the component ID to be unbound
     * @exception NotFoundException thrown if this id was never bound
     * in this context
     */
    public void unbind(ComponentID id) 
	throws NotFoundException, InvalidContextException {
	// the instance name for this component
	String instanceName = "";
	try {
	    instanceName = ((ComponentIDregistry)id).getName();
	} catch (RemoteException exception) {
	    System.out.println("Got exception "+exception+" while getting instance name from ComponentID");
	    exception.printStackTrace();
	}
	// if any prefix of the name matches a suffix of the context name,
	// ignore that portion
	int lastSlashPos = instanceName.lastIndexOf("/");
	String nameRelativeToContext = "";
	String contextName = this.getName();
	while(lastSlashPos != -1 && lastSlashPos != 0) {
	    String possibleMatchingPrefix = instanceName.substring(0,lastSlashPos+1);
	    System.out.println(possibleMatchingPrefix+" may match...");
	    if (contextName.endsWith(possibleMatchingPrefix)) {
		// the name relative to the context is whatever follows 
		// the prefix
		// if nothing follws the prefix, the name is empty
		if ((lastSlashPos+1) < instanceName.length())
		    nameRelativeToContext = instanceName.substring(lastSlashPos+1);
		else
		    nameRelativeToContext = "";
		System.out.println("Prefix matched, name relative to this "+
				   " context is "+nameRelativeToContext);
		break;
	    }
	    lastSlashPos = instanceName.substring(0,lastSlashPos).lastIndexOf("/");
	}
	System.out.println("Attempting to unbind instance "+nameRelativeToContext);
	// get the instances that have been bound with this name
	InstanceInfo [] instances = lookup(nameRelativeToContext);
	System.out.println("Found "+instances.length+" matching instances");
	// iterate through them and check whose id matches this one
	for (int i=0; i<instances.length; i++) {
	    System.out.println("Checking if IDs are equal...");
	    try {
		ComponentIDregistry davID = (ComponentIDregistry) instances[i].getComponentID();
		System.out.println("The stored ID is "+davID);
		if (davID.equals((ComponentIDregistry)id)) {
		    // remove this instance
		    // use strong deletion so empty parent directories will
		    // also be removed
		    System.out.println("They are equal");
		    try {
			if (!nameRelativeToContext.endsWith("/"))
			    nameRelativeToContext += "/";
			String fileName = getName()+nameRelativeToContext+instances[i].getInfo("instance-info")[0].getValue("unique-id")[0];
			System.out.println("Got a matching id, deleting "+fileName);
			DAVUtilityCalls.strongDelete(connection,fileName);
			// generate an instance unbound event here
			// TODO
		    return;
		    } catch (DAVResourceNotFoundException exception) {
			// some resource became stale or there was a 
			// network error
			throw new InvalidContextException();
		    }
		} else {
		    System.out.println("They are not equal");
		} 
	    } catch (Exception exception) {
		System.out.println("Got exception "+exception+" while doing unbind");
		exception.printStackTrace();
	    }
	}
	// did not find a matching instance
	throw new NotFoundException();
    }

    /**
     * @return a string representation of this context
     */
    public String toString() {
	StringBuffer sb = new StringBuffer();
	sb.append("Host: ");
	sb.append(host);
	sb.append("\nPort: ");
	sb.append(port);
	sb.append("\n");
	return sb.toString();
    }

    /**
     * Check if the context is valid.
     * This is done by requesting this context from the server
     * and making sure there was no error. 
     * @exception InvalidContextException thrown if the context
     * is not valid
     */
    private void checkContextValidity() throws InvalidContextException {
	// request the context from the server
	try {
	    DAVServerResponse response = connection.get(getName());
	    // got response, check if the request was successful
	    if (!response.isSuccess())
		throw new InvalidContextException();
	} catch (DAVResourceNotFoundException exception) {
	    throw new InvalidContextException();
	}
    }
}
