/**
 * This class contains a set of utility methods for accessing instance
 * information stored in XML format and for creating XML-ified versions
 * of instance information
 * @author Nirmal (nmukhi)
 * @see InstanceInfo
 */
package registry;

import java.io.File;
import java.io.FileInputStream;
import java.io.ByteArrayOutputStream;
import java.util.Date;
import java.util.Enumeration;
import com.ms.xml.om.Element;
import com.ms.xml.parser.ParseException;
import com.ms.xml.om.Document;
import com.ms.xml.util.Name;
import com.ms.xml.util.XMLOutputStream;
import util.GeneralXMLUtils;
import util.Info;
import util.InfoImplementation;
import util.EnvObj;
import ccacore.ComponentID;
import ccacore.ComponentIDregistry;
import ccacore.ComponentIDImpl;
import ccacore.ComponentIDFactory;
import ids.InformationXMLUtils;
import nexusrmi.RemoteException;

public class InstanceXMLUtils extends GeneralXMLUtils {
    private static String DAV_HOST = "lookout.extreme.indiana.edu";
    private static int DAV_PORT = 4242;


    /**
     * Create information about an instance from its XML description
     * @param xmlInformation the XML description of the instance
     * @return an <var>InstanceInfo</var> object describing 
     * the instance or null if there was a parsing exception of
     * any kind
     */
    public static InstanceInfo createInstanceInfoFromXML(String xmlInformation) {
	// parse the XML string into instance information
	InfoImplementation info = GeneralXMLUtils.getInfoFromXML(xmlInformation);
	// if this returned null, the XML could not be parsed, just return null
	if (info == null)
	    return null;
	// get the name of the instance from the XML information
	Info [] instanceInfo = info.getInfo("instance-info");
	if (instanceInfo.length == 0)
	    return null;
	String [] instanceNames = instanceInfo[0].getValue("instance-name");
	// get the component id of the instance from the XML information
	String [] stringifiedId = instanceInfo[0].getValue("componentID");
	try {
	    ComponentID id = ComponentIDFactory.create(stringifiedId[0]);
	    // create the instance info object
	    return new InstanceInfoImplementation(id,instanceNames[0],info);
	} catch (RemoteException exception) {
	    System.out.println("Got exception "+exception+" while creating instance info from XML");
	    exception.printStackTrace();
	}
	return null;
    }

    /**
     * Create an XML string describing an instance given the corresponding 
     * information object and its name.
     * @param info the <tt>InstanceInfo</tt> object describing the instance
     * @param instanceName the name of the instance. 
     */
    public static String createXMLFromInfo(InstanceInfo info, 
					   String instanceName) {
	// create an XML document
	Document document = new Document();
	document.setOutputStyle(XMLOutputStream.PRETTY);
	document.setVersion("1.0");
	document.setEncoding("us-ascii");
	// create the DTD node
	Element dtdNode = document.createElement(Element.DTD);
	dtdNode.setAttribute("NAME",Name.create("instance-info"));
	dtdNode.setAttribute("URL",Name.create("http://www.cs.indiana.edu/~nmukhi/cat/test/instance-info.dtd"));
	document.addChild(dtdNode,null);
	// create the root node
	Element rootNode = document.createElement(Element.ELEMENT,"instance-info");
	document.addChild(rootNode,null);
	// begin adding children to the document, with the current parent element
	// as the root
	GeneralXMLUtils.createXMLLevelFromInfo(document,info.getInfo("instance-info")[0],rootNode);
	// adjust the component name in the XML tree to match the provided component
	// name
	Enumeration rootChildren = rootNode.getElements();
	while(rootChildren.hasMoreElements()) {
	    Element current = (Element) rootChildren.nextElement();
	    if (current.getTagName().getName().equals("instance-name")) {
		// get the only child of this element, which will be the
		// PCDATA corresponding to the value of the component name
		Element name = current.getChild(0);
		// make sure the name is the same as the name under 
		// which it was bound
		name.setText(instanceName);
		break;
	    }
	}
	// write the XML tree into a file
	try {
	    ByteArrayOutputStream outStream = new ByteArrayOutputStream();
	    document.save(document.createOutputStream(outStream));
	    return outStream.toString();
	} catch (Exception exception) {
	    System.err.println("Could not create XML file, got exception "+exception);
	}
	return null;
    }
	
    /**
     * Create an XML string describing the instance given its name, creation
     * environment and component ID
     * @param id the instance's <var>ComponentID</var>
     * @param instanceName the user provided name for this instance
     * @param env the instance's creation environment
     * @return the XML string describing the instance
     */
    public static String createXMLFromInfo(ComponentID id, String instanceName,
					   EnvObj env, Info componentInfo) {
	// get the required information from the component information
	Info portTypeInfo = componentInfo.getInfo("component-info")[0].
	    getInfo("port-type-list")[0];
	Info portInfo = componentInfo.getInfo("component-info")[0].
	    getInfo("port-list")[0];
	Info paramInfo = componentInfo.getInfo("component-info")[0].
	    getInfo("parameter-block")[0];
	String executableValue = componentInfo.getInfo("component-info")[0].
	    getValue("executable")[0];
	// create an XML document
	Document document = new Document();
	document.setOutputStyle(XMLOutputStream.PRETTY);
	document.setVersion("1.0");
	document.setEncoding("us-ascii");
	// create the DTD node
	Element dtdNode = document.createElement(Element.DTD);
	dtdNode.setAttribute("NAME",Name.create("instance-info"));
	dtdNode.setAttribute("URL",Name.create("http://www.cs.indiana.edu/~nmukhi/cat/test/instance-info.dtd"));
	document.addChild(dtdNode,null);
	// create the root node
	Element rootNode = document.createElement(Element.ELEMENT,"instance-info");
	document.addChild(rootNode,null);
	// create a node for the unique id
	// construct the unique id first
	// let's make it a combination of the username, instance name and 
	// date
	// make sure there is no / in the instance name
	String instanceNameNoSlash = instanceName.replace('/','_');
	String fileName = instanceNameNoSlash+env.get("username")+env.get("creation-time");
	Element uniqueIdNode = document.createElement(rootNode,Element.ELEMENT,
						      Name.create("uniqueID"),null);
	// add the unique id
	Element uniqueIdText = document.createElement(uniqueIdNode,Element.PCDATA,
						      null,fileName);
	System.out.println("Added uniqueID node");
	// create a node for the instance name
	//	System.out.println("Adding instance name node");
	Element nameNode = document.createElement(rootNode,Element.ELEMENT,
						  Name.create("instance-name"),null);
	// add the name
	Element nameNodeText = document.createElement(nameNode,Element.PCDATA,
						      null,instanceName);
	System.out.println("Added instance name node");
	// create a node for the component id
	//	System.out.println("Adding component ID node");
	Element idNode = document.createElement(rootNode,Element.ELEMENT,
						Name.create("componentID"),null);
	// add the component id text
	Element idNodeText = null;
	try {
	    System.out.println("Adding to component id node: "+((ComponentIDregistry)id).getRef());
	    idNodeText = document.createElement(idNode,Element.PCDATA,
						null,((ComponentIDregistry)id).getRef());
	} catch (RemoteException exception) {
	    System.out.println("Got exception "+exception+" while writing component ID to XML node");
	    exception.printStackTrace();
	}
	System.out.println("Added component ID node");
	// add the machine name node
	Element machineNode = document.createElement(rootNode,Element.ELEMENT,
						     Name.create("target-host"),null);
	System.out.println("Added instance name node");
	// add the creation method node
	Element creationMethod = document.createElement(rootNode,Element.ELEMENT,
							Name.create("creation-mechanism"),
							null);
	System.out.println("Added creation method node");
	// add the creation env information
	Element creationInfo = document.createElement(rootNode,Element.ELEMENT,
						      Name.create("creation-info"),
						      null);
	System.out.println("Added creation info node");
	String [] keys = env.getAllNames();
	for (int i=0; i<keys.length; i++) {
	    // create the name-value pair node
	    Element envNVPairNode = document.createElement(creationInfo,Element.ELEMENT,
							   Name.create("name-value-pair"),
							   null);
	    // create a node for this name
	    Element envNameNode = document.createElement(envNVPairNode,Element.ELEMENT,
							 Name.create("name"),
							 null);
	    // add its value
	    Element nodeText = document.createElement(envNameNode,Element.PCDATA,
						      null,keys[i]);
	    // create a node for the value for this key
	    Element valueNode = document.createElement(envNVPairNode,Element.ELEMENT,
						       Name.create("value"),
						       null);
	    // add its value
	    Element valueText = document.createElement(valueNode,Element.PCDATA,
						       null,env.get(keys[i]));
	    System.out.println("Added node "+keys[i]);
	    // if this was the host name, add it to the target host node too
	    if (keys[i].equals("host"))
		document.createElement(machineNode,Element.PCDATA,
				       null,env.get(keys[i]));
	    // if this was the creation mechanism, add it to that node too
	    if (keys[i].equals("creation-mech"))
		document.createElement(creationMethod,Element.PCDATA,null,
				       env.get(keys[i]));
	}
	// add the executable info
	Element executableNode = document.createElement(rootNode,Element.ELEMENT,
							Name.create("executable"),
							null);
	document.createElement(executableNode,Element.PCDATA,null,executableValue);
	// add the port type info
	Element portTypeNode = document.createElement(rootNode,Element.ELEMENT,
						      Name.create("port-type-list"),
						      null);
	GeneralXMLUtils.createXMLLevelFromInfo(document,portTypeInfo,portTypeNode);
	System.out.println("Added port type info");
	// add the port info
	Element portNode = document.createElement(rootNode,Element.ELEMENT,
						  Name.create("port-list"),
						  null);
	GeneralXMLUtils.createXMLLevelFromInfo(document,portInfo,portNode);
	System.out.println("Added port info");
	// add the parameter information, if there is any
	if (paramInfo!=null) {
	    Element paramNode = document.createElement(rootNode,Element.ELEMENT,
						       Name.create("parameter-block"),
						       null);
	    GeneralXMLUtils.createXMLLevelFromInfo(document,paramInfo,paramNode);
	    System.out.println("Added param info");
	}
	// write the XML tree into a file
	try {
	    ByteArrayOutputStream outStream = new ByteArrayOutputStream();
	    document.save(document.createOutputStream(outStream));
	    return outStream.toString();
	} catch (Exception exception) {
	    System.err.println("Could not create XML file for instance, got exception "+exception);
	}
	return null;
    }
    
    /**
     * Extract an instance's unique file name given the XML data describing
     * the instance
     * @param xmlInformation the description of the instance in XML
     * @return the unique file name to be used for this instance or null
     * if there was a parsing error
     */
    public static String getInstanceFileName(String xmlInformation) {
	// parse the provided data and access the information about
	// the file name (unique id)
	String tagName = "uniqueID";
	// find the start of the tag
	int startOfName = xmlInformation.indexOf(tagName);
	startOfName += tagName.length() + 1; // add one for the > at end of tag
	// find the end of the tag
	int endOfName = xmlInformation.indexOf(tagName,startOfName);
	endOfName-=2; // subtract two for the </ at the start of the tag
	// return stuff in between the start and end of the tag
	return xmlInformation.substring(startOfName,endOfName);
    }

    /**
     * Get the default registry host to be used
     * @return the name of the host to be used for the default registry
     */
    private static String getDefaultRegistryHost() {
	return DAV_HOST;
    }
    
    /**
     * Get the default registry port to be used
     * @return the port number to be used for the default registry
     */
    private static int getDefaultRegistryPort() {
	return DAV_PORT;
    }
    
    /**
     * Get the default information directory service (IDS) host to be used
     * @return the name of the host to be used for the IDS
     */
    private static String getDefaultIDSHost() {
	return DAV_HOST;
    }
    
    /**
     * Get the default IDS port to be used
     * @return the port number to be used for the default IDS
     */
    private static int getDefaultIDSPort() {
	return DAV_PORT;
    }
    
    public static void main(String [] args) {
	// create an instance info object for the specified file
	// argv[0] is the XML file to be converted to an Info
	// open the file
	try {
	    File xmlFile = new File(args[0]);
	    // read stuff into a string
	    FileInputStream inStream = new FileInputStream(xmlFile);
	    ByteArrayOutputStream outStream = new ByteArrayOutputStream();
	    int currentByte;
	    currentByte = inStream.read();
	    while(currentByte != -1) {
		// put the current char into the byte array
		outStream.write(currentByte);
		currentByte = inStream.read();
	    }
	    // parse into an info object
	    InstanceInfo infoObj = InstanceXMLUtils.
		createInstanceInfoFromXML(outStream.toString());
	    // print it out
	} catch (Exception exception) {
	    System.err.println("Got exception "+exception);
	    System.exit(1);
	}
    }

}
