/**
 * This class is a handle to a directory on the local filesystem.
 * It is used through the <var>RegistryContext</var> interface
 * to navigate the file hierarchy and add and remove instance 
 * information from it.
 * @author Nirmal (nmukhi)
 */

package registry;

import java.io.File;
import java.io.FileReader;
import java.io.StringWriter;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.FileNotFoundException;
import java.util.Vector;
import java.util.Enumeration;
import com.ms.xml.util.StringInputStream;
import ccacore.ComponentID;
import ccacore.ComponentIDregistry;
import nexusrmi.RemoteException;

public class LocalContext extends Context implements RegistryContext {
    private File contextDirectory;

    /**
     * The default directory where the context will point to
     */
    public static String DEFAULT_DIRECTORY = "resources";

    /**
     * Create a local context pointing to the default directory
     */
    public LocalContext() {
	this(DEFAULT_DIRECTORY);
    }

    /**
     * Create a local context pointing to the specified directory
     */
    public LocalContext(String directory) {
	super(directory);
	contextDirectory = new File(directory);
    }

    /**
     * Looks for instances that have the specified name. There may
     * be more than one since this is a user-provided instance name
     * supplied at the time of the <a href="#bind">bind</a> call and 
     * cannot be guaranteed unique.
     * @param instanceName the name of the instance
     * @return a list of <var>InstanceInfo</var> objects describing
     * the instances with the specified name.
     * @exception NotFoundException if there is no instance with the
     * provided <var>instanceName</var>.
     * @exception InvalidContextException if the context being used
     * becomes invalid for some reason such as its deletion from the 
     * registry
     */
    public InstanceInfo [] lookup(String instanceName) 
	throws NotFoundException, InvalidContextException {
	// treat the instanceName as a subdirectory in this context
	// and look for it
	File instanceDirectory = new File(contextDirectory,instanceName);
	if (!instanceDirectory.exists() || !instanceDirectory.isDirectory()) {
	    // make sure that the context is valid
	    checkContextValidity();
	    // context was valid, so the instance directory was not 
	    // there at all
	    throw new NotFoundException();
	}
	// if the instance directory does not have permissions, throw
	// invalid context exception (to be consistent with DAVContext)
	if (!instanceDirectory.canRead())
	    throw new InvalidContextException();
	// iterate through the files in the instance directory and convert
	// them into info objects
	// initially store info objects in a vector
	Vector instanceInfo = new Vector();
	try {
	    // get all the files/subdirectories in this directory
	    File [] children = instanceDirectory.listFiles();
	    // if the call returned null, there was a problem with the
	    // context
	    if (children == null)
		throw new InvalidContextException();
	    for (int i=0; i<children.length; i++) {
		// get the data from the current file
		// if this child is a subdirectory, ignore it
		if (!children[i].isFile())
		    continue;
		String fileData = readFile(children[i]);
		if (fileData == null) {
		    // something went wrong while reading this file
		    // make sure the context is valid
		    checkContextValidity();
		    // skip this file and go to the next one
		} else {
		    // assume the data is XML describing an instance, and try
		    // parsing it to create instance information
		    InstanceInfo currentInstance = 
			InstanceXMLUtils.createInstanceInfoFromXML(fileData);
		    // if this is null, there was a parsin error, so
		    // ignore this child, otherwise add it to our instance 
		    // vector
		    if (currentInstance != null)
			instanceInfo.add(currentInstance);
		}
	    }
	    // copy instance information from vector and return that
	    InstanceInfo [] instances = new InstanceInfo[instanceInfo.size()];
	    for (int i=0; i<instanceInfo.size(); i++) {
		instances[i] = (InstanceInfo)instanceInfo.elementAt(i);
	    }
	    // return the information on the requested instances
	    return instances;
	} catch (SecurityException exception) {
	    // could not list files in the context
	   throw new InvalidContextException();
	}
    }

    /**
     * Gets a context for the specified subcontext name.
     * @param subcontext the name of a valid context accessible 
     * from this context.
     * @return a handle to the named context
     * @exception BadContextNameException if the specified 
     * context does not correspond to a context
     * @exception InvalidContextException if the context being used
     * becomes invalid for some reason such as its deletion from
     * the registry
     */
    public RegistryContext getSubContext(String subcontext)
	throws BadContextNameException, InvalidContextException, NotFoundException {
	// get the file object for the subcontext
	File subContext = new File(contextDirectory,subcontext);
	// if this does not exist, throw not found exception
	// after checking that the context is valid
	if (!subContext.exists()) {
	    checkContextValidity();
	    throw new NotFoundException();
	}
	// if this does not correspond to a directory, throw 
	// BadContextName exception
	if (!subContext.isDirectory())
	    throw new BadContextNameException();
	// it exists, and is a directory!
	// create a local context for it and return
	return new LocalContext(subContext.getAbsolutePath());
    }

    /**
     * Gets a list of nodes that are immediate children of this context
     * in the registry hierarchy.
     * @return a list of nodes that are children of the current context
     * @exception InvalidContextException if the context being used
     * becomes invalid for some reason
     */
    public RegistryNode [] getChildren() throws InvalidContextException {
	try {
	    // get all children of this context
	    File [] children = contextDirectory.listFiles();
	    // if the call returned null, there was a problem 
	    // with the context
	    if (children == null)
		throw new InvalidContextException();
	    // create a vector of registry nodes to store the nodes in 
	    // the context
	    Vector nodes = new Vector();
	    // iterate through these
	    for (int i=0; i<children.length; i++) {
		File currentChild = children[i];
		if (currentChild.isDirectory()) {
		    // this is a directory, create a local 
		    // context for it
		    nodes.add(new LocalContext(currentChild.getAbsolutePath()));
		} else {
		    // this is a file, assume it describes an instance and try
		    // parsing it
		    String instanceXMLData = readFile(currentChild);
		    // if the data was null, there was an error while reading the
		    // file, so just skip this child
		    if (instanceXMLData != null) {
			InstanceInfo instanceInfo = 
			    InstanceXMLUtils.createInstanceInfoFromXML(instanceXMLData);
			// if the information was null, there was an error while
			// parsing the XML, so ignore this child otherwise add it
			// to our node list
			if (instanceInfo != null)
			    nodes.add(instanceInfo);
		    } else {
			// make sure the context is still valid
			checkContextValidity();
		    }
		}
	    }
	    // now copy the things from the vector to the array which will 
	    // be returned
	    RegistryNode [] registryNodes = new RegistryNode[nodes.size()];
	    for (int i=0; i<nodes.size(); i++) {
		registryNodes[i] = (RegistryNode)nodes.elementAt(i);
	    }
	    return registryNodes;
	} catch (SecurityException exception) {
	    // could not list files in the context
	    throw new InvalidContextException();
	}
    }

    /**
     * Gets a list of all instances that are below this context in the
     * registry hierarchy 
     * @exception InvalidContextException thrown if this context becomes
     * invalid for some reason such as its deletion from the registry
     */
    public InstanceInfo [] getAllInstances() throws InvalidContextException {
	// do an iterative breadth first search of the directory
	// tree 
	Vector directoriesRemaining = new Vector();
	Vector instanceInfo = new Vector();
	// add this context to the list of directories to be searched
	directoriesRemaining.add(contextDirectory);
	// go through all the directories that remain to be searched
	while(directoriesRemaining.size()!=0) {
	    // get the first remaining directory
	    File currentDir = (File)directoriesRemaining.remove(0);
	    try {
		File [] currentChildren = currentDir.listFiles();
		// if the call returned null, there was a problem with the 
		// currentDir
		if (currentChildren == null)
		    throw new InvalidContextException();
		// go through all the children
		for (int i=0; i<currentChildren.length; i++) {
		    File currentChild = currentChildren[i];
		    if (currentChild.isDirectory()) {
			// a subdirectory, add it to the directories remaining to 
			// be searched
			directoriesRemaining.add(currentChild);
		    } else {
			// a file, assume it describes an instance and try
			// parsing it
			String instanceXMLData = readFile(currentChild);
			// if this was null, there was an error while reading the 
			// file, skip this child
			if (instanceXMLData != null) {
			    InstanceInfo info = 
				InstanceXMLUtils.createInstanceInfoFromXML(instanceXMLData);
			    // if this was null, ther was a parsing error, skip the file
			    // otherwise add it to our instance list
			    if (info != null)
				instanceInfo.add(info);
			} else {
			    // make sure the context is still valid
			    checkContextValidity();
			}
		    }
		}
	    } catch (SecurityException exception) {
		// could not list files for some reason
		throw new InvalidContextException();
	    }
	}
	// now copy the things from the vector to the array which will 
	// be returned
	InstanceInfo [] instances = new InstanceInfo[instanceInfo.size()];
	for (int i=0; i<instanceInfo.size(); i++) {
	    instances[i] = (InstanceInfo) instanceInfo.elementAt(i);
	}
	return instances;
    }

    /**
     * Binds the given stringified instance information into this
     * context under the given name. <tt>id</tt> is the reference
     * to the actual running instance.
     */
    protected void bind(String xmlInstanceDescription, String instanceName, 
			ComponentID id)
	throws InvalidContextException {
	// make sure the context is valid before trying to create the
	// instance directory
	checkContextValidity();
	File instanceDir = new File(contextDirectory,instanceName);
	try {
	    // create the collection in which the instance will live
	    // if it doesn't already exist
	    boolean dirCreated = instanceDir.mkdirs();
	    if (dirCreated)
		System.out.println("Created "+instanceDir);
	    else
		System.out.println(instanceDir+" exists");
	    // get the file name to be used to store the instance
	    String instanceFileName = 
		InstanceXMLUtils.getInstanceFileName(xmlInstanceDescription);
	    // put this file in the instance directory with the specified data
	    // first create the instance file
	    File instanceFile = new File(instanceDir,instanceFileName);
	    instanceFile.createNewFile();
	    // now fill it with the data
	    FileOutputStream outStream = new FileOutputStream(instanceFile);
	    StringInputStream inStream = new StringInputStream(xmlInstanceDescription);
	    int currentChar = inStream.read();
	    while (currentChar!=-1) {
		outStream.write(currentChar);
		currentChar = inStream.read();
	    }
	    // close the streams
	    inStream.close();
	    outStream.close();
	    // set the name of the instance the component id refers to
	    try {
		// make sure the component ID refers to the FQN of the instance
		// i.e. tag the context name in front
		// and make sure it ends with a /
		if (!instanceName.endsWith("/"))
		    instanceName += "/";
		((ComponentIDregistry)id).setName(getName()+instanceName);
	    } catch (RemoteException exception) {
		System.out.println("Got exception "+exception+" while setting name of ComponentID");
		exception.printStackTrace();
	    }
	    // generate an instance bound event here
	    // TODO
	} catch (SecurityException exception) {
	    // could not create the instance directory or the instance file
	    throw new InvalidContextException();
	} catch (IOException exception) {
	    // could not write the instance file
	    throw new InvalidContextException();
	}
    }

    /**
     * Removes the binding of the specified id from the registry.
     * @param id the component ID to be unbound
     * @exception NotFoundException thrown if this id was never bound
     * in this context
     */
    public void unbind(ComponentID id) 
	throws NotFoundException, InvalidContextException {
	// get the name of the instance the id refers to
	String instanceName = "";
	try {
	    instanceName = ((ComponentIDregistry)id).getName();
	} catch (RemoteException exception) {
	    System.out.println("Got exception "+exception+" while getting instance name from ComponentID");
	    exception.printStackTrace();
	}
	// if any prefix of the name matches a suffix of the context name,
	// ignore that portion
	int lastSlashPos = instanceName.lastIndexOf("/");
	String nameRelativeToContext = "";
	String contextName = this.getName();
	while(lastSlashPos != -1 && lastSlashPos != 0) {
	    String possibleMatchingPrefix = instanceName.substring(0,lastSlashPos+1);
	    System.out.println(possibleMatchingPrefix+" may match...");
	    if (contextName.endsWith(possibleMatchingPrefix)) {
		// the name relative to the context is whatever follows 
		// the prefix
		// if nothing follws the prefix, the name is empty
		if ((lastSlashPos+1) < instanceName.length())
		    nameRelativeToContext = instanceName.substring(lastSlashPos+1);
		else
		    nameRelativeToContext = "";
		System.out.println("Prefix matched, name relative to this "+
				   " context is "+nameRelativeToContext);
		break;
	    }
	    lastSlashPos = instanceName.substring(0,lastSlashPos).lastIndexOf("/");
	}
	// get the instances with this name
	InstanceInfo [] instances = lookup(nameRelativeToContext);
	System.out.println("Found "+instances.length+" matching instances");
	// iterate through them and check whose id matches this one
	for (int i=0; i<instances.length; i++) {
	    try {
		if (((ComponentIDregistry)(instances[i].getComponentID())).equals((ComponentIDregistry)id)) {
		    // remove this instance
		    // use strong deletion so empty parent directories will
		    // also be removed
		    try {
			String dirName = this.getName()+nameRelativeToContext;
			if (!dirName.endsWith("/"))
			    dirName += "/";
			String fileName = dirName+instances[i].getInfo("instance-info")[0].getValue("unique-id")[0];
			System.out.println("Got a matching id, deleting "+fileName);
			// delete file
			File fileToDelete = new File(fileName);
			File parentDir = fileToDelete.getParentFile();
			boolean deleted = fileToDelete.delete();
			while(deleted && parentDir!=null) {
			    // keep deleting parent directories while it is 
			    // possible to do so
			    fileToDelete = parentDir;
			    parentDir = fileToDelete.getParentFile();
			    deleted = fileToDelete.delete();
			}
			// generate an instance unbound event here
			// TODO
			return;
		    } catch (SecurityException exception) {
			// deletion was not allowed somewhere
			throw new InvalidContextException();
		    }
		}
	    } catch (Exception exception) {
		System.out.println("Got exception "+exception+" while doing unbind");
		exception.printStackTrace();
	    }
	}
	// nothing matched the given component id
	throw new NotFoundException();
    }

    /**
     * Check if this context is valid. This is done by making sure that
     * a directory with the context name exists
     */
    private void checkContextValidity() throws InvalidContextException {
	// check that the context directory exists, is a directory
	// and can be read form and written to
	if (!contextDirectory.exists() || !contextDirectory.canRead() ||
	    !contextDirectory.canWrite() || !contextDirectory.isDirectory())
	    throw new InvalidContextException(contextDirectory+" is not a valid context");
    }
    
    /**
     * Read information from the specified file and return it as 
     * a string
     * @param file the file to be read
     * @return the data in the file or null if the file was not found
     * or an error was encountered while reading it
     */
    private String readFile(File file) {
	// open a stream for reading
	FileReader reader = null;
	try {
	    reader = new FileReader(file);
	} catch (FileNotFoundException exception) {
	    // could not find file
	    return null;
	}
	// open a stream for writing
	StringWriter writer = new StringWriter();
	// read one character at a time till the whole
	// file has been read
	try {
	    // read a character
	    int currentCharacter = reader.read();
	    // write the character as long as we have not reached the end
	    while(currentCharacter != -1) {
		writer.write(currentCharacter);
		currentCharacter = reader.read();
	    }
	    // we've reached the end, return what we have so far
	    reader.close();
	    writer.close();
	    return writer.toString();
	} catch (IOException exception) {
	    // problem while reading, just return null
	    return null;
	}
    }

    /**
     * Creates a string representation of the context and returns it
     * @return a string representation of the context
     */
    public String toString() {
	return contextDirectory.toString();
    }
}
