/**
 * This interface is a handle to the information hierarchy on the registry.
 * It provides methods to traverse the information hierarchy, bind
 * instances into the registry, lookup previously bound instances, etc.
 * Logically a registry context is a pointer to a specific directory
 * on a server that maintains instance information.
 * Details on the API are available from 
 * <a href="http://www.extreme.indiana.edu/ccat/registry_service.html">here</a>.
 * @author Nirmal (nmukhi)
 * @see RegistryNode
 */

package registry;

import util.EnvObj;
import ccacore.ComponentID;
import ids.InformationInfo;

public interface RegistryContext extends RegistryNode {

    /**
     * Looks for instances that have the specified name. There may
     * be more than one since this is a user-provided instance name
     * supplied at the time of the <a href="#bind">bind</a> call and 
     * cannot be guaranteed unique.
     * @param instanceName the name of the instance
     * @return a list of <var>InstanceInfo</var> objects describing
     * the instances with the specified name.
     * @exception NotFoundException if there is not instance with the
     * provided <var>instanceName</var>.
     * @exception InvalidContextException if the context being used
     * becomes invalid for some reason such as its deletion from the 
     * registry
     */
    public InstanceInfo [] lookup(String instanceName) 
	throws NotFoundException, InvalidContextException;

    /**
     * Gets a context for the specified subcontext name.
     * @param subcontext the name of a valid context accessible 
     * from this context.
     * @return a handle to the named context
     * @exception BadContextNameException if the specified 
     * context does not correspond to a context
     * @exception InvalidContextException if the context being used
     * becomes invalid for some reason such as its deletion from
     * the registry
     * @exception NotFoundException if the specified context is not
     * bound in this context.
     */
    public RegistryContext getSubContext(String subcontext)
	throws BadContextNameException, InvalidContextException, NotFoundException;

    /**
     * Gets a list of nodes that are immediate children of this context
     * in the registry hierarchy.
     * @return a list of nodes that are children of the current context
     * @exception InvalidContextException if the context being used
     * becomes invalid for some reason
     */
    public RegistryNode [] getChildren() throws InvalidContextException;

    /**
     * Gets a list of all instances that are below this context in the
     * registry hierarchy 
     * @exception InvalidContextException thrown if this context becomes
     * invalid for some reason such as its deletion fromt he registry
     */
    public InstanceInfo [] getAllInstances() throws InvalidContextException;

    /**
     * Binds the specified <var>id</var> into this context under the 
     * specified name. The environment object contains additional 
     * information about the instance
     * @param id the handle to the component itself
     * @param instanceName a name that can be used to lookup this
     * instance. This need not be unique.
     * @param env the environment object used to create this 
     * instance
     * @see InstanceInfo
     */
    public void bind(ComponentID id, String instanceName, EnvObj env,
		     InformationInfo componentInfo) 
	throws InvalidContextException;

    /**
     * Binds the specified <var>instanceInfo</var> into this context
     * under the specified name. 
     * @param instanceInfo the <tt>InstanceInfo</tt> object describing 
     * the instance
     * @param instanceName a name that can be used to lookup this
     * instance. This need not be unique
     * @see InstanceInfo
     */
    public void bind(InstanceInfo instanceInfo, String instanceName) 
	throws InvalidContextException;

    /**
     * Removes the binding of the specified id from the registry.
     * @param id the component ID to be unbound
     * @exception NotFoundException thrown if this id was never bound
     * in this context
     */
    public void unbind(ComponentID id) throws NotFoundException, InvalidContextException;
}
