/**
 * This class implements the <var>RegistryService</var> interface.
 * Given a named context, it can create an instance of the 
 * appropriate context using a set of context service providers
 * and instantiates the appropriate one.<br>
 * @author Nirmal (nmukhi)
 * @see RegistryContext
 */

package registry;

import util.EnvObj;
import util.Info;
import ccacore.ComponentID;
import ccacore.ComponentIDImpl;

public class RegistryServiceImplementation implements RegistryService {
    // constants
    /**
     * protocol name for local registry
     */
    public static final String LOCAL_PROTOCOL_NAME = "file";

    /**
     * protocol name for DAV registry
     */
    public static final String DAV_PROTOCOL_NAME = "http";

    /**
     * Gets a context for a registry
     * @param context the address of the context, for example
     * http://lookout.extreme.indiana.edu:4242/ccat/instances<br>
     * @return a context object for the named context or null if 
     * the protocol was not recognized
     * @exception NotFoundException thrown if the specified context
     * did not correspond to an actual registry
     * @exception BadContextNameException if the specified context
     * corresponded to a node in the registry which was not a valid
     * context
     */    
    public RegistryContext getContext(String context) 
	throws NotFoundException, BadContextNameException {
	// parse the context name

	// get the name of the protocol to be used
	// this is everything upto the first "://"
	int numCharsBeforeSeparator = context.indexOf("://");
	// if the protocol was not specified, throw a NotFoundException
	if (numCharsBeforeSeparator == -1)
	    throw new NotFoundException("Context name "+context+" had invalid format");
	String protocol = new String(context.substring(0,numCharsBeforeSeparator));

	// the context that will finally be returned
	RegistryContext registryContext = null;
	// the initial resource name on the registry server
	String resourceName = "";

	// if protocol is local, the rest of the name just refers to a local
	// resource
	if (protocol.equals(RegistryServiceImplementation.LOCAL_PROTOCOL_NAME)) {
	    resourceName = new String(context.substring(numCharsBeforeSeparator+2));
	    // now we can create the local context
	    registryContext = new LocalContext(resourceName);
	} else if (protocol.equals(RegistryServiceImplementation.DAV_PROTOCOL_NAME)) {
	    // the portion of the name upto the first / (if any) describes
	    // the server and port (if any)
	    int endOfServerDescription = context.indexOf('/',numCharsBeforeSeparator+3);
	    String serverDescription = "", host = "";
	    int port = -1;
	    if (endOfServerDescription == -1) {
		// no slash, add one
		context = context+"/";
		endOfServerDescription = context.length()-1;
	    }
	    // everything up to the end of the server description is 
	    // server, port stuff
	    // everything after that is the resource name
	    resourceName = context.substring(endOfServerDescription);
	    serverDescription = 
		new String(context.substring(numCharsBeforeSeparator+3,
					     endOfServerDescription));
	    // check if the server description includes a port number
	    // it does if it contains a :
	    int beginningOfPortNumber = serverDescription.lastIndexOf(':');
	    if (beginningOfPortNumber == -1) {
		// no port number
		// the server description just contains the host name
		host = serverDescription;
	    } else {
		// everything up to the port number is the host name
		host = serverDescription.substring(0,beginningOfPortNumber);
		String portNumber = serverDescription.substring(beginningOfPortNumber+1);
		port = new Integer(portNumber).intValue();
	    }
	    // now we know the resource name, host and port so we can create
	    // a DAV context object
	    // if the port number is still -1, it was not specified so let 
	    // the context use the default
	    if (port == -1) {
		registryContext = new DAVContext(host,resourceName);
	    } else {
		registryContext = new DAVContext(host,port,resourceName);
	    }
	}
	return registryContext;
    }

    /**
     * Main to test the registry service code
     * Attempt to pass different context names and see if they are parsed correctly
     * to create the correct context
     */
    public static void main(String args[]) {
	// this test is now outdated
	// will write another one later - Nirmal
/*
	// create the imeplemtation class
	// this will be done internally within the services obj
	RegistryServiceImplementation registryService = new RegistryServiceImplementation();
	// get a reference to the registry service interface alone
	// so I don't begin doing things specific to the impl
	RegistryService service = registryService;
	if (args.length != 1) {
	    System.out.println("Usage\n\tjava RegistryServiceImplementation <initial-context>\nwhere <initial-context> is a pointer to a DAV or local context such as http://lookout.extreme.indiana.edu/testdir or file://u/jsmith/testdir");
	    System.exit(1);
	}
	try {
	    // create a DAV registry context
	    // assume args[0] is he context to be used
	    // for example try http://lookout.extreme.indiana.edu
	    // or file://tmpdir where tmpdir is an existing directory
	    RegistryContext context = service.getContext(args[0]);
	    System.out.println("Created context\n"+context+"\n");
	    System.out.println("\n\n------------------------------\nAttempting to create a dummy instance");
	    // initialize the creation env for an instance
	    EnvObj env = new EnvObj();
	    env.put("username","nmukhi");
	    env.put("LD_LIBRARY_PATH","/foo");
	    env.put("PATH",".:/bin:/usr/bin:/usr/local/bin");
	    env.put("WD","/l/extreme/components/workingdir");
	    env.put("TMPDIR","/tmp");
	    env.put("SHELL","/bin/sh");
	    env.put("registry-host","lookout.extreme.indiana.edu");
	    env.put("registry-port","4242");
	    String creationTime = new java.util.Date().toString().replace(' ','_');
	    env.put("creation-time",creationTime);
	    env.put("creation-mech","gram");
	    env.put("host","bread.extreme.indiana.edu");
	    System.out.println("Creation env is "+env);
	    // set the instance name
	    String instanceName = "myInstance";
	    // create the component ID
	    // this will be obtained as the return value of a call to createInstance
	    // using the CreationService port
	    ComponentID id = new ComponentIDImpl("dummy string");
	    // bind this instance to the registry
	    System.out.println("\n\n------------------------------\nAttempting to bind instance id "+id+" with name "+instanceName);
	    context.bind(id,instanceName,env);
	    // attempt to look it up now
	    System.out.println("\n\n------------------------------\nAttempting to lookup instance "+instanceName);
	    InstanceInfo [] instances = context.lookup(instanceName);
	    System.out.println("There are "+instances.length+" instances with that name in this context, trying to get the one created by nmukhi...");
	    for (int i=0; i<instances.length; i++) {
		InstanceInfo currentInstance = instances[i];
		String userName = currentInstance.getInfo("instance-info")[0].
		    getValue("creator")[0];
		if (userName.equals("nmukhi")) {
		    System.out.println("Instance #"+(i+1)+" created by nmukhi, details on instance are\n"+currentInstance);
		    break;
		}
	    }
	    System.out.println("\n\n------------------------------\nAttempting to create another dummy instance...");
	    // try creating another instance with the same name
	    ComponentID id2 = new ComponentIDImpl("dummy string");
	    // use the same environment with a different username
	    env.put("username","jsmith");
	    System.out.println("Creation env is "+env);
	    System.out.println("\n\n------------------------------\nAttempting to bind instance id "+id2+" with name "+instanceName);
	    context.bind(id2,instanceName,env);
	    // attempt to look it up now
	    System.out.println("\n\n------------------------------\nAttmepting to lookup instance "+instanceName);
	    instances = context.lookup("myInstance");
	    System.out.println("There are "+instances.length+" instances with that name in this context, trying to get the one created by jsmith...");
	    for (int i=0; i<instances.length; i++) {
		InstanceInfo currentInstance = instances[i];
		String userName = currentInstance.getInfo("instance-info")[0].
		    getValue("creator")[0];
		if (userName.equals("jsmith")) {
		    System.out.println("Instance #"+(i+1)+" created by jsmith, details on instance are\n"+currentInstance);
		    break;
		}
	    }
	    ComponentID id3 = new ComponentIDImpl("dummy string");
	    System.out.println("\n\n------------------------------\nAttempting to bind an instance under a very long, name");
	    String longName = "components/dummy/test/ccat/longNamedInstance";
	    System.out.println("Attempting to bind instance id "+id3+" with name "+longName);
	    context.bind(id3,longName,env);
	    System.out.println("\n\n------------------------------\nAttempting to get ALL instances in this context");
	    instances = context.getAllInstances();
	    // iterate through them
	    for (int i=0; i<instances.length; i++) {
		InstanceInfo currentInstance = instances[i];
		String user = currentInstance.getInfo("instance-info")[0].getValue("creator")[0];
		String date = currentInstance.getInfo("instance-info")[0].getValue("creation-time")[0];
		System.out.println("Instance #"+(i+1)+" is named "+currentInstance.getName()+" and was created by "+user+" at time "+date);
	    }
	    // now try unbinding the instance
	    System.out.println("\n\n------------------------------\nAttempting to unbind instance with id "+id3);
	    context.unbind(id3);
	    System.out.println("Now trying to lookup instance "+longName+" (which I just unbound)");
	    try {
		instances = context.lookup(longName);
		System.out.println("Ooops, no NotFoudnException, unbind failed");
	    } catch (NotFoundException e) {
		System.out.println("Got not found exception as excpected");
		System.out.println("Check your local or DAV site to see if empty parent directories of this instance (if any) got deleted");
	    }
	    // get the myInstance subcontext and then get its children
	    System.out.println("\n\n------------------------------\nAttempting to get the myInstance subcontext");
	    RegistryContext subcontext = context.getSubContext("myInstance");
	    System.out.println("Got context "+subcontext.getName());
	    // get the children of this subcontext
	    System.out.println("\n\n------------------------------\nAttempting to get the children of the subcontext myInstance");
	    RegistryNode [] myinstanceChildren = subcontext.getChildren();
	    System.out.println("There are "+myinstanceChildren.length+" children");
	    // iterate through children
	    for (int i=0; i<myinstanceChildren.length; i++) {
		RegistryNode currentChild = myinstanceChildren[i];
		System.out.println("Child #"+(i+1)+" is "+currentChild.getName());
		if (currentChild.isInstance()==1)
		    System.out.println("This is an instance, instance info is\n"+(InstanceInfo)currentChild);
		else
		    System.out.println("This is a context");
	    }
	    System.out.println("\n\n------------------------------\ntesting done\n\n");
	} catch (Exception e) {
	    System.err.println("Got exception "+e+", stack trace is");
	    e.printStackTrace();
	    System.exit(1);
	}*/
    }
}
             
