/**
 * This class is used for general utility methods involving <var>Info</var>
 * objects.
 * @author Nirmal (nmukhi)
 * @see Info
 */

package util;

import java.util.StringTokenizer;

public class GeneralInfoUtils {
    public static String KEY_PATH_SEPARATOR = "/";
    public static String WILD_CARD = "*";
    public static int KEY_INDENT = 4;

    /**
     * This method tries to get a list of values contained in an info object
     * given a hierarchical key path.<br>
     * The key path is a sequence of keys separated by the KEY_PATH_SEPARATOR
     * string. The first key in the path should be a valid key in the provided
     * info object. The last key in the key path cannot be a key into an info
     * object. If at any point a key is not valid, the key path 
     * terminates without reaching a value or a value is reached without 
     * termination of the key path, null is returned.<br>
     * When a WILD_CARD string is given as a key in the key path, it
     * will match any key at that level in the info object.<br>
     * For example,<br>
     * <p>a key path of <tt>foo/bar/name </tt> will return the result of<br>
     * <tt>info.getInfo("foo")[0].getInfo("bar")[0].getValue("name")
     * </p>
     * @param info the <vaInfo</var> object being looked up
     * @param keyPath the sequence of keys used for the lookup
     * @return the value mapped to the specified key path in this info object
     * or null if there was an invalid key or any other error
     */
    public static String [] deepGetValue(Info info, String keyPath) {
	// parse the key path and get info objects along it
	StringTokenizer tokenizer = new StringTokenizer(keyPath,KEY_PATH_SEPARATOR);
	Info currentInfo = info;
	String [] values = null;
	while(tokenizer.hasMoreTokens()) {
	    String currentKey = (String) tokenizer.nextToken();
	    // if this key is a wild card, it is a match for any key
	    if (currentKey.equals(WILD_CARD)) {
		// get an info object, e.g. the one which maps to the first key
		String [] keys = currentInfo.getKeys();
		if (keys.length == 0) {
		    // no match
		    return null;
		} else {
		    // check if there are any keys left in the key path
		    if (tokenizer.countTokens()==0) {
			// no keys left, get any value and try matching
			// to the value
			values = currentInfo.getValue(keys[0]);
			break;
		    } else {
			// there are still keys left in the path
			// get the first info object in the current 
			// info object and follow it
			boolean gotInfo = false;
			int i=0;
			while(!gotInfo && i<keys.length) {
			    Info [] currentInfoList = currentInfo.getInfo(keys[i++]);
			    if (currentInfoList.length != 0) {
				// got a valid list of info objects
				gotInfo= true;
				// look for a match with only the first one
				// mapped under this key
				currentInfo = currentInfoList[0];
			    }
			}
			if (!gotInfo) {
			    // no info objects under current
			    return null;
			}
		    }
		}
	    } else {
		// current key is not a wild card
		if (tokenizer.countTokens() == 0) {
		    // this is the last token, i.e. it is a key
		    // to a value
		    values = currentInfo.getValue(currentKey);
		    break;
		} else {
		    // this is not the last token
		    // get the info object corresponding to the given key
		    Info [] currentInfoList = currentInfo.getInfo(currentKey);
		    if (currentInfoList.length == 0) {
			// no match
			return null;
		    } else {
			// next key will be found in this info object
			// ignore info objects other than the first
			currentInfo = currentInfoList[0];
		    }
		}
	    }
	}
	// values is the list of values to be returned
	return values;
    }

    /**
     * Return a pretty string format of the given info object
     * @param info the info object to be prettied
     */
    public static String prettyToString(Info info) {
	StringBuffer sb = new StringBuffer();
	return levelToPrettyString(sb,info,0);
    }

    private static String levelToPrettyString(StringBuffer buffer,
					      Info info, int levelIndent) {
	// go through this level
	String [] keys = info.getKeys();
	for (int i=0; i<keys.length; i++) {
	    String currentKey = keys[i];
	    // append key indent
	    for (int j=0; j<levelIndent; j++) {
		buffer.append(" ");
	    }
	    // append key string
	    buffer.append(currentKey);
	    buffer.append(":\n");
	    // try to get a value from this key
	    String [] value = info.getValue(currentKey);
	    if (value == null) {
		// get info from this key
		Info [] currentInfo = info.getInfo(currentKey);
		// print this level of info
		for (int j=0; j<currentInfo.length; j++)
		    levelToPrettyString(buffer,currentInfo[j],levelIndent+KEY_INDENT);
	    } else {
		// non null value
		for (int j=0; j<value.length; j++) {
		    // append value indent (= level indent + KEY_INDENT)
		    for (int k=0; k<(levelIndent+KEY_INDENT); k++) {
			buffer.append(" ");
		    }
		    // append value
		    buffer.append(value[j]);
		    buffer.append("\n");
		}
	    }
	}
	return buffer.toString();
    }
}
