/**
 * This class contains a set of XML utility methods.
 * @author Nirmal (nmukhi)
 */
package util;

import com.ms.xml.om.Document;
import com.ms.xml.om.Element;
import com.ms.xml.om.ElementFactory;
import com.ms.xml.parser.ParseException;
import com.ms.xml.util.XMLOutputStream;
import com.ms.xml.util.Name;
import java.util.Enumeration;
import java.io.File;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.DataInputStream;
import java.io.FileInputStream;
import java.io.EOFException;
import java.io.StringWriter;

public class GeneralXMLUtils {
    /**
     * the type of the ID attribute
     */
    private static String ID = "ID";

    /**
     * the type of the IDREF attribute
     */
    private static String IDREF = "IDREF";

    
    /**
     * This method takes some XML information and converts it
     * into an <var>Info</var> object containing the same information.
     * Tag names are used as keys into fields in the <var>Info</var> 
     * object, which may be nested. 
     * @param xmlInformation the XML information to convert to an 
     * <var>Info</var> object
     * @return an <var>Info</var> object containing the XML 
     * information
     */
    public static InfoImplementation getInfoFromXML(String xmlInformation) {
	// load the document, return null in case
	// of a parse error
	Document xmlDoc = new Document();
	try {
	    xmlDoc.load(new ByteArrayInputStream(xmlInformation.getBytes()));
	} catch (ParseException exception) {
	    System.out.println("Got parse exception "+exception);
	    return null;
	}
	// get the root element of this document
	Element root = xmlDoc.getRoot();
	// create an empty info object for the 
	// root element
	InfoImplementation info = new InfoImplementation();
	// get the info for this element
	addInfoForElement(root,info);
	// return the info object
	return info;
    }

    /**
     * This creates an Info object from an element and recursively does that 
     * for its children. A hashtable is used to maintain a mapping from IDs
     * to Info objects so that internal references can be resolved.
     */
    private static void addInfoForElement(Element element, 
					  InfoImplementation info) {
	int numChildren = element.numElements();
	if (numChildren == 1 && element.getChild(0).getType()==Element.PCDATA) {
	    // if this element has only data, put this element's
	    // tag name and text into the info object
	    // assume this has no ID
	    info.putValue(element.getTagName().getName(),element.getText());
	} else {
	    // if this element has children, put this element's tag 
	    // name and a new Info object into the info object to 
	    // be returned
	    InfoImplementation newInfo = new InfoImplementation();
	    // go through the children
	    Enumeration children = element.getElements();
	    while(children.hasMoreElements()) {
		// get the info for each child and put it in this new
		// info table
		Element currentChild = (Element)children.nextElement();
		if (currentChild.getType() == Element.ELEMENT) {
		    addInfoForElement(currentChild,newInfo);
		}
	    }
	    info.putValue(element.getTagName().getName(),newInfo);
	}
    }

    /**
     * This method creates a level in the XML tree from the current info
     * object and then operates on the children of the current Info object
     * to create deeper levels
     */
    public static void createXMLLevelFromInfo(ElementFactory factory, 
					      Info currentInfo, Element parent) {
	// get all the keys at the current level
	// and create elements from them
	String [] keys = currentInfo.getKeys();
	for (int i=0; i<keys.length; i++) {
	    String currentKey = keys[i];
	    // assume this key points to a list of values
	    // get the values for this key
	    String [] keyValue = currentInfo.getValue(currentKey);
	    if (keyValue == null) {
		// assume this key points to a list of info objects
		Info [] keyInfo = currentInfo.getInfo(currentKey);
		// we need to create as many tags as there are values
		// and then operate on each of thier children
		for (int j=0; j<keyInfo.length; j++) {
		    // create an element whose name is the same as the key
		    Element currentParent = factory.createElement(parent,Element.ELEMENT,
								  Name.create(currentKey),
								  null);
		    // work on the children of this info object
		    createXMLLevelFromInfo(factory,keyInfo[j],currentParent);
		}
	    } else {
		// we need to create as many tags as there are values
		for (int j=0; j<keyValue.length; j++) {
		    // create an element whose name is the same as the key
		    // and whose text is the same as the value
		    Element currentParent = factory.createElement(parent,Element.ELEMENT,
								  Name.create(currentKey),
								  null);
		    factory.createElement(currentParent,Element.PCDATA,
					  null,keyValue[j]);
		}
	    }
	}
	// done creating the XML from the provided info object
    }

    public static void main(String [] args) {
	// argv[0] is the XML file to be converted to an Info
	// open the file
	try {
	    File xmlFile = new File(args[0]);
	    // read stuff into a string
	    FileInputStream inStream = new FileInputStream(xmlFile);
	    ByteArrayOutputStream outStream = new ByteArrayOutputStream();
	    int currentByte;
	    currentByte = inStream.read();
	    while(currentByte != -1) {
		// put the current char into the byte array
		outStream.write(currentByte);
		currentByte = inStream.read();
	    }
	    // parse into an info object
	    Info infoObj = GeneralXMLUtils.getInfoFromXML(outStream.toString());
	    // print it out
	    System.out.println("Got\n"+infoObj);
	} catch (Exception exception) {
	    System.err.println("Got exception "+exception);
	    System.exit(1);
	}
    }
}

/**
 * tiny class to represent a named info object used in internal linking
 */
class NamedInfoImplementation extends InfoImplementation {
	private String name;
    
    public NamedInfoImplementation(String name,InfoImplementation info) {
	    super(info);
	    this.name = name;
    }
    
    // access the name
    public String getName() {
	return name;
    }	
}
