/**
 * This is an implementation of the <a href="Info">Info</a>
 * interface that allows access to name/value pairs and name/list
 * pairs
 * @author Nirmal (nmukhi)
 *
 * modified by Madhuri 
 * Iterator class is not present in jdk1.1
 *
 */

package util;

import java.util.Hashtable;
import java.util.Vector;
import java.util.Enumeration;
import java.util.Iterator;

public class InfoImplementation implements Info {
    private Hashtable information;
    Vector orderedKeys;

    /** 
     * Creates an info object with no data.<br>
     */
    public InfoImplementation() {
	information = new Hashtable();
	orderedKeys = new Vector();
    }

    /**
     * Creates an info object from the given 
     * <var>InfoImplementation</var> object
     * @param information the <var>InstanceInfo</var> object which 
     * is the source of the information.
     */
    public InfoImplementation(InfoImplementation info) {
	orderedKeys = info.getOrderedKeys();
	information = info.getInfoTable();
    }

    /**
     * Adds an item of data to the information maintained here.
     * If the key has previously been used to store an <var>Info</var
     * item using <a href="#putValue(String,Info)">putValue(String,Info)</a>
     * this method call will do nothing
     * @param key the key to be used to access this data
     * @param value the data to be stored
     */
    public void putValue(String key, String value) {
	// access the information stored with this key
	Vector infoStored = (Vector)information.get(key);
	if (infoStored == null) {
	    // nothing stored so far, create a new vector
	    // and put it in the table
	    Vector newInfo = new Vector();
	    newInfo.addElement(value);
	    information.put(key,newInfo);
	    orderedKeys.addElement(key);
	} else {
	    // something stored, make sure it is a vector
	    // of Strings and not something else
	    try {
		String objectStored = (String) infoStored.firstElement();
	    } catch (ClassCastException exception) {
		// not a vector of strings, just return
		return;
	    }
	    // ok, we have a vector of strings, add this one to 
	    // the list
	    infoStored.addElement(value);
	}
    }

    /**
     * Adds an Info object to the information maintained 
     * here. Semantics are the same as 
     * <a href="#putValue(String,String)">putValue(String,String)</a>.
     * @param key the key to be used to access this data
     * @param value the data to be stored
     */
    public void putValue(String key, Info value) {
	// access the information stored with this key
	Vector infoStored = (Vector)information.get(key);
	if (infoStored == null) {
	    // nothing stored so far, create a new vector
	    // and put it in the table
	    Vector newInfo = new Vector();
	    newInfo.addElement(value);
	    information.put(key,newInfo);
	    orderedKeys.addElement(key);
	} else {
	    // something stored, make sure it is a vector
	    // of Info objects and not something else
	    try {
		Info objectStored = (Info) infoStored.firstElement();
	    } catch (ClassCastException exception) {
		// not a vector of info objects, just return
		return;
	    }
	    // ok, we have a vector of info ojects, add this one to 
	    // the list
	    infoStored.addElement(value);
	}
    }
	    
    /**
     * Gets the set of values mapped to the supplied key
     * @param key the key
     * @return the values corresponding to <var>key</var>
     * or null if the key did not map to any string
     */
    public String [] getValue(String key) {
	// get the data
	Object value = information.get(key);
	if (value == null) {
	    // nothing matching this key
	    return null;
	}
	try {
	    Vector stringList = (Vector) value;
	    // convert the vector of strings into an array
	    String [] data = new String[stringList.size()];
	    for (int i=0; i<stringList.size(); i++) {
		data[i] = (String)stringList.elementAt(i);
	    }
	    return data;
	} catch (ClassCastException exception) {
	    // some data in the vector wasn't a string
	}
	return null;
    }

    /**
     * Gets the <var>Info</var> objects corresponding to
     * the given key. This will be a list containing only
     * one Info object unless the same key was used multiple 
     * times when the data was filled into the <var>Info</var>
     * object.
     * @param key the key
     * @return the list of <var>Info</var> objects 
     * that <var>key</var> maps to, or an array of size zero
     *  if this key did not map to a list of <var>Info</var> 
     * objects
     */
    public Info [] getInfo(String key) {
	// get the data
	Object value = information.get(key);
	if (value == null) {
	    // nothing matching this key
	    return new Info[0];
	}
	try {
	    Vector infoList = (Vector) value;
	    // convert the vector of info object into an array
	    Info [] info = new Info[infoList.size()];
	    for (int i=0; i<infoList.size(); i++) {
		info[i] = (Info)infoList.elementAt(i);
	    }
	    // return the array of info objects
	    return info;
	} catch (ClassCastException exception) {
	    // something in the vector wasn't an Info object
	}
	return new Info[0];
    }

    /**
     * Gets the keys used to access information 
     * stored in this <var>Info</var> object
     * @return a list of all the keys mapped to
     * anything in this object
     */
    public String [] getKeys() {
	// Iterator keyIterator = orderedKeys.iterator();
	String [] keys = new String[orderedKeys.size()];
	int i=0;
        for (i=0; i<orderedKeys.size(); i++) {
	// while(keyIterator.hasNext()) {
	//    keys[i++] = (String) keyIterator.next();
	    keys[i] = (String) orderedKeys.elementAt(i);
	}
	return keys;
    }
	
    /**
     * Convert this info object into a suitable string representation.
     */
    public String toString() {
	return information.toString();
    }

    /**
     * Accesses the hash table contianing the information
     * @return the hash table containing the information
     */
    private Hashtable getInfoTable() {
	return information;
    }

    /**
     * Accesses the list of ordered keys in the table
     * @return the vector containing the set of ordered keys
     */
    private Vector getOrderedKeys() {
	return orderedKeys;
    }

    public static void main(String [] args) {
	InfoImplementation info = new InfoImplementation();
	info.putValue("foo","1");
	info.putValue("bar","2");
	info.putValue("bar","3");
	System.out.println("Contents are "+info);
    }
}
