/**
 * This is the interface used by classes that possesss a connection to a DAV server
 * and wish to make requests to the server.
 * It encapsulates simple DAV requests that can be made to the server. Note that
 * this does not fully utilize the DAV capabilities, but exposes only the parts 
 * required for the current application. This can easily be extended later to 
 * act as an interface for a fully functional DAV client.
 * Currently this interface allows applications to issue GET, PUT, MKCOL 
 * and DELETE requests to the server. Authentication and other headers are
 * not supported. The response from the server is returned in the form of a 
 * <var>DAVServerResponse</var> object containing an indication of success
 * or failure and the data (if any). Note that failure of the response could
 * be due to many reasons such as invalid authentication, a network error, etc.
 * The only error that is forwarded to the application is absence of a 
 * requested resource; this is communicated to the application via a <var>DAVResourceNotFound</var> 
 * exception.
 * @see DAVConnectionFactory
 * @see DAVServerResponse
 * @see DAVResourceNotFoundException
 * @author Nirmal (nmukhi)
 */

package util.webdav.client;

public interface DAVConnection {
    /**
     * Issues a GET request to the DAV server asking for the specified resource.
     * This returns a response which will indicate success or failure.
     * @param resourceName the name of the resource to get
     * @return a response object containing the server's response
     * @exception DAVResourceNotFoundException thrown if the requested resource is 
     * not present on the server.
     * @see DAVServerResponse
     *
     */
    public DAVServerResponse get(String resourceName) throws DAVResourceNotFoundException;

    /**
     * Issues a PUT request to the DAV server asking that the provided data be 
     * stored on the DAV server in the named resource. 
     * @param resourceName the name of the resource to be created at the server end.
     * @param data the data to be stored in the resource
     * @return the response from the server.
     * @exception DAVResourceNotFoundException thrown if the initial part of the path
     * in <var>resourceName</var> does not refer to a valid resource.
     * @see DAVServerResponse
     *
     */
    public DAVServerResponse put(String resourceName, String data) throws DAVResourceNotFoundException;

    /**
     * Issues a MKCOL request to the DAV server asking that the named resource 
     * be created at the server. 
     * @param resourceName the name of the resource to be created
     * @return the response from the server
     * @exception DAVResourceNotFoundException thrown if the initial part of the path
     * in <var>resourceName</var> does not refer to a valid resource.
     * @see DAVServerResponse
     *
     */
    public DAVServerResponse makeCollection(String resourceName) throws DAVResourceNotFoundException;

    /**
     * Issues a DELETE request to the DAV server asking that the named resource
     * be removed form the server.
     * @param resourceName the name of the resource to be deleted
     * @return the response from the server
     * @exception DAVResourceNotFoundException thrown if the <var>resourceName</var>
     * does not refer to a valid resource on the server.
     * @see DAVServerResponse
     *
     */
    public DAVServerResponse delete(String resourceName) throws DAVResourceNotFoundException;
}
