/**
 * This class extends the basic HTTP 1.1 compliant connection, 
 * <var>HTTPCOnnection</var> to allow additional DAV requests to be made.
 * It also obeys the <var>DAVConnection</var> interface that allows
 * simple requests to be made to the DAV server, hiding most server
 * errors and other details.
 * @author Nirmal (nmukhi)
 *
 */

package util.webdav.client;

import HTTPClient.HTTPConnection;
import HTTPClient.HTTPResponse;
import HTTPClient.ModuleException;
import java.io.IOException;

public class DAVConnectionImplementation extends HTTPConnection implements DAVConnection {

    /**
     * Creates a DAV connection to the specified server at the 
     * specified port.
     *
     */
    public DAVConnectionImplementation(String host, int port) {
	super(host,port);
    }

    /**
     * Issues a GET request to the DAV server asking for the specified resource.
     * This returns a response which will indicate success or failure.
     * @param resourceName the name of the resource to get
     * @return a response object containing the server's response
     * @exception DAVResourceNotFoundException thrown if the requested resource is 
     * not present on the server.
     * @see DAVServerResponseImplementation
     *
     */
    public DAVServerResponse get(String resourceName) 
	throws DAVResourceNotFoundException {
	HTTPResponse response = null;
	// make the request
	try {
	    response = Get(resourceName);
	    if (response != null) {
		// request was made and a response received successfully
		// check if the resource was not found (error code 404 or 410 in HTTP 1.1)
		int statusCode = response.getStatusCode();
		if (statusCode == 404 || statusCode == 410) {
		    // resource was not found
		    throw new DAVResourceNotFoundException
			("The server could not find the resource "+resourceName+", get failed");
		}
		// create a normal response
		return new DAVServerResponseImplementation(response);
	    }
	} catch (IOException exception) {
	    System.err.println("ERROR >> DAV exception while getting response: "+exception);
	    throw new DAVResourceNotFoundException();
	} catch (ModuleException exception) {
	    System.err.println("ERROR >> DAV exception while getting response: "+exception);
	    throw new DAVResourceNotFoundException();
	}
	// response returned was null or the attempt to get generated an exception somewhere
	// return default response, which indicates a failure took place
	return new DAVServerResponseImplementation();
    }
	    

    /**
     * Issues a PUT request to the DAV server asking that the provided data be 
     * stored on the DAV server in the named resource. 
     * @param resourceName the name of the resource to be created at the server end.
     * @param data the data to be stored in the resource
     * @return the response from the server.
     * @exception DAVResourceNotFoundException thrown if the initial part of the path
     * in <var>resourceName</var> does not refer to a valid resource.
     * @see DAVServerResponseImplementation
     *
     */
    public DAVServerResponse put(String resourceName, String data) 
	throws DAVResourceNotFoundException {
	HTTPResponse response = null;
	// make the request
	try {
	    response = Put(resourceName, data);
	    if (response != null) {
		// request was made and a response received successfully
		// check if the resource was not found (error code 404 or 410 in HTTP 1.1)
		int statusCode = response.getStatusCode();
		if (statusCode == 404 || statusCode == 410) {
		    // resource was not found
		    throw new DAVResourceNotFoundException
			("The server could not find the resource "+resourceName+", put failed");
		}
		// create a normal response
		return new DAVServerResponseImplementation(response);
	    }
	} catch (IOException exception) {
	    System.err.println("ERROR >> DAV exception while getting response: "+exception);
	} catch (ModuleException exception) {
	    System.err.println("ERROR >> DAV exception while getting response: "+exception);
	}
	// response returned was null or the attempt to get generated an exception somewhere
	// return default response, which indicates a failure took place
	return new DAVServerResponseImplementation();
    }


    /**
     * Issues a MKCOL request to the DAV server asking that the named resource 
     * be created at the server. 
     * @param resourceName the name of the resource to be created
     * @return the response from the server
     * @exception DAVResourceNotFoundException thrown if the initial part of the path
     * in <var>resourceName</var> does not refer to a valid resource.
     * @see DAVServerResponseImplementation
     *
     */
    public DAVServerResponse makeCollection(String resourceName) 
	throws DAVResourceNotFoundException {
	HTTPResponse response = null;
	// make the request
	try {
	    response = ExtensionMethod("MKCOL",resourceName,(byte [])null,null);
	    if (response != null) {
		// request was made and a response received successfully
		// check if the resource was not found (error code 404 or 410 in HTTP 1.1)
		int statusCode = response.getStatusCode();
		if (statusCode == 409) {
		    // resource was not found
		    throw new DAVResourceNotFoundException
			("The server could not find some part of the prefix of the resource "+resourceName+", makeCollection failed");
		}
		// create a normal response
		return new DAVServerResponseImplementation(response);
	    }
	} catch (IOException exception) {
	    System.err.println("ERROR >> DAV exception while getting response: "+exception);
	} catch (ModuleException exception) {
	    System.err.println("ERROR >> DAV exception while getting response: "+exception);
	}
	// response returned was null or the attempt to get generated an exception somewhere
	// return default response, which indicates a failure took place
	return new DAVServerResponseImplementation();
    }

    /**
     * Issues a DELETE request to the DAV server asking that the named resource
     * be removed from the server.
     * @param resourceName the name of the resource to be deleted
     * @return the response from the server
     * @exception DAVResourceNotFoundException thrown if the <var>resourceName</var>
     * does not refer to a valid resource on the server.
     * @see DAVServerResponseImplementation
     *
     */
    public DAVServerResponse delete(String resourceName) 
	throws DAVResourceNotFoundException {
	HTTPResponse response = null;
	// make the request
	try {
	    response = Delete(resourceName);
	    if (response != null) {
		// request was made and a response received successfully
		// check if the resource was not found (error code 404 or 410 in HTTP 1.1)
		int statusCode = response.getStatusCode();
		if (statusCode == 404 || statusCode == 410) {
		    // resource was not found
		    throw new DAVResourceNotFoundException
			("The server could not find the resource "+resourceName+", delete failed");
		}
		// create a normal response
		return new DAVServerResponseImplementation(response);
	    }
	} catch (IOException exception) {
	    System.err.println("ERROR >> DAV exception while getting response: "+exception);
	} catch (ModuleException exception) {
	    System.err.println("ERROR >> DAV exception while getting response: "+exception);
	}
	// response returned was null or the attempt to get generated an exception somewhere
	// return default response, which indicates a failure took place
	return new DAVServerResponseImplementation();
    }

    /**
     * Tests the DAVConnectionImplementation by issuing various requests
     *
     */
    public static void main(String [] args) {
	// argv[0] is the name of the server
	// argv[1] is the port on which the server is listening
	// create the connection
	System.out.println("Connecting to "+args[0]+" at port "+args[1]);
	int port = new Integer(args[1]).intValue();
	DAVConnection connection = DAVConnectionFactory.createConnection(args[0],port);
	int choice = 0;
	// present the menu
	while (choice != 9) {
	    System.out.print("Options\n1. GET\n2. PUT\n3. MKCOL\n4. DELETE\n5. get children\n6. get all children\n7. create path\n8. strong delete\n9. quit\n\nEnter choice: ");
	    try {
		byte [] b = new byte [64]; // ought to be enough
		choice = System.in.read()-48;
		System.out.println("Choice is "+choice);
		if (choice == -1)
		    break;
		System.out.print("\nEnter resource name: ");
		System.in.skip(System.in.available());
		int size = System.in.read(b);
		String resourceName = new String(b,0,size-1);
		DAVServerResponse response = null;
		// switch on choice
		// ugh! (but this is just a test, switch is ok)
		if (choice == 1) {
		    // get the specified resource
		    response = connection.get(resourceName);
		} else if (choice == 2) {
		    // put data into the specified resource
		    // get the data
		    System.out.print("\nEnter data to be stored: ");
		    size = System.in.read(b);
		    String data = new String(b,0,size);
		    response = connection.put(resourceName, data);
		} else if (choice == 3) {
		    // create the specified resource
		    response = connection.makeCollection(resourceName);
		} else if (choice == 4) {
		    // delete the specified resource
		    response = connection.delete(resourceName);
		} else if (choice == 5) {
		    // get the children of the specified resource
		    String [] children = DAVUtilityCalls.getChildren(connection,resourceName);
		    if (children == null) {
			System.out.println("Either "+resourceName+" did not correspond to a collection, or a network error occurred ");
		    } else {
			System.out.println("Children of "+resourceName+" are: ");
			for (int i=0; i<children.length; i++) {
			    System.out.println(children[i]);
			}
		    }
		} else if (choice == 6) {
		    // get the children of the specified resource
		    String [] children = DAVUtilityCalls.getAllChildren(connection,resourceName);
		    if (children == null) {
			System.out.println("Either "+resourceName+" did not correspond to a collection, or a network error occurred ");
		    } else {
			System.out.println("All children of "+resourceName+" are: ");
			for (int i=0; i<children.length; i++) {
			    System.out.println(children[i]);
			}
		    }
		} else if (choice == 7) {
		    // create the path starting from the specified resource
		    // get the path to be created
		    System.out.print("\nEnter the path to be created: ");
		    size = System.in.read(b);
		    String path = new String(b,0,size);
		    DAVUtilityCalls.createCollectionPath(connection,path,resourceName);
		    System.out.println("Path created");
		} else if (choice == 8) {
		    // string delete the specified resource
		    DAVUtilityCalls.strongDelete(connection,resourceName);
		    System.out.println("Strong deletion done");
		} else if (choice == 9) 
		    break;
		if (choice >=1 && choice <=4)
		    System.out.println("Response to your request was \n"+response+"\n");
	    } catch (Exception exception) {
		System.err.println("Got exception "+exception);
		System.exit(1);
	    }
	}
    }
}

