/**
 * This class encapsulates a response from a DAV server, hiding DAV
 * or HTTP specific details such as status codes. It thus presents
 * a simple interface to applications communicating with a DAV server.
 * @author Nirmal (nmukhi)
 * @see HTTPClient.HTTPResponse
 *
 */

package util.webdav.client;

import HTTPClient.HTTPResponse;
import HTTPClient.ModuleException;
import java.io.IOException;

public class DAVServerResponseImplementation implements DAVServerResponse {
    private boolean success;
    private String actionDescription;
    private byte [] data;

    /**
     * Creates a default DAVServerResponseImplementation object.
     *
     */
    public DAVServerResponseImplementation() {
	// default values for instance variables
	success = false;
	actionDescription = "Unknown";
	data = null;
    }

    /** 
     * Creates a DAVServerResponseImplementation object from an HTTPResponse.
     * @param response the <var>HTTPResponse</var> object from which this
     * object should be created.
     *
     */
    public DAVServerResponseImplementation(HTTPResponse response) {
	this();
	try {
	    int statusCode = response.getStatusCode();
	    if (statusCode >= 200 && statusCode < 300) {
		// successful request
		success = true;
	    }
	    actionDescription = response.getReasonLine();
	    data = response.getData();
	} catch (IOException exception) {
	    System.err.println("ERROR >> DAV exception while getting response: "+exception);
	} catch (ModuleException exception) {
	    System.err.println("ERROR >> DAV exception while getting response: "+exception);
	}
    }
	
    /**
     * @return true if this response refers to a successful request, 
     * false otherwise
     *
     */
    public boolean isSuccess() {
	return success;
    }
    
    /**
     * @return a string describing the action that the server took
     * for the request.
     *
     */
    public String getActionDescription() {
	return actionDescription;
    }
    
    /**
     * @return the data form the server. Requests that cause
     * the server to return data include GET.
     *
     */
    public byte [] getData() {
	return data;
    }

    /**
     * @return a string representation of the response.
     *
     */
    public String toString() {
	StringBuffer sb = new StringBuffer();
	sb.append("Success: ");
	sb.append(success);
	sb.append("\nDescription: ");
	sb.append(actionDescription);
	sb.append("\nData: ");
	sb.append(new String(data));
	return sb.toString();
    }
}

