/**
 * This class contains a set of utility calls that can be made using
 * a <var>DAVConnection</var>.
 * These are static convenience methods usable by applications
 * wishing to make certain kinds f complex requests to DAV servers.
 * @author Nirmal (nmukhi)
 */
package util.webdav.client;

import java.util.Vector;
import java.util.StringTokenizer;

public class DAVUtilityCalls {
    private static String dirTag = "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 3.2 Final//EN\">\n<HTML>\n <HEAD>\n  <TITLE>Index of ";

    /**
     * Uses the provided connection to get a list of information on all resources 
     * that are immediate children of the given resource
     * @param connection the <var>DAVConnection</var> to be used
     * @param resourceName the resource whose children we are interested 
     * in
     * @return an array of strings representing names of the specified 
     * resource's children or null if the specified <var>resourceName</var> 
     * does not correspond to the name of a collection of resources or if 
     * an error occurred while talking to the server.
     * @exception DAVResourceNotFoundException thrown if the specified resource
     * cannot be found on the server
     */
    public static String [] getChildren(DAVConnection connection, String resourceName) 
	throws DAVResourceNotFoundException {
	String [] children = null;

	// get the specified resource
	DAVServerResponse response = connection.get(resourceName);

	// if the request failed, return null
	if (!response.isSuccess())
	    return null;

	// get the data returned by the server
	String data = new String(response.getData());

	// if this corresponds to a file (not a collection), return null
	// TODO
	// this is not the best way of figuring this out, I'll do something 
	// better later
	if (!describesCollection(data))
	    return null;

	// TODO 
	// this is not the best way of extracting the children's names either
	// will do something better later
	// parse the data contents
	// parse using "=" as the token
	StringTokenizer tokenizer = new StringTokenizer(data,"=");
	// count the number of tokens
	// skip the first two tokens to get past the parent directory
	tokenizer.nextToken();
	tokenizer.nextToken();
	int numChildren = tokenizer.countTokens();
	children = new String[numChildren];
	int currentChild = 0;
	while (tokenizer.hasMoreTokens()) {
	    String currentToken = tokenizer.nextToken();
	    // get the name of the child
	    int fileNameEndChar = currentToken.indexOf('"',1);
	    children[currentChild++] = currentToken.substring(1,fileNameEndChar);
	}
	String [] childrenList = new String[currentChild];
	for (int i=0; i<currentChild; i++) {
	    childrenList[i] = children[i];
	}
	return childrenList;
    }

    /**
     * Uses the provided connection to recursively get all leaves 
     * (i.e. files, not collections) under the specified resource
     * @param connection the <var>DAVConnection</var> to be used
     * @param resourceName the resource below which we want the 
     * leaves
     * @return a list of fully qualified of children below the specified
     * resource, or null if there is an error during processing
     * if the given resource name is not a collection, null will be returned.
     * @exception DAVResourceNotFoundException thrown if the specified resource
     * cannot be found on the server
     *
     */
    public static String [] getAllChildren(DAVConnection connection, String resourceName) 
	throws DAVResourceNotFoundException {
	// do an iterative breadth first search of the tree
	Vector collectionsRemaining = new Vector();
	Vector leaves = new Vector();
	// if the resource name doesn't already have a trailing slash, add one
	if (!resourceName.endsWith("/"))
	    resourceName += "/";
	collectionsRemaining.addElement(resourceName);
	while (collectionsRemaining.size()!=0) {
	    // get the children of the first resource on the list
	    // and remove that element from the list
	    String currentCollection = (String)collectionsRemaining.remove(0);
	    String [] currentChildren = getChildren(connection,currentCollection);
	    // if something weird happened while processing this child, return null
	    // something weird could happen if this child did not correspond to a 
	    // collection (should never happen) or there was a network error 
	    // (should rarely happen)
	    // in such cases, the list of children returned will be null as 
	    // specified in getAllChildren's contract
	    if (currentChildren == null)
		return null;
	    // iterate through the list of these children
	    for (int i=0; i<currentChildren.length; i++) {
		String childName = currentCollection+currentChildren[i];
		DAVServerResponse response = connection.get(childName);
		// if the response failed, just return null
		if (!response.isSuccess())
		    return null;
		// response was successful
		// check if this child is a collection
		String data = new String(response.getData());
		if (describesCollection(data)) {
		    // add this child to the list of collections
		    // make sure it has a trailing slash
		    if (!childName.endsWith("/"))
			childName += "/";
		    collectionsRemaining.addElement(childName);
		} else {
		    // add this child to the list of leaves
		    leaves.addElement(childName);
		}
	    }
	}
	// create an array with the leaf names
	String [] leafNames = new String[leaves.size()];
	for (int i=0; i<leaves.size(); i++) {
	    leafNames[i] = (String) leaves.elementAt(i);
	}
	return leafNames;
    }

    /**
     * Creates the specifed path on the server if it doesn't already exist
     * If part of the path already exists, additional collections will
     * be created so that the specified path is valid after the process
     * @param path the path to be created
     * @param startOfPath the place where the path creation should 
     * begin from
     * @connection the DAV connection to be used for this operation
     * @exception DAVResourceNotFoundException exception thrown if the
     * <var>startOfPath</var> did not exist, or there was some network 
     * error or unexpected DAV error
     * @see DAVConnection#makeCollection
     */
    public static void createCollectionPath(DAVConnection connection,String path,
					    String startOfPath) 
	throws DAVResourceNotFoundException {
	// if there is a leading slash, skip it
	if (path.startsWith("/"))
	    path = path.substring(1);
	// break the path up into collections
	StringTokenizer tokenizer = new StringTokenizer(path,"/");
	// iterate through the collections to be created and
	// create each one
	while(tokenizer.hasMoreTokens()) {
	    String currentCollection = tokenizer.nextToken();
	    String collectionName = startOfPath+"/"+currentCollection;
	    // create this collection
	    connection.makeCollection(collectionName);
	    // add on this collection to the path of created collections
	    startOfPath=collectionName;
	}
    }

    /**
     * Deletes the specified resource and all intermediate 
     * collections from the server. This does not delete 
     * intermediate collections if they contain other resources
     * too.
     * @param resourceName the resource to be deleted
     * @param connection the DAV connection to be used for this 
     * operation
     * @exception DAVResourceNotFoundException thrown if some or all
     * parts of the specified resource do not exist on the server 
     */
    public static void strongDelete(DAVConnection connection,String resourceName) 
	throws DAVResourceNotFoundException {
	// if this resource is not a collection, remove it without 
	// other checks
	String collectionName = resourceName;
	if (!isCollection(connection,resourceName)) {
	    connection.delete(resourceName);
	    int lastResourceNameIndex = resourceName.lastIndexOf("/");
	    collectionName = resourceName.substring(0,lastResourceNameIndex);
	}
	// now check if the collection and its parents can be deleted 
	while(!collectionName.equals("/")) {
	    // check if this collection has any children
	    String [] children = getChildren(connection,collectionName);
	    // if there are children, stop the deletion process right here
	    if (children.length != 0)
		break;
	    // no children, so delete this collection
	    connection.delete(collectionName);
	    // check this collection's parent
	    int lastResourceNameIndex = collectionName.lastIndexOf("/");
	    if (lastResourceNameIndex == -1) {
		// we already deleted the first collection in the 
		// original resource name
		break;
	    } else {
		collectionName = collectionName.substring(0,lastResourceNameIndex);
	    }
	}
    }

    /**
     * Checks whether the specified resource corresponds to a collection
     * or not
     * @param resourceName the name of the resource
     * @return true if this corresponds to a collection, false otherwise
     * if <var>resourceName</var> does not correspond to a valid resource
     * or a network error occurred, false will be returned.
     */
    public static boolean isCollection(DAVConnection connection, String resourceName) {
	try {
	    DAVServerResponse response = connection.get(resourceName);
	    // if there was any error, return false
	    if (!response.isSuccess()) {
		return false;
	    }
	    // we made a successful request
	    // get the data corresponding to the resource
	    String data = new String(response.getData());
	    // check if this corresponds to a collection
	    return describesCollection(data);
	} catch (Exception exception) {
	    // network or DAV exception
	    return false;
	}
    }

    /**
     * Checks whether the specified data describes a collection or not
     * @param data data from a DAV resource
     * @return true if this describes a collection, false otherwise
     */
    public static boolean describesCollection(String collectionData) {
	return collectionData.startsWith(dirTag);
    }
}
