
package nexus;

import java.util.Hashtable;
import java.util.Vector;
import java.util.Enumeration;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.net.URL;

public class Nexus extends Object
{
    /*
     * Constants
     */
    public static final int BUFFER_VERSION = 0;
    public static final int MI_PROTO_HEADER_VERSION = 0;
    
    public static final int PROTO_TYPE_LOCAL = 0;
    public static final int PROTO_TYPE_TCP   = 1;
    public static final int PROTO_TYPE_UDP   = 9;
    public static final int PROTO_TYPE_TOTEM = 12;
    public static final int PROTO_TYPE_DEFAULT = PROTO_TYPE_TCP;

    /*
     * Class statics
     */
    private static int debug_level = 0;

    /*
     * Instance variables
     */
    private boolean				destroying;

    private String				node_name;
    private int					context;
    
    private PackageCallbackInterface		callback_object;
    
    private Vector				thread_vector;
    private boolean				thread_vector_done;

    private long				next_endpoint_id;
    private Hashtable				endpoint_id_hashtable;

    private LocalProto				local_proto;
    private TcpProto				tcp_proto;
    private UdpProto				udp_proto;
    private TotemProto				totem_proto;

    /*
     * debug_println()
     */
    public static
    void
    debug_println(int level, String s)
    {
	if (level <= debug_level)
	{
	    System.out.println(s);
	}
    }
    /* debug_println() */
    

    /*
     * debug_array_println()
     */
    public static
    void
    debug_array_println(
	int					level,
	String					prefix,
	Object					a[])
    {
	if (level <= debug_level)
	{
	    Nexus.array_println(prefix, a);
	}
    }
    
    public static
    void
    debug_array_println(
	int					level,
	Object					a[])
    {
	if (level <= debug_level)
	{
	    Nexus.array_println(a);
	}
    }
    /* debug_array_println() */
	 
    
    /*
     * array_println()
     */
    public static
    void
    array_println(
	String					prefix,
	Object					a[])
    {
	System.out.print(prefix);
	array_println(a);
    }

    public static
    void
    array_println(
	Object					a[])
    {
	System.out.print("[");
	for (int i = 0; i < a.length; i++)
	{
	    System.out.print(a[i]);
	    if (i != a.length - 1)
	    {
		System.out.print(",");
	    }
	}
	System.out.println("]");
    }
    /* array_println() */
    
    
    /*
     * Nexus()
     */
    public
    Nexus()
    {
	/*
	 * Initialize internal state
	 */
	destroying = false;
	thread_vector_done = false;
	
    }
    /* Nexus() */

    /*
     * activate()
     */
    public
    boolean
    activate()
    {
	/*
	 * Get interesting properties from JRE and set associated internal
	 * state
	 */
	String					prop_val;

	prop_val = System.getProperty("nexus.debug_level", "0");
	try
	{
	    debug_level = Integer.parseInt(prop_val);
	}
	catch (Exception e)
	{
	    debug_level = 0;
	}
	
	/* Initialize my node and context information */
	try
	{
	    node_name = InetAddress.getLocalHost().getHostName();
	}
	catch (UnknownHostException e)
	{
	    node_name = new String("localhost");
	}
	context = 0;

	/*
	 * Initialize the endpoint id hashtable.
	 * This is used to assign unique integer id's to endpoints,
	 * so that the ids can later be used to map back to
	 * an endpoint.
	 */
	endpoint_id_hashtable = new Hashtable();
	next_endpoint_id = 0;

	/*
	 * Initialize the thread vector
	 * This keeps a complete record of currently
	 * active nexus threads.
	 */
	thread_vector = new Vector(5, 5);
    
	/*
	 * Initialize the various communication modules
	 */
	local_proto = new LocalProto(this);
	tcp_proto = new TcpProto(this);

	return true;
    }
    /* activate() */

    /*
     * deactivate()
     */
    public
    boolean
    deactivate()
    {
	destroying = true;
	local_proto.shutdown();
	tcp_proto.shutdown();
	try { udp_proto.shutdown(); }
	catch (NullPointerException nullX) {}
	try { totem_proto.shutdown(); }
	catch (NullPointerException nullX) {}

	Nexus.debug_println(2, "Nexus.destroy_current_context(): " +
			    "Waiting for threads");

	wait_for_all_threads();

	Nexus.debug_println(2, "Nexus.destroy_current_context(): done");

	return true;
    }

    /*
     * finalize()
     */
    protected
    void
    finalize()
    {
	if (!destroying)
	{
	    deactivate();
	}
    }
    /* finalize() */


    /*
     * node_name()
     */
    public
    String
    node_name()
    {
	return (this.node_name);
    }
    /* node_name() */


    /*
     * destroy_current_context()
     */
    public
    void
    destroy_current_context(
	boolean					non_threaded_handler)
    {
    }
    /* destroy_current_context() */


    /*
     * is_destroying()
     */
    public
    boolean
    is_destroying()
    {
	return (destroying);
    }
    /* is_destroying() */
    
    
    /*
     * add_thread()
     */
    public synchronized
    void
    add_thread(
	Thread					t)
    {
	thread_vector.addElement(t);
    }
    /* add_thread() */
   

    /*
     * remove_thread()
     */
    public synchronized
    void
    remove_thread(
	Thread					t)
    {
	/* removeElement() will change the Enumeration of that Vector */
	if (!thread_vector_done)
	{
	    thread_vector.removeElement(t);
	}
    }
    /* add_thread() */
   

    /*
     * get_final_threads()
     */
    private synchronized
    Enumeration
    get_final_threads()
    {
	thread_vector_done = true;
	return (thread_vector.elements());
    }
    /* get_final_threads() */

    
    /*
     * wait_for_all_threads()
     */
    public
    void
    wait_for_all_threads()
    {
	Thread t;
	boolean interrupted;

	for (Enumeration e = get_final_threads(); e.hasMoreElements(); )
	{
	    t = (Thread) e.nextElement();

	    Nexus.debug_println(2, "Nexus.wait_for_all_threads(): " + 
				"Waiting for thread: "+t.getName());
	    do
	    {
		interrupted = false;
		try { t.join(); }
		catch (InterruptedException intX) { interrupted = true; }
	    } while (interrupted);
  
	    Nexus.debug_println(2, "Nexus.wait_for_all_threads(): " +
				"Done waiting for thread: "+t.getName());
	}
    }
    /* wait_for_all_threads() */
  
	 
    /*
     * data_format()
     *
     * Return my Nexus data format
     */
    public
    int
    data_format()
    {
	return (DataConversion.NEXUS_DC_FORMAT_JAVA); 
    }
    /* data_format() */
    

    /*
     * get_context()
     *
     * Return my context
     */
    public
    int
    context()
    {
	return (this.context);
    }
    /* context() */
    

    /*
     * create_comm_descriptor_table()
     */
    public
    byte[]
    create_comm_descriptor_table(
	ProtoInfo				protoInfo,
	long					endpointID)
    {
	byte[] comm_descriptor_table;
	byte[] comm_descriptor;
	int i;
	
	Nexus.debug_println(2,
			    "Nexus.create_comm_descriptor_table(): " +
			    "proto type = " + protoInfo.protoType);
	
	if (protoInfo.protoType == PROTO_TYPE_TCP)
	{
	    comm_descriptor =
		tcp_proto.get_my_mi_proto(new TcpInfo(),
					  ProtoInterface.USE_BUFFER_EP_ID);
	}
	else if (protoInfo.protoType == PROTO_TYPE_UDP)
	{
	    /* Create a comm_descriptor_table */
	    if (udp_proto == null)
	    {
		udp_proto = new UdpProto(this);
	    }

	    comm_descriptor =
		udp_proto.get_my_mi_proto(protoInfo, endpointID);

	}
	else if (protoInfo.protoType == PROTO_TYPE_TOTEM)
	{
	    /* Create a comm_descriptor_table */
	    if (totem_proto == null) totem_proto = new TotemProto(this);

	    comm_descriptor = 
		totem_proto.get_my_mi_proto(protoInfo, endpointID);
	}
	else
	{
	    Nexus.debug_println(1,
				"Nexus.create_comm_descriptor_table(): " +
				"invalid communications protocol");
	    return null;
	}

	if (comm_descriptor == null)
	{
	    Nexus.debug_println(1,
				"Nexus.create_comm_descriptor_table(): " +
				"get_my_mi_proto failed for protocol=" +
				protoInfo.protoType);
	    return null;
	}

	comm_descriptor_table =
	    new byte[(1 + 4 + node_name.length() + 1 +
		      4 + comm_descriptor.length)];

	/* Note that comm_descriptor_table is always encoded in 32bit BE.  Java
	   format "puts" (the default used by DataConversion.put() ) are
	   acceptable because 32BE and Java have the same-size and format for
	   shorts and ints */
    
	i = 0;

	    /* pack in the version number */
	comm_descriptor_table[i] = MI_PROTO_HEADER_VERSION;
	i += 1;
	    
	    /* pack in the context differentiator */
	DataConversion.put_int(0, comm_descriptor_table, i);
	i += 4;
     
	    /* pack in name as 1-byte chars w/null appended */
	byte b[] = node_name.getBytes();
	System.arraycopy(b, 0, comm_descriptor_table, i, b.length);
	i += b.length;
	comm_descriptor_table[i++] = 0;
    
	    // pack in the udp_proto comm descriptor 
	DataConversion.put_short((short) protoInfo.protoType,
				 comm_descriptor_table, i);
	i += 2;
	DataConversion.put_short((short)comm_descriptor.length,
				 comm_descriptor_table, i);
	i += 2;
	DataConversion.put_byte(comm_descriptor, 0,
				comm_descriptor.length,
				comm_descriptor_table, i);
	
	return comm_descriptor_table;
    }
    /* create_comm_descriptor_table() */

    /*
     * outgoing_instantiate()
     *
     * Go through the comm_descriptors in the passed table,
     * decide which one to use, and return an OutgoingInterface
     * for the appropriate one.
     */
    public
    OutgoingInterface
    outgoing_instantiate(
	byte[]					sp_comm_descriptor_table)
    {
	OutgoingInterface outgoing = null;
	int i;
	String sp_hostname;
	int sp_context;
	int length;
	int sp_proto_type;
	boolean[] is_valid = new boolean[1];
	//array - to allow pass by reference

	try
	{
	    i = 0;

	    /* get version */
	    byte				version;
	    
	    version = sp_comm_descriptor_table[i++];
	    if (version != (byte) Nexus.MI_PROTO_HEADER_VERSION)
	    {
		return null;
	    }
	    
	    /* get context differentiator */
	    sp_context = DataConversion.get_int
		(sp_comm_descriptor_table, i, 1);
	    i += 4;

	    /* find end of name */
	    for (length = 0;
		 sp_comm_descriptor_table[i+length] != 0;
		 length++);

	    /* get name */
	    sp_hostname = new String(sp_comm_descriptor_table, i, length);
	    i += (length + 1);

	    Nexus.debug_println(2,
				"Nexus.outgoing_instantiate():" + 
				" Header sp context=" + sp_context +
				", sp hostname=" + sp_hostname);

	    /* step through the sp protos */
	    while ((i < sp_comm_descriptor_table.length)
		   && (outgoing == null))
	    {
		/* get sp proto_type */
		sp_proto_type = (int)
		    DataConversion.get_short(sp_comm_descriptor_table, i, 1);
		i += 2;
		/* get proto length */
		length = (int)
		    DataConversion.get_short(sp_comm_descriptor_table, i, 1);
		i += 2;
 
		Nexus.debug_println(2, 
				    "Nexus.outgoing_instantiate(): " +
				    "Checking sp proto type=" + sp_proto_type);

		if (sp_proto_type == PROTO_TYPE_TCP)
		{
		    /* construct outgoing interface */
		    outgoing =
			tcp_proto.construct_from_mi_proto(
			    sp_comm_descriptor_table,
			    i,
			    length,
			    sp_hostname,
			    is_valid);

		    if (is_valid[0])
		    {
			if (outgoing == null) // flag for local proto
			{
			    outgoing = local_proto;

			    Nexus.debug_println(
				2,
				"Nexus.outgoing_instantiate(): " +
				"Instantiated to local proto");
			}
			else
			{
			    Nexus.debug_println(
				2,
				"Nexus.outgoing_instantiate(): " +
				"Instantiated to tcp proto");
			}
		    }
		}
		else if (sp_proto_type == PROTO_TYPE_UDP)
		{
		    if (udp_proto == null) udp_proto = new UdpProto(this);

		    /* construct outgoing interface */
		    outgoing = udp_proto.construct_from_mi_proto( 
			sp_comm_descriptor_table, i, length,
			sp_hostname, is_valid);
		}
		else if (sp_proto_type == PROTO_TYPE_TOTEM)
		{
		    if (totem_proto == null)
		    {
			totem_proto = new TotemProto(this);
		    }
		    /* construct outgoing interface */
		    outgoing = totem_proto.construct_from_mi_proto( 
                        sp_comm_descriptor_table, i, length,
                        sp_hostname, is_valid);
		}
		
		i += length;
	    }
	}
	catch (LostPrecisionException x)
	{
	    outgoing = null;
	}

	if (outgoing == null)
	{
	    System.err.println("Nexus.outgoing_instantiate(): " +
			       "could not find a valid protocol");
	}
	
	return (outgoing);
    }
    /* outgoing_instantiate() */

    /*
     * register_endpoint()
     *
     * Register an endpoint with nexus to get back a unique id.
     * This id can be used later to retrieve or unregister the endpoint.
     */
    public synchronized
    long
    register_endpoint(
	Endpoint				ep)
    {
	long endpoint_id = next_endpoint_id++;
	endpoint_id_hashtable.put(new Long(endpoint_id), ep);
	return endpoint_id;
    }
    /* register_endpoint() */


    /*
     * unregister_endpoint()
     *
     * Unregister an endpoint from nexus.
     */
    public synchronized
    void
    unregister_endpoint(
	long					id)
    {
	endpoint_id_hashtable.remove(new Long(id));
    }
    /* unregister_endpoint() */

    
    /*
     * lookup_endpoint()
     *
     * Lookup an endpoint which is registered with Nexus.
     */
    public synchronized Endpoint lookup_endpoint(long id)
    {
	return (Endpoint)(endpoint_id_hashtable.get(new Long(id)));
    }
    /* lookup_endpoint() */
    
    
    /*
     * endpointattr_init()
     */
    public
    EndpointAttr
    endpointattr_init()
    {
	return new EndpointAttr(this);
    }


    /*
     * endpoint_init()
     */
    public
    Endpoint
    endpoint_init(
	EndpointAttr				epattr)
    {
	return new Endpoint(this, epattr);
    }
    /* endpoint_init() */


    /*
     * startpoint_bind()
     */
    public
    Startpoint
    startpoint_bind(
	Endpoint				ep)
    {
	return new Startpoint(this, ep);
    }
    /* endpoint_init() */


    /*
     * buffer_init()
     */
    public
    PutBuffer
    buffer_init(
	int					buffer_size,
	int					num_direct_puts)
    {
	return new PutBuffer(this, buffer_size, num_direct_puts);
    }
    /* buffer_init() */

    
//     /*
//      * send_rsr()
//      */
//     public
//     void
//     send_rsr(
// 	PutBuffer				put_buffer,
// 	Startpoint				sp,
// 	int					handler_id,
// 	boolean					destroy_buffer)
//     {
// 	int start;
// 	int finish;
// 	int data_morph_info_size = 0;
// 	int direct_info_offset = 0;
// 	int other_info_offset = 0;
// 	int full_message_size;	

// 	boolean has_data_morph_info = false;
// 	boolean has_direct_info = false;
// 	boolean has_other_info = false;

// 	byte buffer[];
// 	byte buf[];
// 	byte pad_size = 0;
// 	byte flags = 0;
// 	byte liba[] = sp.liba();

// 	finish = put_buffer.get_current_location();
// 	start = put_buffer.get_header_size();

// 	/* get buffer from PutBuffer */
// 	if (destroy_buffer)
// 	{
// 	    buf = put_buffer.get_buffer();
// 	    put_buffer = null;
// 	}
// 	else
// 	{
// 	    buf = new byte[finish];	
// 	    buffer = put_buffer.get_buffer();
// 	    System.arraycopy(buffer, 0, buf, 0, finish);
// 	}

// 	/* fill in header information from end of header to start */

// 	/* put other info offset, set flag bit */
// 	if (has_other_info)
// 	{
// 	    start -= DataConversion.sizeof_long(1);
// 	    DataConversion.put_long((long)other_info_offset, buf, start);
// 	    flags |= 0x08;
// 	}
    
// 	/* put direct info offset, set flag bit */
// 	if (has_direct_info)
// 	{
// 	    start -= DataConversion.sizeof_long(1);
// 	    DataConversion.put_long((long)direct_info_offset, buf, start);
// 	    flags |= 0x10;
// 	}

// 	/* put handler id */
// 	start -= DataConversion.sizeof_int(1);
// 	DataConversion.put_int(handler_id, buf, start);

// 	/* data_morph info - not yet implemented
// 	   if (has_data_morph_info)
// 	   {
// 	   start -= data_morph_info_size;
// 	   DataConversion.put_byte(?)
// 	   }
// 	   */
    
// 	/* pad total array to 8 byte multiple */
// 	pad_size = (byte)(
// 	    (2 + DataConversion.sizeof_long(1) + 1 + liba.length + 1) % 8);
// 	start -= pad_size;
    
// 	/* set pad size flag, put flag byte */
// 	flags |= pad_size;
// 	start -= 1;
// 	buf[start] = flags;
    
// 	/* put liba and liba length */
// 	start -= liba.length;
// 	DataConversion.put_byte(liba, 0, liba.length, buf, start);
// 	start -= 1; 
// 	buf[start] = (byte)liba.length;
    
// 	/* include space for message length */
// 	start -= DataConversion.sizeof_long(1);
    
// 	/* calculate optimal buffer size, put length */
// 	full_message_size = (finish - start) + 2;
// 	DataConversion.put_long((long)(full_message_size), buf, start);
    
// 	/* put data format */
// 	start -= 1;
// 	buf[start] = (byte)DataConversion.NEXUS_DC_FORMAT_JAVA;
    
// 	/* put proto version number */
// 	start -= 1;
// 	buf[start] = (byte) sp.outgoing().get_Protocol_Version();
    
// 	/* should probably avoid this extra copy, send start and end indexes
// 	   as args to send_rsr() */

// 	/* create buffer of optimal size and copy data */ 
// 	buffer = new byte[full_message_size];
// 	System.arraycopy(buf, start, buffer, 0, buffer.length);

// 	/* send rsr */
// 	sp.outgoing().send_rsr(buffer);
//     }
//     /* send_rsr() */
      

    /*
     * send_rsr()
     */
    public void send_rsr(PutBuffer put_buffer,
			Startpoint sp,
			int handler_id,
			boolean destroy_buffer)
    {
	int 	        start,
			finish,
			data_morph_info_size = 0,
			direct_info_offset = 0,
			other_info_offset = 0,
			full_message_size;	
	boolean 	has_data_morph_info = false,
			has_direct_info = false,
			has_other_info = false;
	byte		buf[],
	                pad_size = 0,
	                flags = 0,
	                liba[] = sp.liba();
	Chunk           oldheader = null, newheader = null;

	if(!destroy_buffer)
	    oldheader = put_buffer.getHeader();
	newheader = put_buffer.newHeader();
	buf = newheader.getData();
	
	finish = put_buffer.getSize();
	start = put_buffer.get_header_size();

	/* fill in header information from end of header to start */

	/* put other info offset, set flag bit */
	if (has_other_info)
	{
		start -= DataConversion.sizeof_long(1);
		DataConversion.put_long((long)other_info_offset, buf, start);
		flags |= 0x08;
	}
	/* put direct info offset, set flag bit */
	if (has_direct_info)
	{
		start -= DataConversion.sizeof_long(1);
		DataConversion.put_long((long)direct_info_offset, buf, start);
		flags |= 0x10;
	}

	/* put handler id */
	start -= DataConversion.sizeof_int(1);
	DataConversion.put_int(handler_id, buf, start);

	/* data_morph info - not yet implemented
	if (has_data_morph_info)
	{
		start -= data_morph_info_size;
		DataConversion.put_byte(?)
	}
	*/

	/* pad total array to 8 byte multiple */
	// System.out.println("\n Nexus:  buffer is " + liba);	
	//  System.out.println("\n Nexus: length of buffer is " + liba.length);	
	pad_size = (byte)(
	    (2 + DataConversion.sizeof_long(1) + 1 + liba.length + 1) % 8);
	start -= pad_size;

	/* set pad size flag, put flag byte */
	flags |= pad_size;
	start -= 1;
	buf[start] = flags;

	/* put liba and liba length */
	start -= liba.length;
	DataConversion.put_byte(liba, 0, liba.length, buf, start);
	start -= 1; 
	buf[start] = (byte)liba.length;

	/* include space for message length */
	start -= DataConversion.sizeof_long(1);

	/* calculate optimal buffer size, put length */
	put_buffer.setHeaderSize(put_buffer.get_header_size() - start + 2);
	full_message_size = put_buffer.getSize();
	//	System.out.println("writing full_message_size = " + full_message_size + " at " + start);
	DataConversion.put_long((long)(full_message_size), buf, start);

	/* put data format */
	start -= 1;
	buf[start] = (byte)DataConversion.NEXUS_DC_FORMAT_JAVA;

	/* put proto version number */
	start -= 1;
	buf[start] = (byte) sp.outgoing().get_Protocol_Version();

	/* send rsr */
	sp.outgoing().send_rsr(put_buffer);
	
	if(!destroy_buffer)
	    put_buffer.setHeader(oldheader);
    } /* send_rsr() */


    /*
     * buffer_dispatch()
     */
    public
    void
    buffer_dispatch(
	byte[]					buffer,
	long					endpointID)
    {
	Endpoint				ep;
	HandlerInterface			handler_object;

	int location = 0;
	int version;
	int data_format;
	int pad_size;
	int liba_size;
	int handler_id;

	byte flags;
	byte liba[];

	boolean has_direct_info;
	boolean has_other_info;

	try
	{

	    /* get data format */
	    version = (int)(buffer[location++]);
  
	    /* get data format */
	    data_format = (int)(buffer[location++]);
  
	    /* skip over message length field */
	    location += DataConversion.sizeof_u_long(1, data_format);
  
	    /* get liba out of buffer */
	    liba_size = (int)(buffer[location++]);
	    liba = new byte[liba_size];
	    DataConversion.get_byte(liba, 0, liba_size, buffer,
				    location, data_format);
	    location += liba_size;
  
	    /* get info from flag byte */
	    flags = buffer[location++];
	    has_direct_info = ((flags & 0x10) != 0) ? true : false;
	    has_other_info = ((flags & 0x08) != 0) ? true : false;
	    pad_size = (int)(flags & 0x07);
	    location += pad_size;
  
	    /* datamorph info - how is its length determined? */
  
	    /* get handler id */
	    handler_id = DataConversion.get_int(buffer, location,
						data_format);
	    location += DataConversion.sizeof_u_int(1, data_format);
  
	    /* skip direct and other info offsets */
	    if (has_direct_info)
	    {
		location += DataConversion.sizeof_u_int(1, data_format);
	    }
	    if (has_other_info)
	    {
		location += DataConversion.sizeof_u_int(1, data_format);
	    }
 
	    if (endpointID == UdpProto.USE_BUFFER_EP_ID)
	    {
		/* use endpoint id from sp's liba - standard for unicast*/
		endpointID =
		    DataConversion.get_long(
			liba,
			0,
			DataConversion.NEXUS_DC_FORMAT_LOCAL);
	    }
	    ep = lookup_endpoint(endpointID);
// 	    System.out.println("\nNexus.buffer_dispatch() ep: " + ep);
// 	    System.out.println("\nNexus.buffer_dispatch() epId: " + endpointID);

	    handler_object = ep.get_handler_object();
// 	    System.out.println("Nexus.java:: HandlerInterface is " + handler_object.getClass());
// 	    System.out.println("\nNexus.buffer_dispatch() handler_object: " + handler_object);
// 	    System.out.println("\nNexus.buffer_dispatch(): this : " + this);
// 	    System.out.println("\nNexus.buffer_dispatch(): buffer : " + buffer);
// 	    System.out.println("\nNexus.buffer_dispatch(): location : " + location);
// 	    System.out.println("\nNexus.buffer_dispatch(): data_format: " + data_format);
// 	    System.out.println("\nNexus.buffer_dispatch(): ep: " + ep);
// 	    System.out.println("\nNexus.buffer_dispatch(): handler_id: " + handler_id);
	    handler_object.invoke_handler(
		ep, 
		new GetBuffer(this, buffer, location, data_format), 
		handler_id);
	}
	catch (LostPrecisionException x)
	{
	    System.out.println
		("LostPrecisionException occurred while dispatching buffer");
	}

    }
    /* buffer_dispatch() */

    
    
    /*
     * allow_attach()
     */
    public
    int
    allow_attach(
	int					port_in,
	AttachApprovalInterface			approval_object,
	Object					approval_object_user_arg)
    {
	return (tcp_proto.allow_attach(port_in,
				       approval_object,
				       approval_object_user_arg));
    }
    /* allow_attach() */


    /*
     * attach_hostname()
     */
    public
    String
    attach_hostname()
    {
	return (tcp_proto.get_hostname());
    }
    /* attach_hostname() */

    
    /*
     * disallow_attach()
     */
    public
    void
    disallow_attach(
	int					port)
    {
	tcp_proto.disallow_attach(port);
    }
    /* disallow_attach() */


    /*
     * attach()
     */
    public
    AttachReturn
    attach(
	URL					url)
    {
	return (tcp_proto.attach(url));
    }
    public AttachReturn attach(String url_string)
    {
	return (tcp_proto.attach(url_string));
    }
    /* attach() */

    
    /*
     * sizeof_TYPE()
     */
    public static
    int
    sizeof_byte(
	int					count)
    {
	return DataConversion.sizeof_byte(count);
    }
    
    public static
    int
    sizeof_char(
	int					count)
    {
	return DataConversion.sizeof_char(count);
    }

    public static
    int
    sizeof_short(
	int					count)
    {
	return DataConversion.sizeof_short(count);
    }
    
    public static
    int
    sizeof_int(
	int					count)
    {
	return DataConversion.sizeof_int(count);
    }
    
    public static
    int
    sizeof_long(
	int					count)
    {
	return DataConversion.sizeof_long(count);
    }
    
    public static
    int
    sizeof_float(
	int					 count)
    {
	return DataConversion.sizeof_float(count);
    }
    
    public static
    int
    sizeof_double(
	int					 count)
    {
	return DataConversion.sizeof_double(count);
    }

}
/* class Nexus */
