/*
 * TcpProto
 */

package nexus;

import java.util.*;
import java.net.*;
import java.io.*;

public class TcpProto
    implements ProtoInterface, URLStreamHandlerFactory
{
    public static final int PROTOCOL_VERSION = Nexus.BUFFER_VERSION + 0;
    public static final int MI_PROTO_VERSION = 0;
    public static final int ATTACH_VERSION = 0;
    
    public static final int TCP_MSG_HEADER_SIZE = 10;
    
    public static final int CLOSE_NORMAL_FLAG     = 128;

    private int        tcp_port;
    private String     tcp_hostname;

    private Nexus      nexus;
    private Queue      message_queue;
    private TcpHandler handler;
    private Vector     listener_vector;
    private boolean    listener_vector_done = false;
    private Vector     reader_vector;
    private boolean    reader_vector_done = false;
    private Vector     outgoing_vector;
    private boolean    outgoing_vector_done = false;

    /*
     * TcpProto()
     */
    public TcpProto(Nexus nexus)
    {
	NexusThread thread;
	Properties properties;
	String property;
	
	this.nexus = nexus;
	
	/*
	 * Initialize my hostname
	 */
	try
	{
	    tcp_hostname = InetAddress.getLocalHost().getHostName();
	}
	catch (UnknownHostException e)
	{
	    tcp_hostname = new String("localhost");
	}

	/*
	 * Setup an x-nexus URL handler so that we can use the
	 * URL class to parse x-nexus URLs.
	 * This doesn't seem to work.
	 * So let's just replace the URLStreamHandlerFactory.
	 */
	/*
	properties = System.getProperties();
	System.out.println("properties="+properties);
	property = properties.getProperty("java.handler.protol.pkgs");
	if (property == null)
	{
	    properties.put("java.handler.protol.pkgs",
			   "nexus");
	}
	else
	{
	    properties.put("java.handler.protol.pkgs",
			   property+"|nexus");
	}
	System.setProperties(properties);
	System.out.println("prop="+System.getProperty("java.handler.protol.pkgs"));
	*/
        // can't do this since we need a working URL stream handler
	// changed by nmukhi 10/27
	//URL.setURLStreamHandlerFactory(this);
	
	/*
	 * Create a message queue.
	 * As messages are received by the various threads watching
	 * the various InputStreams, those threads will put the messages
	 * onto the message queue.  A separate handler thread will deal
	 * with them.
	 */
	message_queue = new Queue();

	/*
	 * Initialize the incoming reader vector.
	 * This keeps a complete record of currently active
	 * Reader's.
	 */
	reader_vector = new Vector(5, 5);

	/*
	 * Initialize the outgoing object reader vector.
	 * This keeps a complete record of currently active
	 * TcpOutgoing connections.
	 */
	outgoing_vector = new Vector(5, 5);

	/*
	 * Start a handler thread.  This thread will watch the
	 * message_queue for incoming messages, and invoke the
	 * handlers for them.
	 */
	handler = new TcpHandler(nexus,
				 this,
				 message_queue);
	thread = new NexusThread(nexus, handler, "tcp_handler");
	thread.setDaemon(true);
	thread.start();

	/*
	 * Create a port to listen on, and create a thread
	 * to watch this port for connections.
	 */
	listener_vector = new Vector(5, 5);
	try
	{
	    ServerSocket server_socket;
	    Listener listener;
	    server_socket = new ServerSocket(0);
	    tcp_port = server_socket.getLocalPort();
	    listener = new Listener(nexus,
				    this,
				    server_socket);
	    thread = new NexusThread(nexus,
				     listener,
				     "tcp_listener_"+tcp_port);
	    thread.setDaemon(true);
	    thread.start();
	}
	catch (Exception e) { e.printStackTrace(); }
	
    } /* TcpProto() */

    public URLStreamHandler createURLStreamHandler(String protocol)
    {
	return (new NexusURLStreamHandler());
    }

    /*
     * shutdown()
     */
    public void shutdown()
    {
	/* Wake up the handler thread */
	message_queue.awaken_all_waiting_threads();

	/* Shut down outgoing connections */
	shutdown_outgoings();
	
	/* Wake up the listener threads */
	shutdown_listeners();
	
	/* Wake up the reader threads */
	shutdown_readers(CLOSE_NORMAL_FLAG);
	
    } /* shutdown() */


    /*
     * shutdown_listeners()
     */
    private void shutdown_listeners()
    {
	Listener l;
	Nexus.debug_println(2, "TcpProto.shutdown_listeners(): entering");
	for (Enumeration e = get_final_listeners(); e.hasMoreElements(); )
	{
	    l = (Listener) e.nextElement();
	    l.shutdown();
	}
	Nexus.debug_println(2, "TcpProto.shutdown_listeners(): exiting");
    } /* shutdown_listeners() */

	 
    /*
     * shutdown_readers()
     */
    private void shutdown_readers(int flag)
    {
	Reader r;
	Nexus.debug_println(2, "TcpProto.shutdown_readers(): entering");
	for (Enumeration e = get_final_readers(); e.hasMoreElements(); )
	{
	    r = (Reader) e.nextElement();
	    r.shutdown();
	}
	Nexus.debug_println(2, "TcpProto.shutdown_readers(): exiting");
    } /* shutdown_readers() */

	 
    /*
     * shutdown_outgoings()
     */
    private void shutdown_outgoings()
    {
	TcpOutgoing o;
	Nexus.debug_println(2, "TcpProto.shutdown_outgoings(): entering");
	for (Enumeration e = get_final_outgoings(); e.hasMoreElements(); )
	{
	    o = (TcpOutgoing) e.nextElement();
	    o.shutdown();
	}
	Nexus.debug_println(2, "TcpProto.shutdown_outgoings(): exiting");
    } /* shutdown_outgoings() */

	 
    /*
     * add_listener()
     */
    public synchronized void add_listener(Listener l)
    {
	listener_vector.addElement(l);
    } /* add_listener() */
   

    /*
     * remove_listener()
     */
    public synchronized void remove_listener(Listener l)
    {
	/* removeElement() will change the Enumeration of that Vector */
	if (!listener_vector_done)
	{
	    listener_vector.removeElement(l);
	}
    } /* remove_listener() */


    /*
     * get_final_listeners()
     */
    public synchronized Enumeration get_final_listeners()
    {
	listener_vector_done = true;
	return (listener_vector.elements());
    } /* get_final_listeners() */

    
    /*
     * add_reader()
     */
    public synchronized void add_reader(Reader r)
    {
	reader_vector.addElement(r);
    } /* add_reader() */
   

    /*
     * remove_reader()
     */
    public synchronized void remove_reader(Reader r)
    {
	/* removeElement() will change the Enumeration of that Vector */
	if (!reader_vector_done)
	{
	    reader_vector.removeElement(r);
	}
    } /* remove_reader() */
   

    /*
     * get_final_readers()
     */
    private synchronized Enumeration get_final_readers()
    {
	reader_vector_done = true;
	return (reader_vector.elements());
    } /* get_final_readers() */

    
    /*
     * add_outgoing()
     */
    public synchronized void add_outgoing(TcpOutgoing o)
    {
	outgoing_vector.addElement(o);
    } /* add_outgoing() */
   

    /*
     * remove_outgoing()
     */
    public synchronized void remove_outgoing(TcpOutgoing o)
    {
	/* removeElement() will change the Enumeration of that Vector */
	if (!outgoing_vector_done)
	{
	    outgoing_vector.removeElement(o);
	}
    } /* remove_outgoing() */


    /*
     * get_final_outgoings()
     */
    public synchronized Enumeration get_final_outgoings()
    {
	outgoing_vector_done = true;
	return (outgoing_vector.elements());
    } /* get_final_outgoings() */

    
    /*
     * get_my_mi_proto()
     *
     * Construct a byte array with my communications descriptor (mi_proto).
     */
    public byte[] get_my_mi_proto(ProtoInfo protoInfo, long endpointID)
    {
        TcpInfo					tcpInfo;
	byte[]					mi_proto;
	byte []					hostname;
	
	tcpInfo = (TcpInfo)protoInfo;

	if (tcp_hostname.equals(nexus.node_name()))
	{
	    hostname = (new byte[0]);
	}
	else
	{
	    hostname = tcp_hostname.getBytes();
	}

        /* Java put; 32bit BE and Java have same int size and format */
	
	mi_proto = new byte[hostname.length + 6];
	mi_proto[0] = (byte) MI_PROTO_VERSION;
	DataConversion.put_int(tcp_port, mi_proto, 1);
	System.arraycopy(hostname, 0, mi_proto, 5, hostname.length);
	mi_proto[hostname.length + 5] = 0;

	return (mi_proto);

    }
    /* get_my_mi_proto() */


    /*
     * construct_from_mi_proto()
     */
    public
    OutgoingInterface
    construct_from_mi_proto(
	byte[]					mi_proto,
	int					start,
	int					size,
	String					node_name,
	boolean[]				is_valid)
    {
	OutgoingInterface outgoing;
	int port;
	String hostname;
	int					length;
	byte					version;

	version = mi_proto[start++];
	if (version != (byte) MI_PROTO_VERSION)
	{
	    throw new Error("TcpProto.construct_from_mi_proto():" +
			    "incompatible protocol version");
	}
	
	try
	{
	    port = DataConversion.get_int(mi_proto, start,
                     DataConversion.NEXUS_DC_FORMAT_32BIT_BE);
	}
	catch (LostPrecisionException x)
	{
	    throw new Error("TcpProto.construct_from_mi_proto():" +
			    "LostPrecisionException occured when extracting " +
			    "port from mi_proto");
	}
	start += 4;
	
	for (length = 0; mi_proto[start+length] != 0; length++);

	if (length == 0)
	{
	    hostname = node_name;
	}
	else
	{
	    hostname = new String(mi_proto, start, length);
	}

	Nexus.debug_println(2,
            "TcpProto.construct_from_mi_proto(): hostname="
             +hostname+",port="+port+",tcp_hostname="
             +tcp_hostname+",tcp_port="+tcp_port);
	
	if ((port == tcp_port) && (hostname.equals(tcp_hostname)))
	{
	    /*
	     * Set to null to tell Nexus.java that it should use the
	     * local_proto.
	     */
	    outgoing = null;
	}
	else
	{
	    outgoing = new TcpOutgoing(nexus, this, hostname, port);
	}
	
	is_valid[0] = true;
	return (outgoing);
    } /* construct_from_mi_proto() */


    /*
     * get_hostname()
     */
    protected String get_hostname()
    {
	return (tcp_hostname);
    } /* get_hostname() */


    /*
     * enqueue_message()
     */
    public void enqueue_message(byte[] message)
    {
	message_queue.add(message);
    } /* enqueue_message() */
    
    
    /*
     * allow_attach()
     */
    public int allow_attach(int port_in,
			    AttachApprovalInterface approval_object,
			    Object approval_object_user_arg)
    {
	int port_out = -1;
    	try
	{
	    ServerSocket server_socket;
	    AttachListener listener;
	    NexusThread thread;
	    server_socket = new ServerSocket(port_in);
	    port_out = server_socket.getLocalPort();
	    listener = new AttachListener(nexus,
					  this,
					  server_socket,
					  approval_object,
					  approval_object_user_arg);
	    thread = new NexusThread(nexus,
				     listener,
				     "tcp_attach_listener_"+port_out);
	    thread.setDaemon(true);
	    thread.start();
	}
	catch (Exception e) { e.printStackTrace(); }
	return (port_out);
    } /* allow_attach() */


    /*
     * find_listener()
     */
    private synchronized Listener find_listener(int port)
    {
	Listener l;
	for (Enumeration e = listener_vector.elements();
	     e.hasMoreElements(); )
	{
	    l = (Listener) e.nextElement();
	    if (l.get_port() == port)
	    {
		return(l);
	    }
	}
	return (null);
    } /* find_listener() */

    
    /*
     * disallow_attach()
     */
    public void disallow_attach(int port)
    {
	Listener l;

	l = find_listener(port);
	if (l != null)
	{
	    l.shutdown();
	}
    } /* disallow_attach() */

    
    /*
     * attach()
     */
    public AttachReturn attach(URL url)
    {
	String host;
	int port;
	String url_string;
	byte[] url_bytes;
	Socket s;
	InputStream in;
	OutputStream out;

	int int_length,
	    body_length;
	int rc = -1;
	Startpoint sp = null;
	
	int data_format;
	int bytes_read, bytes_left, alength;

	data_format = DataConversion.NEXUS_DC_FORMAT_JAVA;
	int_length = DataConversion.sizeof_int(1);
	byte[] bytes = new byte[int_length];
	host = url.getHost();
	port = url.getPort();
	url_string = url.toExternalForm();
	url_bytes = url_string.getBytes();
	
	Nexus.debug_println(1, "TcpProto.attach(): attaching to host="
          +host+", port="+port+", for url="+url_string);

	try
	{
	    s = new Socket(host, port);
	    in = s.getInputStream();
	    out = s.getOutputStream();

	    /* write version */
	    bytes[0] = (byte) TcpProto.ATTACH_VERSION;
	    out.write(bytes, 0, 1);

	    /* write data format */
	    bytes[0] = (byte)data_format;
	    out.write(bytes, 0, 1);

	    /* write body length */
	    body_length = int_length + url_string.length();
	    DataConversion.put_int(body_length, bytes, 0);
	    out.write(bytes);

	    /* write url length */
	    DataConversion.put_int(url_string.length(), bytes, 0);
	    out.write(bytes);

	    /* write url */
	    out.write(url_bytes);

	    /* read data format */
	    in.read(bytes, 0, 1);
	    data_format = (int)bytes[0];

	    /* read body length */
	    bytes = new byte[DataConversion.sizeof_int(1, data_format)];
	    alength = DataConversion.sizeof_int(1, data_format);
	    bytes = new byte[alength];
	    bytes_read = 0;
	    bytes_left = alength;
	    while(bytes_left > 0)
	    {
	        int nbytes = in.read(bytes, bytes_read, bytes_left);
		bytes_read += nbytes;
		bytes_left -= nbytes;
	    }
	    body_length = DataConversion.get_int(bytes, 0, data_format);

	    /* read return code */
	    bytes_read = 0;
	    bytes_left = alength;
	    while(bytes_left > 0)
	    {
		int nbytes = in.read(bytes, bytes_read, bytes_left);
		bytes_read += nbytes;
		bytes_left -= nbytes;
	    }
	    rc = DataConversion.get_int(bytes, 0, data_format);
	    
	    if (rc == 0)
	    {
		/* Read the startpoint */
		sp = new Startpoint(nexus, in, data_format);
		if (sp.is_null())
		{
		    rc = -1;
		}
	    }
	    
	    in.close();
	    out.close();
	    s.close();
	}
	catch (IOException e)
	{
	    rc = -1;
	    sp = null;
	}
	catch (Exception e) { e.printStackTrace(); }
	    
	return (new AttachReturn(rc, sp));
    }
    public AttachReturn attach(String url_string)
    {
	URL url;
	AttachReturn rc = null;
	try
	{
	    url = new URL(url_string);
	    rc = attach(url);
	}
	catch (java.net.MalformedURLException e)
	{
	    rc = new AttachReturn(2, null);
	}
	catch (Exception e) { e.printStackTrace(); }
	
	return (rc);
    } /* attach() */

} /* class TcpProto */


/*
 * class Listener
 */
class Listener implements Runnable
{
    protected Nexus		nexus;
    protected TcpProto		tcp_proto;
    protected ServerSocket	server_socket;
    protected int		port;
    private boolean		shutting_down = false;

    /*
     * Listener()
     */
    public Listener(Nexus nexus,
		    TcpProto tcp_proto,
		    ServerSocket server_socket)
    {
	this.nexus = nexus;
	this.tcp_proto = tcp_proto;
	this.server_socket = server_socket;
	this.port = server_socket.getLocalPort();
	tcp_proto.add_listener(this);
    } /* Listener() */


    /*
     * run()
     */
    public void run()
    {
	Socket s;
	nexus.add_thread(Thread.currentThread());
	while (!nexus.is_destroying() && !shutting_down)
	{
	    try
	    {
		Nexus.debug_println(2, "Listener.run() "+port
                  +": calling accept() in thread "
                  +Thread.currentThread().getName());
		s = server_socket.accept();
		Nexus.debug_println(2, "Listener.run() "+port
                  +": accept() returned");
		if (nexus.is_destroying() || shutting_down)
		{
		    s.close();
		    server_socket.close();
		}
		else
		{
		    handle_connection(s);
		}
	    }
	    catch (Exception e)
	    {
		System.out.println("Listener.run() "+port
                  +": accept() threw exception");
		e.printStackTrace();
	    }
	}
	nexus.remove_thread(Thread.currentThread());
	tcp_proto.remove_listener(this);
	Nexus.debug_println(2, "Listener.run() "+port+": exiting");
    } /* run() */


    /*
     * shutdown()
     */
    public void shutdown()
    {
	Socket s;
	
	if (shutting_down)
	{
	    return;
	}
	
	shutting_down = true;
	try
	{
	    /* Attach to the port to wakeup the blocked thread */
	    Nexus.debug_println(2, "Listener.shutdown() "+port
              +": attaching to wakeup thread");
	    s = new Socket(tcp_proto.get_hostname(),
			   server_socket.getLocalPort());
	    s.close();
	    Nexus.debug_println(2, "Listener.shutdown() "+port
              +": done");
	}
	catch (Exception e) { e.printStackTrace(); }
    } /* shutdown() */


    /*
     * get_port()
     */
    public int get_port()
    {
	return (port);
    } /* get_port() */

    
    /*
     * handle_connection()
     */
    protected void handle_connection(Socket s)
    {
	Reader reader;
	NexusThread thread;
	Nexus.debug_println(2, "Listener.handle_connection() "+port 
          +": creating Reader");
	reader = new Reader(nexus,
			    tcp_proto,
			    s);
	thread = new NexusThread(nexus, reader, "tcp_reader");
	thread.start();
    } /* handle_connection() */
    
} /* class Listener */


/*
 * class AttachListener
 */
class AttachListener extends Listener
{
    protected AttachApprovalInterface	approval_object;
    protected Object			approval_object_user_arg;

    /*
     * AttachListener()
     */
    public AttachListener(Nexus nexus,
			  TcpProto tcp_proto,
			  ServerSocket server_socket,
			  AttachApprovalInterface approval_object,
			  Object approval_object_user_arg)
    {
	super(nexus, tcp_proto, server_socket);
	this.approval_object = approval_object;
	this.approval_object_user_arg = approval_object_user_arg;
    } /* AttachListener() */


    /*
     * handle_connection()
     */
    protected void handle_connection(Socket s)
    {
        InputStream     in;
        OutputStream    out;
        AttachReturn    attach_return;

	byte[] 		url_bytes,
	 		bytes;
	String 		url_string;
	int		data_format,
	 		body_length,
			url_length,
			int_length,
			rc;
	int             bytes_read, bytes_left;
	
	Nexus.debug_println(2, "AttachListener.handle_connection() "
          +port+": Got attach request");

	try
	{
	    in = s.getInputStream();
	    out = s.getOutputStream();

	    /* Read the version and verify it is supported */
	    bytes = new byte[1];
	    in.read(bytes);
	    if (bytes[0] != (byte) TcpProto.ATTACH_VERSION)
	    {
		in.close();
		out.close();
		s.close();
	    }

	    /* Read the data format */
	    bytes = new byte[1];
	    in.read(bytes);
	    data_format = (int)bytes[0];

	    /* Read the body length and URL length args */
	    int_length = DataConversion.sizeof_int(1, data_format);
	    bytes = new byte[2 * int_length];
	    bytes_left = 2 * int_length;
	    bytes_read = 0;
	    while(bytes_left > 0)
	    {
		int nbytes = in.read(bytes, bytes_read, bytes_left);
		bytes_read += nbytes;
		bytes_left -= nbytes;
	    }
	    body_length = DataConversion.get_int(bytes, 0, data_format);
	    url_length = DataConversion.get_int(bytes, int_length,
                                                data_format);
	
	    /* Read the url */
	    bytes = new byte[url_length];
	    bytes_left = url_length;
	    bytes_read = 0;
	    while(bytes_left > 0)
	    {
		int nbytes = in.read(bytes, bytes_read, bytes_left);
		bytes_read += nbytes;
		bytes_left -= nbytes;
	    }
	    url_string = new String(bytes, 0, url_length);

	    /* Call the approval object */
	    attach_return =
              approval_object.attach_approval(approval_object_user_arg,
                                              url_string);

	    /* write data format */
	    bytes[0] = (byte)DataConversion.NEXUS_DC_FORMAT_JAVA;
	    out.write(bytes, 0, 1);

	    int_length = DataConversion.sizeof_int(1);
	    bytes = new byte[int_length];	
	    body_length = int_length;
	    
	    if (   (attach_return == null)
		|| (   (attach_return.status == 0)
		    && (attach_return.sp == null) ) )
	    {
		rc = -3;

		/* write message body length, return code */
		DataConversion.put_int(body_length, bytes, 0);
		out.write(bytes);
		DataConversion.put_int(rc, bytes, 0);
		out.write(bytes);
	    }
	    else if ((rc = attach_return.status) != 0)
	    {
		/* write message body length, return code */
		DataConversion.put_int(body_length, bytes, 0);
		out.write(bytes);
		DataConversion.put_int(rc, bytes, 0);
		out.write(bytes);
	    }
	    else
	    {
		body_length += attach_return.sp.sizeof();

		/* write message body length, return code */
		DataConversion.put_int(body_length, bytes, 0);
		out.write(bytes);
		DataConversion.put_int(rc, bytes, 0);
		out.write(bytes);

		/* Write the startpoint and destroy the local instance */
		attach_return.sp.write(out);
		attach_return.sp.destroy();
	    }

	    in.close();
	    out.close();
	    s.close();
	}
	catch (Exception e) { e.printStackTrace(); }
    } /* handle_connection() */
    
} /* class AttachListener */


/*
 * class Reader
 */
class Reader implements Runnable
{
    private Nexus		nexus;
    private TcpProto		tcp_proto;
    private Socket		socket;
    private InputStream		input_stream;
    private OutputStream	output_stream;
    private boolean		closing_reader = false;
    
    /*
     * Reader()
     */
    public Reader(Nexus nexus,
		  TcpProto tcp_proto,
		  Socket socket)
    {
	this.nexus = nexus;
	this.tcp_proto = tcp_proto;
	this.socket = socket;
	try
	{
	    this.input_stream = socket.getInputStream();
	    this.output_stream = socket.getOutputStream();
	}
	catch (Exception e) { e.printStackTrace(); }
	tcp_proto.add_reader(this);
    } /* Reader() */


    /*
     * run()
     */
    public
    void
    run()
    {
	boolean					reading_header;
	byte []					header;
	byte []					buf;
	int					n_read;
	int					n_left;
	boolean					done;
    
	nexus.add_thread(Thread.currentThread());

	header = new byte[TcpProto.TCP_MSG_HEADER_SIZE];
	reading_header = true;
	n_read = 0;
	n_left = TcpProto.TCP_MSG_HEADER_SIZE;
	buf = header;

	done = false;
	while (!nexus.is_destroying() && !done)
	{
	    int					n;
	    
	    try
	    {
		n = input_stream.read(buf, n_read, n_left);
		Nexus.debug_println(3,
				    "TcpProto.Reader.run(): " +
				    "got " + n + " bytes, n_read=" +
				    n_read + ", n_left=" + n_left);
	    }
	    catch (Exception e)
	    {
		if (!closing_reader)
		{
		    throw new Error("TcpProto.Reader.run(): unexpected " + e +
				    " during read");
		}
		
		done = true;
		continue;
	    }

	    if (n == -1)
	    {
		if (reading_header &&
		    n_read == 2 &&
		    header[0] == (byte) TcpProto.PROTOCOL_VERSION &&
		    header[1] == (byte) TcpProto.CLOSE_NORMAL_FLAG)
		{
		    done = true;
		    continue;
		}
		else
		{
		    throw new Error("TcpProto.Reader.run(): " +
				    "unexpected EOF during read");
		}
	    }
	    
	    n_read += n;
	    n_left -= n;

	    if (n_left > 0)
	    {
		continue;
	    }

	    if (reading_header)
	    {
		int			 	data_format;
		int				full_message_size;
		int				location;

		location = 0;
		
		/*
		 * get protoocol/buffer version
		 */
		int				version;
		
		version = (int) header[location++];
		if (version != TcpProto.PROTOCOL_VERSION)
		{
		    throw new Error("TcpProto.Reader.run(): " +
				    "incompatible protocol version");
		}
		
		/*
		 * get data format
		 */
		data_format = (int)(header[location++]);
		    
		/*
		 * get full message size
		 */
		try
		{
		    full_message_size = (int)
			DataConversion.get_u_long(header, 
						  location,
						  data_format);
		}
		catch (LostPrecisionException x)
		{
		    throw new Error("TcpProto.Reader.run(): " +
				    "unexcepted " + x +
				    " while extracting message size");
		}

		Nexus.debug_println(
		    2,
		    "ReaderThread.run(): " +
		    "Finished reading header: message_size=" +
		    full_message_size);

		/*
		 * Allocate byte array for the message and copy the header
		 * into it
		 */
		buf = new byte[full_message_size];
		
		System.arraycopy(header, 0, buf, 0, n_read);
	
		/*
		 * Setup to read message body
		 */
		reading_header = false;
		n_left = full_message_size - n_read;
		    
		Nexus.debug_println(
		    2,
		    "TcpProto.Reader.run(): " +
		    "Reading body: " + n_left + " bytes");
	    }
	    else
	    {
		/*
		 * Got a complete body.
		 *
		 * enqueue byte array for the handler thread to handle.
		 */

		Nexus.debug_println(
		    2,
		    "TcpProto.Reader.run(): Finished reading body");

		/* enqueue message */
		tcp_proto.enqueue_message(buf);

		/* Setup to read next header */
		reading_header = true;
		buf = header;
		n_read = 0;
		n_left = TcpProto.TCP_MSG_HEADER_SIZE;
	    }
	}

	nexus.remove_thread(Thread.currentThread());
	tcp_proto.remove_reader(this);
    }
    /* run() */


    /*
     * shutdown()
     */
    public synchronized
    void
    shutdown()
    {
	if (!closing_reader)
	{
	    closing_reader = true;
	    try
	    {
		byte[] bytes = new byte[2];
		bytes[0] = (byte) TcpProto.PROTOCOL_VERSION;
		bytes[1] = (byte) TcpProto.CLOSE_NORMAL_FLAG;

		output_stream.write(bytes);

		input_stream.close();
		output_stream.close();
		socket.close();
	    }
	    catch (Exception e) { e.printStackTrace(); }
	}
    } /* shutdown() */
    
} /* class Reader */


/*
 * class TcpHandler
 */
class TcpHandler implements Runnable
{
    private Nexus	nexus;
    private TcpProto	tcp_proto;
    private Queue	message_queue;

    /*
     * TcpHandler()
     */
    public TcpHandler(Nexus nexus,
		      TcpProto tcp_proto,
		      Queue message_queue)
    {
	this.nexus = nexus;
	this.tcp_proto = tcp_proto;
	this.message_queue = message_queue;
    } /* TcpHandler() */


    /*
     * run()
     */
    public void run()
    {
	handle_messages();
	Nexus.debug_println(2, "TcpHandler.run(): handler thread terminating");
    } /* run() */

 
    /*
     * handle_messages()
     */
    private void handle_messages()
    {
	byte[] message;
	// i think this is causing a bug in brian's (student) code.
	// where does this thread die?
	while ((message = (byte[]) message_queue.remove_with_wait()) != null)
	{
	    nexus.buffer_dispatch(message, ProtoInterface.USE_BUFFER_EP_ID);
	}
    } /* handle_messages() */


} /* class TcpHandler */


class NexusURLStreamHandler extends URLStreamHandler
{

    protected URLConnection openConnection(URL u) throws IOException
    {
	 throw (new IOException("Cannot open nexus URL "+u));
    }
}
