
package nexus;

import java.net.*;
import java.io.*;

/*
 * UdpOutgoing
 *
 * Don't need to check IP address for multicast range.  Any socket
 * may send messages to a multicast group.  There is no need for the 
 * socket to be part of the group.
 */
public class UdpOutgoing
    implements OutgoingInterface
{
    private Nexus       nexus;
    private UdpProto    udpProto;
    private String      hostname;
    private InetAddress address;
    private int         port;
    private byte        ttl;

    private boolean  open = false;
    private boolean  shuttingDown = false;

    private DatagramSocket          socket;
    private static int              msgNum = 0;
    private static Object           msgNumLock;


    /*
     * UdpOutgoing()
     */
    public UdpOutgoing(Nexus nexus, UdpProto udpProto,
                       String hostname, int port, byte ttl)
    {
        this.nexus = nexus;
        this.udpProto = udpProto;
        this.hostname = hostname;
        this.port = port;
        this.ttl = ttl;

        // create InetAddress from hostname for DatagramPackets
        try { address = InetAddress.getByName(hostname); }
        catch (UnknownHostException hostX)
        {
            System.err.println("UdpOutgoing.UdpOutgoing(): unable to"
                    + " resolve address for hostname " + hostname);
        }

        msgNumLock = new Object();
        udpProto.add_outgoing(this);
    } /* UdpOutgoing() */

    /*
     * UdpOutgoing.get_Protocol_Version()
     */
    public
    int
    get_Protocol_Version()
    {
	return TcpProto.PROTOCOL_VERSION;
    }
    /* UdpOutgoing.get_Protocol_Version() */
    

    /*
     * send_rsr()
     */
    public void send_rsr(byte[] buffer)
    {
        byte pack[];
        int size;
        int totalLength;
        int bytesSent;
        int bytesLeft;
        int seqNum = 0;

	Nexus.debug_println(2, "UdpOutgoing.send_rsr(): entering");

        totalLength = buffer.length;

        // open the socket if necessary
        if (!open)
        {
        Nexus.debug_println(2, "UdpOutgoing.send_rsr(): " +
                            "opening socket to "+hostname+":"+port);

          if (address.isMulticastAddress())
          {
            try
            {
                socket = new MulticastSocket();
                ((MulticastSocket)socket).setTTL(ttl);
//TODO: test this - joinGroup() necessary?
          //      ((MulticastSocket)socket).joinGroup(address);
                open = true;
            }
            catch (IOException ioX)
            {
                System.err.println("UdpOutgoing.send_rsr(): unable to "
                        + "create multicast socket for "
                        + address + ":" + port);
            }
          }
          else
          { 
            try
            {
                socket = new DatagramSocket();
                open = true;
            }
            catch (SocketException hostX)
            {
                System.err.println("UdpOutgoing.send_rsr(): unable to "
                        + "create datagram socket for "
                        + hostname + ":" + port);
            }
          }
	}

        bytesSent = 0;
        while(bytesSent < totalLength)
        {
            bytesLeft = totalLength - bytesSent;

            if(bytesLeft < UdpProto.UDP_MAX_BODY_SIZE)
            {
                size = bytesLeft; 
            }
            else
            {
                size = UdpProto.UDP_MAX_BODY_SIZE;
            }

            synchronized(msgNumLock)
            {
                pack = UdpMessage.packHeader(msgNum, seqNum, size);
            }

            
            System.arraycopy(buffer, bytesSent, pack,
                             UdpProto.UDP_HDR_SIZE, size);

            this.send(pack, pack.length);

            bytesSent += size;
            seqNum++; 
        }

        synchronized(msgNumLock) { msgNum++; }
    } /* send_rsr() */
    
    public void send_rsr(PutBuffer buffer)
    {
	System.err.println("UdpProto: send_rsr(PutBuffer) not implemented yet.");
    }
    /*
     * send()
     */
    private synchronized void send(byte[] buffer, int size)
    {
        DatagramPacket p;

        try
        {
            p = new DatagramPacket(buffer, size, address, port);
            socket.send(p);
        }
        catch (IOException ioX) 
        {
          try{
            if (socket instanceof MulticastSocket)
              ((MulticastSocket)socket).leaveGroup(address);
          }
          catch (IOException _ioX) {}
          socket.close();
          open = false;
        }
    } /* send() */
    
    /*
     * shutdown()
     */
    public synchronized void shutdown()
    {

    Nexus.debug_println(2, "UdpOutgoing.shutdown()");

	if (!shuttingDown)
	{
            shuttingDown = true;
            if (open)
            {
                socket.close();
                open = false;
            }
	}
	udpProto.remove_outgoing(this);

    Nexus.debug_println(2, "UdpOutgoing.shutdown(): "
                        + "Removed UdpOutgoing - shutdown");
    } /* shutdown() */

} /* UdpOutgoing */
