
package nexus;

import java.util.*;
import java.net.*;
import java.io.*;

/*
 * UdpProto
 */
public class UdpProto
    implements ProtoInterface
{
  // class statics
    public static final int PROTOCOL_VERSION = Nexus.BUFFER_VERSION + 0;
    
  public static final int UDP_HDR_SIZE = 4;
  public static final int UDP_MAX_TRANSMISSION_UNIT = 1024;
  public static final int UDP_MAX_BODY_SIZE =
                          UDP_MAX_TRANSMISSION_UNIT - UDP_HDR_SIZE;

  // object vars
  private Nexus         nexus;
  private Queue         packet_queue;
  private UdpHandler    handler;
  private Vector        reader_vector;
  private boolean       reader_vector_done = false;
  private Vector        outgoing_vector;
  private boolean       outgoing_vector_done = false;
  private byte[]        default_mi_proto;

  /*
   * UdpProto()
   */
  public UdpProto(Nexus nexus)
  {
    NexusThread thread;
    Properties  properties;
    String      property;

    this.nexus = nexus;

    /*
     * Create a message queue.
     * As messages are received by the various threads watching
     * the various InputStreams, those threads will put the messages
     * onto the message queue.  A separate handler thread will deal
     * with them.
     */
    packet_queue = new Queue();

    outgoing_vector = new Vector(5, 5);
    reader_vector = new Vector(5, 5);

    /*
     * Start a handler thread.  This thread will watch the
     * packet_queue for incoming packets, and invoke the
     * handlers for them.
     */
    handler = new UdpHandler(nexus, this, packet_queue);
    thread = new NexusThread(nexus, handler, "udp_handler");
    thread.setDaemon(true);
    thread.start();

  } /* UdpProto() */

  /*
   * shutdown()
   */
  public void shutdown()
  {
    /* Wake up the handler thread */
    packet_queue.awaken_all_waiting_threads();

    /* Shut down outgoing connections */
    shutdown_outgoings();

    /* Shut down incoming connections */
    shutdown_readers();
  } /* shutdown() */

  /*
   * add_reader()
   */
  public synchronized void add_reader(UdpReader reader)
  {
    reader_vector.addElement(reader);
  } /* add_reader() */

  /*
   * remove_reader()
   */
  public synchronized void remove_reader(UdpReader reader)
  {
    /* removeElement() will change the Enumeration of that Vector */
    if (!reader_vector_done)
    {
      reader_vector.removeElement(reader);
    }
  } /* remove_reader() */
				    
  /*
   * get_final_readers()
   */
  private synchronized Enumeration get_final_readers()
  {
    reader_vector_done = true;
    return (reader_vector.elements());
  } /* get_final_readers() */

  /*
   * shutdown_readers()
   */
  private void shutdown_readers()
  {
    UdpReader reader;

    Nexus.debug_println(2, "UdpProto.shutdown_readers(): entering");
    for (Enumeration e = get_final_readers(); e.hasMoreElements(); )
    {
      reader = (UdpReader) e.nextElement();
      reader.shutdown();
    }
    Nexus.debug_println(2, "UdpProto.shutdown_readers(): exiting");
  } /* shutdown_readers() */

  /*
   * add_outgoing()
   */
  public synchronized void add_outgoing(UdpOutgoing o)
  {
    outgoing_vector.addElement(o);
  } /* add_outgoing() */
   
  /*
   * remove_outgoing()
   */
  public synchronized void remove_outgoing(UdpOutgoing o)
  {
    /* removeElement() will change the Enumeration of that Vector */
    if (!outgoing_vector_done) outgoing_vector.removeElement(o);
  } /* remove_outgoing() */

  /*
   * get_final_outgoings()
   */
  public synchronized Enumeration get_final_outgoings()
  {
    outgoing_vector_done = true;
    return (outgoing_vector.elements());
  } /* get_final_outgoings() */

  /*
   * shutdown_outgoings()
   */
  private void shutdown_outgoings()
  {
    UdpOutgoing o;
    Nexus.debug_println(2, "UdpProto.shutdown_outgoings(): entering");
    for (Enumeration e = get_final_outgoings(); e.hasMoreElements(); )
    {
      o = (UdpOutgoing) e.nextElement();
      o.shutdown();
    }
    Nexus.debug_println(2, "UdpProto.shutdown_outgoings(): exiting");
  } /* shutdown_outgoings() */
	 
  /*
   * get_my_mi_proto()
   *
   * Construct a byte array communications descriptor (mi_proto).
   */
  public byte[] get_my_mi_proto(ProtoInfo protoInfo,
                                long endpointID)
  {
    DatagramSocket socket;
    UdpReader      reader;
    UdpInfo        udpInfo;
    byte[]         mi_proto;
    String         address;
    boolean        unicast;

    // get user-specified connection information
    udpInfo = (UdpInfo)protoInfo;

    // if user specifies address
    if (udpInfo.address != null)
    {
      // if multicast address
      if (udpInfo.address.isMulticastAddress())
      {
        unicast = false;
 
        // Multiple multicast readers with the same address 
        // and port are acceptable - they will have unique
        // endpoint id's (allows multiple multicast
        // endpoints to exist in the same context with the
        // same communication descriptors).

        // open socket
	try
	{
          address = udpInfo.address.getHostAddress();

          // create MulticastSocket according to UdpInfo
	  socket = new MulticastSocket(udpInfo.port);
        }
	catch (IOException ioX)
	{
	  System.err.println("UdpProto: Couldn't create incoming connection "
	                       + "for " + udpInfo.address + ":" + udpInfo.port);
          return mi_proto = new byte[1];
	}
        // join multicast group
        try
        {
          ((MulticastSocket)socket).setTTL(udpInfo.ttl);
          ((MulticastSocket)socket).joinGroup(udpInfo.address);

	  // create reader to manage this incoming
          reader = new UdpReader(nexus, this, socket, endpointID);
	  this.add_reader(reader);
          Thread thread = new NexusThread(nexus, reader, "");
          thread.start();
        }
	catch (IOException ioX)
	{
	  System.err.println("UdpProto: Couldn't join Multicast group "
	                     + udpInfo.address + ":" + udpInfo.port);
          try { socket.close(); }
          catch (NullPointerException nullX) {}

          return mi_proto = new byte[1];
	}
      }
      // user-specified unicast address
      else
      {
        unicast = true;

        // TODO 
        // Need to check UdpInfo against existing UdpReaders.
        // If a reader with the same unicast address and port
        // already exists, it should be returned (it would be
        // impossible to open another on the same port).

	try
	{
          address = udpInfo.address.getHostAddress();

          // create DatagramSocket according to UdpInfo
	  socket = new DatagramSocket(udpInfo.port, udpInfo.address);

	  // create reader to manage this incoming
          reader = new UdpReader(nexus, this, socket,
                                 ProtoInterface.USE_BUFFER_EP_ID);
	  this.add_reader(reader);
          Thread thread = new NexusThread(nexus, reader, "");
          thread.start();
        }
	catch (SocketException sockX)
	{
	  System.err.println("UdpProto: Couldn't create incoming "
	         		   + "connection for " + udpInfo.address
				   + ":" + udpInfo.port);
          return mi_proto = new byte[1];
	}
      }
    }
    else // no user-specified address 
    {
      unicast = true;

      // If a default unicast reader already exists, return
      // the mi_proto describing it.
      if (default_mi_proto != null) {
        return default_mi_proto;
      }
      else
      {
        try
        {
          // create generic DatagramSocket
          socket = new DatagramSocket();
          address = (socket.getLocalAddress()).getHostAddress();
  
          // create reader to manage this incoming
          reader = new UdpReader(nexus, this, socket,
                                 ProtoInterface.USE_BUFFER_EP_ID);
          this.add_reader(reader);
          Thread thread = new NexusThread(nexus, reader, "");
          thread.start();
        }
        catch (SocketException sockX)
        {
  	  System.err.println("UdpProto: Couldn't create incoming connection");
            return mi_proto = new byte[1];
        }
      }
    }

    // IP adress as String to bytes
    byte[] ip_addr = address.getBytes();

    // create mi_proto - insert address
    mi_proto = new byte[ip_addr.length + 13];
    System.arraycopy(ip_addr, 0, mi_proto, 12, ip_addr.length);
    mi_proto[ip_addr.length + 12] = 0;

    /* Java put; 32bit BE and Java have same int size and format */

    // place file descriptor number, port, and TTL in proto
    DataConversion.put_int( 0 /*fd*/, mi_proto, 0);
    DataConversion.put_int(socket.getLocalPort(), mi_proto, 4);
    if (socket instanceof MulticastSocket)
    {
      try
      {
        DataConversion.put_int(((MulticastSocket)socket).getTTL(), mi_proto, 8);
      }
      catch (IOException ioX)
      {
        DataConversion.put_int(UdpInfo.RESTRICT_TO_SUBNET, mi_proto, 8);
      }
    }
    else
    {
      DataConversion.put_int(UdpInfo.RESTRICT_TO_SUBNET, mi_proto, 8);
    }

    // save default unicast mi_proto
    if (unicast && (default_mi_proto == null)) {
      default_mi_proto = mi_proto;
    }

    return (mi_proto);

  } /* get_my_mi_proto() */


  /*
   * construct_from_mi_proto()
   */
  public OutgoingInterface construct_from_mi_proto(byte[] mi_proto,
                                                   int start,
                                                   int size,
                                                   String node_name,
                                                   boolean[] is_valid)
  {
    OutgoingInterface outgoing;
    int    port;
    byte   ttl;
    String hostname;
    int    length;

    try
    {
        port = DataConversion.get_int(mi_proto, start + 4,
                 DataConversion.NEXUS_DC_FORMAT_32BIT_BE);
        ttl = (byte)(DataConversion.get_int(mi_proto, start + 8,
                 DataConversion.NEXUS_DC_FORMAT_32BIT_BE));
    }
    catch (LostPrecisionException x)
    {
      System.err.println("LostPrecisionException occured " +
                         "when constructing OutgoingInterface");
      port = -1;
      ttl = -1;
    }

    // get hostname/ip address
    start += 12;  // skip fd, port, ttl
    for (length = 0; mi_proto[start+length] != 0; length++);
    if (length == 0)
    {
      hostname = node_name;
    }
    else
    {
      hostname = new String(mi_proto, start, length);
    }

    outgoing = new UdpOutgoing(nexus, this, hostname, port, ttl);
    is_valid[0] = true;

    return (outgoing);
  } /* construct_from_mi_proto() */

  /*
   * get_hostname()
   */
/*
  protected String get_hostname()
  {
    return (new String(""));//udp_hostname);
  } /* get_hostname() */

  /*
   * enqueue_packet()
   */
  public void enqueue_packet(UdpPacket packet)
  {
    packet_queue.add(packet);
  } /* enqueue_packet() */

} /* UdpProto() */

/*
 * UdpReader
 */
class UdpReader implements Runnable
{
  private Nexus	         nexus;
  private UdpProto       udp_proto;
  private DatagramSocket socket;
  private boolean        closing_reader = false;
  private boolean        done = false;
  private long           endpointID;
    
  /*
   * UdpReader()
   */
  public UdpReader(Nexus nexus,
                UdpProto udp_proto,
                DatagramSocket socket,
                long endpointID)
  {
    this.nexus = nexus;
    this.udp_proto = udp_proto;
    this.socket = socket;
    this.endpointID = endpointID;
  } /* UdpReader() */

  /*
   * run()
   */
  public void run()
  {
    nexus.add_thread(Thread.currentThread());

    while (!nexus.is_destroying() && !done)
    {
      try
      {
        DatagramPacket packet = new DatagramPacket(
          new byte[UdpProto.UDP_MAX_TRANSMISSION_UNIT],
          UdpProto.UDP_MAX_TRANSMISSION_UNIT);
       
        socket.receive(packet);  
        udp_proto.enqueue_packet(new UdpPacket(packet, endpointID));
      }
      catch (Exception e)
      {
        if (!closing_reader)
        {
          System.err.println("ERROR: Reader.run(): " +
                             "Exception while reading: " + e);
          e.printStackTrace();
        }
	done = true;
	continue;
      }
    }
    this.shutdown();
    nexus.remove_thread(Thread.currentThread());
  } /* run() */

  /*
   * shutdown()
   */
  public synchronized void shutdown()
  {
    if (!closing_reader)
    {
      closing_reader = true;
      socket.close();
    }
  } /* shutdown() */
    
} /* UdpReader */


/*
 * class UdpHandler
 */
class UdpHandler implements Runnable
{
  private Nexus    nexus;
  private UdpProto udp_proto;
  private Queue    packet_queue;
  private Vector   msgTable;

  /*
   * UdpHandler()
   */
  public UdpHandler(Nexus nexus,
                    UdpProto udp_proto,
                    Queue packet_queue)
  {
    this.nexus = nexus;
    this.udp_proto = udp_proto;
    this.packet_queue = packet_queue;

    msgTable = new Vector(10, 10);
  } /* UdpHandler() */

  /*
   * run()
   */
  public void run()
  {
    handle_messages();
    Nexus.debug_println(1, "UdpHandler.run(): " +
                        "handler thread terminating");
  } /* run() */
    
  /*
   * handle_messages()
   */
  private void handle_messages()
  {
    UdpPacket inPack;
    byte[] msg;
    int msg_num;
    int seq_num;
    long msg_total_length;
    UdpMessage udpMessage;
    boolean killFlag;

    while ((inPack = (UdpPacket) packet_queue.remove_with_wait()) != null)
    {
      msg = inPack.getData();
      msg_num = (int) msg[0];

      seq_num = (int)msg[1] << 16;
      seq_num += ((int)msg[2] << 8);
      seq_num += ((int)msg[3]);

      /* if new packet */
      if(seq_num == 0)
      {
        int format = (int)(msg[4]);

        try
        { msg_total_length = DataConversion.get_long(msg, 5, format); }
        catch (LostPrecisionException lpX)
        { msg_total_length = -1; }

        udpMessage = new UdpMessage((int)(msg_total_length), msg_num, 
                                    inPack, nexus);

        msgTable.addElement(udpMessage);
      }

      udpMessage = findNode(inPack, msg_num);
      if(udpMessage != null)
      {
        killFlag = udpMessage.addBytes(msg, inPack.getLength(), seq_num);
        if(killFlag)
        {
          msgTable.removeElement(udpMessage);
        }
      }

    }
  } /* handle_messages() */

  private UdpMessage findNode(UdpPacket packet, int msg_num)
  {
    int count;
    UdpMessage tempMsg = null;
    boolean found = false;

    InetAddress address;
    int         port;
    long        endpointID;

    address = packet.getAddress();
    port = packet.getPort();
    endpointID = packet.getEndpointID();

    count = msgTable.size();

    for(int ctr = 0; !found && ctr < count; ctr++)
    {
      try { tempMsg = (UdpMessage)(msgTable.elementAt(ctr)); }
      catch(Exception e) {System.out.println("element not found");}

      if(tempMsg.messageNum == msg_num 
	 && tempMsg.fromAddr.equals(address)
	 && tempMsg.fromPort == port
	 && tempMsg.endpointID == endpointID)
      {
        found = true;
      }
    }

    if(found)
    {
      return tempMsg;
    }

    return null;
  } /* findNode() */

} /* UdpHandler */
