package nexus.com.totem;

import java.lang.*;
import java.net.*;
import java.io.*;

/** header for message exchanged between PG layer and Totem. */
public class InfHeader {

  /** entire length: message and header. */
  int len;                  

  /**  length of header alone. */
  int header_len;           

  /** message type - agreed, safe for data messages. */
  public int type;                 
  // boolean isAgreed(), isSafe(), isMembershipMsg(), ....???
 
  /** number of process groups being sent to. */
  /* doug */
  public
  /* doug */
  int num_sets;             

  /** process group ids of the destinaltion process groups. */
  /* doug */
  public
  /* doug */
  InfId[] pgid;             

  /** Maximum process groups that a message can be sent to simultaneously. */
  static int MAX_PG = 10;   

  /** Length of header when no destination process groups are defined. */
  static int MIN_LENGTH = 16; 

  /** Maximum possible length of header (max destination groups defined)
      -  currently used for SAFE and AGREED messages due to Totem.  */
  static int MAX_LENGTH = 16 + MAX_PG*8;


  /** Creates an empty header of maximum length. */
  public InfHeader() { 
    len = 0; header_len = 0; type = 0; num_sets = 0;
    pgid = new InfId[MAX_PG];
  }

  /** Creates a specific header of length calculated from number of dest.
      @param type defines type of message (agreed or safe) if data message
      or defines if message is a Join/Leave/Membership message.
      @param num_sets defines the number of destination  process groups
      @param pgid an array of destination process groups
      @param msg_len length of actual data content of message */
  public InfHeader(int type, int num_sets, InfId[] pgid, int msg_len) {
    int i;

    this.type       = type;     // type of message
    this.num_sets   = num_sets; // number of groups sent to
    this.pgid       = pgid;     // ids of groups sent to
    this.header_len = 4*4;      // header without pgids
    for (i = 0; i < num_sets; i++)
      header_len += pgid[i].length();

    this.len = header_len + msg_len;// this.header_len = 4*4 + MAX_PG*8;
  }


  /** Creates a specific header of a known header length. 
      @param type defines type of message (agreed or safe - if data message
                  - or if it is Join/Leave/Membership message).
      @param num_sets defines the number of destination  process groups
      @param pgid an array of destination process groups
      @param hdr_len specified length of header 
      @param msg_len length of actual data content of message 
  */
  public InfHeader(int type, int num_sets, InfId[] pgid, int hdr_len, int msg_len) {
    int i;

    this.type       = type;
    this.num_sets   = num_sets;
    this.pgid       = pgid;
    this.header_len = hdr_len;
    this.len        = header_len + msg_len;
  }

  public int msgLength() {
    return len - header_len;
  }
  //-------------------------------------------------------------------------
  /**
    * Print this string for debugging - Can be easily commented out, if no
    * debugging is desired <br>
    * If the output call inside this method is commented out, the compiler
    * should eliminate the call of debug().
    */
  public static final void debug( String str )
  {
    System.err.println( "InfHdr: "+ str );
  }

  //-------------------------------------------------------------------------  
  /** returns the length of the header containing useful information */
  int length() {
    return (16 + num_sets*pgid[0].length());
  }

  /** writes the header to the specified output data stream in the order.
    * length of message, length of header, message type, destination count,
    * and actual destination process group ids.
    *
    * @param out the DataOutputStream to be written to 
    */
  void write(DataOutputStream out) throws IOException {
    int i;

    out.writeInt(len);
    out.writeInt(header_len);
    out.writeInt(type);
    out.writeInt(num_sets);
    for (i = 0; i < num_sets; i++)
      pgid[i].write(out);
  }

  //-----------------------------------------------------------------------
  /**  reads the header from an input data stream in the order <br>
    *  message length, header length, message type, destination count,
    *  actual destination process ids.
    *
    *  @param in the DataInputStream read from 
    */
  void read(DataInputStream in) throws IOException 
  {
    int i;

    len        = in.readInt();
    header_len = in.readInt();
    type       = in.readInt();
    num_sets   = in.readInt();
    for (i = 0; i < num_sets; i++) {
      pgid[i] = new InfId();
      pgid[i].read(in);
    }
  }

  /** Prints out the header information */
  void print() throws IOException 
  {
    int i;

    debug("Header info: ");
    System.out.println(" Header length  = " + header_len);
    System.out.println(" Packet Length  = " + len);
    System.out.println(" Type           = " + type);
    System.out.println(" Number of Sets = " + num_sets);
    for (i = 0; i < num_sets; i++)
      pgid[i].print();
  }
}
