/* 
 * NexusRMI
 *
 * Copyright (c) 1999
 * Fabian Breg
 * Indiana University/Leiden University
 * All Rights Reserved
 *
 * Please refer to the LICENSE file provided in this
 * distribution for full licensing terms and
 * conditions.
 *
 * Please report all bugs/problems to
 * breg@cs.indiana.edu
 *
 ******************************************************
 *
 * Naming.java
 *
 */


package nexusrmi;

import nexusrmi.registry.*;
import nexusrmi.server.*;
import java.net.*;

public final class Naming
{
    // Prevent the creation of instances of this class.
    private Naming() {}

    // This method copied from the original RMI. file: changed to http:
    // Since "rmi" is not an officially recognized protocol, we change
    // the protocol to http: in order to create URL objects from the name
    private static URL cleanURL(String name) throws java.net.MalformedURLException
    {
        URL url = new URL("http:");

        // remove the approved protocol
        if (name.startsWith("rmi:"))
            name = name.substring(4);

        // No protocol must remain
        int colon = name.indexOf(':');
        if (colon >= 0 && colon < name.indexOf('/') )
            throw new java.net.MalformedURLException("Protocol should be rmi:");

        url = new URL(url, name);
        return url;
    }
    
    public static Remote lookup(String url) throws NotBoundException, MalformedURLException,
                                                                     nexusrmi.UnknownHostException, RemoteException
    {
	Remote obj = null;

	// Retrieve information from the URL.
	URL objurl = cleanURL(url);
	String host = objurl.getHost();
	int port = objurl.getPort();
	try
	{
	    if(host.equals(""))
		host = InetAddress.getLocalHost().getHostName();
	}
	catch(java.net.UnknownHostException e)
	{
	    throw new nexusrmi.UnknownHostException(host);
	}

	// Get a stub for the registry.
	// If no registry present, a ConnectException with a descriptive string is thrown
	Registry reg = LocateRegistry.getRegistry(host, port);
	if(reg == null)
	    throw new ConnectException("Connection refused to host " + host);
	else
	    obj = reg.lookup(url); // Let the stub do the real work.
	
	return obj;
    }

    
    public static void bind(String url, Remote obj) throws AlreadyBoundException, MalformedURLException,
                                                                                  nexusrmi.UnknownHostException, RemoteException
    {
	// Retrieve information from the URL.
	URL objurl = cleanURL(url);
	String host = objurl.getHost();
	int port = objurl.getPort();

	// Get a stub for the registry.
	// If no registry present, a ConnectException with a descriptive string is thrown
	Registry reg = LocateRegistry.getRegistry(host, port);
	if(reg == null)
	    throw new ConnectException("Connection refused to host " + host);
	else
	    reg.bind(url, obj); // Let the stub do the real work.
    }
    
    public static void unbind(String url) throws RemoteException, NotBoundException, MalformedURLException, nexusrmi.UnknownHostException
    {
	// Retrieve information from the URL.
	URL objurl = cleanURL(url);
	String host = objurl.getHost();
	int port = objurl.getPort();

	// Get a stub for the registry.
	// If no registry present, a ConnectException with a descriptive string is thrown
	Registry reg = LocateRegistry.getRegistry(host, port);
	if(reg == null)
	    throw new ConnectException("Connection refused to host " + host);
	else
	    reg.unbind(url);
    }

    public static void rebind(String url, Remote obj) throws RemoteException, MalformedURLException, nexusrmi.UnknownHostException
    {
	// Retrieve information from the URL.
	URL objurl = cleanURL(url);
	String host = objurl.getHost();
	int port = objurl.getPort();

	// Get a stub for the registry.
	// If no registry present, a ConnectException with a descriptive string is thrown
	Registry reg = LocateRegistry.getRegistry(host, port);
	if(reg == null)
	    throw new ConnectException("Connection refused to host " + host);
	else
	    reg.rebind(url, obj); // Let the stub do the real work.
    }
    
    public static String[] list(String url) throws RemoteException, MalformedURLException, nexusrmi.UnknownHostException
    {
	String[] res = null;
	
	// Retrieve information from the URL.
	URL objurl = cleanURL(url);
	String host = objurl.getHost();
	int port = objurl.getPort();

	try
	{
	    if(host.equals(""))
		host = InetAddress.getLocalHost().getHostName();
	}
	catch(java.net.UnknownHostException e)
	{
	    throw new nexusrmi.UnknownHostException(host);
	}
	
	// Get a stub for the registry.
	// If no registry present, a ConnectException with a descriptive string is thrown
	Registry reg = LocateRegistry.getRegistry(host, port);
	if(reg == null)
	    throw new ConnectException("Connection refused to host " + host);
	else
	    res = reg.list(); // Let the stub do the real work.

	return res;
    }
}

