/* 
 * NexusRMI
 *
 * Copyright (c) 1999
 * Fabian Breg
 * Indiana University/Leiden University
 * All Rights Reserved
 *
 * Please refer to the LICENSE file provided in this
 * distribution for full licensing terms and
 * conditions.
 *
 * Please report all bugs/problems to
 * breg@cs.indiana.edu
 *
 ******************************************************
 *
 * LocateRegistry.java
 *
 */


package nexusrmi.registry;

import nexusrmi.*;
import java.net.*;

// This class tries to locate or create a registry and returns a stub for it.
public final class LocateRegistry
{
    private static final int DEFAULT_REGISTRY_PORT = 7000;

    // Return the registry on this host for the default port.
    // Return null if no registry running.
    public static Registry getRegistry() throws RemoteException
    {
	Registry r = null;
	try
	{
	    r = getRegistry(InetAddress.getLocalHost().getHostName(), DEFAULT_REGISTRY_PORT);
	}
	catch(java.net.UnknownHostException e)
	{
	    e.printStackTrace();
	    //System.exit(1);
	}
	return r;
    }

    // Return the registry on this host for the given port.
    // Return null if no registry running.
    public static Registry getRegistry(int port) throws RemoteException
    {
	Registry r = null;
	try
	{
	    r = getRegistry(InetAddress.getLocalHost().getHostName(), port);
	}
	catch(java.net.UnknownHostException e)
	{
	    e.printStackTrace();
	    //System.exit(1);
	}
	return r;
    }
    
    // Return the registry on the given host for the default port.
    // Return null if no registry running.
    public static Registry getRegistry(String host) throws RemoteException, nexusrmi.UnknownHostException
    {
	return getRegistry(host, DEFAULT_REGISTRY_PORT);
    }
    
    // Return the registry on the given host for the given port.
    // Return null if no registry running.
    public static Registry getRegistry(String host, int port) throws RemoteException, nexusrmi.UnknownHostException
    {
	Registry reg = null;
	
	try
	{
	    reg = new RegistryStub(host, port);
	}
	catch(Exception e)
	{
	    // This is not what Java RMI does, however, but this allows you to check
	    // whether a registry is running, and to create one if not.
	    reg = null;
	}
	return reg;
    }

    // Create a registry on the local host for the given port and return a stub for it.
    public static Registry createRegistry(int port) throws RemoteException
    {
	Registry reg = null;
	
	try
	{
	    RMIRegistry rmireg = new RMIRegistry(port, false); // Create the registry and
	    Thread registryThread = new Thread(rmireg); // run it in a separate thread.
	    registryThread.start();
	    reg = new RegistryStub(InetAddress.getLocalHost().getHostName(), port);
	}
	catch(java.net.UnknownHostException e)
	{
	    // This should not happen since the local host always exists.
	    e.printStackTrace();
	    //System.exit(1);
	}
	catch(NoRegistryException e)
	{
	    // The registry stub throws a NoRegistryException if no registry is found, but since we have
	    // just created one this should not happen.
	    e.printStackTrace();
	    //System.exit(1);
	}
	return reg;
    }
}
