/* 
 * NexusRMI
 *
 * Copyright (c) 1999
 * Fabian Breg
 * Indiana University/Leiden University
 * All Rights Reserved
 *
 * Please refer to the LICENSE file provided in this
 * distribution for full licensing terms and
 * conditions.
 *
 * Please report all bugs/problems to
 * breg@cs.indiana.edu
 *
 ******************************************************
 *
 * RMIRegistry.java
 *
 */


package nexusrmi.registry;

import nexus.Nexus;
import nexus.AttachApprovalInterface;
import nexus.AttachReturn;
import nexus.HandlerInterface;
import nexus.Endpoint;
import nexus.EndpointAttr;
import nexus.Startpoint;
import nexus.GetBuffer;
import nexus.PutBuffer;
import nexusrmi.*;
import nexusrmi.server.*;
import java.util.Vector;

// This class is a combination of a registry remote object and a skeleton 
// for it.
public class RMIRegistry implements Runnable, HandlerInterface, AttachApprovalInterface
{
    private static final int DEFAULT_PORT = 7000; // Madhu chose this one.

    // These numbers should correspond to the numbers in the RegistryStub 
    // class.
    private final int GF_HANDLER_ID = 0;
    private final int RETURN_HANDLER_ID = 5;
    private final int EXCEPTION_HANDLER_ID = 15;
    
    private final int LOOKUP_HANDLER_ID = 100;
    private final int REBIND_HANDLER_ID = 101;
    private final int BIND_HANDLER_ID = 102;
    private final int UNBIND_HANDLER_ID = 103;
    private final int LIST_HANDLER_ID = 104;

    
    private Nexus nxs;
    private boolean RecdFromClient = false; 
    private int portNumber;
    private Endpoint registryEndpoint;

    private Serialize ser;

    private Vector codebaseTable = new Vector();      // store codebase
    private Vector serializationTable = new Vector(); // store object implementing the serialization protocol
    private Vector startpointTable = new Vector();    // store startpoints to the remote object
    private Vector objectNameTable = new Vector();    // store fully qualified class name for remote object
    private Vector NameTable = new Vector();          // store the name under which the remote object is registered.
    // This stores just the filename of the URL to catch the event that multiple servers register the same object.
    // The hostname would make these names different and thus allow the registering of the object. Formally,
    // however the list() method should return full URL names.
                                                   
  
    private int objectName = 0;
    private boolean verbose = false;
    
    public RMIRegistry(int port, boolean vf) throws ExportException
    {
	nxs = Communication.getNexus();
	portNumber = port;
	EndpointAttr registryEpattr = nxs.endpointattr_init();
	registryEpattr.set_handler_object(this);
	registryEndpoint = nxs.endpoint_init(registryEpattr);
	int newport = nxs.allow_attach(portNumber, this, this);
	if(newport < 0)
	    throw new ExportException("Port already in use", new java.net.BindException());
	
	String hostname = nxs.attach_hostname();
	ser = new Nexusize();
	verbose = vf;
	if(verbose) {
	    System.out.println("");
	    registryOut("Accepting requests on " + hostname + ":" + portNumber + " ... ");
	}
    }

    public static void main(String args[])
    {
	int i = 0;
	int port = DEFAULT_PORT;
	boolean vf = false;

        while(i < args.length)
	{
	    if(args[i].equals("-v"))
		vf = true;
	    else
	    {
		try
		{
		    port = Integer.parseInt(args[i]);
		}
		catch(Exception e)
		{
		    System.err.println("Usage: nexusrmiregistry [-v] [portnumber]");
		}
	    }
	    i++;
	}
	try
	{
	    RMIRegistry m = new RMIRegistry(port, vf);
	    m.run();
	}
	catch(RemoteException e)
	{
	    e.printStackTrace();
	}
    }

    private void registryOut(String msg)
    {
	System.out.println("[NexusRMIRegistry]: " + msg);
    }
    private void registryErr(String msg)
    {
	System.out.println("[NexusRMIRegistry error]: " + msg);
    }
    
    public void run()
    {
	waitForRequest();
    }

    protected void finalize() throws Throwable
    {
	nxs.disallow_attach(portNumber);
	registryEndpoint.destroy();
	nxs.destroy_current_context(false);
    }

    public AttachReturn attach_approval(Object user_arg, String url)
    {
	Startpoint registryStartpoint = nxs.startpoint_bind(registryEndpoint);
	return(new AttachReturn(0,registryStartpoint));
    }

    public String getName(String urlName)
    {
	return urlName.substring(urlName.lastIndexOf('/') + 1);
    }
    
    public synchronized void invoke_handler(Endpoint ep, GetBuffer buffer, int hid)
    {
	char c;
	ser.initRead(buffer, null);
	ser.startReadHeader();
	Startpoint repSP = ser.readHPCxxHeaderStartpoint();
	int offset = ser.readHPCxxHeaderOffset();
	int handler_id = ser.readHPCxxHeaderHandlerid();
	ser.startReadData();
	
	switch(handler_id)
	{      
	case REBIND_HANDLER_ID:
	    {
		// Store the entry from the buffer in the table. If an object with
		// the same name already exists, replace it.
 		String serclass = (String)ser.readFinalObject(java.lang.String.class, null, null);
		String codebase = (String)ser.readFinalObject(java.lang.String.class, null, null);
		Startpoint objectStartpoint = (Startpoint)ser.readFinalObject(nexus.Startpoint.class, null, null);
 		String objectName_s = (String)ser.readFinalObject(java.lang.String.class, null, null);
 		String Name_s = getName((String)ser.readFinalObject(java.lang.String.class, null, null));

		if(verbose)
		    registryOut("Storing in registry: " + Name_s);

		int index = NameTable.indexOf(Name_s);
		if(index < 0)
		{   // Just store
		    codebaseTable.addElement(codebase);
		    serializationTable.addElement(serclass);
		    startpointTable.addElement(objectStartpoint);
		    objectNameTable.addElement(objectName_s);
		    NameTable.addElement(Name_s);
		}
		else
		{   // Replace entry
		    codebaseTable.setElementAt(codebase, index);
		    serializationTable.setElementAt(serclass, index);
		    startpointTable.setElementAt(objectStartpoint, index);
		    objectNameTable.setElementAt(objectName_s, index);
		}

		// Return empty message, no exceptions should occur at the registry
		ser.initWrite();
		ser.startWriteHeader();
		ser.writeHPCxxHeaderHandlerid(REBIND_HANDLER_ID);
		ser.startWriteData();
		nxs.send_rsr((PutBuffer)ser.getBuffer(), repSP, RETURN_HANDLER_ID, false);
		repSP.destroy();
		break;
	    }
	case LOOKUP_HANDLER_ID:
	    {
		// Return the table entry corresponding to the given name.
		// Return NotBoundException if no entry with the given name exists.
 		String Name_s = getName((String)ser.readFinalObject(java.lang.String.class, null, null));
		if(verbose)
		    registryOut("Looking up: " + Name_s);

		int indexOfEntry = getIndex(Name_s);
		ser.initWrite();
		ser.startWriteHeader();

		if(indexOfEntry < 0)
		{   // Return exception
		    String ex = "nexusrmi.NotBoundException";

		    ser.writeHPCxxHeaderHandlerid(EXCEPTION_HANDLER_ID);
		    ser.startWriteData();
 		    ser.writeFinalObject(java.lang.String.class, ex);

		    nxs.send_rsr((PutBuffer)ser.getBuffer(), repSP, RETURN_HANDLER_ID, false);
		}
		else
		{   // Return stub information
		    String codebase = (String)codebaseTable.elementAt(indexOfEntry);
		    String serclass = (String)serializationTable.elementAt(indexOfEntry);
		    Startpoint skelStartpoint = (Startpoint)startpointTable.elementAt(indexOfEntry);
		    String NameToSend = (String)NameTable.elementAt(indexOfEntry);
		    String objectNameToSend = (String)objectNameTable.elementAt(indexOfEntry);

		    ser.writeHPCxxHeaderHandlerid(LOOKUP_HANDLER_ID);
		    ser.startWriteData();
 		    ser.writeFinalObject(java.lang.String.class, serclass);
		    ser.writeFinalObject(java.lang.String.class, codebase);
		    ser.writeFinalObject(nexus.Startpoint.class, skelStartpoint);
 		    ser.writeFinalObject(java.lang.String.class, objectNameToSend);

		    nxs.send_rsr((PutBuffer)ser.getBuffer(), repSP, RETURN_HANDLER_ID, false);
		}
		repSP.destroy();
		break;
	    }
	case BIND_HANDLER_ID:
	    {
		// Store the entry from the buffer in the table. If an object with
		// the same name already exists, return AlreadyBoundException..
 		String serclass = (String)ser.readFinalObject(java.lang.String.class, null, null);
		String codebase = (String)ser.readFinalObject(java.lang.String.class, null, null);
		Startpoint objectStartpoint = (Startpoint)ser.readFinalObject(nexus.Startpoint.class, null, null);
 		String objectName_s = (String)ser.readFinalObject(java.lang.String.class, null, null);
 		String Name_s = getName((String)ser.readFinalObject(java.lang.String.class, null, null));
		// Juan
		registryOut("Storing in registry: " + Name_s);

		int index = NameTable.indexOf(Name_s);
		ser.initWrite();
		ser.startWriteHeader();
		
		if(index < 0)
		{   // Store the entry, return empty message
		    codebaseTable.addElement(codebase);
		    serializationTable.addElement(serclass);
		    startpointTable.addElement(objectStartpoint);
		    objectNameTable.addElement(objectName_s);
		    NameTable.addElement(Name_s);

		    ser.writeHPCxxHeaderHandlerid(BIND_HANDLER_ID);
		    ser.startWriteData();
		    nxs.send_rsr((PutBuffer)ser.getBuffer(), repSP, RETURN_HANDLER_ID, false);
		}
		else
		{   // Return exception
		    String ex = "nexusrmi.AlreadyBoundException";

		    ser.writeHPCxxHeaderHandlerid(EXCEPTION_HANDLER_ID);
		    ser.startWriteData();
 		    ser.writeFinalObject(java.lang.String.class, ex);

		    nxs.send_rsr((PutBuffer)ser.getBuffer(), repSP, RETURN_HANDLER_ID, false);
		}
		repSP.destroy();
		break;
	    }
	case UNBIND_HANDLER_ID: // unbind
	    {
 		String Name_s = getName((String)ser.readFinalObject(java.lang.String.class, null, null));
		if(verbose)
		    registryOut("Removing from registry: " + Name_s);

		int indexOfEntry = getIndex(Name_s);
		ser.initWrite();
		ser.startWriteHeader();
		
		if(indexOfEntry < 0)
		{
		    String ex = "nexusrmi.NotBoundException";

		    ser.writeHPCxxHeaderHandlerid(EXCEPTION_HANDLER_ID);
		    ser.startWriteData();
 		    ser.writeFinalObject(java.lang.String.class, ex);

		    nxs.send_rsr((PutBuffer)ser.getBuffer(), repSP, RETURN_HANDLER_ID, false);
		}
		else
		{
		    serializationTable.removeElementAt(indexOfEntry);
		    startpointTable.removeElementAt(indexOfEntry);
		    NameTable.removeElementAt(indexOfEntry);
		    objectNameTable.removeElementAt(indexOfEntry);

		    ser.writeHPCxxHeaderHandlerid(UNBIND_HANDLER_ID);
		    ser.startWriteData();
		    nxs.send_rsr((PutBuffer)ser.getBuffer(), repSP, RETURN_HANDLER_ID, false);
		}
		repSP.destroy();
		break;
	    }
	case LIST_HANDLER_ID:
	    {
		// Return a String array consisting of the names of all objects stored in the registry.
		if(verbose)
		    registryOut("Returning list of bound objects.");

		int nrEntries = NameTable.size();

		ser.initWrite();
		ser.startWriteHeader();
		ser.writeHPCxxHeaderHandlerid(LIST_HANDLER_ID);
		ser.startWriteData();
		ser.writeint(nrEntries);
		for(int i = 0; i < nrEntries; i++)
 		    ser.writeFinalObject(java.lang.String.class, (String)NameTable.elementAt(i));

		nxs.send_rsr((PutBuffer)ser.getBuffer(), repSP, RETURN_HANDLER_ID, false);
		repSP.destroy();
		break;
	    }
	default :
	    // Should not happen.
	    registryErr("Failed to get the correct handler id " + handler_id);
	    break;
	}
    }

    // Get the index of the entry containing the object with the given name.
    public int getIndex(String urlString)
    {
	for(int i =0; i < NameTable.size(); i++)
	    if ( urlString.equals(NameTable.elementAt(i)))
		return i;

	return -1;
    }

    private synchronized void waitForRequest()
    {
	while(!RecdFromClient)
	{
	    try
	    {
		wait();
	    }
	    catch(Exception e)
	    {
		e.printStackTrace();
		//System.exit(1);
	    }
	}
    }
}
