/* 
 * NexusRMI
 *
 * Copyright (c) 1999
 * Fabian Breg
 * Indiana University/Leiden University
 * All Rights Reserved
 *
 * Please refer to the LICENSE file provided in this
 * distribution for full licensing terms and
 * conditions.
 *
 * Please report all bugs/problems to
 * breg@cs.indiana.edu
 *
 ******************************************************
 *
 * RegistryStub.java
 *
 */


package nexusrmi.registry;

import java.net.*;
import java.lang.reflect.*;
import nexus.Nexus;
import nexus.AttachApprovalInterface;
import nexus.AttachReturn;
import nexus.HandlerInterface;
import nexus.Endpoint;
import nexus.EndpointAttr;
import nexus.Startpoint;
import nexus.GetBuffer;
import nexus.PutBuffer;
import nexusrmi.*;
import nexusrmi.server.*;
import nexusrmi.registry.*;

class RegistryStub extends RemoteStub implements Registry, RegistryHandler
{
    public static final int REGISTRY_PORT = 7000;

    // These numbers should correspond to the numbers in the RMIRegistry class.
    private final int GF_HANDLER_ID = 0;
    private final int RETURN_HANDLER_ID = 5;
    private final int EXCEPTION_HANDLER_ID = 15;
    
    private final int LOOKUP_HANDLER_ID = 100;
    private final int REBIND_HANDLER_ID = 101;
    private final int BIND_HANDLER_ID = 102;
    private final int UNBIND_HANDLER_ID = 103;
    private final int LIST_HANDLER_ID = 104;

    private Class RemoteStubClass = null;

    // Create a registry stub by connecting to a registry (done in getStartpoint()).
    public RegistryStub(String host, int port) throws nexusrmi.UnknownHostException, NoRegistryException
    {
	super(Communication.getNexus(), null, "nexusrmi.Nexusize", null);
	try
	{
	    RemoteStubClass = Class.forName("nexusrmi.server.RemoteStub");
	    if(port < 0)
		port = REGISTRY_PORT;
	    
	    URL RegistryUrl = new URL("http", host, port, "");
	    stubStartpoint = Communication.getStartpoint(RegistryUrl);
	}
	catch(Exception e)
	{
	    e.printStackTrace();
	    //System.exit(1);
	}
	// We cannot return an error flag, so we throw an exception :-)
	if(stubStartpoint == null)
	    throw new NoRegistryException();
    }

    // The next five methods all communicate with the registry like any stub
    // communicates with a skeleton object. This method is described in the
    // RemoteStub class.
    
    // Return a stub to the given remote object
    public Remote lookup(String url) throws RemoteException, NotBoundException, AccessException
    {
	Object stub = null;
	int index = 0;
	URL cbaseurl = null;

	index = getFreeEntry();
	ReplyHandler lh = new ReplyHandler(nxs, this, index, serClass, codebase);
	EndpointAttr epa = nxs.endpointattr_init();
	epa.set_handler_object(lh);
	Endpoint ep = nxs.endpoint_init(epa);
	Startpoint repStartpoint = nxs.startpoint_bind(ep);

	Serialize ser = new Nexusize();
	ser.initWrite();
	ser.startWriteHeader();
        ser.writeHPCxxHeaderStartpoint(repStartpoint);
        ser.writeHPCxxHeaderOffset(0);
        ser.writeHPCxxHeaderHandlerid(LOOKUP_HANDLER_ID);
	ser.startWriteData();
	ser.writeFinalObject(java.lang.String.class, url);

	nxs.send_rsr((PutBuffer)ser.getBuffer(), stubStartpoint, GF_HANDLER_ID, false);
	waitForReply(index);

	if(ex[index] != null)
	{
	    if(ex[index] instanceof NotBoundException)
		throw (NotBoundException)ex[index];
	    else
		throw new RemoteException("lookup failed", ex[index]);
	}

	Startpoint serverSP = (Startpoint)(((Object[])result[index])[0]);
	String stubName = (String)(((Object[])result[index])[1]);
	String serProt = (String)(((Object[])result[index])[2]);
	String cbase = (String)(((Object[])result[index])[3]);
	try
	{
	    cbaseurl = new URL(cbase);
	}
	catch(Exception e)
	{
	    cbaseurl = null;
	}
	try
	{
	    if(!RemoteStubClass.isAssignableFrom(RMIClassLoader.loadClass(cbaseurl, stubName)))
	        stubName += "_Stub";
	}
	catch(Exception e)
	{
	    e.printStackTrace();
	}
	// Create the stub from the received information
	try
	{
	    stub = UnicastRemoteObject.createStub(stubName, serverSP, serProt, cbase);
	}
	catch(Exception e)
	{
	    e.printStackTrace();
	    //System.exit(1);
	}
	repStartpoint.destroy();
	ep.destroy();
	return (Remote)stub;
    }
    
    public void bind(String url, Remote obj) throws RemoteException, AlreadyBoundException, AccessException
    {
	int index = 0;

	index = getFreeEntry();
	ReplyHandler bh = new ReplyHandler(nxs, this, index, serClass, codebase);
	EndpointAttr epa = nxs.endpointattr_init();
	epa.set_handler_object(bh);
	Endpoint ep = nxs.endpoint_init(epa);
	Startpoint repStartpoint = nxs.startpoint_bind(ep);

	// Get a startpoint for the remote object from the UnicastRemoteObject table
	Startpoint skelStartpoint = UnicastRemoteObject.getStartpoint(obj);
	// Get the serialization class name from the UnicastRemoteObject table
	String serclass = UnicastRemoteObject.getSerClass(obj);
	// Get fully qualified classname
	String objName = obj.getClass().getName();
	// Get the codebase
	String cbase = UnicastRemoteObject.getCodebase(obj);

	Serialize ser = new Nexusize();
	ser.initWrite();
	ser.startWriteHeader();
        ser.writeHPCxxHeaderStartpoint(repStartpoint);
        ser.writeHPCxxHeaderOffset(0);
        ser.writeHPCxxHeaderHandlerid(BIND_HANDLER_ID);
	ser.startWriteData();
	ser.writeFinalObject(java.lang.String.class, serclass);
	ser.writeFinalObject(java.lang.String.class, cbase);
	ser.writeFinalObject(nexus.Startpoint.class, skelStartpoint);
	ser.writeFinalObject(java.lang.String.class, objName);
	ser.writeFinalObject(java.lang.String.class, url);
	    
	nxs.send_rsr((PutBuffer)ser.getBuffer(), stubStartpoint, GF_HANDLER_ID, false);
	waitForReply(index);
	
	if(ex[index] != null)
	{
	    if(ex[index] instanceof AlreadyBoundException)
		throw (AlreadyBoundException)ex[index];
	    else
		throw new RemoteException("bind failed", ex[index]);
	}
	repStartpoint.destroy();
	ep.destroy();
    }
    
    public void unbind(String url) throws RemoteException, NotBoundException, AccessException
    {
	int index = 0;
	
	index = getFreeEntry();
	ReplyHandler uh = new ReplyHandler(nxs, this, index, serClass, codebase);
	EndpointAttr epa = nxs.endpointattr_init();
	epa.set_handler_object(uh);
	Endpoint ep = nxs.endpoint_init(epa);
	Startpoint repStartpoint = nxs.startpoint_bind(ep);

	Serialize ser = new Nexusize();
	ser.initWrite();
	ser.startWriteHeader();
        ser.writeHPCxxHeaderStartpoint(repStartpoint);
        ser.writeHPCxxHeaderOffset(0);
        ser.writeHPCxxHeaderHandlerid(UNBIND_HANDLER_ID);
	ser.startWriteData();
	ser.writeFinalObject(java.lang.String.class, url);

	nxs.send_rsr((PutBuffer)ser.getBuffer(), stubStartpoint, GF_HANDLER_ID, false);
	waitForReply(index);

	if(ex[index] != null)
	{
	    if(ex[index] instanceof NotBoundException)
		throw (NotBoundException)ex[index];
	    else
		throw new RemoteException("unbind failed", ex[index]);
	}
	repStartpoint.destroy();
	ep.destroy();
    }
    
    public void rebind(String url, Remote obj) throws RemoteException, AccessException
    {
	int index = 0;
	
	index = getFreeEntry();
	ReplyHandler rh = new ReplyHandler(nxs, this, index, serClass, codebase);
	EndpointAttr epa = nxs.endpointattr_init();
	epa.set_handler_object(rh);
	Endpoint ep = nxs.endpoint_init(epa);
	Startpoint repStartpoint = nxs.startpoint_bind(ep);

	// Get startpoint for the remote object from the UnicastRemoteObject table
	Startpoint skelStartpoint = UnicastRemoteObject.getStartpoint(obj);
	// Get the serialization class name from the UnicastRemoteObject table
	String serclass = UnicastRemoteObject.getSerClass(obj);
	// Get fully qualified classname for the remote object
	String objName = obj.getClass().getName();
	// Get codebase
	String codebase = UnicastRemoteObject.getCodebase(obj);

	Serialize ser = new Nexusize();
	ser.initWrite();
	ser.startWriteHeader();
        ser.writeHPCxxHeaderStartpoint(repStartpoint);
        ser.writeHPCxxHeaderOffset(0);
        ser.writeHPCxxHeaderHandlerid(REBIND_HANDLER_ID);
	ser.startWriteData();
	ser.writeFinalObject(java.lang.String.class, serclass);
	ser.writeFinalObject(java.lang.String.class, codebase);
	ser.writeFinalObject(nexus.Startpoint.class, skelStartpoint);
	ser.writeFinalObject(java.lang.String.class, objName);
	ser.writeFinalObject(java.lang.String.class, url);

	nxs.send_rsr((PutBuffer)ser.getBuffer(), stubStartpoint, GF_HANDLER_ID, false);
	waitForReply(index);

	if(ex[index] != null)
	    throw new RemoteException("Rebind failed", ex[index]);

	repStartpoint.destroy();
	ep.destroy();
    }
    
    public String[] list() throws RemoteException, AccessException
    {
	int index = 0;
	
	index = getFreeEntry();
	ReplyHandler lh = new ReplyHandler(nxs, this, index, serClass, codebase);
	EndpointAttr epa = nxs.endpointattr_init();
	epa.set_handler_object(lh);
	Endpoint ep = nxs.endpoint_init(epa);
	Startpoint repStartpoint = nxs.startpoint_bind(ep);

	Serialize ser = new Nexusize();
	ser.initWrite();
	ser.startWriteHeader();
        ser.writeHPCxxHeaderStartpoint(repStartpoint);
        ser.writeHPCxxHeaderOffset(0);
        ser.writeHPCxxHeaderHandlerid(LIST_HANDLER_ID);
	ser.startWriteData();
	nxs.send_rsr((PutBuffer)ser.getBuffer(), stubStartpoint, GF_HANDLER_ID, false);
	waitForReply(index);

	if(ex[index] != null)
	    throw new RemoteException("list failed", ex[index]);

	repStartpoint.destroy();
	ep.destroy();
	return (String[])result[index];
    }

    // These are the methods from the RegistryHandler interface.
    // Not sure whether the guys from Sun had this in mind, however...
    // These methods create a RegistryStub and return it.
    public Registry registryStub(String host, int port) throws nexusrmi.UnknownHostException, RemoteException
    {
	Registry reg = null;

	try
	{
	    reg = new RegistryStub(host, port);
	}
	catch(NoRegistryException e)
	{
	    e.printStackTrace();
	    //System.exit(1);
	}
	return reg;
    }

    public Registry registryImpl(int port) throws RemoteException
    {
	Registry reg = null;
	try
	{
	    reg = registryStub(InetAddress.getLocalHost().getHostName(), port);
	}
	catch(RemoteException e)
	{
	    throw e;
	}
	catch(Exception e)
	{
	    e.printStackTrace();
	    //System.exit(1);
	}
	return reg;
    }

    protected void finalize() throws Throwable
    {
	super.finalize();
    }
}

// The handler object for RegistryStub objects
// For a description on how to use this object see the RemoteStub class
class ReplyHandler implements HandlerInterface
{
    // Again, these numbers should correspond to the same definitions in other files.
    private final int GF_HANDLER_ID = 0;
    private final int RETURN_HANDLER_ID = 5;
    private final int EXCEPTION_HANDLER_ID = 15;
    
    private final int LOOKUP_HANDLER_ID = 100;
    private final int REBIND_HANDLER_ID = 101;
    private final int BIND_HANDLER_ID = 102;
    private final int UNBIND_HANDLER_ID = 103;
    private final int LIST_HANDLER_ID = 104;
    
    private RegistryStub ri;
    private int index;
    private Nexus nxs;
    private Class serClass;
    private String codebase;
    
    public ReplyHandler(Nexus nxs, RegistryStub ri, int index, Class ser, String codebase)
    {
	this.nxs = nxs;
	this.ri = ri;
	this.index = index;
	this.serClass = ser;
	this.codebase = codebase;
    }

    public void invoke_handler(Endpoint ep, GetBuffer buffer, int hid)
    {
	Serialize ser = new Nexusize();
	ser.initRead(buffer, codebase);
	ser.startReadHeader();
	int handler_id = ser.readHPCxxHeaderHandlerid();
	ser.startReadData();

	switch(handler_id)
	{
	case LOOKUP_HANDLER_ID:
	    {
		Object[] dummy = new Object[4];

 		dummy[2] = (String)ser.readFinalObject(java.lang.String.class, null, null);
		dummy[3] = (String)ser.readFinalObject(java.lang.String.class, null, null);
		dummy[0] = (Startpoint)ser.readFinalObject(nexus.Startpoint.class, null, null);
 		dummy[1] = (String)ser.readFinalObject(java.lang.String.class, null, null);

		ri.setResult(index, dummy);
	    }
	    break;
	case REBIND_HANDLER_ID:
	    {
		ri.setResult(index, null);
	    }
	    break;
	case BIND_HANDLER_ID:
	    {
		ri.setResult(index, null);
	    }
	    break;
	case UNBIND_HANDLER_ID:
	    {
		ri.setResult(index, null);
	    }
	    break;
	case LIST_HANDLER_ID:
	    {
		int n = ser.readint();
		String[] res = new String[n];
		for(int i = 0; i < n; i++)
 		    res[i] = (String)ser.readFinalObject(java.lang.String.class, null, null);
		ri.setResult(index, res);
	    }
	    break;
	case EXCEPTION_HANDLER_ID:
	    {
		try
		{
 		    String ExName = (String)ser.readFinalObject(java.lang.String.class, null, null);
		    Class ExClass = Class.forName(ExName);
		    ri.setException(index, (Exception)ExClass.newInstance());
		}
		catch(Exception e)
		{
		    e.printStackTrace();
		    //System.exit(1);
		}
	    }
	    break;
	default:
	    // Should not happen
	    System.err.println("Invalid handler id="+handler_id);
	}
    }
}
