/* 
 * NexusRMI
 *
 * Copyright (c) 1999
 * Fabian Breg
 * Indiana University/Leiden University
 * All Rights Reserved
 *
 * Please refer to the LICENSE file provided in this
 * distribution for full licensing terms and
 * conditions.
 *
 * Please report all bugs/problems to
 * breg@cs.indiana.edu
 *
 ******************************************************
 *
 * RemoteStrub.java
 *
 */


package nexusrmi.server;

import nexus.Nexus;
import nexus.Startpoint;
import nexusrmi.*;

// This class is the superclass of all stubs
// --------------------------------------------------------------------
// In general a stub works as follows:
// The stub contains methods that can be invoked on the corresponding
// remote object. Each such method sends a message containing method ID
// and parameters to the associated remote object and receives a reply
// that it must return or an exception that must be thrown. To allow
// multiple threads to concurrently execute a remote method, a table
// is maintained at the stub, which contains the status of the invocation,
// the returned result if available and the returned exception if
// available in the status, result and ex array respectively. Each
// method in the stub consists of the following steps:
//
// - Obtain a free entry in the table, using the getFreeEntry() method
//   from this class.
// - Create a handler that handles the reply (described later).
// - Create a startpoint for that handler using the getStartpoint()
//   method from the Communication class.
// - Create and fill a PutBuffer with the information to be sent to
//   the remote object.
// - Send the buffer with send_rsr() from the Nexus class.
// - Invoke waitForReply() from this class to wait for the reply.
// - If a reply has arrived, check the ex-array at the obtained index
//   to check whether a result or exception is returned. Only one of
//   the ex or result entry should be set; the other should be null.
// - Release the table entry with freeEntry() from this class.
// - Throw the exception or return the result.
//
// The handler object that handles the reply has to be an instance of
// a class that implements the HandlerInterface and thus contains an
// invoke_handler() method. This invoke_handler() method typically
// contains a switch statement with a case label for every method id
// for which it can receive a reply, as well as a special exception
// handler case. Within each method case, this object extracts the
// return object, which it returns to the stub with the setResult()
// method from the RemoteStub class. This method also wakes up the
// stub. The exception handler extracts the strings describing the
// exception and uses the reflection classes to create the exception
// object, which it returns to the client using the setException()
// method from this class, which also wakes up the stub.
// --------------------------------------------------------------------

public abstract class RemoteStub extends RemoteObject
{
    private final int MAX_ENTRIES = 10;
    private final int FREE = 0;
    private final int BUSY = 1;
    private final int READY = 2;

    protected Class serClass;
    protected String codebase;
    protected Startpoint stubStartpoint;

    protected Nexus nxs;
    protected Object[] result;
    protected Exception[] ex;
    protected int[] status;
    
    protected RemoteStub(String ser, String codebase)
    {
    	super();
	try
	{
	    this.serClass = Class.forName(ser);
	}
	catch(Exception e)
	{
	    e.printStackTrace();
	    //System.exit(1);
	}
	this.codebase = codebase;
    }

    // These two methods are part of the original RMI and are
    // not used in this implementation.
    protected RemoteStub(RemoteRef ref, String ser)
    {
        super(ref);
	try
	{
	    this.serClass = Class.forName(ser);
	}
	catch(Exception e)
	{
	    e.printStackTrace();
	    //System.exit(1);
	}
    }

    static protected void setRef(RemoteStub stub, RemoteRef ref)
    {
	stub.ref = ref;
    }

    // This is the constructor that we use.
    public RemoteStub(Nexus nxs, Startpoint sp, String ser, String codebase)
    {
	super();
	this.nxs = nxs;
	this.stubStartpoint = sp;
	try
	{
	    this.serClass = Class.forName(ser);
	}
	catch(Exception e)
	{
	    e.printStackTrace();
	    //System.exit(1);
	}
	this.codebase = codebase;
	
	result = new Object[MAX_ENTRIES];
	ex = new Exception[MAX_ENTRIES];
	status = new int[MAX_ENTRIES];
    }

    
    // Block until a free entry can be obtained from the table and than
    // return the index of that entry
    protected synchronized int getFreeEntry()
    {
	int i = 0;
	boolean found = false;

	try
	{
	    while(!found)
	    {
		while((i < MAX_ENTRIES) && (status[i] != FREE))
		    i++;
		
		if(i == MAX_ENTRIES)
		    wait();
		else
		    found = true;
	    }
	    
	    result[i] = null;
	    ex[i] = null;
	    status[i] = BUSY;
	}
	catch(InterruptedException e)
	{
	    e.printStackTrace();
	    //System.exit(1);
	}
	return i;
    }

    // Block until the handler object has received the reply message and
    // has set the return object.
    protected synchronized void waitForReply(int index)
    {
	try
	{
	    while(status[index] != READY)
		wait();
	}
	catch(InterruptedException e)
	{
	    e.printStackTrace();
	    //System.exit(1);
	}
    }

    // Free the given table entry.
    public synchronized void freeEntry(int index)
    {
	result[index] = null;
	ex[index] = null;
	status[index] = FREE;
	notifyAll();
    }

    // Put the exception object in the table and notify stub.
    public synchronized void setException(int index, Exception e)
    {
	result[index] = null;
	ex[index] = e;
	status[index] = READY;
	notifyAll();
    }

    // Could be used by the stub to obtain the exception, but
    // currently it just executes the array indexing itself :-)
    public Exception getException(int index)
    {
	return ex[index];
    }

    // Put the return object in the table and notify the stub.
    public synchronized void setResult(int index, Object res)
    {
	result[index] = res;
	ex[index] = null;
	status[index] = READY;
	notifyAll();
    }

    // Could be used by the stub to obtain the result, but
    // currently it just executes the array indexing itself :-)
    public Object getResult(int index)
    {
	return result[index];
    }

    public Startpoint getStubStartpoint()
    {
	return stubStartpoint;
    }

    public String getCodebase()
    {
	return codebase;
    }

    public Class getSerClass()
    {
	return serClass;
    }
    
    public boolean equals(RemoteStub other)
    {
	boolean res = true;
	if(other == null)
	    return false;
	
	Startpoint osp = other.getStubStartpoint();
	if((stubStartpoint == null) && (osp != null))
	    return false;
	if(!stubStartpoint.equals(osp))
	    return false;

	String ocb = other.getCodebase();
	if((codebase == null) && (ocb != null))
	    return false;
	if(!codebase.equals(ocb))
	    return false;

	Class osc = other.getSerClass();
	if((serClass == null) && (osc != null))
	    return false;
	if(serClass.equals(osc))
	    return false;

	return true;
    }

    protected void finalize() throws Throwable
    {
	super.finalize();
	stubStartpoint.destroy();
    }
}
