/* 
 * NexusRMI
 *
 * Copyright (c) 1999
 * Fabian Breg
 * Indiana University/Leiden University
 * All Rights Reserved
 *
 * Please refer to the LICENSE file provided in this
 * distribution for full licensing terms and
 * conditions.
 *
 * Please report all bugs/problems to
 * breg@cs.indiana.edu
 *
 ******************************************************
 *
 * javas.c
 *
 */


#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <assert.h>
#include <errno.h>

#include "class.h"

/* These are the specific actions that can be performed on a parameter of returnvalue */
#define DUMP_FORMAL    0
#define DUMP_TYPE      1
#define DUMP_SIZEOF    2
#define DUMP_MARSHAL   3
#define DUMP_CAST      4
#define DUMP_STUB_UNM  5
#define DUMP_SKEL_UNM  6
#define DUMP_ACTUAL    7
#define DUMP_DECL      8

#define BASE_HANDLER_ID 200

/* The type names for the primitive types */
static char *type_name[] = {"void", "boolean", "byte", "short", "int",     "long", "float", "double", "char",       "????", "????"};
/* The wrapper names for the primitive types */
static char *wrap_name[] = {"????", "Boolean", "Byte", "Short", "Integer", "Long", "Float", "Double", "Character",  "????", "????"};
/* Initializers for all types */
static char *type_init[] = {"????", "true",    "0",    "0",     "0",       "0",    "0.0f",  "0.0",    "'\\0'",      "null", "null"};

static char *full_name;
static char *pack_name;
static char *server_name;
static char *stub_name;
static char *stub_file;
static char *skel_name;
static char *skel_file;

static FILE *fp;
static int handler_id;
static int base_handler_id;
static int handler_id_offset = 0;

short verbose;

/* miscellaneous routines
   ********************** */

void *alloc_mem(size_t nbytes)
{
    void *mem = malloc(nbytes);
    if(!mem)
    {
	fprintf(stderr, "Out of memory.");
	exit(EXIT_FAILURE);
    }
    return mem;
}

void *realloc_mem(void *oldmem, size_t nbytes)
{
    void *newmem = realloc(oldmem, nbytes);
    if(!newmem)
    {
	fprintf(stderr, "Out of memory.");
	exit(EXIT_FAILURE);
    }
    return newmem;
}

/* Return a string consisting of the given amount of spaces */
static char *space(int n)
{
    static char *string = NULL;

    int i;

    if(string)
	string = (char *)realloc(string, (n + 1) * sizeof(char));
    else
	string = (char *)malloc((n + 1) * sizeof(char));
    
    for(i = 0; i < n; i++)
	string[i] = ' ';
    string[i] = '\0';
    return string;
}

/* Convert byte code representation for types to a number */
static int get_type_id(char *t)
{
    switch(*t)
    {
    case 'V': return 0;
    case 'Z': return 1;
    case 'B': return 2;
    case 'S': return 3;
    case 'I': return 4;
    case 'J': return 5;
    case 'F': return 6;
    case 'D': return 7;
    case 'C': return 8;
    case 'L': return 9;
    case '[': return 10;
    }
    javas_out(-1, "Unexpected fall-through in get_type_id().");
    return -1;
}

/* Advance to the next parameter. */
static void next_param(char **p)
{
    if(**p == 'L')
    {
	(*p)++;
	while(**p != ';')
	    (*p)++;
	
	(*p)++;
    }
    else /* **p == '[' */
    {
	while(**p == '[')
	    (*p)++;
	if(**p == 'L')
	    next_param(p);
	else
	    (*p)++;
    }
}

/* Bytecode uses a slash (/) as a separator in fully qualified names,
   while Java likes dots (.) better.
*/
static char *replace_slash(char *s)
{
    int i;
    char *res = (char *)strdup(s);
    
    for(i = 0; i < strlen(res); i++)
	if(res[i] == '/')
	    res[i] = '.';
    return res;
}

/* This will construct the various names from the given parameters
   string representation.
*/
static void construct_names(void)
{
    int i = 0, last_point = 0;

    pack_name = NULL;
    server_name = NULL;
    stub_name = NULL;
    stub_file = NULL;
    skel_name = NULL;
    skel_file = NULL;
    
    while(i < strlen(full_name))
    {
	if(full_name[i] == '.')
	    last_point = i;

	i++;
    }

    if(last_point == 0)
	server_name = (char *)strdup(full_name);
    else
    {
	pack_name = (char *)malloc((last_point + 1)* sizeof(char));
	strncpy(pack_name, full_name, last_point);
	pack_name[last_point] = '\0';
	server_name = (char *)malloc(strlen(full_name) - last_point);
	for(i = last_point + 1; i < strlen(full_name); i++)
	    server_name[i - last_point - 1] = full_name[i];
	server_name[i - last_point - 1] = '\0';
    }

    stub_name = (char *)malloc((strlen(server_name) + 6) * sizeof(char));
    sprintf(stub_name, "%s%s", server_name, "_Stub");
    stub_file = (char *)malloc((strlen(stub_name) + 6) * sizeof(char));
    sprintf(stub_file, "%s%s", stub_name, ".java");
    skel_name = (char *)malloc((strlen(server_name) + 6) * sizeof(char));
    sprintf(skel_name, "%s%s", server_name, "_Skel");
    skel_file = (char *)malloc((strlen(skel_name) + 6) * sizeof(char));
    sprintf(skel_file, "%s%s", skel_name, ".java");
}

/* Remove some names */
static void delete_names(void)
{
    if(pack_name)
	free(pack_name);
    if(server_name)
	free(server_name);
    if(stub_name)
	free(stub_name);
    if(stub_file)
	free(stub_file);
    if(skel_name)
	free(skel_name);
    if(skel_file)
	free(skel_file);
}

/* Get the symbolic representation for the handler identifier for a method.
   c contains the class containing the method m represented by an fm_ptr.
   The handler identification number is still needed to construct unique
   names in case of method overloading.
*/
static char *get_handler_name(char *c, fm_ptr m, int handler_id)
{
    static char *string = NULL;
    int i;
    char *orig, *suffix = "_HANDLER_ID";

    orig = (char *)strdup(get_name(c, m));
    for(i = 0; i < strlen(orig); i++)
	orig[i] = toupper(orig[i]);

    if(string)
	string = (char *)realloc(string, (strlen(orig) + strlen(suffix) + 12) * sizeof(char));
    else
	string = (char *)malloc((strlen(orig) + strlen(suffix) + 12) * sizeof(char));
    
    sprintf(string, "%s%s_%d%c", orig, suffix, handler_id, '\0');
    free(orig);
    return string;
}

/* Extract the class name for the current object reference */
static char *get_class_name(char *p)
{
    static char *string = NULL;
    int i, len = 1;

    i = 0;
    while(p[i++] != ';')
	len++;

    if(string)
	string = (char *)realloc(string, len * sizeof(char));
    else
	string = (char *)malloc(len * sizeof(char));

    i = 0;
    while(p[i] != ';')
    {
	if(p[i] == '/')
	    string[i] = '.';
        else
	    string[i] = p[i];
	i++;
    }
    string[i] = '\0';
    return string;
}

/* Generate an expression that allocates space for the first dimension of an array,
   the size of which will be specified by parameter l<al>, where al indicates the
   dimension to be allocated. It will generate something like "classname[l1][][]".
*/
static char *gen_array_type(char *p)
{
    static char *string = NULL;
    char *elt = NULL;
    int dims = 0;
    int i;

    while(*p == '[')
    {
	dims++;
	p++;
    }

    if(*p == 'L')
	elt = (char *)strdup(get_class_name(++p));
    else
	elt = (char *)strdup(type_name[get_type_id(p)]);

    if(string)
	string = (char *)realloc(string, (strlen(elt) + (dims * 2) + 1) * sizeof(char));
    else
	string = (char *)malloc((strlen(elt) + (dims * 2) + 1) * sizeof(char));

    sprintf(string, "%s", elt);

    for(i = 0; i < dims; i++)
    {
	strcat(string, "[]");
    }

    free(elt);
    return string;
}

/* Check whether the given class is serializable */
static int is_nexusable(char *class)
{
    int i = 0;
    char *cn = (char *)strdup(class);;

    while(cn[i] != ';')
	i++;
    cn[i] = '\0';
    return implements(cn, "java/io/Serializable");
}

/* Dump the name of remote interface i contained in class 'class'.
   The 'first' parameter determines whether a comma is to be
   prepended or not.
*/
static int dump_remote_interface(char *class, char *i, int first)
{
    char *n;

    if(!strcmp(i, "nexusrmi/Remote") || implements(i, "nexusrmi/Remote"))
    {
	if(!first)
	    fputs(", ", fp);
	
	n = replace_slash(i);
	fputs(n, fp);
	free(n);
	return 0;
    }
    else
	return first;
}

/* Iterate over all interfaces that the given class and its superclass
   implements and execute the given function.
*/
static void intf_traverse(char *class, int (*action)(char *, char *, int))
{
    char **implements;
    int cur_imps = 10, nr_imps = 0, implemented, n;
    int first = 1, j;
    char *i;

    implements = (char **)malloc(cur_imps * sizeof(char *));
    while(class)
    {
	trav_class_init(class, 0, 0, 1);
	while(i = get_next_interface())
	{
	    implemented = 0;
	    n = 0;
	    while(!implemented && n < nr_imps)
	    {
		if(!strcmp(i, implements[n]))
		    implemented = 1;
		n++;
	    }
	    if(!implemented)
	    {
		if(nr_imps == cur_imps)
		{
		    cur_imps += 10;
		    implements = (char **)realloc(implements, cur_imps * sizeof(char *));
		}
		implements[nr_imps] = (char *)strdup(i);
		first = action(class, i, first);
		nr_imps++;
	    }
	}

	class = get_superclass(class);
    }
    for(j = 0; j < nr_imps; j++)
	free(implements[j]);
    free(implements);
}

/* Dump type specific code. The action indicates what kind of code to produce, p
   is points to the string representing the type of the current parameter of return
   value. After this method p will point to the next parameter if available or to
   nothing in case it pointed to a return value.
   The following are the representations of types in bytecode:
   V                           stands for void,
   Z                           stands for boolean,
   B                           stands for byte,
   S                           stands for short,
   I                           stands for integer,
   J                           stands for long,
   F                           stands for float,
   D                           stands for double,
   C                           stands for char,
   Lfully.qualified.classname; stands for object references
   [type                       stands for an array of type type, where type
                               is itself a type in bytecode representation.

   For instance:
   ICLjava.lang.String;[[D
   stands for a list of four parameters of types int, char, java.lang.String and double[][].
*/
static void handle_type(char **p, char *varname, int action, int indent, int al)
{
    int type_id;
    int dims;

    type_id = get_type_id(*p);
    
    switch(type_id)
    {
    case 0:
	assert(action == DUMP_TYPE);
	fprintf(fp, "%svoid", space(indent));
	break;
    case 1:
    case 2:
    case 3:
    case 4:
    case 5:
    case 6:
    case 7:
    case 8:
	switch(action)
	{
	case DUMP_TYPE:
	    fprintf(fp, "%s%s", space(indent), type_name[type_id]);
	    break;
	case DUMP_ACTUAL:
	    fprintf(fp, " %s", varname);
	    break;
	case DUMP_FORMAL:
	    fprintf(fp, "%s%s", space(indent), type_name[type_id]);
	    if(strcmp(varname, ""))
	       fprintf(fp, " ");
	    fprintf(fp, "%s", varname);
	    break;
	case DUMP_SIZEOF:
	    fprintf(fp, "%sbufsize += ser.sizeof%s(%s);\n", space(indent), type_name[type_id], varname);
	    break;
	case DUMP_MARSHAL:
	    fprintf(fp, "%sser.write%s(%s);\n", space(indent), type_name[type_id], varname);
	    break;
	case DUMP_CAST:
	    fprintf(fp, "((%s)r).%sValue()", wrap_name[type_id], type_name[type_id]);
	    break;
	case DUMP_STUB_UNM:
	    fprintf(fp, "%s%s = new %s(ser.read%s());\n", space(indent), varname, wrap_name[type_id], type_name[type_id]);
	    break;
	case DUMP_SKEL_UNM:
	    fprintf(fp, "%s%s = ser.read%s();\n", space(indent), varname, type_name[type_id]);
	    break;
	case DUMP_DECL:
	    fprintf(fp, "%s%s %s = %s;\n", space(indent), type_name[type_id], varname, type_init[type_id]);
	    break;
	}
	(*p)++;
	break;
    case 9:
	{
	    short final;
	    char *cn, *sc;
	    char *q = (*p) + 1;
	    
	    cn = (char *)strdup(q);
	    sc = strchr(cn, ';');
	    *sc = '\0';
	    final = is_final_class(cn);

	    switch(action)
	    {
	    case DUMP_TYPE:
		fprintf(fp, "%s%s", space(indent), get_class_name(q));
		break;
	    case DUMP_ACTUAL:
		fprintf(fp, "%s", varname);
		break;
	    case DUMP_FORMAL:
		fprintf(fp, "%s%s", space(indent), get_class_name(q));
		if(strcmp(varname, ""))
		    fprintf(fp, " ");
		fprintf(fp, "%s", varname);
		break;
	    case DUMP_CAST:
		if(!strncmp(*p, "java/lang/String", 16))
		    fprintf(fp, "(String)r");
		else /* if(is_nexusable(q)) */
		    fprintf(fp, "(%s)r", get_class_name(q));
		break;
	    case DUMP_DECL:
		fprintf(fp, "%s%s %s = %s;\n", space(indent), get_class_name(q), varname, type_init[type_id]);
		break;
	    case DUMP_SIZEOF:
		if(final)
		    fprintf(fp, "%sbufsize += ser.sizeofFinalObject(%s.class, %s);\n", space(indent), get_class_name(q), varname);
		else		      
		    fprintf(fp, "%sbufsize += ser.sizeofObject(%s);\n", space(indent), varname);
		break;
	    case DUMP_MARSHAL:
		if(final)
		    fprintf(fp, "%sser.writeFinalObject(%s.class, %s);\n", space(indent), get_class_name(q), varname);
		else
		    fprintf(fp, "%sser.writeObject(%s);\n", space(indent), varname);
		break;
	    case DUMP_STUB_UNM:
	    case DUMP_SKEL_UNM:
		if(final)
		    fprintf(fp, "%s%s = (%s)ser.readFinalObject(%s.class, null, null);\n", space(indent), varname, get_class_name(q), get_class_name(q));
		else
		    fprintf(fp, "%s%s = (%s)ser.readObject(null, null);\n", space(indent), varname, get_class_name(q));
		break;
	    }
	    free(cn);
	    next_param(p);
	}
	break;
    case 10:
	switch(action)
	{
	case DUMP_TYPE:
	    dims = 0;
	    while(**p == '[')
	    {
		(*p)++;
		dims++;
	    }
	    handle_type(p, varname, DUMP_TYPE, indent, al);
	    while(dims-- > 0)
		fprintf(fp, "[]");
	    break;
	case DUMP_ACTUAL:
	    fprintf(fp, " %s", varname);
	    while(**p == '[')
		(*p)++;

	    if(**p == 'L')
		next_param(p);
	    else
		(*p)++;
	    break;
	case DUMP_FORMAL:
	    dims = 0;
	    while(**p == '[')
	    {
		(*p)++;
		dims++;
	    }
	    handle_type(p, varname, DUMP_TYPE, indent, al);
	    while(dims-- > 0)
		fprintf(fp, "[]");

	    if(strcmp(varname, ""))
		fprintf(fp, " ");
	    fprintf(fp, "%s", varname);
	    break;
	case DUMP_SIZEOF:
	    fprintf(fp, "%sbufsize += ser.sizeofObject(%s);\n", space(indent), varname);
	    next_param(p);
	    break;
	case DUMP_MARSHAL:
	    fprintf(fp, "%sser.writeObject(%s);\n", space(indent), varname);
	    next_param(p);
	    break;
	case DUMP_CAST:
	    fprintf(fp, "(");
	    handle_type(p, varname, DUMP_TYPE, 0, al);
	    fprintf(fp, ")r");
	    break;
	case DUMP_STUB_UNM:
	case DUMP_SKEL_UNM: {
	    fprintf(fp, "%s%s = (%s)ser.readObject(null, null);\n", space(indent), varname, gen_array_type(*p));
	    next_param(p);
	    break;
	}
	case DUMP_DECL:
	    dims = 0;
	    while(**p == '[')
	    {
		(*p)++;
		dims++;
	    }
	    handle_type(p, varname, DUMP_TYPE, indent, al);
	    while(dims-- > 0)
		fprintf(fp, "[]");

	    fprintf(fp, " %s = null;\n", varname);
	    break;
	}
	break;
    }
}

/* Iterate over all parameters and perform action on them */
static void handle_parameters(char *param, int action, int indent, int dump_paramname)
{
    int i = 1;
    char *name;
    int first = 1;
    
    name = malloc((strlen("param") + 10) * sizeof(char));
    while(*param != ')')
    {
	if((action == DUMP_FORMAL) || (action == DUMP_ACTUAL))
	    if(!first)
		fprintf(fp, ", ");
	    else
		first = 0;
	if(dump_paramname)
	    sprintf(name, "param%d", i++);
	else
	    sprintf(name, "");
	handle_type(&param, name, action, indent, 0);
    }
    free(name);
}    

/* Perform action on the return type */
static void handle_return_type(char *return_type, int action, int indent)
{
    if((action == DUMP_MARSHAL) || (action == DUMP_SIZEOF))
	handle_type(&return_type, "result", action, indent, 0);
    else if(action == DUMP_STUB_UNM)
	handle_type(&return_type, "res", action, indent, 0);
    else
	handle_type(&return_type, "", action, indent, 0);
}

/* Dump the remote method header in the stub */
static void dump_method_header(char *name, char *param, char *return_type, char **ex, int i, int dump_paramname)
{
    int first = 1;
    
    fprintf(fp, "public ");
    handle_return_type(return_type, DUMP_TYPE, 0);
    fprintf(fp, " %s(", name);
    handle_parameters(param, DUMP_FORMAL, 0, dump_paramname);
    fprintf(fp, ") throws ");
    while(--i >= 0)
	if(strcmp(ex[i], "nexusrmi.RemoteException"))
	    if(first)
	    {
		fprintf(fp, "%s", ex[i]);
		first = 0;
	    }
	    else
		fprintf(fp, ", %s", ex[i]);

    if(first)
	fprintf(fp, "nexusrmi.RemoteException");
    else
	fprintf(fp, ", nexusrmi.RemoteException");
}


/* Dump code for all remote methods for the stub */
static void dump_stub_method(char *c, fm_ptr m, int handler_id)
{
    char *param;
    char *return_type;
    char *method_name;
    char *exs[100], *ex;
    int i = 0;
    
    method_name = get_name(c, m);
    param = get_typedescriptor(c, m);
    return_type = strrchr(param, ')') + 1;
    param++;
    while(ex = get_next_exception())
	exs[i++] = replace_slash(ex);

    fprintf(fp, "    ");
    dump_method_header(method_name, param, return_type, exs, i, 1);
    fprintf(fp, "\n");
    fprintf(fp, "    {\n");
    if(*return_type != 'V')
    {
	handle_return_type(return_type, DUMP_TYPE, 8);
	fprintf(fp, " res = %s;\n\n", type_init[get_type_id(return_type)]);
    }
    
    fprintf(fp, "        int index = getFreeEntry();\n");
    fprintf(fp, "        %sHandler handler = new %sHandler(nxs, this, index, serClass, codebase);\n", stub_name, stub_name);
    fprintf(fp, "        nexus.EndpointAttr epa = nxs.endpointattr_init();\n");
    fprintf(fp, "        epa.set_handler_object(handler);\n");
    fprintf(fp, "        nexus.Endpoint ep = nxs.endpoint_init(epa);\n");
    fprintf(fp, "        nexus.Startpoint sp = nxs.startpoint_bind(ep);\n");
    
    fprintf(fp, "        nexusrmi.Serialize ser = null;\n");
    fprintf(fp, "        try\n");
    fprintf(fp, "        {\n");
    fprintf(fp, "            ser = (nexusrmi.Serialize)serClass.newInstance();\n");
    fprintf(fp, "        }\n");
    fprintf(fp, "        catch(Exception e)\n");
    fprintf(fp, "        {\n");
    fprintf(fp, "            e.printStackTrace();\n");
    fprintf(fp, "            System.exit(1);\n");
    fprintf(fp, "        }\n");
    fprintf(fp, "        ser.initWrite();\n");
    fprintf(fp, "        ser.startWriteHeader();\n");
    fprintf(fp, "        ser.writeHPCxxHeaderStartpoint(sp);\n");
    fprintf(fp, "        ser.writeHPCxxHeaderOffset(0);\n");
    fprintf(fp, "        ser.writeHPCxxHeaderHandlerid(%s);\n", get_handler_name(c, m, handler_id));
    fprintf(fp, "        ser.startWriteData();\n");
/*     fprintf(fp, "        sp.serializeWrite(ser);\n"); */
/*     fprintf(fp, "        ser.writeint(0);\n"); */
/*     fprintf(fp, "        ser.writeint(%s);\n",  */
    handle_parameters(param, DUMP_MARSHAL, 8, 1);
    fprintf(fp, "        nxs.send_rsr((nexus.PutBuffer)ser.getBuffer(), stubStartpoint, GF_HANDLER_ID, true);\n");
    fprintf(fp, "        waitForReply(index);\n");
    fprintf(fp, "        if(ex[index] != null)\n");
    fprintf(fp, "        {\n");
    while(--i >= 0)
	if(strcmp(exs[i], "nexusrmi.RemoteException"))
	{
	    fprintf(fp, "            if(ex[index] instanceof %s)\n", exs[i]);
	    fprintf(fp, "                throw (%s)ex[index];\n", exs[i]);
	}
    
    fprintf(fp, "            if(ex[index] instanceof nexusrmi.RemoteException)\n");
    fprintf(fp, "                throw (nexusrmi.RemoteException)ex[index];\n");
    fprintf(fp, "        }\n");
    if(*return_type != 'V')
    {
	fprintf(fp, "        Object r = result[index];\n");
	fprintf(fp, "        res = ");
	handle_return_type(return_type, DUMP_CAST, 0);
	fprintf(fp, ";\n");
    }
    fprintf(fp, "        freeEntry(index);\n");
    fprintf(fp, "        sp.destroy();\n");
    fprintf(fp, "        ep.destroy();\n");
    if(*return_type != 'V')
	fprintf(fp, "        return res;\n");
    fprintf(fp, "    }\n\n");
}

/* Dump receive code for every remote method for the stub */
static void dump_stub_case(char *c, fm_ptr m, int handler_id)
{
    char *param;
    char *return_type;
    char *method_name;

    method_name = get_name(c, m);
    param = get_typedescriptor(c, m);
    return_type = strrchr(param, ')') + 1;
    param++;
    
    fprintf(fp, "            case %s:\n", get_handler_name(c, m, handler_id));
    fprintf(fp, "                {\n");

    if(*return_type != 'V')
    {
	switch(get_type_id(return_type))
	{
	case 1:
	case 2:
	case 3:
	case 4:
	case 5:
	case 6:
	case 7:
	case 8:
	    fprintf(fp, "                    %s res = null;\n", wrap_name[get_type_id(return_type)]);
	    break;
	case 9:
	    fprintf(fp, "                    %s res = null;\n", get_class_name(return_type + 1));
	    break;
	case 10:
	    handle_return_type(return_type, DUMP_TYPE, 20);
	    fprintf(fp, " res = null;\n");
	    break;
	}
	handle_return_type(return_type, DUMP_STUB_UNM, 20);
	fprintf(fp, "                    stub.setResult(index, res);\n");
    }
    else
	fprintf(fp, "                    stub.setResult(index, null);\n");
    fprintf(fp, "                }\n");
    fprintf(fp, "                break;\n");
}

/* Dump skeleton code for every remote method */
static void dump_skel_case(char *c, fm_ptr m, int handler_id)
{
    char *param;
    char *return_type;
    char *method_name;
    char *exs[100], *ex;
    int i = 0;
    
    method_name = get_name(c, m);
    param = get_typedescriptor(c, m);
    return_type = strrchr(param, ')') + 1;
    param++;
    while(ex = get_next_exception())
	exs[i++] = replace_slash(ex);
    
    fprintf(fp, "            case %s:\n", get_handler_name(c, m, handler_id));
    fprintf(fp, "                {\n");
    fprintf(fp, "                    %s serverObject = (%s)ep.get_user_object();\n", server_name, server_name);
    fprintf(fp, "                    if(doLogging)\n");
    fprintf(fp, "                        nexusrmi.server.RemoteServer.getLog().println(\"");
    dump_method_header(method_name, param, return_type, exs, i, 0);
    fprintf(fp, "\");\n");
    handle_parameters(param, DUMP_DECL, 20, 1);
    handle_parameters(param, DUMP_SKEL_UNM, 20, 1);
    fprintf(fp, "                    status = 1;\n");
    if(*return_type != 'V')
    {
	fprintf(fp, "                    ");
	handle_return_type(return_type, DUMP_TYPE, 0);
	fprintf(fp, " result = ");
    }
    else
	fprintf(fp, "                    ");

    fprintf(fp, "serverObject.%s(", method_name);
    handle_parameters(param, DUMP_ACTUAL, 20, 1);
    fprintf(fp, ");\n");

    fprintf(fp, "                    status = 0;\n\n");
    
    fprintf(fp, "                    ser.initWrite();\n");
    fprintf(fp, "                    ser.startWriteHeader();\n");
    fprintf(fp, "                    ser.writeHPCxxHeaderHandlerid(%s);\n", get_handler_name(c, m, handler_id));
    fprintf(fp, "                    ser.startWriteData();\n");

    if(*return_type != 'V')
	handle_return_type(return_type, DUMP_MARSHAL, 20);

    fprintf(fp, "                    nxs.send_rsr((nexus.PutBuffer)ser.getBuffer(), replySP, RETURNHANDLER_ID, true);\n");
    fprintf(fp, "                }\n");
    fprintf(fp, "                break;\n");
}

/* Dump a list of handler identifiers */
static void dump_handler_ids(char *c, fm_ptr m, int handler_id)
{
    fprintf(fp, "    private final int %s = %d;\n", get_handler_name(c, m, handler_id), handler_id);
}

static short is_remote_method(fm_ptr m, char *meth, char *sig, char *remclass)
{
    char *i;
    int is_remote_method;
	    
    while(remclass)
    {
	trav_class_init(remclass, 0, 0, 1);
	if(!is_static_method(m))
	    while(i = get_next_interface())
	    {
/* 		printf("        Searching in interface %s\n", i); */
		if(implements(i, "nexusrmi/Remote") && contains_imethod(i, meth, sig))
		    return 1;
	    }

	remclass = get_superclass(remclass);
    }
    return 0;
}

/* Call action on every remote method in the given class */
static void remote_method_traverse(char *class, void (*action)(char *, fm_ptr, int))
{
    char *remclass = class;
    char **methods_added;
    char **signatures;
    char *meth, *sig;
    int nr_methods = 0, cur_methods = 10, j, method_added;
    fm_ptr m;

    methods_added = (char **)alloc_mem(cur_methods * sizeof(char *));
    signatures = (char **)alloc_mem(cur_methods * sizeof(char *));
    
    handler_id = base_handler_id;
    while(class)
    {
/* 	printf("Class = %s\n", class); */
	trav_class_init(class, 0, 1, 0);
	while(m = get_next_method())
	{
	    meth = get_name(class, m);
	    sig = get_typedescriptor(class, m);
/* 	    printf("    Method = %s(%s)\n", meth, sig); */

	    if(is_remote_method(m, meth, sig, remclass))
	    {
		method_added = 0;
		j = 0;
		while(!method_added && (j < nr_methods))
		{
		    if(!strcmp(methods_added[j], meth) && !strcmp(signatures[j],sig))
			method_added = 1;
		    j++;
		}
		if(!method_added)
		{
		    if(nr_methods == cur_methods)
		    {
			cur_methods += 10;
			methods_added = (char **)realloc_mem(methods_added, cur_methods * sizeof(char *));
			signatures = (char **)realloc_mem(signatures, cur_methods * sizeof(char *));
		    }
		    methods_added[nr_methods] = (char *)strdup(meth);
		    signatures[nr_methods] = (char *)strdup(sig);
		    action(class, m, handler_id++);
		    nr_methods++;
/* 		    printf("            Added this method\n"); */
		}
	    }
	}
	class = get_superclass(class);
    }
/*     printf("--\n"); */
    for(j = 0; j < nr_methods; j++)
    {
	free(methods_added[j]);
	free(signatures[j]);
    }
    free(methods_added);
    free(signatures);
}

/* Dump the stub code */
static void generate_stub(void)
{
    javas_out(2, "Generating stub.");

    fp = fopen(stub_file, "w");

    if(pack_name)
	fprintf(fp, "package %s;\n\n", pack_name);
    
    fprintf(fp, "public final class %s extends nexusrmi.server.RemoteStub implements ", stub_name);
    intf_traverse(full_name, dump_remote_interface);
    
    fprintf(fp, "\n");
    fprintf(fp, "{\n");
    remote_method_traverse(full_name, dump_handler_ids);
    fprintf(fp, "    private final int GF_HANDLER_ID = 0;\n");
    fprintf(fp, "    private final int RETURNHANDLER_ID = 5;\n");
    fprintf(fp, "    private final int EXCEPTION_HANDLER_ID = 15;\n\n");

    fprintf(fp, "    public %s(nexus.Nexus nxs, nexus.Startpoint sp, String ser, String codebase)\n", stub_name);
    fprintf(fp, "    {\n");
    fprintf(fp, "        super(nxs, sp, ser, codebase);\n");
    fprintf(fp, "    }\n\n");

    remote_method_traverse(full_name, dump_stub_method);
    fprintf(fp, "    protected void finalize() throws Throwable\n");
    fprintf(fp, "    {\n");
    fprintf(fp, "        super.finalize();\n");
    fprintf(fp, "    }\n");
    fprintf(fp, "}\n\n");

    fprintf(fp, "class %sHandler implements nexus.HandlerInterface\n", stub_name);
    fprintf(fp, "{\n");
    remote_method_traverse(full_name, dump_handler_ids);
    fprintf(fp, "    private final int EXCEPTION_HANDLER_ID = 15;\n\n");

    fprintf(fp, "    private nexus.Nexus nxs;\n");
    fprintf(fp, "    private Class serClass;\n");
    fprintf(fp, "    private String codebase;\n");
    fprintf(fp, "    private %s stub;\n", stub_name);
    fprintf(fp, "    private int index;\n\n");
    
    fprintf(fp, "    public %sHandler(nexus.Nexus nxs, %s stub, int index, Class serClass, String codebase)\n", stub_name, stub_name);
    fprintf(fp, "    {\n");
    fprintf(fp, "        this.nxs = nxs;\n");
    fprintf(fp, "        this.stub = stub;\n");
    fprintf(fp, "        this.index = index;\n");
    fprintf(fp, "        this.serClass = serClass;\n");
    fprintf(fp, "        this.codebase = codebase;\n");
    fprintf(fp, "    }\n\n");

    fprintf(fp, "    public void invoke_handler(nexus.Endpoint ep, nexus.GetBuffer buffer, int handler_id)\n");
    fprintf(fp, "    {\n");
    fprintf(fp, "        nexusrmi.Serialize ser = null;\n");
    fprintf(fp, "        try\n");
    fprintf(fp, "        {\n");
    fprintf(fp, "            try\n");
    fprintf(fp, "            {\n");
    fprintf(fp, "                ser = (nexusrmi.Serialize)serClass.newInstance();\n");
    fprintf(fp, "            }\n");
    fprintf(fp, "            catch(Exception e)\n");
    fprintf(fp, "            {\n");
    fprintf(fp, "                e.printStackTrace();\n");
    fprintf(fp, "                System.exit(1);\n");
    fprintf(fp, "            }\n");
    fprintf(fp, "            ser.initRead(buffer, codebase);\n");
    fprintf(fp, "            ser.startReadHeader();\n");
    fprintf(fp, "            int id = ser.readHPCxxHeaderHandlerid();\n\n");
    fprintf(fp, "            ser.startReadData();\n");
    fprintf(fp, "            switch(id)\n");
    fprintf(fp, "            {\n");
    remote_method_traverse(full_name, dump_stub_case);
    fprintf(fp, "            case EXCEPTION_HANDLER_ID:\n");
    fprintf(fp, "                {\n");
    fprintf(fp, "                    Throwable ex;\n");
    fprintf(fp, "                    Class[] fp;\n");
    fprintf(fp, "                    Object[] ap;\n");
    fprintf(fp, "                    int eid = ser.readbyte();\n");
    fprintf(fp, "                    String pack = (String)ser.readFinalObject(java.lang.String.class, null, null);\n");
    fprintf(fp, "                    String cname = (String)ser.readFinalObject(java.lang.String.class, null, null);\n");
    fprintf(fp, "                    String fqen;\n");
    fprintf(fp, "                    if(pack.length() == 0)\n");
    fprintf(fp, "                        fqen = cname;\n");
    fprintf(fp, "                    else\n");
    fprintf(fp, "                        fqen = pack + \".\" + cname;\n\n");

    fprintf(fp, "                    ex = (Throwable)Class.forName(fqen).newInstance();\n");
    fprintf(fp, "                    switch(eid)\n");
    fprintf(fp, "                    {\n");
    fprintf(fp, "                    case 1: ex = (Exception)new nexusrmi.ServerError(cname, (Error)ex); break;\n");
    fprintf(fp, "                    case 2: ex = (Exception)new nexusrmi.ServerException(cname, (Exception)ex); break;\n");
    fprintf(fp, "                    case 3: ex = (Exception)new nexusrmi.ServerRuntimeException(cname, (Exception)ex); break;\n");
    fprintf(fp, "                    case 4: ex = (Exception)new nexusrmi.UnexpectedException(cname, (Exception)ex); break;\n");
    fprintf(fp, "                    default: break;\n");
    fprintf(fp, "                    }\n");
    fprintf(fp, "                    stub.setException(index, (Exception)ex);\n");
    fprintf(fp, "                }\n");
    fprintf(fp, "                break;\n");
    fprintf(fp, "            default:\n");
    fprintf(fp, "                {\n");
    fprintf(fp, "                    System.out.println(\"Invalid handler ID.\");\n");
    fprintf(fp, "                }\n");
    fprintf(fp, "                break;\n");
    fprintf(fp, "            }\n");
    fprintf(fp, "        }\n");
    fprintf(fp, "        catch(Exception e)\n");
    fprintf(fp, "        {\n");
    fprintf(fp, "            stub.setException(index, new nexusrmi.UnexpectedException(\"RemoteException\", e));\n");
    fprintf(fp, "        }\n");
    fprintf(fp, "    }\n");
    fprintf(fp, "}\n");    
    fclose(fp);
}

/* Dump the skeleton code */
static void generate_skel(void)
{
    javas_out(2, "Generating skeleton.");

    fp = fopen(skel_file, "w");
    
    if(pack_name)
	fprintf(fp, "package %s;\n\n", pack_name);
    
    fprintf(fp, "public final class %s implements nexus.HandlerInterface\n", skel_name);
    fprintf(fp, "{\n");
    fprintf(fp, "    nexus.Nexus nxs;\n");
    fprintf(fp, "    boolean doLogging = false;\n");
    fprintf(fp, "    Class serClass;\n\n");
    
    fprintf(fp, "    public %s(nexus.Nexus nxs, String ser)\n", skel_name);
    fprintf(fp, "    {\n");
    fprintf(fp, "        this.nxs = nxs;\n");
    fprintf(fp, "        String logCalls = System.getProperty(\"java.rmi.server.logCalls\");\n");
    fprintf(fp, "        if((logCalls != null) && (logCalls.equals(\"true\")))\n");
    fprintf(fp, "            doLogging = true;\n");
    fprintf(fp, "        try\n");
    fprintf(fp, "        {\n");
    fprintf(fp, "            this.serClass = Class.forName(ser);\n");
    fprintf(fp, "        }\n");
    fprintf(fp, "        catch(Exception e)\n");
    fprintf(fp, "        {\n");
    fprintf(fp, "            e.printStackTrace();\n");
    fprintf(fp, "            System.exit(1);\n");
    fprintf(fp, "        }\n");
    fprintf(fp, "    }\n\n");

    fprintf(fp, "    public void invoke_handler(nexus.Endpoint ep, nexus.GetBuffer inbuf, int handler_id)\n");
    fprintf(fp, "    {\n");
    fprintf(fp, "        %sUpcallThread handler = new %sUpcallThread(this, nxs, serClass, ep, inbuf, handler_id, doLogging);\n", skel_name, skel_name);
    fprintf(fp, "        handler.start();\n");
    fprintf(fp, "    }\n");
    fprintf(fp, "}\n\n");

    fprintf(fp, "class %sUpcallThread extends Thread\n", skel_name);
    fprintf(fp, "{\n");
    fprintf(fp, "    private final int GF_HANDLER_ID = 0;\n");
    fprintf(fp, "    private final int RETURNHANDLER_ID = 5;\n");
    fprintf(fp, "    private final int EXCEPTION_HANDLER_ID = 15;\n");
    
    fprintf(fp, "    private nexus.Nexus nxs;\n");
    fprintf(fp, "    private Class serClass;\n");
    fprintf(fp, "    private nexus.Endpoint ep;\n");
    fprintf(fp, "    private nexus.GetBuffer inbuf;\n");
    fprintf(fp, "    private int handler_id;\n");
    fprintf(fp, "    private %s skel;\n", skel_name);
    fprintf(fp, "    private boolean doLogging = false;\n");

    remote_method_traverse(full_name, dump_handler_ids);
    fprintf(fp, "\n");
    
    fprintf(fp, "    public %sUpcallThread(%s skel, nexus.Nexus nxs, Class ser, nexus.Endpoint ep, nexus.GetBuffer inbuf, int handler_id, boolean doLogging)\n",
	    skel_name, skel_name);
    fprintf(fp, "    {\n");
    fprintf(fp, "        this.skel = skel;\n");
    fprintf(fp, "        this.serClass = ser;\n");
    fprintf(fp, "        this.nxs = nxs;\n");
    fprintf(fp, "        this.ep = ep;\n");
    fprintf(fp, "        this.inbuf = inbuf;\n");
    fprintf(fp, "        this.handler_id = handler_id;\n");
    fprintf(fp, "        this.doLogging = doLogging;\n");
    fprintf(fp, "    }\n\n");

    fprintf(fp, "    public void run()\n");
    fprintf(fp, "    {\n");
    fprintf(fp, "        int status = 0;\n");
    fprintf(fp, "        nexusrmi.Serialize ser = null;\n");
    fprintf(fp, "        nexus.Startpoint replySP = null;\n");
    fprintf(fp, "        try\n");
    fprintf(fp, "        {\n");
    fprintf(fp, "            ser = (nexusrmi.Serialize)serClass.newInstance();\n");
    fprintf(fp, "        }\n");
    fprintf(fp, "        catch(Exception e)\n");
    fprintf(fp, "        {\n");
    fprintf(fp, "            e.printStackTrace();\n");
    fprintf(fp, "            System.exit(1);\n");
    fprintf(fp, "        }\n");
    fprintf(fp, "        ser.initRead(inbuf, null);\n\n");
    
    fprintf(fp, "        try\n");
    fprintf(fp, "        {\n");
    fprintf(fp, "            ser.startReadHeader();\n");
    fprintf(fp, "            replySP = ser.readHPCxxHeaderStartpoint();\n");
    fprintf(fp, "            int offset = ser.readHPCxxHeaderOffset();\n");
    fprintf(fp, "            int key = ser.readHPCxxHeaderHandlerid();\n");
    fprintf(fp, "            ser.startReadData();\n");
    fprintf(fp, "            switch(key)\n");
    fprintf(fp, "            {\n");
    remote_method_traverse(full_name, dump_skel_case);
    fprintf(fp, "            default:\n");
    fprintf(fp, "                {\n");
    fprintf(fp, "                    System.err.println(\"Invalid handler ID.\");\n");
    fprintf(fp, "                }\n");
    fprintf(fp, "                break;\n");
    fprintf(fp, "            }\n");
    fprintf(fp, "            replySP.destroy();\n");
    fprintf(fp, "        }\n");
    fprintf(fp, "        catch(Throwable e)\n");
    fprintf(fp, "        {\n");
    fprintf(fp, "            e.printStackTrace();\n");
    fprintf(fp, "            int id;\n");
    fprintf(fp, "            if(status == 1)\n");
    fprintf(fp, "                if(e instanceof Error)\n");
    fprintf(fp, "                    id = 1;\n");
    fprintf(fp, "                else if(e instanceof RuntimeException)\n");
    fprintf(fp, "                    id = 3;\n");
    fprintf(fp, "                else if(e instanceof nexusrmi.RemoteException)\n");
    fprintf(fp, "                    id = 2;\n");
    fprintf(fp, "                else\n");
    fprintf(fp, "                    id = 0;\n");
    fprintf(fp, "            else\n");
    fprintf(fp, "                id = 4;\n\n");
    
    fprintf(fp, "            String pack = null;\n");
    fprintf(fp, "            String cname = null;\n");
    fprintf(fp, "            String fqen = e.getClass().getName();\n");
    fprintf(fp, "            int lastdot = fqen.lastIndexOf('.');\n");
    fprintf(fp, "            if(lastdot < 0)\n");
    fprintf(fp, "            {\n");
    fprintf(fp, "                pack = new String(\"\");\n");
    fprintf(fp, "                cname = fqen;\n");
    fprintf(fp, "            }\n");
    fprintf(fp, "            else\n");
    fprintf(fp, "            {\n");
    fprintf(fp, "                pack = fqen.substring(0, lastdot);\n");
    fprintf(fp, "                cname = fqen.substring(lastdot + 1);\n");
    fprintf(fp, "            }\n\n");
    			     
    fprintf(fp, "            ser.initWrite();\n");
    fprintf(fp, "            ser.startWriteHeader();\n");
    fprintf(fp, "            ser.writeHPCxxHeaderHandlerid(EXCEPTION_HANDLER_ID);\n");
    fprintf(fp, "            ser.startWriteData();\n");
    fprintf(fp, "            ser.writebyte((byte)id);\n");
    fprintf(fp, "            ser.writeFinalObject(java.lang.String.class, pack);\n");
    fprintf(fp, "            ser.writeFinalObject(java.lang.String.class, cname);\n\n");
    			     
    fprintf(fp, "            nxs.send_rsr((nexus.PutBuffer)ser.getBuffer(), replySP, RETURNHANDLER_ID, true);\n");
    fprintf(fp, "        }\n");
    fprintf(fp, "    }\n");
    fprintf(fp, "}\n");
    fclose(fp);
}

/* The main program.
   Process arguments and generate stub and skeleton for each argument
*/
int main(int argc, char **argv)
{
    int i, j, iinc = 1;
    
    for(i = 1; i < argc; i += iinc)
    {
	iinc = 1;
	if(argv[i][0] == '-')
	{
	    for(j = 1; j < strlen(argv[i]); j++)
	    {
		switch(argv[i][j])
		{
		case 'h':
		case '?':
		    fprintf(stderr, "Usage: javas [options] remote_object ...\n");
		    fprintf(stderr, "\n");
		    fprintf(stderr, "Options:\n");
		    fprintf(stderr, "-o offset     : start handler id's at 200 + offset\n");
		    fprintf(stderr, "-v            : output verbose messages\n");
		    fprintf(stderr, "-h / -?       : print this help\n");
		    exit(EXIT_SUCCESS);
		case 'v':
		    verbose = 1;
		    break;
		case 'o':
		    if((argc <= i + 1) || (argv[i + 1] == NULL))
			fprintf(stderr, "No argument following -o, 0 assumed\n");
		    else
		    {
			handler_id_offset = atoi(argv[i + 1]);
			if(errno == EINVAL)
			    fprintf(stderr, "Invalid argument following -o, 0 assumed\n");
			else
			{
			    argv[i + 1] = NULL;
			    iinc = 2;
			}
		    }
		    break;
		default:
		    fprintf(stderr, "Unknown option: -%c, ignored\n", argv[i][j]);
		}
	    }
	    argv[i] = NULL;
	}
    }
    
    base_handler_id = BASE_HANDLER_ID + handler_id_offset;
    for(i = 1; i < argc; i++)
	if(argv[i])
	{
	    full_name = argv[i];
    
	    construct_names();

/* 	    printf("full_name   = %s\n", full_name); */
/* 	    printf("pack_name   = %s\n", pack_name); */
/* 	    printf("server_name = %s\n", server_name); */
/* 	    printf("stub_name   = %s\n", stub_name); */
/* 	    printf("stub_file   = %s\n", stub_file); */
/* 	    printf("skel_name   = %s\n", skel_name); */
/* 	    printf("skel_file   = %s\n", skel_file); */
	    
	    if(!read_classes(full_name))
	    {
		generate_stub();
		generate_skel();
	    }
	    
	    delete_names();
	    base_handler_id = handler_id;
	}
    
    return 0;
}

