/* 
 * NexusRMI
 *
 * Copyright (c) 1999
 * Fabian Breg
 * Indiana University/Leiden University
 * All Rights Reserved
 *
 * Please refer to the LICENSE file provided in this
 * distribution for full licensing terms and
 * conditions.
 *
 * Please report all bugs/problems to
 * breg@cs.indiana.edu
 *
 ******************************************************
 *
 * nexusize.c
 *
 */

/* ********************************************************
   *** INCLUDE FILES and DEFINITIONS                    ***
   ********************************************************  */

#include <stdarg.h>
#include <assert.h>
#include "class.h"
#include "global.h"
#include "y.tab.h"

#define NEXUS_WRITE_BLANK_FINALS          0
#define NEXUS_WRITE_REST                  1
#define NEXUS_READ                        2
#define CONSTRUCTOR                       3

/* Some symbol names that we are going to need.
   These will be initialized in nexusize_init()
*/
static symb_ptr *put_meths;
static symb_ptr *get_meths;

static symb_ptr nexusCreateInstance;
static symb_ptr container;
static symb_ptr classs;
static symb_ptr Object;
static symb_ptr Serialize;
static symb_ptr SerObject;
static symb_ptr serializeWrite;
static symb_ptr serializeRead;
static symb_ptr size;
static symb_ptr SerialObject;
static symb_ptr writeObject;
static symb_ptr readObject;
static symb_ptr writeFinalObject;
static symb_ptr readFinalObject;
static symb_ptr tmp;
static symb_ptr null;

/* Global variables */
static stmt_ptr current_class;
static stmt_ptr cur_stmt;
static stmt_ptr nexusWriteMeth;
static stmt_ptr nexusReadMeth;
static expr_ptr package;
static expr_ptr cur_expr;

static int nexusable;
static int add_cons;
static int action;
static int loop_nest;

static int field_blank;
static char *field_name;

static int has_cons;

char *this_class;
char *super_class;


/* Check whether class is serializable, ie implements the Serializable interface.
   wrapper function for class loading function
*/
static int is_serializable(char *class)
{
    int res = implements(class, "java/io/Serializable");
    if(res)
	if(!strncmp("java/", class, 5))
	    res = 0;
    return res;
}

/* Check whether the given class is remote, ie implements the Remote interface.
   wrapper function for class loading function
*/
static int is_remote(char *class)
{
    return implements(class, "nexusrmi/Remote");
}

static int is_nonstatic_class(char *cn, stmt_ptr class_body)
{
    if(!class_body)
	return 0;
    if(class_body->kind == CONCAT_S)
	return is_nonstatic_class(cn, class_body->u.concat.s1) ||
	       is_nonstatic_class(cn, class_body->u.concat.s2);
    if(class_body->kind == CLASS_S)
	return (!strcmp(cn, class_body->u.header.name->u.name.id->symbol)) &&
	       (!(class_body->u.header.mod & STA_B));
    return 0;
}

static int is_member_class_in(char *cn, stmt_ptr in_class)
{
    char *cn1, *cn2;
    int res;

    cn1 = strrchr(cn, '$') + 1;
    cn2 = in_class->u.header.name->u.name.id->symbol;
    res = is_nonstatic_class(cn1, in_class->u.header.body);
    return res;
}

static int is_member_class(stmt_ptr s)
{
    return (s->u.header.container &&
	    (s->u.header.container->kind == CLASS_S) &&
	    !(s->u.header.mod & STA_B));
}

static void field_is_blank(stmt_ptr s)
{
    if((s->kind == DECL_S) &&
       (s->u.decl.name) &&
       (s->u.decl.name->u.name.info == FIELD_N) &&
       (!strcmp(s->u.decl.name->u.name.id->symbol, field_name)) &&
       (s->u.decl.mod & FIN_B) &&
       (s->u.decl.init == NULL))
	field_blank = 1;
}

static int is_blank(fm_ptr f)
{
    field_name = get_name(this_class, f);
    field_blank = 0;
    stmt_traverse(current_class, field_is_blank, NULL);
    return field_blank;
}

/* Convert bytecode representation for type to a number. */
static int get_type_id(char *type)
{
    switch(type[0])
    {
    case 'V' : return 0;
    case 'Z' : return 1;
    case 'B' : return 2;
    case 'S' : return 3;
    case 'I' : return 4;
    case 'J' : return 5;
    case 'C' : return 6;
    case 'F' : return 7;
    case 'D' : return 8;
    case 'L' : return 9;
    case '[' : return 10;
    }
    javar_out(-1, "Unexpected fall-through in get_type_id().");
    return -1;
}

/* Bytecode uses slashes (/) instead of dots (.) as a separater,
   so we replace these.
*/
static void replace_slash(char *s)
{
    int i;
    for(i = 0; i < strlen(s); i++)
	if(s[i] == '/')
	    s[i] = '.';
}

/* Get a string representation of the fully qualified name. qual is
   typically given a pointer to the package expression, while the
   name is obtained from the class header.
*/
/* static char *get_full_name(expr_ptr qual, char *name) */
/* { */
/*     char *s; */
/*     char *q; */
/*     int len; */
    
/*     if(qual) */
/*     { */
/* 	q = get_full_name(qual->u.name.lcomp, qual->u.name.id->symbol); */
/* 	len = strlen(q) + strlen(name); */
/* 	s = (char *)make_mem((len + 2) * sizeof(char)); */
/* 	sprintf(s, "%s.%s%c", q, name, '\0'); */
/* 	free(q); */
/*     } */
/*     else */
/*     { */
/* 	len = strlen(name); */
/* 	s = (char *)make_mem((len + 1) * sizeof(char)); */
/* 	strcpy(s, name); */
/* 	s[len] = '\0'; */
/*     } */
/*     return s; */
/* } */
static char *get_pack_name(expr_ptr q)
{
    char *q1 = NULL, *q2 = NULL, *res;
    int len = 1;
    
    if(!q)
	return NULL;
    
    q1 = q->u.name.id->symbol;
    len += strlen(q1);
    q2 = get_pack_name(q->u.name.lcomp);
    if(q2)
	len += strlen(q2) + 1;
    
    res = (char *)make_mem(len * sizeof(char));
    strcpy(res, "");
    if(q2)
    {
	strcat(res, q2);
	strcat(res, ".");
/* 	free(q2); */
    }
    strcat(res, q1);
    return res;
}

/* static char *get_class_name(stmt_ptr c) */
/* { */
/*     char *c1 = NULL, *c2 = NULL, *res; */
/*     int len = 1; */
    
/*     if(!c) */
/* 	return NULL; */

/*     if(c->u.header.name) */
/*     { */
/* 	c1 = c->u.header.name->u.name.id->symbol; */
/* 	len += strlen(c1); */
/*     } */
/*     c2 = get_class_name(c->u.header.container); */
/*     if(c2) */
/* 	len += strlen(c2) + 1; */

/*     res = (char *)make_mem(len * sizeof(char)); */
/*     strcpy(res, ""); */
/*     if(c2) */
/*     { */
/* 	strcat(res, c2); */
/* 	strcat(res, "$"); */
/*     } */
/*     if(c1) */
/* 	strcat(res, c1); */
/*     return res; */
/* } */

static char *get_class_name(expr_ptr c)
{
    char *c1 = NULL, *c2 = NULL, *res;
    int len = 1;
    
    if(!c)
	return NULL;

    if(c->kind == ID_)
    {
	return (char *)strdup(c->u.name.id->symbol);
    }

    if(c->kind == '$')
    {
	c1 = get_class_name(c->u.op.lhs);
	c2 = get_class_name(c->u.op.rhs);
	if(c1)
	    len += strlen(c1);
	if(c1 && c2)
	    len++;
	if(c2)
	    len += strlen(c2);

	res = (char *)make_mem(len * sizeof(char));
	strcpy(res, "");
	
	if(c1)
	{
	    strcat(res, c1);
/* 	    free(c1); */
	}
	if(c1 && c2)
	    strcat(res, "$");
	if(c2)
	{
	    strcat(res, c2);
/* 	    free(c2); */
	}
	return res;
    }
    javar_out(-1, "Unexpected function exit in get_class_name().");
    return NULL;
}

static char *get_full_name(expr_ptr q, expr_ptr c)
{
    char *qual = NULL;
    char *class = NULL;
    char *s;
    int len = 1;

    qual = get_pack_name(q);
    if(qual)
	len += strlen(qual) + 1;
    class = get_class_name(c);
    if(class)
	len += strlen(class);
    
    s = (char *)make_mem(len * sizeof(char));
    strcpy(s, "");
    if(qual)
    {
	strcat(s, qual);
	strcat(s, ".");
/* 	free(qual); */
    }
    if(class)
    {
	strcat(s, class);
/* 	free(class); */
    }
    return s;
}

/* Return a pointer to the ast representation of the fully qualified name 'name'. */
static expr_ptr construct_name(char *name)
{
    expr_ptr res;
    char *q, *n, *t;
    int l;

    t = strrchr(name, '.');
    if(t)
    {
	t++;
	l = strlen(t);
	n = (char *)make_mem((l + 1) * sizeof(char));
	strncpy(n, t, l);
	n[l] = '\0';
	l = strlen(name) - strlen(t);
	q = (char *)make_mem((l) * sizeof(char));
	strncpy(q, name, l);
	q[l - 1] = '\0';
	res = make_name(construct_name(q), insert_table(n), 0);
	free(n);
	free(q);
    }
    else
    {
	res = make_name(NULL, insert_table(name), 0);
    }
    return res;
}

/* Return a pointer to the ast representation for the type given in
   bytecode notation.
*/
static type_ptr make_type(char *tname)
{
    type_ptr tp = NULL;
    char *t = (char *)strdup(tname);

    if(tname[0] == '[')
	tp = make_array_type(make_type(t + 1));
    else if(tname[0] == 'L')
    {
	char *cname = (char *)strdup(t + 1);
	cname[strlen(cname) - 1] = '\0';
	replace_slash(cname);
	tp = make_c_i_type(construct_name(cname));
	free(cname);
    }
    else
    {
	switch(t[0])
	{
	case 'Z':
	    tp = make_primitive_type(BOOL_T);
	    break;
	case 'B':
	    tp = make_primitive_type(BYTE_T);
	    break;
	case 'S':
	    tp = make_primitive_type(SHORT_T);
	    break;
	case 'I':
	    tp = make_primitive_type(INT_T);
	    break;
	case 'J':
	    tp = make_primitive_type(LONG_T);
	    break;
	case 'C':
	    tp = make_primitive_type(CHAR_T);
	    break;
	case 'F':
	    tp = make_primitive_type(FLOAT_T);
	    break;
	case 'D':
	    tp = make_primitive_type(DOUBLE_T);
	    break;
	}
    }
    free(t);
    return tp;
}

/* Generate a pointer to the ast representation for an array allocation expression.
   Returns an expression that looks like this: new [nr][][], where the number of
   dimensions is equal to the number of dimensions of tp.
*/
static expr_ptr array_alloc_expr(char *tp, symb_ptr nr)
{
    char *bt = tp;
    expr_ptr dim = make_op(']', make_op('[', NULL, make_name(NULL, nr, 0)), NULL);
    expr_ptr t = dim;
    type_ptr base_type;
    
    while(*bt == '[')
    {
	bt++;
	t->u.op.rhs = make_op('[', NULL, NULL);
	t = t->u.op.rhs;
    }

    if(get_type_id(bt) < 9)
	base_type = make_primitive_type(get_type_id(bt));
    else
    {
	char *elt_type = (char *)make_mem((strlen(bt) - 1) * sizeof(char));
	strncpy(elt_type, bt + 1, strlen(bt) - 2);
	elt_type[strlen(bt) - 2] = '\0';
	replace_slash(elt_type);
        base_type = make_c_i_type(make_name(NULL, insert_table(elt_type), 0));
    }
    return make_te(NEW_, base_type, dim, NULL);
}

/* Produce type specific serialization code for the field with the given type and name. */
static void *handle_type_expr(char *type, expr_ptr name)
{
    int type_id = get_type_id(type);
    void *res = NULL;

    switch(type_id)
    {
    case 0:
	assert(0); /* Should never happen */
	break;
    case 1:
    case 2:
    case 3:
    case 4:
    case 5:
    case 6:
    case 7:
    case 8:
	if((action == NEXUS_WRITE_BLANK_FINALS) || (action == NEXUS_WRITE_REST))
	{
	    /* SerObject.write<primitive>(<name>); */
	    res = make_simple(EXP_S,
				  make_op('i',
					  make_op('.',
						  make_name(NULL, SerObject, 0),
						  make_name(NULL, put_meths[type_id - 1], 0)),
					  name));
	}
	else if((action == NEXUS_READ) || (action == CONSTRUCTOR))
	{
	    /* <name> = SerObject.read<primitive>(); */
	    res = make_simple(EXP_S,
				  make_op('=',
					  name,
					  make_op('i',
						  make_op('.',
							  make_name(NULL, SerObject, 0),
							  make_name(NULL, get_meths[type_id - 1], 0)),
						  NULL)));
	}
	else
	    javar_out(-1, "Illegal action in handle_type_expr(): %d", action);
	
	break;
    case 9:
    case 10:
	{
	    short final_class = 0;
	    char *cn = NULL;
	    expr_ptr ro_param = NULL;
	    
	    if(type_id == 9)
	    {
		cn = (char *)strdup(type + 1);
		cn[strlen(cn) - 1] = '\0';
		final_class = is_final_class(cn);
		ro_param = make_thissup(THIS_);
		    
		if(cn && is_member_class_in(cn, current_class))
		    ro_param = make_op(',', ro_param, make_thissup(THIS_));
		else
		    ro_param = make_op(',', ro_param, make_name(NULL, null, 0));
	    }
	    else
		ro_param = make_op(',', make_thissup(THIS_), make_name(NULL, null, 0));
	    
	    if(!final_class)
	    {
		if((action == NEXUS_WRITE_BLANK_FINALS) || (action == NEXUS_WRITE_REST))
		{
		    /* SerObject.writeObject(<name>); */
		    res = make_simple(EXP_S,
					  make_op('i',
						  make_op('.',
							  make_name(NULL, SerObject, 0),
							  make_name(NULL, writeObject, 0)),
						  name));
		}			      
		else if((action == NEXUS_READ) || (action == CONSTRUCTOR))
		{
		    /* <name> = (<type>)SerObject.ReadObject(); */
		    res = make_simple(EXP_S,
					  make_op('=',
						  name,
						  make_cast(make_type(type),
							    NULL,
							    make_op('i',
								    make_op('.',
									    make_name(NULL, SerObject, 0),
									    make_name(NULL, readObject, 0)),
								    ro_param))));
		}
		else
		    javar_out(-1, "Illegal action in handle_type_expr(): %d", action);
	    }
	    else
	    {
		replace_slash(cn);
/* 		char *cns = (char *)make_mem((strlen(cn) + 3) * sizeof(char)); */
/* 		sprintf(cns, "\"%s\"", cn); */
		if((action == NEXUS_WRITE_BLANK_FINALS) || (action == NEXUS_WRITE_REST))
		    /* SerObject.writeFinalObject(<cn>, <name>); */
		    res = make_simple(EXP_S,
					  make_op('i',
						  make_op('.',
							  make_name(NULL, SerObject, 0),
							  make_name(NULL, writeFinalObject, 0)),
						  make_op(',',
							  make_op('.',
								  construct_name(cn),
								  make_name(NULL, classs, 0)),
/* 							  make_literal(-1, cns, 0), */
							  name)));
						      
		else if((action == NEXUS_READ) || (action == CONSTRUCTOR))
		{
		    /* <name> = (<type>)SerObject.readFinalObject(<cn>); */
		    res = make_simple(EXP_S,
					  make_op('=',
						  name,
						  make_cast(make_type(type),
							    NULL,
							    make_op('i',
								    make_op('.',
									    make_name(NULL, SerObject, 0),
									    make_name(NULL, readFinalObject, 0)),
								    make_op(',',
									    make_op('.',
										    construct_name(cn),
										    make_name(NULL, classs, 0)),
									    ro_param)))));
		}
		else
		    javar_out(-1, "Illegal action in handle_type_expr(): %d", action);
	    }
	    free(cn);
	}
	break;
    }
    return res;
}

/* This is a function that can be given to field_traverse(). It will call handle_type(). */
static void handle_field(fm_ptr f)
{
    stmt_ptr s;
    expr_ptr e;
    loop_nest = 0;
    
    if((action == NEXUS_WRITE_BLANK_FINALS) &&
       !(f->access_flags & (ACC_STATIC | ACC_TRANSIENT)) &&
       (f->access_flags & ACC_FINAL) &&
       is_blank(f))
    {
	s = handle_type_expr(get_typedescriptor(this_class, f),
			     make_name(NULL, insert_table(get_name(this_class, f)), 0));
	cur_stmt = concat(cur_stmt, s);
    }
    else if((action == NEXUS_WRITE_REST) &&
       !(f->access_flags & (ACC_STATIC | ACC_TRANSIENT)) &&
       (!(f->access_flags & ACC_FINAL)))
    {
	s = handle_type_expr(get_typedescriptor(this_class, f),
			     make_name(NULL, insert_table(get_name(this_class, f)), 0));
	cur_stmt = concat(cur_stmt, s);
    }
    else if((action == NEXUS_READ) &&
	    !(f->access_flags & (ACC_FINAL | ACC_STATIC | ACC_TRANSIENT)))
    {
	s = handle_type_expr(get_typedescriptor(this_class, f),
			     make_name(NULL, insert_table(get_name(this_class, f)), 0));
	cur_stmt = concat(cur_stmt, s);
    }
    else if((action == CONSTRUCTOR) &&
	    !(f->access_flags & (ACC_STATIC | ACC_TRANSIENT)) &&
	    (f->access_flags & (ACC_FINAL)) &&
	    is_blank(f))
    {
	s = handle_type_expr(get_typedescriptor(this_class, f),
			     make_name(NULL, insert_table(get_name(this_class, f)), 0));
	cur_stmt = concat(cur_stmt, s);
    }
}

/* Call 'action' with every field in class 'class' */
static void field_traverse(char *class, void (*action)(fm_ptr))
{
    fm_ptr f;
    
    trav_class_init(class, 1, 1, 1);
    while(f = get_next_field())
	action(f);
}
			   
/* Add the nexusWrite() method to class s */
static void add_nexus_write()
{
    nexusable = 0;

    if(!is_remote(this_class) && (is_serializable(this_class)))
    {
	cur_stmt = NULL;

	if(super_class && is_serializable(super_class))
	{
	    /* super.serializeWrite(SerObject); */
	    cur_stmt = concat(cur_stmt,
			      make_simple(EXP_S,
					  make_op('i',
						  make_op('.',
							  make_thissup(SUPER_),
							  make_name(NULL, serializeWrite, 0)),
						  make_name(NULL, SerObject, 0))));
	}
	else if(strcmp(super_class, "java/lang/Object"))
	{
	    javar_out(-1, "superobject of %s is not nexusable", this_class);
	}

	/* Handle every field */
	action = NEXUS_WRITE_BLANK_FINALS;
	field_traverse(this_class, handle_field);
	action = NEXUS_WRITE_REST;
	field_traverse(this_class, handle_field);

	/* public void serializeWrite(nexusrmi.Serialize SerObject)
	   {
	       ...
	   }
	*/
	current_class->u.header.body = concat(current_class->u.header.body,
				  make_header(METHOD_S, PUB_B, serializeWrite, make_primitive_type(VOID_T), 0,
					      make_declexpr(give_m_t_c(0,
								       make_c_i_type(make_name(NULL, Serialize, TYPE_N)),
								       make_decl(SerObject, 0, NULL),
								       LOCPAR_N)),
					      NULL,
					      cur_stmt, 0));
    }
}

/* Add nexusRead() method to class s. */
static void add_nexus_read()
{
    if((!is_remote(this_class)) && (is_serializable(this_class)))
    {
	cur_stmt = NULL;
	
	if(super_class && is_serializable(super_class))
	{
	    /* super.serializeRead(SerObject); */
	    cur_stmt = concat(cur_stmt,
			      make_simple(EXP_S,
					  make_op('i',
						  make_op('.',
							  make_thissup(SUPER_),
							  make_name(NULL, serializeRead, 0)),
						  make_name(NULL, SerObject, 0))));
	}

	/* handle every field */
	action = NEXUS_READ;
	field_traverse(this_class, handle_field);
	cur_stmt = concat(cur_stmt, make_simple(RET_S, make_thissup(THIS_)));

	/* public Object serializeRead(nexusrmi.Serialize SerObject)
	   {
	       ...
	   }
	*/
	current_class->u.header.body = concat(current_class->u.header.body,
					      make_header(METHOD_S, PUB_B,
							  serializeRead, make_c_i_type(make_name(NULL, Object, 0)), 0,
					      make_declexpr(give_m_t_c(0,
								       make_c_i_type(make_name(NULL, Serialize, TYPE_N)),
								       make_decl(SerObject, 0, NULL),
								       LOCPAR_N)),
					      NULL,
					      cur_stmt, 0));
    }
}


/* Add an empty constructor to class s */
static void add_constructor()
{
    cur_stmt = NULL;
    if(super_class && is_serializable(super_class))
    {
	/* super(SerObject); */
	cur_stmt = make_simple(EXP_S,
			       make_op('i',
				       make_thissup(SUPER_),
				       make_name(NULL, SerObject, 0)));
    }
    
    if(!is_remote(this_class) && (is_serializable(this_class)))
    {
	action = CONSTRUCTOR;
	field_traverse(this_class, handle_field);
	
	current_class->u.header.body = concat(current_class->u.header.body,
					      make_header(CONS_S,
							  PUB_B,
							  current_class->u.header.name->u.name.id,
							  NULL, 0,
							  make_declexpr(give_m_t_c(0,
										   make_c_i_type(make_name(NULL, Serialize, 0)),
										   make_decl(SerObject, 0, NULL),
										   LOCPAR_N)),
							  NULL, cur_stmt, 0));
    }
}

static void check_cons(stmt_ptr s)
{
    if(s->kind == CONS_S)
	has_cons = 1;
}

static void add_empty_constructor(stmt_ptr s)
{
    has_cons = 0;
    stmt_traverse(s, check_cons, NULL);
    if(!has_cons)
	current_class->u.header.body = concat(current_class->u.header.body,
					      make_header(CONS_S, PUB_B,
							  current_class->u.header.name->u.name.id,
							  NULL, 0, NULL, NULL, NULL, 0));
}

/* driver routine */
static void add_nexus_meths(stmt_ptr s)
{
    if(s->kind == PCK_S)
	package = s->u.arg;
    
    if(s->kind == CLASS_S)
    {
/* 	printf("Getting full name for: "); */
/* 	dump_expr(package, stdout); */
/* 	printf(" & "); */
/* 	dump_expr(s->u.header.fqcn, stdout); */
/* 	printf("\n"); */
	
	this_class = get_full_name(package, s->u.header.fqcn);
	printf("%s...", this_class);

	/* Read the class file corresponding to class s. */
	if(!read_classes(this_class))
	{
	    super_class = get_superclass(this_class);

	    if((!is_remote(this_class)) && (is_serializable(this_class)))
	    {
		current_class = s;
		/* Add the SerialObject interface */
		s->u.header.e1 = make_te('t',
					 make_c_i_type(make_name(NULL, SerialObject, 0)),
					 s->u.header.e1,
					 NULL);

		/* Add parameterless / bodyless constructor */
		add_empty_constructor(s);
		
		/* Add serialization constructor */
		add_constructor();
		
		/* Add the serialization methods. */
		add_nexus_write();
		add_nexus_read();
	    }
	}
	printf(" done\n");
	free(this_class);
    }
}



/* Initialize
   ********** */
   
static void nexusize_init(void)
{
    /* Initialize the symbols needed. */
    nexusCreateInstance = insert_table("nexusCreateInstance");
    container = insert_table("container");
    classs = insert_table("class");
    Object = insert_table("Object");
    
    put_meths = (symb_ptr *)make_mem(8 * sizeof(symb_ptr));
    put_meths[0] = insert_table("writeboolean");
    put_meths[1] = insert_table("writebyte");
    put_meths[2] = insert_table("writeshort");
    put_meths[3] = insert_table("writeint");
    put_meths[4] = insert_table("writelong");
    put_meths[5] = insert_table("writechar");
    put_meths[6] = insert_table("writefloat");
    put_meths[7] = insert_table("writedouble");

    get_meths = (symb_ptr *)make_mem(8 * sizeof(symb_ptr));
    get_meths[0] = insert_table("readboolean");
    get_meths[1] = insert_table("readbyte");
    get_meths[2] = insert_table("readshort");
    get_meths[3] = insert_table("readint");
    get_meths[4] = insert_table("readlong");
    get_meths[5] = insert_table("readchar");
    get_meths[6] = insert_table("readfloat");
    get_meths[7] = insert_table("readdouble");

    Serialize = insert_table("nexusrmi.Serialize");
    SerObject = insert_table("SerObject");
    serializeWrite = insert_table("serializeWrite");
    serializeRead = insert_table("serializeRead");
    size = insert_table("size");
    SerialObject = insert_table("nexusrmi.SerialObject");
    writeObject = insert_table("writeObject");
    readObject = insert_table("readObject");
    writeFinalObject = insert_table("writeFinalObject");
    readFinalObject = insert_table("readFinalObject");

    tmp = insert_table("tmp");
    null = insert_table("null");
}

/* The main routine. */
void nexusize(void)
{
    cur_stmt = NULL;
    nexusWriteMeth = NULL;
    nexusReadMeth = NULL;
    package = NULL;
    
    nexusize_init();
    stmt_traverse(comp_unit, add_nexus_meths, NULL);
}
