

/* *************
   *** JAVAS ***
   *********************************************************
   *** Copyright (c) 1997                                ***
   *** Fabian Breg and Aart J.C. Bik  Indiana University *** 
   *** All Rights Reserved                               ***
   *********************************************************
   *** Please refer to the LICENSE file for
   *** licensing terms and conditions.
   *** */

/* ********************************************************
   *** INCLUDE FILES and DEFINITIONS                    ***
   ********************************************************  */

#include "class.h"

/* ********************************************************
   *** EXTERNAL VARIABLES                               ***
   ******************************************************** */

/* PRIVATE
   ******* */

static short          error     = 0;
static char          *class_env = NULL;
static class_ptr      classes   = NULL;
static u1_int        *classname = NULL;
static FILE          *file      = NULL;

static u4_int         magic;
static u2_int         minor_version, major_version;

static u2_int         constant_pool_count = 0u;
static constant_ptr  *constant_pool       = NULL;

static u2_int         access_flags, this_class, super_class;

static u2_int         interfaces_count = 0u;
static u2_int        *interfaces       = NULL;
static u2_int         fields_count     = 0u;
static fm_ptr        *fields           = NULL;
static u2_int         methods_count    = 0u;
static fm_ptr        *methods          = NULL;
static u2_int         attributes_count = 0u;
static attribute_ptr *attributes       = NULL;

/* Traversal Variables
   ******************* */

static class_ptr      ftrav_class = NULL;
static class_ptr      mtrav_class = NULL;
static class_ptr      itrav_class = NULL;
static u2_int         trav_fc    = 0u;
static u2_int         trav_mc    = 0u;
static u2_int         trav_ic    = 0u;

static u2_int         trav_ex    = 0u;
static attribute_ptr  trav_exc   = NULL;

/* ********************************************************
   *** PRIVATE FUNCTIONS                                ***
   ******************************************************** */

/* **********************************
   *** Error and Warning Messages ***
   ***************************************************************
   *** level == -1 : FATAL ERROR,     EXIT PROGRAM             ***
   *** level ==  0 : ERRROR,          SET ERROR FLAG           ***
   *** level ==  1 : STRONG MESSAGE                            ***
   *** level ==  2 : MESSAGE,         PRINT ONLY FOR '-v'      ***
   *************************************************************** */

void javas_out(s1_int level, char *fmt, ...)  {
  va_list argv;

  if (level == 0)
     error = 1; 
  else if (level > 1 && ! verbose)
    return;
    
  va_start(argv, fmt);

  if (classname)
    fprintf(stderr, "%s:: ", (char *)classname);

  vfprintf(stderr, fmt, argv);

  putc('\n', stderr);

  va_end(argv);

  if (level < 0)
    exit(1);
}

/* Memory Allocation Functions
   [for size <= 0, size == 1 is used
    because some systems return NULL for malloc(0);]
   ************************************************* */

static void *make_mem(int size) { 
  void *p = calloc(((size > 0) ? size : 1), sizeof(u1_int)); 
  if (! p) 
    javas_out(-1, "Out of Memory");
  return p; 
}

/* ***********************************
   *** Move Record To General Info ***
   *********************************** */

static void move_out(void) {

  if (! classes)
    javas_out(-1, "incorrect invocation of move_out");

  magic               = classes -> magic;
  minor_version       = classes -> minor_version;
  major_version       = classes -> major_version;
  constant_pool_count = classes -> constant_pool_count;
  constant_pool       = classes -> constant_pool;
  access_flags        = classes -> access_flags;
  this_class          = classes -> this_class;
  super_class         = classes -> super_class;
  interfaces_count    = classes -> interfaces_count;
  interfaces          = classes -> interfaces;
  fields_count        = classes -> fields_count;
  fields              = classes -> fields;
  methods_count       = classes -> methods_count;
  methods             = classes -> methods;
  attributes_count    = classes -> attributes_count;
  attributes          = classes -> attributes;
}

static void move_in(void) {

  /* Construct New Class Node */

  class_ptr new   = (class_ptr) make_mem(sizeof(struct class_node));

  new -> next      = classes;
  new -> classname = (classname) ? (u1_int *) strdup((char *) classname) : NULL;

  new -> magic               = magic;
  new -> minor_version       = minor_version;
  new -> major_version       = major_version;
  new -> constant_pool_count = constant_pool_count;
  new -> constant_pool       = constant_pool;
  new -> access_flags        = access_flags;
  new -> this_class          = this_class;
  new -> super_class         = super_class;
  new -> interfaces_count    = interfaces_count;
  new -> interfaces          = interfaces;
  new -> fields_count        = fields_count;
  new -> fields              = fields;
  new -> methods_count       = methods_count;
  new -> methods             = methods;
  new -> attributes_count    = attributes_count;
  new -> attributes          = attributes;

  /* Hook In List */

  classes = new;

  /* Test Consistency */

  { u2_int  ind = constant_pool[ this_class ] -> u.indices.index1;
    char    *nm = (char *) constant_pool[ ind        ] -> u.utf8.s;

    if ((! nm) || (! classname) || (strcmp(nm, (char *) classname)))
      javas_out(1, "file name does not match fully qualified name");
  }
}

/* Check constant pool entry
   ************************* */

static u1_int valid_cp_entry(u1_int tag, u2_int entry, char *mess) {
  if ((entry != 0u) && (entry < constant_pool_count)) {

   constant_ptr p = constant_pool[entry];

   if ((! p) || (p -> tag != tag)) {
     javas_out(0, "invalid reference of %s to constant pool (%u)", mess, entry);
     return 0u;
   }
  }
  else {
    javas_out(0, "invalid index of %s into constant pool (%u)", mess, entry);
    return 0u;
  }
  return 1u;
}

/* read u1_int, u2_int, and u4_int routines 
   **************************************** */

static u1_int read_u1(void) {
  int u = fgetc(file);
  if (u == EOF)
    javas_out(0, "unexpected EOF");
  return (u1_int) u;
}

static u2_int read_u2(void) {
  u1_int u1 = read_u1();
  u1_int u2 = read_u1();
  return B2U2(u1,u2);
}

static u4_int read_u4(void) {
  u1_int u1 = read_u1();
  u1_int u2 = read_u1();
  u1_int u3 = read_u1();
  u1_int u4 = read_u1();
  return B2U4(u1,u2,u3,u4);
}

/* Read Constant Pool
   (entry constant_pool[0] is reserved, but included in count)
   *********************************************************** */

static void read_constant_pool(void) {
  u4_int i, j; /* wide counters */

  constant_pool_count = read_u2();
  constant_pool       = NULL;

  if (constant_pool_count == 0u)
    return;

  /* Construct the constant pool */

  constant_pool = (constant_ptr *) 
     make_mem( constant_pool_count * sizeof(constant_ptr) );

  constant_pool[0] = NULL;

  for (i = 1u; i < constant_pool_count; i++) {

     constant_pool[i] = (constant_ptr) make_mem( sizeof(struct constant_node) );
     constant_pool[i] -> tag = read_u1();

     switch(constant_pool[i] -> tag) {

       case CONSTANT_Class:
       case CONSTANT_String:

	 constant_pool[i] -> u.indices.index1 = read_u2();
         constant_pool[i] -> u.indices.index2 = 0u;
	 break;

       case CONSTANT_Fieldref:
       case CONSTANT_Methodref:
       case CONSTANT_InterfaceMethodref:
       case CONSTANT_NameAndType:

         constant_pool[i] -> u.indices.index1 = read_u2();
         constant_pool[i] -> u.indices.index2 = read_u2();
         break;

       case CONSTANT_Integer:
       case CONSTANT_Float:

	   constant_pool[i] -> u.data.val1 = read_u4();
	   constant_pool[i] -> u.data.val2 = 0u;
	   break;

       case CONSTANT_Long:
       case CONSTANT_Double:

	   constant_pool[i] -> u.data.val1 = read_u4();
	   constant_pool[i] -> u.data.val2 = read_u4();

	   /* These entries make next entry invalid!
	      ************************************** */

	   constant_pool[ ++i ] = NULL;
	   break;

       case CONSTANT_Utf8:

         /* Read-in constant string value (represented as BYTE sequence) */

        { u2_int  len = read_u2();
	  u1_int *s   = (u1_int *) make_mem((1+len) * sizeof(u1_int));

	  for (j = 0u; j < len; j++) 
	     s[j] = read_u1();
  	  s[len] = '\0'; 

          constant_pool[i] -> u.utf8.l = len;
	  constant_pool[i] -> u.utf8.s = s;
        }
        break;

       default:
	 javas_out(-1, "invalid constant pool tag (%u)", constant_pool[i] -> tag);
     }
  }
}

/* Read Interfaces
   *************** */

static void read_interfaces(void) {
  u4_int i; /* wide counter */

  interfaces_count = read_u2();
  interfaces       = NULL;

  if (interfaces_count != 0u) {
    interfaces = (u2_int *) make_mem(interfaces_count * sizeof(u2_int));
    for (i = 0u; i < interfaces_count; i++) 
      interfaces[i] = read_u2();
  }
}

/* Read Attributes
   *************** */

static attribute_ptr *read_attributes(u2_int ac) {
  if (ac != 0u) {
    u4_int i, j; /* wide counters */
    u4_int len;

    attribute_ptr *new = (attribute_ptr *) make_mem(ac * sizeof(attribute_ptr));

    for (i = 0u; i < ac; i++) {

      new[i] = (attribute_ptr) make_mem(sizeof(struct attribute_node));

      new[i] -> attribute_name_index =       read_u2();
      new[i] -> attribute_length     = len = read_u4();

      if (len == U4MAX)
        javas_out(-1, "Sorry, my internal u4_int counter will wrap around");

      new[i] -> info = (u1_int *) make_mem(len * sizeof(u1_int));

      for (j = 0u; j < len; j++)
        new[i] -> info[j] = read_u1();
    }

    return new;
  }
  return NULL;
}

/* Read Fields
   *********** */

static void read_fields(void) {
  u4_int i; /* wide counter */

  fields_count = read_u2();
  fields       = NULL;

  if (fields_count != 0u) {
    fields = (fm_ptr *) make_mem(fields_count * sizeof(fm_ptr));
    for (i = 0u; i < fields_count; i++) {
      fields[i] = (fm_ptr) make_mem(sizeof(struct fm_node));
      fields[i] -> access_flags     = read_u2();
      fields[i] -> name_index       = read_u2();
      fields[i] -> descr_index      = read_u2();
      fields[i] -> attributes_count = read_u2();
      fields[i] -> attributes       = read_attributes(fields[i] -> 
					  attributes_count);
    }
  }
}

/* Read Methods
   ************ */

static void read_methods(void) {
  u4_int i; /* wide counter */

  methods_count = read_u2();
  methods       = NULL;

  if (methods_count != 0u) {
    methods = (fm_ptr *) make_mem(methods_count * sizeof(fm_ptr));
    for (i = 0u; i < methods_count; i++) {
      methods[i] = (fm_ptr) make_mem(sizeof(struct fm_node));
      methods[i] -> access_flags     = read_u2();
      methods[i] -> name_index       = read_u2();
      methods[i] -> descr_index      = read_u2();
      methods[i] -> attributes_count = read_u2();
      methods[i] -> attributes       = read_attributes(methods[i] -> 
					     attributes_count);
    }
  }
}

/* Read Class-File
   *************** */

static void read_classfile(void) {

  if (! file)
    javas_out(-1, "incorrect invocation of read_classfile()");

  javas_out(2, " -- reading class file");

  /* Read magic 0xCAFEBABE string and version
     **************************************** */

  magic = read_u4();

  if (magic != 0xCAFEBABE) {
    javas_out(0, "not a class file");
    goto Einde;
  }

  minor_version = read_u2();
  major_version = read_u2();

  /* Read constant pool
     ****************** */

  read_constant_pool();

  /* Read flags and class info
     ************************* */

  access_flags = read_u2();
  this_class   = read_u2();
  super_class  = read_u2();

  /* Read interfaces, fields, and methods
     ************************************ */

  read_interfaces();
  read_fields();
  read_methods();

  /* Read attributes
     *************** */

  attributes_count = read_u2();
  attributes       = read_attributes(attributes_count);

  if (fgetc(file) != EOF)
    javas_out(1, "additional bytes in class file ignored");

Einde:
  fclose(file);
  file = NULL;
}

/* Check Field/Method
   ****************** */

static void check_fm(fm_ptr f) {
  valid_cp_entry(CONSTANT_Utf8, f -> name_index,  "fm name index");
  valid_cp_entry(CONSTANT_Utf8, f -> descr_index, "fm descriptor index");
}

/* Check Class File
   **************** */

static void check_classfile(void) {
  u4_int i; /* wide counter */

  javas_out(2, " -- verifying class file");

  /* Check class references */

  valid_cp_entry(CONSTANT_Class, this_class, "this");

  if (super_class)
    valid_cp_entry(CONSTANT_Class, super_class, "super");

  /* Check constant pool */

  for (i = 1u; i < constant_pool_count; i++) 
     if (constant_pool[i])  
       switch (constant_pool[i] -> tag) {
	 
	  case CONSTANT_Class:

	     valid_cp_entry(CONSTANT_Utf8, 
	              constant_pool[i] -> u.indices.index1,
	              "Class");
	     break;

	  case CONSTANT_Fieldref:
	  case CONSTANT_Methodref:
	  case CONSTANT_InterfaceMethodref:

	     valid_cp_entry(CONSTANT_Class,       
		      constant_pool[i] -> u.indices.index1,
	              "ref");
	     valid_cp_entry(CONSTANT_NameAndType, 
		      constant_pool[i] -> u.indices.index2,
	              "ref");
	     break;

	  case CONSTANT_String:

	     valid_cp_entry(CONSTANT_Utf8, 
	              constant_pool[i] -> u.indices.index1,
  	              "String");
	     break;

	  case CONSTANT_NameAndType:

	     valid_cp_entry(CONSTANT_Utf8, 
	              constant_pool[i] -> u.indices.index1,
	              "N_and_T");
	     valid_cp_entry(CONSTANT_Utf8, 
	              constant_pool[i] -> u.indices.index2,
	              "N_and_T");

             { constant_ptr c = 
		constant_pool[constant_pool[i] -> u.indices.index2];

               if (c -> u.utf8.l == 0)
		 javas_out(0, "invalid field/method descriptor");
             }
	     break;
    }

  /* Check interfaces */

  for (i = 0u; i < interfaces_count; i++)  
    valid_cp_entry(CONSTANT_Class, interfaces[i], "interface");

  /* Check Fields */

  for (i = 0u; i < fields_count; i++) 
    check_fm(fields[i]);

  /* Check Methods */

  for (i = 0u; i < methods_count; i++) 
    check_fm(methods[i]);
}

/* Release Memory of a Constant Pool Entry
   *************************************** */

static void del_cp(constant_ptr c) {
  if (c) {
    if ((c -> tag == CONSTANT_Utf8) && (c -> u.utf8.s))
      free(c -> u.utf8.s);
    free(c);
  }
}

/* Release Memory of Attributes
   **************************** */

static void del_attributes(attribute_ptr *atts, u2_int ac) {
  if (atts) {
    u4_int i;
    for (i = 0u; i < ac; i++) {
      attribute_ptr a = atts[i];
      if (a) {
        if (a -> info)
          free(a -> info);
        free(a);
      } 
    }
    free(atts);
  }
}

/* Release Memory of Class File
   **************************** */

static void delete_classfile(void) {

  u4_int i; /* wide counters */

  javas_out(2, " -- deleting class file");

  /* Delete Constant Pool
     ******************** */

  for (i = 1u; i < constant_pool_count; i++) 
    if (constant_pool[i])
      del_cp(constant_pool[i]);

  if (constant_pool)
    free(constant_pool);

  constant_pool_count = 0u;
  constant_pool       = NULL;

  /* Delete Interfaces
     ***************** */

  if (interfaces)
    free(interfaces);

  interfaces = NULL;

  /* Delete Fields
     ************* */

  if (fields) {
    for (i = 0u; i < fields_count; i++) 
      if (fields[i]) {

	del_attributes(fields[i] -> attributes, 
		       fields[i] -> attributes_count);

        free(fields[i]);
      }
    free(fields);
  }
  
  fields_count = 0u;
  fields       = NULL;

  /* Delete Methods
     ************** */

  if (methods) {
    for (i = 0u; i < methods_count; i++) 
      if (methods[i]) {

	del_attributes(methods[i] -> attributes, 
		       methods[i] -> attributes_count);

        free(methods[i]);
      }
    free(methods);
  }
  
  methods_count = 0u;
  methods       = NULL;

  /* Delete Attributes
     ***************** */

  del_attributes(attributes, attributes_count);

  attributes_count = 0u;
  attributes       = NULL;
}

/* Determine File Name of Class File
   ********************************* */

static FILE *find_file(u1_int *cname) {
  u4_int  j   = 0u, k, l1 = strlen((char *) cname), l;
  FILE   *res = NULL;
  char   *nm;

  /* Read CLASSPATH from Environment */

  if (! class_env) {
    char *t   = getenv("CLASSPATH");
    class_env = (t) ? (char *)strdup(t) : (char *)strdup(".");
  }

  l  = strlen(class_env);
  nm = make_mem(l1 + l + 10);

  /* Loop over all Prefixes */

  while (j < l) {
    k = j;

    while ((class_env[k] != ':') && (k < l)) k++;

    if ((j < k) && (k <= l)) {
      char save;

      if (k < l) {
        save =  class_env[k];
        class_env[k] = '\0';
      }

      if (class_env[k-1] == '/')
        sprintf(nm, "%s%s.class",  class_env+j, (char *) cname);
      else
        sprintf(nm, "%s/%s.class", class_env+j, (char *) cname);

      res = fopen(nm, "r");
    
      if (k < l)
        class_env[k] = save;

      if (res)
	goto Einde;
    }
    j = k+1;
  }

Einde:
  free(nm); 

  return res; 
}

/* Read Single Class File
   ************************ */

static void recursive_read(class_ptr);

static void read_in(u1_int *cname) {

  class_ptr scan = classes;

  if (! cname)
    javas_out(-1, "what do you want?");

  for ( ; scan; scan = scan -> next)
    if (strcmp((char *) scan -> classname, (char *) cname) == 0)
      break;

  if (! scan) { /* not previously read */
  
    file = find_file(cname);
    
    if (file) {

      classname = cname;

      read_classfile();
      check_classfile();
      move_in();

      recursive_read(classes);
    }
    else
      javas_out(0, "cannot find %s.class", cname);
  }
}

static void extract(u1_int *d) {
  if (d) {
    u4_int j  = 0u, k; /* wide counters */
    u2_int l2 = (u2_int) strlen((char *) d);

    while (j < l2) {
      if (d[j] == 'L') {
        k = j;
        while ((d[k] != ';') && (k < l2)) k++;
        if (k < l2) {
	  d[k] = '\0';
          read_in(d+j+1);
	  d[k] = ';'; /* restore */
          j = k;
        }
       else 
        javas_out(0, "illegal format for descriptor %s", d); 
      } 
      j++;
    }
  }
}

static void recursive_read(class_ptr c) {
  u4_int i; /* wide counters */

  u2_int        l = c -> constant_pool_count;
  constant_ptr *p = c -> constant_pool;
  fm_ptr       *f;

  for (i = 0u; i < l; i++) {
    if (p[i]) {
      switch(p[i] -> tag) {

	case CONSTANT_Class:

	   { u2_int  ind = p[i]   -> u.indices.index1;
	     u1_int *cnm = p[ind] -> u.utf8.s;

	     if (cnm) {
	       if (cnm[0] != '[')
	         read_in(cnm);
               else
	         extract(cnm);
             }
           }
	   break;

	case CONSTANT_NameAndType:
	   { u2_int ind = p[i] -> u.indices.index2;
	     extract(p[ind] -> u.utf8.s);
           }
	  break;
      }
    }
  }

  f = c -> fields;
  l = c -> fields_count;
  
  if (f)
    for (i = 0u; i < l; i++) {
      u2_int ind = f[i] -> descr_index;
      extract(p[ind] -> u.utf8.s);
    }

  f = c -> methods;
  l = c -> methods_count;

  if (f)
    for (i = 0u; i < l; i++) {
      u2_int ind = f[i] -> descr_index;
      extract(p[ind] -> u.utf8.s);
    }
}

/* ********************************************************
   *** PUBLIC FUNCTIONS                                 ***
   ******************************************************** */

/* Read Class Files
   **************** */

short read_classes(char *name) {

  error = 0;

  if (name) {

    int i, l = strlen(name);

    for (i = 0; i < l; i++) 
      if (name[i] == '.')
	name[i] = '/';

    read_in((u1_int *) name);

    if (error)
      delete_classes();
  }
  else
    error = 1;

  return error;
}

/* Delete Class Files
   ****************** */

void delete_classes(void) {

  ftrav_class = NULL;
  mtrav_class = NULL;
  itrav_class = NULL;
  
  while (classes) {
    class_ptr next = classes -> next;
  
    move_out();

    javas_out(2, "delete %s", classes -> classname);

    delete_classfile();

    if (classes -> classname)
      free(classes -> classname);

    free(classes); 

    classname = NULL;
    classes   = next;
  }
}

/* Auxiliary Search Function
   ************************* */

static class_ptr find_class(char *class) {
  class_ptr s = classes;
  for (; s; s = s -> next)
    if (strcmp((char *) s -> classname, class) == 0)
      return s;
  return NULL;
}

/* Access Functions 
   **************** */

char *get_superclass(char *class) {
  class_ptr s = find_class(class);
  if ((s) && (s -> super_class)) {
    u2_int ind = s -> constant_pool[s -> super_class] -> u.indices.index1;
    return (char *) s -> constant_pool[ind] -> u.utf8.s;
  }
  return NULL;
}

char *get_name(char *class, fm_ptr f) {
  class_ptr s = find_class(class);
  if (s) 
    return (char *) s -> constant_pool[f -> name_index] -> u.utf8.s;
  return NULL;
}

char *get_typedescriptor(char *class, fm_ptr f) {
  class_ptr s = find_class(class);
  if (s)
    return (char *) s -> constant_pool[f -> descr_index] -> u.utf8.s;
  return NULL;
}

/* Test Functions
   ************** */

short implements(char *class, char *interface) {
  class_ptr s = find_class(class);
  if (s) {
    u4_int        i;   
    u2_int        cnt  = s -> interfaces_count;
    u2_int       *intf = s -> interfaces;
    constant_ptr *pool = s -> constant_pool;


    /* Test Interfaces */

    for (i = 0u; i < cnt; i++)  {
       u2_int  ind = pool[intf[i]] -> u.indices.index1;
       u1_int *inm = pool[ind    ] -> u.utf8.s;


       if ( (strcmp(interface, (char *) inm) == 0) ||
	    (implements((char *) inm, interface)) ) 
         return 1;
    }

     /* Test Superclass */

     if (s -> super_class != 0) {
       u2_int  supi = pool[s -> super_class] -> u.indices.index1;
       u1_int *sup  = pool[supi]              -> u.utf8.s;

       if ((sup) && (implements((char *) sup, interface))) 
         return 1;
    }
  }
  return 0;
}

short contains_imethod(char *class, char *nm1, char *td1)  {
  class_ptr s = find_class(class);
  if (s) {
    u4_int        i;   
    u2_int        cnt  = s -> methods_count;
    fm_ptr       *mets = s -> methods;
    constant_ptr *pool = s -> constant_pool;


    /* Test your own methods */

    for (i = 0u; i < cnt; i++)  {
      fm_ptr f   = mets[i];
      char  *nm2 = (char *) pool[f -> name_index]  -> u.utf8.s;
      char  *td2 = (char *) pool[f -> descr_index] -> u.utf8.s;

      if ((strcmp(nm1, nm2) == 0) &&
	  (strcmp(td1, td2) == 0))
        return 1;
    }

     /* Test inherited methods */
    for(i = 0u; i < s->interfaces_count; i++)
    {
     if (s -> interfaces[i] != 0) {
       u2_int  supi = pool[s -> interfaces[i]] -> u.indices.index1;
       u1_int *sup  = pool[supi]               -> u.utf8.s;

       if ((sup) && (contains_imethod((char *) sup, nm1, td1))) 
         return 1;
     }
    }
  }
  return 0;
}

short contains_method(char *class, char *nm1, char *td1)  {
  class_ptr s = find_class(class);
  if (s) {
    u4_int        i;   
    u2_int        cnt  = s -> methods_count;
    fm_ptr       *mets = s -> methods;
    constant_ptr *pool = s -> constant_pool;


    /* Test your own methods */

    for (i = 0u; i < cnt; i++)  {
      fm_ptr f   = mets[i];
      char  *nm2 = (char *) pool[f -> name_index]  -> u.utf8.s;
      char  *td2 = (char *) pool[f -> descr_index] -> u.utf8.s;

      if ((strcmp(nm1, nm2) == 0) &&
	  (strcmp(td1, td2) == 0))
        return 1;
    }

     /* Test inherited methods */

     if (s -> super_class != 0) {
       u2_int  supi = pool[s -> super_class] -> u.indices.index1;
       u1_int *sup  = pool[supi]              -> u.utf8.s;

       if ((sup) && (contains_method((char *) sup, nm1, td1))) 
         return 1;
    }
  }
  return 0;
}

/* Traversal Functions 
   ******************* */

void trav_class_init(char *class, short fi, short mi, short ii) {
  if (fi)
  {
    ftrav_class = find_class(class);
    trav_fc    = 0u;
  }
  if (mi)
  {
    mtrav_class = find_class(class);
    trav_mc    = 0u;
  }
  if (ii)
  {
    itrav_class = find_class(class);
    trav_ic    = 0u;
  }
}

fm_ptr get_next_field(void) {
  if ((ftrav_class) && (trav_fc < ftrav_class -> fields_count))
    return ftrav_class -> fields[trav_fc++];
  return NULL;
}

fm_ptr get_next_method(void) {

  trav_ex  = 0u;
  trav_exc = NULL;

  if ((mtrav_class) && (trav_mc < mtrav_class -> methods_count)) {

    fm_ptr m = mtrav_class -> methods[trav_mc++];

    /* Reset Exception Traversal Information */

    attribute_ptr *atts = m -> attributes;
    u2_int         ac   = m -> attributes_count;
    u4_int         i;

    if (atts)
     for (i = 0u; i < ac; i++) {
      constant_ptr ua = mtrav_class -> constant_pool[atts[i] -> attribute_name_index];
      if ((ua) && (strcmp((char *) ua -> u.utf8.s, "Exceptions") == 0))
	trav_exc = atts[i];
     }
     else javas_out(10, "no  atts\n");

    return m;
  }
  return NULL;
}

char *get_next_interface(void) {
  if ((itrav_class) && (trav_ic < itrav_class -> interfaces_count)) {
    u2_int ind1 =   itrav_class -> interfaces[trav_ic++];
    u2_int ind2 =   itrav_class -> constant_pool[ind1] -> u.indices.index1;
    return (char *) itrav_class -> constant_pool[ind2] -> u.utf8.s;
  }
  return NULL;
}

char *get_next_exception(void) {
  if ((mtrav_class) && (trav_exc)) {
    u1_int *locb       = trav_exc -> info;
    u2_int  num_of_exc = B2U2(locb[0u], locb[1u]);
    u2_int  j          = trav_ex++;

    if (j < num_of_exc) {
      u2_int ind1 = B2U2(locb[2u+2u*j],locb[3u+2u*j]);
      u2_int ind2 =   mtrav_class -> constant_pool[ind1] -> u.indices.index1;
      return (char *) mtrav_class -> constant_pool[ind2] -> u.utf8.s;
    }
  }
  return NULL;
}

short is_final_class(char *class) {
  class_ptr s = find_class(class);
  if (s) 
    return ((s->access_flags & ACC_FINAL) && !(s->access_flags & ACC_INTERFACE));
  fprintf(stderr, "is_final: class %s not found\n", class);
  return 0;
}

short is_static_method(fm_ptr m) {
  return (m->access_flags & ACC_STATIC);
}

#ifdef STAND_ALONE
short verbose = 0;

main(int argc, char *argv[]) {
  char     *cn = argv[1];
  fm_ptr    f;
  char     *c;

  read_classes(cn);

  trav_class_init(cn, 1, 1, 1);

  printf("*** %s ***\n", cn);

  while (f = get_next_field()) 
    printf("field %s, %s\n", get_name(cn, f), get_typedescriptor(cn, f));

  while (f = get_next_method()) {

    printf("method %s, %s\n", get_name(cn, f), get_typedescriptor(cn, f));
    printf("contains SELF = %i\n",
	contains_method(cn, get_name(cn,f), get_typedescriptor(cn,f)));
    printf("contains joho()V = %i\n",
	contains_method(cn, 
	"joho" , "()V" ));

    while (c = get_next_exception()) 
      printf(" -- throws %s\n", c);
  }

  while (c = get_next_interface()) 
    printf("interface %s\n", c);

}
#endif

