

/* *************
   *** JAVAR ***
   ****************************************************
   *** Copyright (c) 1996/1997                      ***
   *** Aart J.C. Bik             Indiana University *** 
   *** All Rights Reserved                          ***
   ****************************************************
   *** Please refer to the LICENSE file distributed ***
   *** with this software for further details on    ***
   *** the licensing terms and conditions.          ***
   ***                                              ***
   *** Please, report all bugs, comments, etc.      ***
   *** to: ajcbik@extreme.indiana.edu               ***
   ****************************************************
   *** semantic.c : semantic analysis 
   ***
   ***
   *** Your courtesy in mentioning the use of this compiler 
   *** in any scientific work that presents results obtained
   *** by using (extensions or modifications of) this compiler
   *** is highly appreciated.
   ***
   *** NOTE! 
   *** This implementation does NOT provide a complete Java 
   *** front-end (unicode escapes are not supported
   *** and the semantic analysis is rather limited).
   ***
   *** */

/* This file has been modified from the original to support
 * Java 1.1 language.
 *
 *     Fabian Breg.
 */

/* ********************************************************
   *** INCLUDE FILES and DEFINITIONS                    ***
   ********************************************************  */

#include "global.h"
#include "y.tab.h"

/* ********************************************************
   *** EXTERNAL VARIABLES                               ***
   ******************************************************** */

/* PRIVATE
   ******* */

static int default_cnt;

/* ********************************************************
   *** PRIVATE FUNCTIONS                                ***
   ******************************************************** */

/* Test Validity of Type Conversion:
   note that here and throughout this module, rather 
   liberal tests are used because:
    1) JAVAR cannot always determine the type
       of (sub)expressions (unknown types are NULL)
    2) JAVAR has no notion of the class hierarchy
   ************************************************* */

static short is_valid(type_ptr t1, type_ptr t2) {

  if ((t1) && (t1 -> kind == ARRAY_T) &&
      (t2) && (t2 -> kind == ARRAY_T) )
    return is_valid(t1 -> u.aref, t2 -> u.aref);
  else 
   return
   ( ((IS_BOOL(t1)) && (IS_BOOL(t2))) ||
     (( IS_NUM(t1)) && ( IS_NUM(t2))) ||
     (( IS_REF(t1)) && ( IS_REF(t2))) );
}

/* Unary Type Promotion (on numeric types)
   *************************************** */

static type_ptr un_prom(type_ptr t) {
  if ((t) && (IS_INTEGRAL1(t)))           /* CERTAINLY byte/short/char/int */
    return make_primitive_type(INT_T);
  else
    return t;
}

/* Binary Type Promotion (on numeric types)
   **************************************** */

static type_ptr bin_prom(type_ptr t1, type_ptr t2) {
  if ((t1) && (t2)) {
  
    /* Both types are CERTAIN */

    if ((t1 -> kind == DOUBLE_T) || (t2 -> kind == DOUBLE_T))
      return make_primitive_type(DOUBLE_T);
    else if ((t1 -> kind == FLOAT_T) || (t2 -> kind == FLOAT_T)) 
      return make_primitive_type(FLOAT_T);
    else if ((t1 -> kind == LONG_T) || (t2 -> kind == LONG_T)) 
      return make_primitive_type(LONG_T);
    else 
      return make_primitive_type(INT_T);
    }
  else 
    return NULL;   /* remains unkown */
}

/* Traverse at ONE level to count defauls-labels of switch
   ******************************************************* */

static void check_default_labs(stmt_ptr s) {
  if (s) {
    if (s -> kind == CONCAT_S) {
      check_default_labs(s -> u.concat.s1); 
      check_default_labs(s -> u.concat.s2); 
    }
    else if (s -> kind == DEF_S) {
      if (default_cnt++)
	javar_out(0, "invalid multiple default labels");
    }
  }
}

/* Handle Declaration of Fields, 
   Methods, and Constructors
  (FIRST PASS TRAVERSAL ROUTINE) 
  ******************************** */

static void decl_s(stmt_ptr s) {

  if (s -> kind == LAB_S) {

    if (activate_lab(s))
      javar_out(0, "ambigous use of statement label %s", 
		   id_table(s -> u.lab.label -> u.name.entry));

  }
  else if (s -> kind == DECL_S) {

    /* Check Declaration of Fields */

    expr_ptr nm = s -> u.decl.name;
    if ((nm -> u.name.info == FIELD_N) && (set_info_table(nm, s)))
        javar_out(0, "invalid re-declaratation of field %s", 
		    id_table(nm -> u.name.entry));
  }
  else if ( (s -> kind == CLASS_S) || (s -> kind == INTERF_S)  ||
	    (s -> kind == CONS_S)  || (s -> kind == METHOD_S) )   { 

    /* Check Declaration of Methods and Constructors */

    expr_ptr nm = s -> u.header.name;
    if (set_info_table(nm, s))
        javar_out(0, "invalid re-declaratation of %s %s", 
        ((s -> kind == CLASS_S)  || (s -> kind == INTERF_S)) 
	? "class/interface" : "method/constructor", id_table(nm -> u.name.entry));

    if (s -> kind == CONS_S) {
      expr_ptr classname = cur_class -> u.header.name;

      if ((! classname) || (nm -> u.name.entry != classname -> u.name.entry))
        javar_out(0, "constructor name does not match class name");
    }
    else if ((s -> kind == METHOD_S) && (s -> u.header.mod & ABS_B)) {
        if (! (cur_class -> u.header.mod & ABS_B))
  	  javar_out(0, "non-abstract class contains abstract method %s", 
                      id_table(nm -> u.name.entry));
    }
  }
}

/* Semantically Validate Expression 
   (SECOND PASS TRAVERSAL ROUTINE)
   ********************************** */

static void valid_e(expr_ptr e) {

  switch (e -> kind) {

    /* LITERALS 
       (TYPE has already been determined during construction)
       ****************************************************** */

    case LITERAL_:
      break;

    /* Names
       ***** */

    case ID_:
      classify(e, e -> u.name.info);
      break;

    /* TE Operators 
       (TYPE has already been determined during construction)
       ****************************************************** */

    case INSTOF_:

       if ((! (IS_REF(e -> u.te.tp))) || (! (IS_REF(e -> u.te.ex -> type))))
	 javar_out(0, "invalid operand for instanceof expression");
       break;

    case 'c': /* CAST */

       if (! is_valid(e -> u.te.tp, e -> u.te.ex -> type))
	 javar_out(0, "invalid cast expression");
       break;

    /* List-like Operators
       ******************* */

    case 'i':

	e -> type = e -> u.op.lhs -> type;
	break;

    case '[':
	 
       if ((e -> u.op.rhs) && (! (IS_INTEGRAL1(e -> u.op.rhs -> type))))
	 javar_out(0, "invalid subscript expression");

       if (e -> u.op.lhs) {
         type_ptr tl =  e -> u.op.lhs -> type;

         if (tl) { /* Known Type */

	   if (tl -> kind == ARRAY_T)
	     e -> type = tl -> u.aref; 
	   else
	     javar_out(0, "invalid array reference expression");
         }
       }
       break;

    case '?':

        if (IS_BOOL(e -> u.op.lhs -> type))
          e -> type = e -> u.op.rhs -> type;
        else
          javar_out(0, "invalid conditional expression for conditional operator");
        break;

    case ':':

        if (is_valid(e -> u.op.lhs -> type, e -> u.op.rhs -> type)) {
          if (equal_type(e -> u.op.lhs -> type, e -> u.op.rhs -> type))
	    e -> type = e -> u.op.lhs -> type;
	}
        else
	  javar_out(0, "invalid operand for conditional operator");
        break;

    case '.':

       { expr_ptr n =  e -> u.op.rhs, args = NULL;

	 if ((n) && (n -> kind == ID_)) {
	   if (n -> u.name.info == MISCM_N) {
             if ((cur_inv) && (cur_inv -> u.op.lhs == e))
               args = cur_inv -> u.op.rhs;
             else
               javar_out(-1, "lost method ID context");
           }
         }
         else if((n) && (n->kind == NEW_)) {
	 }
	 else
            javar_out(-1, "AST of '.' corrupt");

         /* the ID in an expression (Q).id() or (Q).id */

         det_meaning(e, n, (n -> u.name.info == MISCM_N) ? METH_N : FIELD_N, args);
	
 	 e -> type = e -> u.op.rhs -> type;
       }
       break;

    /* Ordinary Operators
       ****************** */
  
    case THIS_: /* (TYPE : class) */
    case SUPER_:

      if ( ((cur_method) && (cur_method -> u.header.mod & STA_B)) ||
           ((cur_decl)   && (cur_decl   -> u.decl.mod   & STA_B)) )
        javar_out(0, "cannot make static reference to this/super");
      else if (cur_class) {
	 if (e -> kind == THIS_)
	   e -> type = make_c_i_type(cur_class -> u.header.name);
      }
      else
	javar_out(-1, "AST corrupt, strange occurrence of this/super");
      break;

    case PP_:
    case MM_:

      { expr_ptr op = (e -> u.op.lhs) ? e -> u.op.lhs : e -> u.op.rhs;

	if (IS_NUM(op -> type))
          e -> type = op -> type;
	else
	  javar_out(0, "invalid operand for %fix operator",
			(e -> u.op.lhs) ? "post" : "pre");
      }
      break;

    case '+': 

      /* Deal with Possible String concatenation in Binary + first */

      if ((e -> u.op.lhs) && ((IS_STR(e -> u.op.lhs -> type)) || 
  			      (IS_STR(e -> u.op.rhs -> type)))) {
         e -> type = ( ((e -> u.op.lhs -> type) && (IS_STR(e -> u.op.lhs -> type))) || 
	               ((e -> u.op.rhs -> type) && (IS_STR(e -> u.op.rhs -> type))) )
		   ? make_c_i_type(MK_STRING)  /* one is CERTAINLY String  */
		   : NULL;                     /* remains unknown */
         break;
      } 

    case '-':  /* MAY FALL THROUGH */

      if (e -> u.op.lhs) { /* Binary */

        if ((IS_NUM(e -> u.op.lhs -> type)) && (IS_NUM(e -> u.op.rhs -> type)))
          e -> type = bin_prom(e -> u.op.lhs -> type, e -> u.op.rhs -> type);
	else
	  javar_out(0, "invalid operand for additive operator");
      }
      else { /* Unary */

	if (IS_NUM(e -> u.op.rhs -> type))
          e -> type = un_prom(e -> u.op.rhs -> type);
	else
	  javar_out(0, "invalid operand for unary additive operator");
      }
      break;

    case '~': /* bitwise complement operator */

      if (IS_INTEGRAL2(e -> u.op.rhs -> type)) 
        e -> type = un_prom(e -> u.op.rhs -> type);
      else
        javar_out(0, "invalid operand for bitwise complement operator");
      break;

    case '!': /* locigal complement operator */

      if (IS_BOOL(e -> u.op.rhs -> type)) 
        e -> type = make_primitive_type(BOOL_T);
      else
        javar_out(0, "invalid operand for logical complement operator");
      break;

    case '*':  /* multiplicative operators */
    case '/':
    case '%':  

      if ((IS_NUM(e -> u.op.lhs -> type)) && (IS_NUM(e -> u.op.rhs -> type)))
        e -> type = bin_prom(e -> u.op.lhs -> type, e -> u.op.rhs -> type);
      else
        javar_out(0, "invalid operand for multiplicative operator");
      break;

    case LL_:  /* shift operators */
    case RR_:
    case RRR_:

      if ((IS_INTEGRAL2(e -> u.op.lhs -> type)) && (IS_INTEGRAL2(e -> u.op.rhs -> type)))
	/* Only use left-hand-side operand */
        e -> type = un_prom(e -> u.op.lhs -> type);
      else
        javar_out(0, "invalid operand for shift operator");
      break;

    case '<':  /* relational operators */
    case '>':
    case LE_:
    case GE_:

      if ((IS_NUM(e -> u.op.lhs -> type)) && (IS_NUM(e -> u.op.rhs -> type)))
        e -> type = make_primitive_type(BOOL_T);
      else
        javar_out(0, "invalid operand for relational operator");
      break;

    case EQ_:  /* equality operators */
    case NE_:

      if (is_valid(e -> u.op.lhs -> type, e -> u.op.rhs -> type))  
        e -> type = make_primitive_type(BOOL_T);
      else
        javar_out(0, "invalid operand for equality operator");
      break;

    case '&':  /* bitwise and logical operators */
    case '^':
    case '|':
	 
      if ((IS_BOOL(e -> u.op.lhs -> type)) && (IS_BOOL(e -> u.op.rhs -> type))) 
        e -> type = ((e -> u.op.lhs -> type) || (e -> u.op.rhs -> type))
		  ? make_primitive_type(BOOL_T)  /* one is CERTAIN  */
		  : NULL;                        /* remains unknown */
      else if ((IS_NUM(e -> u.op.lhs -> type)) && (IS_NUM(e -> u.op.rhs -> type)))
        e -> type = bin_prom(e -> u.op.lhs -> type, e -> u.op.rhs -> type);
      else
       javar_out(0, "invalid operand for operator %c", e -> kind);
      break;

    case AND_: /* conditional operators */
    case OR_:

      if ((IS_BOOL(e -> u.op.lhs -> type)) && (IS_BOOL(e -> u.op.rhs -> type))) 
        e -> type = make_primitive_type(BOOL_T);
      else
        javar_out(0, "invalid operand for conditional operator");
      break;

    case '=':

      if (is_valid(e -> u.op.lhs -> type, e -> u.op.rhs -> type))  
        e -> type = e -> u.op.lhs -> type;
      else
        javar_out(0, "invalid operand for assignment operator");
      break;

    case PIS_:

      if (IS_STR(e -> u.op.lhs -> type)) {
        e -> type = e -> u.op.lhs -> type;
        break;
      }

    default:  /* MAY FALL THROUGH */

      if ((e -> kind >= PIS_) && (e -> kind <= RRRIS_)) {

        if ( (is_valid(e -> u.op.lhs -> type, e -> u.op.rhs -> type))              &&  
	     ((IS_NUM(e -> u.op.lhs -> type)) || (IS_BOOL(e -> u.op.lhs -> type))) &&
	     ((IS_NUM(e -> u.op.rhs -> type)) || (IS_BOOL(e -> u.op.rhs -> type))) )
          e -> type = e -> u.op.lhs -> type;
        else
          javar_out(0, "invalid operand for compound assignment operator");
      }
      break;
  }
}

/* Semantically Validate Statement 
   (SECOND PASS TRAVERSAL ROUTINE)
   ********************************* */

static void valid_s(stmt_ptr s) {

  switch (s -> kind) {

     /* Simple Statements
	***************** */

     case BRK_S:
      check_transfer(1); 
      break;

     case CONT_S:
      check_transfer(0);
      check_target(s -> u.arg);
      break;

     case RET_S:
       if (cur_method) {
         type_ptr tp = cur_method -> u.header.tp;

          if (! cur_method -> u.header.name)
            javar_out(0, "invalid return in static initializer");
          else if ((! s -> u.arg) && (tp) && (tp -> kind != VOID_T))
            javar_out(0, "missing value in return statement");
          else if (s -> u.arg) {
	  
	     if (! tp) 
               javar_out(0, "invalid return of value in constructor");
             else if (tp -> kind == VOID_T) 
               javar_out(0, "invalid return of value in void-method");
             else if (! (is_valid(tp, s -> u.arg -> type)))
               javar_out(0, "invalid return expression");
          }
       }
       break;

     case CASE_S:

     { expr_ptr e = s -> u.arg;

        if ((! e) || (! (IS_INTEGRAL1(e -> type))))
          javar_out(0, "invalid case expression");
     }
     break;

     /* Control Statements
	****************** */

     case FOR_S:

     {  expr_ptr e = s -> u.arg -> u.op.rhs -> u.op.lhs;

        if ((e) && (! (IS_BOOL(e -> type))))
           javar_out(0, "invalid control expression in for statement");
     }
     break;

     case IF_S:
     case WHILE_S:
     case DOWHILE_S:

       if ((! s -> u.arg) || (! (IS_BOOL(s -> u.arg -> type))))
         javar_out(0, "invalid control expression");
       break;

     case SWITCH_S:
     
     { expr_ptr e = s -> u.control.e;

        if ((! e) || (! (IS_INTEGRAL1(e -> type))))
          javar_out(0, "invalid switch expression");

	default_cnt = 0;

	if (s -> u.control.s1) 
          check_default_labs(s -> u.control.s1 -> u.header.body);
     }
     break;

     case SYNC_S:

     { expr_ptr e = s -> u.arg;

       if ((! e) || (! (IS_REF(e -> type))))
          javar_out(0, "invalid expression in synchronized statement");
     }
     break;

     /* Declaratation Statements
	************************ */

     case DECL_S: 
     
     { expr_ptr nm = s -> u.decl.name;

       /* Check Local/Parameter Declarations */

        if ((nm -> u.name.info == LOCPAR_N) && (set_info_table(nm, s)))
            javar_out(0, "invalid re-declaratation of local/parameter %s", 
                          id_table(nm -> u.name.entry));
    
       /* Check Final Field for initializer */

       if ((s -> u.decl.mod & FIN_B) && (! s -> u.decl.init))
         javar_out(2, "no initializer given for final field %s",         /* JAVA 1.1 */
		    id_table(nm -> u.name.entry));

       /* Type Test for initializer */
      
      if ((s -> u.decl.init) && 
	  (! (is_valid(s -> u.decl.tp, s -> u.decl.init -> type))))
        javar_out(0, "invalid initializer");

     } 
     break;
  }
}

/* ********************************************************
   *** PUBLIC FUNCTIONS                                 ***
   ******************************************************** */

void semantic_validate(void) {
  
  /* Two-Pass Semantic Checking */

  stmt_traverse(comp_unit, decl_s,  NULL);
  stmt_traverse(comp_unit, valid_s, valid_e);
}
